Package: lavaSearch2
Type: Package
Title: Tools for Model Specification in the Latent Variable Framework
Version: 2.0.3
Date: 2024-02-22
Authors@R: c(
    person("Brice", "Ozenne", role = c("aut", "cre"), email = "brice.mh.ozenne@gmail.com", comment = c(ORCID = "0000-0001-9694-2956"))
	)
URL: https://github.com/bozenne/lavaSearch2
BugReports: https://github.com/bozenne/lavaSearch2/issues
Description: Tools for model specification in the latent variable framework
    (add-on to the 'lava' package). The package contains three main functionalities:
    Wald tests/F-tests with improved control of the type 1 error in small samples,
    adjustment for multiple comparisons when searching for local dependencies,
    and adjustment for multiple comparisons when doing inference for multiple latent variable models. 
License: GPL-3
Encoding: UTF-8
Depends: R (>= 2.10), ggplot2, lava (>= 1.6.4)
Imports: abind, doParallel, MASS, Matrix, methods, multcomp, mvtnorm,
        nlme, parallel, Rcpp, reshape2, sandwich, stats, utils
Suggests: clubSandwich, data.table, foreach, emmeans, lme4, lmerTest,
        numDeriv, pbapply, pbkrtest, R.rsp, riskRegression, survival,
        testthat
LinkingTo: Rcpp, RcppArmadillo
VignetteBuilder: R.rsp
NeedsCompilation: yes
RoxygenNote: 7.2.3
Packaged: 2024-02-22 11:18:47 UTC; hpl802
Author: Brice Ozenne [aut, cre] (<https://orcid.org/0000-0001-9694-2956>)
Maintainer: Brice Ozenne <brice.mh.ozenne@gmail.com>
Repository: CRAN
Date/Publication: 2024-02-23 09:10:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-28 05:54:20 UTC; unix
Archs: lavaSearch2.so.dSYM

Package: lavaSearch2
Type: Package
Title: Tools for Model Specification in the Latent Variable Framework
Version: 2.0.3
Date: 2024-02-22
Authors@R: c(
    person("Brice", "Ozenne", role = c("aut", "cre"), email = "brice.mh.ozenne@gmail.com", comment = c(ORCID = "0000-0001-9694-2956"))
	)
URL: https://github.com/bozenne/lavaSearch2
BugReports: https://github.com/bozenne/lavaSearch2/issues
Description: Tools for model specification in the latent variable framework
    (add-on to the 'lava' package). The package contains three main functionalities:
    Wald tests/F-tests with improved control of the type 1 error in small samples,
    adjustment for multiple comparisons when searching for local dependencies,
    and adjustment for multiple comparisons when doing inference for multiple latent variable models. 
License: GPL-3
Encoding: UTF-8
Depends: R (>= 2.10), ggplot2, lava (>= 1.6.4)
Imports: abind, doParallel, MASS, Matrix, methods, multcomp, mvtnorm,
        nlme, parallel, Rcpp, reshape2, sandwich, stats, utils
Suggests: clubSandwich, data.table, foreach, emmeans, lme4, lmerTest,
        numDeriv, pbapply, pbkrtest, R.rsp, riskRegression, survival,
        testthat
LinkingTo: Rcpp, RcppArmadillo
VignetteBuilder: R.rsp
NeedsCompilation: yes
RoxygenNote: 7.2.3
Packaged: 2024-02-22 11:18:47 UTC; hpl802
Author: Brice Ozenne [aut, cre] (<https://orcid.org/0000-0001-9694-2956>)
Maintainer: Brice Ozenne <brice.mh.ozenne@gmail.com>
Repository: CRAN
Date/Publication: 2024-02-23 09:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 08:25:22 UTC; unix
Archs: lavaSearch2.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("sCorrect<-",default)
S3method(addLink,lvm)
S3method(addLink,lvm.reduced)
S3method(autoplot,calibrateType1)
S3method(autoplot,intDensTri)
S3method(autoplot,modelsearch2)
S3method(calibrateType1,lvm)
S3method(calibrateType1,lvmfit)
S3method(checkData,lvm)
S3method(clean,lvm)
S3method(coef,lvmfit2)
S3method(coef2,lvmfit)
S3method(coef2,lvmfit2)
S3method(coefCov,lvm)
S3method(coefCov,lvmfit)
S3method(coefCov,multigroup)
S3method(coefExtra,lvm)
S3method(coefExtra,lvmfit)
S3method(coefExtra,multigroup)
S3method(coefIndexModel,lvm)
S3method(coefIndexModel,lvmfit)
S3method(coefIndexModel,multigroup)
S3method(coefIndexModel,multigroupfit)
S3method(coefIntercept,lvm)
S3method(coefIntercept,lvmfit)
S3method(coefIntercept,multigroup)
S3method(coefRef,lvmfit)
S3method(coefReg,lvm)
S3method(coefReg,lvmfit)
S3method(coefReg,multigroup)
S3method(coefType,lvm)
S3method(coefType,lvmfit)
S3method(coefType,multigroup)
S3method(coefVar,lvm)
S3method(coefVar,lvmfit)
S3method(coefVar,multigroup)
S3method(compare,lvmfit2)
S3method(compare2,lvmfit)
S3method(compare2,lvmfit2)
S3method(confint,lvmfit2)
S3method(confint2,lvmfit)
S3method(confint2,lvmfit2)
S3method(createContrast,character)
S3method(createContrast,list)
S3method(createContrast,lvmfit)
S3method(createContrast,lvmfit2)
S3method(createContrast,mmm)
S3method(effects,lvmfit2)
S3method(effects2,lvmfit)
S3method(effects2,lvmfit2)
S3method(estimate2,list)
S3method(estimate2,lvm)
S3method(estimate2,lvmfit)
S3method(estimate2,mmm)
S3method(extractData,lvmfit)
S3method(findNewLink,lvm)
S3method(getNewLink,modelsearch2)
S3method(getNewModel,modelsearch2)
S3method(getStep,modelsearch2)
S3method(getVarCov2,lvmfit)
S3method(getVarCov2,lvmfit2)
S3method(glht,lvmfit2)
S3method(glht2,lvmfit)
S3method(glht2,lvmfit2)
S3method(glht2,mmm)
S3method(hessian2,lvmfit)
S3method(hessian2,lvmfit2)
S3method(iid,lvmfit2)
S3method(iid2,lvmfit)
S3method(iid2,lvmfit2)
S3method(iidJack,default)
S3method(information,lvmfit2)
S3method(information2,lvmfit)
S3method(information2,lvmfit2)
S3method(leverage2,lvmfit)
S3method(leverage2,lvmfit2)
S3method(model.tables,lvmfit)
S3method(model.tables,lvmfit2)
S3method(model.tables2,lvmfit)
S3method(model.tables2,lvmfit2)
S3method(modelsearch2,lvmfit)
S3method(moments2,lvm)
S3method(moments2,lvmfit)
S3method(nStep,modelsearch2)
S3method(nobs2,lvmfit)
S3method(nobs2,lvmfit2)
S3method(print,intDensTri)
S3method(print,modelsearch2)
S3method(print,summary.glht2)
S3method(residuals,lvmfit2)
S3method(residuals2,lvmfit)
S3method(residuals2,lvmfit2)
S3method(sCorrect,default)
S3method(score,lvmfit2)
S3method(score2,lvmfit)
S3method(score2,lvmfit2)
S3method(setLink,lvm)
S3method(summary,calibrateType1)
S3method(summary,glht2)
S3method(summary,lvmfit2)
S3method(summary,modelsearch2)
S3method(summary2,lvmfit)
S3method(summary2,lvmfit2)
S3method(var2dummy,list)
S3method(var2dummy,lvm)
S3method(vcov,lvmfit2)
S3method(vcov2,lvmfit)
S3method(vcov2,lvmfit2)
export("sCorrect<-")
export(addLink)
export(calcDistMaxBootstrap)
export(calcDistMaxIntegral)
export(calcType1postSelection)
export(calibrateType1)
export(checkData)
export(clean)
export(coef2)
export(coefCov)
export(coefExtra)
export(coefIndexModel)
export(coefIntercept)
export(coefRef)
export(coefReg)
export(coefType)
export(coefVar)
export(combineFormula)
export(compare2)
export(confint2)
export(createContrast)
export(effects2)
export(estimate2)
export(extractData)
export(findNewLink)
export(formula2character)
export(gaussian_weight.estimate.hook)
export(gaussian_weight_gradient.lvm)
export(gaussian_weight_hessian.lvm)
export(gaussian_weight_logLik.lvm)
export(gaussian_weight_method.lvm)
export(gaussian_weight_objective.lvm)
export(gaussian_weight_score.lvm)
export(getNewLink)
export(getNewModel)
export(getStep)
export(getVarCov2)
export(glht2)
export(hessian2)
export(iid2)
export(iid2plot)
export(iidJack)
export(information2)
export(initVarLink)
export(initVarLinks)
export(intDensTri)
export(leverage2)
export(matrixPower)
export(model.tables2)
export(modelsearch2)
export(moments2)
export(nStep)
export(nobs2)
export(residuals2)
export(sCorrect)
export(sampleRepeated)
export(score2)
export(setLink)
export(skeleton)
export(summary2)
export(transformSummaryTable)
export(tryWithWarnings)
export(vcov2)
import(Rcpp)
import(lava)
importFrom(MASS,mvrnorm)
importFrom(Matrix,bdiag)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,autoplot)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,text)
importFrom(methods,is)
importFrom(multcomp,glht)
importFrom(mvtnorm,pmvnorm)
importFrom(mvtnorm,pmvt)
importFrom(mvtnorm,qmvnorm)
importFrom(mvtnorm,qmvt)
importFrom(mvtnorm,rmvnorm)
importFrom(parallel,detectCores)
importFrom(parallel,makeCluster)
importFrom(parallel,stopCluster)
importFrom(reshape2,melt)
importFrom(sandwich,estfun)
importFrom(stats,anova)
importFrom(stats,coef)
importFrom(stats,confint)
importFrom(stats,cov)
importFrom(stats,effects)
importFrom(stats,formula)
importFrom(stats,logLik)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.tables)
importFrom(stats,predict)
importFrom(stats,qqnorm)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,update)
importFrom(stats,vcov)
importFrom(utils,methods)
importFrom(utils,packageVersion)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
useDynLib(lavaSearch2, .registration=TRUE)

### summary2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 10 2017 (10:57) 
## Version: 
## Last-Updated: feb 18 2019 (13:36) 
##           By: Brice Ozenne
##     Update #: 310
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - summary2
#' @title Summary with Small Sample Correction
#' @description Summary with small sample correction.
#' @name summary2
#'
#' @param object a \code{gls}, \code{lme} or \code{lvm} object.
#' @param digit [integer > 0] the number of decimal places to use when displaying the summary.
#' @param df [logical] should the degree of freedoms of the Wald statistic be computed using the Satterthwaite correction?
#' Otherwise the degree of freedoms are set to \code{Inf}, i.e. a normal distribution is used instead of a Student's t distribution when computing the p-values.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias?
#' See \code{\link{sCorrect}} for more details.
#' @param robust [logical] should the robust standard errors be used instead of the model based standard errors?
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param ... arguments passed to the \code{summary} method of the object.
#' 
#' @seealso \code{\link{sCorrect}} for more detail about the small sample correction.
#'
#' @details \code{summary2} is the same as \code{summary}
#' except that it first computes the small sample correction (but does not store it).
#' So if \code{summary2} is to be called several times,
#' it is more efficient to pre-compute the quantities for the small sample correction
#' using \code{sCorrect} and then call \code{summary2}.
#' 
#' @examples
#' m <- lvm(Y~X1+X2)
#' set.seed(10)
#' d <- lava::sim(m, 2e1)
#'
#' ## Gold standard
#' summary(lm(Y~X1+X2, d))$coef
#' 
#' ## gls models
#' library(nlme)
#' e.gls <- gls(Y~X1+X2, data = d, method = "ML")
#' summary(e.gls)$tTable
#' sCorrect(e.gls, cluster = 1:NROW(d)) <- FALSE ## no small sample correction
#' summary2(e.gls)$tTable
#' 
#' sCorrect(e.gls, cluster = 1:NROW(d)) <- TRUE ## small sample correction
#' summary2(e.gls)$tTable
#' 
#' ## lvm models
#' e.lvm <- estimate(m, data = d)
#' summary(e.lvm)$coef
#' 
#' sCorrect(e.lvm) <- FALSE ## no small sample correction
#' summary2(e.lvm)$coef
#' 
#' sCorrect(e.lvm) <- TRUE ## small sample correction
#' summary2(e.lvm)$coef
#' 
#' @concept small sample inference
#' @export
`summary2` <-
  function(object, ...) UseMethod("summary2")

## * summary2.lm
#' @rdname summary2
#' @method summary2 lm
#' @export
summary2.lm <- function(object, df = TRUE, bias.correct = TRUE, ...){
    sCorrect(object, df = df) <- bias.correct
    return(summary2(object, df = df, ...))
}
## * summary2.gls
#' @rdname summary2
#' @method summary2 gls
#' @export
summary2.gls <- function(object, df = TRUE, bias.correct = TRUE, cluster = NULL, ...){
    sCorrect(object, df = df, cluster = cluster) <- bias.correct
    return(summary2(object, df = df, ...))
}

## * summary2.lme
#' @rdname summary2
#' @method summary2 lme
#' @export
summary2.lme <- summary2.lm

## * summary2.lvmfit
#' @rdname summary2
#' @method summary2 lvmfit
#' @export
summary2.lvmfit <- summary2.lm

## * summary2.lm2
#' @rdname summary2
#' @method summary2 lm2
#' @export
summary2.lm2 <- function(object, 
                         digit = max(3, getOption("digit")),
                         robust = FALSE,
                         df = TRUE,
                         ...){

### ** perform Wald test
    name.param <- names(coef(object))
    n.param <- length(name.param)

    tTable.all <- compare2(object,
                           par = name.param,
                           robust = robust,
                           df = df,
                           F.test = FALSE,
                           as.lava = FALSE)
    tTable <- tTable.all[1:n.param,c("estimate","std","statistic","p-value","df")]
    dimnames(tTable) <- list(name.param,
                             c("Value","Std.Error","t-value","p-value","df")
                             )

### ** get summary
    class(object) <- setdiff(class(object),c("lm2"))
    object.summary <- summary(object, digits = digit, ...)
    
### ** update summary
    object.summary$coefficients <- tTable

### ** export
    return(object.summary)

}
## * summary2.gls2
#' @rdname summary2
#' @method summary2 gls2
#' @export
summary2.gls2 <- function(object, 
                          digit = max(3, getOption("digit")),
                          robust = FALSE,
                          df = TRUE,
                          ...){
    
    ### ** perform Wald test
    name.param <- names(coef(object))
    n.param <- length(name.param)

    tTable.all <- compare2(object,
                           par = name.param,
                           robust = robust,
                           df = df,
                           F.test = FALSE,
                           as.lava = FALSE)
    tTable <- tTable.all[1:n.param,c("estimate","std","statistic","p-value","df")]
    dimnames(tTable) <- list(name.param,
                             c("Value","Std.Error","t-value","p-value","df")
                             )

    ### ** get summary
    class(object) <- setdiff(class(object),c("gls2","lme2"))
    object.summary <- summary(object, digits = digit, ...)
    
    ### ** update summary
    object.summary$tTable <- tTable

    ### ** export
    return(object.summary)
}

## * summary2.lme2
#' @rdname summary2
#' @method summary2 lme2
#' @export
summary2.lme2 <- summary2.gls2

## * summary2.lvmfit2
#' @rdname summary2
#' @method summary2 lvmfit2
#' @export
summary2.lvmfit2 <- function(object, cluster = NULL, robust = FALSE, df = TRUE, ...){

    
### ** perform Wald test
    param <- lava::pars(object)
    name.param <- names(param)
    n.param <- length(param)

    table.all <- compare2(object,
                          par = name.param,
                          robust = robust,
                          cluster = cluster,
                          df = df,
                          F.test = FALSE,
                          as.lava = FALSE)
    table.coef <- table.all[1:n.param,c("estimate","std","statistic","p-value","df")]
    dimnames(table.coef) <- list(name.param,
                                 c("Estimate", "Std. Error", "t-value", "P-value", "df")
                                 )

### ** get summary
    class(object) <- setdiff(class(object),"lvmfit2")
    object.summary <- summary(object, ...)
    if(!is.null(object$cluster) || inherits(object,"lvm.missing")){
        
        ## if(robust == FALSE){
        ##     stop("Can only print summary for robust standard errors \n",
        ##          "when the object contain a cluster variable \n")
        ## }
        colnames(object.summary$coef) <- c("Estimate","Std. Error","Z-value","P-value")
        object.summary$coef[,"Z-value"] <- NA

        colnames(object.summary$coefmat) <- c("Estimate","Std. Error","Z-value","P-value", "std.xy")
        object.summary$coefmat[,"Z-value"] <- ""
        
    }
    
    ## find digit
    vec.char <- setdiff(object.summary$coefmat[,"Estimate"],"")
    digit <- max(c(nchar(gsub(".","",vec.char,fixed = TRUE)))-1,1)

    ### ** update summary
    ### *** vcov
    object.summary$vcov <- attr(object$dVcov, "vcov.param")[name.param,name.param]    

### *** coef
    lava.rownames <- rownames(object.summary$coef)
    ## add rows corresponding to reference parameters
    missing.rows <- setdiff(lava.rownames,rownames(table.coef))
    if(length(missing.rows)>0){
        addon <- object.summary$coef[missing.rows,
                                     c("Estimate","Std. Error","Z-value","P-value"),
                                     drop=FALSE]
        colnames(addon)[3] <- "t-value"
        table.coef <- rbind(table.coef, cbind(addon,df=NA))
    }

    ## re-order table according to lava
    table.coef <- table.coef[intersect(lava.rownames,rownames(table.coef)),,drop=FALSE]
    ## remove unappropriate p.values
    lava.NApvalue <- which(is.na(object.summary$coef[,"P-value"]))
    table.coef[intersect(lava.rownames[lava.NApvalue],rownames(table.coef)),"P-value"] <- NA
    object.summary$coef <- table.coef
    
    
    ### *** coefmat
    name.label0 <- trimws(rownames(CoefMat(object, labels = 0, level = 9)), which = "both")
    index.titleVariance <- which(name.label0=="Residual Variances:")
    if(length(index.titleVariance)>0){
        ## rename variance parameters from Y to Y~~Y
        index.vcov <- (index.titleVariance+1):length(name.label0)
        index.var <- setdiff(index.vcov,grep("~~",name.label0,fixed=TRUE)) ## exclude covariance parameters that are already correctly named
        name.label0[index.var] <- paste0(name.label0[index.var],lava.options()$symbols[2],name.label0[index.var])
    }

    table.coefmat <- object.summary$coefmat
    colnames(table.coefmat)[3:5] <- c("t-value","P-value","df")
    
    ## mimic lava:::CoefMat (called by lava:::summary.lvmfit)    
    e2add <- format(round(table.coef[,"Estimate"], max(1, digit - 1)), digits = digit - 1)
    e2add <- gsub(" NA","",e2add)
    sd2add <- format(round(table.coef[,"Std. Error"], max(1, digit - 1)), digits = digit - 1)
    sd2add <- gsub(" NA","",sd2add)
    df2add <- as.character(round(table.coef[,"df"],2))    
    df2add[is.na(df2add)] <- ""
    t2add <- format(round(table.coef[,"t-value"], max(1, digit - 1)), digits = digit - 1)
    t2add <- gsub(" NA","",t2add)

    p2add <- formatC(table.coef[,"P-value"], digits = digit - 1, format = "g",  preserve.width = "common", flag = "")
    p2add <- gsub(" NA","",p2add)
    p2add[table.coef[,"P-value"] < 1e-12] <- "  <1e-12"

    M2add <- cbind(e2add,sd2add,t2add,p2add,df2add)
    table.coefmat[match(rownames(table.coef), name.label0),] <- M2add

    table.coefmat[object.summary$coefmat[,"P-value"]=="","P-value"] <- ""
    object.summary$coefmat <- table.coefmat

### ** Export
    if(robust){
        colnames(object.summary$coefmat)[2] <- "robust SE"
        colnames(object.summary$coef)[2] <- "robust SE"
    }
    return(object.summary)    
}



##----------------------------------------------------------------------
### summary2.R ends here

### test1-sCorrect-summary2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: apr  4 2018 (13:29) 
## Version: 
## Last-Updated: jan 17 2022 (16:51) 
##           By: Brice Ozenne
##     Update #: 67
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)

    printDF <- function(object, bias.correct){
        colDF <- summary2(object, bias.correct = bias.correct)$coef[,"df",drop=FALSE]
        n.coef <- NROW(colDF)
        vec.end <- c(rep(",",n.coef-1),")")
        vec.start <- c("c(", rep("",n.coef-1))        
        df <- data.frame(paste0(vec.start,"\"",rownames(colDF),"\""),
                         "=",
                         paste0(colDF[,1],vec.end))
        names(df) <- NULL
        print(df, row.names = FALSE)
    }

}

lava.options(symbols = c("~","~~"))
library(nlme)
context("sCorrect (dVcov-SatterthwaiteCorrection)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * linear regression 
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d)

## ** test df
test_that("linear regression: Satterthwaite (df)", {
    df <- c("Y1" = 50,
            "Y1~X1" = 50,
            "Y1~X2" = 50,
            "Y1~GenderFemale" = 50,            
            "Y1~~Y1" = 12.5)
    expect_equal(as.double(df),
                 summary2(e.lvm, ssc = "none")$coef$df)
})

test_that("linear regression: Satterthwaite + SSC (df)", {
    df <- c("Y1" =   46,
            "Y1~X1" =   46,
            "Y1~X2" =   46,
            "Y1~GenderFemale" =   46,
            "Y1~~Y1" = 11.5)
    expect_equal(as.double(df),
                 summary2(e.lvm, ssc = "residuals")$coef$df)
})

## ** robust standard error
test_that("linear regression: robust SE", {
    ## printDF(e.lvm, bias.correct = TRUE)
    lava.options(df.robust = 1)
    eS0 <- summary2(e.lvm, robust = TRUE, df = "satterthwaite")$coef
    
    df <- c("Y1" =   46,
            "Y1~X1" =   46,
            "Y1~X2" =   46,
            "Y1~GenderFemale" =   46,
            "Y1~~Y1" = 11.5)
    expect_equal(as.double(df),
                 eS0$df, tol = 1e-2)
    
    eS1 <- summary2(e.lvm, robust = TRUE, df = "satterthwaite")$coef
    eS2 <- summary2(e.lvm, robust = TRUE, df = "satterthwaite", cluster = 1:n)$coef
    expect_equal(eS1,eS2)
   
})

## * linear regression with constrains 
## ** model fit
e.lvm <- estimate(lvm(Y1[0:2]~X1+1*X2), data = d)

e.lvm2 <- estimate(lvm(Y1~beta*X1+beta*X2), d)


## ** test df
test_that("linear regression with constrains: Satterthwaite (df)", {
    expect_equal(summary2(e.lvm)$coef$df,c(Inf)) ## Inf since the variance coefficient is known
    ## printDF(e.lvm2, bias.correct = FALSE)
    df <- c("Y1~X1" =   50,
            "Y1" =   50,
            "Y1~~Y1" = 12.5)
    expect_equal(summary2(e.lvm2, ssc = "none")$coef$df,
                 as.double(df))
})

test_that("linear regression with constrains: Satterthwaite + SSC (df)", {
    expect_equal(summary2(e.lvm)$coef$df,c(Inf)) ## Inf since the variance coefficient is known
    ## printDF(e.lvm2, bias.correct = TRUE)
    df <- c("Y1~X1" =   48,
            "Y1" =   48,
            "Y1~~Y1" = 12)
    expect_equal(summary2(e.lvm2)$coef$df,
                 as.double(df))
})

## * multiple linear regression 
## ** model fit
ls.lm <- list(lm(Y1~X1,d),lm(Y2~X2,d),lm(Y3~X1+X3,d))
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

## ** test df
test_that("multiple linear regression: Satterthwaite (df)", {
    ## printDF(e.lvm, bias.correct = FALSE)
    df <- c("Y1~X1" =   50,
            "Y2~X2" =   50,
            "Y3~X1" =   50,
            "Y3~X3" =   50,
            "Y1~~Y1" = 12.5,
            "Y2~~Y2" = 12.5,
            "Y3~~Y3" = 12.5,
            "Y1" =   50,
            "Y2" =   50,
            "Y3" =   50)
    expect_equal(summary2(e.lvm, ssc = "none")$coef[names(df),"df"],
                 as.double(df)) ## 
    
})

test_that("multiple linear regression: Satterthwaite + SSC (df)", {
    ## printDF(e.lvm, bias.correct = TRUE)
    df <- c("Y1~X1" =    48,
            "Y2~X2" =    48,
            "Y3~X1" =    47,
            "Y3~X3" =    47,
            "Y1~~Y1" =    12,
            "Y2~~Y2" =    12,
            "Y3~~Y3" = 11.75,
            "Y1" =    48,
            "Y2" =    48,
            "Y3" =    47)
    expect_equal(summary2(e.lvm)$coef[names(df),"df"],
                 as.double(df)) ## 
    
})

## * multiple linear regression with constrains 
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+1*X2,Y2~2*X3+2*X1,Y3~X2), data = d)

## ** test df
test_that("multiple linear regression with constrains: Satterthwaite (df)", {
    ## printDF(e.lvm, bias.correct = FALSE)
    df <- c("Y1~X1" =   50,
            "Y1~X2" =   NA,
            "Y2~X1" =   NA,
            "Y2~X3" =   NA,
            "Y3~X2" =   50,
            "Y1~~Y1" = 12.5,
            "Y2~~Y2" = 12.5,
            "Y3~~Y3" = 12.5,
            "Y1" =   50,
            "Y2" =   50,
            "Y3" =   50)
    expect_equal(summary2(e.lvm, ssc = "none")$coef[names(df),"df"],
                 as.double(df)) ## 
    
})

test_that("multiple linear regression with constrains: Satterthwaite + SSC (df)", {
    ## printDF(e.lvm, bias.correct = TRUE)
    df <- c("Y1~X1" =    48,
            "Y1~X2" =    NA,
            "Y2~X1" =    NA,
            "Y2~X3" =    NA,
            "Y3~X2" =    48,
            "Y1~~Y1" =    12,
            "Y2~~Y2" = 12.25,
            "Y3~~Y3" =    12,
            "Y1" =    48,
            "Y2" =    49,
            "Y3" =    48)
    expect_equal(summary2(e.lvm)$coef[names(df),"df"],
                 as.double(df)) ## 
    
})

## * multiple linear regression with covariance links 
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+X2,Y2~X3+X1,Y3~X2,Y1~~Y2),d)

## ** test df
test_that("multiple linear regression with covariance: Satterthwaite (df)", {
    ## printDF(e.lvm, bias.correct = FALSE)
    df <- c("Y1~X1" = 50.0023249929247,
            "Y1~X2" = 50.0557533452502,
            "Y2~X1" = 50.1412333709522,
            "Y2~X3" = 50.0557533452502,
            "Y3~X2" =               50,
            "Y1~~Y1" =             12.5,
            "Y1~~Y2" = 14.4382586892588,
            "Y2~~Y2" =             12.5,
            "Y3~~Y3" =             12.5,
            "Y1" = 51.0449669789772,
            "Y2" = 50.0000667169911,
            "Y3" =               50)
    expect_equal(summary2(e.lvm, ssc = "none")$coef[names(df),"df"],
                 as.double(df), tol = 1e-6) ## 
    
})

test_that("multiple linear regression with covariance: Satterthwaite +SSC (df)", {
    ## printDF(e.lvm, bias.correct = TRUE)
    df <- c("Y1~X1" = 47.0021511585814,
            "Y1~X2" = 47.0515840309539,
            "Y2~X1" = 47.1306527469107,
            "Y2~X3" = 47.0515840309539,
            "Y3~X2" =               48,
            "Y1~~Y1" =            11.75,
            "Y1~~Y2" = 13.6588307493286,
            "Y2~~Y2" =            11.75,
            "Y3~~Y3" =               12,
            "Y1" = 47.9656506208306,
            "Y2" = 47.0000617288762,
            "Y3" =               48)
    expect_equal(summary2(e.lvm)$coef[names(df),"df"],
                 as.double(df), tol = 1e-6) ## 
    
})



## * two factors model (regression)
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3,
           eta1~eta2))

e.lvm <- lava::estimate(m, d)

test_that("Two factor model: Satterthwaite +SSC (df)", {
    GS <- model.tables(estimate2(e.lvm))
    expect_equal(summary2(e.lvm)$coef[rownames(GS),"df"],
                 GS$df)
})
######################################################################
### test1-sCorrect-summary2.R ends here

### test1-sCorrect-dVcov.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  8 2018 (14:56) 
## Version: 
## Last-Updated: feb 11 2019 (13:43) 
##           By: Brice Ozenne
##     Update #: 114
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
library(nlme)
context("sCorrect (dVcov-SatterthwaiteCorrection)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * linear regression [lm,gls,lvm]
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d)
e.lm <- lm(Y1~X1+X2+Gender, data = d)
e.gls <- gls(Y1~X1+X2+Gender, data = d, method = "ML")

## ** check dVcov
test_that("linear regression: Satterthwaite", {
    X <- model.matrix(e.lm)
    sigma2 <- coef(e.lvm)["Y1~~Y1"]
    dI <- bdiag(crossprod(X)/sigma2^2,n/(sigma2^3))
    vcov <- solve(bdiag(crossprod(X)/sigma2,n/(2*sigma2^2)))
    GS <- vcov %*% dI %*% vcov

    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(as.double(test.lvm), as.double(GS))

    test.lm <- sCorrect(e.lm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lm <- sCorrect(e.lm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lm, GS.lm)
    expect_equal(as.double(test.lm), as.double(GS))

    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE, cluster = "Id")$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE, cluster = "Id")$dVcov.param
    expect_equal(test.gls, GS.gls)
    expect_equal(as.double(test.gls), as.double(GS))
})

## ** check dVcov (Satterthwaite + small sample correct) [should be moved to another file]
test_that("linear regression: Satterthwaite + small sample correction (dVcov))", {
    X <- model.matrix(e.lm)
    sigma2 <- sigma(e.lm)^2
    dI <- bdiag(crossprod(X)/sigma2^2,(n-NCOL(X))/(sigma2^3))
    vcov <- solve(bdiag(crossprod(X)/sigma2,(n-NCOL(X))/(2*sigma2^2)))
    GS <- vcov %*% dI %*% vcov

    test.lvm <- sCorrect(e.lvm, adjust.Omega = TRUE, adjust.n = TRUE, numeric.derivative = FALSE)$dVcov.param
    expect_equal(as.double(test.lvm), as.double(GS), tol = 1e-5)

    test.lm <- sCorrect(e.lm, adjust.Omega = TRUE, adjust.n = TRUE, numeric.derivative = FALSE)$dVcov.param
    expect_equal(as.double(test.lm), as.double(GS), tol = 1e-5)
    test.gls <- sCorrect(e.gls, adjust.Omega = TRUE, adjust.n = TRUE, numeric.derivative = FALSE, cluster = "Id")$dVcov.param
    expect_equal(as.double(test.gls), as.double(GS), tol = 1e-5)
})

## * linear regression with constrains [lvm]
## ** model fit
e.lvm <- estimate(lvm(Y1[0:2]~X1+1*X2), data = d)

e.lvm2 <- estimate(lvm(Y1~beta*X1+beta*X2), d)

## ** check dVcov
test_that("linear regression with constrains: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)

    test.lvm2 <- sCorrect(e.lvm2, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm2 <- sCorrect(e.lvm2, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm2, GS.lvm2)    
})

## * multiple linear regression [lvm,gls]
## ** model fit
ls.lm <- list(lm(Y1~X1,d),lm(Y2~X2,d),lm(Y3~X1+X3,d))
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

e.lvm2 <- estimate(lvm(Y1[mu:sigma1]~ beta1*X1 + beta2*X2,
                       Y2[mu:sigma2]~ beta1*X1 + beta2*X2,
                       Y3[mu:sigma3]~ beta1*X1 + beta2*X2),
                       data = d)
e.gls <- gls(value ~ X1 + X2,
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             weight = varIdent(form = ~1|variable),
             method = "ML")

test_that("gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm2)), as.double(logLik(e.gls)))
})

## ** check dVcov (Satterthwaite)
test_that("multiple linear regression: Satterthwaite (dVcov)", {
    X <- lapply(ls.lm, model.matrix)
    sigma2 <- list(coef(e.lvm)["Y1~~Y1"],
                   coef(e.lvm)["Y2~~Y2"],
                   coef(e.lvm)["Y3~~Y3"])
    dI <- mapply(X,sigma2, FUN = function(x,y){
        bdiag(crossprod(x)/y^2,n/(y^3))
    })
    vcov <- mapply(X,sigma2, FUN = function(x,y){
        solve(bdiag(crossprod(x)/y,n/(2*y^2)))
    })
    GS <- mapply(vcov, dI, FUN = function(x,y){
        x %*% y %*% x
    })
    name.coef.lvm <- names(coef(e.lvm))
    
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(as.double(test.lvm[grep("Y1",name.coef.lvm),grep("Y1",name.coef.lvm),"Y1~~Y1"]),
                 as.double(GS[[1]]))
    expect_equal(as.double(test.lvm[grep("Y2",name.coef.lvm),grep("Y2",name.coef.lvm),"Y2~~Y2"]),
                 as.double(GS[[2]]))
    expect_equal(as.double(test.lvm[grep("Y3",name.coef.lvm),grep("Y3",name.coef.lvm),"Y3~~Y3"]),
                 as.double(GS[[3]]))

    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE,
                         numeric.derivative = FALSE, cluster = "Id")$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE,
                       numeric.derivative = TRUE, cluster = "Id")$dVcov.param
    expect_equal(test.gls, GS.gls)    
})

## ** multiple linear regression (Satterthwaite + small sample correct) [should be moved to another file]
test_that("multiple linear regression: Satterthwaite + small sample correction (dVcov)", {
    X <- lapply(ls.lm, model.matrix)
    sigma2 <- lapply(ls.lm, function(x){sigma(x)^2})
    dI <- mapply(X,sigma2, FUN = function(x,y){
        bdiag(crossprod(x)/y^2,(n-NCOL(x))/(y^3))
    })
    vcov <- mapply(X,sigma2, FUN = function(x,y){
        solve(bdiag(crossprod(x)/y,(n-NCOL(x))/(2*y^2)))
    })
    GS <- mapply(vcov, dI, FUN = function(x,y){
        x %*% y %*% x
    })
    name.coef.lvm <- names(coef(e.lvm))

    test.lvm <- sCorrect(e.lvm, adjust.Omega = TRUE, adjust.n = TRUE, numeric.derivative = FALSE)$dVcov.param
    expect_equal(as.double(test.lvm[grep("Y1",name.coef.lvm),grep("Y1",name.coef.lvm),"Y1~~Y1"]),
                 as.double(GS[[1]]))
    expect_equal(as.double(test.lvm[grep("Y2",name.coef.lvm),grep("Y2",name.coef.lvm),"Y2~~Y2"]),
                 as.double(GS[[2]]))
    expect_equal(as.double(test.lvm[grep("Y3",name.coef.lvm),grep("Y3",name.coef.lvm),"Y3~~Y3"]),
                 as.double(GS[[3]]))

    test.gls <- sCorrect(e.gls, adjust.Omega = TRUE, adjust.n = TRUE, numeric.derivative = FALSE, cluster = "Id")$dVcov.param
    
})

## * multiple linear regression with constrains [lvm]
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+1*X2,Y2~2*X3+2*X1,Y3~X2), data = d)

## ** check dVcov
test_that("multiple linear regression with constrains: Satterthwaite", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
})


## * multiple linear regression with covariance links [lvm]
## ** model fit
e.lvm <- estimate(lvm(Y1~X1+X2,Y2~X3+X1,Y3~X2,Y1~~Y2),d)

## ** check dVcov
test_that("multiple linear regression with covariance: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
})


## * mixed model: Compound symmetry [lvm,gls,lme]
## ** model fit
m <- lvm(Y1[mu1:sigma]~1*eta,
         Y2[mu2:sigma]~1*eta,
         Y3[mu3:sigma]~1*eta,
         eta~X1+Gender)
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corCompSymm(form=~ 1|Id),
             data = dLred,
             method = "ML")

test_that("compound symmetry: lme/gls equivalent to lvm",{
    expect_equal(as.double(logLik(e.lme)),as.double(logLik(e.lvm)))
    expect_equal(as.double(logLik(e.gls)),as.double(logLik(e.lvm)))
})


## ** check dVcov
test_that("compound symmetry: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.043665672, 0.001141085, 0.086044394),
                 tol = 1e-8) ## compare with previous version
   
    test.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lme, GS.lme)
   
    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.gls, GS.gls)
})

## * mixed model: CS with different variances [lvm,lme]
## ** model fit 
m <- lvm(c(Y1[mu1:sigma1]~1*eta,
           Y2[mu2:sigma2]~1*eta,
           Y3[mu3:sigma3]~1*eta,
           eta~X1+Gender))
latent(m) <- ~eta
e.lvm <- estimate(m, d)

e.lme <- nlme::lme(value ~ variable + X1 + Gender,
                   random =~1| Id,
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ variable + X1 + Gender,
                   correlation = corCompSymm(form = ~1| Id),
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

test_that("lme equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    ## gls does not give the same likelihood
    ## expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** check dVcov
test_that("compound symmetry with different variances: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm) ## internal consistency
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.043665672, 0.001163405, 0.158962412),
                 tol = 1e-8) ## compare with previous version
    
    test.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lme, GS.lme)

    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.gls, GS.gls)
})


## * mixed model: Unstructured [lvm,gls,lme]
## ** model fit 
m <- lvm(Y1[mu1:sigma]~1*eta,
         Y2[mu2:sigma]~1*eta,
         Y3[mu3:sigma]~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             ## weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             ## weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** check dVcov
test_that("Unstructured: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.069511713, 0.001092602, 0.129487510),
                 tol = 1e-8) ## compare with previous version
    ## ERROR: inversion?
    ## test.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    ## GS.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    ## expect_equal(test.lme, GS.lme)

    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.gls, GS.gls)
})


## * mixed model: Unstructured with weights [lvm,gls,lme]
## ** model fit
m <- lvm(Y1~1*eta,
         Y2~1*eta,
         Y3~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** check dVcov
test_that("Unstructured with different variances: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.08912488, 0.00188713, 0.18430283),
                 tol = 1e-8) ## compare with previous version
    ## summary2(e.lvm)
    
    ## ERROR: the model is overparametrized
    ## test.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    ## GS.lme <- sCorrect(e.lme, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    ## expect_equal(test.lme, GS.lme)

    test.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.gls <- sCorrect(e.gls, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.gls, GS.gls)
})


## * LVM: factor model
## ** model fit
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1))
regression(m) <- eta1~X1+X2

e.lvm <- estimate(m,d)

## ** check dVcov
test_that("factor model: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.297105651, 0.001184003, 0.360920233),
                 tol = 1e-8) ## compare with previous version
})



## * LVM: factor model with constrains
## ** model fit
e.lvm <- estimate(lvm(Y1~1*eta+1*X2,Y2~1*eta,Y3~1*eta),
                  data = d)

e.lvm2 <- estimate(lvm(Y1~1*eta+X2,
                       Y2~lambda*eta+X2,
                       Y3~lambda*eta,
                       eta ~ beta*X2+beta*X1),
                   data = d)

## ** check dVcov
test_that("factor model: Satterthwaite", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.035030251, 0.004314892, 0.289075943),
                 tol = 1e-8) ## compare with previous version    
})

## * LVM: 2 factor model
## ** model fit
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(m) <- eta1~X1+X2
latent(m) <- ~eta1+eta2

e.lvm <- estimate(m,d)

## ** check dVcov
test_that("2 factor model: Satterthwaite", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-1.407457972, 0.000378982, 1.299890022),
                 tol = 1e-8) ## compare with previous version   
})

## * LVM: 2 factor model with constrains
## ** model fit
m <- lvm(Y1~1*eta1+X2,Y2~lambda*eta1+X2,Y3~lambda*eta1,eta1 ~ beta*X2+beta*X1,
         Z1~0+eta2,Z2~lambda*eta2,Z3~eta2)
e.lvm <- estimate(m, d)
e2.lvm <- e.lvm

## ** check dVcov
test_that("2 factor model with constrains: Satterthwaite", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.9719462803, 0.0005765745, 1.0173093948),
                 tol = 1e-8) ## compare with previous version   
})

## * LVM: 2 factor model (covariance)
## ** model fit
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
covariance(m) <- eta1 ~ eta2
latent(m) <- ~eta1+eta2

e.lvm <- estimate(m,d)

## ** check dVcov
test_that("2 factor model with covariance: Satterthwaite", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-1.0091216732, 0.0004489561, 1.0301898081),
                 tol = 1e-8) ## compare with previous version   
})

## * LVM: 2 factor model (correlation LV)
## ** model fit
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(m) <- eta2 ~ X1
regression(m) <- eta1 ~ eta2+X2+X3

e.lvm <- estimate(m,d)

## ** check dVcov
test_that("2 factor model with correlation: Satterthwaite (dVcov)", {
    test.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = FALSE)$dVcov.param
    GS.lvm <- sCorrect(e.lvm, adjust.Omega = FALSE, adjust.n = FALSE, numeric.derivative = TRUE)$dVcov.param
    expect_equal(test.lvm, GS.lvm)
    expect_equal(c(min(test.lvm),mean(test.lvm),max(test.lvm)),
                 c(-0.7338483344, 0.0002132907, 0.8107582027),
                 tol = 1e-8) ## compare with previous version
})

### test1-sCorrect-missingValues.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  7 2018 (13:39) 
## Version: 
## Last-Updated: feb 11 2019 (14:32) 
##           By: Brice Ozenne
##     Update #: 36
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
.coef2 <- lavaSearch2:::.coef2
library(nlme)
context("sCorrect (dealing with missing values)")

## * simulation
n <- 2e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
           measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]
dLred2 <- dL[dL$variable %in% c("Y1","Y2"),]
dLred3 <- dLred2
dLred3[dL$variable == "Y2","Id"] <- paste0("-",dLred3[dL$variable == "Y2","Id"])

## ** t test
## formula:
## df = \frac{ 2 * s_pool^2 }{ var(s_pool^2) }
##    = \frac{ ( s_X^2/m + s_Y^2/n )^2}{( s_X^4/(m(m-1)) + s_Y^4/(n(n-1)))}

## using the t test function
e.ttest <- t.test(d$Y1, d$Y2)
e.ttest$parameter

## by hand
sX1 <- var(d$Y1)/n
sX2 <- var(d$Y2)/n
df <- (sX1+sX2)^2/(sX1^2/(n-1) + sX2^2/(n-1))
df-e.ttest$parameter

## *** using gls - no missing data
e.gls <- gls(value ~ variable, data = dLred2,
             weights = varIdent(form = ~1|variable),
             method = "ML")

expect_equal(as.double(coef(e.gls)[2]),
             as.double(diff(e.ttest$estimate)))

test_that("t test (full data)", {
    eS.gls <- summary2(e.gls, cluster = "Id")$tTable
    expect_equal(unname(e.ttest$parameter),
                 eS.gls["variableY2","df"],
                 tol = 1e-3)
    expect_equal(unname(e.ttest$p.value),
                 eS.gls["variableY2","p-value"],
                 tol = 1e-5)
})

## *** using gls - missing data
eNA.gls <- gls(value ~ variable, data = dLred3,
               weights = varIdent(form = ~1|variable),
               method = "ML")

expect_equal(as.double(coef(eNA.gls)[2]),
             as.double(diff(e.ttest$estimate)))
## getVarCov2(eNA.gls, cluster = "Id")



test_that("t test (missing data)", {
    eNAS.gls <- summary2(eNA.gls, cluster = "Id")$tTable
    
    expect_equal(unname(e.ttest$parameter),
                 eNAS.gls["variableY2","df"],
                 tol = 1e-3)
    expect_equal(unname(e.ttest$p.value),
                 eNAS.gls["variableY2","p-value"],
                 tol = 1e-5)
})

## * LVM: factor model
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1))
regression(m) <- eta1~X1+X2

e.lvm <- estimate(m,d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- TRUE

## ** complete case analysis
missing.Row <- d[1,]
missing.Row[,"Id"] <- -1
missing.Row[,c("Y1","Y2","Y3")] <- NA
eNA.lvm <- estimate(m, rbind(d,missing.Row), missing = FALSE)

test_that("complete case analysis (factor model)", {
    eNA2.lvm <- eNA.lvm
    sCorrect(eNA2.lvm) <- TRUE

    expect_equal(eNA2.lvm$sCorrect,
                 e2.lvm$sCorrect)
})

## ** full information
missing.Row <- d[1,]
missing.Row[,"Id"] <- -1
missing.Row[,c("Y1","Y2")] <- NA
eNA.lvm <- estimate(m, rbind(d,missing.Row), missing = TRUE)

test_that("full information (factor model)", {
    eNA2.lvm <- eNA.lvm
    sCorrect(eNA2.lvm) <- TRUE
    ## sCorrect(eNA2.lvm, numeric.derivative = TRUE) <- TRUE
    summary2(eNA2.lvm)
})



##----------------------------------------------------------------------
### test1-sCorrect-missingValues.R ends here

### skeleton.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov  8 2017 (10:35) 
## Version: 
## Last-Updated: feb  8 2019 (11:48) 
##           By: Brice Ozenne
##     Update #: 1025
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - skeleton
#' @title Pre-computation for the Score
#' @description Pre-compute quantities that are necessary to compute the score of a lvm model.
#' @name skeleton
#' 
#' @param object a \code{lvm} object.
#' @param df.param.all [data.frame] output of \code{\link{coefType}} containing the type of each coefficient.
#' @param param2originalLink [named character vector] matching between the name of the coefficient in lava and their label.
#' @param B,alpha.XGamma,Lambda,Psi [matrix] pre-computed matrix.
#' @param OD [list] the pre-computed quantities for the second derivatives. 
#' @param as.lava [logical] should the name of the links be used to name the coefficient?
#' Otherwise uses the labels (when defined) of each coefficient.
#' @param name.endogenous [character vector] name of the endogenous variables
#' @param name.latent [character vector] name of the latent variables
#' @param p [numeric vector, optional] vector of coefficients at which to evaluate the score.
#' @param data [data.frame, optional] data set.
#' @param ... [internal] only used by the generic method.
#' 
#' @details
#' When the use specify names for the coefficients (e.g. Y1[mu:sigma]) or uses constrains (Y1~beta*X1), \code{as.lava=FALSE} will use the names specified by the user (e.g. mu, sigma, beta) while \code{as.lava=TRUE} will use the name of the first link defining the coefficient.
#'
#' @examples
#' \dontrun{
#' skeleton <- lavaSearch2::skeleton
#' skeleton.lvm <- lavaSearch2::skeleton.lvm
#' skeleton.lvmfit <- lavaSearch2::skeleton.lvmfit
#' 
#' ## without constrain
#' m <- lvm(Y1~X1+X2+eta,Y2~X3+eta,Y3~eta)
#' latent(m) <- ~eta
#' 
#' e <- estimate(m, lava::sim(m,1e2))
#' M.data <- as.matrix(model.frame(e))
#'
#' skeleton(e$model, as.lava = TRUE,
#'          name.endogenous = endogenous(e), n.endogenous = 3,
#'          name.latent = latent(e), 
#'          update.value = FALSE)
#' skeleton(e, data = M.data, p = pars(e), as.lava = TRUE,
#'          name.endogenous = endogenous(e), n.endogenous = 3,
#'          name.latent = latent(e), 
#'          update.value = TRUE)
#'
#' ## with constrains
#' m <- lvm(Y[mu:sigma] ~ beta*X1+X2)
#' e <- estimate(m, lava::sim(m,1e2))
#' M.data <- as.matrix(model.frame(e))
#'
#' skeleton(e$model, as.lava = TRUE,
#'          name.endogenous = "Y", n.endogenous = 1,
#'          name.latent = NULL, 
#'          update.value = FALSE)$skeleton
#' 
#' skeleton(e, data = M.data, p = pars(e), as.lava = FALSE,
#'          name.endogenous = "Y", n.endogenous = 1,
#'          name.latent = NULL, 
#'          update.value = FALSE)$skeleton
#' 
#'}
#' @concept small sample inference
#' @concept derivative of the score equation
#' @keywords internal
`skeleton` <-
    function(object, ...) UseMethod("skeleton")


## * skeleton.lvm
#' @rdname skeleton
skeleton.lvm <- function(object, as.lava,
                         name.endogenous, name.latent,
                         ...){
    detail <- Y <- NULL ## [:for CRAN check] subset

    n.endogenous <- length(name.endogenous)
    n.latent <- length(name.latent)
    
### ** prepare
    df.param.all  <- coefType(object, as.lava = FALSE)
    if(as.lava){
        param2originalLink <- subset(df.param.all, subset = !is.na(lava), select = c("originalLink", "param"))
        param2originalLink <- stats::setNames(param2originalLink$originalLink, param2originalLink$param)
    }else{
        param2originalLink <- subset(df.param.all, subset = !is.na(lava), select = "param", drop = TRUE)
        param2originalLink <- stats::setNames(param2originalLink, param2originalLink)
    }
    df.param.detail <- subset(df.param.all, subset = !is.na(detail)) ## important cannot be lava because we need to keep track of the constrained parameters
    
    skeleton <- list()
    value <- list()
    skeleton$type <- setNames(df.param.all[!is.na(df.param.all$lava),"detail"], df.param.all[!is.na(df.param.all$lava),"name"])
    skeleton$toUpdate <- stats::setNames(c(rep(FALSE,8),TRUE,TRUE,TRUE,TRUE),
                                         c("nu","K","Lambda","Sigma","alpha","Gamma","B","Psi",
                                           "extra","mu","Omega","param"))
   
### ** Measurement model
    
    ## *** nu
    df.param.nu <-  subset(df.param.detail, subset = detail=="nu", select = c("value", "param", "Y", "name"))
    df.param.nu <- df.param.nu[order(df.param.nu$name),]
    value$nu <- stats::setNames(df.param.nu$value,df.param.nu$Y)
    skeleton$nu <- stats::setNames(param2originalLink[df.param.nu$param],df.param.nu$Y)
    skeleton$toUpdate["nu"] <- any(is.na(value$nu))
    
    ## *** X K
    df.param.K <- subset(df.param.detail, subset = detail == "K", select = c("value", "param", "X", "Y"))
    df.param.K <- df.param.K[order(df.param.K$Y),]

    if(NROW(df.param.K)>0){
        value$K <- stats::setNames(lapply(1:n.endogenous, function(iEndogenous){ # iEndogenous <- 1
            subset(df.param.K, subset = Y == name.endogenous[iEndogenous], select = "value", drop = TRUE)
        }), name.endogenous)
            
        skeleton$K <- stats::setNames(lapply(1:n.endogenous, function(iEndogenous){
            param2originalLink[subset(df.param.K, subset = Y == name.endogenous[iEndogenous], select = "param", drop = TRUE)]
        }), name.endogenous)
    
        skeleton$XK <- stats::setNames(lapply(1:n.endogenous, function(iEndogenous){
            subset(df.param.K, subset = Y == name.endogenous[iEndogenous], select = "X", drop = TRUE)
        }), name.endogenous)
        
        skeleton$toUpdate["K"] <- any(unlist(lapply(value$K,is.na)))
    }
    
    ## *** Lambda
    if(n.latent>0){
        ## define matrix
        value$Lambda <- matrix(0,nrow = n.latent, ncol = n.endogenous,
                               dimnames = list(name.latent,name.endogenous))
        skeleton$Lambda <- matrix(as.character(NA),nrow = n.latent, ncol = n.endogenous,
                                  dimnames = list(name.latent,name.endogenous))
        ## update according to the model
        df.param.Lambda <- subset(df.param.detail, subset = detail == "Lambda", select = c("X","Y","param","value","name"))
        df.param.Lambda$index <- match(df.param.Lambda$X, name.latent) + n.latent * (match(df.param.Lambda$Y, name.endogenous)-1)
        df.param.Lambda <- df.param.Lambda[order(df.param.Lambda$name),]

        ## store in the Lambda matrix the name of the coefficient and their pre-computed values
        dfNA.tempo <- subset(df.param.Lambda, subset = is.na(value))
        skeleton$Lambda[dfNA.tempo$index] <- stats::setNames(param2originalLink[dfNA.tempo$param],dfNA.tempo$Y)
        dfNNA.tempo <- subset(df.param.Lambda, subset = !is.na(value))
        value$Lambda[dfNNA.tempo$index] <- stats::setNames(dfNNA.tempo$value,dfNNA.tempo$Y)
        value$Lambda[!is.na(skeleton$Lambda)] <- NA

        skeleton$toUpdate["Lambda"] <- any(is.na(value$Lambda))
    }

    ## *** Sigma    
    ## define matrix
    value$Sigma <- matrix(0,nrow = n.endogenous, ncol = n.endogenous,
                          dimnames = list(name.endogenous,name.endogenous))
    skeleton$Sigma <- matrix(as.character(NA),nrow = n.endogenous, ncol = n.endogenous,
                             dimnames = list(name.endogenous,name.endogenous))
    
    ## update according to the model
    df.param.Sigma <- subset(df.param.detail,
                             subset = detail %in% c("Sigma_var","Sigma_cov"),
                             select = c("X","Y","param","value","name"))
    df.param.Sigma$index <- match(df.param.Sigma$X, name.endogenous) + n.endogenous*(match(df.param.Sigma$Y, name.endogenous)-1)

    dfNA.tempo <- subset(df.param.Sigma, subset = is.na(value))
    skeleton$Sigma[dfNA.tempo$index] <- param2originalLink[dfNA.tempo$param]
    dfNNA.tempo <- subset(df.param.Sigma, subset = !is.na(value))
    value$Sigma[dfNNA.tempo$index] <- dfNNA.tempo$value

    ## symmetrize
    skeleton$Sigma <- symmetrize(skeleton$Sigma, update.upper = TRUE)
    value$Sigma <- symmetrize(value$Sigma, update.upper = TRUE)
    value$Sigma[!is.na(skeleton$Sigma)] <- NA

    skeleton$toUpdate["Sigma"] <- any(is.na(value$Sigma))

### ** Structural model
    if(n.latent>0){
        ## *** alpha 
        df.param.alpha <-  subset(df.param.detail,
                                  subset = detail=="alpha",
                                  select = c("value","param","Y"))
        value$alpha <- stats::setNames(df.param.alpha$value,df.param.alpha$Y)
        skeleton$alpha <- param2originalLink[stats::setNames(df.param.alpha$param,df.param.alpha$Y)]

        skeleton$toUpdate["alpha"] <- any(is.na(value$alpha))
        
        ## *** X Gamma
        df.param.Gamma <- subset(df.param.detail,
                                 subset = detail=="Gamma",
                                 select = c("value","param","X","Y"))
        
        if(NROW(df.param.Gamma)>0){
            
            value$Gamma <- stats::setNames(lapply(1:n.latent, function(iLatent){
                subset(df.param.Gamma, subset = Y==name.latent[iLatent], select = "value", drop = TRUE)
            }), name.latent)
            
            skeleton$Gamma <- stats::setNames(lapply(1:n.latent, function(iLatent){ # iLatent <- 1
                param2originalLink[subset(df.param.Gamma, subset = Y==name.latent[iLatent], select = "param", drop = TRUE)]
            }), name.latent)
    
            skeleton$XGamma <- stats::setNames(lapply(1:n.latent, function(iLatent){
                subset(df.param.Gamma, subset = Y==name.latent[iLatent], select = "X", drop = TRUE)
            }), name.latent)

            skeleton$toUpdate["Gamma"] <- any(unlist(lapply(value$Gamma,is.na)))
        }
        
        ## *** B
        ## define matrix
        value$B <- matrix(0,nrow = n.latent, ncol = n.latent,
                          dimnames = list(name.latent,name.latent))
        skeleton$B <- matrix(as.character(NA),nrow = n.latent, ncol = n.latent,
                             dimnames = list(name.latent,name.latent))

        if(any("B" %in% df.param.all$detail)){
            ## update according to the model
            df.param.B <- subset(df.param.detail,
                                 subset = detail == "B",
                                 select = c("X", "Y", "param", "value", "name"))
            df.param.B$index <- match(df.param.B$X, name.latent) + n.latent*(match(df.param.B$Y, name.latent)-1)
            dfNA.tempo <- subset(df.param.B, subset = is.na(value))            
            skeleton$B[dfNA.tempo$index] <- param2originalLink[dfNA.tempo$param]
            dfNNA.tempo <- subset(df.param.B, subset = is.na(value))
            value$B[dfNNA.tempo$index] <- dfNNA.tempo$value
            value$B[!is.na(skeleton$B)] <- NA

            skeleton$toUpdate["B"] <- any(is.na(value$B))
        }
    
        ## *** Psi    
        ## define matrix
        value$Psi <- matrix(0,nrow = n.latent, ncol = n.latent,
                            dimnames = list(name.latent,name.latent))
        skeleton$Psi <- matrix(as.character(NA),nrow = n.latent, ncol = n.latent,
                               dimnames = list(name.latent,name.latent))

        ## update according to the model
        df.param.Psi <- subset(df.param.all,
                               subset = detail %in% c("Psi_var","Psi_cov"),
                               select = c("X", "Y", "param", "value", "Y", "name"))
                               
        df.param.Psi$index <- match(df.param.Psi$X, name.latent) + n.latent*(match(df.param.Psi$Y, name.latent)-1)

        dfNA.tempo <- subset(df.param.Psi, subset = is.na(value))      
        skeleton$Psi[dfNA.tempo$index] <- param2originalLink[dfNA.tempo$param]
        dfNNA.tempo <- subset(df.param.Psi, subset = is.na(value))
        value$Psi[dfNNA.tempo$index] <- dfNNA.tempo$value

        ## symmetrize
        skeleton$Psi <- symmetrize(skeleton$Psi, update.upper = TRUE)
        value$Psi <- symmetrize(value$Psi, update.upper = TRUE)
        value$Psi[!is.na(skeleton$Psi)] <- NA

        skeleton$toUpdate["Psi"] <- any(is.na(value$Psi))
    }

### ** prepare matrix for updating the variance parameter according to the adjusted Omega
    index.matrix <- data.frame(index = which(upper.tri(skeleton$Sigma, diag = TRUE)),
                               which(upper.tri(skeleton$Sigma, diag = TRUE), arr.ind = TRUE)
                               )
    index.keep <- intersect(which(df.param.all$detail %in% c("Sigma_var","Sigma_cov","Psi_var","Psi_cov")),
                            which(!is.na(df.param.all$lava)))
    name.var <- df.param.all[index.keep,"name"]

    name.rhs <- paste(name.endogenous[index.matrix[,"row"]],
                      lava.options()$symbols[2],
                      name.endogenous[index.matrix[,"col"]],
                      sep = "")
    n.rhs <- length(name.rhs)
    
    A <- matrix(0, nrow = n.rhs, ncol = length(name.var),
                dimnames = list(name.rhs, name.var))
    vec.Sigma <- skeleton$Sigma[index.matrix$index]
    for(i in which(!is.na(vec.Sigma))){
        A[i, vec.Sigma[i]] <- 1
    }

    if(n.latent>0){
        index.Psi <- rbind(cbind(index = which(value$Psi!=0),
                                 which(value$Psi!=0, arr.ind = TRUE)),
                           cbind(index = which(is.na(value$Psi)),
                                 which(is.na(value$Psi), arr.ind = TRUE))
                           )
    }else{
        index.Psi <- NULL        
    }

    toUpdate2 <- c(nu = FALSE,
                   K = FALSE,
                   Lambda = FALSE,
                   Sigma = TRUE,
                   alpha = FALSE,
                   Gamma = FALSE,
                   B = FALSE,
                   Psi = TRUE,
                   extra = TRUE,
                   mu = FALSE,
                   Omega = FALSE,
                   param = FALSE)

    index.LambdaB <- names(skeleton$type)[which(skeleton$type %in% c("Lambda","B"))]
    adjustMoment <- list(index.matrix = index.matrix,
                         index.Psi = index.Psi,
                         index.LambdaB = index.LambdaB,
                         toUpdate = toUpdate2,
                         A = A,
                         name.endogenous = name.endogenous,
                         name.latent = name.latent,
                         name.var = name.var,
                         n.rhs = n.rhs)


### ** export
    return(list(skeleton = skeleton,
                value = value,
                df.param = df.param.all,
                adjustMoment = adjustMoment, 
                param2originalLink = param2originalLink)
           )
}


## * skeleton.lvmfit
#' @rdname skeleton
skeleton.lvmfit <- function(object, param, data,
                            name.endogenous, name.latent,
                            ...){
    
    n.endogenous <- length(name.endogenous)
    n.latent <- length(name.latent)
    n.data <- NROW(data)
    
    skeleton <- object$conditionalMoment$skeleton
    toUpdate <- skeleton$toUpdate
    value <- object$conditionalMoment$value
    
### ** Update skeleton with the current values
    ## *** nu
    if(toUpdate["nu"]){
        index.update <- which(!is.na(skeleton$nu))
        value$nu[index.update] <- param[skeleton$nu[index.update]]
    }
    
    ## *** K
    if(toUpdate["K"]){
        for(iY in 1:n.endogenous){ # iY <- 3
            if(length(skeleton$K[[iY]])>0){
                index.update <- which(!is.na(skeleton$K[[iY]]))
                value$K[[iY]][index.update] <- param[skeleton$K[[iY]][index.update]]
            }
        }
    }

    ## *** Lambda
    if(toUpdate["Lambda"]){
        index.update <- which(!is.na(skeleton$Lambda))
        value$Lambda[index.update] <- param[skeleton$Lambda[index.update]]
    }
    
    ## *** Sigma
    if(toUpdate["Sigma"]){
        index.update <- which(!is.na(skeleton$Sigma))
        value$Sigma[index.update] <- param[skeleton$Sigma[index.update]]
    }

    ## *** mu
    if(toUpdate["mu"]){ ## linear predictor (measurement model without latent variable)   
        value$nu.XK <- matrix(0, nrow = n.data, ncol = n.endogenous, byrow = TRUE,
                                       dimnames = list(NULL,name.endogenous))
        for(iY in 1:n.endogenous){ # iY <- 1
            iY2 <- name.endogenous[iY]
            if(length(value$K[[iY2]])>0){
                value$nu.XK[,iY2] <- value$nu[iY2] + data[,skeleton$XK[[iY2]],drop=FALSE] %*% value$K[[iY2]]
            }else{
                value$nu.XK[,iY2] <- value$nu[iY2]
            }
        }
    }
        
### ** Structural model
    if(n.latent>0){
        ## *** alpha
        if(toUpdate["alpha"]){
            index.update <- which(!is.na(skeleton$alpha))
            value$alpha[index.update] <- param[skeleton$alpha[index.update]]
        }
        
        ## *** Gamma
        if(toUpdate["Gamma"]){
            for(iLatent in 1:n.latent){
                if(length(skeleton$Gamma[[iLatent]])>0){
                    index.update <- which(!is.na(skeleton$Gamma[[iLatent]]))
                    value$Gamma[[iLatent]][index.update] <- param[skeleton$Gamma[[iLatent]][index.update]]
                }
            }
        }
        
        ## *** B
        if(toUpdate["B"] && length(skeleton$B)>0){
            index.update <- which(!is.na(skeleton$B))
            value$B[index.update] <- param[skeleton$B[index.update]]
        }
        
        ## *** Psi
        if(toUpdate["Psi"] && length(skeleton$Psi)>0){
            index.update <- which(!is.na(skeleton$Psi))
            value$Psi[index.update] <- param[skeleton$Psi[index.update]]
        }
        
        ## *** mu
        if(toUpdate["mu"]){ ## linear predictor (latent variable)            
            value$alpha.XGamma <- matrix(0,nrow = n.data, ncol = n.latent, byrow = TRUE,
                                         dimnames = list(NULL,name.latent))
        
            for(iLatent in 1:n.latent){
                iLatent2 <- name.latent[iLatent]
                if(length(value$Gamma[[iLatent2]])>0){
                    value$alpha.XGamma[,iLatent2] <- value$alpha[iLatent2] + data[,skeleton$XGamma[[iLatent2]],drop=FALSE] %*% value$Gamma[[iLatent2]]
                }else{
                    value$alpha.XGamma[,iLatent2] <- value$alpha[iLatent2]
                }
            }
            value$iIB <- solve(diag(1,n.latent,n.latent)-value$B)            
            value$alpha.XGamma.iIB <- value$alpha.XGamma %*% value$iIB
        }
        
        ## *** extra
        if(toUpdate["extra"]){
            value$iIB.Lambda <-  value$iIB %*% value$Lambda    
            value$Psi.iIB <- value$Psi %*% value$iIB
            value$tLambda.tiIB.Psi.iIB <- t(value$iIB.Lambda) %*% value$Psi.iIB
        }
    }

### ** Export
    return(value)
}


## * skeletonDtheta
#' @rdname skeleton
`skeletonDtheta` <-
    function(object, ...) UseMethod("skeletonDtheta")
## * skeletonDtheta.gls
#' @rdname skeleton
skeletonDtheta.gls <- function(object, class.cor, class.var, X, 
                               sigma2.base0, Msigma2.base0, M.corcoef, ref.group,
                               index.lower.tri, indexArr.lower.tri,
                               name.endogenous, n.endogenous, cluster, n.cluster,
                               var.coef, name.varcoef, name.otherVar, n.varcoef,
                               cor.coef, name.corcoef, n.corcoef,
                               index.Omega, update.mean, update.variance,
                               ...){

    out <- list()
    ## ** mean
    if(update.mean == TRUE){
        name.X <- colnames(X)
        out$dmu <- lapply(name.X, function(iCoef){ # iCoef <- name.X[1]
            dmu.tempo <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                                dimnames = list(NULL, name.endogenous))
            for(iC in 1:n.cluster){ ## iC <- 5
                dmu.tempo[iC,index.Omega[[iC]]] <- X[cluster==iC,iCoef]
            }
        
            return(dmu.tempo)
        })
        names(out$dmu) <- name.X
    }
    
    ## ** variance
    if(update.variance == TRUE){
        out$dOmega <- vector(mode = "list", length = n.corcoef + n.varcoef)
        names(out$dOmega) <- c(name.corcoef, name.varcoef)

        ## *** dispersion coefficient
        out$dOmega[["sigma2"]] <- diag(sigma2.base0, nrow = n.endogenous, ncol = n.endogenous)
   
        if("NULL" %in% class.cor == FALSE){
            out$dOmega[["sigma2"]][index.lower.tri] <- Msigma2.base0[index.lower.tri] * cor.coef[M.corcoef[index.lower.tri]]
            out$dOmega[["sigma2"]] <- symmetrize(out$dOmega[["sigma2"]])      
        }
        dimnames(out$dOmega[["sigma2"]]) <-  list(name.endogenous, name.endogenous)

        ## *** multiplicative factors
        if("NULL" %in% class.var == FALSE){

            for(iVar in name.otherVar){ # iVar <- name.otherVar
                iTest.endogenous <- ref.group %in% iVar
                out$dOmega[[iVar]] <- var.coef["sigma2"]*diag(iTest.endogenous,
                                                              nrow = n.endogenous, ncol = n.endogenous)

                if("NULL" %in% class.cor == FALSE){
                    index.iVar <- which(rowSums(indexArr.lower.tri==which(iTest.endogenous))>0)

                    ##  d sqrt(x) / d x = 1/(2 sqrt(x)) = sqrt(x) / (2*x)
                    out$dOmega[[iVar]][index.lower.tri[index.iVar]] <- var.coef["sigma2"]*out$dOmega[["sigma2"]][index.lower.tri[index.iVar]]/(2*var.coef[iVar])
                    out$dOmega[[iVar]] <- symmetrize(out$dOmega[[iVar]])
                }
            
                dimnames(out$dOmega[[iVar]]) <- list(name.endogenous, name.endogenous)            
            }
        }
    
        ## ** correlation
        if("NULL" %in% class.cor == FALSE){
            for(iVar in name.corcoef){
                out$dOmega[[iVar]] <- Msigma2.base0 * var.coef["sigma2"] * (M.corcoef==iVar)
            }
        }
    }
    
    ### ** export
    return(out)
}

## * skeletonDtheta.lme
#' @rdname skeleton
skeletonDtheta.lme <- function(object, name.endogenous, n.endogenous,
                               name.rancoef, ...){

    out <- skeletonDtheta.gls(object,
                              name.endogenous = name.endogenous,
                              n.endogenous = n.endogenous,
                              ...)

    out$dOmega[[name.rancoef]] <- matrix(1, nrow = n.endogenous, ncol = n.endogenous,
                                         dimnames = list(name.endogenous,name.endogenous)
                                         )

    return(out)
}

## * skeletonDtheta.lvm
#' @rdname skeleton
skeletonDtheta.lvm <- function(object, data,
                               df.param.all, param2originalLink,
                               name.endogenous, name.latent, ...){

    factitious <- marginal <- param <- value <- X <- Y <- NULL ## [:for CRAN check] subset

    n.endogenous <- length(name.endogenous)
    n.latent <- length(name.latent)

    df.param <- subset(df.param.all, subset = is.na(value) & marginal == FALSE & factitious == FALSE)
    Utype.by.detail <- tapply(df.param$detail, df.param$param, function(x){length(unique(x))})
    if(any(Utype.by.detail>1)){
        stop("cannot constrain two coefficients of different types to be equal \n")
    }
    name.param <- subset(df.param, subset = !duplicated(param), select = param, drop = TRUE)
    n.param <- length(name.param)

    name.originalLink <- as.character(param2originalLink)

### ** prepare
    n.data <- NROW(data)
    name.data <- colnames(data)
    
    mean.param <- c("nu","K","alpha","Gamma","Lambda","B")
    vcov.param <- c("Sigma_var","Sigma_cov","Psi_var","Psi_cov","Lambda","B")    
    dmu <- list()
    dOmega <- list()
    dLambda <- list()
    dB <- list()
    dPsi <- list()

    toUpdate <- stats::setNames(vector(mode = "logical", n.param),name.originalLink)
    
    ### ** Compute derivative or prepare for the derivative
    for(iName in name.param){ # iName <- name.param[1]

        iName2 <- as.character(param2originalLink[iName])
        iType <- unique(subset(df.param, subset = (param == iName), select = "detail", drop = TRUE))
        iY <- subset(df.param, subset = param %in% iName, select = Y, drop = TRUE)
        iX <- subset(df.param, subset = param %in% iName, select = X, drop = TRUE)

        ## *** derivative regarding the mean        
        if(iType %in% mean.param){            
            if(iType=="nu"){
                dmu[[iName2]] <- matrix(as.numeric(name.endogenous %in% iY),
                                        nrow = n.data, ncol = n.endogenous, byrow = TRUE,
                                        dimnames = list(NULL, name.endogenous))
                toUpdate[iName2] <- FALSE
            }else if(iType=="K"){
                dmu[[iName2]] <- matrix(0, nrow = n.data, ncol = n.endogenous, byrow = TRUE,
                                        dimnames = list(NULL, name.endogenous))
                for(Y.tempo in unique(iY)){                    
                    dmu[[iName2]][,Y.tempo] <- rowSums(data[,iX[iY == Y.tempo],drop=FALSE])
                }
                toUpdate[iName2] <- FALSE
            }else if(iType=="alpha"){
                dmu[[iName2]] <- matrix(as.numeric(name.latent %in% unique(iY)), nrow = n.data, ncol = n.latent, byrow = TRUE,
                                        dimnames = list(NULL, name.latent))                
                toUpdate[iName2] <- TRUE
            }else if(iType=="Gamma"){
                dmu[[iName2]] <- matrix(0, nrow = n.data, ncol = n.latent, byrow = TRUE,
                                        dimnames = list(NULL, name.latent))
                for(Y.tempo in unique(iY)){ # Y.tempo <- "eta"
                    dmu[[iName2]][,Y.tempo] <- rowSums(data[,iX[iY == Y.tempo],drop=FALSE])
                }
                toUpdate[iName2] <- TRUE
            }
        }
        
        ## *** derivative regarding the residual variance covariance
        if(iType %in% vcov.param){
            
            if(iType=="Sigma_var"){
                dOmega[[iName2]] <- matrix(0,
                                           nrow = n.endogenous, ncol = n.endogenous, byrow = TRUE,
                                           dimnames = list(name.endogenous, name.endogenous))
                dOmega[[iName2]][match(iX, name.endogenous) + (match(iY, name.endogenous) - 1) * n.endogenous] <- 1
                toUpdate[iName2] <- FALSE
            }else if(iType=="Sigma_cov"){
                dOmega[[iName2]] <- matrix(0,
                                           nrow = n.endogenous, ncol = n.endogenous, byrow = TRUE,
                                           dimnames = list(name.endogenous, name.endogenous))
                dOmega[[iName2]][match(iX, name.endogenous) + (match(iY, name.endogenous) - 1) * n.endogenous] <- 1
                dOmega[[iName2]][match(iY, name.endogenous) + (match(iX, name.endogenous) - 1) * n.endogenous] <- 1
                toUpdate[iName2] <- FALSE
            }
            
        }        

        ## *** matrices
        if(iType=="Lambda"){            
            dLambda[[iName2]] <- matrix(0,
                                        nrow = n.latent, ncol = n.endogenous, byrow = TRUE,
                                        dimnames = list(name.latent, name.endogenous))
            dLambda[[iName2]][match(iX, name.latent) + (match(iY, name.endogenous) - 1) * n.latent] <- 1            
            toUpdate[iName2] <- TRUE
        }else if(iType=="B"){
            dB[[iName2]] <- matrix(0,
                                   nrow = n.latent, ncol = n.latent, byrow = TRUE,
                                   dimnames = list(name.latent, name.latent))
            dB[[iName2]][match(iX, name.latent) + (match(iY, name.latent) - 1) * n.latent] <- 1
            toUpdate[iName2] <- TRUE
        }else if(iType=="Psi_var"){
            dPsi[[iName2]] <- matrix(0,
                                     nrow = n.latent, ncol = n.latent, byrow = TRUE,
                                     dimnames = list(name.latent, name.latent))
            dPsi[[iName2]][match(iX, name.latent) + (match(iY, name.latent) - 1) * n.latent] <- 1
            toUpdate[iName2] <- TRUE
        }else if(iType=="Psi_cov"){
            dPsi[[iName2]] <- matrix(0,
                                     nrow = n.latent, ncol = n.latent, byrow = TRUE,
                                     dimnames = list(name.latent, name.latent))
            dPsi[[iName2]][match(iX, name.latent) + (match(iY, name.latent) - 1) * n.latent] <- 1
            dPsi[[iName2]][match(iY, name.latent) + (match(iX, name.latent) - 1) * n.latent] <- 1            
            toUpdate[iName2] <- TRUE
        } 
    }
### ** export
    return(list(
        dmu = dmu,
        dOmega = dOmega,
        dLambda = dLambda,
        dB = dB,
        dPsi = dPsi,
        toUpdate = toUpdate
    ))
}


## * skeletonDtheta.lvmfit
#' @rdname skeleton
skeletonDtheta.lvmfit <- function(object, name.endogenous, name.latent, ...){

### ** Import information
    n.endogenous <- length(name.endogenous)
    n.latent <- length(name.latent)

    ## from Moment
    type <- object$conditionalMoment$skeleton$type
    iIB.Lambda <- object$conditionalMoment$value$iIB.Lambda
    alpha.XGamma.iIB <- object$conditionalMoment$value$alpha.XGamma.iIB
    tLambda.tiIB.Psi.iIB <- object$conditionalMoment$value$tLambda.tiIB.Psi.iIB

    ## from dMoment.init
    dmu <- object$conditionalMoment$dMoment.init$dmu
    dOmega <- object$conditionalMoment$dMoment.init$dOmega
    dLambda <- object$conditionalMoment$dMoment.init$dLambda
    dB <- object$conditionalMoment$dMoment.init$dB
    dPsi <- object$conditionalMoment$dMoment.init$dPsi
    toUpdate <- object$conditionalMoment$dMoment.init$toUpdate
    name2Update <- names(toUpdate)
    type2Update <- type[name2Update]
    
### ** Update partial derivatives

    ## *** mean coefficients
    type2Update.meanparam <- type2Update[type2Update %in% c("alpha","Lambda","Gamma","B")]
    name2Update.meanparam <- names(type2Update.meanparam)
    n2Update.meanparam <- length(name2Update.meanparam)
        
    if(n2Update.meanparam>0){
        for(iP in 1:n2Update.meanparam){ # iP <- 1
            iType <- type2Update.meanparam[iP]
            iName <- name2Update.meanparam[iP]
            
            if(iType == "alpha"){
                dmu[[iName]] <- dmu[[iName]] %*% iIB.Lambda
            }else if(iType == "Gamma"){
                dmu[[iName]] <- dmu[[iName]] %*% iIB.Lambda 
            }else if(iType == "Lambda"){
                dmu[[iName]] <- alpha.XGamma.iIB %*% dLambda[[iName]]
            }else if(iType == "B"){
                dmu[[iName]] <- alpha.XGamma.iIB %*% dB[[iName]] %*% iIB.Lambda
            }

            colnames(dmu[[iName]]) <- name.endogenous
        }
    }

    ## *** variance-covariance coefficients
    type2Update.vcovparam <- type2Update[type2Update %in% c("Psi_var","Psi_cov","Lambda","B")]
    name2Update.vcovparam <- names(type2Update.vcovparam)
    n2Update.vcovparam <- length(name2Update.vcovparam)

    if(n2Update.vcovparam>0){
        for(iP in 1:n2Update.vcovparam){ # iP <- 1
            iType <- type2Update.vcovparam[iP]
            iName <- name2Update.vcovparam[iP]
        
            if(iType %in% "Psi_var"){
                dOmega[[iName]] <-  t(iIB.Lambda) %*% dPsi[[iName]] %*% iIB.Lambda
            }else if(iType %in% "Psi_cov"){
                dOmega[[iName]] <-  t(iIB.Lambda) %*% dPsi[[iName]] %*% iIB.Lambda
            }else if(iType == "Lambda"){
                dOmega[[iName]] <- tLambda.tiIB.Psi.iIB %*% dLambda[[iName]]
                dOmega[[iName]] <- dOmega[[iName]] + t(dOmega[[iName]])
            }else if(iType == "B"){
                dOmega[[iName]] <- tLambda.tiIB.Psi.iIB %*% dB[[iName]] %*% iIB.Lambda
                dOmega[[iName]] <- dOmega[[iName]] + t(dOmega[[iName]])
            }

            colnames(dOmega[[iName]]) <- name.endogenous
            rownames(dOmega[[iName]]) <- name.endogenous
        }
    }

### ** Export
    return(list(dmu = dmu, dOmega = dOmega))

}

## * skeletonDtheta2
#' @rdname skeleton
`skeletonDtheta2` <-
    function(object, ...) UseMethod("skeletonDtheta2")

## * skeletonDtheta2.gls
#' @rdname skeleton
skeletonDtheta2.gls <- function(object, dOmega = NULL,
                                class.cor = NULL, class.var = NULL,
                                M.corcoef = NULL, n.endogenous = NULL,
                                index.lower.tri = NULL, indexArr.lower.tri = NULL,
                                var.coef = NULL, name.otherVar = NULL, name.varcoef = NULL, n.varcoef = NULL,
                                cor.coef = NULL, name.corcoef = NULL,
                                ...){

    ## ** import information
    if(is.null(dOmega)){
        dOmega <- object$conditionalMoment$dOmega
    }
    if(is.null(class.cor)){
        class.cor <- object$conditionalMoment$skeleton$class.cor
    }
    if(is.null(class.var)){
        class.var <- object$conditionalMoment$skeleton$class.var
    }
    if(is.null(M.corcoef)){
        M.corcoef <- object$conditionalMoment$skeleton$M.corcoef
    }
    if(is.null(n.endogenous)){
        n.endogenous <- object$conditionalMoment$skeleton$n.endogenous
    }
    if(is.null(index.lower.tri)){
        index.lower.tri <- object$conditionalMoment$skeleton$index.lower.tri
    }
    if(is.null(indexArr.lower.tri)){
        indexArr.lower.tri <- object$conditionalMoment$skeleton$indexArr.lower.tri
    }
    if(is.null(var.coef)){
        var.coef <- object$conditionalMoment$skeleton$var.coef
    }
    if(is.null(name.varcoef)){
       name.varcoef <- object$conditionalMoment$skeleton$name.varcoef
    }
    if(is.null(name.otherVar)){
        name.otherVar <- object$conditionalMoment$skeleton$name.otherVar
    }
    if(is.null(n.varcoef)){
       n.varcoef <- object$conditionalMoment$skeleton$n.varcoef
    }
    if(is.null(cor.coef)){
        cor.coef <- object$conditionalMoment$skeleton$cor.coef
    }
    if(is.null(name.corcoef)){
       name.corcoef <- object$conditionalMoment$skeleton$name.corcoef
    }
    out <- list(d2Omega = list(), d2mu = NULL)

    ## ** derivative: dispersion parameter with other variance parameter
    if("NULL" %in% class.var == FALSE){
        for(iVar in name.otherVar){ ## iVar <- name.otherVar[1]
            out$d2Omega[["sigma2"]][[iVar]] <- dOmega[[iVar]]/var.coef["sigma2"]
        }
    }

    ## ** derivative: dispersion parameter with correlation parameters
    if("NULL" %in% class.cor == FALSE){
        for(iVar in name.corcoef){
            out$d2Omega[["sigma2"]][[iVar]] <- dOmega[[iVar]]/var.coef["sigma2"]
        }
    }

    ## ** derivative: correlation parameter with other variance parameters
    if("NULL" %in% class.var == FALSE && "NULL" %in% class.cor == FALSE){
        M.corvalue <- matrix(1, nrow = n.endogenous, ncol = n.endogenous)
        M.corvalue[index.lower.tri] <- cor.coef[M.corcoef[index.lower.tri]]
        M.corvalue <- symmetrize(M.corvalue, update.upper = TRUE)

            for(iVar1 in name.otherVar){ ## iVar <- name.otherVar[1]

                iIndex.var1 <- which(name.varcoef == iVar1)
                
                ## var var
                for(iVar2 in name.varcoef[iIndex.var1:n.varcoef]){

                    ##
                    M.tempo <- c(1,-1)[(iVar1==iVar2)+1] * dOmega[[iVar1]]/(2*var.coef[iVar2])

                    ## remove null derivative on the diagonal
                    diag(M.tempo) <- 0

                    ## remove null derivative outside the diagonal
                    iIndex.var2 <- which(name.varcoef == iVar2)
                    
                    index0 <- union(which(rowSums(indexArr.lower.tri==iIndex.var1)==0),
                                    which(rowSums(indexArr.lower.tri==iIndex.var2)==0))
                    M.tempo[index.lower.tri[index0]] <- 0
                    M.tempo <- symmetrize(M.tempo, update.upper = TRUE)

                    out$d2Omega[[iVar1]][[iVar2]] <- M.tempo
                }                
                
                ## var cor
                for(iVar2 in name.corcoef){                    
                    M.tempo <- dOmega[[iVar1]]/M.corvalue
                    M.tempo[M.corcoef!=iVar2] <- 0
                    if(any(M.tempo!=0)){
                        out$d2Omega[[iVar1]][[iVar2]] <- M.tempo
                    }
                }

            }
    }

    return(out)
}

## * skeletonDtheta2.lme
#' @rdname skeleton
skeletonDtheta2.lme <- skeletonDtheta2.gls

## * skeletonDtheta2.lm
#' @rdname skeleton
skeletonDtheta2.lm <- function(object, ...){
    return(list(d2mu = NULL, d2Omega = NULL))    
}

## * skeletonDtheta2.lvm
#' @rdname skeleton
skeletonDtheta2.lvm <- function(object, data, df.param.all,
                                param2originalLink, name.latent, ...){

    detail <- factitious <- marginal <- param <- value <- Y <- NULL ## [:for CRAN check] subset
    
    df.param <- subset(df.param.all, is.na(value) & marginal == FALSE & factitious == FALSE)
    dfred.param <- subset(df.param, subset = !duplicated(param))
    
    n.latent <- length(name.latent)
    n.data <- NROW(data)

### ** identify all combinations of coefficients with second derivative
    grid.mean <- list()

    grid.mean$alpha.B <- .combinationDF(dfred.param,
                                        detail1 = "alpha", name1 = "alpha",
                                        detail2 = "B", name2 = "B")

    grid.mean$alpha.Lambda <- .combinationDF(dfred.param,
                                             detail1 = "alpha", name1 = "alpha",
                                             detail2 = "Lambda", name2 = "Lambda")

    grid.mean$Gamma.B <- .combinationDF(dfred.param,
                                        detail1 = "Gamma", name1 = "Gamma",
                                        detail2 = "B", name2 = "B")

    grid.mean$Gamma.Lambda <- .combinationDF(dfred.param,
                                             detail1 = "Gamma", name1 = "Gamma",
                                             detail2 = "Lambda", name2 = "Lambda")
    
    grid.mean$Lambda.B <- .combinationDF(dfred.param,
                                        detail1 = "Lambda", name1 = "Lambda",
                                        detail2 = "B", name2 = "B")

    grid.mean$B.B <- .combinationDF(dfred.param,
                                    detail1 = "B", name1 = "B1",
                                    detail2 = "B", name2 = "B2")

    n.mean <- lapply(grid.mean, NROW)
    

    grid.vcov <- list()
    
    grid.vcov$Psi.Lambda <- .combinationDF(dfred.param,
                                           detail1 = c("Psi_var","Psi_cov"), name1 = "Psi",
                                           detail2 = "Lambda", name2 = "Lambda")

    grid.vcov$Psi.B <- .combinationDF(dfred.param,
                                      detail1 = c("Psi_var","Psi_cov"), name1 = "Psi",
                                      detail2 = "B", name2 = "B")

    grid.vcov$Lambda.B <- .combinationDF(dfred.param,
                                         detail1 = "Lambda", name1 = "Lambda",
                                         detail2 = "B", name2 = "B")

    grid.vcov$Lambda.Lambda <- .combinationDF(dfred.param,
                                              detail1 = "Lambda", name1 = "Lambda1",
                                              detail2 = "Lambda", name2 = "Lambda2")

    grid.vcov$B.B <- .combinationDF(dfred.param,
                                    detail1 = "B", name1 = "B1",
                                    detail2 = "B", name2 = "B2")
    
    n.vcov <- lapply(grid.vcov, NROW)
    
### ** convert back to lava names
    grid.mean <- lapply(grid.mean, function(x){ ## x <- grid.mean[[2]]
        if(length(x)>0){
            x[,1] <- param2originalLink[x[,1]]
            x[,2] <- param2originalLink[x[,2]]
        }
        return(x)
    })

    grid.vcov <- lapply(grid.vcov, function(x){ ## x <- grid.vcov[[2]]
        if(length(x)>0){
            x[,1] <- param2originalLink[x[,1]]
            x[,2] <- param2originalLink[x[,2]]
        }
        return(x)
    })

### ** prepare export
    if(any(unlist(n.mean)>0)){
        xx <- lapply(grid.mean, function(x){
            if(NROW(x)>0){
                colnames(x) <- c("x","y")
            }
            return(x)
        })
        collapseGrid <- do.call(rbind, xx)
        name.tempo <- as.character(unique(collapseGrid[[1]]))
        d2mu <- lapply(name.tempo, function(x){
            iIndex <- which(collapseGrid[[1]]==x)
            v <- vector(mode = "list", length(iIndex))
            names(v) <- collapseGrid[[2]][iIndex]
            return(v)
        })
        names(d2mu) <- name.tempo
    }else{
        d2mu <- list()
    }
    
    if(any(unlist(n.vcov)>0)){
        xx <- lapply(grid.vcov, function(x){
            if(NROW(x)>0){
                colnames(x) <- c("x","y")
            }
            return(x)
        })
        collapseGrid <- do.call(rbind, xx)
        name.tempo <- as.character(unique(collapseGrid[[1]]))
        d2Omega <- lapply(name.tempo, function(x){
            iIndex <- which(collapseGrid[[1]]==x)
            v <- vector(mode = "list", length(iIndex))
            names(v) <- collapseGrid[[2]][iIndex]
            return(v)
        })
        names(d2Omega) <- name.tempo
    }else{
        d2Omega <- list()
    }
    
    ## ** prepare alpha.B and alpha.Lambda
    if(any(df.param$detail == "alpha")){
        name.alpha <- subset(df.param, subset = !duplicated(param) & detail == "alpha", select = "param", drop = TRUE)
        ls.Malpha <- list()
        for(iName in name.alpha){ # iName <- name.alpha[1]

            iParam <- df.param[df.param$name == iName, "param"]
            iY <- subset(df.param, subset = param %in% iParam, select = Y, drop = TRUE)
            ls.Malpha[[iName]] <- matrix(as.numeric(name.latent %in% unique(iY)),
                                         nrow = n.data, ncol = n.latent, byrow = TRUE,
                                         dimnames = list(NULL, name.latent))
            
        }
    }
    if(n.mean$alpha.B>0){
        for(iP in 1:n.mean$alpha.B){ ## iP <- 1
            iName1 <- grid.mean$alpha.B[iP,"alpha"]
            iName2 <- grid.mean$alpha.B[iP,"B"]
            
            d2mu[[iName1]][[iName2]] <- ls.Malpha[[iName1]]
        }
    }
    if(n.mean$alpha.Lambda>0){
        for(iP in 1:n.mean$alpha.Lambda){ ## iP <- 1
            iName1 <- grid.mean$alpha.Lambda[iP,"alpha"]
            iName2 <- grid.mean$alpha.Lambda[iP,"Lambda"]
            
            d2mu[[iName1]][[iName2]] <- ls.Malpha[[iName1]]
        }
    }
    
    ## ** Store X for Gamma
    if(n.mean$Gamma.Lambda>0){
        for(iP in 1:n.mean$Gamma.Lambda){ ## iP <- 1
            iName1 <- grid.mean$Gamma.Lambda[iP,"Gamma"]
            iName2 <- grid.mean$Gamma.Lambda[iP,"Lambda"]

            iParam <- df.param[df.param$name == iName1, "param"]
            iX <- subset(df.param.all, subset = param %in% iParam, select = "X", drop = TRUE)
            iY <- subset(df.param.all, subset = param %in% iParam, select = "Y", drop = TRUE)

            d2mu[[iName1]][[iName2]] <- matrix(0, nrow = n.data, ncol = n.latent, byrow = TRUE,
                                               dimnames = list(NULL, name.latent))
            for(Y.tempo in unique(iY)){
                d2mu[[iName1]][[iName2]][,Y.tempo] <- rowSums(data[,iX[iY == Y.tempo],drop=FALSE])
            }
        }
    }
    
    if(n.mean$Gamma.B>0){
        for(iP in 1:n.mean$Gamma.B){ ## iP <- 1
            iName1 <- grid.mean$Gamma.B[iP,"Gamma"]
            iName2 <- grid.mean$Gamma.B[iP,"B"]
            
            iParam <- df.param[df.param$name == iName1, "param"]
            iX <- subset(df.param.all, subset = param %in% iParam, select = "X", drop = TRUE)
            iY <- subset(df.param.all, subset = param %in% iParam, select = "Y", drop = TRUE)

            d2mu[[iName1]][[iName2]] <- matrix(0, nrow = n.data, ncol = n.latent, byrow = TRUE,
                                               dimnames = list(NULL, name.latent))
            for(Y.tempo in unique(iY)){
                d2mu[[iName1]][[iName2]][,Y.tempo] <- rowSums(data[,iX[iY == Y.tempo],drop=FALSE])
            }
        }
    }

### ** export
    toUpdate <- unlist(lapply(c(n.mean,n.vcov), function(x){x>0}))
    return(list(grid.mean = grid.mean,
                n.mean = n.mean,                
                grid.vcov = grid.vcov,
                n.vcov = n.vcov,
                d2mu = d2mu,
                d2Omega = d2Omega,
                toUpdate = toUpdate
                ))
}

## * skeletonDtheta2.lvmfit
#' @rdname skeleton
skeletonDtheta2.lvmfit <- function(object, ...){
    
### ** Import information
    n.endogenous <- NCOL(object$conditionalMoment$Omega)

    ## from Moment
    Psi <- object$conditionalMoment$value$Psi
    Lambda <- object$conditionalMoment$value$Lambda
    iIB <- object$conditionalMoment$value$iIB
    Psi.iIB <- object$conditionalMoment$value$Psi.iIB
    iIB.Lambda <- object$conditionalMoment$value$iIB.Lambda
    alpha.XGamma.iIB <- object$conditionalMoment$value$alpha.XGamma.iIB
    type <- object$conditionalMoment$skeleton$type

    ## from dMoment.init
    dLambda <- object$conditionalMoment$dMoment.init$dLambda
    dB <- object$conditionalMoment$dMoment.init$dB
    dPsi <- object$conditionalMoment$dMoment.init$dPsi
    
    ## from d2Moment.init
    d2mu <- object$conditionalMoment$d2Moment.init$d2mu
    d2Omega <- object$conditionalMoment$d2Moment.init$d2Omega

    grid.mean <- object$conditionalMoment$d2Moment.init$grid.mean
    grid.vcov <- object$conditionalMoment$d2Moment.init$grid.vcov

    n.mean <- object$conditionalMoment$d2Moment.init$n.mean
    n.vcov <- object$conditionalMoment$d2Moment.init$n.vcov

    toUpdate <- object$conditionalMoment$d2Moment.init$toUpdate
    ##    names(object$conditionalMoment$d2Moment)
    
### ** second order partial derivatives
    if(any(toUpdate)){
        
        ## *** mean coefficients        
        if(toUpdate["alpha.B"]){
            for(iP in 1:n.mean$alpha.B){ # iP <- 1
                iName1 <- grid.mean$alpha.B[iP,"alpha"]
                iName2 <- grid.mean$alpha.B[iP,"B"]

                d2mu[[iName1]][[iName2]] <- d2mu[[iName1]][[iName2]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
            }
        }
        
        if(toUpdate["alpha.Lambda"]){
            for(iP in 1:n.mean$alpha.Lambda){ # iP <- 1
                iName1 <- grid.mean$alpha.Lambda[iP,"alpha"]
                iName2 <- grid.mean$alpha.Lambda[iP,"Lambda"]

                d2mu[[iName1]][[iName2]] <- d2mu[[iName1]][[iName2]] %*% iIB %*% dLambda[[iName2]]
                
            }
        }

        if(toUpdate["Gamma.B"]){
            for(iP in 1:n.mean$Gamma.B){ # iP <- 1
                iName1 <- grid.mean$Gamma.B[iP,"Gamma"]
                iName2 <- grid.mean$Gamma.B[iP,"B"]

                d2mu[[iName1]][[iName2]] <- d2mu[[iName1]][[iName2]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
            }
        }        

        if(toUpdate["Gamma.Lambda"]){
            for(iP in 1:n.mean$Gamma.Lambda){ # iP <- 1
                iName1 <- grid.mean$Gamma.Lambda[iP,"Gamma"]
                iName2 <- grid.mean$Gamma.Lambda[iP,"Lambda"]                

                d2mu[[iName1]][[iName2]] <- d2mu[[iName1]][[iName2]] %*% iIB %*% dLambda[[iName2]]
            }
        }        

        if(toUpdate["Lambda.B"]){
            for(iP in 1:n.mean$Lambda.B){ # iP <- 1
                iName1 <- grid.mean$Lambda.B[iP,"Lambda"]
                iName2 <- grid.mean$Lambda.B[iP,"B"]

                d2mu[[iName1]][[iName2]] <- alpha.XGamma.iIB %*% dB[[iName2]] %*% iIB %*% dLambda[[iName1]]
            }
        }

        if(toUpdate["B.B"]){
            for(iP in 1:n.mean$B.B){ # iP <- 1
                iName1 <- grid.mean$B.B[iP,"B1"]
                iName2 <- grid.mean$B.B[iP,"B2"]

                term1 <- alpha.XGamma.iIB %*% dB[[iName2]] %*% iIB %*% dB[[iName1]] %*% iIB.Lambda
                term2 <- alpha.XGamma.iIB %*% dB[[iName1]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
                d2mu[[iName1]][[iName2]] <- term1 + term2
            }
        }

        ## *** variance-covariance coefficients
        if(toUpdate["Psi.Lambda"]){
            for(iP in 1:n.vcov$Psi.Lambda){ # iP <- 1
                iName1 <- grid.vcov$Psi.Lambda[iP,"Psi"]
                iName2 <- grid.vcov$Psi.Lambda[iP,"Lambda"]

                term1 <- t(dLambda[[iName2]]) %*% t(iIB) %*% dPsi[[iName1]] %*% iIB.Lambda                
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
            }
        }

        if(toUpdate["Psi.B"]){
            for(iP in 1:n.vcov$Psi.B){ # iP <- 1
                iName1 <- grid.vcov$Psi.B[iP,"Psi"]
                iName2 <- grid.vcov$Psi.B[iP,"B"]

                term1 <- t(iIB.Lambda) %*% t(dB[[iName2]]) %*% t(iIB) %*% dPsi[[iName1]] %*% iIB.Lambda
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
            }
        }

        if(toUpdate["Lambda.B"]){
            for(iP in 1:n.vcov$Lambda.B){ # iP <- 1
                iName1 <- grid.vcov$Lambda.B[iP,"Lambda"]
                iName2 <- grid.vcov$Lambda.B[iP,"B"]

                term1 <- t(dLambda[[iName1]]) %*% t(iIB) %*% t(dB[[iName2]]) %*% t(iIB) %*% Psi %*% iIB.Lambda
                term2 <- t(dLambda[[iName1]]) %*% t(iIB) %*% Psi %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
                ## term2 <- tLambda.tiIB.Psi.iIB %*% dB[[iName2]] %*% iIB %*% dLambda[[iName1]]                
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1) + term2 + t(term2)
            }
        }

        if(toUpdate["Lambda.Lambda"]){
            for(iP in 1:n.vcov$Lambda.Lambda){ # iP <- 1
                iName1 <- grid.vcov$Lambda.Lambda[iP,"Lambda1"]
                iName2 <- grid.vcov$Lambda.Lambda[iP,"Lambda2"]
                
                term1 <- t(dLambda[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% dLambda[[iName2]]
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
            }
        }

        if(toUpdate["B.B"]){
            for(iP in 1:n.vcov$B.B){ # iP <- 1
                iName1 <- grid.vcov$B.B[iP,"B1"]
                iName2 <- grid.vcov$B.B[iP,"B2"]

                term1 <- t(iIB.Lambda) %*% t(dB[[iName2]]) %*% t(iIB) %*% t(dB[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% Lambda
                term2 <- t(iIB.Lambda) %*% t(dB[[iName1]]) %*% t(iIB) %*% t(dB[[iName2]]) %*% t(iIB) %*% Psi.iIB %*% Lambda
                term3 <- t(iIB.Lambda) %*% t(dB[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% dB[[iName2]] %*% iIB %*% Lambda
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1) + term2 + t(term2) + term3 + t(term3)
            }
        }

    }

### ** Export
    return(list(d2mu = d2mu, d2Omega = d2Omega))

}

## * .combination
#' @title Form all Unique Combinations Between two Vectors
#' @description Form all unique combinations between two vectors (removing symmetric combinations).
#' @name combination
#'
#' @param ... [vectors] elements to be combined.
#'
#' @return A matrix, each row being a different combination.
#' 
#' @examples
#' .combination <- lavaSearch2:::.combination
#' 
#' .combination(1,1)
#' .combination(1:2,1:2)
#' .combination(c(1:2,1:2),1:2)
#' 
#' .combination(alpha = 1:2, beta = 3:4)
#'
#' @keywords internal
.combination <- function(...){

    ## ** normalize arguments
    dots <- list(...)
    if(length(dots)!=2){
        stop("can only handle two vectors \n")
    }
    test.null <- unlist(lapply(dots,is.null))    
    if(any(test.null)){
        return(NULL)
    }
    dots <- lapply(dots,unique)

    ## ** form all combinations
    grid <- expand.grid(dots, stringsAsFactors = FALSE) 
    
    ## ** remove combinations (b,a) when (a,b) is already there
    name1 <- paste0(grid[,1],grid[,2])
    name2 <- paste0(grid[,2],grid[,1])

    if(NROW(grid)>1 && any(name1 %in% name2)){ 

        n.grid <- NROW(grid)
        test.duplicated <- c(FALSE,sapply(2:n.grid, function(iG){
            any(name2[iG] %in% name1[1:(iG-1)]) ## find duplicates
        }))

        grid <- grid[test.duplicated==FALSE,]
    }

    ## ** export
    return(grid)        
}


## * .combinationDF
.combinationDF <- function(data,
                           detail1, detail2,
                           name1, name2){

    detail <- NULL # [:for CRAN check] subset
    
    if(any(detail1 %in% data$detail) && any(detail2 %in% data$detail) ){
        ls.args <- list(subset(data, subset = detail %in% detail1, select = "param", drop = TRUE),
                        subset(data, subset = detail %in% detail2, select = "param", drop = TRUE))
        names(ls.args) <- c(name1,name2)
    
        return(do.call(.combination, args = ls.args))
        
    }else{
        
        return(numeric(0))
        
    }
}



##----------------------------------------------------------------------
### skeleton.R ends here

### leverage2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 19 2018 (17:58) 
## Version: 
## Last-Updated: jan 23 2024 (10:25) 
##           By: Brice Ozenne
##     Update #: 178
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * documentation - leverage2
#' @title Leverage With Small Sample Correction.
#' @description Extract leverage values from a latent variable model, with small sample correction. 
#' @name leverage2
#' 
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param format [character] Use \code{"wide"} to return the residuals in the wide format (one row relative to each sample).
#' Otherwise use \code{"long"} to return the residuals in the long format.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details The leverage are defined as the partial derivative of the fitted values with respect to the observations.
#' \deqn{
#' leverage_i = \frac{\partial \hat{Y}_i}{\partial Y_i}
#' }
#' See Wei et al. (1998). \cr \cr
#' 
#' When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the leverage.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#' 
#' @return a matrix containing the leverage relative to each sample (in rows)
#' and each endogenous variable (in column).
#'
#' @references Bo-Cheng Wei et al., Generalized Leverage and its applications (1998), Scandinavian Journal of Statistics 25:1:25-37.
#' 
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' m <- lvm(Y1~eta,Y2~eta,Y3~eta)
#' latent(m) <- ~eta
#' d <- lava::sim(m,20, latent = FALSE)
#'
#' #### latent variable models ####
#' e.lvm <- estimate(m, data = d)
#' leverage2(e.lvm)
#' 
#' @concept estimator
#' @keywords smallSampleCorrection
#' 
#' @export
`leverage2` <-
    function(object, format, ssc, ...) UseMethod("leverage2")

## * leverage2.lvmfit
#' @rdname leverage2
#' @export
leverage2.lvmfit <- function(object, format = "wide", ssc = lava.options()$ssc, ...){

    return(leverage2(estimate2(object, ssc = ssc, ...), format = format))

}

## * leverage2.lvmfit2
#' @rdname leverage2
#' @export
leverage2.lvmfit2 <- function(object, format = "wide", ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    format <- match.arg(format, choices = c("long","wide"))

    if(format == "wide"){
        return(object$sCorrect$leverage)
    }else if(format == "long"){
        endogenous <- colnames(object$sCorrect$leverage)
        n.endogenous <- length(endogenous)
        
        outW <- data.frame(cluster = 1:NROW(object$sCorrect$leverage), object$sCorrect$leverage)
        outL <- stats::na.omit(stats::reshape(outW,
                                              idvar = "id",
                                              direction = "long",
                                              varying = list(endogenous),
                                              timevar = "endogenous",
                                              v.names = "leverage"))
        rownames(outL) <- NULL

        
        outL$endogenous <- factor(outL$endogenous, levels = 1:n.endogenous, labels = endogenous)
        reorder <- match(interaction(object$sCorrect$old2new.order$XXclusterXX.old,object$sCorrect$old2new.order$XXendogenousXX.old),
                         interaction(outL$cluster,outL$endogenous))
        return(outL[reorder,])
    }

}

## * .leverage2
.leverage2 <- function(Omega, epsilon, dmu, dOmega, vcov.param,
                       name.pattern, missing.pattern, unique.pattern,
                       endogenous, n.endogenous, param, param.mean, param.var, n.cluster){

    n.pattern <- NROW(unique.pattern)
    n.param <- length(param)
    leverage <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                       dimnames = list(NULL, endogenous))
    if(length(param.mean)==0){
        leverage[] <- 0
        return(leverage)
    }
    if(is.null(vcov.param)){

        stop("Cannot compute the leverage values without the variance-covariance matrix of the coefficients. \n")

    }

    scoreY <- array(0, dim = c(n.cluster, n.endogenous, n.param),
                    dimnames = list(NULL, endogenous, param))
    
    for(iP in 1:n.pattern){ ## iP <- 1 
        iIndex <- missing.pattern[[iP]]
        iY <- which(unique.pattern[iP,]==1)
        
        iOmega <- Omega[iY,iY,drop=FALSE]
        iOmegaM1 <- chol2inv(chol(iOmega))
        iOmegaM1.epsilon <- epsilon[iIndex,iY,drop=FALSE] %*% iOmegaM1
            
        ## derivative of the score regarding Y
        for(iParam in param){
            
            if(iParam %in% param.mean){
                if(length(iY)>1){
                    scoreY[iIndex,iY,iParam] <- scoreY[iIndex,iY,iParam] + t(dmu[iParam,iY,iIndex]) %*% iOmegaM1 
                }else{
                    scoreY[iIndex,iY,iParam] <- scoreY[iIndex,iY,iParam] + dmu[iParam,iY,iIndex] * iOmegaM1[1,1] 
                }
            }
            if(iParam %in% param.var){
                scoreY[iIndex,iY,iParam] <- scoreY[iIndex,iY,iParam] + 2 * iOmegaM1.epsilon %*% dOmega[[iParam]][iY,iY,drop=FALSE] %*% iOmegaM1
            }

        

        }

        ## leverage
        for(iiY in iY){ ## iiY <- iY[2]
            
            if(length(param.mean)==1){
                leverage[iIndex,iiY] <- dmu[param.mean,iiY,iIndex] * (scoreY[iIndex,iiY,] %*% vcov.param)[,param.mean]
            }else if(n.param==1){
                leverage[iIndex,iiY] <- dmu[param.mean,iiY,iIndex] * vcov.param * scoreY[iIndex,iiY,]
            } else{
                leverage[iIndex,iiY] <- rowSums(t(dmu[param.mean,iiY,iIndex]) * (scoreY[iIndex,iiY,] %*% vcov.param)[,param.mean,drop=FALSE] )
            }
            ## dmu2 <- matrix(0, nrow = n.param, ncol = length(iIndex), dimnames = list(param,NULL))
            ## dmu2[param.mean,] <- dmu[param.mean,iiY,iIndex]
            ## diag( t(dmu2) %*% vcov.param %*% t(scoreY[iIndex,iiY,]) )
        }
    }
    return(leverage)            
}

##----------------------------------------------------------------------
### leverage2.R ends here

### test1-sCorrect-lava.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  6 2018 (10:40) 
## Version: 
## Last-Updated: apr  4 2018 (14:20) 
##           By: Brice Ozenne
##     Update #: 137
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## source("c:/Users/hpl802/Documents/GitHub/lavaSearch2/tests/testthat/test1-iid2-lava.R")

## * header
rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
.coef2 <- lavaSearch2:::.coef2
library(nlme)
context("sCorrect (replicate lava results)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
           measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * linear regression [lm,gls,lvm]
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d)
e.lm <- lm(Y1~X1+X2+Gender, data = d)
e.gls <- gls(Y1~X1+X2+Gender, data = d, method = "ML")

e2.lvm <- e.lvm
e2.gls <- e.gls
e2.lm <- e.lm

sCorrect(e2.lvm) <- FALSE
sCorrect(e2.gls, cluster = 1:n) <- FALSE
sCorrect(e2.lm) <- FALSE

## ** check score, iid, residuals, vcov, compare2 at ML
test_that("linear regression (at ML) internal consistency",{
    expect_equivalent(e2.lvm$sCorrect$Omega,e2.lm$sCorrect$Omega)
    expect_equivalent(e2.gls$sCorrect$Omega,e2.lm$sCorrect$Omega)

    expect_equivalent(e2.lvm$sCorrect$vcov.param,e2.lm$sCorrect$vcov.param)
    expect_equivalent(e2.gls$sCorrect$vcov.param,e2.lm$sCorrect$vcov.param)

    expect_equivalent(e2.lvm$sCorrect$leverage,e2.lm$sCorrect$leverage)
    expect_equivalent(e2.gls$sCorrect$leverage,e2.lm$sCorrect$leverage)

    expect_equivalent(e2.lvm$sCorrect$score,e2.lm$sCorrect$score)
    expect_equivalent(e2.gls$sCorrect$score,e2.lm$sCorrect$score)

    expect_equivalent(e2.lvm$sCorrect$epsilon,e2.lm$sCorrect$epsilon)
    expect_equivalent(e2.gls$sCorrect$epsilon,e2.lm$sCorrect$epsilon)

    expect_equivalent(e2.lvm$sCorrect$dVcov.param,e2.lm$sCorrect$dVcov.param)
    expect_equivalent(e2.gls$sCorrect$dVcov.param,e2.lm$sCorrect$dVcov.param)    

    expect_equivalent(iid2(e2.gls), iid2(e2.lvm))    
    expect_equivalent(iid2(e2.lm), iid2(e2.lvm))    
})

test_that("linear regression (at ML) compare to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(e2.lvm$sCorrect$n.corrected==e.lvm$data$n)
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))

    GS <- iid(e.lm)
    expect_equivalent(iid2(e2.lvm)[,1:length(colnames(GS))], GS)
    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))

    ## NOTE: iid in lava uses numerical derivative to compute the information matrix
    ## this is why there is not a perfect matching between iid2.lvm and iid.lvm
    
    ## expect_equal(as.double(iid2(e2.lvm)), as.double(iid(e.lvm)))
    
    ## Error: as.double(iid2(e2.lvm)) not equal to as.double(iid(e.lvm)).
    ## 245/250 mismatches (average diff: 1.46e-06)
    ## [1] -0.01755 - -0.01756 ==  6.95e-07
    ## [2] -0.02036 - -0.02037 ==  6.63e-07
    ## [3]  0.04413 -  0.04413 ==  4.34e-07
    ## [4]  0.00615 -  0.00615 == -9.30e-07
    ## [5]  0.00236 -  0.00236 ==  2.20e-07
    ## [6]  0.01338 -  0.01338 ==  7.35e-07
    ## [7] -0.00482 - -0.00482 == -4.08e-08
    ## [8]  0.04877 -  0.04877 ==  3.21e-07
    ## [9] -0.00811 - -0.00811 == -1.35e-06 

    expect_true(all(leverage2(e2.lvm) == 0))

    C1 <- compare2(e2.lvm, par = c("Y1~X1","Y1~X2"))
    C2 <- lava::compare(e.lvm, par = c("Y1~X1","Y1~X2"))
    expect_equal(unname(C1$statistic),
                 unname(C2$statistic/NROW(C1$estimate))
                 )

})

## ** check score not at ML
S1 <- score2(e2.lvm, param = coef(e2.lvm)+1)
S2 <- score2(e.lvm, param = coef(e.lvm)+1, bias.correct = FALSE)

S3 <- score2(e2.lm, param = .coef2(e2.gls)+1) ## not .coef2(e2.lm) because different estimate of the variance
S4 <- score2(e.lm, param = .coef2(e.gls)+1, bias.correct = FALSE)

S5 <- score2(e2.gls, param = .coef2(e2.gls)+1)
S6 <- score2(e.gls, param = .coef2(e.gls)+1, cluster = 1:n, bias.correct = FALSE)

GS <- score(e.lvm, p = coef(e.lvm)+1, indiv = TRUE)

test_that("linear regression (at ML + 1) compare to lava",{
    expect_equal(S1, GS)
    expect_equal(S2, GS)
    
    expect_equal(as.double(S3), as.double(GS))
    expect_equal(as.double(S4), as.double(GS))
    expect_equal(as.double(S5), as.double(GS))
    expect_equal(as.double(S6), as.double(GS))
})

p <- length(coef(e.lvm))
S1 <- score2(e2.lvm, param = coef(e2.lvm)+1:p)
S2 <- score2(e.lvm, param = coef(e.lvm)+1:p, bias.correct = FALSE)

S3 <- score2(e2.lm, param = .coef2(e2.gls)+1:p) ## not .coef2(e2.lm) because different estimate of the variance
S4 <- score2(e.lm, param = .coef2(e.gls)+1:p, bias.correct = FALSE)

S5 <- score2(e2.gls, param = .coef2(e2.gls)+1:p)
S6 <- score2(e.gls, param = .coef2(e.gls)+1:p, cluster = 1:n, bias.correct = FALSE)
GS <- score(e.lvm, p = coef(e.lvm)+1:p, indiv = TRUE)

test_that("linear regression (at ML + 1:p) compare to lava",{
    
    expect_equal(S1, GS)
    expect_equal(S2, GS)
    
    expect_equal(as.double(S3), as.double(GS))
    expect_equal(as.double(S4), as.double(GS))
    expect_equal(as.double(S5), as.double(GS))
    expect_equal(as.double(S6), as.double(GS))
    
})


## * linear regression with constrains [lvm]
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1[0:2]~X1+1*X2), data = d)

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

e.lvm2 <- estimate(lvm(Y1~beta*X1+beta*X2), d)

e2.lvm2 <- e.lvm2
sCorrect(e2.lvm2) <- FALSE

## ** check score, iid, residuals, vcov, compare2 at ML
test_that("linear regression: constrains",{
    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(e2.lvm$sCorrect$n.corrected==e.lvm$data$n)
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))
    
    expect_equal(score2(e.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv = TRUE))
})

## * multiple linear regression [lvm,gls]
## ** model fit and sCorrect
ls.lm <- list(lm(Y1~X1,d),lm(Y2~X2,d),lm(Y3~X1+X3,d))
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

e.lvm2 <- estimate(lvm(Y1[mu:sigma1]~ beta1*X1 + beta2*X2,
                       Y2[mu:sigma2]~ beta1*X1 + beta2*X2,
                       Y3[mu:sigma3]~ beta1*X1 + beta2*X2),
                       data = d)
e.gls <- gls(value ~ X1 + X2,
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             weight = varIdent(form = ~1|variable),
             method = "ML")

test_that("gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm2)), as.double(logLik(e.gls)))
})


e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

e2.lvm2 <- e.lvm2
sCorrect(e2.lvm2) <- FALSE

e2.gls <- e.gls
sCorrect(e2.gls, cluster = "Id") <- FALSE

## ** check score, iid, residuals, vcov, compare2 at ML

test_that("multiple linear regression (at ML) internal consistency",{
    param <- attr(e2.gls$sCorrect$param,"mean.coef")

    expect_equal(unname(e2.lvm2$sCorrect$Omega),
                 unname(e2.gls$sCorrect$Omega),
                 tolerance = 1e-5)
    expect_equal(unname(e2.lvm2$sCorrect$residuals),
                 unname(e2.gls$sCorrect$residuals),
                 tolerance = 1e-5)
    expect_equal(unname(e2.lvm2$sCorrect$score[,c("Y1","Y1~X1","Y1~X2")]),
                 unname(e2.gls$sCorrect$score[,c("(Intercept)","X1","X2")]),
                 tolerance = 1e-5)
})

e2.lvm2$sCorrect$residuals[1:5,]
e2.gls$sCorrect$residuals[1:5,]

test_that("multiple linear regression (at ML) compare to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equal(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))

    GS <- lapply(ls.lm,iid)
    test <- iid2(e2.lvm)
    expect_equivalent(test[,grep("^Y1$|^Y1~X",colnames(test))], GS[[1]])
    expect_equivalent(test[,grep("^Y2$|^Y2~X",colnames(test))], GS[[2]])
    expect_equivalent(test[,grep("^Y3$|^Y3~X",colnames(test))], GS[[3]])
    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))
    expect_true(all(leverage2(e2.lvm) == 0))

    C1 <- compare2(e2.lvm, par = c("Y1~X1","Y2~X2","Y3~X1"))
    C2 <- lava::compare(e.lvm, par = c("Y1~X1","Y2~X2","Y3~X1"))
    expect_equal(unname(C1$statistic),
                 unname(C2$statistic/NROW(C1$estimate))
                 )
})

## ** check score not at ML
test_that("multiple linear regression (at ML + 1) compare to lava",{
    S1 <- score2(e2.lvm, param = coef(e2.lvm)+1)
    S2 <- score2(e.lvm, param = coef(e.lvm)+1, bias.correct = FALSE)

    GS <- score(e.lvm, p = coef(e.lvm)+1, indiv = TRUE)

    expect_equal(S1, GS)
    expect_equal(S2, GS)
})

test_that("multiple linear regression (at ML + 1:p) compare to lava",{
    p <- length(coef(e.lvm))
    S1 <- score2(e2.lvm, param = coef(e2.lvm)+1:p)
    S2 <- score2(e.lvm, param = coef(e.lvm)+1:p, bias.correct = FALSE)

    GS <- score(e.lvm, p = coef(e.lvm)+1:p, indiv = TRUE)
    
    expect_equal(S1, GS)
    expect_equal(S2, GS)
})

## * multiple linear regression with constrains [lvm]
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1~X1+1*X2,Y2~2*X3+2*X1,Y3~X2), data = d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE    

## ** check score, residuals, vcov
test_that("multiple linear regressions: constrains",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equal(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))

    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))
    expect_true(all(leverage2(e2.lvm) == 0))

    C1 <- compare2(e2.lvm, par = c("Y1~X1"))
    C2 <- lava::compare(e.lvm, par = c("Y1~X1"))
    expect_equal(unname(C1$statistic),
                 unname(C2$statistic/NROW(C1$estimate))
                 )
})

## * multiple linear regression with covariance links [lvm]
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1~X1+X2,Y2~X3+X1,Y3~X2,Y1~~Y2),d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score, residuals, vcov at ML
test_that("multiple linear regression, covariance link (at ML)",{
    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equal(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv=TRUE))
    
    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))
    expect_true(all(leverage2(e2.lvm) == 0))
})

## ** check score not at ML
test_that("multiple linear regression, covariance link (not at ML: +1)",{
    expect_equal(score2(e.lvm, p = coef(e.lvm)+1, bias.correct = FALSE),
                 score(e.lvm, p = coef(e.lvm)+1, indiv=TRUE))
})

test_that("multiple linear regression, covariance link (not at ML: +1:p)",{
    newcoef <- coef(e.lvm)+0.1*(1:length(coef(e.lvm)))
    expect_equal(score2(e.lvm, p = newcoef, bias.correct = FALSE),
                 score(e.lvm, p = newcoef, indiv = TRUE))
})

## * mixed model: Compound symmetry [lvm,gls,lme]
## ** model fit and sCorrect
m <- lvm(Y1[mu1:sigma]~1*eta,
         Y2[mu2:sigma]~1*eta,
         Y3[mu3:sigma]~1*eta,
         eta~X1+Gender)
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corCompSymm(form=~ 1|Id),
             data = dLred,
             method = "ML")

test_that("compound symmetry: lme/gls equivalent to lvm",{
    expect_equal(as.double(logLik(e.lme)),as.double(logLik(e.lvm)))
    expect_equal(as.double(logLik(e.gls)),as.double(logLik(e.lvm)))
})

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

e2.lme <- e.lme
sCorrect(e2.lme) <- FALSE

e2.gls <- e.gls
sCorrect(e2.gls) <- FALSE


## ** check score, residuals, vcov at ML
test_that("compound symmetry (at ML) internal consistency",{
    param.nlme <- names(coef(e.gls))
    param.lava <- c("eta","Y2","Y3","eta~X1","eta~GenderFemale")
    
    expect_equal(unname(e2.lvm$sCorrect$Omega),unname(e2.lme$sCorrect$Omega), tol = 1e-5)
    expect_equal(unname(e2.lme$sCorrect$Omega),unname(e2.gls$sCorrect$Omega), tol = 1e-5)

    expect_equal(unname(e2.lvm$sCorrect$vcov.param[param.lava,param.lava]),
                 unname(e2.lme$sCorrect$vcov.param[param.nlme,param.nlme]), tol = 1e-5)
    expect_equal(unname(e2.gls$sCorrect$vcov.param[param.nlme,param.nlme]),
                 unname(e2.lme$sCorrect$vcov.param[param.nlme,param.nlme]), tol = 1e-5)
    
    expect_equal(unname(e2.lvm$sCorrect$score[,param.lava]),
                 unname(e2.lme$sCorrect$score[,param.nlme]), tol = 1e-5)
    expect_equal(unname(e2.gls$sCorrect$score[,param.nlme]),
                 unname(e2.lme$sCorrect$score[,param.nlme]), tol = 1e-5)

    expect_equivalent(e2.lvm$sCorrect$residuals,e2.lme$sCorrect$residuals)
    expect_equivalent(e2.gls$sCorrect$residuals,e2.lme$sCorrect$residuals)

    expect_equivalent(e2.lvm$sCorrect$dVcov.param[param.lava,param.lava,"Y1~~Y1"],
                      e2.lme$sCorrect$dVcov.param[param.nlme,param.nlme,"sigma2"])

    expect_equivalent(iid2(e2.gls)[,param.nlme], iid2(e2.lme)[,param.nlme])    
    expect_equivalent(iid2(e2.lme)[,param.nlme], iid2(e2.lvm)[,param.lava])    
})

test_that("compound symmetry (at ML) compare to lava",{
    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equal(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv = TRUE))
    
    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))
    expect_true(all(leverage2(e2.lvm) == 0))
})


## * mixed model: CS with different variances [lvm,lme]
## ** model fit and sCorrect
m <- lvm(c(Y1[mu1:sigma1]~1*eta,
           Y2[mu2:sigma2]~1*eta,
           Y3[mu3:sigma3]~1*eta,
           eta~X1+Gender))
latent(m) <- ~eta
e.lvm <- estimate(m, d)

e.lme <- nlme::lme(value ~ variable + X1 + Gender,
                   random =~1| Id,
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ variable + X1 + Gender,
                   correlation = corCompSymm(form = ~1| Id),
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

test_that("lme equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    ## gls does not give the same likelihood
    ## expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

e2.lme <- e.lme
sCorrect(e2.lme) <- FALSE

e2.gls <- e.gls
sCorrect(e2.gls) <- FALSE

## ** check score, residuals, vcov at ML
test_that("compound symmetry with weights (at ML) internal consistency",{
    param.nlme <- names(coef(e.gls))
    param.lava <- c("eta","Y2","Y3","eta~X1","eta~GenderFemale")
    
    expect_equal(unname(e2.lvm$sCorrect$Omega),unname(e2.lme$sCorrect$Omega), tol = 1e-5)

    expect_equal(unname(e2.lvm$sCorrect$vcov.param[param.lava,param.lava]),
                 unname(e2.lme$sCorrect$vcov.param[param.nlme,param.nlme]), tol = 1e-5)
    
    expect_equal(unname(e2.lvm$sCorrect$score[,param.lava]),
                 unname(e2.lme$sCorrect$score[,param.nlme]), tol = 1e-5)

    expect_equal(unname(e2.lvm$sCorrect$residuals),
                 unname(e2.lme$sCorrect$residuals), tol = 1e-5)
})

test_that("compound symmetry with weights (at ML) compare to lava",{
    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equal(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv = TRUE))
    
    expect_equal(e2.lvm$sCorrect$residuals, residuals2(e2.lvm))
    expect_true(all(leverage2(e2.lvm) == 0))
})

## * mixed model: Unstructured [lvm,gls,lme]
## ** model fit and sCorrect
m <- lvm(Y1[mu1:sigma]~1*eta,
         Y2[mu2:sigma]~1*eta,
         Y3[mu3:sigma]~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             ## weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             ## weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## e2.lme <- e.lme
## sCorrect(e2.lme) <- FALSE
## does not work because the model is overparametrized

e2.gls <- e.gls
sCorrect(e2.gls) <- FALSE

## ** check score, residuals, vcov at ML
test_that("Unstructured (at ML) internal consistency",{
    param.nlme <- names(coef(e.gls))
    param.lava <- c("eta","Y2","Y3","eta~X1","eta~GenderFemale")

    expect_equal(unname(e2.gls$sCorrect$Omega),unname(e2.lvm$sCorrect$Omega), tol = 1e-4)

    expect_equal(unname(e2.lvm$sCorrect$vcov.param[param.lava,param.lava]),
                 unname(e2.gls$sCorrect$vcov.param[param.nlme,param.nlme]), tol = 1e-4)
    
    expect_equal(unname(e2.lvm$sCorrect$score[,param.lava]),
                 unname(e2.gls$sCorrect$score[,param.nlme]), tol = 1e-4)
    
    expect_equal(unname(e2.lvm$sCorrect$residuals),
                 unname(e2.gls$sCorrect$residuals), tol = 1e-4)
})

test_that("Unstructured (at ML) compare to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))
    
})

## * mixed model: Unstructured with weights [lvm,gls,lme]
## ** model fit and sCorrect
m <- lvm(Y1~1*eta,
         Y2~1*eta,
         Y3~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

e2.lme <- e.lme
## sCorrect(e2.lme) <- FALSE
## does not work because the model is overparametrized

e2.gls <- e.gls
sCorrect(e2.gls) <- FALSE

## ** check score, residuals, vcov at ML
test_that("Unstructured with weights (at ML) internal consistency",{
    param.nlme <- names(coef(e.gls))
    param.lava <- c("eta","Y2","Y3","eta~X1","eta~GenderFemale")

    expect_equal(unname(e2.lvm$sCorrect$Omega),unname(e2.gls$sCorrect$Omega), tol = 1e-3)

    expect_equal(unname(e2.lvm$sCorrect$vcov.param[param.lava,param.lava]),
                 unname(e2.gls$sCorrect$vcov.param[param.nlme,param.nlme]), tol = 1e-3)
    
    expect_equal(unname(e2.lvm$sCorrect$score[,param.lava]),
                 unname(e2.gls$sCorrect$score[,param.nlme]), tol = 1e-3)

    expect_equal(unname(e2.lvm$sCorrect$residuals),
                 unname(e2.gls$sCorrect$residuals), tol = 1e-3)

})

test_that("Unstructured with weights (at ML) compare to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))
    
})

## * LVM: factor model
## ** model fit and sCorrect
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1))
regression(m) <- eta1~X1+X2

e.lvm <- estimate(m,d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score, residuals, vcov at ML
test_that("factor model (at ML) compared to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))

})

## ** check score not at ML
param <- coef(e.lvm)
test <- score2(e.lvm, p = param+1, bias.correct = FALSE)
GS <- score(e.lvm, p = param+1, indiv = TRUE)

test_that("factor model (not at ML: +1)",{
    expect_equal(test, GS)
})

test <- score2(e.lvm, p = param+0.1*(1:length(param)), bias.correct = FALSE)
GS <- score(e.lvm, p = param+0.1*(1:length(param)), indiv=TRUE)
test_that("factor model (not at ML: +1:p)",{
    expect_equal(test, GS)
})

## * LVM: factor model with constrains
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1~1*eta+1*X2,Y2~1*eta,Y3~1*eta),
                  data = d)


e.lvm2 <- estimate(lvm(Y1~1*eta+X2,
                       Y2~lambda*eta+X2,
                       Y3~lambda*eta,
                       eta ~ beta*X2+beta*X1),
                   data = d)

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE
e2.lvm2 <- e.lvm2
sCorrect(e2.lvm2) <- FALSE

## ** check score at ML
test_that("factor model: fixed coefficients",{
    expect_equal(score2(e2.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv = TRUE))

    expect_equivalent(e2.lvm$sCorrect$vcov.param,
                      vcov(e.lvm))
})

test_that("factor model: constrains",{
    expect_equal(score2(e2.lvm2, bias.correct = FALSE),
                 score(e.lvm2, indiv = TRUE))

    expect_equivalent(e2.lvm2$sCorrect$vcov.param,
                      vcov(e.lvm2))
})


## * LVM: 2 factor model
## ** model fit and sCorrect
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(m) <- eta1~X1+X2
latent(m) <- ~eta1+eta2

e.lvm <- estimate(m,d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score, residuals, vcov at ML
test_that("2 factor model (at ML) compared to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))

})


## ** check score not at ML
param <- coef(e.lvm)
test <- score2(e.lvm, p = param+1, bias.correct = FALSE)
GS <- score(e.lvm, p = param+1, indiv = TRUE)

test_that("2 factor model (not at ML: +1)",{
    expect_equal(test, GS)
})

test <- score2(e.lvm, p = param+0.1*(1:length(param)), bias.correct = FALSE)
GS <- score(e.lvm, p = param+0.1*(1:length(param)), indiv=TRUE)
test_that("2 factor model (not at ML: +1:p)",{
    expect_equal(test, GS)
})

## * LVM: 2 factor model with constrains
## ** model fit and sCorrect
m <- lvm(Y1~1*eta1+X2,Y2~lambda*eta1+X2,Y3~lambda*eta1,eta1 ~ beta*X2+beta*X1,
         Z1~0+eta2,Z2~lambda*eta2,Z3~eta2)
e.lvm <- estimate(m, d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score at ML
test_that("2 factor model: constrains",{
    expect_equal(score2(e2.lvm, bias.correct = FALSE),
                 score(e.lvm, indiv = TRUE))

    expect_equivalent(e2.lvm$sCorrect$vcov.param,
                      vcov(e.lvm))
})

## * LVM: 2 factor model (covariance)
## ** model fit and sCorrect
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
covariance(m) <- eta1 ~ eta2
latent(m) <- ~eta1+eta2

e.lvm <- estimate(m,d)
e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score, residuals, vcov at ML
test_that("2 factor model with covariance (at ML) compared to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))

})

## ** check score not at ML
param <- coef(e.lvm)
test <- score2(e.lvm, p = param+1, bias.correct = FALSE)
GS <- score(e.lvm, p = param+1, indiv = TRUE)

test_that("2 factor model with covariance (not at ML: +1)",{
    expect_equal(test, GS)
})

test <- score2(e.lvm, p = param+0.1*(1:length(param)), bias.correct = FALSE)
GS <- score(e.lvm, p = param+0.1*(1:length(param)), indiv=TRUE)
test_that("2 factor model with covariance (not at ML: +1:p)",{
    expect_equal(test, GS)
})

## * LVM: 2 factor model (correlation LV)
## ** model fit and sCorrect
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(m) <- eta2 ~ X1
regression(m) <- eta1 ~ eta2+X2+X3

e.lvm <- estimate(m,d)

e2.lvm <- e.lvm
sCorrect(e2.lvm) <- FALSE

## ** check score, residuals, vcov at ML
test_that("2 factor model with correlation (at ML) compared to lava",{

    expect_equivalent(e2.lvm$sCorrect$vcov.param, vcov(e.lvm))
    expect_true(all(e2.lvm$sCorrect$leverage==0))
    expect_true(all(e2.lvm$sCorrect$n.corrected==e.lvm$data$n))
    expect_equivalent(e2.lvm$sCorrect$residuals, residuals(e.lvm))
    expect_equal(e2.lvm$sCorrect$param, coef(e.lvm))
    expect_equal(e2.lvm$sCorrect$score, score(e.lvm, indiv = TRUE))
    expect_equal(score2(e.lvm, bias.correct = FALSE), score(e.lvm, indiv = TRUE))

})

## ** check score not at ML
param <- coef(e.lvm)
test <- score2(e.lvm, p = param+1, bias.correct = FALSE)
GS <- score(e.lvm, p = param+1, indiv = TRUE)

test_that("2 factor model (not at ML: +1)",{
    expect_equal(test, GS)
})

test <- score2(e.lvm, p = param+0.1*(1:length(param)), bias.correct = FALSE)
GS <- score(e.lvm, p = param+0.1*(1:length(param)), indiv=TRUE)
test_that("2 factor model (not at ML: +1:p)",{
    expect_equal(test, GS)
})


##----------------------------------------------------------------------
### test1-sCorrect-lava.R ends here

### coefType.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (14:38) 
## Version: 
## last-updated: Jan 11 2022 (09:55) 
##           By: Brice Ozenne
##     Update #: 882
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - coefType
#' @title Extract the Type of Each Coefficient
#' @description Extract the type of each coefficient of a \code{lvm} object.
#' @name coefType
#' 
#' @param object a \code{lvm} or \code{lvmfit} object. 
#' @param data [data.frame, optional] the dataset. Help to identify the categorical variables.
#' @param as.lava [logical] export the type of coefficients mimicking \code{lava:::coef}.
#' @param ... arguments to be passed to \code{lava::coef}
#'
#' @details A lvm can be written as a measurement model:
#' \deqn{Y_i = \nu + \Lambda \eta_i + K X_i + \epsilon_i}
#' and a structural model:
#' \deqn{\eta_i = \alpha + B \eta_i + \Gamma X_i + \zeta_i}
#' where \eqn{\Psi}   is the variance covariance matrix of the residuals \eqn{\zeta} \cr
#' and   \eqn{\Sigma} is the variance covariance matrix of the residuals \eqn{\epsilon}. \cr \cr
#'
#' \code{coefType} either returns the Latin/Greek letter corresponding to the coefficients
#' or it groups them:
#' \itemize{
#' \item intercept: \eqn{\nu} and \eqn{\alpha}.
#' \item regression: \eqn{\Lambda}, \eqn{K}, \eqn{B}, and \eqn{\Gamma}.
#' \item covariance: extra-diagonal terms of \eqn{\Sigma} and \eqn{\Psi}.
#' \item variance: diagonal of \eqn{\Sigma} and \eqn{\Psi}.
#' }
#'
#' A link denotes a relationship between two variables.
#' The coefficient are used to represent the strength of the association between two variable, i.e. the strength of a link.
#' A coefficient may corresponds to the strength of one or several link.
#'
#' @return \code{coefType} returns a \code{data.frame} when \code{as.lava=FALSE}:
#' \itemize{
#' \item name: name of the link
#' \item Y: outcome variable
#' \item X: regression variable in the design matrix (could be a transformation of the original variables, e.g. dichotomization).
#' \item data: original variable
#' \item type: type of link
#' \item value: if TRUE, the value of the link is set and not estimated.
#' \item marginal: if TRUE, the value of the link does not impact the estimation.
#' \item detail: a more detailed description of the type of link (see the details section)
#' \item lava: name of the coefficient in lava
#' }
#' When \code{as.lava=TRUE}, \code{coefType} returns a named vector containing the type of each coefficient.
#' 
#' @examples 
#' #### regression ####
#' m <- lvm(Y~X1+X2)
#' e <- estimate(m, lava::sim(m, 1e2))
#' 
#' coefType(m)
#' coefType(e)
#'
#' #### LVM ####
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' covariance(m) <- y1~y2
#'
#' m.Sim <- m
#' categorical(m.Sim, labels = c("a","b","c")) <- ~x2
#' e <- estimate(m, lava::sim(m.Sim, 1e2))
#' 
#' coefType(m)
#' coefType(e)
#'
#' ## additional categorical variables 
#' categorical(m, labels = as.character(1:3)) <- "X1"
#'
#' coefType(m, as.lava = FALSE)
#'
#' #### LVM with constrains ####
#' m <- lvm(c(Y1~0+1*eta1,Y2~0+1*eta1,Y3~0+1*eta1,
#'           Z1~0+1*eta2,Z2~0+1*eta2,Z3~0+1*eta2))
#' latent(m) <- ~eta1 + eta2
#' e <- estimate(m, lava::sim(m,1e2))
#' 
#' coefType(m)
#' coefType(e)
#' 
#' #### multigroup ####
#' m <- lvm(Y~X1+X2)
#' eG <- estimate(list(m,m), list(lava::sim(m, 1e2), lava::sim(m, 1e2)))
#' coefType(eG)
#'
#' @concept extractor
#' @export
`coefType` <-
  function(object, as.lava, ...) UseMethod("coefType")

## * coefType.lvm
#' @rdname coefType
#' @export
coefType.lvm <- function(object, as.lava = TRUE, data = NULL, ...){

    externalLink <- type <- NULL ## [:for CRAN check] subset
    
    ## ** extract all coef
    index.all <- which(!is.na(object$M), arr.ind = FALSE)
    ls.name <- list()
    ls.X <- list()
    ls.Y <- list()
    ls.type <- list()
    ls.value <- list()
    ls.param <- list()
    ls.marginal <- list()

    ## ** intercept
    n.intercept <- length(object$mean)
    if(n.intercept>0){
        ls.name$intercept <- names(object$mean)    

        ls.Y$intercept <- ls.name$intercept
        ls.X$intercept <- rep(NA, n.intercept)    
        ls.type$intercept <- rep("intercept", n.intercept)
        ls.value$intercept <- lapply(object$mean, function(iP){if(is.numeric(iP)){iP}else{NA}})
        ls.param$intercept <- unlist(Map(function(iPar,iFix,iName){if(iFix){NA}else if(!is.na(iPar)){iPar} else {iName}},
                                         iPar = unlist(object$mean),
                                         iFix = !is.na(ls.value$intercept),
                                         iName = ls.name$intercept)
                                     )
        ls.marginal$intercept <-  ls.name$intercept %in% exogenous(object)
    }
    
    ## ** regression
    arrIndex.regression <- which(object$M==1, arr.ind = TRUE)
    index.regression <- which(object$M==1, arr.ind = FALSE)
    n.regression <- length(index.regression)
    if(n.regression>0){
    
        ls.Y$regression <- colnames(object$M)[arrIndex.regression[,"col"]]
        ls.X$regression <- rownames(object$M)[arrIndex.regression[,"row"]]
        ls.name$regression <- paste0(ls.Y$regression,
                                     lava.options()$symbols[1],
                                     ls.X$regression)
    
        ls.type$regression <- rep("regression", n.regression)    
        ls.value$regression <- object$fix[index.regression]
        ls.param$regression <- unlist(Map(function(iPar,iFix,iName){if(iFix){NA}else if(!is.na(iPar)){iPar} else {iName}},
                                          iPar = object$par[index.regression],
                                          iFix = !is.na(ls.value$regression),
                                          iName = ls.name$regression)
                                      )
        ls.marginal$regression <- rep(FALSE,n.regression)
    }

    ## ** covariance
    M.cov <- object$cov
    M.cov[upper.tri(M.cov)] <- 0
    
    arrIndex.vcov <- which(M.cov==1, arr.ind = TRUE)
    index.vcov <- which(M.cov==1, arr.ind = FALSE)
    n.vcov <- length(index.vcov)
    if(n.vcov>0){
        
        Y.vcov <- colnames(object$cov)[arrIndex.vcov[,"col"]]
        X.vcov <- rownames(object$cov)[arrIndex.vcov[,"row"]]
    
        name.vcov <- paste0(Y.vcov,
                            lava.options()$symbols[2],
                            X.vcov)
    
        value.vcov <- object$covfix[index.vcov]
        param.vcov <- unlist(Map(function(iPar,iFix,iName){if(iFix){NA}else if(!is.na(iPar)){iPar} else {iName}},
                                 iPar = object$covpar[index.vcov],
                                 iFix = !is.na(value.vcov),
                                 iName = name.vcov)
                             )

        index.variance <- which(arrIndex.vcov[,1]==arrIndex.vcov[,2])
        ls.name$variance <- name.vcov[index.variance]
        n.variance <- length(ls.name$variance)
        ls.Y$variance <- Y.vcov[index.variance]
        ls.X$variance <- X.vcov[index.variance]
        ls.type$variance <- rep("variance", n.variance)
        ls.value$variance <- value.vcov[index.variance]
        ls.param$variance <- param.vcov[index.variance]
        ls.marginal$variance <- ls.name$variance %in% paste0(exogenous(object),lava.options()$symbols[2],exogenous(object))
    
        index.covariance <- which(arrIndex.vcov[,1]!=arrIndex.vcov[,2])
        ls.name$covariance <- name.vcov[index.covariance]
        n.covariance <- length(ls.name$covariance)
        ls.Y$covariance <- Y.vcov[index.covariance]
        ls.X$covariance <- X.vcov[index.covariance]
        ls.type$covariance <- rep("covariance", n.covariance)
        ls.value$covariance <- value.vcov[index.covariance]
        ls.param$covariance <- param.vcov[index.covariance]
        ls.marginal$covariance <- rep(FALSE, n.covariance)
    }
    
    ## ** external coefficients
    n.external <- length(object$expar)
    if(n.external>0){
        ls.name$external <- names(object$expar)
        ls.type$external <- rep("external", n.external)

        ls.X$external <- rep(NA,n.external)
        for(iX in names(object$attributes$ordinalparname)){ ## iX <- "X1"
            ls.X$external[ls.name$external %in% object$attributes$ordinalparname[[iX]]] <- iX
        }
        ls.Y$external <- rep(NA,n.external)
        ls.value$external <- unlist(object$exfix)
        ls.param$external <- unlist(Map(function(iPar,iFix,iName){if(iFix){NA}else if(!is.na(iPar)){iPar} else {iName}},
                                        iPar = rep(NA,n.external),
                                        iFix = !is.na(ls.value$external),
                                        iName = ls.name$external)
                                    )
        ls.marginal$external <-  rep(FALSE, n.external)
    }

    ## ** merge
    df.param <- data.frame(name = unlist(ls.name),
                           Y = unlist(ls.Y),
                           X = unlist(ls.X),
                           data = unlist(ls.X),
                           type = unlist(ls.type),
                           value = unlist(ls.value),
                           param = unlist(ls.param),
                           marginal = unlist(ls.marginal),
                           stringsAsFactors = FALSE)
    df.param[which(df.param$X %in% latent(object)),"data"] <- NA
    
    ## ** categorical variables
    if(!is.null(object$attributes$ordinalparname)){
        resCar <- defineCategoricalLink(object, link = df.param$name, data = data)
        
        ## normal coefficients
        resCar.Nexternal <- subset(resCar,
                                   subset = is.na(externalLink),
                                   select = c("link","type","factitious","level","originalLink","externalLink"))
        ## rename according to the main data frame
        match.tempo <- match(c("link","type"),
                             names(resCar.Nexternal))
        names(resCar.Nexternal)[match.tempo] <- c("name","distribution") 
        
        df.Nexternal <- merge(subset(df.param, subset = type != "external"),
                              resCar.Nexternal, by = "name")

        ## external coefficients
        resCar.external <- subset(resCar,
                                  subset = !is.na(externalLink),
                                  select = c("link", "endogenous", "exogenous", "type", "factitious", "level", "originalLink", "externalLink"))
        resCar.external$X <- paste0(resCar.external$exogenous,
                                    resCar.external$level)

        ## rename according to the main data frame
        match.tempo <- match(c("link","endogenous","exogenous","type"),
                             names(resCar.external))
        names(resCar.external)[match.tempo] <- c("name","Y","data","distribution")
        resCar.external$param <- resCar.external$name

        for(iCol in c("type","value","marginal")){ # iCol <- "type"
            name2col <- stats::setNames(df.param[[iCol]],df.param$name)
            resCar.external[,iCol] <- name2col[resCar.external$originalLink]
        }
        df.param <- rbind(resCar.external[,names(df.Nexternal),drop=FALSE],
                          df.Nexternal)
    }else{
        df.param$factitious <- FALSE
        df.param$level <- as.character(NA)
        df.param$externalLink <-  as.character(NA)
        df.param$originalLink <- df.param$name
    }

    ## ** original link
    coef.lava <- coef(object, labels = 0)
    coef2.lava <- coef(object, labels = 1)
    
    ## ** merge with lava
    name.coef <- names(coef.lava)

    index.keep <- which(df.param$type!="external" & df.param$factitious == FALSE & df.param$marginal == FALSE)
    df.param$detail <- as.character(NA)
    df.param[index.keep, "detail"] <- detailName(object,
                                                 name.coef = df.param[index.keep, "name"],
                                                 type.coef = df.param[index.keep, "type"])
    df.param$lava <- name.coef[match(df.param$originalLink,coef.lava)]
    df.param[df.param$factitious,c("param","lava")] <- as.character(NA)
    df.param <- df.param[order(df.param$type,df.param$detail,df.param$name),,drop=FALSE]
    df.param$originalLink[is.na(df.param$lava)] <- NA
    ## df.param$param[is.na(df.param$lava)] <- NA
    rownames(df.param) <- NULL

    ## ** export
    if(as.lava){
        ## add extra mean as links
        vec.extra <- unique(stats::na.omit(df.param$externalLink))
        if(length(vec.extra)>0){
            df.extra <- data.frame(name = vec.extra, type = "extra",
                                   lava = name.coef[match(vec.extra,coef.lava)],
                                   stringsAsFactors = FALSE)
            df.param <- rbind(df.param[,c("name", "type", "lava")],
                              df.extra)
        }
        ## 
        out <- stats::setNames(df.param$type, df.param$name)
        ## out <- out[!duplicated(names(out))]
        return(out[coef.lava])    
    }else{
        df.param$detail <- factor(df.param$detail,
                                  levels = c("nu","alpha","K","Gamma","Lambda","B","Sigma_var","Sigma_cov","sigma2","sigma2k","cor","Psi_var","Psi_cov",NA))
        df.param <- df.param[order(df.param$detail,df.param$param),]
        df.param$detail <- as.character(df.param$detail)
        return(df.param)
    }
}

## * coefType.lvmfit
#' @rdname coefType
#' @export
coefType.lvmfit <- function(object, as.lava = TRUE, ...){ 

    ## ** find type of the coefficients in the original model
    df.param <- coefType(object$model0, as.lava = FALSE, ...)
    
    ## ** export
    if(as.lava){
        out <- subset(df.param, subset = !is.na(lava), select = c("type", "name"))
        out <- stats::setNames(out$type, out$name)
        coef.lava <- names(stats::coef(object))
        return(out[coef.lava])    
    }else{        
        return(df.param)
    }
}

## * coefType.multigroup
#' @rdname coefType
#' @export
coefType.multigroup <- function(object, as.lava = TRUE, ...){

    n.model <- length(object$lvm)
    df.param <- NULL
    
    for(iModel in 1:n.model){ # iModel <- 2

        df.param <- rbind(df.param,
                          cbind(coefType(object$lvm[[iModel]], as.lava = FALSE), model = iModel)
                          )
      
    }

    df.param$name <- paste0(df.param$model,"@",df.param$name)
    
    ## *** export
    if(as.lava){        
        out <- subset(df.param, subset = !is.na(lava), select = c("type", "name"))
        out <- stats::setNames(out$type, out$name)
        return(out)
    }else{
        return(df.param)
    }
}


## * detailName (needed for coefType)
detailName <- function(object, name.coef, type.coef){
    ls.links <- initVarLinks(name.coef)
    index.loading <- setdiff(which(ls.links$var2 %in% latent(object)),
                             which(type.coef %in% c("covariance","variance")))
    if(length(index.loading)>0){
        type.coef[index.loading] <- "loading"
    }

    index.measurement <- which(ls.links$var1 %in% endogenous(object))
    if(length(index.measurement)>0){
        type.coef[index.measurement] <- as.character(factor(type.coef[index.measurement],
                                                       levels = c("intercept","regression","loading","covariance","variance"),
                                                       labels = c("nu","K","Lambda","Sigma_cov","Sigma_var")))
    }
    index.structural <- setdiff(1:length(type.coef),index.measurement)
    if(length(index.structural)>0){
        type.coef[index.structural] <- as.character(factor(type.coef[index.structural],
                                                      levels = c("intercept","regression","loading","covariance","variance"),
                                                      labels = c("alpha","Gamma","B","Psi_cov","Psi_var")))
    }
    return(type.coef)
}

##----------------------------------------------------------------------
### coefType.R ends here

### sCorrect.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: Jan 22 2022 (13:50) 
## Version: 
## Last-Updated: Jan 22 2022 (14:05) 
##           By: Brice Ozenne
##     Update #: 12
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

#' @title Depreciated Method For Small Sample Correction
#' @description Depreciated method for small sample correction, now replaced by the \code{\link{estimate2}} method.
#' @name sCorrect
#' 
#' @param object,x a \code{lvmfit} object.
#' @param value not used.
#' @param ... not used.
#'

#' @rdname sCorrect
#' @export
`sCorrect` <-
    function(object, ...) UseMethod("sCorrect")

#' @rdname sCorrect
#' @export
`sCorrect.default` <- function(object, ...){
    .Defunct("estimate2", package = "lavaSearch2", msg = "Function sCorrect has been removed from lavaSearch2 version 2.0.0 or higher and replace by the estimate2 method. \n")
}

#' @rdname sCorrect
#' @export
`sCorrect<-` <-
  function(x, ..., value) UseMethod("sCorrect<-")

#' @rdname sCorrect
#' @export
`sCorrect<-.default` <- function(x, ..., value){
    .Defunct("estimate2", package = "lavaSearch2", msg = "Function sCorrect has been removed from lavaSearch2 version 2.0.0 or higher and replace by the estimate2 method. \n")
}

##----------------------------------------------------------------------
### sCorrect.R ends here

### test1-sCorrect-validObjects.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  6 2018 (10:42) 
## Version: 
## Last-Updated: Jan 11 2022 (09:07) 
##           By: Brice Ozenne
##     Update #: 68
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R    
    library(testthat)
    library(lavaSearch2)
}
library(data.table)
lava.options(symbols = c("~","~~"))
context("sCorrect (warnings and errors for invalid objects/arguments)")

## * Simulation
n <- 100
m.sim <- lvm(Y~X1+X2,G~1)
categorical(m.sim,K=3,label=c("a","b","c")) <- ~G+X2
set.seed(10)
d <- lava::sim(m.sim,n,latent=FALSE)

## * estimate2 for lvm objects

## ** error for multigroup lvm
## check in sCorrect.R 
suppressWarnings(e <- estimate(list(lvm(Y~X1),lvm(Y~X1),lvm(Y~X1)), data = split(d,d$G)))
test_that("error for multigroup models", {
    expect_error(estimate2(e))
})

## ** error for tobit lvm
## check in sCorrect.R
e <- estimate(lvm(G~X1), data = d)
test_that("error for tobit models", {
    expect_error(estimate2(e))
})

## ** error for lvm with transform variables
## check in sCorrect.R
m <- lvm(Y~X1)
transform(m,Id~X1) <- function(x){1:NROW(x)}
d.tempo <- lava::sim(m, n)
e <- estimate(m, data = d.tempo)
test_that("error when using transform", {
    expect_error(estimate2(e))
})

## * estimate2 with data.table
dt <- as.data.table(d)
e <- estimate(lvm(Y~X1+X2+G), data = dt)
test_that("ok for data.table objects", {
    expect_true(inherits(estimate2(e), "lvmfit2"))
})

##----------------------------------------------------------------------
### test1-sCorrect-validObjects.R ends here

### test1-sCorrect-smallSampleCorrection.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  7 2018 (12:08) 
## Version: 
## Last-Updated: jul 31 2020 (10:23) 
##           By: Brice Ozenne
##     Update #: 57
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
## library(clubSandwich)
library(nlme)
calcFactor <- function(object){
    return((object$dims$N - object$dims$p)/(object$dims$N - object$dims$p * (object$method == "REML")))
}

context("sCorrect (Satterthwaite + small sample correction)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]
dLred$variable.factor <- as.factor(dLred$variable)

## * linear regression [lm,gls,lvm]
## ** model fit and sCorrect
e.lm <- lm(Y1~X1+X2, data = d)

## ** iid2 matches clubSandwich
test_that("iid2.lm/iid2.lvm matches clubSandwich", {
    ## V.GS <- clubSandwich::vcovCR(e.lm, type = "CR2", cluster = d$Id)
    
    eHC2.iid2.lm <- iid2(e.lm, bias.correct = TRUE)
    V.lm <- crossprod(eHC2.iid2.lm)

    ## no more equal to clubSandwich since version 1.2.1 
    ## > lavaSearch2 uses the average bias to correct the residuals instead of the individual bias
    
    ## expect_equal(as.matrix(V.GS),
    ##              V.lm[rownames(V.GS),colnames(V.GS)],
    ##              tol = 1e-7)
})


## * multiple linear regression with constrains [lvm, gls]
## ** model fit and sCorrect
e.gls <- gls(value ~ X1 + X2,
             data = dLred,
             weight = varIdent(form = ~1|variable),
             method = "ML")
e.lvm <- estimate(lvm(Y1[mu:sigma1]~ beta1*X1 + beta2*X2,
                      Y2[mu:sigma2]~ beta1*X1 + beta2*X2,
                      Y3[mu:sigma3]~ beta1*X1 + beta2*X2),
                  data = d)

factor <- calcFactor(e.gls)
index.coef <- 1:3

test_that("gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.gls)))
})
 
## ** HC0/HC1
iid2HC0.gls <- iid2(e.gls, bias.correct = FALSE, cluster = "Id")
iid2HC0.lvm <- iid2(e.lvm, bias.correct = FALSE)

test_that("iid2.gls/iid2.lvm matches clubSandwich (HC0-HC1)", {
    expect_equal(unname(iid2HC0.gls[,index.coef]),
                 unname(iid2HC0.lvm[,index.coef]),
                 tol = 1e-5)
    
    VHC0.gls <- crossprod(iid2HC0.gls)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR0", cluster = dLred$Id) * factor^2
    ## expect_equal(as.double(GS),as.double(VHC0.gls), tolerance = 1e-10)
    
    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR1", cluster = dLred$Id) * factor^2
    ## VHC1.gls <- VHC0.gls*n/(n-1)
    ## expect_equal(as.double(GS),as.double(VHC1.gls), tolerance = 1e-10)
})

## ** HC2
iid2HC2.gls <- iid2(e.gls, bias.correct = TRUE, n.iter = 1, cluster = "Id")
iid2HC2.lvm <- iid2(e.lvm, bias.correct = TRUE, n.iter = 1)

test_that("iid2.lvm matches clubSandwich (HC2)", {
    expect_equal(unname(iid2HC2.gls[,index.coef]),
                 unname(iid2HC2.lvm[,index.coef]),
                 tol = 1e-5)

    VHC2.gls <- crossprod(iid2HC2.gls)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR2", cluster = dLred$Id) * factor^2

    ## no more equal to clubSandwich since version 1.2.1 
    ## > lavaSearch2 uses the average bias to correct the residuals instead of the individual bias
    ## expect_equal(as.double(GS),as.double(VHC2.gls), tolerance = 1e-5)
})

## * mixed model: CS [lvm,gls,lme]
## ** model fit and sCorrect
m <- lvm(c(Y1[0:sigma]~1*eta,
           Y2[0:sigma]~1*eta,
           Y3[0:sigma]~1*eta,
           eta~X1+X2))
latent(m) <- ~eta
e.lvm <- estimate(m, d)

e.lme <- nlme::lme(value ~ X1 + X2,
                   random =~1| Id,
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ X1 + X2,
                   correlation = corCompSymm(form = ~1| Id),
                   data = dLred, method = "ML")
index.coef <- 1:length(coef(e.gls))

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

factor <- calcFactor(e.gls)
 
## ** HC0/HC1
iid2HC0.gls <- iid2(e.gls, bias.correct = FALSE)
iid2HC0.lme <- iid2(e.lme, bias.correct = FALSE)
iid2HC0.lvm <- iid2(e.lvm, bias.correct = FALSE)

test_that("iid2.gls/iid2.lme/iid2.lvm matches clubSandwich (HC0-HC1)", {
    expect_equal(unname(iid2HC0.gls[,index.coef]),
                 unname(iid2HC0.lvm[,index.coef]),
                 tol = 1e-5)

    expect_equal(unname(iid2HC0.gls[,index.coef]),
                 unname(iid2HC0.lme[,index.coef]),
                 tol = 1e-5)
    
    VHC0.gls <- crossprod(iid2HC0.gls)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR0", cluster = dLred$Id) * factor^2
    ## expect_equal(as.double(GS),as.double(VHC0.gls), tolerance = 1e-10)

    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR1", cluster = dLred$Id) * factor^2
    ## VHC1.gls <- VHC0.gls*n/(n-1)
    ## expect_equal(as.double(GS),as.double(VHC1.gls), tolerance = 1e-10)
})

## ** HC2
iid2HC2.gls <- iid2(e.gls, bias.correct = TRUE, n.iter = 1)
iid2HC2.lme <- iid2(e.lme, bias.correct = TRUE, n.iter = 1)
iid2HC2.lvm <- iid2(e.lvm, bias.correct = TRUE, n.iter = 1)

test_that("iid2.gls/iid2.lme/iid2.lvm matches clubSandwich (HC2)", {
    expect_equal(unname(iid2HC2.gls[,index.coef]),
                 unname(iid2HC2.lvm[,index.coef]),
                 tol = 1e-5)

    expect_equal(unname(iid2HC2.gls[,index.coef]),
                 unname(iid2HC2.lme[,index.coef]),
                 tol = 1e-5)

    VHC2.gls <- crossprod(iid2HC2.gls)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.gls, type = "CR2", cluster = dLred$Id) * factor^2
    ## no more equal to clubSandwich since version 1.2.1 
    ## > lavaSearch2 uses the average bias to correct the residuals instead of the individual bias
    ## expect_equal(as.double(GS),as.double(VHC2.gls), tolerance = 1e-10)
})

## * mixed model: CS with different variances [lvm,gls]
## ** model fit and sCorrect
m <- lvm(c(Y1[0:sigma1]~1*eta,
           Y2[0:sigma2]~1*eta,
           Y3[0:sigma3]~1*eta,
           eta~X1+X2))
latent(m) <- ~eta
e.lvm <- estimate(m, d)

e.lme <- nlme::lme(value ~ X1 + X2,
                   random =~1| Id,
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ X1 + X2,
                   correlation = corCompSymm(form = ~1| Id),
                   weights = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")
index.coef <- 1:length(coef(e.gls))

test_that("lme equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    ## gls does not give the same likelihood
    ## expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** HC0/HC1
iid2HC0.gls <- iid2(e.gls, bias.correct = FALSE)
iid2HC0.lme <- iid2(e.lme, bias.correct = FALSE)
iid2HC0.lvm <- iid2(e.lvm, bias.correct = FALSE)

test_that("iid2.lme/iid2.lvm matches clubSandwich (HC0-HC1)", {
    expect_equal(unname(iid2HC0.lme[,index.coef]),
                 unname(iid2HC0.lvm[,index.coef]),
                 tol = 1e-5)

    VHC0.lme <- crossprod(iid2HC0.lme)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR0", cluster = dLred$Id)
    ## expect_equal(as.double(GS),as.double(VHC0.lme), tolerance = 1e-10)

    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR1", cluster = dLred$Id)
    ## VHC1.lme <- VHC0.lme*n/(n-1)
    ## expect_equal(as.double(GS),as.double(VHC1.lme), tolerance = 1e-10)
})

## ** HC2
iid2HC2.gls <- iid2(e.gls, bias.correct = TRUE, n.iter = 1)
iid2HC2.lme <- iid2(e.lme, bias.correct = TRUE, n.iter = 1)
iid2HC2.lvm <- iid2(e.lvm, bias.correct = TRUE, n.iter = 1)

test_that("iid2.lme/iid2.lvm matches clubSandwich (HC2)", {
    expect_equal(unname(iid2HC2.lme[,index.coef]),
                 unname(iid2HC2.lvm[,index.coef]),
                 tol = 1e-5)

    VHC2.lme <- crossprod(iid2HC2.lme)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR2", cluster = dLred$Id)
    ## no more equal to clubSandwich since version 1.2.1 
    ## > lavaSearch2 uses the average bias to correct the residuals instead of the individual bias
    ## expect_equal(as.double(GS),as.double(VHC2.lme), tolerance = 1e-5)
})


## * mixed model: Unstructured
## ** model fit and sCorrect
m <- lvm(c(Y1~0+1*eta,
           Y2~0+1*eta,
           Y3~0+1*eta,
           eta~X1+X2))
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.gls <- nlme::gls(value ~ X1 + X2,
                   correlation = corSymm(form =~ 1| Id),
                   weight = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

e.lme <- nlme::lme(value ~ X1 + X2,
                   random =~ 1|Id,
                   correlation = corSymm(),
                   weight = varIdent(form =~ 1|variable),
                   data = dLred, method = "ML")

index.coef <- 1:length(coef(e.gls))

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** HC0/HC1
iid2HC0.gls <- iid2(e.gls, bias.correct = FALSE)
## iid2HC0.lme <- iid2(e.lme, bias.correct = FALSE)
## does not work because the model is overparametrized
iid2HC0.lvm <- iid2(e.lvm, bias.correct = FALSE)

test_that("iid2.gls/iid2.lme/iid2.lvm matches clubSandwich (HC0-HC1)", {
    expect_equal(unname(iid2HC0.gls[,index.coef]),
                 unname(iid2HC0.lvm[,index.coef]),
                 tol = 1e-5)

    VHC0.lvm <- crossprod(iid2HC0.lvm)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR0", cluster = dLred$Id)
    ## expect_equal(as.double(GS),as.double(VHC0.lvm), tolerance = 1e-5)

    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR1", cluster = dLred$Id)
    ## VHC1.lvm <- VHC0.lvm*n/(n-1)
    ## expect_equal(as.double(GS),as.double(VHC1.lvm), tolerance = 1e-5)
})

## ** HC2
iid2HC2.gls <- iid2(e.gls, bias.correct = TRUE, n.iter = 1)
expect_error(iid2HC2.lme <- iid2(e.lme, bias.correct = TRUE, n.iter = 1))
## does not work because the model is overparametrized
iid2HC2.lvm <- iid2(e.lvm, bias.correct = TRUE, n.iter = 1)

test_that("iid2.gls/iid2.lme/iid2.lvm matches clubSandwich (HC2)", {
    expect_equal(unname(iid2HC2.gls[,index.coef]),
                 unname(iid2HC2.lvm[,index.coef]),
                 tol = 1e-5)

    VHC2.lvm <- crossprod(iid2HC2.lvm)[index.coef,index.coef]
    ## GS <- clubSandwich::vcovCR(e.lme, type = "CR2", cluster = dLred$Id)
    ## no more equal to clubSandwich since version 1.2.1 
    ## > lavaSearch2 uses the average bias to correct the residuals instead of the individual bias
    ## expect_equal(as.double(GS),as.double(VHC2.lvm), tolerance = 1e-5)
})


##----------------------------------------------------------------------
### test1-sCorrect-smallSampleCorrection.R ends here

### test1-sCorrect-adjustedResiduals.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  7 2018 (12:21) 
## Version: 
## Last-Updated: jan 17 2022 (14:11) 
##           By: Brice Ozenne
##     Update #: 32
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
context("sCorrect (residuals)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * linear regression [lvm]
## ** model fit and sCorrect
e.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d)
e.lm <- lm(Y1~X1+X2+Gender, data = d)
e.gls <- gls(Y1~X1+X2+Gender, data = d, method = "ML")

e2.lvm <- estimate2(e.lvm)


## ** test adjusted residuals
test_that("residuals2 match residuals.lm (lm adjusted)", {
    epsilon.lm <- residuals(e.lm)
    ## X <- model.matrix(e.lm, d)
    ## iH <- diag(1,n,n) - X %*% solve(t(X) %*% X) %*% t(X)
    ## GS1 <- epsilon.lm/diag(iH)^(1/2)

    ## match expectation
    expect_equal(mean(e2.lvm$sCorrect$residuals), 0)

    ## match variance
    expect_equal(mean(e2.lvm$sCorrect$residuals^2), sigma(e.lm)^2,
                 tol = 1e-8)
    
    ## mean(GS1^2)
    ## mean(residuals(e.lm, type = "pearson")^2)
    ## mean(e2.lm$sCorrect$residuals^2)
    ## sigma(e2.lm)^2

    ## match individual components
    factor <- sqrt(NROW(epsilon.lm)/(NROW(epsilon.lm) - length(coef(e.lm))))
    expect_equal(as.double(e2.lvm$sCorrect$residuals),
                 as.double(epsilon.lm)*factor)
})

## * multivariate linear models
## ** model fit
ls.lm <- list(lm(Y1~X1,d),lm(Y2~X2,d),lm(Y3~X1+X3,d))
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

e2.lvm <- estimate2(e.lvm)

test_that("residuals2 match residuals.lm", {

    epsilon.lm <- lapply(ls.lm, residuals)
    ## X <- lapply(ls.lm, model.matrix)
    ## iH <- lapply(X, function(x){
    ##     diag(1,NROW(x),NROW(x)) - x %*% solve(t(x) %*% x) %*% t(x)
    ## })
    ## GS <- mapply(iH,epsilon.lm, FUN = function(x,y){
    ##     y/diag(x)^(1/2)
    ## })

    ## match expectation
    expect_equal(as.double(colMeans(e2.lvm$sCorrect$residuals)),
                 c(0,0,0),
                 tol = 1e-8)
    
    ## match variance
    expect_equal(unname(colMeans(e2.lvm$sCorrect$residuals^2)),
                 sapply(ls.lm,sigma)^2,
                 tol = 1e-8)

    ## match individual components
    ls.GS <- lapply(ls.lm, function(x){
        iEpsilon <- residuals(x)
        factor <- sqrt(NROW(iEpsilon)/(NROW(iEpsilon) - length(coef(x))))
        return(iEpsilon*factor)
    })

    
    expect_equal(unname(e2.lvm$sCorrect$residuals),
                 unname(do.call(cbind,ls.GS)))
})

## * mixed model: CS [lvm]
m <- lvm(c(Y1[0:sigma]~1*eta,
           Y2[0:sigma]~1*eta,
           Y3[0:sigma]~1*eta,
           eta~X1+X2))
latent(m) <- ~eta
e.lvm <- estimate(m, d)

e.lme <- nlme::lme(value ~ X1 + X2,
                   random =~1| Id,
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ X1 + X2,
                   correlation = corCompSymm(form = ~1| Id),
                   data = dLred, method = "ML")

e2.lvm <- estimate2(e.lvm)

test_that("residuals2 in mixed models", {

    ## mean
    expect_equal(colSums(e2.lvm$sCorrect$residuals),
                 colSums(residuals(e.lvm)))

    ## variance
    getVarCov2(e.lvm)/var(e2.lvm$sCorrect$residuals)
    ## not 1....
    
})

##----------------------------------------------------------------------
### test1-sCorrect-adjustedResiduals.R ends here

### test1-cluster.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: apr 26 2018 (10:37) 
## Version: 
## Last-Updated: feb 11 2019 (11:03) 
##           By: Brice Ozenne
##     Update #: 6
##----------------------------------------------------------------------
## 
### Commentary: 
## Check that the robust standard error of lavaSearch2 matches those of lava
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
context("cluster argument (replicate lava results)")

## * simulation
n <- 2e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){round(runif(n = n, min = 0, max = 30))}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
table(d$Id)

## * linear regression [lvm]
## ** model fit and sCorrect
test_that("lm - robust vcov",{
    e.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d)
    sCorrect(e.lvm) <- FALSE
    eR.lvm <- estimate(lvm(Y1~X1+X2+Gender), data = d, cluster ="Id")
    
    n.Id <- length(unique(d$Id))
    GS <- vcov(eR.lvm)/(n.Id/(n.Id-1))
    test <- crossprod(iid2(e.lvm, cluster = "Id"))
    expect_true(all(abs(GS/test-1)<1e-2))
    ## GS/test

    summary(eR.lvm)$coef
    compare2(e.lvm, par = c("Y1~X1"),
             robust = TRUE, cluster = "Id")
##    compare2(e.lvm, par = c("Y1~X1"))
})

######################################################################
### test1-cluster.R ends here

### test1-sCorrect-effects2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan 19 2022 (11:40) 
## Version: 
## Last-Updated: jan 19 2022 (11:57) 
##           By: Brice Ozenne
##     Update #: 3
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))
context("sCorrect (effects2)")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)

## * effects2
m <- lvm(Y1~eta1,Y2~eta1+X2,Y3~eta1,
         Z1~eta2+X2,Z2~eta2,Z3~eta2,
         eta2~eta1+X2)

test_that("LM - effects2 correction", {

    e <- estimate(lvm(Y1~X2+Z1,Z1~X2), d)
    e2 <- estimate2(e)

    test <- effects2(e2, linfct = c("Y1~X2|total","Y1~X2|direct","Y1~X2|indirect"))
    expect_equal(summary(test, test = adjusted("none"))$table2$df,
                 c(51.82375, 47.00000, 48.04664), tol = 1e-3)
})

test_that("LVM - effects2 no correction", {

    e <- estimate(m, d)
    e0 <- estimate2(e, ssc = FALSE, df = FALSE) ## no correction

    GS <- effects(e, from = "X2", to = "Z1")
    test1 <- effects(e0, linfct = "Z1~X2")
    test3 <- effects(e0, linfct = c("Z1~X2|total","Z1~X2|direct","Z1~X2|indirect"))

    expect_equal(as.double(GS$coef[c("total","direct","indirect")]),
                 summary(test3, test = adjusted("none"))$table2[,"estimate"],
                 tol = 1e-6)
    expect_equal(as.double(summary(test1, test = adjusted("none"))$table2[1,]),
                 as.double(summary(test3, test = adjusted("none"))$table2[1,]),
                 tol = 1e-6)
    expect_equal(as.double(sqrt(diag(GS$vcov[1:3,1:3]))),
                 summary(test3, test = adjusted("none"))$table2[,"se"],
                 tol = 1e-6)


    e2 <- estimate2(e) ## correction
    test12 <- effects(e2, linfct = "Z1~X2")
    test32 <- effects(e2, linfct = c("Z1~X2|total","Z1~X2|direct","Z1~X2|indirect"))
})


##----------------------------------------------------------------------
### test1-sCorrect-effects2.R ends here

### test1-sCorrect-validObjects.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  6 2018 (10:42) 
## Version: 
## Last-Updated: feb 11 2019 (13:44) 
##           By: Brice Ozenne
##     Update #: 63
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
rm(list = ls())
if(FALSE){ ## already called in test-all.R    
    library(testthat)
    library(lavaSearch2)
}
library(data.table)
library(nlme)
lava.options(symbols = c("~","~~"))
context("sCorrect (warnings and errors for invalid objects/arguments)")

## * Simulation
n <- 100
m.sim <- lvm(Y~X1+X2,G~1)
categorical(m.sim,K=3,label=c("a","b","c")) <- ~G+X2
set.seed(10)
d <- lava::sim(m.sim,n,latent=FALSE)

## * sCorrect for lvm objects

## ** error for multigroup lvm
## check in sCorrect.R 
suppressWarnings(e <- estimate(list(lvm(Y~X1),lvm(Y~X1),lvm(Y~X1)), data = split(d,d$G)))
test_that("error for multigroup models", {
    expect_error(sCorrect(e))
})

## ** error for tobit lvm
## check in sCorrect.R
if(require(lava.tobit)){
    e <- estimate(lvm(G~X1), data = d)
    test_that("error for tobit models", {
        expect_error(sCorrect(e))
    })
}

## ** error for lvm with transform variables
## check in sCorrect.R
m <- lvm(Y~X1)
transform(m,Id~X1) <- function(x){1:NROW(x)}
d.tempo <- lava::sim(m, n)
e <- estimate(m, data = d.tempo)
test_that("error when using transform", {
    expect_error(sCorrect(e))
})

## * sCorrect for nlme objects

## ** warning for the satterthwaite approx. with REML
## check in sCorrect.R 
e <- gls(Y~X1, data = d, correlation = corCompSymm(form =~1|G),
         method = "REML")
test_that("no warning when using no correction", {
    expect_warning(sCorrect(e, df = FALSE, trace = 0) <- FALSE, regexp = NA)
})

## ** warning for the small sample correction estimated with REML
## check in sCorrect.R
e <- gls(Y~X1, data = d, correlation = corCompSymm(form =~1|G))
test_that("warning when using nlme with REML and small sample correction", {
    expect_warning(sCorrect(e, df = FALSE, trace = 0) <- TRUE)
})

## ** error for more than one random effect
## check in Utils-nlme.R (.getVarCov)
e <- lme(Y~X1, random =~ 1|G / X2, data = d, method = "ML")
test_that("error when using nlme with several random effects", {
    expect_error(sCorrect(e, trace = 0))
})

## ** error for more non-standard correlation shape
## check in Utils-nlme.R (.getIndexOmega2)
e <- gls(Y~X1, data = d, correlation = corAR1(form =~1|G), method = "ML")
test_that("error when using nlme with non standard correlation", {
    expect_error(sCorrect(e, trace = 0))
})

## ** error for more non-standard variance shape
## check in Utils-nlme.R (.getIndexOmega2)
e <- gls(Y~X1, data = d, weight = varExp(form =~X1), method = "ML")
test_that("error when using nlme with non standard variance", {
    expect_error(sCorrect(e, cluster = 1:NROW(d), trace = 0))
})


## * sCorrect with data.table

e <- estimate(lvm(Y~X1+X2+G), data = as.data.table(d))
test_that("ok for data.table objects", {
    sCorrect(e) <- FALSE
    sCorrect(e) <- TRUE
})

##----------------------------------------------------------------------
### test1-sCorrect-validObjects.R ends here

### sCorrect-glht2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 29 2017 (12:56) 
## Version: 
## Last-Updated: jan 23 2024 (10:25) 
##           By: Brice Ozenne
##     Update #: 811
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## User 
### Code:

## * Documentation - glht2
#' @title General Linear Hypothesis Testing With Small Sample Correction
#' @description Test linear hypotheses on coefficients from a latent variable models with small sample corrections.
#' @name glht2
#' 
#' @param object,model a \code{lvmfit}, \code{lvmfit2}, or \code{mmm} object.
#' @param linfct [matrix or vector of character] the linear hypotheses to be tested. Same as the argument \code{par} of \code{\link{createContrast}}.
#' @param rhs [vector] the right hand side of the linear hypotheses to be tested.
#' @param robust [logical] should robust standard error be used? 
#' Otherwise rescale the influence function with the standard error obtained from the information matrix.
#' @param cluster  [integer vector] the grouping variable relative to which the observations are iid.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... [logical] arguments passed to lower level methods.
#'
#' @details
#' Whenever the argument linfct is not a matrix, it is passed to the function \code{createContrast} to generate the contrast matrix and, if not specified, rhs. \cr \cr
#'
#' Since only one degree of freedom can be specify in a glht object and it must be an integer, the degree of freedom of the denominator of an F test simultaneously testing all hypotheses is retained, after rounding. \cr \cr
#'
#' Argument rhs and null are equivalent.
#' This redondance enable compatibility between \code{lava::compare}, \code{compare2}, \code{multcomp::glht}, and \code{glht2}.
#' @return A \code{glht} object.
#' 
#' @seealso
#' \code{\link{createContrast}} to create contrast matrices. \cr
#' \code{\link{estimate2}} to pre-compute quantities for the small sample correction.
#' 
#' @concept multiple comparisons
#'
#' @examples
#' library(multcomp)
#' 
#' ## Simulate data
#' mSim <- lvm(c(Y1,Y2,Y3)~ beta * eta, Z1 ~ E, Z2 ~ E, Age[40:5]~1)
#' latent(mSim) <- "eta"
#' set.seed(10)
#' n <- 1e2
#'
#' df.data <- lava::sim(mSim, n, latent = FALSE, p = c(beta = 1))
#'
#' #### Inference on a single model ####
#' e.lvm <- estimate(lvm(Y1~E), data = df.data)
#' summary(glht2(e.lvm, linfct = c("Y1~E + Y1","Y1")))
#' 
#' #### Inference on separate models ####
#' ## fit separate models
#' lvmX <- estimate(lvm(Z1 ~ E), data = df.data)
#' lvmY <- estimate(lvm(Z2 ~ E + Age), data = df.data)
#' lvmZ <- estimate(lvm(c(Y1,Y2,Y3) ~ eta, eta ~ E), 
#'                  data = df.data)
#'
#' #### create mmm object #### 
#' e.mmm <- mmm(X = lvmX, Y = lvmY, Z = lvmZ)
#'
#' #### create contrast matrix ####
#' resC <- createContrast(e.mmm, linfct = "E")
#'
#' #### adjust for multiple comparisons ####
#' e.glht2 <- glht2(e.mmm, linfct = c(X="E"), df = FALSE)
#' summary(e.glht2)
#'
#' @concept multiple comparison
#' @export
`glht2` <-
    function(object, ...) UseMethod("glht2")


## * glht2.lvmfit
#' @rdname glht2
#' @export
glht2.lvmfit <- function(object, linfct, rhs = NULL, robust = FALSE, cluster = NULL, ssc = lava.options()$ssc, df = lava.options()$df, ...){
    return(glht2(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...), linfct = linfct, rhs = rhs, robust = robust, cluster = cluster))

}

## * glht2.lvmfit2
#' @rdname glht2
#' @export
glht2.lvmfit2 <- function(object, linfct, rhs = NULL,
                          robust = FALSE, cluster = NULL,
                          ...){

    out <- compare2(object, linfct = linfct, rhs = rhs,
                    robust = robust, cluster = cluster,
                    as.lava = FALSE, F.test = FALSE, ...)
        
    return(out)
}


## * glht2.mmm
#' @rdname glht2
#' @export
glht2.mmm <- function (object, linfct, rhs = 0,
                       robust = FALSE, cluster = NULL,
                       ...){

    ## ** check the class of each model
    n.object <- length(object)
    name.object <- names(object)    
    if(is.null(name.object)){
        stop("Argument \'object\' must be named list. \n")
    }

    test.lvmfit <- sapply(object, inherits, what = "lvmfit")
    if(any(test.lvmfit == 0)){
        index.wrong <- which(test.lvmfit == 0)
        stop("Argument \'object\' must be a list of objects that inherits from lvmfit. \n",
             "Incorrect element(s): ",paste(index.wrong, collapse = " "),".\n")
    }
    test.lvmfit2 <- sapply(object, inherits, what = "lvmfit2")
    if(any(test.lvmfit2 == 0)){
        for(iO in which(test.lvmfit2==0)){
            object[[iO]] <- estimate2(object[[iO]], dVcov.robust = robust, ...)
        }
    }
    
    ## ** define the contrast matrix
    out <- list()
    if (is.character(linfct)){
        resC <- createContrast(object, linfct = linfct, rowname.rhs = FALSE)
        linfct <- resC$contrast
        ls.contrast <- resC$mlf
        if("rhs" %in% names(match.call()) == FALSE){
            rhs <- resC$null
        }
    }else if(is.matrix(linfct)){
        ls.contrast <- lapply(name.object, function(x){ ## x <- name.object[2]
            iColnames <- grep(paste0("^",x,": "), colnames(linfct), value = FALSE, fixed = FALSE)
            iRownames <- rowSums(linfct[,iColnames]!=0)>0
            linfct[iRownames, iColnames,drop=FALSE]            
        })
        names(ls.contrast) <- name.object
        contrast <- linfct
        if("rhs" %in% names(match.call()) == FALSE){ ## left rhs to default value
            rhs <- rep(0, NROW(contrast))
        }else if(length(rhs)!=NROW(contrast)){
            stop("mismatch between the dimensions of argument \'rhs\' and argument \'contrast\' \n")
        }
    }else{
        stop("Argument \'linfct\' must be a matrix or a vector of characters. \n",
             "Consider using  out <- createContrast(...) and pass out$contrast to linfct. \n")
    }

    ## ** check whether it is possible to compute df
    ## i.e. are linear hypothesis model specific?
    test.df <- all(unlist(lapply(object, function(iModel){iModel$sCorrect$df == "satterthwaite"})))
    if(test.df){
        n.hypo <- NROW(linfct)
        ls.modelPerTest <- lapply(1:n.hypo, function(iHypo){ ## iHypo <- 1
            iContrast <- linfct[iHypo,]
            iNames <- names(iContrast)[abs(iContrast)>0]
            iModels <- unlist(lapply(strsplit(iNames, split = ":"),"[[",1))
            return(length(unique(iModels)))
        })
        
        if(any(unlist(ls.modelPerTest)>1)){
            stop("Cannot compute the degrees of freedom for tests performed across several models \n",
                 "Consider setting the argument \'df\' to FALSE \n")
        }    
    }

    ## ** Total number of observations
    if(!is.null(cluster)){
        ls.cluster <- lapply(object, function(iO){extractData(iO, rm.na = FALSE)[[cluster]]})
        Ucluster <- unique(unlist(ls.cluster))
        n.cluster <- length(Ucluster)
    }
    
    ## ** Extract influence functions from all models
    ls.res <- lapply(1:n.object, function(iM){ ## iM <- 1

        ## *** Pre-compute quantities
        if(!inherits(object[[iM]],"lvmfit2")){
            object[[iM]] <- estimate2(object[[iM]], ...)
        }
        out$param <- coef(object[[iM]], as.lava = FALSE)
        name.param <- names(out$param)
        name.object.param <- paste0(name.object[iM],": ",name.param)
        out$param <- stats::setNames(out$param, name.object.param)
        
        ## *** Compute df for each test
        if(!is.na(object[[iM]]$sCorrect$df)){
            ## here null does not matter since we only extract the degrees of freedom
            iContrast <- ls.contrast[[iM]]
            colnames(iContrast) <- name.param
            iWald <- compare2(object[[iM]], linfct = iContrast, as.lava = FALSE, F.test = FALSE)
            out$df <- iWald$df
        }else{
            out$df <- Inf
        }
        ## *** get iid decomposition
        iid.tempo <- iid(object[[iM]], robust = robust, cluster = cluster, as.lava = FALSE)
        if(!is.null(cluster)){
            out$iid <- matrix(NA, nrow = n.cluster, ncol = length(name.param),
                              dimnames = list(Ucluster, name.param))
            out$iid[attr(iid.tempo,"cluster"),] <- iid.tempo
        }else{
            out$iid <- iid.tempo
        }
        colnames(out$iid) <- name.object.param

        ## *** get se
        if(robust){
            out$se <- sqrt(diag(crossprod(iid.tempo)))
        }else{
            out$se <- sqrt(diag(vcov(object[[iM]], as.lava = FALSE)))
        }
        return(out)
        
    })
    seq.df <- unlist(lapply(ls.res,"[[","df"))
    seq.param <- unlist(lapply(ls.res,"[[","param"))

    if(test.df){
        df.global <- round(stats::median(seq.df), digits = 0)
    }else{
        df.global <- 0
    }
    ls.iid <- lapply(ls.res,"[[","iid")
    ls.se <- lapply(ls.res,"[[","se")
    n.obs <- unique(unlist(lapply(ls.iid, NROW)))
    if(length(n.obs)>1){
        stop("Mismatch between the number of observations in the iid \n",
             "Likely to be due to the presence of missing values \n",
             "Consider specifying the \'cluster\' argument \n")
    }
    M.iid <- do.call(cbind,ls.iid)
    diag.se <- diag(do.call(c,ls.se))
    if(any(is.na(M.iid))){
       M.iid[is.na(M.iid)] <- 0
    }
    vcov.object <- diag.se %*% stats::cov2cor(crossprod(M.iid)) %*% diag.se ## same as multcomp:::vcov.mmm
    dimnames(vcov.object) <- list(colnames(M.iid), colnames(M.iid))
    
    ## ** sanity check
    name.param <- names(seq.param)
    if(!identical(colnames(linfct),name.param)){
        stop("Column names of the contrast matrix does not match the one of the coefficients \n")
    }
    if(!identical(colnames(vcov.object),name.param)){
        stop("Column names of the variance covariance matrix does not match the one of the coefficients \n")
    }
    if(!identical(rownames(vcov.object),name.param)){
        stop("Rownames names of the variance covariance matrix does not match the one of the coefficients \n")
    
    }

    ## ** convert to the appropriate format    
    out <- list(model = object,
                linfct = linfct,
                rhs = unname(rhs),
                coef = seq.param,
                vcov = vcov.object,
                df = df.global,
                alternative = "two.sided",
                type = NULL,
                robust = robust)
    class(out) <- c("glht2","glht")
        
    ### ** export
    return(out)    
}


## * glht.lvmfit2
#' @rdname glht2
#' @export
glht.lvmfit2 <- function(model, linfct, rhs = NULL,
                         robust = FALSE, cluster = NULL,
                         ...){

    out <- compare2(model, linfct = linfct, rhs = rhs,
                    robust = robust, cluster = cluster,
                    as.lava = FALSE, F.test = FALSE, ...)
        
    return(out)
}

## * .calcClosure
.calcClosure <- function(name, estimate, covariance, type, df){

    n.hypo <- length(name)
    correlation <- stats::cov2cor(covariance)

    ## ** create all possible hypotheses
    ls.closure <- lapply(n.hypo:1, function(iNtest){ ## iNtest <- 1  
        iList <- list(M = utils::combn(name, m = iNtest))
        iList$vec <- apply(iList$M, 2, paste, collapse = ",")
        return(iList)
    })

    ## ** compute all p.values
    for(iLevel in 1:length(ls.closure)){ ## iLevel <- 1
        ls.closure[[iLevel]]$test <- t(apply(ls.closure[[iLevel]]$M, 2, function(iHypo){
            index <- which(name %in% iHypo)
            if(type == "chisq"){
                return(.ChisqTest(estimate[index], covariance = covariance[index,index,drop=FALSE], df = df))
            }else if(type == "max"){
                return(.tTest(estimate[index],
                              covariance = covariance[index,index,drop=FALSE],
                              correlation = correlation[index,index,drop=FALSE], df = df))
            }
        }))
        rownames(ls.closure[[iLevel]]$test) <- ls.closure[[iLevel]]$vec
    }
    
    ## ** find all hypotheses in the closure related to an individual hypothesis
    ls.hypo <- vector(mode = "list", length = n.hypo)
    for(iHypo in 1:n.hypo){ ## iHypo <- 1
        ls.hypo[[iHypo]] <- do.call(rbind,lapply(ls.closure, function(iClosure){ ## iClosure <- 1
            iIndex <- which(colSums(iClosure$M==name[iHypo])>0)
            data.frame(hypothesis = iClosure$vec[iIndex],
                       statistic = as.double(iClosure$test[iIndex,"statistic"]),
                       p.value = as.double(iClosure$test[iIndex,"p.value"]))
        }))
    }
    names(ls.hypo) <- name
        
    ## ** adjusted p.values
    vec.p.value <- unlist(lapply(ls.hypo, function(x){max(x$p.value)}))
    return(list(closure = ls.closure,
                test = ls.hypo,
                p.value = vec.p.value))
    
}

## * .tTest
.tTest <- function(estimate, covariance, correlation, df, ...){
    df1 <- length(estimate)
    statistic <- max(abs(estimate/sqrt(diag(covariance))))
    if(is.null(df)){
        distribution <-  "gaussian"
    }else{
        distribution <- "student"
    }
    p.value <- .calcPmaxIntegration(statistic, p = df1, Sigma = correlation, df = df,
                                    distribution = distribution)
    return(c("statistic" = statistic,
             "p.value" = p.value))
}

## * .ChisqTest
.ChisqTest <- function(estimate, covariance, df, ...){
    df1 <- length(estimate)
    ## q * statistic ~ chisq or fisher
    statistic <- as.double(matrix(estimate, nrow = 1) %*% solve(covariance) %*% matrix(estimate, ncol = 1)) / df1
    if(!is.null(df)){
        return(c("statistic" = statistic,
                 "p.value" = 1-stats::pf(statistic, df1 = df1, df2 = df)))
    }else{
        return(c("statistic" = statistic,
                 "p.value" = 1-stats::pchisq(statistic, df = df1)))
        
    }
}


 

### test-mmm2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 29 2017 (15:22) 
## Version: 
## Last-Updated: Jan 12 2022 (14:46) 
##           By: Brice Ozenne
##     Update #: 130
##----------------------------------------------------------------------
## 
### Commentary: 
## Test battery:
##  - list of linear regressions:
##      Compare multcomp:::glht to lavaSearch2:::glht2 (no small sample adjustment)
##      * standard error derived from the information matrix
##      * robust standard error derived from the iid decomposition
##
##  - latent variable models:
##      Compare multcomp:::glht to lavaSearch2:::glht2 (no small sample adjustment)
##      * standard error derived from the information matrix, with or without second member
##
##  - list of latent variable model (for linear regressions):
##      Compare multcomp:::glht to lavaSearch2:::glht2 (no small sample adjustment)
##      * standard error derived from the information matrix
##
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

library(multcomp)
library(sandwich)
library(emmeans)
lava.options(symbols = c("~","~~"))

context("multcomp - mmm")

## * simulation
mSim <- lvm(c(Y1,Y2,Y3,Y4)~ beta * eta,
            E ~ 1, Y1 ~ 0.25*T1 + 0.5*T2 + 0.05*T3)
latent(mSim) <- "eta"
set.seed(10)
n <- 1e2

df.data <- lava::sim(mSim, n, latent = FALSE, p = c(beta = 1))
df.data$eY1 <- exp(df.data$Y1)

## * linear regressions with logical constrains
e.lm <- lm(Y1 ~ T1 + T2 + T3, data = df.data)
e.lvm <- estimate(lvm(Y1 ~ T1 + T2 + T3), data = df.data)
## summary(e.lm)

test_that("glht vs. glht2 (logical constrains)", {
    e.glht <- glht(e.lm, linfct = c("T2-T1=0",
                                    "T2-T3=0",
                                    "T1-T3=0"))
    ## summary(e.glht, test = adjusted("none"))
    ## summary(e.glht, test = adjusted("bonferroni"))

    e.glht2 <- glht2(e.lvm, linfct = c("Y1~T2-Y1~T1=0",
                                       "Y1~T2-Y1~T3=0",
                                       "Y1~T1-Y1~T3=0"))
    
    expect_equal(unname(e.glht$vcov),unname(e.glht2$vcov[1:4,1:4]), tol = 1e-6)
    expect_equal(unname(e.glht$coef),unname(e.glht2$coef[1:4]), tol = 1e-6)

    eS.glht <- summary(e.glht, test = adjusted("Shaffer"))
    eS.glht2 <- summary(e.glht2, test = adjusted("Shaffer"))

    expect_equivalent(eS.glht$test[c("coefficients","sigma","tstat","pvalues")],
                      eS.glht2$test[c("coefficients","sigma","tstat","pvalues")], tol = 1e-6)
})

test_that("glht2 (back-transformation)", {
    e.log.lvm <- estimate(lvm(log(eY1) ~ T1 + T2 + T3), data = df.data)

    e.glht2 <- glht2(e.log.lvm, linfct = c("eY1~T1","eY1~T2","eY1~T3"))
    df.glht2 <- summary(e.glht2, transform = "exp", test = adjusted("none"))$table2

    e.glht2.bis <- glht2(e.log.lvm, linfct = "eY1~T3")
    df.glht2.bis <- summary(e.glht2.bis, transform = exp, test = adjusted("none"))$table2

    expect_equal(as.double(df.glht2[3,]) , as.double(df.glht2.bis[1,]))
})

## * list of linear regressions
name.Y <- setdiff(endogenous(mSim),"E")[1:2]
n.Y <- length(name.Y)

ls.lm <- lapply(name.Y, function(iY){
    eval(parse( text = paste("lm(",iY,"~E, data = df.data)")))
})
names(ls.lm) <- name.Y
class(ls.lm) <- "mmm"

ls.lvm <- lapply(name.Y, function(iY){
    eval(parse( text = paste("estimate(lvm(",iY,"~E), data = df.data)")))
})
names(ls.lvm) <- name.Y
class(ls.lvm) <- "mmm"

test_that("glht vs. glht2 (list lm): information std", {
    e.glht <- glht(ls.lm, mlf("E = 0"))
    e.glht2 <- glht2(ls.lvm, linfct = "E")
    e.glht2C <- glht2(ls.lvm, linfct = createContrast(ls.lvm, linfct = "E")$contrast)

    eS.glht <- summary(e.glht)
    eS.glht2 <- summary(e.glht2)
    eS.glht2C <- summary(e.glht2C)

    expect_equivalent(eS.glht2$test, eS.glht2C$test, tol = 1e-6)
    expect_equal(unname(eS.glht$test$tstat), unname(eS.glht2$test$tstat), tol = 1e-6)
})
     
test_that("glht vs. glht2 (list ml): robust std", {
    e.glht <- summary(glht(ls.lm, mlf("E = 0"), vcov = sandwich))
    e.lava <- rbind(estimate(ls.lm[[1]])$coefmat[2,,drop=FALSE],
                    estimate(ls.lm[[2]])$coefmat[2,,drop=FALSE])
    ## no correction for the score
    e.glht0 <- summary(glht2(ls.lvm, linfct = "E", robust = TRUE, ssc = "residuals0"))
    ## correction for the score by inflating the residuals such that they have correct variance
    e.glht2 <- summary(glht2(ls.lvm, linfct = "E", robust = TRUE))
    e.glht2C <- summary(glht2(ls.lvm, linfct = createContrast(ls.lvm, linfct = "E")$contrast, robust = TRUE))

    expect_equivalent(e.glht0$test$tstat, e.glht$test$tstat, tol = 1e-6)
    ## cannot compare p.values
    ## because some are based on a student law and others on a gaussian law

    expect_equivalent(e.glht2$test, e.glht2C$test, tol = 1e-6)
    expect_equivalent(e.glht2$test$tstat, e.glht$test$tstat*sqrt(coef(estimate2(ls.lvm[[1]], ssc = "none"))["Y1~~Y1"])/sigma(ls.lm[[1]]), tol = 1e-6)
})


test_that("glht vs. calcDistMaxIntegral", {
    e.glht <- glht(ls.lm, mlf("E = 0"), vcov = sandwich)
    res.GS <- summary(e.glht)

    iid.tempo <- do.call(cbind,lapply(ls.lm, iid)) %*% t(e.glht$linfct)
    beta <- unlist(lapply(ls.lm, coef)) %*% t(e.glht$linfct)
    beta.var <- diag(crossprod(iid.tempo))
    z.value <- beta/sqrt(beta.var)
    res.Search <- calcDistMaxIntegral(as.vector(z.value),
                                      iid = iid.tempo, quantile.compute = FALSE,
                                      df = NULL, trace = FALSE, alpha = 0.05)
    
    expect_equal(as.double(res.Search$p.adjust),
                 as.double(res.GS$test$pvalues),
                 tolerance = attr(res.GS$test$pvalues, "error")
                 )
    ## cannot compare p.values
    ## because some are based on a student law and others on a gaussian law
})



## * lvm
## names(df.data)

m.lvm <- lvm(c(Y1,Y2,Y3)~ eta, eta ~ E,
             Y1~Y4)
e.lvm <- estimate(m.lvm, df.data)

test_that("glht vs. glht2 (lvm): information std", {

    resC <- createContrast(e.lvm, linfct = c("eta~E","Y2=1","Y3=1"))
    e.glht.null <- glht(e.lvm, linfct = resC$contrast)
    e.glht.H1 <- glht(e.lvm, linfct = resC$contrast, rhs = resC$null)

    e.glht2.null <- glht2(e.lvm, linfct = resC$contrast, rhs = rep(0,3),
                          ssc = "none")
    e.glht2.H1 <- glht2(e.lvm, linfct = resC$contrast, rhs = resC$null,
                        ssc = "none")


    eS.glht.null <- summary(e.glht.null)
    eS.glht.H1 <- summary(e.glht.H1)
    eS.glht2.null <- summary(e.glht2.null)
    eS.glht2.H1 <- summary(e.glht2.H1)

    expect_equal(unname(eS.glht.null$test$tstat),
                 unname(eS.glht2.null$test$tstat))
    expect_equal(unname(eS.glht.H1$test$tstat),
                 unname(eS.glht2.H1$test$tstat))
    ## cannot compare p.values
    ## because some are based on a student law and others on a gaussian law
})

## * list of lvm
mmm.lvm <- mmm(Y1 = estimate(lvm(Y1~E), data = df.data),
               Y2 = estimate(lvm(Y2~E), data = df.data),
               Y3 = estimate(lvm(Y3~E), data = df.data),
               Y4 = estimate(lvm(Y4~E), data = df.data)
               )

test_that("glht2 (list lvm): information std", {

    ##    
    resC <- createContrast(mmm.lvm, linfct = paste0("Y",1:4,": Y",1:4,"~E"))
    lvm2.glht <- glht2(mmm.lvm, linfct = resC$contrast,
                       ssc = NA, robust = FALSE)
    lvm2.sglht <- summary(lvm2.glht)    
    expect_equal(lvm2.sglht$df,100)
    
    lvm3.glht <- glht2(mmm.lvm, linfct = resC$contrast,
                       rhs = rnorm(4),
                       ssc = NA, robust = FALSE)
    lvm3.sglht <- summary(lvm3.glht)    
    expect_equal(lvm3.sglht$df,100)
})

##----------------------------------------------------------------------
### test-mmm2.R ends here

### estimate2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 16 2018 (16:38) 
## Version: 
## Last-Updated: feb 15 2019 (14:08) 
##           By: Brice Ozenne
##     Update #: 864
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * estimate2
#' @title Compute Bias Corrected Quantities.
#' @description Compute bias corrected residuals variance covariance matrix
#' and information matrix.
#' Also provides the leverage values and corrected sample size when adjust.n is set to TRUE.
#' @name estimate2
#' 
#' @keywords internal
.estimate2 <- function(object, epsilon, n.cluster,
                       name.param, name.endogenous, name.meanparam, name.varparam,
                       index.Omega,
                       adjust.Omega, adjust.n, tol, n.iter, trace){

    ## ** Prepare
    Omega <- object$conditionalMoment$Omega
    dmu <- object$conditionalMoment$dmu
    dOmega <- object$conditionalMoment$dOmega
    
    name.hybridparam <- intersect(name.meanparam, name.varparam)

    n.param <- length(name.param)
    n.meanparam <- length(name.meanparam)
    n.varparam <- length(name.varparam)
    n.hybridparam <- length(name.hybridparam)

    n.endogenous <- length(name.endogenous)
    grid.meanparam <- .combination(name.meanparam, name.meanparam)    
    n.grid.meanparam <- NROW(grid.meanparam)
    grid.varparam <- .combination(name.varparam, name.varparam)
    n.grid.varparam <- NROW(grid.varparam)

    ## check low diagonal
    name2num <- setNames(1:n.param,name.param)
    if(!all(name2num[grid.meanparam[,1]]-name2num[grid.meanparam[,2]]>=0)){
        stop("Incorrect allocation of the computation of the information matrix (mean parameter) \n")
    }
    name2num <- setNames(1:n.param,name.param)
    if(!all(name2num[grid.varparam[,1]]-name2num[grid.varparam[,2]]>=0)){
        stop("Incorrect allocation of the computation of the information matrix (variance parameter) \n")
    }
    ##
    
    leverage <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                       dimnames = list(NULL, name.endogenous))
    ls.dmu <- vector(mode = "list", length = n.cluster)
    for(iC in 1:n.cluster){ # iC <- 1
        if(is.null(index.Omega)){            
            leverage[iC,] <- 0
            ls.dmu[[iC]] <- matrix(0, nrow = n.param, ncol = n.endogenous,
                                   dimnames = list(name.param, name.endogenous))
            ls.dmu[[iC]][name.meanparam,] <- do.call(rbind, lapply(dmu[name.meanparam],function(x){x[iC,]}))
        }else{
            leverage[iC,index.Omega[[iC]]] <- 0
            ls.dmu[[iC]] <- matrix(0, nrow = n.param, ncol = length(index.Omega[[iC]]),
                                   dimnames = list(name.param, name.endogenous[index.Omega[[iC]]]))
            ls.dmu[[iC]][name.meanparam,] <- do.call(rbind, lapply(dmu[name.meanparam],function(x){x[iC,index.Omega[[iC]]]}))
        }        
    }
    
    ## ** Initialisation (i.e. first iteration without correction)
    if(any(eigen(Omega)$value<=0)){
        stop("the residual variance-covariance matrix is not positive definite \n")
    }

    if(is.null(index.Omega)){
        n.corrected <- rep(n.cluster, n.endogenous)
    }else{
        n.corrected <- NULL
    }
    ls.Psi <- vector(mode = "list", length = n.cluster)

    Omega.adj <- Omega
    if(!adjust.n){
       epsilon.adj <- epsilon
    }

    if(trace>0){
        cat("* Reconstruct estimated information matrix ")
    }

    iInfo <- .information2(dmu = dmu,
                           dOmega = dOmega,
                           Omega = Omega,
                           n.corrected = n.corrected,
                           leverage = leverage, index.Omega = index.Omega, n.cluster = n.cluster,
                           grid.meanparam = grid.meanparam,
                           n.grid.meanparam = n.grid.meanparam,
                           grid.varparam = grid.varparam,
                           n.grid.varparam = n.grid.varparam,
                           name.param = name.param,
                           n.param = n.param)
    iVcov.param <- try(chol2inv(chol(iInfo)), silent = TRUE)
    if(inherits(iVcov.param, "try-error")){
        iVcov.param <- solve(iInfo)
    }
    if(trace>0){
        cat("- done \n")
    }
    
    ## ** Loop    
    if(adjust.Omega || adjust.n){
        if(trace>0){
            cat("* iterative small sample correction: ")
        }
        iIter <- 0
        iTol <- Inf
        Omega_save <- Omega
        iOmega.adj <- Omega.adj
    }else{
        iIter <- Inf
        iTol <- -Inf        
    }
    
    while(iIter < n.iter & iTol > tol){
        if(trace>0){
            cat("*")
        }

        ## *** Step (i-ii): compute individual bias, expected bias
        Psi <- matrix(0, nrow = n.endogenous, ncol = n.endogenous,
                      dimnames = list(name.endogenous, name.endogenous))
        M.countCluster <- matrix(0, nrow = n.endogenous, ncol = n.endogenous,
                                 dimnames = list(name.endogenous, name.endogenous))
        for(iC in 1:n.cluster){
            ## individual bias
            ls.Psi[[iC]] <- t(ls.dmu[[iC]])  %*% iVcov.param %*% ls.dmu[[iC]]
            ## cumulated bias            
            if(is.null(index.Omega)){
                Psi <- Psi + ls.Psi[[iC]]
                M.countCluster <- M.countCluster + 1
            }else{
                Psi[index.Omega[[iC]],index.Omega[[iC]]] <- Psi[index.Omega[[iC]],index.Omega[[iC]]] + ls.Psi[[iC]]
                M.countCluster[index.Omega[[iC]],index.Omega[[iC]]] <- M.countCluster[index.Omega[[iC]],index.Omega[[iC]]] + 1
            }
        }

        ## update
        for(iPsi in 1:length(Psi)){
            if(M.countCluster[iPsi]>0){
                Psi[iPsi] <- Psi[iPsi]/M.countCluster[iPsi]
            }
        }
        
        ## *** Step (iii): compute leverage
        if(adjust.n){
            epsilon.adj <- .adjustResiduals(Omega = Omega.adj,
                                            Psi = Psi,
                                            epsilon = epsilon,
                                            index.Omega = index.Omega,
                                            name.endogenous = name.endogenous,
                                            n.endogenous = n.endogenous,
                                            n.cluster = n.cluster)

            leverage <- .adjustLeverage(Omega = Omega.adj,
                                        epsilon = epsilon.adj,
                                        ls.dmu = ls.dmu,
                                        dOmega = dOmega,
                                        vcov.param = iVcov.param,
                                        index.Omega = index.Omega,
                                        name.endogenous = name.endogenous,
                                        n.endogenous = n.endogenous,
                                        name.varparam = name.varparam,
                                        n.varparam = n.varparam,
                                        n.cluster = n.cluster)

            n.corrected <- rep(n.cluster, n.endogenous) - colSums(leverage, na.rm = TRUE)
        }
        
        ## *** Step (v): correct residual covariance matrix, estimates, and derivatives
        if(adjust.Omega){
            ## corrected residual covariance variance
            Omega.adj <- Omega + Psi
            
            ## correct estimates
            object$conditionalMoment <- .adjustMoment(object, Omega = Omega.adj)
            dOmega <- object$conditionalMoment$dOmega
            ## conditionalMoment.adj$param - coef(object)
           
        }

        ## *** Step (vii): expected information matrix
        iInfo <- .information2(dmu = dmu,
                               dOmega = dOmega,
                               Omega = Omega.adj,
                               n.corrected = n.corrected,
                               leverage = leverage,
                               index.Omega = index.Omega,
                               n.cluster = n.cluster,
                               grid.meanparam = grid.meanparam,
                               n.grid.meanparam = n.grid.meanparam,
                               grid.varparam = grid.varparam,
                               n.grid.varparam = n.grid.varparam,
                               name.param = name.param,
                               n.param = n.param)
        iVcov.param <- try(chol2inv(chol(iInfo)), silent = TRUE)
        if(inherits(iVcov.param, "try-error")){
            iVcov.param <- solve(iInfo)
        }
        
        ## *** Update cv
        iIter <- iIter + 1
        iTol <- norm(Omega.adj-Omega_save, type = "F")
        Omega_save <- Omega.adj
        ## cat("Omega.adj: ",Omega.adj," | n:",n.corrected," | iTol:",iTol,"\n")
    }
    
    ## ** Post processing
    if(!is.infinite(iIter)){

        if(iTol > tol){
            warning("small sample correction did not reach convergence after ",iIter," iterations \n")

            if(trace>0){
                cat(" - incomplete \n")
            }
        }else{
            if(trace>0){
                cat(" - done \n")
            }
        }
        
    }

    vcov.param <- try(chol2inv(chol(iInfo)), silent = TRUE)
    if("try-error" %in% class(vcov.param)){
        errorMessage <- vcov.param
        vcov.param <- solve(iInfo)
        attr(vcov.param, "warning") <- errorMessage
    }
    dimnames(vcov.param) <- dimnames(iInfo)

    ## update object
    object$conditionalMoment$Omega <- Omega.adj
    object$dVcov <- list(param = object$conditionalMoment$param,
                         score = NULL,
                         vcov.param = vcov.param,
                         dVcov.param = NULL,
                         Omega = Omega.adj,
                         residuals = epsilon.adj,
                         leverage = leverage,
                         n.corrected = rep(n.cluster, n.endogenous) - colSums(leverage, na.rm = TRUE),
                         opt = list(objective = iTol, iterations = iIter, convergence = (iTol <= tol), grid.meanparam = grid.meanparam, grid.varparam = grid.varparam))

    ## ** Export
    return(object)
}

## * .adjustResiduals
.adjustResiduals <- function(Omega, Psi, epsilon,
                             index.Omega,
                             name.endogenous, n.endogenous, n.cluster){

    if(is.null(index.Omega)){ ## no missing values
        
        Omega.chol <- matrixPower(Omega, symmetric = TRUE, power = 1/2)
        H <- Omega %*% Omega - Omega.chol %*% Psi %*% Omega.chol
        HM1 <- tryCatch(matrixPower(H, symmetric = TRUE, power = -1/2), warning = function(w){w})
        if(inherits(HM1,"warning")){
            stop("Cannot compute the adjusted residuals \n",
                 "Estimated bias too large compared to the estimated variance-covariance matrix \n",
                 "Consider setting argument \'adjust.n\' to FALSE when calling sCorrect \n")
        }
        epsilon.adj <- epsilon %*% Omega.chol %*% HM1 %*% Omega.chol
        
    }else{ ## missing values
        
        epsilon.adj <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                              dimnames = list(NULL, name.endogenous))

        for(iC in 1:n.cluster){
            iIndex <- index.Omega[[iC]]
            iOmega <- Omega[iIndex,iIndex,drop=FALSE]
            iOmega.chol <- matrixPower(iOmega, symmetric = TRUE, power = 1/2)
            iH <- iOmega %*% iOmega - iOmega.chol %*% Psi[iIndex,iIndex,drop=FALSE] %*% iOmega.chol
            iHM1 <- tryCatch(matrixPower(iH, symmetric = TRUE, power = -1/2), warning = function(w){w})
            if(inherits(iHM1,"warning")){
                stop("Cannot compute the adjusted residuals \n",
                     "Estimated bias too large compared to the estimated variance-covariance matrix \n",
                     "Consider setting argument \'adjust.n\' to FALSE when calling sCorrect \n")
            }
            epsilon.adj[iC,iIndex] <- epsilon[iC,iIndex] %*% iOmega.chol %*% iHM1 %*% iOmega.chol
        }
        
    }
    dimnames(epsilon.adj) <- list(NULL,name.endogenous)
    return(epsilon.adj)
}

## * .adjustLeverage
.adjustLeverage <- function(Omega, epsilon, ls.dmu, dOmega, vcov.param,
                            index.Omega,
                            name.endogenous, n.endogenous, name.varparam, n.varparam, n.cluster){

    ## ** prepare
    leverage <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                       dimnames = list(NULL, name.endogenous))

    if(is.null(index.Omega)){
        iIndex <- 1:n.endogenous
        iOmegaM1 <- chol2inv(chol(Omega)) ## solve(Omega)
        iOmegaM1.dOmega.OmegaM1 <- lapply(dOmega, function(x){iOmegaM1 %*% x %*% iOmegaM1})
    }

    ## ** compute
    for(iC in 1:n.cluster){                 # iC <- 1
        if(!is.null(index.Omega)){
            iIndex <- index.Omega[[iC]]
            iOmegaM1 <- chol2inv(chol(Omega[iIndex,iIndex,drop=FALSE]))
            iOmegaM1.dOmega.OmegaM1 <- lapply(dOmega, function(x){iOmegaM1 %*% x[iIndex,iIndex] %*% iOmegaM1})
        }
        ## derivative of the score regarding Y
        scoreY <- ls.dmu[[iC]] %*% iOmegaM1

        for(iP in 1:n.varparam){ ## iP <- 1
            scoreY[name.varparam[iP],] <- scoreY[name.varparam[iP],] + 2 * epsilon[iC,iIndex] %*% iOmegaM1.dOmega.OmegaM1[[name.varparam[iP]]]
        }
        ## leverage
        leverage[iC,iIndex] <- colSums(vcov.param %*% ls.dmu[[iC]] * scoreY) ## NOTE: dimensions of ls.dmu and scoreY matches even when there are missing values
                                        # same as
                                        # diag(t(ls.dmu[[iC]])  %*% iVcov.param %*% scoreY)
    }

    return(leverage)            
}

## * .adjustMoment
`.adjustMoment` <-
    function(object, ...) UseMethod(".adjustMoment")

## * .adjustMoment.lm
.adjustMoment.lm <- function(object, Omega){

    object$conditionalMoment$param["sigma2"] <- as.double(Omega)
    return(object$conditionalMoment)
    
}

## * .adjustMoment.gls
.adjustMoment.gls <- function(object, Omega, ...){

    ## ** extract information
    class.cor <- object$conditionalMoment$skeleton$class.cor
    class.var <- object$conditionalMoment$skeleton$class.var
    name.corcoef <- object$conditionalMoment$skeleton$name.corcoef
    name.otherVar <- object$conditionalMoment$skeleton$name.otherVar
    name.varcoef <- object$conditionalMoment$skeleton$name.varcoef
    ref.group <- object$conditionalMoment$skeleton$ref.group
    M.corcoef <- object$conditionalMoment$skeleton$M.corcoef
    name.endogenous <- object$conditionalMoment$skeleton$name.endogenous
    n.endogenous <- object$conditionalMoment$skeleton$n.endogenous
    
    ## ** identify parameters

    if(identical(class.var, "NULL")){
        object$conditionalMoment$param["sigma2"] <- mean(diag(Omega))
    }else{            
        index.Sigma2 <- which(ref.group %in% name.otherVar == FALSE)
        object$conditionalMoment$param["sigma2"] <- mean(diag(Omega)[index.Sigma2])

        vec.k <- tapply(diag(Omega)/Omega[index.Sigma2,index.Sigma2], ref.group, mean)            
        object$conditionalMoment$param[name.otherVar] <- vec.k[name.otherVar]
    }

    if(identical(class.cor, "NULL")){
        ## do nothing
    }else if("corCompSymm" %in% class.cor){
        object$conditionalMoment$param[name.corcoef] <- mean(stats::cov2cor(Omega)[lower.tri(Omega)])
    }else if("corSymm" %in% class.cor){
        vec.cor <- tapply(stats::cov2cor(Omega)[lower.tri(Omega)],
                          M.corcoef[lower.tri(Omega)],
                          mean)            
        object$conditionalMoment$param[name.corcoef] <- vec.cor[name.corcoef]
    } 

    ## ** update conditional moments
    object$conditionalMoment$Omega <- .getVarCov2(object,
                                                  param = object$conditionalMoment$param,
                                                  attr.param = attributes(object$conditionalMoment$param),
                                                  name.endogenous = name.endogenous,
                                                  n.endogenous = n.endogenous,
                                                  ref.group = ref.group)
    
    ## ** update first derivative of the conditional variance
    object$conditionalMoment$dOmega <- skeletonDtheta(object, class.cor = class.cor, class.var = class.var, 
                                                      sigma2.base0 = object$conditionalMoment$skeleton$sigma2.base0,
                                                      Msigma2.base0 = object$conditionalMoment$skeleton$Msigma2.base0,
                                                      M.corcoef = M.corcoef, ref.group = ref.group,
                                                      index.lower.tri = object$conditionalMoment$skeleton$index.lower.tri,
                                                      indexArr.lower.tri = object$conditionalMoment$skeleton$indexArr.lower.tri,
                                                      name.endogenous =  name.endogenous, n.endogenous = n.endogenous,
                                                      cluster = object$conditionalMoment$skeleton$cluster,
                                                      n.cluster = object$conditionalMoment$skeleton$n.cluster,
                                                      var.coef = object$conditionalMoment$param[name.varcoef],
                                                      name.varcoef = name.varcoef, name.otherVar = name.otherVar,
                                                      n.varcoef = object$conditionalMoment$skeleton$n.varcoef,
                                                      cor.coef = object$conditionalMoment$param[name.corcoef],
                                                      name.corcoef = name.corcoef,
                                                      n.corcoef = object$conditionalMoment$skeleton$n.corcoef,
                                                      update.mean = FALSE, update.variance = TRUE, ...)$dOmega

    ## ** export
    ## names(object$conditionalMoment)
    ## object$conditionalMoment$param["sigma2"] <- as.double(Omega)
    return(object$conditionalMoment)
    
}

## * .adjustMoment.lme
.adjustMoment.lme <- function(object, Omega){

    name.rancoef <- attr(object$conditionalMoment$param,"ran.coef")
    
    ## ** Identify random effect
    if(!identical(object$conditionalMoment$skeleton$class.cor,"NULL")){
        stop("Does not know how to identify the correlation coefficients when corStruct is not NULL \n")
    }
    object$conditionalMoment$param[name.rancoef] <- mean(Omega[lower.tri(Omega)])

    ## ** save derivative regarding random effect
    save <- object$conditionalMoment$dOmega$ranCoef1

    ## ** compute moments 
    conditionalMoment <- .adjustMoment.gls(object, Omega = Omega - object$conditionalMoment$param["ranCoef1"],
                                           name.rancoef = name.rancoef)

    ## ** restaure derivative regarding random effect
    conditionalMoment$dOmega$ranCoef1 <- save
    return(conditionalMoment)
}

## * .adjustMoment.lvmfit
.adjustMoment.lvmfit <- function(object, Omega){

    ## ** extract info
    n.endogenous <- NROW(Omega)
    df.param <- object$conditionalMoment$df.param
    
    index.matrix <- object$conditionalMoment$adjustMoment$index.matrix
    index.Psi <- object$conditionalMoment$adjustMoment$index.Psi
    A <- object$conditionalMoment$adjustMoment$A
    name.var <- object$conditionalMoment$adjustMoment$name.var
    n.rhs <- object$conditionalMoment$adjustMoment$n.rhs
    index.LambdaB <- object$conditionalMoment$adjustMoment$index.LambdaB
    name.endogenous <- object$conditionalMoment$adjustMoment$name.endogenous
    name.latent <- object$conditionalMoment$adjustMoment$name.latent
    
    skeleton <- object$conditionalMoment$skeleton

    param <- object$conditionalMoment$param
    Lambda <- object$conditionalMoment$value$Lambda
    iIB <- object$conditionalMoment$value$iIB
    iIB.Lambda <- object$conditionalMoment$value$iIB.Lambda
    dLambda <- object$conditionalMoment$dMoment.init$dLambda
    dB <- object$conditionalMoment$dMoment.init$dB

    ## ** right hand side of the equation
    eq.rhs <- Omega[index.matrix$index]

    ## ** left hand side of the equation
    if(NROW(index.Psi)>0){
        n.index.Psi <- NROW(index.Psi)
        n.latent <- NROW(skeleton$Psi)        
        Z <- iIB %*% Lambda

        ## A = t(Z) Psi Z + Sigma
        ## (t(Z) Psi Z)_{ij} = \sum_{k,l} Z_{k,i} Psi_{k,l} Z_{l,j}
        for(iIndex in 1:n.rhs){ # iIndex <- 1
            iRow <- index.matrix[iIndex,"row"]
            iCol <- index.matrix[iIndex,"col"]
            for(iPsi in 1:n.index.Psi){
                iRowPsi <- index.Psi[iPsi,"row"]
                iColPsi <- index.Psi[iPsi,"col"]
                A[iIndex,skeleton$Psi[iRowPsi,iColPsi]] <- A[iIndex,skeleton$Psi[iRowPsi,iColPsi]] + Z[iRowPsi,iRow]*Z[iColPsi,iCol]
            }
        }
    }

    ## ** solve equation
    ## microbenchmark::microbenchmark(svd = {asvd <- svd(A) ; asvd$v %*% diag(1/asvd$d) %*% t(asvd$u) %*% eq.rhs;},
    ## qr = qr.coef(qr(A), eq.rhs),
    ## Rcpp = OLS_cpp(A, eq.rhs),
    ## RcppTry = try(OLS_cpp(A, eq.rhs)[,1], silent = TRUE),
    ## Rcpp2 = OLS2_cpp(A, eq.rhs),
    ## OLS1 = solve(crossprod(A), crossprod(A, eq.rhs)),
    ## OLS2 = solve(t(A) %*% A) %*% t(A) %*% eq.rhs,
    ## OLS_stats = stats::lsfit(x = A, y = eq.rhs),
    ## OLS_LINPACK = .Call(stats:::C_Cdqrls, x = A, y = eq.rhs, tolerance = 1e-7, FALSE)$coefficients, times = 500)
    if(lava.options()$method.estimate2=="svd"){
        asvd <- svd(A)
        iSolution <- try((asvd$v %*% diag(1/asvd$d) %*% t(asvd$u) %*% eq.rhs)[,1], silent = TRUE)
    }else if(lava.options()$method.estimate2=="ols"){
        iSolution <- try(OLS_cpp(A, eq.rhs)[,1], silent = TRUE)
    }else{
        stop("unknown OLS methods \n")
    }
    
    if(inherits(iSolution, "try-error")){
        if(abs(det(t(A) %*% A)) <  1e-10){            
            stop("Singular matrix: cannot update the estimates \n")
        }else{
            stop(iSolution)
        }
    }

    ## ** update parameters in conditional moments
    object$conditionalMoment$param[name.var] <- setNames(iSolution, name.var)

    ## ** update conditional moments
    object$conditionalMoment$skeleton$toUpdate <- object$conditionalMoment$adjustMoment$toUpdate
    object$conditionalMoment$value <- skeleton.lvmfit(object,
                                                      param = param,
                                                      data = NULL,
                                                      name.endogenous = name.endogenous,
                                                      name.latent = name.latent)
    object$conditionalMoment$Omega <- Omega
    

    ## ** update first derivative of the conditional variance
    if(length(index.LambdaB)>0){
        object$conditionalMoment$dMoment.init$toUpdate[] <- FALSE
        object$conditionalMoment$dMoment.init$toUpdate[index.LambdaB] <- TRUE

        object$conditionalMoment$dOmega <- skeletonDtheta.lvmfit(object,
                                                                name.endogenous = name.endogenous,
                                                                name.latent = name.latent)$dOmega
    }
    ## ** export
    return(object$conditionalMoment)
}

##----------------------------------------------------------------------
### estimate2.R ends here

### test1-adjustMoment.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar 27 2018 (09:50) 
## Version: 
## Last-Updated: jul 16 2018 (16:49) 
##           By: Brice Ozenne
##     Update #: 22
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:


## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}
lava.options(symbols = c("~","~~"))

.adjustMoment <- lavaSearch2:::.adjustMoment
context("adjustMoment")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
           measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * multiple linear regression
## ** no constrains
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

test_that("linear regression - no constrains",{
    Omega <- getVarCov2(e.lvm)
    e.lvm$conditionalMoment <- conditionalMoment(e.lvm,
                                                 data = d,
                                                 param = coef(e.lvm),
                                                 first.order = TRUE,
                                                 second.order = FALSE,
                                                 usefit = TRUE)    
    solution <- .adjustMoment(e.lvm, Omega = Omega)

    expect_equal(solution$param, coef(e.lvm)[names(solution$param)], tol = 1e-8)
})

## ** constrains and covariance link
m <- lvm(Y1[mu1:sigma]~X1,
         Y2[mu2:sigma]~X2,Y3~X1+X3,
         Y2~~Y3)
e.lvm <- estimate(m, data = d)

test_that("linear regression - constrains and covariance",{
    Omega <- getVarCov2(e.lvm)
    e.lvm$conditionalMoment <- conditionalMoment(e.lvm,
                                                 data = d,
                                                 param = coef(e.lvm),
                                                 first.order = TRUE,
                                                 second.order = FALSE,
                                                 usefit = TRUE)    
    solution <- .adjustMoment(e.lvm, Omega = Omega)

    expect_equal(solution$param, coef(e.lvm)[names(solution$param)], tol = 1e-8)
})

## * factor model
m <- lvm(Y1~eta,
         Y2~eta+X2,
         Y3~eta,
         Z1~eta, Z1~~Y1,Z1~~Y2,
         eta~X1+X3)
e.lvm <- estimate(m, d)

test_that("factor model",{
    Omega <- getVarCov2(e.lvm)
    e.lvm$conditionalMoment <- conditionalMoment(e.lvm,
                                                 data = d,
                                                 param = coef(e.lvm),
                                                 first.order = TRUE,
                                                 second.order = FALSE,
                                                 usefit = TRUE)    
    solution <- .adjustMoment(e.lvm, Omega = Omega)

    expect_equal(solution$param, coef(e.lvm)[names(solution$param)], tol = 1e-8)
})

## * two factor model
## ** correlation
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3,
           eta1~eta2))

e.lvm <- estimate(m, d)

test_that("two factor model - correlation",{
    Omega <- getVarCov2(e.lvm)
    e.lvm$conditionalMoment <- conditionalMoment(e.lvm,
                                                 data = d,
                                                 param = coef(e.lvm),
                                                 first.order = TRUE,
                                                 second.order = FALSE,
                                                 usefit = TRUE)    
    solution <- .adjustMoment(e.lvm, Omega = Omega)

    expect_equal(solution$param, coef(e.lvm)[names(solution$param)], tol = 1e-8)
})

## ** covariance
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3,
           eta1~~eta2))

e.lvm <- estimate(m, d)

test_that("two factor model - covariance",{
    Omega <- getVarCov2(e.lvm)
    e.lvm$conditionalMoment <- conditionalMoment(e.lvm,
                                                 data = d,
                                                 param = coef(e.lvm),
                                                 first.order = TRUE,
                                                 second.order = FALSE,
                                                 usefit = TRUE)    
    solution <- .adjustMoment(e.lvm, Omega = Omega)

    expect_equal(solution$param, coef(e.lvm)[names(solution$param)], tol = 1e-8)
})

##
##----------------------------------------------------------------------
### test1-adjustMoment.R ends here

#' @docType package
#' @name lavaSearch2
#' @aliases lavaSearch2, lavaSearch2-package
#'
#' @title Tools for Model Specification in the Latent Variable Framework
#' @description
#' The package contains three main functionalities:
#' \itemize{
#' \item \code{\link{compare2}}, \code{\link{summary2}}: Wald tests/robust Wald tests/F-tests/robust F-tests with improved control of the type 1 error in small samples.
#' \item \code{\link{glht2}}: adjustment for multiple comparisons when doing inference for multiple latent variable models.
#' \item \code{\link{modelsearch2}}: searching for local dependencies with adjustment for multiple comparisons.
#' }
#' It contains other useful functions such as:
#'\itemize{
#' \item \code{\link{calibrateType1}}: simulation study of the type 1 error of Wald tests.
#' \item \code{\link{createContrast}}: user-friendly function generating a contrast matrix.
#' \item \code{\link{getVarCov2}}: reconstruct the conditional variance covariance matrix.
#' \item \code{\link{iidJack}}: extract the jackknife iid decomposition.
#' }
#' @details
#' The latent variable models (LVM) considered in this package can be written \cr
#' as a measurement model:
#' \deqn{Y_i = \nu + \eta_i \Lambda + X_i K + \epsilon_i}
#' and a structural model:
#' \deqn{\eta_i = \alpha + \eta_i B + X_i \Gamma + \zeta_i}
#' where \eqn{\Sigma} is the variance covariance matrix of the residuals \eqn{\epsilon}, \cr
#' and \eqn{\Psi}   is the variance covariance matrix of the residuals \eqn{\zeta}. \cr
#' 
#' The corresponding conditional mean is:
#' \deqn{
#' \mu_i(\theta) = E[Y_i|X_i] = \nu + (\alpha + X_i \Gamma) (1-B)^{-1} \Lambda + X_i K
#' }
#' \deqn{
#' \Omega(\theta) = Var[Y_i|X_i] = \Lambda^{t} (1-B)^{-t} \Psi (1-B)^{-1} \Lambda + \Sigma
#' }
#'
#' The package aims to provides tool for testing linear hypotheses on the model coefficients
#' \eqn{\nu}, \eqn{\Lambda}, \eqn{K}, \eqn{\Sigma},
#' \eqn{\alpha}, \eqn{B}, \eqn{\Gamma}, \eqn{\Psi}.
#' Searching for local dependency enable to test whether the proposed model is too simplistic and if so to identify which additional coefficients should be added to the model.
#' 
#'
#'
#' 
#' @section Limitations:
#' 
#' 'lavaSearch2' has been design for Gaussian latent variable models.
#' This means that it may not work / give valid results:
#' \itemize{
#' \item in presence of censored or binary outcomes.
#' \item with stratified models (i.e. object of class \code{multigroup}).
#' } 
#' 

#' @useDynLib lavaSearch2, .registration=TRUE
#' @import lava
#' @importFrom ggplot2  aes_string autoplot
#' @importFrom graphics par plot text
#' @importFrom MASS mvrnorm
#' @importFrom Matrix bdiag
#' @importFrom methods is
#' @importFrom multcomp glht 
#' @importFrom mvtnorm pmvnorm qmvnorm rmvnorm qmvt pmvt
#' @importFrom parallel detectCores makeCluster stopCluster
#' @import Rcpp
#' @importFrom reshape2 melt
#' @importFrom sandwich estfun
#' @importFrom stats anova coef confint cov effects formula logLik model.frame model.matrix model.tables predict qqnorm quantile residuals update vcov
#' @importFrom utils methods packageVersion setTxtProgressBar tail txtProgressBar
#' 
NULL



  

### defineCategoricalLink.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 26 2017 (16:35) 
## Version: 
## last-updated: aug  6 2018 (15:32) 
##           By: Brice Ozenne
##     Update #: 156
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * documentation - defineCategoricalLink
#' @title Identify Categorical Links in LVM
#' @description Identify categorical links in latent variable models.
#' @name defineCategoricalLink
#' 
#' @param object a \code{lvm} model.
#' @param link [character] the links to be analyzed. If \code{NULL}, all the coefficients from the lvm model are used instead.
#' @param data [data.frame] the dataset that will be used to fit the model. If \code{NULL}, a simulated data will be generated from the model.
#'
#' @return a \code{data.frame} with a description of each link in rows. \cr
#' The column factitious identify the links that will be replaced with other links
#' (e.g. "Y1~X1" becomes "Y1~X1b" and "Y1~X1c"). \cr
#' 
#' @examples
#' \dontrun{
#' defineCategoricalLink <- lavaSearch2:::defineCategoricalLink
#' defineCategoricalLink.lvm <- lavaSearch2:::defineCategoricalLink.lvm
#' 
#' ## linear model
#' m <- lvm(Y1~X1+X2,Y2~X1+X3)
#' categorical(m, K = 3) <- "X1"
#' try(defineCategoricalLink(m)) # error
#'
#' categorical(m, K = 3, labels = 1:3) <- "X1"
#' defineCategoricalLink(m)
#' defineCategoricalLink(m, "Y~X1")
#' defineCategoricalLink(m, "X1:0|1")
#' defineCategoricalLink(m, "X1:1|2")
#' defineCategoricalLink(m, c("X1:0|1", "X1:1|2"))
#' defineCategoricalLink(m, c("Y~X1","Y~X2"))
#' defineCategoricalLink(m, c("Y~X2","Y~X1"))
#'
#' ## latent variable model
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' covariance(m) <- y1~y2
#' categorical(m, labels = as.character(1:3)) <- "X1"
#'
#' defineCategoricalLink(m)
#'}
#' 
#' @concept setter
#' @keywords internal 
`defineCategoricalLink` <-
  function(object, link, data) UseMethod("defineCategoricalLink")


## * defineCategoricalLink.lvm
#' @rdname defineCategoricalLink
defineCategoricalLink.lvm <- function(object, link = NULL, data = NULL){

    ### ** normalize arguments
    if(is.null(link)){
        link <- stats::coef(object)
    }
    if(is.null(data)){
        data <- lava::sim(object, 1e2)
    }
    
    ### ** identify the type of regression variable (continuous or categorical)
    index.cat <- which(link %in% unlist(object$attributes$ordinalparname))
    index.Ncat <- setdiff(1:length(link), index.cat)
    link.Ncat <- setdiff(link[index.Ncat], names(object$attributes$ordinalparname))

    ### ** caracterize links involving categorical variables    
    if(length(index.cat)>0){
        link.cat <- link[index.cat]
        xCAT <- lava_categorical2dummy(object, data)$x

        ## *** find exogenous variable
        X.name.cat <- sapply(link.cat, function(iL){
            test <- unlist(lapply(object$attributes$ordinalparname, function(vec.coef){iL %in% vec.coef}))
            return(names(object$attributes$ordinalparname)[test])
        })
        UX.name.cat <- unique(X.name.cat)
    
        ## *** find the level of the exogenous variable
        X.level.cat <- unlist(lapply(UX.name.cat, function(iL){ 
            if(iL %in% names(xCAT$attributes$labels)){
                labels <- xCAT$attributes$labels[[iL]]
                index.label <- which(object$attributes$ordinalparname[[iL]] %in% link.cat)                
                return(labels[1+index.label])
            }else {
                stop("Categorical variables must have labels. Specify argument \'labels\' when calling categorical. \n")
            }            
        }))

        ## *** find endogenous variable
        M.link <- xCAT$M[paste0(X.name.cat,X.level.cat),,drop = FALSE]
        M.link <- cbind(M.link, as.numeric(rowSums(M.link)==0))

        convertion.back <- stats::setNames(X.name.cat,paste0(X.name.cat,X.level.cat))
                
        indexLink <- which(M.link==1, arr.ind = TRUE)
        Y.name.allcat <- colnames(M.link)[indexLink[,"col"]]
        X.name.allcat <- as.character(convertion.back[rownames(M.link)][indexLink[,"row"]])
        
        ## *** characterize all links
        Xcat.name.allcat <- rownames(M.link)[indexLink[,"row"]]
        X.level.allcat <- as.character(X.level.cat[indexLink[,"row"]])
        external.link.allcat <- link[index.cat[indexLink[,"row"]]]
        original.link.allcat <- paste0(Y.name.allcat, lava.options()$symbol[1], X.name.allcat)
        original.link.allcat[Y.name.allcat == ""] <- gsub("~","",original.link.allcat[Y.name.allcat == ""])
        cat.link.allcat <- paste0(Y.name.allcat, lava.options()$symbol[1], Xcat.name.allcat)
        cat.link.allcat[Y.name.allcat == ""] <- gsub("~","",cat.link.allcat[Y.name.allcat == ""])
        
        df.cat <- data.frame(link = cat.link.allcat,
                             endogenous = Y.name.allcat,
                             exogenous = X.name.allcat,
                             type = "categorical",
                             factitious = FALSE,
                             level = X.level.allcat,
                             originalLink = original.link.allcat,
                             externalLink = external.link.allcat,
                             stringsAsFactors = FALSE)       

    }else{
            df.cat <- NULL
    }

### ** caracterize links involving continuous variables    
    if(length(index.Ncat)>0){

        var.tempo <- initVarLinks(link.Ncat)
        Y.name.Ncat <- var.tempo$var1
        X.name.Ncat <- var.tempo$var2
        test.factitious <- X.name.Ncat %in% names(object$attributes$ordinalparname)

        df.Ncat <- data.frame(link = link.Ncat,
                              endogenous = Y.name.Ncat,
                              exogenous = X.name.Ncat,
                              type = "continuous",
                              factitious = test.factitious,
                              level = NA,
                              originalLink = link.Ncat,
                              externalLink = NA,
                              stringsAsFactors = FALSE)       
    }else{
        df.Ncat <- NULL
    }
    
    ### ** export
    out <- rbind(df.Ncat,df.cat)
    return(out)
}

#----------------------------------------------------------------------
### defineCategoricalLink.R ends here

### Scorrect-confint2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: Jan  4 2022 (10:59) 
## Version: 
## Last-Updated: jan 23 2024 (10:27) 
##           By: Brice Ozenne
##     Update #: 101
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:


## * Documentation
#' @title Confidence Intervals With Small Sample Correction
#' @description Extract confidence intervals of the coefficients from a latent variable model.
#' Similar to \code{lava::confint} but with small sample correction.
#' @name confint2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param robust [logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] when \code{TRUE} uses the same names as when using \code{stats::coef}.
#' @param transform [function] transformation to be applied.
#' @param conf.level [numeric, 0-1] level of the confidence intervals.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
#' 
#' @return A data.frame with a row per coefficient.
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`confint2` <- function(object, robust, cluster, transform,
                       as.lava, conf.level, ...) UseMethod("confint2")

## * Examples
#' @rdname confint2
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' dW <- sampleRepeated(10, format = "wide")
#' set.seed(10)
#' dL <- sampleRepeated(10, format = "long")
#' dL$time2 <- paste0("visit",dL$time)
#' 
#' #### latent variable models ####
#' e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
#' confint(e.lvm)
#' confint2(e.lvm)
#' confint2(e.lvm, as.lava = FALSE)

## * confint2.lvmfit
#' @export
confint2.lvmfit <- function(object, robust = FALSE, cluster = NULL,
                            transform = NULL, as.lava = TRUE, conf.level = 0.95, 
                            ssc = lava.options()$ssc, df = lava.options()$df, ...){

    return(confint(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...),
                   robust = robust, cluster = cluster, as.lava = as.lava, conf.level = conf.level,
                   transform = transform))

}

## * confint2.lvmfit2
#' @export
confint2.lvmfit2 <- function(object, robust = FALSE, cluster = NULL,
                            transform = NULL, as.lava = TRUE, conf.level = 0.95,  ...){

    out <- model.tables(object, robust = robust, cluster = cluster, transform = transform,
                        as.lava = as.lava, conf.level = conf.level, ...)[,c("lower","upper"),drop=FALSE]
    
    if(as.lava){
        colnames(out) <- c(paste0(100*(1-conf.level)/2," %"),paste0(100*(1-(1-conf.level)/2)," %"))
    }
    return(out)

}

## *  confint.lvmfit2
#' @export
confint.lvmfit2 <- function(object, parm = NULL, level = NULL, ...){ ## necessary as confint must contain arguments parm and level

    if(!is.null(parm)){
        warning("Argument \'parm\' is ignored. \n")
    }
    if(!is.null(level)){
        warning("Argument \'level\' is ignored. \n")
    }
    
    return(confint2(object, ...))

}

##----------------------------------------------------------------------
### Scorrect-confint2.R ends here

### clean.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 27 2018 (14:35) 
## Version: 
## Last-Updated: Jan 11 2022 (16:00) 
##           By: Brice Ozenne
##     Update #: 14
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * clean (Documentation)
#' @title Simplify a lvm object
#' @description Remove variables with no link.
#' @name clean
#' 
#' @param x \code{lvm}-object
#' @param rm.exo  should exogenous variables with no links be removed from the object? 
#' @param rm.endo  should endogenous variables with no links be removed from the object?
#' @param rm.latent  should latent variables with no links be removed from the object?
#' @param ... additional arguments to lower level functions
#'
#' @export
`clean` <-
  function(x,...) UseMethod("clean")

## * clean (examples)
#' @rdname clean
#' @examples 
#' 
#' m <- lvm()
#' m <- regression(m, x=paste0("x",1:5),y="y1")
#' m <- regression(m, x=paste0("x",1:5),y="y2")
#' covariance(m) <- y1~y2
#'
#' cancel(m) <- y1 ~ x1
#' cancel(m) <- y2 ~ x1
#' clean(m)
#'
#' m <- lvm(y1 ~ eta + x1, y2 ~ eta, y3 ~ eta + x2)
#' latent(m) <- ~eta
#' clean(m)
#' m
#' cancel(m) <- y1 ~ eta
#' cancel(m) <- y2 ~ eta
#' cancel(m) <- y3 ~ eta
#' clean(m)

## * clean.lvm
#' @export
clean.lvm <- function(x, rm.exo = TRUE, rm.endo = TRUE, rm.latent = TRUE, ...){

    var.latent <- latent(x)
    var.exogenous <- exogenous(x)
    var.endogenous <- endogenous(x)

    M.reg <- x$index$A
    M.cov <- x$index$P - diag(diag(x$index$P))
    varClean <- NULL

    if(rm.exo && length(var.exogenous) > 0){
        indexClean.reg <- which(rowSums(M.reg[var.exogenous,,drop = FALSE]!=0)==0)
        indexClean.cov <- which(rowSums(M.cov[var.exogenous,,drop = FALSE]!=0)==0)
        indexClean <- intersect(indexClean.reg, indexClean.cov)
        varClean <- c(varClean, var.exogenous[indexClean])
    }
    if(rm.endo && length(var.endogenous)>0){
        indexClean.reg <- which(colSums(M.reg[,var.endogenous,drop = FALSE]!=0)==0)
        indexClean.cov <- which(colSums(M.cov[,var.endogenous,drop = FALSE]!=0)==0)
        indexClean <- intersect(indexClean.reg, indexClean.cov)
        varClean <- c(varClean, var.endogenous[indexClean])
    }
    if(rm.latent && length(var.latent)>0){
        indexClean.Rreg <- which(rowSums(M.reg[var.latent,,drop = FALSE]!=0)==0)
        indexClean.Rcov <- which(rowSums(M.cov[var.latent,,drop = FALSE]!=0)==0)
        indexClean.Creg <- which(colSums(M.reg[,var.latent,drop = FALSE]!=0)==0)
        indexClean.Ccov <- which(colSums(M.cov[,var.latent,drop = FALSE]!=0)==0)
        indexClean <- intersect(intersect(indexClean.Rreg, indexClean.Rcov),
                                intersect(indexClean.Creg, indexClean.Ccov))
        varClean <- c(varClean, var.latent[indexClean])
    }

    if(length(varClean)>0){
        x <- kill(x, value =  varClean, ...)
    }
    return(x)
}


######################################################################
### clean.R ends here

### sCorrect-vcov2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: dec 11 2019 (13:55) 
## Version: 
## Last-Updated: jan 23 2024 (10:27) 
##           By: Brice Ozenne
##     Update #: 102
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * vcov2 (documentation)
#' @title  Variance-Covariance With Small Sample Correction
#' @description  Extract the variance-covariance matrix from a latent variable model.
#' Similar to \code{stats::vcov} but with small sample correction.
#' @name vcov2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param robust [logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#' 
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the variance-covariance matrix.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @return A matrix with as many rows and columns as the number of coefficients.
#' 
#' @examples
#' #### simulate data ####
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' #### linear models ####
#' e.lm <- lm(formula.lvm,data=d)
#'
#' #### latent variable models ####
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' vcov0 <- vcov(e.lvm)
#' vcovSSC <- vcov2(e.lvm)
#' 
#' vcovSSC/vcov0
#' vcovSSC[1:4,1:4]/vcov(e.lm)
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`vcov2` <-
    function(object, robust, cluster, as.lava, ...) UseMethod("vcov2")

## * vcov2.lvmfit
#' @rdname vcov2
#' @export
vcov2.lvmfit <- function(object, robust = FALSE, cluster = NULL, as.lava = TRUE, ssc = lava.options()$ssc, ...){

    return(vcov(estimate2(object, ssc = ssc, ...), robust = robust, cluster = cluster, as.lava = as.lava))

}

## * vcov2.lvmfit2
#' @rdname vcov2
#' @export
vcov2.lvmfit2 <- function(object, robust = FALSE, cluster = NULL, as.lava = TRUE, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }
    

    if(robust){
        out <- crossprod(iid(object, cluster = cluster, as.lava = as.lava))
    }else{
        if(!is.null(cluster)){
            warning("Argument \'cluster\' disregarded when argument \'robust\' is FALSE. \n")
        }
        out <- object$sCorrect$vcov.param
    }

    out <- out[names(object$sCorrect$skeleton$originalLink2param),
               names(object$sCorrect$skeleton$originalLink2param),
               drop=FALSE]
    if(as.lava==FALSE){
        dimnames(out) <- list(as.character(object$sCorrect$skeleton$originalLink2param),
                              as.character(object$sCorrect$skeleton$originalLink2param))
    }

    return(out)
}

## * vcov.lvmfit2
#' @rdname vcov2
#' @export
vcov.lvmfit2 <- vcov2.lvmfit2



######################################################################
### sCorrect-vcov2.R ends here

### sCorrect-summary2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 10 2017 (10:57) 
## Version: 
## Last-Updated: jan 23 2024 (10:26) 
##           By: Brice Ozenne
##     Update #: 554
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - summary2
#' @title Latent Variable Model Summary After Small Sample Correction
#' @description Summarize a fitted latent variable model.
#' Similar to \code{stats::summary} with small sample correction.
#' @name summary2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param digit [integer > 0] the number of decimal places to use when displaying the summary.
#' @param robust [logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... [logical] arguments passed to lower level methods.
#' 
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @details \code{summary2} is the same as \code{summary}
#' except that it first computes the small sample correction (but does not store it).
#' So if \code{summary2} is to be called several times,
#' it is more efficient to pre-compute the quantities for the small sample correction
#' using \code{sCorrect} and then call \code{summary2}.
#'
#' \code{summary2} returns an object with an element \code{table2} containing the estimates, standard errors, degrees of freedom,
#' upper and lower limits of the confidence intervals, test statistics, and p-values.
#' 
#' @examples
#' #### simulate data ####
#' m <- lvm(Y~X1+X2)
#' set.seed(10)
#' d <- lava::sim(m, 2e1)
#'
#' #### latent variable models ####
#' e.lvm <- estimate(m, data = d)
#' summary(e.lvm)$coef
#' 
#' summary2(e.lvm)
#' summary2(e.lvm, ssc = "none")
#' 
#' @concept small sample inference
#' @export
`summary2` <-
  function(object, robust, cluster, digit, ...) UseMethod("summary2")

## * summary2.lvmfit
#' @rdname summary2
#' @export
summary2.lvmfit <- function(object, robust = FALSE, cluster = NULL, digit = max(5, getOption("digit")), ssc = lava.options()$ssc, df = lava.options()$df, ...){

    return(summary(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...), robust = robust, cluster = NULL, digit = digit))

}

## * summary2.lvmfit2
#' @rdname summary2
#' @export
summary2.lvmfit2 <- function(object, robust = FALSE, cluster = NULL, digit = max(5, getOption("digit")), ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    ## ** table with se, df, confint, p-value for the corrected parameters
    tableS.all <- model.tables(object, robust = robust, cluster = cluster, as.lava = TRUE)
    name.param <- rownames(tableS.all)
    n.param <- length(name.param)

    ## ** get and  normalize lava summary
    object0 <- object
    class(object0) <- setdiff(class(object0),c("lvmfit2"))
    object.summary <- summary(object0, digits = digit)
    
    previous.summary <- object.summary$coef
    object.summary$coef <- tableS.all[name.param,c("estimate","se","statistic","df","p.value"),drop=FALSE]


    ## find digit
    vec.char <- setdiff(object.summary$coefmat[,"Estimate"],"")
    digit <- max(c(nchar(gsub(".","",vec.char,fixed = TRUE)))-1,1)

    ## ** update summary
    ## *** vcov
    object.summary$vcov <- attr(object$dVcov, "vcov.param")[name.param,name.param]    

    ## *** coef
    lava.rownames <- rownames(previous.summary)
    
    ## *** coefmat
    name.label0 <- trimws(rownames(CoefMat(object0, labels = 0, level = 9)), which = "both")
    index.titleVariance <- which(name.label0=="Residual Variances:")
    if(length(index.titleVariance)>0){
        ## rename variance parameters from Y to Y~~Y
        index.vcov <- (index.titleVariance+1):length(name.label0)
        index.var <- setdiff(index.vcov,grep("~~",name.label0,fixed=TRUE)) ## exclude covariance parameters that are already correctly named
        name.label0[index.var] <- paste0(name.label0[index.var],lava.options()$symbols[2],name.label0[index.var])
    }
    table.coefmat <- object.summary$coefmat
    if(object$sCorrect$df=="satterthwaite"){
        colnames(table.coefmat)[3:5] <- c("t-value","P-value","df")
    }else{
        colnames(table.coefmat)[3:5] <- c("Z-value","P-value","df")
    }

    ## mimic lava:::CoefMat (called by lava:::summary.lvmfit)
    table.coef <- object.summary$coef
    e2add <- format(round(table.coef[,"estimate"], max(1, digit - 1)), digits = digit - 1)
    e2add <- gsub(" NA","",e2add)
    sd2add <- format(round(table.coef[,"se"], max(1, digit - 1)), digits = digit - 1)
    sd2add <- gsub(" NA","",sd2add)
    df2add <- as.character(round(table.coef[,"df"],2))    
    df2add[is.na(df2add)] <- ""
    t2add <- format(round(table.coef[,"statistic"], max(1, digit - 1)), digits = digit - 1)
    t2add <- gsub(" NA","",t2add)

    p2add <- formatC(table.coef[,"p.value"], digits = digit - 1, format = "g",  preserve.width = "common", flag = "")
    p2add <- gsub(" NA","",p2add)
    p2add[table.coef[,4] < 1e-12] <- "  <1e-12"

    M2add <- cbind(e2add,sd2add,t2add,p2add,df2add)
    table.coefmat[,"df"] <- ""
    table.coefmat[match(rownames(table.coef), name.label0),] <- M2add

    table.coefmat[object.summary$coefmat[,4]=="",4] <- ""
    object.summary$coefmat <- table.coefmat

    ## ** Export
    if(robust){
        colnames(object.summary$coefmat)[2] <- "robust SE"
        colnames(object.summary$coef)[2] <- "robust SE"
    }

    ## ** gather all results in one table
    object.summary$table2 <- data.frame(matrix(NA, nrow = n.param, ncol = 7,
                                               dimnames = list(name.param,
                                                               c("estimate","se","df","lower","upper","statistic","p.value"))
                                               ), stringsAsFactors = FALSE)

    object.summary$table2$estimate <- tableS.all[name.param,"estimate"]
    object.summary$table2$se <- tableS.all[name.param,"se"]
    object.summary$table2$df <- tableS.all[name.param,"df"]
    object.summary$table2$lower <- object.summary$table2$estimate + object.summary$table2$se * stats::qt(p=0.025, df = object.summary$table2$df)
    object.summary$table2$upper <- object.summary$table2$estimate + object.summary$table2$se * stats::qt(p=0.975, df = object.summary$table2$df)
    object.summary$table2$statistic <- tableS.all[name.param,"statistic"]
    object.summary$table2$p.value <- tableS.all[name.param,"p.value"]

    ## ** export
    return(object.summary)
    
}

## * summary.lvmfit2
#' @rdname summary2
#' @export
summary.lvmfit2 <- summary2.lvmfit2


##----------------------------------------------------------------------
### Scorrect-summary2.R ends here

### score2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (16:43) 
## Version: 
## last-updated: feb 11 2019 (13:33) 
##           By: Brice Ozenne
##     Update #: 2262
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - score2
#' @title  Extract the Individual Score
#' @description  Extract the individual score from a Gaussian linear model.
#' @name score2
#'
#' @param object a linear model or a latent variable model
#' @param param [optional] the fitted parameters.
#' @param data [optional] the data set.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias? Only relevant if the \code{sCorrect} function has not yet be applied to the object.
#' @param ... arguments to be passed to \code{sCorrect}.
#'
#' @details If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the influence function.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#'
#' @return A matrix containing the score relative to each sample (in rows)
#' and each model coefficient (in columns).
#' 
#' @examples
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' ## linear model
#' e.lm <- lm(formula.lvm,data=d)
#' score.tempo <- score2(e.lm, bias.correct = FALSE)
#' colMeans(score.tempo)
#'
#' ## latent variable model
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' score.tempo <- score2(e.lvm, bias.correct = FALSE)
#' range(score.tempo-score(e.lvm, indiv = TRUE))
#'
#' @concept small sample inference
#' @export
`score2` <-
  function(object, ...) UseMethod("score2")

## * score2.lm
#' @rdname score2
#' @export
score2.lm <- function(object, param = NULL, data = NULL, bias.correct = TRUE, ...){
    sCorrect(object, param = param, data = data, df = FALSE, ...) <- bias.correct

    ### ** export
    return(object$sCorrect$score)
}

## * score2.gls
#' @rdname score2
#' @export
score2.gls <- score2.lm

## * score2.lme
#' @rdname score2
#' @export
score2.lme <- score2.lm

## * score2.lvmfit
#' @rdname score2
#' @export
score2.lvmfit <- score2.lm

## * score2.lm2
#' @rdname score2
#' @export
score2.lm2 <- function(object, param = NULL, data = NULL, ...){

    ### ** compute the score
    if(!is.null(param) || !is.null(data)){
        args <- object$sCorrect$args
        args$df <- FALSE
        object$sCorrect <- do.call(sCorrect,
                                   args = c(list(object, param = param, data = data),
                                            args))
    }

    ### ** export
    return(object$sCorrect$score)

}

## * score2.gls2
#' @rdname score2
#' @export
score2.gls2 <- score2.lm2

## * score2.lme2
#' @rdname score2
#' @export
score2.lme2 <- score2.lm2

## * score2.lvmfit
#' @rdname score2
#' @export
score2.lvmfit2 <- score2.lm2

## * .score2
#' @title Compute the Corrected Score.
#' @description Compute the corrected score when there is no missing value.
#' @name score2-internal
#' 
#' @param n.cluster [integer >0] the number of observations.
#' 
#' @keywords internal
.score2 <- function(epsilon, Omega, OmegaM1, dmu, dOmega,                    
                    name.param, name.meanparam, name.varparam,
                    index.Omega, n.cluster, indiv){

### ** Prepare
    test.global <- is.null(index.Omega)
    out.score <- matrix(0, nrow = n.cluster, ncol = length(name.param),
                        dimnames = list(NULL,name.param))
            
### ** global
    if(test.global){
        epsilon.OmegaM1 <- epsilon %*% OmegaM1

        ## *** Compute score relative to the mean coefficients
        for(iP in name.meanparam){ # iP <- 1
            out.score[,iP] <- out.score[,iP] + rowSums(dmu[[iP]] * epsilon.OmegaM1)
        }
        
        ## *** Compute score relative to the variance-covariance coefficients
        for(iP in name.varparam){ # iP <- 1
            term2 <- - 1/2 * tr(OmegaM1 %*% dOmega[[iP]])            
            term3 <- 1/2 * rowSums(epsilon.OmegaM1 %*% dOmega[[iP]] * epsilon.OmegaM1)
            out.score[,iP] <- out.score[,iP] + as.double(term2) + term3
        }        
    }


### ** individual specific
    if(!test.global){

        for(iC in 1:n.cluster){
            iIndex <- index.Omega[[iC]]
            iEpsilon.OmegaM1 <- OmegaM1[[iC]] %*% cbind(epsilon[iC,iIndex])


            ## *** Compute score relative to the mean coefficients
            for(iP in name.meanparam){ # iP <- name.meanparam[1]
                out.score[iC,iP] <- out.score[iC,iP] + dmu[[iP]][iC,iIndex] %*% iEpsilon.OmegaM1
            }

            ## *** Compute score relative to the variance-covariance coefficients
            for(iP in name.varparam){ # iP <- name.varparam[1]
                term2 <- - 1/2 * tr(OmegaM1[[iC]] %*% dOmega[[iP]][iIndex,iIndex,drop=FALSE])
                term3 <- 1/2 * sum(iEpsilon.OmegaM1 * dOmega[[iP]][iIndex,iIndex,drop=FALSE] %*% iEpsilon.OmegaM1)
                out.score[iC,iP] <- out.score[iC,iP] + as.double(term2) + term3 
            }
        }
        
    }

    ### ** export
    if(indiv==FALSE){
        out.score <- colSums(out.score)
    }
    return(out.score)
}


#----------------------------------------------------------------------
### score2.R ends her

### test1-conditionalMoment.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar 27 2018 (09:50) 
## Version: 
## Last-Updated: jan 19 2022 (11:00) 
##           By: Brice Ozenne
##     Update #: 102
##----------------------------------------------------------------------
## 
### Commentary: 
## Compare the computation of the score/information matrix/residuals between sCorrect and lava
## Compare the computation of the hessian/derivative of the information matrix using analytical formulae vs. numerical derivatives
##
## NOTE: iid in lava uses numerical derivative to compute the information matrix
## this is why there is not a perfect matching between iid2.lvm and iid.lvm
### Change Log:
##----------------------------------------------------------------------
## 
### Code:


## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
    library(nlme)
}
lava.options(symbols = c("~","~~"))
context("conditionalMoment")

test.secondOrder <- TRUE

## * simulation
cat("- simulation \n")
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3,
              X4~1,X5~1))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","X4","X5","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))

## * multiple linear regression
cat("- multiple linear regression \n")
    
## ** no constrains
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)
e.lm <- lm(Y1~X1, data = d)
lvm2lm <- c("(Intercept)" = "Y1", "X1" = "Y1~X1", "sigma2" = "Y1~~Y1")

test_that("linear regression (ML) - no constrains",{
    expect_equal(as.double(logLik(e.lvm)), -277.9339, tol = 1e-6)
    
    test.lvm <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"satterthwaite"}else{"none"}, derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)
    
    expect_equal(lava::score(e.lvm, indiv = TRUE),
                 score2(test.lvm, indiv = TRUE),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(coef(e.lvm),
                 coef2(test.lvm),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
    
    if(test.secondOrder){
        
        X <- lapply(list(Y1~X1,Y2~X2,Y3~X1+X3), model.matrix, d)
        sigma2 <- list(coef(e.lvm)["Y1~~Y1"],
                       coef(e.lvm)["Y2~~Y2"],
                       coef(e.lvm)["Y3~~Y3"])

        dI <- mapply(X,sigma2, FUN = function(iX,iSigma){
            bdiag(crossprod(iX)/iSigma^2,n/(iSigma^3))
        })
        vcov <- mapply(X,sigma2, FUN = function(iX,iSigma){
            solve(bdiag(crossprod(iX)/iSigma,n/(2*iSigma^2)))
        })
        GS <- mapply(vcov, dI, FUN = function(x,y){
            as.matrix(x %*% y %*% x)
        })
        nameCoef.Y <- lapply(list("Y1","Y2","Y3"), function(iY){grep(iY, names(coef(e.lvm)), value = TRUE)})

        testN.lvm <- estimate2(e.lvm, ssc = "none", df = "satterthwaite", derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)
        expect_equal(test.lvm$sCorrect$hessian,testN.lvm$sCorrect$hessian, tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dVcov.param, testN.lvm$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(unname(test.lvm$sCorrect$dVcov.param[nameCoef.Y[[1]],nameCoef.Y[[1]],"Y1~~Y1"]), unname(GS[[1]]), tol = 1e-7)
        expect_equal(unname(test.lvm$sCorrect$dVcov.param[nameCoef.Y[[2]],nameCoef.Y[[2]],"Y2~~Y2"]), unname(GS[[2]]), tol = 1e-7)
        expect_equal(unname(test.lvm$sCorrect$dVcov.param[nameCoef.Y[[3]],nameCoef.Y[[3]],"Y3~~Y3"]), unname(GS[[3]]), tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dRvcov.param,testN.lvm$sCorrect$dRvcov.param, tol = 1e-7)
    }
})

test_that("linear regression (ML+1) - no constrains",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(lava::score(e.lvm, p = newcoef, indiv = TRUE),
                 score2(test.lvm, indiv = TRUE),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm, p = newcoef),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

## ** constrains and covariance link
m <- lvm(Y1[mu1:sigma]~X1,
         Y2[mu2:sigma]~X2,
         Y3~X1+X3,
         Y2~~Y3)
e.lvm <- estimate(m, data = d)

test_that("linear regression - constrains and covariance",{
    expect_equal(as.double(logLik(e.lvm)), -272.5088, tol = 1e-6)

    test <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"satterthwaite"}else{NA}, derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)
    
    expect_equal(unname(lava::score(e.lvm, indiv = TRUE)),
                 unname(score2(test, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test)),
                 tol = 1e-8)
    expect_equal(unname(coef(e.lvm)),
                 unname(coef2(test)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN <- estimate2(e.lvm, ssc = "none", df = "satterthwaite", derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)
        keep.param <- dimnames(test$sCorrect$dVcov.param)[[3]]
        zero.param <- setdiff(dimnames(testN$sCorrect$dVcov.param)[[3]], keep.param)
        expect_equal(test$sCorrect$hessian,testN$sCorrect$hessian, tol = 1e-7)
        expect_equal(test$sCorrect$dVcov.param,testN$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test$sCorrect$dRvcov.param,testN$sCorrect$dRvcov.param, tol = 1e-7)

        ## compare to previous versions
        GS <- matrix(c(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.00118648, 0.01999964, 0, -0.00554516, 3.94e-06, -3.01e-05, 0, 0, 0, 0, 0.01999964, -4.014e-05, 0, -1.68e-06, 0.00086362, -3.14e-05, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -0.00554516, -1.68e-06, 0, -0.02591598, 1.841e-05, -0.00014069, 0, 0, 0, 0, 3.94e-06, 0.00086362, 0, 1.841e-05, -0.01884834, -0.00143231, 0, 0, 0, 0, -3.01e-05, -3.14e-05, 0, -0.00014069, -0.00143231, -0.01661326, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.05139595, 0.05281454, 0, 0, 0, 0, 0, 0, 0, 0.05139595, -0.02433603, 0.0498916, 0, 0, 0, 0, 0, 0, 0, 0.05281454, 0.0498916, 0), 
                     nrow = 10, 
                     ncol = 10, 
                     dimnames = list(c("Y1", "Y2", "Y3", "Y1~X1", "Y2~X2", "Y3~X1", "Y3~X3", "Y1~~Y1", "Y3~~Y3", "Y2~~Y3"),c("Y1", "Y2", "Y3", "Y1~X1", "Y2~X2", "Y3~X1", "Y3~X3", "Y1~~Y1", "Y3~~Y3", "Y2~~Y3")) 
                     ) 
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y1","Y2~~Y3"],GS[,"Y1"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y2","Y2~~Y3"],GS[,"Y2"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y3","Y2~~Y3"],GS[,"Y3"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y1~X1","Y2~~Y3"],GS[,"Y1~X1"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y2~X2","Y2~~Y3"],GS[,"Y2~X2"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y3~X1","Y2~~Y3"],GS[,"Y3~X1"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y3~X3","Y2~~Y3"],GS[,"Y3~X3"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y1~~Y1","Y2~~Y3"],GS[,"Y1~~Y1"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y3~~Y3","Y2~~Y3"],GS[,"Y3~~Y3"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),"Y2~~Y3","Y2~~Y3"],GS[,"Y2~~Y3"],tol = 1e-6)
        expect_equivalent(test$sCorrect$dVcov.param[rownames(GS),colnames(GS),"Y2~~Y3"],GS,tol = 1e-6)
    }
})

test_that("linear regression (ML+1) - constrains and covariance",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")

    expect_equal(coef(test.lvm), newcoef, tol = 1e-8)

    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})


## * mixed model
cat("- mixed model \n")

## ** Compound symmetry
m <- lvm(Y1[mu1:sigma]~1*eta,
         Y2[mu2:sigma]~1*eta,
         Y3[mu3:sigma]~1*eta,
         eta~X1+Gender)
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corCompSymm(form=~ 1|Id),
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             method = "ML")

test_that("Compound symmetry", {
    expect_equal(as.double(logLik(e.lme)),as.double(logLik(e.lvm)), tol = 1e-3)
    expect_equal(as.double(logLik(e.gls)),as.double(logLik(e.lvm)), tol = 1e-3)
    expect_equal(as.double(logLik(e.lvm)), -259.8317, tol = 1e-6)

    test.lvm <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"Satterthwaite"}else{"none"},
                          derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)

    expect_equal(as.double(getVarCov(e.gls)),
                 as.double(getVarCov2(test.lvm)),
                 tol = 1e-3)
    expect_equal(unname(lava::score(e.lvm, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(coef(e.lvm)),
                 unname(coef2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN.lvm <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                               derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)

        expect_equal(test.lvm$sCorrect$hessian,testN.lvm$sCorrect$hessian, tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dVcov.param,testN.lvm$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dRvcov.param,testN.lvm$sCorrect$dRvcov.param, tol = 1e-7)

        ## compare to previous versions
        GS <- matrix(c(0.02860806, -0.02, -0.02, -0.00089185, -0.01529786, 0, 0, -0.02, 0.04, 0.02, 0, 0, 0, 0, -0.02, 0.02, 0.04, 0, 0, 0, 0, -0.00089185, 0, 0, 0.00645539, 0.00105925, 0, 0, -0.01529786, 0, 0, 0.00105925, 0.02723009, 0, 0, 0, 0, 0, 0, 0, 0.05700596, -0.01900199, 0, 0, 0, 0, 0, -0.01900199, 0.03500504), 
                     nrow = 7, 
                     ncol = 7, 
                     dimnames = list(c("eta", "Y2", "Y3", "eta~X1", "eta~GenderFemale", "Y1~~Y1", "eta~~eta"),c("eta", "Y2", "Y3", "eta~X1", "eta~GenderFemale", "Y1~~Y1", "eta~~eta")) 
                     )
        expect_equal(test.lvm$sCorrect$dVcov.param[rownames(GS),colnames(GS),"Y1~~Y1"],GS,tol = 1e-6)
    }
})

test_that("mixed model (ML+1) - CS",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

## ** Unstructured 
m <- lvm(Y1~1*eta,
         Y2~1*eta,
         Y3~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             weights = varIdent(form =~ 1|variable),
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             weights = varIdent(form =~ 1|variable),
             data = dL[dL$variable %in% c("Y1","Y2","Y3"),],
             method = "ML")

test_that("Unstructured", {
    expect_equal(as.double(logLik(e.lme)),as.double(logLik(e.lvm)), tol = 1e-3)
    expect_equal(as.double(logLik(e.gls)),as.double(logLik(e.lvm)), tol = 1e-3)
    expect_equal(as.double(logLik(e.lvm)), -258.8121, tol = 1e-6)

    test.lvm <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"Satterthwaite"}else{NA},
                          derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)

    expect_equal(as.double(getVarCov(e.gls)),
                 as.double(getVarCov2(test.lvm)),
                 tol = 1e-3)
    expect_equal(unname(lava::score(e.lvm, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(coef(e.lvm)),
                 unname(coef2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN.lvm <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                               derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)

        expect_equal(test.lvm$sCorrect$hessian,testN.lvm$sCorrect$hessian, tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dVcov.param,testN.lvm$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test.lvm$sCorrect$dRvcov.param,testN.lvm$sCorrect$dRvcov.param, tol = 1e-7)

        ## compare to previous versions
        GS <- matrix(c(0.02337325, -0.02, -0.02, -0.00034949, -0.00599479, 0, 0, 0, 0, 0, 0, -0.02, 0.02, 0.02, 0, 0, 0, 0, 0, 0, 0, 0, -0.02, 0.02, 0.02, 0, 0, 0, 0, 0, 0, 0, 0, -0.00034949, 0, 0, 0.00252968, 0.00041509, 0, 0, 0, 0, 0, 0, -0.00599479, 0, 0, 0.00041509, 0.0106707, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.14197442, 0, 0, 0, 0.02650713, 0.02772787, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.02650713, 0, 0, 0, 0.04155753, 0.00895138, 0, 0, 0, 0, 0, 0.02772787, 0, 0, 0, 0.00895138, 0.04452667), 
                     nrow = 11, 
                     ncol = 11, 
                     dimnames = list(c("eta", "Y2", "Y3", "eta~X1", "eta~GenderFemale", "Y1~~Y1", "eta~~eta", "Y2~~Y2", "Y3~~Y3", "Y1~~Y2", "Y1~~Y3"),c("eta", "Y2", "Y3", "eta~X1", "eta~GenderFemale", "Y1~~Y1", "eta~~eta", "Y2~~Y2", "Y3~~Y3", "Y1~~Y2", "Y1~~Y3")) 
                     ) 
        expect_equal(test.lvm$sCorrect$dVcov.param[rownames(GS),colnames(GS),"Y1~~Y1"],GS,tol = 1e-6)
    }

})

test_that("mixed model (ML+1) - UN",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

## * factor model
cat("- factor model \n")

m <- lvm(Y1~eta,
         Y2~eta+X2,
         Y3~eta,
         Z1~eta, Z1~~Y1,Z1~~Y2,
         eta~X1+X3)
e.lvm <- estimate(m, d)

test_that("factor model",{
    expect_equal(as.double(logLik(e.lvm)), -334.2583, tol = 1e-6)

    test <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"Satterthwaite"}else{NA},
                      derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)
    
    expect_equal(lava::score(e.lvm, indiv = TRUE),
                 score2(test, indiv = TRUE),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test)),
                 tol = 1e-8)
    expect_equal(coef(e.lvm),
                 coef2(test),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                           derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)

        expect_equal(test$sCorrect$hessian,testN$sCorrect$hessian, tol = 1e-7)
        expect_equal(test$sCorrect$dVcov.param,testN$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test$sCorrect$dRvcov.param,testN$sCorrect$dRvcov.param, tol = 1e-7)

        ## compare to previous versions
        GS <- matrix(c(0.04792556, 0.04673097, 0.04673097, -0.00085306), 
                     nrow = 2, 
                     ncol = 2, 
                     dimnames = list(c("Y2", "Y3"),c("Y2", "Y3")) 
                     ) 
        expect_equal(test$sCorrect$dVcov.param[paste0("Y",2:3),paste0("Y",2:3),"Y2~eta"],GS,tol = 1e-6)
    }
})

test_that("factor model (ML+1)",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

## * two factor model
cat("- two factor model \n")

## ** correlation
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3,
           eta1~eta2))

e.lvm <- estimate(m, d)

test_that("two factor model - correlation",{
    expect_equal(as.double(logLik(e.lvm)), -518.3538, tol = 1e-6)

    test <- estimate2(e.lvm, ssc = "none", df = if(test.secondOrder){"Satterthwaite"}else{NA},
                      derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)
    
    expect_equal(lava::score(e.lvm, indiv = TRUE),
                 score2(test, indiv = TRUE),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test)),
                 tol = 1e-8)
    expect_equal(coef(e.lvm),
                 coef2(test),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                          derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)

        expect_equal(test$sCorrect$hessian,testN$sCorrect$hessian, tol = 1e-7)
        expect_equal(test$sCorrect$dVcov.param,testN$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test$sCorrect$dRvcov.param,testN$sCorrect$dRvcov.param, tol = 1e-7)
    
        ## compare to previous versions
        GS <- matrix(c(-0.01772622, 0.01772622, -0.01540713, 0.07276457, 0.01772622, -0.01772622, 0.01540713, -0.07276457, -0.01540713, 0.01540713, -0.0133915, 0.06324503, 0.07276457, -0.07276457, 0.06324503, -0.29869238), 
                     nrow = 4, 
                     ncol = 4, 
                     dimnames = list(c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3"),c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3")) 
                     )
        keep.param <-  c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3")
        expect_equal(test$sCorrect$dVcov.param[keep.param,keep.param,"eta1~eta2"],GS,tol = 1e-6)
    }
})

test_that("two factor model (ML+1) - correlation",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

## ** covariance
m <- lvm(c(Y1~eta1,Y2~eta1,Y3~eta1+X1,eta1~X1,
           Z1~eta2,Z2~eta2,Z3~eta2+X3,eta2~X4,
           eta1~~eta2))

e.lvm <- estimate(m, d)

test_that("two factor model - covariance",{
    expect_equal(as.double(logLik(e.lvm)), -502.7317, tol = 1e-6)

    test <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                      derivative = "analytic", hessian = TRUE, dVcov.robust = TRUE)
    
    expect_equal(lava::score(e.lvm, indiv = TRUE),
                 score2(test, indiv = TRUE),
                 tol = 1e-8)
    expect_equal(lava::information(e.lvm),
                 unname(information2(test)),
                 tol = 1e-8)
    expect_equal(coef(e.lvm),
                 coef2(test),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm)),
                 unname(residuals2(test, type = "response")),
                 tol = 1e-8)

    if(test.secondOrder){
        testN <- estimate2(e.lvm, ssc = "none", df = "Satterthwaite",
                           derivative = "numeric", hessian = TRUE, dVcov.robust = TRUE)
    
        expect_equal(test$sCorrect$hessian,testN$sCorrect$hessian, tol = 1e-7)
        expect_equal(test$sCorrect$dVcov.param,testN$sCorrect$dVcov.param, tol = 1e-7)
        expect_equal(test$sCorrect$dRvcov.param,testN$sCorrect$dRvcov.param, tol = 1e-7)

        GS <- matrix(c(-0.01561465, 0.01447463, -0.01301386, 0.05822372, 0.01447463, -0.01341929, 0.01206254, -0.05397059, -0.01301386, 0.01206254, -0.01091509, 0.04865601, 0.05822372, -0.05397059, 0.04865601, -0.21734976), 
                     nrow = 4, 
                     ncol = 4, 
                     dimnames = list(c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3"),c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3")) 
                     ) 
        keep.param <-  c("Z1~~Z1", "eta2~~eta2", "Z2~~Z2", "Z3~~Z3")
        expect_equal(test$sCorrect$dVcov.param[keep.param,keep.param,"eta1~~eta2"],GS,tol = 1e-6)
    }
})

test_that("two factor model (ML+1) - covariance",{
    newcoef <- coef(e.lvm)+1
    test.lvm <- estimate2(e.lvm, param = newcoef, ssc = "none", df = "none")
    
    expect_equal(unname(lava::score(e.lvm, p = newcoef, indiv = TRUE)),
                 unname(score2(test.lvm, indiv = TRUE)),
                 tol = 1e-8)
    expect_equal(unname(lava::information(e.lvm, p = newcoef)),
                 unname(information2(test.lvm)),
                 tol = 1e-8)
    expect_equal(unname(residuals(e.lvm, p = newcoef)),
                 unname(residuals2(test.lvm, type = "response")),
                 tol = 1e-8)
})

##
##----------------------------------------------------------------------
### test1-conditionalMoment.R ends here

### tryWithWarnings.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 29 2017 (09:52) 
## Version: 
## last-updated: feb  5 2018 (16:07) 
##           By: Brice Ozenne
##     Update #: 20
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:


##' @title Run an Expression and Catch Warnings and Errors
##' @description Similar to \code{try} but also returns warnings.
##' 
##' @param expr the line of code to be evaluated
##' @details
##' from https://stackoverflow.com/questions/4948361/how-do-i-save-warnings-and-errors-as-output-from-a-function
##' 
##' @return A list containing:
##' \itemize{
##' \item value the result of the evaluation of the expression
##' \item warnings warning(s) generated during the evaluation of the expression
##' \item error error generated during the evaluation of the expression
##' }
##'
##' @examples
##' FctTest <- function(x){
##'   return(log(x))
##' }
##' tryWithWarnings(FctTest(-1))
##' tryWithWarnings(FctTest(1))
##' tryWithWarnings(FctTest(xxxx))
##'
##' @export
tryWithWarnings <- function(expr) {
    myWarnings <- NULL
    myError <- NULL
    FCT.envir <- environment()
    wHandler <- function(w) {
        assign("myWarnings", value = c(myWarnings, list(w)), envir = FCT.envir)
        invokeRestart("muffleWarning")
    }
    eHandler <- function(e) {
        assign("myError", value = e, envir = FCT.envir)
        return(NULL)
    }
    val <- withCallingHandlers(tryCatch(expr, error = eHandler), warning = wHandler)
    
    out <- list(value = val,
                warnings = myWarnings,
                error = myError)
    
  return(out)
}


#----------------------------------------------------------------------
### tryWithWarnings.R ends here

### score2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (16:43) 
## Version: 
## last-updated: jan 23 2024 (10:26) 
##           By: Brice Ozenne
##     Update #: 2408
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - score2
#' @title  Score With Small Sample Correction
#' @description  Extract the (individual) score a the latent variable model.
#' Similar to \code{lava::score} but with small sample correction.
#' @name score2
#'
#' @param object,x a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param indiv [logical] If \code{TRUE}, the score relative to each observation is returned. Otherwise the total score is returned.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @return When argument indiv is \code{TRUE}, a matrix containing the score relative to each sample (in rows)
#' and each model coefficient (in columns). Otherwise a numeric vector of length the number of model coefficients.
#' 
#' @examples
#' #### simulate data ####
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' #### linear models ####
#' e.lm <- lm(Y~X1+X2+X3, data = d)
#' 
#' #### latent variable models ####
#' m.lvm <- lvm(formula.lvm)
#' e.lvm <- estimate(m.lvm,data=d)
#' e2.lvm <- estimate2(m.lvm,data=d)
#' score.tempo <- score(e2.lvm, indiv = TRUE)
#' colSums(score.tempo)
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`score2` <-
  function(object, indiv, cluster, as.lava, ...) UseMethod("score2")

## * score2.lvmfit
#' @rdname score2
#' @export
score2.lvmfit <- function(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ssc = lava.options()$ssc, ...){

    return(lava::score(estimate2(object, ssc = ssc, ...), indiv = indiv, cluster = cluster, as.lava = as.lava))

}

## * score2.lvmfit2
#' @rdname score2
#' @export
score2.lvmfit2 <- function(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...){
    
    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }
    ## ** define cluster
    if(length(cluster) == 1 && (is.numeric(cluster) || is.character(cluster) || is.factor(cluster))){
        data <- object$sCorrect$data
        if(length(cluster)==1){                
            if(cluster %in% names(data) == FALSE){
                stop("Invalid \'cluster\' argument \n",
                     "Could not find variable \"",cluster,"\" in argument \'data\' \n")
            }
            cluster <- data[[cluster]]
        }
        cluster.index <- as.numeric(factor(cluster, levels = unique(cluster)))            
        n.cluster <- length(unique(cluster.index))
    }else if(is.vector(cluster)){
        cluster.index <- as.numeric(factor(cluster, levels = unique(cluster)))
        n.cluster <- length(unique(cluster.index))
    }else if(is.null(cluster)){ ## NOTE: cluster is a function in the survival package
        cluster <- NULL
        n.cluster <- object$sCorrect$cluster$n.cluster
        cluster.index <- 1:n.cluster
    }else{
        stop("Do not know how to handle argument cluster of class ",class(cluster),"\n")
    }

    ## ** get score
    score <- object$sCorrect$score
    if(!is.null(cluster)){ ## aggregate score by cluster
        score <- rowsum(score, group = cluster.index, reorder = FALSE)
    }
    
    ## ** export
    score <- score[,names(object$sCorrect$skeleton$originalLink2param),drop=FALSE]
    if(as.lava==FALSE){
        colnames(score) <- as.character(object$sCorrect$skeleton$originalLink2param)
    }
    if(!is.null(cluster)){
        index2.cluster <- tapply(1:length(cluster),cluster,list)
        attr(score,"cluster") <- names(index2.cluster)
    }
    
    if(indiv){
        return(score)
    }else{
        return(colSums(score))
    }
}

## * score.lvmfit2
#' @rdname score2
#' @export
score.lvmfit2 <- function(x, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...){## necessary as first argument of score must be x 
    score2(x, indiv = indiv, cluster = cluster, as.lava = as.lava, ...)
}

## * .score2
#' @title Compute the Corrected Score.
#' @description Compute the corrected score.
#' @name score2-internal
#' 
#' @param n.cluster [integer >0] the number of observations.
#' 
#' @keywords internal
.score2 <- function(dmu, dOmega, epsilon, OmegaM1,
                    missing.pattern, unique.pattern, name.pattern,
                    name.param, name.meanparam, name.varparam,
                    n.cluster, weights){
    if(lava.options()$debug){cat(".score2\n")}

    ## ** Prepare
    out.score <- matrix(NA, nrow = n.cluster, ncol = length(name.param),
                        dimnames = list(NULL,name.param))
    n.pattern <- length(name.pattern)
    
    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iOmegaM1 <- OmegaM1[[iPattern]]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        iEpsilon.OmegaM1 <- epsilon[iIndex,iY,drop=FALSE] %*% iOmegaM1
        out.score[iIndex,] <- 0 ## initialize (keep NA for missing values)

        ## *** Compute score relative to the mean coefficients
        for(iP in name.meanparam){ # iP <- "Y3~eta"
            out.score[iIndex,iP] <- out.score[iIndex,iP] + rowSums(dmu[[iP]][iIndex,iY,drop=FALSE] * iEpsilon.OmegaM1)
        }
        
        ## *** Compute score relative to the variance-covariance coefficients
        for(iP in name.varparam){ # iP <- "eta~~eta"
            term2 <- - 1/2 * tr(iOmegaM1 %*% dOmega[[iP]][iY,iY,drop=FALSE])            
            term3 <- 1/2 * rowSums(iEpsilon.OmegaM1 %*% dOmega[[iP]][iY,iY,drop=FALSE] * iEpsilon.OmegaM1)
            out.score[iIndex,iP] <- out.score[iIndex,iP] + as.double(term2) + term3
        }        
    }

    ## ** export
    if(!is.null(weights)){
        out.score <- sweep(out.score, STATS = weights, MARGIN = 1, FUN = "*")
    }
    return(out.score)
}


#----------------------------------------------------------------------
### score2.R ends her

### summary.modelsearch2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 30 2017 (10:46) 
## Version: 
## last-updated: Jan 11 2022 (16:44) 
##           By: Brice Ozenne
##     Update #: 116
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * method summary.modelsearch2
#' @title summary Method for modelsearch2 Objects
#' @description summary method for modelsearch2 objects.
#'
#' @param object output of the \code{modelsearch2} function.
#' @param print should the summary be printed in the terminal.
#' @param ... [internal] only used by the generic method.
#' 
#' @method summary modelsearch2
#'
#' @details
#' The column \code{dp.Info} contains the percentage of extended models (i.e. model with one additional link)
#' for which the information matrix evaluated at the value of the parameters of the initial model is non positive definie.
#' 
#' @export
summary.modelsearch2 <- function(object, print = TRUE, ...){

    p.value <- NULL # [:for CRAN check] subset
    
    ## ** extract data from object
    n.step <- nStep(object)
    tableTest <- do.call(rbind,lapply(object$sequenceTest, function(iTest){
        out <- iTest[which.max(iTest$statistic), c("link","statistic","p.value","adjusted.p.value","dp.Info","selected","nTests")]
        out[,"dp.Info"] <- mean(iTest[,"dp.Info"], na.rm = TRUE)
        return(out)
    }))
    n.selected <- sum(tableTest$selected)

    keep.cols <- c("link","nTests","statistic","adjusted.p.value")
    ## ** output
    out <- list()
    out$message.pre <- paste0("Sequential search for local dependence using the score statistic \n")
    if(n.selected==0){
        out$message.pre <- c(out$message.pre,
                             "The variable selection procedure did not retain any variable \n")
    }else{
        out$message.pre <- c(out$message.pre,
                             paste0("The variable selection procedure retained ",n.selected," variable",
                                    if(n.selected>1){"s"},":\n")
                             )     
    }

    out$table <- tableTest
    rownames(out$table) <- NULL
    out$message.post <- paste0("Confidence level: ",1-object$alpha," (two sided, adjustement: ",object$method.p.adjust,")\n")  

    ## ** display
    if(print){
        cat(out$message.pre,sep="")
        print(out$table)
        cat(out$message.post,sep="")
        if(any(stats::na.omit(out$table[,"dp.Info"])<1)){
            cat("WARNING: some of the score tests could not be correctly computed, probably because extended models are not all identifiable\n",
                "        consider using the argument \'link\' to specify only identifiable models \n")
        }
    }
    
    ## ** export
    return(invisible(out))
}



#----------------------------------------------------------------------
### summary.modelsearch2.R ends here

### calcDistMax.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: jun 21 2017 (16:44) 
## Version: 
## last-updated: jun 20 2019 (10:28) 
##           By: Brice Ozenne
##     Update #: 630
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * documentation
#' @title Adjust the p.values Using the Quantiles of the Max Statistic
#' @description Adjust the p.values using the quantiles of the max statistic.
#' @name calcDistMax
#'
#' @param statistic [numeric vector] the observed Wald statistic.
#' Each statistic correspond to a null hypothesis (i.e. a coefficient) that one wish to test.
#' @param iid [matrix] zero-mean iid decomposition of the coefficient used to compute the statistic.
#' @param df [numeric] the degree of freedom defining the multivariate Student's t distribution.
#' If \code{NULL} the multivariate Gaussian distribution will be used instead.
#' @param iid.previous [matrix, EXPERIMENTAL] zero-mean iid decomposition of previously tested coefficient.
#' @param quantile.compute [logical] should the rejection quantile be computed?
#' @param quantile.previous [numeric, EXPERIMENTAL] rejection quantiles of the previously tested hypotheses. If not \code{NULL} the values should correspond the variable in to the first column(s) of the argument \code{iid.previous}.
#' @param method [character] the method used to compute the p-values.
#' @param alpha [numeric 0-1] the significance cutoff for the p-values.
#' When the p-value is below, the corresponding link will be retained.
#' @param cpus [integer >0] the number of processors to use.
#' If greater than 1, the computation of the p-value relative to each test is performed in parallel. 
#' @param cl [cluster] a parallel socket cluster generated by \code{parallel::makeCluster}
#' that has been registered using \code{registerDoParallel}.
#' @param n.sim [integer >0] the number of bootstrap simulations used to compute each p-values.
#' Disregarded when the p-values are computed using numerical integration.
#' @param n.repmax [integer >0] the maximum number of rejection for each bootstrap sample before switching to a new bootstrap sample.
#' Only relevant when conditioning on a previous test.
#' Disregarded when the p-values are computed using numerical integration.
#' @param trace [logical] should the execution of the function be traced?
#'
#' @return A list containing
#' \itemize{
#' \item p.adjust: the adjusted p-values.
#' \item z: the rejection threshold.
#' \item Sigma: the correlation matrix between the test statistic.
#' \item correctedLevel: the alpha level corrected for conditioning on previous tests.
#' }
#' 
#' @examples 
#' library(mvtnorm)
#'
#' set.seed(10)
#' n <- 100
#' p <- 4
#' link <- letters[1:p]
#' n.sim <- 1e3 # number of bootstrap simulations 
#'
#' #### test - not conditional ####
#' X.iid <- rmvnorm(n, mean = rep(0,p), sigma = diag(1,p))
#' colnames(X.iid) <- link
#' statistic <- setNames(1:p,link)
#'
#' 
#' r1 <- calcDistMaxIntegral(statistic = statistic, iid = X.iid, 
#'             trace = FALSE, alpha = 0.05, df = 1e6) 
#' 
#' r3 <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
#'             method = "residual",
#'             trace = FALSE, alpha = 0.05, n.sim = n.sim)
#'
#' r4 <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
#'             method = "wild",
#'             trace = FALSE, alpha = 0.05, n.sim = n.sim)
#' 
#' rbind(integration = c(r1$p.adjust, quantile = r1$z),
#'       bootResidual = c(r3$p.adjust, quantile = r3$z),
#'       bootWild    = c(r4$p.adjust, quantile = r4$z))
#'
#' #### test - conditional ####
#' \dontrun{
#' Z.iid <- rmvnorm(n, mean = rep(0,p+1), sigma = diag(1,p+1))
#' seqQuantile <- qmvnorm(p = 0.95, delta = rep(0,p+1), sigma = diag(1,p+1), 
#'                     tail = "both.tails")$quantile
#' 
#' r1c <- calcDistMaxIntegral(statistic = statistic, iid = X.iid,
#'             iid.previous = Z.iid, quantile.previous =  seqQuantile, 
#'             trace = FALSE, alpha = 0.05, df = NULL)
#' 
#' r3c <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
#'             iid.previous = Z.iid, quantile.previous =  seqQuantile, method = "residual",
#'             trace = FALSE, alpha = 0.05, n.sim = n.sim)
#'
#' r4c <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
#'             iid.previous = Z.iid, quantile.previous =  seqQuantile, method = "wild",
#'             trace = FALSE, alpha = 0.05, n.sim = n.sim)
#'
#' rbind(integration = c(r1c$p.adjust, quantile = r1c$z),
#'       bootResidual = c(r3c$p.adjust, quantile = r3c$z),
#'       bootWild    = c(r4c$p.adjust, quantile = r4c$z))
#' }
#' @concept modelsearch
#' @concept post-selection inference


## * calcDistMaxIntegral
#' @rdname calcDistMax
#' @export
calcDistMaxIntegral <- function(statistic, iid, df, 
                                iid.previous = NULL, quantile.previous = NULL,
                                quantile.compute = lava.options()$search.calc.quantile.int,
                                alpha, cpus = 1, cl = NULL, trace){

    ## ** normalize arguments
    p.iid <- NCOL(iid)
    n <- NROW(iid)
    conditional <- length(quantile.previous)
    if(length(quantile.previous)>1){
        stop("Can only condition on one previous step \n")
    }
    if(is.null(df)){
        distribution.statistic <- "gaussian"
    }else{
        distribution.statistic <- "student"
    }
    
    iid.all <- cbind(iid,iid.previous)
    index.new <- 1:NCOL(iid)
    index.previous <- setdiff(1:NCOL(iid.all),index.new)
    p.iid.all <- NCOL(iid.all)
   
    ## ** Compute the correlation matrix between the test statistics
    # center to be under the null
    # scale since we want the distribution of the Wald statistic (i.e. statistic with unit variance)
    iid.statistic <- scale(iid.all, center = TRUE, scale = TRUE)
    Sigma.statistic <- stats::cov(iid.statistic, use = "pairwise.complete.obs")
    out <- list(p.adjust = NULL, z = NULL, Sigma = Sigma.statistic[index.new,index.new,drop=FALSE])

    ## ** Definition of the functions used to compute the quantiles
    warperQ <- function(alpha){
        .calcQmaxIntegration(alpha = alpha, p = p.iid,
                             Sigma = Sigma.statistic[index.new,index.new,drop=FALSE],
                             df = df, distribution = distribution.statistic)
    }
    warperP <- function(index){
        .calcPmaxIntegration(statistic = statistic[index], p = p.iid,
                             Sigma = Sigma.statistic[index.new,index.new,drop=FALSE],
                             df = df, distribution = distribution.statistic)
    }
        
    ## ** correction for conditioning on the previous steps
    if(conditional==TRUE){
        out$correctedLevel <- calcType1postSelection(1-alpha, quantile.previous =  quantile.previous,
                                                      mu = rep(0,p.iid.all), Sigma = Sigma.statistic,
                                                      distribution =  distribution.statistic,
                                                      df = df)
        alpha <- 1-out$correctedLevel
    }else{
        out$correctedLevel <- NA
    }

    if(quantile.compute){       
        out$z <- warperQ(alpha)
    }else{
        out$z <- NA
    }

    ## ** start parallel computation
    init.cpus <- (cpus > 1 && is.null(cl))
    if(init.cpus){            
        ## define cluster
        if(trace>0){
            cl <- parallel::makeCluster(cpus, outfile = "")
        }else{
            cl <- parallel::makeCluster(cpus)
        }
        ## link to foreach
        doParallel::registerDoParallel(cl)
    }
    
    ## ** Computation
    if(trace > 0){ cat("Computation of multivariate student probabilities to adjust the p.values \n") }
    if(cpus > 1){
        ## *** parallel computations

        if(trace>0){
            pb <- utils::txtProgressBar(max = length(index.new), style = 3)                   
        }

        ## export package
        parallel::clusterCall(cl, fun = function(x){
            suppressPackageStartupMessages(requireNamespace("mvtnorm", quietly = TRUE))
        })
        
        value <- NULL # [:for CRAN check] foreach
        out$p.adjust <- foreach::`%dopar%`(
                                     foreach::foreach(value = 1:length(statistic),
                                                      .export = c(".calcPmaxIntegration"),
                                                      .combine = "list"),
                                     {
                                         if(trace>0){utils::setTxtProgressBar(pb, value)}
                                         return(warperP(index.new[value]))
                                     })

        if(trace>0){close(pb)}
            
    }else{
        ## *** sequential computations
        if(trace>0){      
            out$p.adjust <- pbapply::pblapply(1:length(statistic), function(iStat){ warperP(index.new[iStat]) })            
        }else{
            out$p.adjust <- lapply(1:length(statistic), function(iStat){ warperP(index.new[iStat]) })
        }
                        
    }
    out$error <- stats::setNames(unlist(lapply(out$p.adjust, function(x){attr(x,"error")})), names(statistic))
    out$p.adjust <- stats::setNames(unlist(out$p.adjust), names(statistic))

    ## ** end parallel computation
    if(init.cpus){
        parallel::stopCluster(cl)
    }

    
    ## ** export
    return(out)
}

## * calcDistMaxBootstrap
#' @rdname calcDistMax
#' @export
calcDistMaxBootstrap <- function(statistic, iid, iid.previous = NULL, quantile.previous = NULL,
                                 method, alpha, cpus = 1, cl = NULL, n.sim, trace, n.repmax = 100){

    ## ** normalize arguments
    n <- NROW(iid)
    conditional <- length(quantile.previous)>0
    if(length(quantile.previous)>1){
        stop("Can only condition on one previous step \n")
    }

    iid.all <- cbind(iid,iid.previous)
    index.new <- 1:NCOL(iid)
    index.previous <- setdiff(1:NCOL(iid.all),index.new)

    ## ** Function used for the simulations
    warperBoot <- .bootMaxDist
    
    ## ** Compute the correlation matrix between the test statistics
    # center to be under the null
    # scale since we want the distribution of the Wald statistic (i.e. statistic with unit variance)
    iid.statistic <- scale(iid.all, center = TRUE, scale = TRUE)
    Sigma.statistic <- stats::cov(iid.statistic, use = "pairwise.complete.obs")

    ## ** start parallel computation
    init.cpus <- (cpus > 1 && is.null(cl))
    if(init.cpus){            
        ## define cluster
        if(trace>0){
            cl <- parallel::makeCluster(cpus, outfile = "")
        }else{
            cl <- parallel::makeCluster(cpus)
        }
        ## link to foreach
        doParallel::registerDoParallel(cl)
    }

    ## ** Computation
    if(trace > 0){ cat("Bootsrap simulations to get the 95% quantile of the max statistic: ") }

    if(cpus>1){
        n.simCpus <- rep(round(n.sim/cpus),cpus)
        n.simCpus[1] <- n.sim-sum(n.simCpus[-1])
        
        i <- NULL # [:for CRAN check] foreach
        distMax <- foreach::`%dopar%`(
                                foreach::foreach(i = 1:cpus, .packages =  c("MASS"),
                                                 .export = "calcDistMax",
                                                 .combine = "c"),{
                                                     replicate(n.simCpus[i],
                                                               warperBoot(iid = iid.all, sigma = Sigma.statistic,
                                                                          n = n, method = method,
                                                                          index.new = index.new, index.previous = index.previous,
                                                                          quantile.previous = quantile.previous, n.repmax = n.repmax))
                                                 })
        
    }else{

        if(trace>0){
            distMax <- pbapply::pbsapply(1:n.sim, warperBoot, method = method,
                                         iid = iid.all, sigma = Sigma.statistic, n = n,                                         
                                         index.new = index.new, index.previous = index.previous,
                                         quantile.previous = quantile.previous, n.repmax = n.repmax)
            
        }else{
            distMax <- sapply(1:n.sim, warperBoot, method = method,
                              iid = iid.all, sigma = Sigma.statistic, n = n,
                              index.new = index.new, index.previous = index.previous,
                              quantile.previous = quantile.previous, n.repmax = n.repmax)
        }
        
    }
     
    if(trace > 0){ cat("done \n") }

    ## ** end parallel calculation
    if(init.cpus){
        parallel::stopCluster(cl)
    }

    ## ** export
    out <- list()
    out$z <- stats::quantile(distMax, probs = 1-alpha, na.rm = TRUE)
    out$p.adjust <- sapply(abs(statistic), function(x){mean(distMax>x,na.rm=TRUE)})
    out$Sigma <- Sigma.statistic
    out$correctedLevel <- NA
    return(out)
}

## * .calcQmaxIntegration: numerical integration to compute the critical threshold
.calcQmaxIntegration <- function(alpha, p, Sigma, df, distribution){

    if(distribution == "gaussian"){
        if(p==1){
            q.alpha <- stats::qnorm(1-alpha, mean = 0, sd = 1)
        }else{
            q.alpha <- mvtnorm::qmvnorm(1-alpha,
                                        mean = rep(0,p),
                                        corr = Sigma,
                                        tail = "both.tails")$quantile
        }
    }else if(distribution == "student"){
        if(p==1){
            q.alpha <- stats::qt(1-alpha, df = df)
        }else{
            q.alpha <- mvtnorm::qmvt(1-alpha,
                                     delta = rep(0,p),
                                     corr = Sigma,
                                     df = df,
                                     tail = "both.tails")$quantile
        }
    }

    return(q.alpha)
}
    
## * .calcPmaxIntegration_firstStep: numerical integration to compute the p.values
.calcPmaxIntegration <- function(statistic, p, Sigma, df, distribution){
    value <- abs(statistic)
    if(!is.na(value)){
        if(distribution == "gaussian"){
            if(p==1){
                out <- stats::pnorm(value, mean = 0, sd = Sigma)-stats::pnorm(-value, mean = 0, sd = Sigma)
                attr(out, "error") <- 0
            }else{      
                out <- mvtnorm::pmvnorm(lower = -value, upper = value,
                                        mean = rep(0, p), corr = Sigma)
            }
        }else if(distribution == "student"){
            if(p==1){
                out <- stats::pt(value, df = df)-stats::pt(-value, df = df)
                attr(out, "error") <- 0
            }else{
                out <- mvtnorm::pmvt(lower = -value, upper = value,
                                     delta = rep(0, p), corr = Sigma, df = df)
            }
        }
        return(1-out)
    }else{
        return(NA)
    }   
}

## * .bootMaxDist: bootstrap simulation
.bootMaxDist <- function(iid, sigma, n, method,
                         index.new, index.previous, quantile.previous, n.repmax,
                         ...){

    iRep <- 0
    cv <- FALSE

    while(iRep < n.repmax && cv == FALSE){

        ## ** resample to obtain a new influence function
        if(method == "residual"){
            iid.sim <- MASS::mvrnorm(n,rep(0,NCOL(sigma)),sigma)                    
        }else if(method == "wild"){
            e <- stats::rnorm(n,mean=0,sd=1)
            iid.sim <- sapply(1:NCOL(sigma),function(x){e*iid[,x]})        
        }
        if(!is.null(quantile.previous)){
            iid.previous <- iid.sim[,index.previous]
            test.previous <- apply(iid.previous,2,function(x){sqrt(n)*mean(x)/stats::sd(x)})
            max.previous <- max(abs(test.previous))        
            if(max.previous<quantile.previous){
                iRep <- iRep + 1
            }else{
                iid.sim <- iid.sim[,index.new]
                cv <- TRUE
            }
        }else{
            cv <- TRUE
        }
    }
    
    ## ** compute the bootstrap test statistic
    if(cv){
        Test <- apply(iid.sim,2,function(x){sqrt(n)*mean(x)/stats::sd(x)})
    }else{
        Test <- NA
    }
    return(max(abs(Test)))
}

#----------------------------------------------------------------------
### calcDistMax.R ends here

### calcType1postSelection.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 31 2017 (16:42) 
## Version: 
## last-updated: feb  4 2019 (09:40) 
##           By: Brice Ozenne
##     Update #: 130
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - calcType1postSelection 
##' @title Compute the Type 1 Error After Selection [EXPERIMENTAL]
##' @description Compute the type 1 error after selection [EXPERIMENTAL].
##' @name calcType1postSelection
##' 
##' @param level [numeric 0-1] expected coverage.
##' @param mu [numeric vector] the expectation of the joint distribution of the test statistics
##' @param Sigma [matrix] the variance-covariance of the joint distribution of the test statistics.
##' @param quantile.previous [numeric] significance quantile used at the previous step.
##' @param df [integer > 0] the degree of freedom of the joint Student's t distribution.
##' Only used when \code{distribution="pvmt"}.
##' @param n [integer > 0] number of points for the numerical integration
##' @param distribution [character] distribution of the test statistics.
##' Can be \code{"pmvnorm"} (normal distribution) or \code{"pvmt"} (Student's t distribution)
##' @param correct [logical] if true, correct the level to account for previous testings.
##' @param ... arguments passed to \code{\link{intDensTri}}.  
##'
##' @details The number of tests at the current step (i.e. after selection) is assumed to be
##' one less than the number of tests at the previous step (i.e. before selection).
##'
##' Arguments \code{mu} and \code{Sigma} must contain the moments for the vector of test statistics
##' before and after selection (in that order).
##' 
##' @return [numeric] the type 1 error.
##' @author Brice Ozenne
##' @examples
##' library(mvtnorm)
##' n <- 350
##' 
##' #### only 2 tests
##' Sigma <- rbind(c(1,0,0),c(0,1,1),c(0,1,1))
##' z2 <- qmvnorm(0.95, mean = rep(0,2), sigma = Sigma[1:2,1:2], tail = "both.tails")$quantile
##' 
##' ## no selection since strong effect
##' mu <- c(10,0,0)
##' calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
##'                         mu = mu, Sigma = Sigma, correct = TRUE)
##'
##' ## strong selection
##' \dontrun{
##' mu <- c(0,0,0)
##' levelC <- calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
##'                         mu = mu, Sigma = Sigma)
##' print(levelC) # more liberal than without selection
##' calcType1postSelection(levelC, quantile.previous = z2, distribution = "gaussian",
##'                         mu = mu, Sigma = Sigma, correct = FALSE)
##' }
##'
##' #### 3 tests
##' Sigma <- diag(1,5,5)
##' Sigma[4,2] <- 1
##' Sigma[2,4] <- 1
##' Sigma[5,3] <- 1
##' Sigma[3,5] <- 1
##' 
##' z2 <- qmvnorm(0.95, mean = mu[1:3], sigma = Sigma[1:3,1:3], tails = "both.tails")$quantile
##' 
##' ## no selection since strong effect
##' \dontrun{
##' mu <- c(10,0,0,0,0)
##' calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
##'                         mu = mu, Sigma = Sigma, correct = TRUE)
##' 
##' ## strong selection
##' mu <- c(0,0,0,0,0)
##' levelC <- calcType1postSelection(0.95, quantile.previous = z2,
##'                         mu = mu, Sigma = Sigma, distribution = "gaussian")
##' calcType1postSelection(levelC, quantile.previous = z2, distribution = "gaussian",
##'                         mu = mu, Sigma = Sigma, correct = FALSE)
##' }
##'
##' @concept post-selection inference


## * calcType1postSelection
##' @rdname calcType1postSelection
##' @export
calcType1postSelection <- function(level, mu, Sigma, quantile.previous, distribution, df,
                                    n = 10, correct = TRUE, ...){

    ## ** normalisation of the arguments
    p <- length(mu)
    if(p %% 2 == 0){
        stop("\'mu\' must have uneven length\n",
             "since it contains the means for the test statistics at the previous step (p+1)/2",
             "and those at the current step (p-1)/2")
    }
    if(NCOL(Sigma)!=p || NROW(Sigma)!=p){
        stop("\'Sigma\' must be a pxp matrix where p is the length of \'mu\' \n")
    }
    nTest <- (p-1)/2
    distribution <- match.arg(distribution, choices = c("student","gaussian"))
    
    if(distribution=="student"){
        pdist <- "pmvt"
        qdist <- "qmvt"
        args.dist <- list(delta = rep(0,nTest),
                          sigma = Sigma[1:nTest,1:nTest,drop=FALSE],
                          tail = "both.tails", df = df)
    }else if(distribution=="gaussian"){
        pdist <- "pmvnorm"
        qdist <- "qmvnorm"
        args.dist <- list(delta = rep(0,nTest),
                          sigma = Sigma[1:nTest,1:nTest,drop=FALSE],
                          tail = "both.tails")
    }    

    ## ** wrapper
    warper <- function(x){ # x <- 0.95## level
        # P[A|B] = 1-P[\bar{A}|B]  = 1-P[\bar{A},B]/P[B]
        ## Current quantile
        args.dist$p <- x
        newQuantile <- do.call(qdist, args = args.dist)$quantile
        
        ## Compute type one error
        num <- intDensTri(mu = mu, Sigma = Sigma, df = df, n=n,
                          x.min = quantile.previous, z.max = rep(newQuantile, nTest),
                          distribution = pdist, ...) # P[\bar{A},B]
        # autoplot(num, coord.plot = c("x","z"))
        denum <- intDensTri(mu = mu[1:(nTest+1)], Sigma = Sigma[1:(nTest+1),1:(nTest+1)], df = df, n=n,
                            x.min = quantile.previous, z.max = NULL,
                            distribution = pdist, ...) # P[B]
        out <- 1-num$value/denum$value
        return(out)
    }

    ## ** compute type1 error
    if(correct){
        out <- stats::optim(par = level, fn = function(x){
            obj <- abs((1-level)-warper(x))
            # cat("level:",x," obj:",obj,"\n")
            return(obj)
        },
        method = "Brent", lower = 0.51, upper = 0.999,
        control = list(abstol = 1e-4, reltol = 1e-4)
        )$par

    }else{
        out <- warper(level)
    }
    
    return(out)
}

#----------------------------------------------------------------------
### calcType1postSelection.R ends here

### test-calcType1postSelection.R --- 
##----------------------------------------------------------------------
## author: Brice Ozenne
## created: sep 26 2017 (16:54) 
## Version: 
## last-updated: feb 11 2019 (13:45) 
##           By: Brice Ozenne
##     Update #: 23
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
##
##
### Code:
## * header
rm(list = ls())
if(TRUE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

library(testthat)
library(mvtnorm)
lava.options(symbols = c("~","~~"))

context("calcType1postSelection")

## * tests
n <- 350

test_that("check type 1 error - unrealistic case with independent statistics", {
    
    p <- 5
    p.all <- 2*p+1
    zp <- qmvnorm(0.95, mean = rep(0,p), sigma = diag(1,p,p), tail = "both.tails")$quantile
    type1Test <- calcType1postSelection(0.95,
                                        quantile.previous = zp,
                                        distribution = "gaussian",
                                        mu = rep(0,p.all),
                                        Sigma = diag(1,p.all,p.all),
                                        correct = FALSE)

    expect_equal(type1Test,0.05, tol = 0.01)
    
})

test_that("check type 1 error - 3 independent tests", {


    #### 3 tests
    Sigma <- diag(1,5,5)
    Sigma[4,2] <- 1
    Sigma[2,4] <- 1
    Sigma[5,3] <- 1
    Sigma[3,5] <- 1
 
    ## no selection since strong effect
    mu <- c(10,0,0,0,0)
    z2 <- qmvnorm(0.975, mean = mu[1:3], sigma = Sigma[1:3,1:3])$quantile

    type1Test <- calcType1postSelection(0.95, quantile.previous = z2,
                                        distribution = "gaussian",
                                        mu = mu, Sigma = Sigma, correct = FALSE)

    expect_equal(type1Test,0.05, tol = 0.01)

    ## conservative type 1 error due to selection
    mu <- c(0,0,0,0,0)
    z2 <- qmvnorm(0.975, mean = mu[1:3], sigma = Sigma[1:3,1:3])$quantile

    type1Test <- calcType1postSelection(0.95, quantile.previous = z2,
                                        distribution = "gaussian",
                                        mu = mu, Sigma = Sigma, correct = FALSE)

    expect_true(type1Test<0.05)
})

#----------------------------------------------------------------------
### test-calcType1postSelection.R ends here

### Objective_gaussian_weight.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 17 2020 (16:29) 
## Version: 
## Last-Updated: Jan 12 2022 (12:31) 
##           By: Brice Ozenne
##     Update #: 136
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

##' @title Estimate LVM With Weights
##' @description Estimate LVM with weights.
##' @name gaussian_weight
##'
##' @param x,object A latent variable model
##' @param data dataset
##' @param estimator name of the estimator to be used
##' @param type must be "cond"
##' @param p parameter value
##' @param weights weight associated to each iid replicate.
##' @param S empirical variance-covariance matrix between variable
##' @param n number of iid replicates
##' @param mu empirical mean
##' @param debug,reindex,mean,constrain,indiv additional arguments not used
##' @param ... passed to lower level functions.
##' 
##' @examples
##' #### linear regression with weights ####
##'
##' ## data
##' df <- data.frame(Y = c(1,2,2,1,2),
##'                  X = c(1,1,2,2,2),
##'                  missing = c(0,0,0,0,1),
##'                  weights = c(1,1,2,1,NA))
##'
##' ## using lm
##' e.lm.GS <- lm(Y~X, data = df)
##' e.lm.test <- lm(Y~X, data = df[df$missing==0,], weights = df[df$missing==0,"weights"])
##' 
##' ## using lvm
##' m <- lvm(Y~X)
##' e.GS <- estimate(m, df)
##' ## e.lava.test <- estimate(m, df[df$missing==0,], weights = df[df$missing==0,"weights"])
##' ## warnings!!
##' e.test <- estimate(m, data = df[df$missing==0,],
##'                    weights = df[df$missing==0,"weights"],
##'                    estimator = "gaussian_weight")
##' 

## * gaussian_weight.estimate.hook
##' @rdname gaussian_weight
##' @export
gaussian_weight.estimate.hook <- function(x, data, estimator, ...){
    dots <- list(...)
    if(identical(estimator,"gaussian_weight")){
        xe <- suppressWarnings(estimate(x, data = data, control = list(iter.max = 0))) ## initialize coefficients
        x$sCorrect <- moments2(xe, data = data, param = NULL, ## initCoef,
                               initialize = TRUE, usefit = FALSE, score = TRUE, information = TRUE, hessian = FALSE, vcov = TRUE, residuals = TRUE, leverage = FALSE, dVcov = FALSE, dVcov.robust = FALSE)
    }
    return( c(list(x=x, data=data, estimator = estimator),dots) )
}

##' @rdname gaussian_weight
##' @export
gaussian_weight_method.lvm <- "nlminb2"

## * gaussian_weight_logLik.lvm
##' @rdname gaussian_weight
##' @export
`gaussian_weight_logLik.lvm` <- function(object, type="cond", p, data, weights,...) {
    ## ** compute mu and Omega
    if(type!="cond"){
        stop("Not implemented for other types than \"cond\"\n ")
    }
    cM <- moments2(object, param = p, data = data, weights = weights,
                   initialize = FALSE, usefit = TRUE, score = FALSE, information = FALSE, hessian = FALSE, vcov = FALSE, residuals = TRUE, leverage = FALSE, dVcov = FALSE, dVcov.robust = FALSE)
    
    ## ** prepare
    name.pattern <- cM$missing$name.pattern
    missing.pattern <- cM$missing$pattern
    unique.pattern <- cM$missing$unique.pattern
    n.pattern <- length(name.pattern)
    
    OmegaM1 <- cM$moment$OmegaM1
    residuals <- cM$residuals
    logLik <- 0

    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iOmegaM1 <- OmegaM1[[iPattern]]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)
        iResiduals <- residuals[iIndex,iY,drop=FALSE]
        iM <- length(iY)
        if(is.null(weights)){
            logLik <- logLik - (cM$cluster$n.cluster/2) * (iM * log(2*pi) - log(det(iOmegaM1))) - sum((iResiduals %*% iOmegaM1) * iResiduals)/2
        }else{
            logLik <- logLik - (sum(weights)/2) * (iM * log(2*pi) - log(det(iOmegaM1))) - sum(weights[,1]/2 * rowSums((iResiduals %*% iOmegaM1) * iResiduals))
        }
    }
    return(logLik)
}

##' @rdname gaussian_weight
##' @export
`gaussian_weight_objective.lvm` <- function(x, ...) {
    logLik <- gaussian_weight_logLik.lvm(object = x,...)
    return(-logLik)
}

## * gaussian_weight_score.lvm
##' @rdname gaussian_weight
##' @export
gaussian_weight_score.lvm <- function(x, data, p, S, n, mu=NULL, weights=NULL, debug=FALSE, reindex=FALSE, mean=TRUE, constrain=TRUE, indiv=FALSE,...) {

    ## if(constrain){
    ##     stop("gaussian_weight_score.lvm does not handle constrain")
    ## }
    ## if(reindex){
    ##     stop("gaussian_weight_score.lvm does not handle reindex")
    ## }
    ## if(!mean){
    ##     stop("gaussian_weight_score.lvm only handles mean")
    ## }
    
    ## ** compute moments
    cM <- moments2(x, param = p, data = data, weights = weights,
                   initialize = FALSE, usefit = TRUE, score = TRUE, information = FALSE, hessian = FALSE, vcov = FALSE, residuals = FALSE, leverage = FALSE, dVcov = FALSE, dVcov.robust = FALSE)

    ## ** export
    if(indiv){
        return(cM$score)
    }else{
        return(colSums(cM$score))
    }
}

## * gaussian_weight_gradient.lvm
##' @rdname gaussian_weight
##' @export
gaussian_weight_gradient.lvm <-  function(...) {
    return(-gaussian_weight_score.lvm(...))
}

## * gaussian_weight_hessian.lvm
##' @rdname gaussian_weight
##' @export
`gaussian_weight_hessian.lvm` <- function(x, p, n, weights=NULL,...) {

    ## ** compute moments
    cM <- moments2(x, param = p, weights = weights,
                   initialize = FALSE, usefit = TRUE, score = FALSE, information = TRUE, hessian = FALSE, vcov = FALSE, residuals = FALSE, leverage = FALSE, dVcov = FALSE, dVcov.robust = FALSE)

    ## ** export
    return(cM$information)
}

######################################################################
### Objective_gaussian_weight.R ends here

### utils-nlme.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 15 2017 (17:29) 
## Version: 
## Last-Updated: dec 10 2018 (23:28) 
##           By: Brice Ozenne
##     Update #: 671
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * .coef2
#' @title Export Mean and Variance Coefficients
#' @description Export mean and variance coefficients
#' from a \code{lm}, \code{gls}, or \code{lme} object.
#' @name coef2-internal
#'
#' @param object a \code{lm}, \code{gls} or \code{lme} object.
#' @param name.Y [character] the name of the endogenous variable. Used to name certain variance parameters.
#' 
#' @details The variance coefficients that are exported are the residual variance of each outcome. 
#' This is \eqn{\sigma^2} for the first one and \eqn{k^2 \sigma^2} for the remaining ones.
#'
#' @return A numeric vector named with the names of the coefficient with three attributes:
#' \itemize{
#' \item mean.coef: the name of the mean coefficients.
#' \item var.coef: the name of the variance coefficients.
#' \item cor.coef:  the name of the correlation coefficients.
#' }
#'
#' @concept extractor
#' @keywords internal
`.coef2` <-
    function(object) UseMethod(".coef2")

## * .coef2.lm
#' @rdname coef2-internal
.coef2.lm <- function(object){
    coef.object <- coef(object)
    p <- c(coef.object,sigma2=sigma(object)^2)
    attr(p, "mean.coef") <- names(coef.object)
    attr(p, "var.coef") <- "sigma2"
    attr(p, "cor.coef") <- NULL
    return(p)
}

## * .coef2.gls
#' @rdname coef2-internal
.coef2.gls <- function(object){

    ## *** mean coefficients
    mean.coef <- stats::coef(object)

    ## *** variance coefficients
    var.coef <- c(sigma2 = stats::sigma(object)^2)
    if(!is.null(object$modelStruct$varStruct)){
        var.coef <- c(var.coef,
                      stats::coef(object$modelStruct$varStruct, unconstrained = FALSE, allCoef = FALSE)^2)          
    }

    ## *** covariance coefficients
    if(!is.null(object$modelStruct$corStruct)){
        cor.coef <- stats::coef(object$modelStruct$corStruct, unconstrained = FALSE)

        n.var <- length(var.coef)
        n.cor <- length(cor.coef)

        ## check unstructured covariance matrix
        if(!is.null(object$modelStruct$varStruct) && ((n.var*(n.var-1))/2 == n.cor)){

            vecgroup <- attr(unclass(object$modelStruct$corStruct), "group")
            veccov.cor <- unname(unlist(attr(object$modelStruct$corStruct, "covariate")))
            veccov.var <- attr(object$modelStruct$varStruct, "groups")

            table.covvar <- table(veccov.cor,veccov.var)
            newlevels.cor <- colnames(table.covvar)[apply(table.covvar, 1, which.max)]
            veccov.cor2 <- factor(veccov.cor, levels = 0:max(veccov.cor), labels = newlevels.cor)
            
            if(identical(as.character(veccov.cor2),as.character(veccov.var))){

                cor.coefName <- apply(utils::combn(newlevels.cor, m = 2), MARGIN = 2, FUN = paste, collapse = "")
                names(cor.coef) <- paste0("corCoef",cor.coefName)

            }else{
                names(cor.coef) <- paste0("corCoef",1:length(cor.coef))
            }
            
            
        }else{
            names(cor.coef) <- paste0("corCoef",1:length(cor.coef))
        }
        
        
    }else{
        cor.coef <- NULL
    }

    p <- c(mean.coef, cor.coef, var.coef)
    attr(p, "mean.coef") <- names(mean.coef)
    attr(p, "var.coef") <- names(var.coef)
    attr(p, "cor.coef") <- names(cor.coef)
    return(p)
}




## * .coef2.lme
#' @rdname coef2-internal
.coef2.lme <- function(object){

     ## *** mean coefficients
    mean.coef <- nlme::fixef(object)

    ## *** variance coefficients
    var.coef <- c(sigma2 = stats::sigma(object)^2)
    if(!is.null(object$modelStruct$varStruct)){
       var.coef <- c(var.coef,
                      stats::coef(object$modelStruct$varStruct, unconstrained = FALSE, allCoef = FALSE)^2)   
    }

    ## *** random effect coefficients
    random.coef <- as.double(nlme::getVarCov(object))    
    names(random.coef) <- paste0("ranCoef",1:length(random.coef))

     ## *** correlation coefficients
    if(!is.null(object$modelStruct$corStruct)){
        cor.coef <- stats::coef(object$modelStruct$corStruct, unconstrained = FALSE)
        names(cor.coef) <- paste0("corCoef",1:length(cor.coef))
    }else{
        cor.coef <- NULL
    }
    
    p <- c(mean.coef, cor.coef, var.coef, random.coef)
    attr(p, "mean.coef") <- names(mean.coef)
    attr(p, "var.coef") <- names(var.coef)
    attr(p, "cor.coef") <- names(cor.coef)
    attr(p, "ran.coef") <- names(random.coef)
    return(p)
}

## * .getFormula2
`.getFormula2` <-
    function(object) UseMethod(".getFormula2")

## * .getFormula2.gls
.getFormula2.gls <- function(object){
    return(evalInParentEnv(object$call$model))
}

## * .getFormula2.lme
.getFormula2.lme <- function(object){
    return(evalInParentEnv(object$call$fixed))
}


## * .getCluster2
#' @title Reconstruct the Cluster Variable from a nlme Model
#' @description Reconstruct the cluster variable from a nlme model.
#' @name getCluster2-internal
#'
#' @param object a \code{gls} or \code{lme} object.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' Only required for \code{gls} models with no correlation argument.
#' @param data [data.frame] the data set.
#' @param ... [internal] Only used by the generic method.
#'  
#' @return A list containing:
#' \itemize{
#' \item cluster: the cluster index for each observation.
#' \item n.cluster: the number of clusters.
#' }
#'
#' @concept extractor
#' @keywords internal
`.getCluster2` <-
    function(object, ...) UseMethod(".getCluster2")

## * .getCluster2.gls
#' @rdname getCluster2-internal
.getCluster2.gls <- function(object, cluster, data, ...){

### ** get cluster
    if(is.null(object$modelStruct$corStruct)){
        if(missing(cluster)){
            stop("cluster must be specified for gls object with no correlation structure \n")
        }        
        if(length(cluster) == 1 && is.character(cluster)){
            if(cluster %in% names(data) == FALSE){
                stop("Variable \"",cluster,"\" not in data \n")
            }
            ## cluster <- as.numeric(factor(data[[cluster]], levels = unique(data[[cluster]])))
            cluster <- as.numeric(as.factor(data[[cluster]]))
            
        }else if(length(cluster)==NROW(data)){
            cluster <- as.numeric(as.factor(cluster))
        }else{
            stop("length of cluster and data do not match \n")
        }
    }else{
        cluster <- as.numeric(nlme::getGroups(object))
    }
    n.cluster <- length(unique(cluster))

### ** reorder cluster according to the data ordering
    levels.cluster <- unique(cluster)
    cluster <- as.numeric(factor(cluster, levels = levels.cluster))

### ** export
    return(list(cluster = cluster,
                levels.cluster = levels.cluster,
                n.cluster = n.cluster))
}

## * .getCluster2.lme
#' @name getCluster2-internal
.getCluster2.lme <- function(object, ...){

### ** get cluster
    if(NCOL(object$groups)!=1){
        stop("cannot only handle one random effect \n")
    }
    cluster <- as.numeric(nlme::getGroups(object))
    n.cluster <- length(unique(cluster))

### ** reorder cluster according to the data ordering
    levels.cluster <- unique(cluster)
    cluster <- as.numeric(as.factor(cluster))
    ## cluster <- as.numeric(factor(cluster, levels = levels.cluster))

### ** export
    return(list(cluster = cluster,
                levels.cluster = levels.cluster,
                n.cluster = n.cluster))
}

## * .getIndexOmega2
#' @title Extract the name of the endogenous variables
#' @description Extract the name of the endogenous variables from a nlme model.
#' @name getIndexOmega2-internal
#'
#' @param object a \code{gls} or \code{lme} object.
#' @param param [numeric vector] the mean and variance coefficients.
#' @param attr.param [character vector] the type of each coefficients (mean or variance).
#' @param name.Y [character] name of the endogenous variable.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' Only required for \code{gls} models with no correlation argument.
#' @param data [data.frame] the data set.
#' @param ... [internal] Only used by the generic method.
#'  
#' @return A list containing:
#' \itemize{
#' \item index.Omega: [list of integer vector] For each cluster of observations,
#' the index of the endogenous variable relative to each observation.
#' \item n.endogenous: [integer] the number of endogenous variables.
#' \item name.endogenous: [character vector] the name of the endogenous variables.
#' \item ref.group: [character vector] the levels of the variable defining the variance component in a generic covariance matrix.
#' }
#'
#' @concept extractor
#' @keywords internal
`.getIndexOmega2` <-
    function(object, ...) UseMethod(".getIndexOmega2")

## * .getIndexOmega2.gls
#' @rdname getIndexOmega2-internal
.getIndexOmega2.gls <- function(object, param, attr.param, 
                                name.Y, cluster, levels.cluster, data){

    class.cor <- class(object$modelStruct$corStruct)
    class.re <- class(object$modelStruct$reStruct)
    if("NULL" %in% class.cor == FALSE){
        formula.cor <- attr(object$modelStruct$corStruct,"formula")
        varIndex.cor <- all.vars(nlme::getCovariateFormula(formula.cor))
    }else{
        varIndex.cor <- NULL
    }

    class.var <- class(object$modelStruct$varStruct)
    if("NULL" %in% class.var == FALSE){
        formula.var <- attr(object$modelStruct$varStruct,"formula")
        varIndex.var <- all.vars(nlme::getCovariateFormula(formula.var))
        groupValue.var <- attr(object$modelStruct$varStruct,"groups")
        if("NULL" %in% class.cor == FALSE || "NULL" %in% class.re == FALSE){ ## nlme automaticly sort data when corStruct or reSruct
            groupValue.var <- groupValue.var[order(order(cluster))] ## undo automatic sort
        } 
    }else{
        varIndex.var <- NULL
    }

    ## ** Check admissible var-cor structure
    validClass.cor <- c("NULL","corCompSymm","corSymm","corStruct")
    if(any(class.cor %in% validClass.cor == FALSE)){
        stop("Can only handle corStruct of class \"corCompSymm\" or \"corSymm\"\n")
    }

    validClass.var <- c("NULL","varIdent","varFunc")
    if(any(class.var %in% validClass.var == FALSE)){
        stop("Can only handle varStruct of class \"varIdent\"\n")
    }

    ## ** Check compatible ordering between var and cor
    if(length(varIndex.cor) > 0 && length(varIndex.cor) > 0 && !identical(varIndex.cor,varIndex.cor)){
        stop("Inconsistency between the left hand side of the formula in corStruct and the left hand side of the formula in varStruct. \n",
             "it should be something like: correlation = corStruct(form = ~index|groupA) \n",
             "                             weight = varStruct(form = ~index|groupB) \n")
    }
    
    ## ** Identify the index and name of the endogenous variables
    if("NULL" %in% class.var && "NULL" %in% class.cor){ ## basic lme models or lm-ish models
        ## order of the variables does not matter
        index.Omega <- tapply(cluster,cluster,function(iC){list(1:length(iC))})
        norm <- FALSE
    }else if("NULL" %in% class.var && "corCompSymm" %in% class.cor){
        ## order of the variables does not matter
        index.Omega <- attr(object$modelStruct$corStruct, "covariate")[levels.cluster]
        norm <- TRUE
    }else if(length(varIndex.cor)!=0){ 
        ## order of the variables matters: use index variable in corStruct
        index.Omega <- attr(object$modelStruct$corStruct, "covariate")[levels.cluster]
        norm <- TRUE
    }else if(length(varIndex.var)!=0){
        ## order of the variables matters: using index variable in varStruct
        index.tempo <- data[[varIndex.var]]
        if(!is.numeric(index.tempo)){
            stop("The variable in the left hand side of the formula in varStruct must be numeric \n")
        }
        if(!is.null(object$na.action)){
            index.tempo <- index.tempo[-object$na.action]
        }
        index.Omega <- tapply(index.tempo, cluster, function(iC){list(iC)})
        norm <- TRUE
    }else{
        ## order of the variables matters: check missing values
        if("NULL" %in% class.var == FALSE){
            test.duplicated <- unique(unlist(tapply(groupValue.var, cluster, duplicated)))
        }else{
            groupValue.cor <- attr(object$modelStruct$corStruct, "covariate")
            ref.tempo <- unname(sort(groupValue.cor[[1]]))
            test.identical <- unique(unlist(lapply(groupValue.cor,function(x){ # x <- groupValue.cor[[1]]
                identical(unname(sort(x)),ref.tempo)
            })))
            test.duplicated <- TRUE
        }
        ## recover order
        if(("NULL" %in% class.var == FALSE) && all(test.duplicated==FALSE)){
            ## from varIdent when no missing values
            tempo <- as.numeric(factor(groupValue.var, levels = attr(object$modelStruct$varStruct,"groupNames")))
            index.Omega <- tapply(tempo, cluster, function(iC){iC})
            norm <- FALSE
        }else{
            ## from the order of the data
            index.Omega <- tapply(cluster,cluster,function(iC){list(1:length(iC))})
            norm <- FALSE
            if("NULL" %in% class.cor == FALSE){
                if(any(test.identical == FALSE)){
                    warning("The residuals covariance matrice is subset based on the ordering of the data\n",
                            "It is safer to define the ordering within group by adding a variable in the left hand side of the formula in corStruct \n",
                            "e.g. correlation = corStruct(form = ~index|group) \n")
                }
            }else{
                warning("The attribution of the repetition number is based on the ordering of the data\n",
                        "It is safer to define the ordering within group by adding a variable in the left hand side of the formula in varStruct \n",
                        "e.g. correlation = varStruct(form = ~index|group) \n")
            }
        }
    }
    attr(object$modelStruct$varStruct,"groupNames")
    attr(object$modelStruct$varStruct,"groups")

    ## ** Normalize index.Omega
    if(norm){
        level.index <- unique(unlist(index.Omega))
        convertion <- setNames(order(level.index), level.index)
        index.Omega <- lapply(index.Omega, function(x){as.double(convertion[as.character(x)])})
    }

    ## ** Define the name and number of endogenous variables
    vecIndex.Omega <- unlist(index.Omega)
    n.endogenous <- max(vecIndex.Omega)
    name.endogenous <- paste0(name.Y,".",1:n.endogenous)

    if("corSymm" %in% class.cor){
        ## second order polynomial equation
        ## m(m-1)/2 = n.cor
        ## i.e. m = 1/2 + sqrt(1+8 n.cor)/2
        cor.coef <- param[attr.param$cor.coef]
        if(n.endogenous != ( 1 + sqrt(1 + 8 * length(cor.coef)) ) / 2){
            stop("The values of ",varIndex.cor," does not match the number of correlation coefficients \n",
                 "i.e. the maximum value of ",varIndex.cor," should equal ( 1 + sqrt(1 + 8 * n.cor.coef)) ) / 2 \n"
                 )
        }
    }

    if("NULL" %in% class.var == FALSE){

        groupValue.var.ordered <- groupValue.var[order(cluster)] ## reorder by cluster
        table.unique <- tapply(1:length(vecIndex.Omega),vecIndex.Omega,function(x){
            length(unique(groupValue.var.ordered[x]))
        })
        if(any(table.unique!=1)){
            stop("The residual covariance matrix should not differ between clusters \n")
        }                
        ref.group <- groupValue.var.ordered[!duplicated(vecIndex.Omega)]
    }else{
        ref.group <- NULL
    }
        
    ## ** Export
    return(list(index.Omega = index.Omega,
                n.endogenous = n.endogenous,
                name.endogenous = name.endogenous,
                ref.group = ref.group))
}

## * .getIndexOmega2.lme
#' @rdname getIndexOmega2-internal
.getIndexOmega2.lme <- .getIndexOmega2.gls

## * .getVarCov2
#' @title Reconstruct the Marginal Variance Covariance Matrix from a nlme Model
#' @description Reconstruct the marginal variance covariance matrix from a nlme model.
#' @name getVarCov2-internal
#'
#' @param object a \code{gls} or \code{lme} object
#' @param param [numeric vector] the mean and variance coefficients.
#' @param attr.param [character vector] the type of each coefficients (mean or variance).
#' @param name.endogenous [character vector] name of each repetition of the endogenous variable. 
#' @param n.endogenous [integer >0] number of repetitions of the endogenous variable.
#' @param ref.group [character vector] the levels of the variable defining the variance component in a generic covariance matrix.
#' @param ... [internal] Only used by the generic method.
#'  
#' @return [matrix] the marginal variance covariance matrix for a full sample.
#'
#' @concept extractor
#' @keywords internal
`.getVarCov2` <-
    function(object, ...) UseMethod(".getVarCov2")

## * .getVarCov2.gls
#' @rdname getVarCov2-internal
.getVarCov2.gls <- function(object, param, attr.param,
                            name.endogenous, n.endogenous, ref.group, ...){

    ## ** Extract information
    var.coef <- param[attr.param$var.coef]
    cor.coef <- param[attr.param$cor.coef]

    ## ** Diagonal terms
    if(length(ref.group)>0){        
        factor.varcoef <- setNames(c(1,var.coef[-1]),
                                   attr(object$modelStruct$varStruct,"groupNames"))
        sigma2.base <- var.coef["sigma2"] * factor.varcoef[ref.group]
    }else{
        sigma2.base <- rep(var.coef["sigma2"], n.endogenous)
    }
    ## re-order according to the order of the correlation coefficients (if possible)
    if(length(cor.coef)>1 & length(var.coef)>1){
        cor.level <- gsub("corCoef","",names(cor.coef))
        var.level <- names(sigma2.base)
        var.permlevel <- .allPermutations(var.level)

        M.try <- apply(var.permlevel, MARGIN = 1, function(iLevel){
            all(apply(utils::combn(iLevel, m = 2), MARGIN = 2, FUN = paste, collapse = "") == cor.level)
        })
        if(any(M.try)){
            sigma2.base <- sigma2.base[var.permlevel[which(M.try),]]
        }
    }
    
    Omega <- diag(as.double(sigma2.base),
                  nrow = n.endogenous, ncol = n.endogenous)

    ## ** Extra-diagonal terms
    if(length(cor.coef)>0){
        index.lower <- which(lower.tri(Omega))
        index.lower.arr <- which(lower.tri(Omega),arr.ind = TRUE)
        vec.sigma.tempo <- apply(index.lower.arr,1,function(x){prod(sqrt(sigma2.base[x]))})        
        Omega[index.lower] <- cor.coef*vec.sigma.tempo
        Omega <- symmetrize(Omega, update.upper = TRUE)
    }    

    ## ** names
    if(all(!duplicated(names(sigma2.base)))){
        dimnames(Omega) <- list(names(sigma2.base), names(sigma2.base))
    }else{
        dimnames(Omega) <- list(name.endogenous, name.endogenous)
    }

    
    ## ** export
    return(Omega)
}

## * .getVarCov2.lme
#' @rdname getVarCov2-internal
.getVarCov2.lme <- function(object, param, attr.param, ...){

    ## ** prepare with gls
    out <- .getVarCov2.gls(object, param = param, attr.param = attr.param, ...)

    ## ** add contribution of the random effect
    ran.coef <- param[attr.param$ran.coef]
    out <- out + ran.coef

    ## ** export
    return(out)    
}

## * .allPermutations
## .allPermutations(1:3)
## .allPermutations(2:3)
.allPermutations <- function(vec){
    X <- lapply(vec, function(x){
        cbind(x, .allPermutations(setdiff(vec, x)))
    })
    return(unname(do.call(rbind,X)))
}

##----------------------------------------------------------------------
### utils-nlme.R ends here

### test-Utils-nlme.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 16 2017 (10:36) 
## Version: 
## Last-Updated: dec 10 2018 (23:44) 
##           By: Brice Ozenne
##     Update #: 69
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

library(nlme)
lava.options(symbols = c("~","~~"))

context("Utils-nlme")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~1*eta,Y2~1*eta,Y3~1*eta,Y4~1*eta,eta~G+Gender))
latent(mSim) <- ~eta
categorical(mSim, labels = c("M","F")) <- ~Gender
transform(mSim,Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
dW <- lava::sim(mSim,n,latent = FALSE)
dW <- dW[order(dW$Id),,drop=FALSE]
dL <- reshape2::melt(dW,id.vars = c("G","Id","Gender"), variable.name = "time")
dL <- dL[order(dL$Id),,drop=FALSE]
dL$time.num <- as.numeric(dL$time)

## * t.test
test_that("invariant to the order in the dataset", {
    e1.gls <- gls(Y1 ~ Gender, data = dW[order(dW$Id),],
                  weights = varIdent(form = ~1|Gender),
                  method = "ML")

    out1 <- getVarCov2(e1.gls, cluster = dW$Id)
    index.cluster <- as.numeric(names(out1$index.Omega))
    expect_true(all(diff(index.cluster)>0))

    e2.gls <- gls(Y1 ~ Gender, data = dW[order(dW$Gender),],
                  weights = varIdent(form = ~1|Gender),
                  method = "ML")
    out2 <- getVarCov2(e2.gls, cluster = dW$Id)
    index.cluster <- as.numeric(names(out2$index.Omega))
    expect_true(all(diff(index.cluster)>0))
})

## * Heteroschedasticity
e.gls <- nlme::gls(value ~ time + G + Gender,
                   weights = varIdent(form =~ 1|time),
                   data = dL, method = "ML")

test_that("Heteroschedasticity", {
    vec.sigma <- c(1,coef(e.gls$modelStruct$varStruct, unconstrained = FALSE))
    expect_equal(diag(vec.sigma^2 * sigma(e.gls)^2),
                 unname(getVarCov2(e.gls, cluster = "Id")$Omega))
})

## * Compound symmetry
e.lme <- nlme::lme(value ~ time + G + Gender,
                   random = ~ 1|Id,
                   data = dL,
                   method = "ML")
e.lme.bis <- nlme::lme(value ~ time + G + Gender,
                       random = ~ 1|Id,
                       correlation = corCompSymm(),
                       data = dL,
                       method = "ML")
e.gls <- nlme::gls(value ~ time + G + Gender,
                   correlation = corCompSymm(form=~ 1|Id),
                   data = dL, method = "ML")

test_that("Compound symmetry", {
    expect_equal(unclass(getVarCov(e.gls)),
                 unname(getVarCov2(e.gls)$Omega))

    expect_equal(unname(getVarCov(e.lme, type = "marginal", individuals = 1)[[1]]),
                 unname(getVarCov2(e.lme)$Omega))

    expect_equal(unname(getVarCov(e.lme.bis, type = "marginal", individuals = 1)[[1]]),
                 unname(getVarCov2(e.lme.bis)$Omega))
})

## * Unstructured 
e.lme <- nlme::lme(value ~ time + G + Gender,
                   random = ~ 1|Id,
                   correlation = corSymm(form =~ time.num|Id),
                   data = dL,
                   method = "ML")
e.gls <- nlme::gls(value ~ time + G + Gender,
                   correlation = corSymm(form=~ time.num|Id),
                   data = dL, method = "ML")


test_that("Unstructured ", {
    expect_equal(unclass(getVarCov(e.gls)),
                 unname(getVarCov2(e.gls)$Omega))

    expect_equal(unname(getVarCov(e.lme, type = "marginal", individuals = 1)[[1]]),
                 unname(getVarCov2(e.lme)$Omega))
})

## * Unstructured with weights
e.lme <- nlme::lme(value ~ time + G + Gender,
                   random = ~ 1|Id,
                   correlation = corSymm(form =~ time.num|Id),
                   weight = varIdent(form = ~ 1|time),
                   data = dL,
                   method = "ML")
e.gls <- nlme::gls(value ~ time + G + Gender,
                   correlation = corSymm(form =~ time.num|Id),
                   weight = varIdent(form = ~ 1|time),
                   data = dL, method = "ML")

test_that("Unstructured with weights", {
    expect_equal(unclass(getVarCov(e.gls)),
                 unname(getVarCov2(e.gls)$Omega))

    expect_equal(unname(getVarCov(e.lme, type = "marginal", individuals = 1)[[1]]),
                 unname(getVarCov2(e.lme)$Omega))
})

## * Unstructured with missing data
## http://publicifsv.sund.ku.dk/~jufo/courses/rm2018/vasscores.txt
## fix bug in the ordering of getVarCov2 due to different ordering of treatment in arguments weight and correlation

## ** data management
## butils::object2script(dfW.score)
dfW <- data.frame("id" = c("1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "30"), 
                  "group" = c("AC", "AB", "AB", "BC", "BC", "AC", "AB", "AC", "BC", "AC", "BC", "AB", "AB", "BC", "AB", "AC", "BC", "AC", "AC", "AC", "BC", "AB", "AB", "BC", "AB", "AB", "BC", "AC", "BC", "AC"), 
                  "vasaucA" = c( 51.0,  42.0,  54.0, NA, NA,  16.5,  58.5, 129.0, NA,  52.5, NA,  23.5,  98.0, NA, 177.0,  67.0, NA,  55.0,  79.5,   3.5, NA,  33.0,   9.5, NA,  47.5,  66.5, NA,  85.5, NA, 143.5), 
                  "vasaucB" = c(NA,  35.0,  62.0,  64.0,  80.5, NA,  33.5, NA,  59.0, NA,  32.5,  13.0, 120.0, 102.0, 166.5, NA, 138.0, NA, NA, NA, 161.5,  53.5,  13.5, 116.5,  68.0, 104.5, 103.0, NA,  36.0, NA), 
                  "vasaucC" = c( 48.5, NA, NA,  65.0,  94.5,  19.5, NA, 102.0,  56.5,  78.5,  18.0, NA, NA,  14.0, NA,  51.0, 168.5,  10.0,  28.0,   3.5, 127.0, NA, NA,  36.5, NA, NA,  33.5,  45.0,   7.5, 132.0))

level.Id <- sort(as.numeric(as.character(dfW$id)))

dfW$id <- factor(dfW$id, levels = level.Id)
dfW$group <- as.factor(dfW$group)
dfL <- reshape2::melt(dfW, id.vars = c("id","group"),
                      measure.vars = c("vasaucA","vasaucB","vasaucC"),
                      value.name = "vasauc",
                      variable.name = "treatment")
dfL <- dfL[order(dfL$id, dfL$treatment),]
dfL$treatment <- gsub("vasauc","",dfL$treatment)
dfL$treatment <- as.factor(dfL$treatment)
dfL$treatment.num <- as.numeric(dfL$treatment)


dfL2 <- dfL
dfL2$id <- as.character(dfL2$id)
dfL2[dfL2$id == "2","id"] <- "0"
dfL2[dfL2$id == "1","id"] <- "2"
dfL2[dfL2$id == "0","id"] <- "1"
dfL2$id <- factor(dfL2$id, levels = level.Id)
dfL2 <- dfL2[order(dfL2$id,dfL2$treatment),]


## ** fit model
e.gls <- gls(vasauc ~ treatment,
             correlation = corSymm(form =~ treatment.num | id),
             weights = varIdent(form =~ 1|treatment),
             na.action = na.omit,
             data = dfL)
logLik(e.gls)

e.gls2 <- gls(vasauc ~ treatment,
             correlation = corSymm(form =~ treatment.num | id),
             weights = varIdent(form =~ 1|treatment),
             na.action = na.omit,
             data = dfL2)
logLik(e.gls2)

## ** extract covariance matrix
Sigma <- unname(getVarCov2(e.gls)$Omega)
Sigma2 <- unname(getVarCov2(e.gls2)$Omega)

expect_equal(Sigma, Sigma2, tol = 1e-5)
## allcoef <- lavaSearch2:::.coef2.gls(e.gls)
## sigmaBase <- allcoef["sigma2"] * c(A=1,allcoef["B"],allcoef["C"])

## AB
expect_equal(Sigma[c(1,2),c(1,2)],
             unclass(nlme::getVarCov(e.gls2, individual = 1)),
             tol = 1e-5)
expect_equal(Sigma[c(1,2),c(1,2)],
             unclass(nlme::getVarCov(e.gls, individual = 2)),
             tol = 1e-5)
## sqrt(sigmaBase["A"] * sigmaBase["B"]) * allcoef["corCoefAB"]

## AC
expect_equal(Sigma[c(1,3),c(1,3)],
             unclass(nlme::getVarCov(e.gls2, individual = 2)),
             tol = 1e-5)
expect_equal(Sigma[c(1,3),c(1,3)],
             unclass(nlme::getVarCov(e.gls, individual = 1)),
             tol = 1e-5)
## sqrt(sigmaBase["A"] * sigmaBase["C"]) * allcoef["corCoefAC"]

## BC
expect_equal(Sigma[c(2,3),c(2,3)],
             unclass(nlme::getVarCov(e.gls2, individual = 4)),
             tol = 1e-5)
expect_equal(Sigma[c(2,3),c(2,3)],
             unclass(nlme::getVarCov(e.gls, individual = 4)),
             tol = 1e-5)
## sqrt(sigmaBase["B"] * sigmaBase["C"]) * allcoef["corCoefBC"]


## * 2 random effect model (error)
e.lme <- nlme::lme(value ~ time + G + Gender,
                   random=~1|Id/Gender,
                   data = dL,
                   method = "ML")

expect_error(getVarCov2(e.lme))

## * PET dataset

df.PET <- data.frame("ID" = c( 925, 2020, 2059, 2051, 2072, 2156, 2159, 2072, 2020, 2051, 2231,
                              2738, 2231, 2777,  939,  539, 2738, 2777,  925, 2156, 2159, 2059), 
                     "session" = c("V", "V", "V", "V", "V", "V", "V", "C", "C", "C", "C",
                                   "C", "V", "C", "C", "V", "V", "V", "C", "C", "C", "C"), 
                     "PET" = c(-2.53, -6.74, -8.17, -2.44, -3.54, -1.27, -0.55, -0.73, -1.42,  3.35,
                               -2.11,  2.60, -4.52,  0.99, -1.02, -1.78, -5.86,  1.20, NA, NA, NA, NA)
                     )
df.PET$session.index <- as.numeric(as.factor(df.PET$session))


e.lme <- lme(PET ~ session,
             random = ~ 1 | ID,
             weights = varIdent(form=~session.index|session),
             na.action = "na.omit",
             data = df.PET)
test_that("getVarCov2 - NA", {
    expect_equal(matrix(c( 7.893839, 1.583932, 1.583932, 4.436933), 2, 2),
                 unname(getVarCov2(e.lme)$Omega), tol = 1e-6, scale = 1)
})


##----------------------------------------------------------------------
### test-Utils-nlme.R ends here

### Scorrect-model.tables.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: Jan 12 2022 (09:41) 
## Version: 
## Last-Updated: jan 23 2024 (10:27) 
##           By: Brice Ozenne
##     Update #: 31
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Estimate, Confidence Intervals, and P-value With Small Sample Correction
#' @description Extract estimate, standard error, confidence intervals and p-values associated to each coefficient of a latent variable model.
#' Similar to \code{lava::confint} but with small sample correction.
#' @name confint2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param robust [logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] when \code{TRUE} uses the same names as when using \code{stats::coef}.
#' @param transform [function] transformation to be applied.
#' @param conf.level [numeric, 0-1] level of the confidence intervals.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
#' 
#' @return A data.frame with a row per coefficient.
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`model.tables2` <- function(object, robust, cluster, transform,
                            as.lava, conf.level, ...) UseMethod("model.tables2")

## * model.tables.lvmfit
#' @export
model.tables.lvmfit <- function(x, ...){

    x.confint <- confint(x, ...)
    out <- cbind(summary(x, ...)$coef, upper = NA, lower = NA)
    colnames(out)[1:4] <- c("estimate","se","statistic","p.value")
    out <- out[,c("estimate","se","lower","upper","statistic","p.value")]
    out[rownames(x.confint),"lower"] <- x.confint[,1]
    out[rownames(x.confint),"upper"] <- x.confint[,2]

    return(out)
}

## * model.tables2.lvmfit
#' @export
model.tables2.lvmfit <- function(object, robust = FALSE, cluster = NULL,
                                 transform = NULL, as.lava = TRUE, conf.level = 0.95, 
                                 ssc = lava.options()$ssc, df = lava.options()$df, ...){

    return(model.tables(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...),
                        robust = robust, cluster = cluster, as.lava = as.lava, conf.level = conf.level,
                        transform = transform))

}

## * model.tables.lvmfit2
#' @export
model.tables2.lvmfit2 <- function(object, robust = FALSE, cluster = NULL,
                            transform = NULL, as.lava = TRUE, conf.level = 0.95,  ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    ## ** new model parameters
    param <- coef(object, as.lava = as.lava)
    name.param <- names(param)
    n.param <- length(name.param)

    ## ** new Wald test
    type <- object$sCorrect$skeleton$type
    type <- type[!is.na(type$lava),]
    null <- stats::setNames(rep(0, n.param),name.param)
    if(any(type$detail %in% c("Sigma_var","Psi_var"))){
        param.var <- type[type$detail %in% c("Sigma_var","Psi_var"),"param"]
        if(as.lava){
            null[names(object$sCorrect$skeleton$originalLink2param)[match(param.var,object$sCorrect$skeleton$originalLink2param)]] <- NA
        }else{
            null[object$sCorrect$skeleton$originalLink2param[match(param.var,object$sCorrect$skeleton$originalLink2param)]] <- NA
        }
    }
    table.all <- compare2(object,
                          linfct = name.param,
                          rhs = null,
                          robust = robust,
                          cluster = NULL,
                          F.test = FALSE,
                          as.lava = FALSE,
                          sep = c("",""))

    tableS.all <- summary(table.all, test = multcomp::adjusted("none"), transform = transform, conf.level = conf.level, rowname.rhs = FALSE)$table2

    if(as.lava){
        tableS.all <- tableS.all[names(object$sCorrect$skeleton$originalLink2param),,drop=FALSE]
    }else{
        tableS.all <- tableS.all[as.character(object$sCorrect$skeleton$originalLink2param),,drop=FALSE]
    }
    return(tableS.all)

}

## *  model.tables.lvmfit2
#' @export
model.tables.lvmfit2 <- function(x, ...){ ## necessary as model.tables must have x has first argument
    return(model.tables2(x, ...))
}

##----------------------------------------------------------------------
### Scorrect-model.tables.R ends here

### autoplot.calibrateType1.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: apr  5 2018 (13:20) 
## Version: 
## Last-Updated: Jan 11 2022 (16:46) 
##           By: Brice Ozenne
##     Update #: 30
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
##' @title Graphical Display of the Bias or Type 1 Error
##' @description Graphical display of the bias or type 1 error
##' for the output of \code{\link{calibrateType1}}.
##' @name autoplot_calibrateType1
##' 
##' @param object output of \code{\link{calibrateType1}}.
##' @param type [character] if type equals \code{"bias"} the bias will be displayed.
##' Otherwise if it equals \code{"type1error"} the type 1 error will be displayed.
##' @param plot [logical] should the plot be displayed?
##' @param color.threshold [character] the color for the line representing the expected value(s).
##' @param type.bias [character] if type.bias equals \code{"absolute"} the absolute bias will be used.
##' Otherwise if it equals \code{"relative"} the relative bias will be used.
##' Only relevant when type equals \code{"bias"}.
##' @param alpha [numeric, 0-1] the significance threshold to consider.
##' Only relevant when type equals \code{"type1error"}.
##' @param nrow.legend [integer, >0] the number of rows for the legend.
##' Only relevant when type equals \code{"type1error"}.
##' @param name2label [named character vector] the label for the legend.
##' The vector should contain the method names (see details).
##' Only relevant when type equals \code{"type1error"}.
##' @param color [character vector] a vector of colours to be used to color the lines.
##' Only relevant when type equals \code{"type1error"}.
##' @param keep.method [character vector] the methods names for which the type 1 error should be displayed.
##' Only relevant when type equals \code{"type1error"}.
##' @param ... [internal] Only used by the generic method.
##' 
##' @details Method names:
##' \itemize{
##' \item \code{p.Ztest}
##' \item \code{p.Satt}
##' \item \code{p.KR}
##' \item \code{p.robustZtest}
##' \item \code{p.robustSatt}
##' \item \code{p.robustKR}
##' }
##' @return An list containing:
##' \itemize{
##' \item plot: a ggplot object.
##' \item data: the dataset used to generate the ggplot object.
##' }
##' 


## * plotBias.calibrateType
##' @rdname autoplot_calibrateType1
##' @method autoplot calibrateType1
##' @export
autoplot.calibrateType1 <- function(object, type = "bias", plot = TRUE, color.threshold = "red",
                                    type.bias = "absolute",
                                    alpha = 0.05, nrow.legend = NULL, name2label = NULL, color = NULL, keep.method = NULL,
                                    ...){

    type <- match.arg(type, choices = c("bias","type1error"))

    ## ** display bias
    if(type == "bias"){
        ## *** data
        type.bias <- match.arg(type.bias, choices = c("absolute","relative"))

        if(type.bias == "absolute"){
            object$bias$bias.ML <- object$bias$estimate.truth-object$bias$estimate.ML
            object$bias$bias.MLcorrect <- object$bias$estimate.truth-object$bias$estimate.MLcorrect
        }else if(type.bias == "relative"){
            object$bias$bias.ML <- object$bias$estiamte.truth-object$bias$estiamte.ML
            object$bias$bias.MLcorrect <- object$bias$estimate.truth-object$bias$estimate.MLcorrect
        }

        df1 <- cbind(object$bias[,c("n","rep","name","type","bias.ML")], corrected = FALSE)
        names(df1)[5] <- "bias"
        df2 <- cbind(object$bias[,c("n","rep","name","type","bias.MLcorrect")], corrected = TRUE)
        names(df2)[5] <- "bias"

        df.gg <- rbind(df1,df2)

        ## *** display
        gg <- ggplot2::ggplot() + ggplot2::geom_boxplot(data = df.gg, aes_string(x = "type", y = "bias")) 
        gg <- gg + ggplot2::facet_wrap(corrected ~ n, labeller = ggplot2::label_both)
        df.line <- data.frame(x = c(-Inf,Inf), y = 0)
        gg <- gg + ggplot2::geom_line(data = df.line, aes_string(x = "x", y = "y"), color = color.threshold)
        
    }

    ## ** display type 1 error
    if(type == "type1error"){
        df.gg <- summary(object, alpha = alpha, type = type, display = FALSE)
        
        ## *** display
        if(is.null(keep.method)){
            keep.method <- as.character(unique(df.gg$method))
        }
        if(is.null(name2label)){
            name2label <- stats::setNames(unique(paste0(df.gg$statistic,", ",df.gg$correction)),unique(df.gg$method))
        }
        if(is.null(color)){
            ## from ggthemes::colorblind_pal()(8)
            color <- c("#000000","#E69F00","#56B4E9","#009E73","#F0E442","#0072B2","#D55E00","#CC79A7")
        }

        label.method <- name2label[keep.method]
        n.method <- length(keep.method)

       y.range <- range(c(alpha,df.gg$type1error))
        gg <- ggplot2::ggplot(df.gg, aes_string(x = "n", y = "type1error", group = "method", color = "method", shape = "method"))
        gg <- gg + ggplot2::geom_point(size = 3) + ggplot2::geom_line(size = 2)
        gg <- gg + ggplot2::facet_grid(~link, labeller = ggplot2::label_parsed)
        gg <- gg + ggplot2::geom_abline(intercept = alpha, slope = 0, color = color.threshold)
        gg <- gg + ggplot2::xlab("sample size")
        gg <- gg + ggplot2::ylab("type 1 error rate")
        gg <- gg + ggplot2::theme(legend.position = "bottom")
        gg <- gg + ggplot2::coord_cartesian(ylim = y.range)

        if(!is.null(nrow.legend)){
            gg <- gg + ggplot2::guides(color = ggplot2::guide_legend(nrow=nrow.legend,byrow=TRUE))
        }
        
        gg <- gg + ggplot2::scale_color_manual("",
                                               breaks = keep.method,
                                               label = label.method,
                                               values = color[1:n.method])
        gg <- gg + ggplot2::scale_shape_manual("",
                                               breaks = keep.method,
                                               label = label.method,
                                               values = seq(15,by=1,length=n.method))
        
         
    }
    
    ## ** export
    if(plot){
        print(gg)
    }
    rownames(df.gg) <- NULL
    return(invisible(list(plot = gg,
                          data = df.gg)))
}


######################################################################
### autoplot.calibrateType1.R ends here

### vcov2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar 12 2018 (16:38) 
## Version: 
## Last-Updated: jul 31 2020 (10:44) 
##           By: Brice Ozenne
##     Update #: 13
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - vcov2
#' @title  Extract the Variance Covariance Matrix of the Model Parameters
#' @description  Extract the variance covariance matrix of the model parameters from a Gaussian linear model.
#' @name vcov2
#'
#' @param object a linear model or a latent variable model
#' @param param [optional] the fitted parameters.
#' @param data [optional] the data set.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias? Only relevant if the \code{sCorrect} function has not yet be applied to the object.
#' @param ... arguments to be passed to \code{sCorrect}.
#' 
#' @details If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the influence function.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#'
#' @return A matrix.
#' 
#' @examples
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' ## linear model
#' e.lm <- lm(formula.lvm,data=d)
#' vcov.tempo <- vcov2(e.lm, bias.correct = TRUE)
#' vcov.tempo[rownames(vcov(e.lm)),colnames(vcov(e.lm))]/vcov(e.lm)
#'
#' ## latent variable model
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' vcov.tempo <- vcov2(e.lvm, bias.correct = FALSE)
#' vcov.tempo/vcov(e.lvm)
#'
#' @concept small sample inference
#' @export
`vcov2` <-
  function(object, ...) UseMethod("vcov2")

## * vcov2.lm
#' @rdname vcov2
#' @export
vcov2.lm <- function(object, param = NULL, data = NULL, bias.correct = TRUE, ...){

    sCorrect(object, param = param, data = data, df = FALSE, ...) <- bias.correct

    ### ** export
    return(object$sCorrect$vcov.param)
}

## * vcov2.gls
#' @rdname vcov2
#' @export
vcov2.gls <- vcov2.lm

## * vcov2.lme
#' @rdname vcov2
#' @export
vcov2.lme <- vcov2.lm

## * vcov2.lvmfit
#' @rdname vcov2
#' @export
vcov2.lvmfit <- vcov2.lm

## * vcov2.lm2
#' @rdname vcov2
#' @export
vcov2.lm2 <- function(object, param = NULL, data = NULL, ...){

    ### ** compute the score
    if(!is.null(param) || !is.null(data)){
        args <- object$sCorrect$args
        args$df <- FALSE
        object$sCorrect <- do.call(sCorrect,
                                   args = c(list(object, param = param, data = data),
                                            args))
    }

    ### ** export
    return(object$sCorrect$vcov.param)

}

## * vcov2.gls
#' @rdname vcov2
#' @export
vcov2.gls2 <- vcov2.lm2

## * vcov2.lme
#' @rdname vcov2
#' @export
vcov2.lme2 <- vcov2.lm2

## * vcov2.lvmfit
#' @rdname vcov2
#' @export
vcov2.lvmfit2 <- vcov2.lm2


##----------------------------------------------------------------------
### vcov2.R ends here

### conditionalMoment.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 27 2017 (16:59) 
## Version: 
## last-updated: jan 23 2024 (10:28) 
##           By: Brice Ozenne
##     Update #: 1955
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * moments2 - documentation
#' @title Compute Key Quantities of a Latent Variable Model
#' @description Compute conditional mean, conditional variance, their first and second derivative regarding model parameters, as well as various derivatives of the log-likelihood.
#' @name moments2
#' 
#' @param object a latent variable model.
#' @param data [data.frame] dataset if different from the one used to fit the model.
#' @param param [numeric vector] value of the model parameters if different from the estimated ones.
#' @param initialize [logical] Pre-compute quantities dependent on the data but not on the parameters values.
#' @param usefit [logical] Compute key quantities based on the parameter values.
#' @param update.dmoment [logical] should the first derivative of the moments be computed/updated?
#' @param update.d2moment [logical] should the second derivative of the the moments be computed/updated?
#' @param score [logical] should the score be output?
#' @param information [logical] should the expected information be output?
#' @param hessian [logical] should the hessian be output?
#' @param vcov [logical] should the variance-covariance matrix based on the expected information be output?
#' @param dVcov [logical] should the derivative of the variance-covariance matrix be output?
#' @param dVcov.robust [logical]  should the derivative of the robust variance-covariance matrix be output?
#' @param Psi [matrix]  Average first order bias in the residual variance. Only necessary for computing adjusted residuals.
#' 
#' @details For lvmfit objects, there are two levels of pre-computation:
#' \itemize{
#' \item a basic one that do no involve the model coefficient (\code{conditionalMoment.lvm}).
#' \item an advanced one that require the model coefficients (\code{conditionalMoment.lvmfit}). 
#' }
#' 
#' @examples
#' m <- lvm(Y1~eta,Y2~eta,Y3~eta)
#' latent(m) <- ~eta
#'
#' d <- lava::sim(m,1e2)
#' e <- estimate(m, d)
#'
#' ## basic pre-computation
#' res1 <- moments2(e, data = d, initialize = TRUE, usefit = FALSE,
#'                 score = TRUE, information = TRUE, hessian = TRUE, vcov = TRUE,
#'                 dVcov = TRUE, dVcov.robust = TRUE, residuals = TRUE, leverage = FALSE,
#'                 derivative = "analytic")
#' res1$skeleton$param$Sigma
#' 
#' ## full pre-computation
#' res2 <- moments2(e, param = coef(e), data = d, initialize = TRUE, usefit = TRUE,
#'                 score = TRUE, information = TRUE, hessian = TRUE, vcov = TRUE,
#'                 dVcov = TRUE, dVcov.robust = TRUE, residuals = TRUE, leverage = FALSE,
#'                 derivative = "analytic")
#' res2$moment$Omega
#'
#' @concept small sample inference
#' @concept derivative of the score equation
#' 
#' @keywords internal
#' @export
`moments2` <-
    function(object, param, data, weights, Omega, Psi,
             initialize, usefit,
             update.dmoment, update.d2moment, score, information, hessian, vcov, dVcov, dVcov.robust, residuals, leverage, derivative) UseMethod("moments2")

## * moments2.lvm
#' @rdname moments2
#' @export
moments2.lvm <- function(object, param = NULL, data = NULL, weights = NULL, Omega = NULL, Psi = NULL,
                         initialize = TRUE, usefit = TRUE,
                         update.dmoment = TRUE, update.d2moment = TRUE, score = TRUE, information = TRUE, hessian = TRUE, vcov = TRUE, dVcov = TRUE, dVcov.robust = TRUE, residuals = TRUE, leverage = TRUE,
                         derivative = "analytic"){
    if(lava.options()$debug){cat("moments2 \n")}
    
    ## ** sanity checks
    if(initialize == FALSE && is.null(object$sCorrect)){
        stop("Initialization of the moments missing. \n",
             "Consider setting the argument \'initialize\' to TRUE \n")
    }
    if(!missing(derivative)){
        derivative <- match.arg(derivative, choices = c("analytic","numeric"))
    }
    
    ## ** initialize
    if(score || information || hessian || vcov || dVcov || dVcov.robust){
        first.order <- TRUE
    }else{
        first.order <- FALSE
    }
    if(hessian || dVcov || dVcov.robust){
        second.order <- TRUE
    }else{
        second.order <- FALSE
    }

    ## NOTE: the estimation of the leverage depends on the information matrix and vice versa.
    ##       moments2 is a single run function called by estimate2 which take care of iterating until reaching a stable point
    previous.vcov.param <- object$sCorrect$vcov.param

    ## ** skeleton
    if(initialize){
        out <- list()

        ## *** get information from object
        out$endogenous <- lava::endogenous(object, format = "wide")
        out$latent <- lava::latent(object)

        if(is.null(data)){
            out$data <- extractData(object, design.matrix = FALSE, as.data.frame = TRUE,
                                    envir = parent.env(environment()), rm.na = TRUE)
        }else{
            out$data <- as.data.frame(data)
        }
        out$X <- as.matrix(out$data[,lava::manifest(object),drop=FALSE])
        out$cluster <- .getGroups2(object, data = out$data, endogenous = out$endogenous)

        reserved.name <- c("XXvalueXX","XXendogenousXX","XXendogenousXX.index","XXclusterXX")
        if(any(colnames(out$X) %in% reserved.name)){
            stop("\"",paste(reserved.name[colnames(out$X) %in% reserved.name], collapse="\" \""),"\" should not correspond to variable names \n",
                 "It is used internally for data manipulation \n")
        }

        ## *** reshape dataset (convert to long format)
        X.latent <- matrix(NA, nrow = NROW(out$X), ncol = length(out$latent),
                           dimnames = list(NULL, out$latent))

        X.long <- reshape2::melt(data.frame(XXclusterXX = out$cluster$name.cluster, out$X, X.latent),
                                 id.vars = "XXclusterXX",
                                 measure.vars = c(out$endogenous,out$latent),
                                 variable.name = "XXendogenousXX",
                                 value.name = "XXvalueXX")
        X.long <- cbind(X.long,out$X[X.long$XXclusterXX,,drop=FALSE]) ## add all variable (endo and exo) in case some endo are regressors

        X.wide <- out$X[,out$endogenous,drop=FALSE]

        out$old2new.order <- X.long[,c("XXclusterXX","XXendogenousXX","XXvalueXX")]
        names(out$old2new.order) <- c("XXclusterXX.old","XXendogenousXX.old","XXvalueXX.old")
        X.long <- X.long[order(X.long$XXclusterXX,X.long$XXendogenousXX),]
        out$old2new.order$XXclusterXX.new <- out$old2new.order$XXclusterXX
        out$old2new.order$XXendogenousXX.new <- out$old2new.order$XXendogenousXX
        out$old2new.order$XXvalueXX.new <- out$old2new.order$XXvalueXX
                
        ## *** identify missing pattern
        pattern <- X.wide
        pattern[!is.na(pattern)] <- 1
        pattern[is.na(pattern)] <- 0

        if(!is.null(object$call$missing) && object$call$missing==FALSE){
            unique.pattern <- unique(pattern[rowSums(pattern==0)==0,,drop=FALSE])
            if(length(unique.pattern)==0){
                stop("All clusters contain at least one missing value - cannot perform complete case analysis. \n",
                     "Consider setting the argument \'missing\' to TRUE. \n")
            }
            XXclusterXX.NA <- unique(X.long$XXclusterXX[rowSums(is.na(X.long[,out$endogenous,drop=FALSE]))>0])
            X.long[X.long$XXclusterXX %in% XXclusterXX.NA,c("XXvalueXX",out$endogenous)] <- NA ## add missing value to the other observation of the same cluster
        }else{
            unique.pattern <- unique(pattern[rowSums(pattern==1)>0,,drop=FALSE])
        }
        
        name.pattern <- apply(unique.pattern, MARGIN = 1, FUN = paste0, collapse = "")
        rownames(unique.pattern) <- name.pattern

        out$missing$pattern <- tapply(1:out$cluster$n.cluster,apply(pattern, MARGIN = 1, FUN = paste0, collapse=""),list)[name.pattern]
        out$missing$unique.pattern <- unique.pattern
        out$missing$name.pattern <- name.pattern

        ## *** initialize conditional moments
        out$skeleton <- skeleton(object, X = X.long,
                                 endogenous = out$endogenous, latent = out$latent,
                                 n.cluster = out$cluster$n.cluster,
                                 index.Omega = out$cluster$index.Omega)

        ## *** initialize partial derivatives of the conditional moments
        out$skeleton <- skeletonDtheta(out$skeleton,
                                       X = X.long,
                                       endogenous = out$endogenous, latent = out$latent,
                                       missing.pattern = out$missing$pattern,
                                       unique.pattern = out$missing$unique.pattern,
                                       name.pattern = out$missing$name.pattern,
                                       n.cluster = out$cluster$n.cluster,
                                       index.Omega = out$cluster$index.Omega)

        ## *** initialize second order partial derivatives of the conditional moments
        ## GS <- skeletonDtheta2(out$skeleton)
        out$skeleton <- skeletonDtheta2(out$skeleton)

        ## *** weights
        if(!is.null(object$weights)){
            out$weights <- object$weights[,1]
        }

    }else{
        ## subset to remove existing results
        rm.name <- c("moment","dmoment","d2moment","score","vcov.param","information","hessian","dInformation","dVcov.param","dRvcov.param","leverage","residuals")
        out <- object$sCorrect[setdiff(names(object$sCorrect),rm.name)]
    }

    ## ** update according to the value of the model coefficients
    if(usefit){
        if(is.null(param)){
            param.tempo <- stats::coef(object, type = 2, labels = 1)
            out$param <- stats::setNames(param.tempo[,"Estimate"],rownames(param.tempo))[out$skeleton$Uparam]
            ## out$name.param <- out$name.param[names(stats::coef(object))]
        }else{
            if(all(names(param) %in% out$skeleton$Uparam)){ ## using user-defined names
                ## e.g. mu1 mu2 Y1~X1 Y2~X1 sigma
                out$param[names(param)] <- param
            }else if(all(names(param) %in% names(out$skeleton$originalLink2param))){ ## using original link
                ## e.g. Y1 Y2 Y1~X1 Y2~X1 Y1~~Y1
                out$param[out$skeleton$Uparam[match(names(param),names(out$skeleton$originalLink2param))]] <- param
            }else{
                stop("Could not find model parameter(s) corresponding to the name(s): \"",paste(setdiff(names(param),c(out$skeleton$Uparam,names(out$skeleton$Uparam))), collapse="\" \""),"\" \n")
            }
        }
        if(is.null(weights)){
            weights <- object$weights[,1]
        }
        
        out$name.param <- stats::setNames(out$skeleton$type[!is.na(out$skeleton$type$lava),"param"],
                                   out$skeleton$type[!is.na(out$skeleton$type$lava),"originalLink"])

        ## *** conditional moments
        out$moment <- updateMoment(skeleton = out$skeleton$param,
                                   value = out$skeleton$value,
                                   toUpdate = out$skeleton$toUpdate.moment,
                                   param = out$param, Omega = Omega,
                                   name.pattern = out$missing$name.pattern,
                                   unique.pattern = out$missing$unique.pattern,
                                   endogenous = out$endogenous,
                                   latent = out$latent,
                                   n.cluster = out$cluster$n.cluster)

        ## *** first order derivatives
        if(update.dmoment && first.order){
            out$dmoment <- updateDMoment(moment = out$moment,
                                         skeleton = out$skeleton,
                                         param = out$param)
        }

        
        ## *** second order derivatives
        if(update.d2moment && second.order){
            out$d2moment <- updateD2Moment(moment = out$moment,
                                           skeleton = out$skeleton,
                                           param = out$param)
        }
        

        ## *** update residuals
        if(residuals || score || information || vcov || hessian || leverage){
            if(!is.null(attr(Omega,"Omega.residuals"))){
                OOmega <- attr(Omega,"Omega.residuals")
            }else{
                OOmega <- out$moment$Omega
            }

            out$residuals <- .adjustResiduals(epsilon = out$skeleton$param$endogenous - out$moment$mu,
                                              Omega = OOmega, Psi = Psi, ## Note: if Psi is null returns epsilon i.e. no adjustment
                                              name.pattern = out$missing$name.pattern, missing.pattern = out$missing$pattern, unique.pattern = out$missing$unique.pattern,
                                              endogenous = out$endogenous, n.cluster = out$cluster$n.cluster)

        }
        ## mean(out$residuals^2)
        ## out$moment$Omega
        
        ## *** score
        if(score){
            out$score <- .score2(dmu = out$dmoment$dmu,
                                 dOmega = out$dmoment$dOmega,                    
                                 epsilon = out$residuals,
                                 OmegaM1 = out$moment$OmegaM1.missing.pattern,
                                 missing.pattern = out$missing$pattern,
                                 unique.pattern = out$missing$unique.pattern,
                                 name.pattern = out$missing$name.pattern,
                                 name.param = out$skeleton$Uparam,
                                 name.meanparam = out$skeleton$Uparam.mean,
                                 name.varparam = out$skeleton$Uparam.var,
                                 n.cluster = out$cluster$n.cluster,
                                 weights = weights)
        }

        ## *** leverage
        if((leverage || information || vcov) && !is.null(previous.vcov.param)){
            if(!is.null(attr(Omega,"Omega.leverage"))){
                OOmega <- attr(Omega,"Omega.leverage")
            }else{
                OOmega <- out$moment$Omega
            }
            if(!is.null(attr(Omega,"dOmega.leverage"))){
                ddOOmega <- attr(Omega,"dOmega.leverage")
            }else{
                ddOOmega <- out$dmoment$dOmega
            }

            out$leverage <- .leverage2(Omega = OOmega, 
                                       epsilon = out$residuals,
                                       dmu = aperm(abind::abind(out$dmoment$dmu, along = 3), perm = c(3,2,1)),
                                       dOmega = ddOOmega,
                                       vcov.param = previous.vcov.param,
                                       name.pattern = out$missing$name.pattern,
                                       missing.pattern = out$missing$pattern,
                                       unique.pattern = out$missing$unique.pattern,
                                       endogenous = out$endogenous,
                                       n.endogenous = length(out$endogenous),
                                       param = out$skeleton$Uparam,
                                       param.mean = out$skeleton$Uparam.mean,
                                       param.var = out$skeleton$Uparam.var,
                                       n.cluster = out$cluster$n.cluster)
        }

        ## *** information matrix
        if(information || vcov){
            if(!is.null(attr(Omega,"dOmega.leverage"))){
                ddOOmega <- attr(Omega,"dOmega.leverage")
            }else{
                ddOOmega <- out$dmoment$dOmega
            }

            out$information <- .information2(dmu = out$dmoment$dmu,
                                             dOmega = ddOOmega,##out$dmoment$dOmega,
                                             OmegaM1 = out$moment$OmegaM1.missing.pattern,
                                             missing.pattern = out$missing$pattern,
                                             unique.pattern = out$missing$unique.pattern,
                                             name.pattern = out$missing$name.pattern,
                                             grid.mean = out$skeleton$grid.dmoment$mean, 
                                             grid.var = out$skeleton$grid.dmoment$var, 
                                             name.param = out$skeleton$Uparam,
                                             leverage = out$leverage,
                                             n.cluster = out$cluster$n.cluster,
                                             weights = weights)
        }

        if(vcov){
            out$vcov.param  <- .info2vcov(out$information, attr.info = FALSE)
        }
    
        ## *** hessian
        if(hessian || (derivative == "analytic") && dVcov.robust){
            out$hessian <- .hessian2(dmu = out$dmoment$dmu,
                                     dOmega = out$dmoment$dOmega,
                                     d2mu = out$d2moment$d2mu,
                                     d2Omega = out$d2moment$d2Omega,
                                     epsilon = out$residuals,                                     
                                     OmegaM1 = out$moment$OmegaM1.missing.pattern,
                                     missing.pattern = out$missing$pattern,
                                     unique.pattern = out$missing$unique.pattern,
                                     name.pattern = out$missing$name.pattern,
                                     grid.mean = out$skeleton$grid.dmoment$mean, 
                                     grid.var = out$skeleton$grid.dmoment$var, 
                                     grid.hybrid = out$skeleton$grid.dmoment$hybrid, 
                                     name.param = out$skeleton$Uparam,
                                     leverage = out$leverage,
                                     n.cluster = out$cluster$n.cluster,
                                     weights = weights)
        }

        ## *** dVcov.param (model based variance, analytic)
        if((dVcov || dVcov.robust) && (derivative == "analytic")){
            out$dInformation <- .dInformation2(dmu = out$dmoment$dmu,
                                               dOmega = out$dmoment$dOmega,
                                               d2mu = out$d2moment$d2mu,
                                               d2Omega = out$d2moment$d2Omega,
                                               OmegaM1 = out$moment$OmegaM1.missing.pattern,
                                               missing.pattern = out$missing$pattern,
                                               unique.pattern = out$missing$unique.pattern,
                                               name.pattern = out$missing$name.pattern,
                                               grid.3varD1 = out$skeleton$grid.3varD1,
                                               grid.2meanD1.1varD1 = out$skeleton$grid.2meanD1.1varD1,
                                               grid.2meanD2.1meanD1 = out$skeleton$grid.2meanD2.1meanD1,
                                               grid.2varD2.1varD1 = out$skeleton$grid.2varD2.1varD1,
                                               name.param = out$skeleton$Uparam,
                                               leverage = out$leverage,
                                               n.cluster = out$cluster$n.cluster,
                                               weights = weights)

            ## delta method
            out$dVcov.param <- .dVcov.param(vcov.param = out$vcov.param,
                                            dInformation = out$dInformation,
                                            n.param = length(out$skeleton$Uparam),
                                            name.param = out$skeleton$Uparam)
        }

        ## *** dRvcov.param  (robust variance, analytic)
        if(dVcov.robust && (derivative == "analytic")){

        out$dRvcov.param <- .dRvcov.param(score = out$score,
                                          hessian = out$hessian,
                                          vcov.param = out$vcov.param,
                                          dVcov.param = out$dVcov.param,
                                          n.param = length(out$skeleton$Uparam),
                                          name.param = out$skeleton$Uparam)

    }

    ## *** dVcov.param and dRvcov.param (numeric derivatives)
    if((dVcov || dVcov.robust) && derivative == "numeric"){
        test.package <- try(requireNamespace("numDeriv"), silent = TRUE)
        if(inherits(test.package,"try-error")){
            stop("There is no package \'numDeriv\' \n",
                 "This package is necessary when argument \'numeric.derivative\' is TRUE \n")
        }

        ## range(out$score - .warper.numDev(value = out$param, object = object, type = "score"))
        ## range(out$hessian - .warper.numDev(value = out$param, object = object, type = "hessian"))
        ## range(out$vcov.param - .warper.numDev(value = out$param, object = object, type = "vcov.model"))

        param <- out$param
        name.param <- names(out$param)
        n.param <- length(param)
        n.cluster <- out$cluster$n.cluster
        object2 <- object
        object2$sCorrect <- out

        ## *** hessian
        ## print(range(.warper.numDev(param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "score")-out$score))
        num.hessian <- numDeriv::jacobian(.warper.numDev, x = param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "score", method = "Richardson")

        out$hessian <- aperm(array(num.hessian, dim = c(n.cluster,n.param,n.param),
                                               dimnames = list(NULL, name.param, name.param)), perm = 3:1)
        
        ## *** dInformation
        ## print(range(.warper.numDev(param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "information")-out$information))
        num.information <- numDeriv::jacobian(.warper.numDev, x = param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "information", method = "Richardson")

        out$dInformation <- array(num.information, dim = c(n.param,n.param,n.param),
                                              dimnames = list(name.param, name.param, name.param))

        ## *** dVcov.param
        ## print(range(.warper.numDev(param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "vcov")-out$vcov.param))
        num.dVcov.param <- numDeriv::jacobian(.warper.numDev, x = param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "vcov", method = "Richardson")

        out$dVcov.param <- array(num.dVcov.param, dim = c(n.param,n.param,n.param),
                                             dimnames = list(name.param, name.param, name.param))


        ## *** dRvcov.param
        num.dRvcov.param <- numDeriv::jacobian(.warper.numDev, x = param, object = object2, weights = weights, Omega = Omega, Psi = Psi, type = "vcov.robust", method = "Richardson")

        out$dRvcov.param <- array(num.dRvcov.param, dim = c(n.param,n.param,n.param),
                                              dimnames = list(name.param, name.param, name.param))
    }        
    
        
    }

    ## ** export
    return(out)
}

## * moments2.lvmfit
#' @rdname moments2
#' @export
moments2.lvmfit <- moments2.lvm

## * .wraper.numDev (helper)
.warper.numDev <- function(value, object, type, weights, Psi = NULL, Omega = NULL){ # x <- p.obj
    ## CANNOT DO DIRECTLY VIA moments2
    ## (because Omega should not be fixed in updateMoment but it should be fixed (as well as Psi) when updating the residuals)
    
    out <- object$sCorrect
    out$moment <- updateMoment(skeleton = out$skeleton$param,
                               value = out$skeleton$value,
                               toUpdate = out$skeleton$toUpdate.moment,
                               param = value, Omega = NULL,
                               name.pattern = out$missing$name.pattern,
                               unique.pattern = out$missing$unique.pattern,
                               endogenous = out$endogenous,
                               latent = out$latent,
                               n.cluster = out$cluster$n.cluster)

    out$dmoment <- updateDMoment(moment = out$moment,
                                 skeleton = out$skeleton,
                                 param = out$param)

    if(type %in% c("score","vcov.robust")){
        out$residuals <- .adjustResiduals(epsilon = out$skeleton$param$endogenous - out$moment$mu,
                                          Omega = Omega, Psi = Psi, ## FIXED
                                          name.pattern = out$missing$name.pattern, missing.pattern = out$missing$pattern, unique.pattern = out$missing$unique.pattern,
                                          endogenous = out$endogenous, n.cluster = out$cluster$n.cluster)

        out$score <- .score2(dmu = out$dmoment$dmu,
                             dOmega = out$dmoment$dOmega,                    
                             epsilon = out$residuals,
                             OmegaM1 = out$moment$OmegaM1.missing.pattern,
                             missing.pattern = out$missing$pattern,
                             unique.pattern = out$missing$unique.pattern,
                             name.pattern = out$missing$name.pattern,
                             name.param = out$skeleton$Uparam,
                             name.meanparam = out$skeleton$Uparam.mean,
                             name.varparam = out$skeleton$Uparam.var,
                             n.cluster = out$cluster$n.cluster,
                             weights = weights)
    }

    if(type %in% c("information","vcov","vcov.robust")){
        out$information <- .information2(dmu = out$dmoment$dmu,
                                         dOmega = out$dmoment$dOmega,
                                         OmegaM1 = out$moment$OmegaM1.missing.pattern,
                                         missing.pattern = out$missing$pattern,
                                         unique.pattern = out$missing$unique.pattern,
                                         name.pattern = out$missing$name.pattern,
                                         grid.mean = out$skeleton$grid.dmoment$mean, 
                                         grid.var = out$skeleton$grid.dmoment$var, 
                                         name.param = out$skeleton$Uparam,
                                         leverage = out$leverage, ## FIXED
                                         n.cluster = out$cluster$n.cluster,
                                         weights = weights)
    }

    if(type %in% c("vcov","vcov.robust")){
        out$vcov  <- .info2vcov(out$information, attr.info = FALSE)
    }

    if(type %in% c("vcov.robust")){
        out$vcov.robust  <- out$vcov %*% crossprod(out$score) %*% out$vcov
    }


    return(out[[type]])
}


## * .info2vcov (helper)
#' @title Inverse the Information Matrix
#' @description Compute the inverse of the information matrix.
#' @name vcov2-internal
#'
#' @param information [matrix] information matrix to be inverted.
#' @param attr.info [logical] should the information matrix be returned as an attribute?
#' 
#' @keywords internal
.info2vcov <- function(information, attr.info = FALSE){
    vcov <- try(chol2inv(chol(information)), silent = TRUE)
    if(inherits(vcov, "try-error")){
        vcov <- try(solve(information), silent = TRUE)
        if(inherits(vcov, "try-error")){ ## try by block
            cat("Singular information matrix: try to inverse it by block \n")
            information.N0 <- abs(information)>1e-10
            remaining.var <- colnames(information)
            vcov <- matrix(0, nrow = NROW(information), ncol = NCOL(information),
                           dimnames = dimnames(information))
            while(length(remaining.var)>0){
                current.set <- remaining.var[1]
                new.set <- unique(unlist(apply(information.N0[current.set,,drop=FALSE],1,function(iRow){list(names(iRow[iRow==1]))})))
                while(length(current.set)<length(new.set)){
                    current.set <- new.set
                    new.set <- unique(unlist(apply(information.N0[current.set,,drop=FALSE],1,function(iRow){list(names(iRow[iRow==1]))})))
                }
                if(length(new.set)>0){
                    iTry <- try(solve(information[current.set,current.set]), silent = TRUE)
                    if(inherits(iTry,"try-error")){
                        vcov[current.set,current.set] <- NA
                    }else{
                        vcov[current.set,current.set] <- iTry
                    }
                }
                remaining.var <- setdiff(remaining.var, current.set)
            }
        }
    }
    if(attr.info){
        attr(vcov,"information") <- information
    }
    if(!inherits(vcov, "try-error")){
        dimnames(vcov) <- dimnames(information)
    }
    return(vcov)
}

### InTriGauss.R --- 
##----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 14 2017 (11:49) 
## Version: 
## last-updated: maj 23 2018 (09:42) 
##           By: Brice Ozenne
##     Update #: 495
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:



## * Documentation - intDensTri
##' @title Integrate a Gaussian/Student Density over a Triangle
##' @name intDensTri
##' @description Consider a univariate random variable X,
##' two multivariate random variables Y and Z,
##' and t1 and t2 two real numbers.
##' This function can compute either
##' P[|X|>t1,|X]>|Y1|,...,|X]>|Yp|] if zmin is not specified,
##' P[|Z1|<t2,...,|Zq|<t2,|X|>t1,|X]>|Y1|,...,|X]>|Yp|] if zmin is specified.
##' 
##' @param mu [numeric vector] the expectation of the joint distribution.
##' @param Sigma [matrix] the variance-covariance of the joint distribution.
##' @param df [integer > 0] the degree of freedom of the joint Student's t distribution.
##' Only used when \code{distribution="pvmt"}.
##' @param n [integer > 0] number of points for the numerical integration.
##' @param x.min [numeric] the minimum value along the x axis.
##' @param z.max [numeric vector, optional] the maximum value along the z axis.
##' Define the dimension of Z.
##' @param type [character] the type of mesh to be used.
##' Can be \code{\"raw\"}, \code{\"double\"}, or \code{\"fine\"}.
##' @param prune [integer >0] number of standard deviations after which the domain ends along the x axis. 
##' @param proba.min [numeric 0-1] the probability used to find the maximum value along the x axis.
##' Only used if \code{prune} is not specified.
##' @param distribution [character] type of joint distribution.
##' Can be \code{"pmvnorm"} (normal distribution) or \code{"pvmt"} (Student's t distribution)
##' @details
##' Argument \code{type}: \itemize{
##' \item \code{\"raw\"}: mesh with points inside the domain
##' \item \code{\"double\"}: mesh with points outside the domain
##' \item \code{\"fine\"}: mesh with points inside the domain plus additional rectangles trying to fill the missing domain.
##' }
##'
##' Argument \code{Sigma} and \code{mu}:
##' define the mean and variance-covariance of the random variables X, Y, Z
##' (in this order). The length of the argument \code{z.max} is used to define the dimension of Z.
##' The dimension of X is always 1.
##'
##' @return A numeric.
##' 
##' @examples
##' library(mvtnorm)
##' 
##' p <- 2
##' Sigma <- diag(p)
##' mu <- rep(0, p)
##' 
##' ## bivariate normal distribution
##' z2 <- qmvt(0.975, mean = mu, sigma = Sigma, df = 1e3)$quantile
##'
##' # compute integral
##' intDensTri(mu = mu, Sigma = Sigma, n=5, x.min=0, type = "fine")$value-1/2
##' intDensTri(mu = mu, Sigma = Sigma, n=30, x.min=0, type = "raw")$value-1/2
##' intDensTri(mu = mu, Sigma = Sigma, n=50, x.min=0, type = "raw")$value-1/2
##'
##' intDensTri(mu = mu, Sigma = Sigma, df = 5, n=5, x.min=0, distribution = "pmvt")$value-1/2
##' res <- intDensTri(mu = mu, Sigma = Sigma, df = 5, n=10, x.min=0, distribution = "pmvt")
##' res$value-1/2
##' ggplot2::autoplot(res)
##' 
##' ## trivariate normal distribution
##' \dontrun{
##' p <- 3
##' Sigma <- diag(p)
##' mu <- rep(0, p)
##'
##' res2 <- intDensTri(mu = mu, Sigma = Sigma, n=5, x.min = 0, z.max = 10)
##' ggplot2::autoplot(res2)
##' ggplot2::autoplot(res2, coord.plot = c("x","z1"))
##' res2
##' }
##' 
##' #### when the distribution is far from 0
##' \dontrun{
##' eq1 <- intDensTri(mu = c(10,0), Sigma = diag(1,2), 
##'                   x.min = 2, n=10)
##' eq1$value-1
##' ggplot2::autoplot(eq1)
##'
##' eq2 <- intDensTri(mu = c(10,0,0), Sigma = diag(1,3),
##'                   x.min=2, z.max = 10, type = "raw",
##'                   n=10)
##' ggplot2::autoplot(eq2, coord.plot = c("y1","z1"))
##' eq2$value-1
##'
##' ## more variables
##' p <- 5
##' Sigma <- diag(p)
##' mu <- rep(0, p)
##'
##' res2 <- intDensTri(mu = mu, Sigma = Sigma, n=5, x.min = 1, z.max = c(2,2))
##' res2$grid
##' }
##'
##' @concept post-selection inference

## * intDensTri
##' @rdname intDensTri
##' @export 
intDensTri <- function(mu, Sigma, df, n, x.min, z.max = NULL,
                       type = "double", proba.min = 1e-6, prune = NULL, distribution = "pmvnorm"){

     interior <- NULL ## [:for CRAN check] subset
    
    ## ** normalize arguments
    type <- match.arg(type, c("raw","fine","double"))
    
    p <- length(mu)
    d.z <- length(z.max)
    d.y <- p - d.z - 1
    z.max <- unique(z.max)
    
    if(is.null(prune)){
        if(distribution=="pmvnorm"){
            prune <- ceiling(abs(qmvnorm(proba.min, sigma = diag(1,d.y))$quantile))
        }else if(distribution=="pmvt"){
            prune <- ceiling(abs(mvtnorm::qmvt(proba.min, sigma = diag(1,d.y), df = df)$quantile))
        }else {
            stop("distribution must be either \"pmvnorm\" or \"pmvt\" \n")
        }
    }
    coordX.max <- (mu[1]+x.min) + prune * sqrt(diag(Sigma)[1])
    
    ls.args <- list(mu,Sigma)    
    if(distribution=="pmvt"){
        names(ls.args) <- c("delta","sigma")
        ls.args$df <- df
    }else{
        names(ls.args) <- c("mean","sigma")
    }

    ## ** create the grid of points to integrate over
    resGrid <- createGrid(n,
                          xmin = x.min, xmax = coordX.max, d.y = d.y, 
                          d.z = d.z, zmax = z.max, 
                          fine = (type=="fine"), double = FALSE
                          )
    grid <- resGrid$grid
    seqNames.min <- resGrid$seqNames.min
    seqNames.max <- resGrid$seqNames.max

    if(type=="double"){
        grid.double <- createGrid(n,
                                  xmin = x.min, xmax = coordX.max, d.y = d.y, 
                                  d.z = d.z, zmax = z.max, 
                                  fine = FALSE, double = TRUE
                                  )$grid
        grid <- rbind(cbind(grid,interior=TRUE),
                      cbind(grid.double,interior=FALSE))
    }

    ## ** integration
    total.area <- 0
    grid$area <- apply(grid, 1, function(x){
        do.call(distribution, args = c(lower = list(c(x[seqNames.min])),
                                       upper = list(c(x[seqNames.max])),
                                       ls.args))
    })


    if(type=="double"){
        grid.interior <- subset(grid, interior==TRUE,select = c("area", "weight"))
        area.interior <- sum(grid.interior$area * grid.interior$weight)
        grid.exterior <- subset(grid, interior==FALSE,select = c("area", "weight"))
        area.exterior <- sum(grid.exterior$area * grid.exterior$weight)
        total.area <- area.interior + (area.exterior-area.interior)/2
    }else{        
        total.area <- sum(grid$area * grid$weight)
    }

    ## ** export
    out <- list()
    out$value <- total.area    
    out$prune <- prune
    out$type <- type
    out$grid <- grid
    class(out) <- "intDensTri"
    return(out)
}

## * autoplot.intDensTri
#' @title 2D-display of the Domain Used to Compute the Integral
#' @description 2D-display of the domain used to compute the integral.
#'
#' @param object output of the function \code{intDensTri}.
#' @param coord.plot [character vector] the x and y coordinates. Can be \code{"x"}, \code{"y1"} to \code{"yd"}, \code{"z"} if \code{zmin} was specified when calling \code{intDensTri}.
#' @param plot [logical] should the plot be displayed?
#' @param ... [internal] Only used by the generic method.
#'
#' @return A \code{ggplot} object.
#' @seealso \code{\link{intDensTri}}
#' 
#' @method autoplot intDensTri
#' @export
autoplot.intDensTri <- function(object, coord.plot=c("x","y1"), plot = TRUE, ...){

    if(length(coord.plot) != 2){
        stop("coord.plot must have length 2 \n")
    }

    gg.data <- object$grid
    gg.data$index <- as.factor(gg.data$index)
    gg.data$weight <- as.factor(gg.data$weight)
    
    if("x" %in% coord.plot == FALSE){
        x.ref <- min(abs(unique(gg.data$x.min)))
        gg.data <- rbind(gg.data[gg.data$x.min == x.ref,,drop=FALSE],
                         gg.data[gg.data$x.max == -x.ref,,drop=FALSE])
    }

    if(object$type=="fine"){
        gg.grid <- ggplot2::ggplot(gg.data, aes_string(xmin = paste0(coord.plot[1],".min"), ymin = paste0(coord.plot[2],".min"),
                                                      xmax = paste0(coord.plot[1],".max"), ymax = paste0(coord.plot[2],".max"),
                                                      fill = "index", alpha = "weight"))
        gg.grid <- gg.grid + ggplot2::scale_alpha_manual(values = c(0.45,1))
    }else if(object$type=="raw"){
        gg.grid <- ggplot2::ggplot(gg.data, aes_string(xmin = paste0(coord.plot[1],".min"), ymin = paste0(coord.plot[2],".min"),
                                                      xmax = paste0(coord.plot[1],".max"), ymax = paste0(coord.plot[2],".max"),
                                                      fill = "index"))        
    }else if(object$type=="double"){
        gg.grid <- ggplot2::ggplot(gg.data, aes_string(xmin = paste0(coord.plot[1],".min"), ymin = paste0(coord.plot[2],".min"),
                                                      xmax = paste0(coord.plot[1],".max"), ymax = paste0(coord.plot[2],".max"),
                                                      fill = "index", alpha = "interior"))
        gg.grid <- gg.grid + ggplot2::scale_alpha_manual(values = c(0.4,1))
    }
    gg.grid <- gg.grid + ggplot2::geom_rect() + ggplot2::xlab(coord.plot[1]) + ggplot2::ylab(coord.plot[2])
    gg.grid <- gg.grid + ggplot2::geom_abline(slope = 1,color = "black") + ggplot2::geom_abline(slope = -1,color = "black")

    if(plot){
        print(gg.grid)
    }
    return(invisible(gg.grid))
}


## * print.intDensTri
#' @method print intDensTri
#' @export
print.intDensTri <- function(x, ...){
    interior <- NULL
    
    x.min <- min(abs(x$grid$x.min))
    x.max <- max(abs(x$grid$x.max))
    n.rectangle <- switch(x$type,
                          "double" = NROW(x$grid[interior==TRUE]),
                          NROW(x$grid)
                          )
    
    cat("integral=",x$value,"\n",
        "computed with ",n.rectangle," rectangles \n",
        "with x ranging from ",x.min," to ",x.max,"\n",
        sep = "")

    return(NULL)
}
#----------------------------------------------------------------------
### InTriGauss.R ends here

### test-intDensTri.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 31 2017 (16:32) 
## Version: 
## last-updated: Jan 12 2022 (16:33) 
##           By: Brice Ozenne
##     Update #: 16
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * header
##rm(list = ls(all.names = TRUE))
if(TRUE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

library(mvtnorm)
lava.options(symbols = c("~","~~"))

context("intDensTri")

## * tests
# {{{ around 0
test_that("Integrate standard gaussian density (2D)", {
    p <- 2
    Sigma <- diag(p)
    mu <- rep(0, p)

    for(n in c(5,10,20,50,100)){ # n <- 5
        res <- intDensTri(mu = mu, Sigma = Sigma, n=n, x.min=0)
        expect_equal(res$value,
                     1/2,
                     tol = 1e-6)
    }
})

test_that("Integrate standard gaussian density (3D)", {
    p <- 3
    Sigma <- diag(p)
    mu <- rep(0, p)

    for(n in c(5,10,20,50,100)){
        res <- intDensTri(mu = mu, Sigma = Sigma, n=n, x.min=0, z.max=10)
        expect_equal(res$value,
                     1/2,
                     tol = 1e-6)
    }
})
# }}}

# {{{ far from 0
test_that("Integrate standard gaussian density (2D)", {
    p <- 2
    Sigma <- diag(p)
    mu <- c(10,0)

    for(n in c(5,10,20,50,100)){
        res <- intDensTri(mu = mu, Sigma = Sigma, n=n, x.min=0)
        expect_equal(res$value,
                     1,
                     tol = 1e-6)
    }
})

test_that("Integrate standard gaussian density (3D)", {
    p <- 3
    Sigma <- diag(p)
    mu <- c(10,0,0)

    for(n in c(5,10,20,50,100)){
        res <- intDensTri(mu = mu, Sigma = Sigma, n=n, x.min=0, z.max=10)
        expect_equal(res$value,
                     1,
                     tol = 1e-6)
    }
})
# }}}


#----------------------------------------------------------------------
### test-intDensTri.R ends here

### conditionalMoment.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 27 2017 (16:59) 
## Version: 
## last-updated: feb  8 2019 (11:47) 
##           By: Brice Ozenne
##     Update #: 1139
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * conditionalMoment - documentation
#' @title Prepare the Computation of score2
#' @description Compute the conditional mean and variance,
#' and their first and second derivative regarding the model parameters.
#' @name conditionalMoment
#' 
#' @param object,x a latent variable model.
#' @param data [data.frame] data set.
#' @param formula [formula] two-sided linear formula.
#' @param param,p [numeric vector] the fitted coefficients.
#' @param attr.param [character vector] the type of each coefficient
#' (e.g. mean or variance coefficient).
#' @param ref.group [character vector] the levels of the variable defining the variance component in a generic covariance matrix.
#' @param second.order [logical] should the terms relative to the third derivative of the likelihood be be pre-computed?
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param n.cluster [integer >0] the number of i.i.d. observations.
#' @param n.endogenous [integer >0] the number of outcomes.
#' @param usefit,value [logical] If TRUE the coefficients estimated by the model are used to pre-compute quantities. Only for lvmfit objects.
#' @param name.endogenous [character vector, optional] name of the endogenous variables
#' @param name.latent [character vector, optional] name of the latent variables
#' @param ... [internal] only used by the generic method or by the <- methods.
#' 
#' @details For lvmfit objects, there are two levels of pre-computation:
#' \itemize{
#' \item a basic one that do no involve the model coefficient (\code{conditionalMoment.lvm}).
#' \item an advanced one that require the model coefficients (\code{conditionalMoment.lvmfit}). 
#' }
#' 
#' @examples
#' m <- lvm(Y1~eta,Y2~eta,Y3~eta)
#' latent(m) <- ~eta
#'
#' d <- lava::sim(m,1e2)
#' e <- estimate(m, d)
#'
#' ## basic pre-computation
#' res1 <- conditionalMoment(e, data = d,
#'                          first.order = FALSE, second.order = FALSE,
#'                          name.endogenous = endogenous(e),
#'                          name.latent = latent(e), usefit = FALSE)
#' res1$skeleton$Sigma
#' 
#' ## full pre-computation
#' res2 <- conditionalMoment(e, param = coef(e), data = d,
#'                          first.order = FALSE, second.order = FALSE,
#'                          name.endogenous = endogenous(e),
#'                          name.latent = latent(e), usefit = TRUE
#' )
#' res2$value$Sigma
#'
#' @concept small sample inference
#' @concept derivative of the score equation
#' 
#' @keywords internal
#' @export
`conditionalMoment` <-
  function(object, ...) UseMethod("conditionalMoment")


## * conditionalMoment.lm
#' @rdname conditionalMoment
#' @export
conditionalMoment.lm <- function(object, data, param,
                                 name.endogenous,
                                 first.order, second.order, ...){

    out <- list(param = param,
                name.3deriv = "sigma2")
    
    ## design matrix
    X <- model.matrix(formula(object), data)

    ## linear predictor
    out$mu <- X %*% param[colnames(X)]
    
    ## residuals variance
    out$Omega <- matrix(param["sigma2"], nrow = 1, ncol = 1,
                    dimnames = list(name.endogenous, name.endogenous))
    
    ## ** first order
    if(first.order){
        out$dmu <- lapply(1:NCOL(X), function(i){
            M <- X[,i,drop=FALSE]
            colnames(M) <- name.endogenous
            return(M)
        })
        names(out$dmu) <- colnames(X)
        out$dOmega = list(sigma2 = matrix(1))
    }

    ## ** second order
    if(second.order){
        out$d2mu <- NULL
        out$d2Omega <- NULL
    }

    ## ** export
    return(out)
}

## * conditionalMoment.gls
#' @rdname conditionalMoment
#' @export
conditionalMoment.gls <- function(object, data, formula, 
                                  param, attr.param, ref.group,
                                  first.order, second.order,
                                  index.Omega, vec.OmegaMat, cluster, n.cluster,
                                  name.endogenous, n.endogenous,
                                  ...){

    if(first.order == FALSE && second.order == TRUE){
        stop("Cannot pre-compute quantities for the second order derivatives ",
             "without those for the first order derivatives \n")
    }
### ** prepare

    ## *** coefficients
    name.varcoef <- attr.param$var.coef
    name.corcoef <- attr.param$cor.coef
    n.varcoef <- length(name.varcoef)
    n.corcoef <- length(name.corcoef)
    var.coef <- param[name.varcoef]
    cor.coef <- param[name.corcoef]
        
    class.var <- class(object$modelStruct$varStruct)
    class.cor <- class(object$modelStruct$corStruct)

    
    ## *** design matrix    
    X <- stats::model.matrix(formula, data)
    X <- X[,attr.param$mean.coef,drop=FALSE] ## drop unused columns (e.g. factor with 0 occurence)    
    attr(X,"assign") <- NULL
    attr(X,"contrasts") <- NULL
    
    ## *** variance terms
    if("NULL" %in% class.var == FALSE){
        name.otherVar <- setdiff(names(var.coef),"sigma2")
        factor.varcoef <- setNames(c(1,var.coef[name.otherVar]),
                                   attr(object$modelStruct$varStruct,"groupNames"))
        sigma2.base0 <- factor.varcoef[ref.group]        
    }else{
        name.otherVar <- NULL
        sigma2.base0 <- setNames(rep(1, n.endogenous), name.endogenous)
    }
    sigma2.base <- sigma2.base0 * var.coef["sigma2"]

    ## *** corelation terms
    if("NULL" %in% class.cor == FALSE){
        M.corcoef <- matrix("", n.endogenous, n.endogenous,
                            dimnames = list(name.endogenous,name.endogenous))
        M.corcoef[which(lower.tri(M.corcoef))] <- name.corcoef
        M.corcoef <- symmetrize(M.corcoef)

        index.lower.tri <- which(lower.tri(M.corcoef))
        indexArr.lower.tri <- which(lower.tri(M.corcoef), arr.ind = TRUE)

        Msigma2.base0 <- matrix(0, n.endogenous, n.endogenous,
                                dimnames = list(name.endogenous, name.endogenous))
        Msigma2.base0[index.lower.tri] <- apply(indexArr.lower.tri, 1, function(x){sqrt(prod(sigma2.base0[x]))})
        Msigma2.base0 <- symmetrize(Msigma2.base0)
    }else{
        M.corcoef <- NULL
        Msigma2.base0 <- NULL
        index.lower.tri <- NULL
        indexArr.lower.tri <- NULL
    }

    ## *** export
    out <- list(param = param,
                name.3deriv = c(name.varcoef,name.corcoef),
                skeleton = list(class.cor = class.cor,
                                class.var = class.var,
                                sigma2.base0 = sigma2.base0,
                                Msigma2.base0 = Msigma2.base0,
                                ref.group = ref.group,
                                n.endogenous = n.endogenous,
                                name.endogenous = name.endogenous,
                                M.corcoef = M.corcoef,
                                index.lower.tri = index.lower.tri,
                                indexArr.lower.tri = indexArr.lower.tri,
                                var.coef = var.coef,
                                name.varcoef = name.varcoef,
                                n.varcoef = n.varcoef,
                                name.otherVar = name.otherVar,
                                ref.group = ref.group,
                                cor.coef = cor.coef,
                                name.corcoef = name.corcoef,
                                n.corcoef = n.corcoef,
                                cluster  = cluster,
                                n.cluster = n.cluster))
    
### ** Reconstruct conditional mean
    ## transpose necessary because of the way index.OmegaMat was computed
    out$mu <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                     dimnames = list(NULL, name.endogenous))
    out$mu[vec.OmegaMat] <- X %*% param[colnames(X)]
    
    
### ** Reconstruct conditional variance covariance matrix
    out$Omega <- .getVarCov2(object,
                             param = param,
                             attr.param = attr.param,
                             name.endogenous = name.endogenous,
                             n.endogenous = n.endogenous,
                             ref.group = ref.group)
    
### ** first order
    if(first.order){
        outD1 <- skeletonDtheta(object,
                                class.cor = class.cor,
                                class.var = class.var,
                                X = X,  
                                sigma2.base0 = sigma2.base0,
                                Msigma2.base0 = Msigma2.base0,
                                M.corcoef = M.corcoef,
                                ref.group = ref.group,
                                name.endogenous = name.endogenous,
                                n.endogenous = n.endogenous,
                                index.lower.tri = index.lower.tri,
                                indexArr.lower.tri = indexArr.lower.tri,
                                cluster = cluster,
                                n.cluster = n.cluster,
                                var.coef = var.coef,
                                name.varcoef = name.varcoef,
                                name.otherVar = name.otherVar,
                                n.varcoef = n.varcoef,
                                cor.coef = cor.coef,
                                name.corcoef = name.corcoef,
                                n.corcoef = n.corcoef,
                                index.Omega = index.Omega,
                                update.mean = TRUE, update.variance = TRUE,
                                ...) ## ... to pass coef.rancoef
        out$dmu <- outD1$dmu
        out$dOmega <- outD1$dOmega
    }
    
### ** second order
    if(second.order){
        out$d2mu <- NULL
        out$d2Omega <- skeletonDtheta2(object,
                                       dOmega = out$dOmega,
                                       class.cor = class.cor,
                                       class.var = class.var,
                                       M.corcoef = M.corcoef,
                                       n.endogenous = n.endogenous,
                                       index.lower.tri = index.lower.tri,
                                       indexArr.lower.tri = indexArr.lower.tri,
                                       var.coef = var.coef,
                                       name.varcoef = name.varcoef,
                                       name.otherVar = name.otherVar,
                                       n.varcoef = n.varcoef,
                                       cor.coef = cor.coef,
                                       name.corcoef = name.corcoef)
    }
    
### ** export
    return(out)
    
}

## * conditionalMoment.lme
#' @rdname conditionalMoment
#' @export
conditionalMoment.lme <- function(object, attr.param, ...){

    name.rancoef <- attr.param$ran.coef
    out <- conditionalMoment.gls(object, attr.param = attr.param,
                                 name.rancoef = name.rancoef, ...)

    ##  the derivative regarding the random effect is added by skeletonDtheta.lme
    out$name.3deriv <- c(out$name.3deriv, name.rancoef)
    out$skeleton$class.ran <- class(object$modelStruct$reStruct)

    return(out)
}
## * conditionalMoment.lvm
#' @rdname conditionalMoment
#' @export
conditionalMoment.lvm <- function(object, data,
                                  first.order, second.order,
                                  name.endogenous, name.latent,
                                  ...){

    if(first.order == FALSE && second.order == TRUE){
        stop("Cannot pre-compute quantities for the second order derivatives ",
             "without those for the first order derivatives \n")
    }

### ** Initialize conditional moments   
    Moment <- skeleton(object,
                       name.endogenous = name.endogenous, 
                       name.latent = name.latent, 
                       as.lava = TRUE)

### ** Initialize partial derivatives of the conditional moments
    if(first.order){
        dMoment <- skeletonDtheta(object, data = data,
                                  df.param.all = Moment$df.param,
                                  param2originalLink = Moment$param2originalLink,
                                  name.endogenous = name.endogenous, 
                                  name.latent = name.latent)
    }else{
        dMoment <- NULL
    }
    
### ** Initialize second order partial derivatives of the conditional moments
    if(second.order){
        d2Moment <- skeletonDtheta2(object, data = data,
                                    df.param.all = Moment$df.param,
                                    param2originalLink = Moment$param2originalLink,
                                    name.latent = name.latent)
    }else{
        d2Moment <- NULL
    }

### ** Export
    return(c(Moment, list(dMoment.init = dMoment, d2Moment.init = d2Moment)))
}
    
    
## * conditionalMoment.lvmfit
#' @rdname conditionalMoment
#' @export
conditionalMoment.lvmfit <- function(object, data, param, 
                                     first.order, second.order, usefit,
                                     ...){

### ** normalize arguments
    name.endogenous <- endogenous(object)
    n.endogenous <- length(name.endogenous)
    name.latent <- latent(object)
    n.latent <- length(name.latent)

    data <- as.matrix(data[,lava::manifest(object),drop=FALSE])

### ** initialize
    if(is.null(object$conditionalMoment)){       
        object$conditionalMoment <- conditionalMoment(lava::Model(object),
                                                      data = data,
                                                      first.order = first.order,
                                                      second.order = second.order,
                                                      name.endogenous = name.endogenous,
                                                      name.latent = name.latent)

        ##  param with non-zero third derivative
        type.3deriv <- c("alpha","Gamma","Lambda","B","Psi_var","Sigma_var","Psi_cov","Sigma_cov")
        index.keep <- intersect(which(!is.na(object$conditionalMoment$df.param$lava)),
                                which(object$conditionalMoment$df.param$detail %in% type.3deriv)
                                )    
        object$conditionalMoment$name.3deriv <- object$conditionalMoment$df.param[index.keep, "originalLink"]
    }

### ** update according to the value of the model coefficients
    if(usefit){

        ## *** conditional moments
        object$conditionalMoment$value <- skeleton(object, data = data, param = param,
                                                   name.endogenous = name.endogenous, 
                                                   name.latent = name.latent)

        if(object$conditionalMoment$skeleton$toUpdate["param"]){
            object$conditionalMoment$param <- coef(object)
        }
        if(object$conditionalMoment$skeleton$toUpdate["mu"]){            
            if(n.latent==0){
                object$conditionalMoment$mu <- object$conditionalMoment$value$nu.XK
            }else{
                object$conditionalMoment$mu <- object$conditionalMoment$value$nu.XK + object$conditionalMoment$value$alpha.XGamma.iIB %*% object$conditionalMoment$value$Lambda
            }            
        }
        if(object$conditionalMoment$skeleton$toUpdate["Omega"]){
            object$conditionalMoment$Omega <- getVarCov2(object)
        }
        
        ## *** first order derivatives
        if(first.order){            
            out <- skeletonDtheta(object,
                                  name.endogenous = name.endogenous, 
                                  name.latent = name.latent)
            object$conditionalMoment$dmu <- out$dmu
            object$conditionalMoment$dOmega <- out$dOmega            
        }

        ## *** second order derivatives
        if(second.order){
            out2 <- skeletonDtheta2(object)
            object$conditionalMoment$d2mu <- out2$d2mu
            object$conditionalMoment$d2Omega <- out2$d2Omega
        }
       
    }
     
### ** Export
    return(object$conditionalMoment)
}

### sCorrect-information.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 19 2018 (14:17) 
## Version: 
## Last-Updated: jan 23 2024 (10:25) 
##           By: Brice Ozenne
##     Update #: 446
##----------------------------------------------------------------------
## 
### Commentary: 
## Compute information, hessian, and first derivative of information
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - information2
#' @title  Expected Information With Small Sample Correction. 
#' @description  Extract the expected information matrix from a latent variable model.
#' Similar to \code{lava::information} but with small sample correction.
#' @name information2
#'
#' @param object,x a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the information matrix.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @return A matrix with as many rows and columns as the number of coefficients.
#' 
#' @examples
#' #### simulate data ####
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#' 
#' #### linear models ####
#' e.lm <- lm(formula.lvm,data=d)
#' 
#' #### latent variable models ####
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' information(e.lvm)
#' information2(e.lvm)
#' information2(e.lvm)[1:4,1:4] -  solve(vcov(e.lm))
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`information2` <-
  function(object, as.lava, ssc, ...) UseMethod("information2")

## * information2.lvmfit
#' @rdname information2
#' @export
information2.lvmfit <- function(object, as.lava = TRUE, ssc = lava.options()$ssc, ...){

    return(information(estimate2(object, ssc = ssc, ...), as.lava = as.lava))

}

## * information2.lvmfit2
#' @rdname information2
#' @export
information2.lvmfit2 <- function(object, as.lava = TRUE, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    out <- object$sCorrect$information[names(object$sCorrect$skeleton$originalLink2param),
                                       names(object$sCorrect$skeleton$originalLink2param),
                                       drop=FALSE]
    if(as.lava==FALSE){
        dimnames(out) <- list(as.character(object$sCorrect$skeleton$originalLink2param),
                              as.character(object$sCorrect$skeleton$originalLink2param))
    }
    return(out)
}

## * information.lvmfit2
#' @rdname information2
#' @export
information.lvmfit2 <- function(x, ...){ ## necessary as first argument of information must be x 
    information2(x, ...)
}

## * .information2
#' @title Compute the Expected Information Matrix From the Conditional Moments
#' @description Compute the expected information matrix from the conditional moments.
#' @name information2-internal
#' 
#' @details \code{calc_information} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.information2 <- function(dmu, dOmega, OmegaM1,
                          missing.pattern, unique.pattern, name.pattern,
                          grid.mean, grid.var, name.param,
                          leverage, weights = NULL, n.cluster){
    if(lava.options()$debug){cat(".information2\n")}
    if(is.null(weights)){weights <- rep(1,n.cluster)}
    
    ## ** Prepare
    n.grid.mean <- NROW(grid.mean)
    n.grid.var <- NROW(grid.var)
    n.param <- length(name.param)
    n.pattern <- length(name.pattern)

    Info <- matrix(0, nrow = n.param, ncol = n.param,
                   dimnames = list(name.param,name.param))
    if(length(dmu)>0){
        index.mean <- 1:n.grid.mean
    }else{
        index.mean <- NULL
    }
    if(length(dOmega)>0){
        index.var <- 1:n.grid.var
    }else{
        index.var <- NULL
    } 

    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iOmegaM1 <- OmegaM1[[iPattern]]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        if(!is.null(leverage)){
            iN.corrected <- sum(weights[iIndex]) - colSums(leverage[iIndex,iY,drop=FALSE])
        }else{
            iN.corrected <- sum(weights[iIndex]) 
        }
        
        ## *** Information relative to the mean parameters
        for(iG in index.mean){ # iG <- 1
            iP1 <- grid.mean[iG,1]
            iP2 <- grid.mean[iG,2]
            Info[iP1,iP2] <- Info[iP1,iP2] + sum(rowSums(dmu[[iP1]][iIndex,iY,drop=FALSE] %*% iOmegaM1 * dmu[[iP2]][iIndex,iY,drop=FALSE])*weights[iIndex])
        }

        ## *** Information realtive to the variance parameters
        for(iG in index.var){ # iG <- 2
            iP1 <- grid.var[iG,1]
            iP2 <- grid.var[iG,2]

            ## NOTE: normally tr(ABAC)=tr(ACAB) but because of the factor n.correct this is no more the case
            ##       so the information may slightly dependent on the ordering of the parameters
            iDiag <- diag(iOmegaM1 %*% dOmega[[iP1]][iY,iY,drop=FALSE] %*% iOmegaM1 %*% dOmega[[iP2]][iY,iY,drop=FALSE])
                        
            Info[iP1,iP2] <- Info[iP1,iP2] + 1/2*sum(iDiag*iN.corrected)
        }        
    }

    ## ** Make Info a symmetric matrix
    Info <- symmetrize(Info, update.upper = NULL)
        
    ## ** export
    return(Info)
}

### sCorrect-getIndexOmega.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 25 2019 (10:52) 
## Version: 
## Last-Updated: jan 17 2022 (14:26) 
##           By: Brice Ozenne
##     Update #: 103
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Identify the Endogenous Variables
#' @description Identify the endogenous variables, i.e., returns a vector with length the number of observations,
#' whose values are the index of the repetitions.
#' @name getIndexOmega
#'
#' @param object a \code{lvmfit} object.
#' @param data dataset.
#' @param ... [internal] Only used by the generic method.
#'  
#' @concept extractor
#' @keywords internal
`.getIndexOmega` <-
    function(object, data, ...) UseMethod(".getIndexOmega")

## * Examples
#' @examples
#' \dontrun{
#' #### simulate data ####
#' set.seed(10)
#' dW <- sampleRepeated(10, format = "wide")
#' set.seed(10)
#' dL <- sampleRepeated(10, format = "long")
#' dL$time2 <- paste0("visit",dL$time)
#' 
#' #### lvm model ####
#' e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
#' ## lavaSearch2:::.getIndexOmega(e.lvm, data = dW)
#' }

## * .getIndexOmega.lvm
#' @rdname getIndexOmega
.getIndexOmega.lvm <- function(object, data, ...){

    ## ** check missing value in exogenous variables
    name.exogenous <- exogenous(object)
    missing.var <- name.exogenous[name.exogenous %in% names(data) == FALSE]

    if(length(missing.var)>0){
        cat2bin <- var2dummy(object$model, var = names(data), data = data)
        name.exogenous[name.exogenous %in% missing.var] <- names(cat2bin)[cat2bin %in% missing.var]
        name.exogenous <- unique(name.exogenous)
    }
    test.na <- Reduce("+",lapply(name.exogenous, function(iCol){is.na(data[[iCol]])}))
    if(any(test.na>0)){
        stop("Does not support missing values in exogenous variables. \n",
             "Consider removing the corresponding rows in the dataset. \n")
    }

    ## ** index.Omega
    n.obs <- NROW(data)
    name.endogenous <- endogenous(object)
    n.endogenous <- length(name.endogenous)

    M.index <- matrix(1:n.endogenous, nrow = n.obs, ncol = n.endogenous, byrow = TRUE)
    index.na <- which(is.na(as.data.frame(data)[,name.endogenous]))
    if(length(index.na)>0){
        M.index[index.na] <- NA
    }
    return(as.double(t(M.index)))
}

## * .getIndexOmega.lvmfit
#' @rdname getIndexOmega
.getIndexOmega.lvmfit <- .getIndexOmega.lvm

######################################################################
### sCorrect-getIndexOmega.R ends here

### information2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 19 2018 (14:17) 
## Version: 
## Last-Updated: feb 11 2019 (16:50) 
##           By: Brice Ozenne
##     Update #: 275
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - information2
#' @title  Extract The full Information Matrix
#' @description  Extract the full information matrix from a Gaussian linear model.
#' @name information2
#'
#' @param object a linear model or a latent variable model
#' @param ... arguments to be passed to \code{vcov2}.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#'
#' @return A matrix.
#' 
#' @examples
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' ## linear model
#' e.lm <- lm(formula.lvm,data=d)
#' info.tempo <- vcov2(e.lm, bias.correct = TRUE)
#' info.tempo[names(coef(e.lm)),names(coef(e.lm))] - vcov(e.lm)
#'
#' ## latent variable model
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' vcov.tempo <- vcov2(e.lvm, bias.correct = FALSE)
#' round(vcov.tempo %*% information(e.lvm), 5)
#'
#' @concept small sample inference
#' @export
`information2` <-
  function(object, ...) UseMethod("information2")

## * information2.lm
#' @rdname information2
#' @export
information2.lm <- function(object, ...){
    return(solve(vcov2(object, ...)))
}

## * information2.gls
#' @rdname information2
#' @export
information2.gls <- information2.lm

## * information2.lme
#' @rdname information2
#' @export
information2.lme <- information2.lm

## * information2.lvmfit
#' @rdname information2
#' @export
information2.lvmfit <- information2.lm

## * information2.lm2
#' @rdname information2
#' @export
information2.lm2 <- function(object, ...){
    return(solve(vcov2(object, ...)))
}

## * information2.gls2
#' @rdname information2
#' @export
information2.gls2 <- information2.lm2

## * information2.lme2
#' @rdname information2
#' @export
information2.lme2 <- information2.lm2

## * information2.lvmfit
#' @rdname information2
#' @export
information2.lvmfit2 <- information2.lm2

## * .information2
#' @title Compute the Expected Information Matrix From the Conditional Moments
#' @description Compute the expected information matrix from the conditional moments.
#' @name information2-internal
#' 
#' @details \code{.information2} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.information2 <- function(dmu, dOmega,
                          Omega, n.corrected,
                          index.Omega, leverage, n.cluster,
                          grid.meanparam, n.grid.meanparam,
                          grid.varparam, n.grid.varparam,
                          name.param, n.param){

### ** Prepare
    test.global <- is.null(index.Omega)
    if(!test.global){
        OmegaM1 <- lapply(1:n.cluster, function(iC){
            return(chol2inv(chol(Omega[index.Omega[[iC]],index.Omega[[iC]]])))
        })
    }else{
        OmegaM1 <- chol2inv(chol(Omega))
    }
    
    Info <- matrix(0, nrow = n.param, ncol = n.param,
                   dimnames = list(name.param,name.param))
    if(length(dmu)>0){
        index.meanparam <- 1:n.grid.meanparam
    }else{
        index.meanparam <- NULL
    }
    if(length(dOmega)>0){
        index.varparam <- 1:n.grid.varparam
    }else{
        index.varparam <- NULL
    } 

### ** Global    
    if(test.global){
        ## *** Information relative to the mean parameters
        for(iG in index.meanparam){ # iG <- 1
            iP1 <- grid.meanparam[iG,1]
            iP2 <- grid.meanparam[iG,2]

            Info[iP1,iP2] <- Info[iP1,iP2] + sum(dmu[[iP1]] %*% OmegaM1 * dmu[[iP2]])
        }

        ## *** Information realtive to the variance parameters
        for(iG in index.varparam){ # iG <- 1
            iP1 <- grid.varparam[iG,1]
            iP2 <- grid.varparam[iG,2]

            iDiag <- diag(OmegaM1 %*% dOmega[[iP1]] %*% OmegaM1 %*% dOmega[[iP2]])
            Info[iP1,iP2] <- Info[iP1,iP2] + 1/2*sum(iDiag*n.corrected)
        }
    }

### ** Individual specific (missing data)
    if(!test.global){
        ## *** Information relative to the mean parameters
        for(iC in 1:n.cluster){ # iC <- 1
            iIndex <- index.Omega[[iC]]

            for(iG in index.meanparam){ # iG <- 1
                iP1 <- grid.meanparam[iG,1]
                iP2 <- grid.meanparam[iG,2]

                Info[iP1,iP2] <- Info[iP1,iP2] + sum(dmu[[iP1]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] * dmu[[iP2]][iC,iIndex,drop=FALSE])            
            }

            ## *** Information relative to the variance parameters
            for(iG in index.varparam){ # iG <- 1
                iP1 <- grid.varparam[iG,1]
                iP2 <- grid.varparam[iG,2]
                iDiag <- diag(OmegaM1[[iC]] %*% dOmega[[iP1]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP2]][iIndex,iIndex,drop=FALSE])
                Info[iP1,iP2] <- Info[iP1,iP2] + 1/2 * sum(iDiag * (1 - leverage[iC,iIndex]))
            }
        }        
    }

    ## ** Make Info a symmetric matrix
    Info <- symmetrize(Info, update.upper = NULL)
        
    ## ** export
    return(Info)
}

## * .hessian2
#' @title Compute the Hessian Matrix From the Conditional Moments
#' @description Compute the Hessian matrix from the conditional moments.
#' @name information2-internal
#' 
#' @details \code{.hessian} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.hessian2 <- function(dmu, d2mu, dOmega, d2Omega, 
                      Omega, n.corrected,
                      index.Omega, leverage, n.cluster,
                      grid.meanparam, n.grid.meanparam,
                      grid.varparam, n.grid.varparam,
                      name.param, n.param, residuals){

    ## ** Prepare
    test.global <- is.null(index.Omega)
    if(!test.global){
        OmegaM1 <- lapply(1:n.cluster, function(iC){
            return(chol2inv(chol(Omega[index.Omega[[iC]],index.Omega[[iC]]])))
        })
    }else{
        OmegaM1 <- chol2inv(chol(Omega))
    }
    
    hessian <- array(0, dim = c(n.param, n.param, n.cluster),
                     dimnames = list(name.param,name.param,NULL))
    if(length(dmu)>0){
        index.meanparam <- 1:n.grid.meanparam
    }else{
        index.meanparam <- NULL
    }
    if(length(dOmega)>0){
        index.varparam <- 1:n.grid.varparam
    }else{
        index.varparam <- NULL
    }

    if((n.grid.meanparam>0) && (n.grid.varparam>0)){
        name.meanparam <- unique(unlist(grid.meanparam[,c("Var1","Var2")]))
        name.varparam <- unique(unlist(grid.varparam[,c("Var1","Var2")]))

        grid.hybridparam <- .combination(name.meanparam,name.varparam)
        n.hybridparam <- NROW(grid.hybridparam)
        index.hybridparam <- 1:n.hybridparam
    }else{
        grid.hybridparam <- NULL
        n.hybridparam <- 0
        index.hybridparam <- NULL
    }

    ## ** Global    
    if(test.global){
        ## *** second derivative relative to the mean parameters
        for(iG in index.meanparam){ # iG <- 1
            iP1 <- grid.meanparam[iG,1]
            iP2 <- grid.meanparam[iG,2]

            if(grid.meanparam[iG,"deriv12"]){
                term1 <- rowSums((d2mu[[iP1]][[iP2]] %*% OmegaM1) * residuals)
            }else if(grid.meanparam[iG,"deriv21"]){
                term1 <- rowSums((d2mu[[iP2]][[iP1]] %*% OmegaM1) * residuals)
            }else{
                term1 <- 0
            }
            term2 <- -rowSums((dmu[[iP1]] %*% OmegaM1) * dmu[[iP2]])
            hessian[iP1,iP2,] <- hessian[iP1,iP2,] + term1 + term2
            hessian[iP2,iP1,] <- hessian[iP1,iP2,]
        }

        ## *** second derivative relative to the variance parameters
        for(iG in index.varparam){ # iG <- 1
            iP1 <- grid.varparam[iG,1]
            iP2 <- grid.varparam[iG,2]

            term1a <- - diag(OmegaM1 %*% dOmega[[iP1]] %*% OmegaM1 %*% dOmega[[iP2]])
            term2 <- - rowSums((residuals %*% OmegaM1 %*% dOmega[[iP2]] %*% OmegaM1 %*% dOmega[[iP1]] %*% OmegaM1) * residuals)
            if(grid.varparam[iG,"deriv12"]){
                term1b <- diag(OmegaM1 %*% d2Omega[[iP1]][[iP2]])
                term3 <- 1/2 * rowSums((residuals %*% OmegaM1 %*% d2Omega[[iP1]][[iP2]] %*% OmegaM1) * residuals)
            }else if(grid.varparam[iG,"deriv21"]){
                term1b <- diag(OmegaM1 %*% d2Omega[[iP2]][[iP1]])
                term3 <- 1/2 * rowSums((residuals %*% OmegaM1 %*% d2Omega[[iP2]][[iP1]] %*% OmegaM1) * residuals)
            }else{
                term1b <- 0
                term3 <- 0
            }
            hessian[iP1,iP2,] <- hessian[iP1,iP2,] - 1/2 * rowSums( sweep(1-leverage, FUN = "*", STATS = term1a + term1b, MARGIN = 2) ) + term2 + term3
            hessian[iP2,iP1,] <- hessian[iP1,iP2,]
        }
        ## *** second derivative relative to the mean and variance parameters
        for(iG in index.hybridparam){ # iG <- 1
            iP1 <- grid.hybridparam[iG,1]
            iP2 <- grid.hybridparam[iG,2]

            term1 <- - rowSums((dmu[[iP1]] %*% OmegaM1 %*% dOmega[[iP2]] %*% OmegaM1) * residuals)
            if(!is.null(dmu[[iP2]]) && !is.null(dOmega[[iP1]])){
                term2 <- - rowSums((dmu[[iP2]] %*% OmegaM1 %*% dOmega[[iP1]] %*% OmegaM1) * residuals)
            }else{
                term2 <- 0
            }
            
            hessian[iP1,iP2,] <- hessian[iP1,iP2,] + term1 + term2
            hessian[iP2,iP1,] <- hessian[iP1,iP2,]
        }
    }

    ## ** Individual specific (missing data)
    if(!test.global){

        ## *** Information relative to the mean parameters
        for(iC in 1:n.cluster){ # iC <- 1
            iIndex <- index.Omega[[iC]]
            
            ## *** second derivative relative to the mean parameters
            for(iG in index.meanparam){ # iG <- 1
                iP1 <- grid.meanparam[iG,1]
                iP2 <- grid.meanparam[iG,2]

                if(grid.meanparam[iG,"deriv12"]){
                    term1 <- sum((d2mu[[iP1]][[iP2]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                }else if(grid.meanparam[iG,"deriv21"]){
                    term1 <- sum((d2mu[[iP2]][[iP1]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                }else{
                    term1 <- 0
                }
                term2 <- -sum((dmu[[iP1]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * dmu[[iP2]][iC,iIndex,drop=FALSE])
                hessian[iP1,iP2,iC] <- hessian[iP1,iP2,iC] + term1 + term2
                hessian[iP2,iP1,iC] <- hessian[iP1,iP2,iC]
            }

            ## *** second derivative relative to the variance parameters
            for(iG in index.varparam){ # iG <- 1
                iP1 <- grid.varparam[iG,1]
                iP2 <- grid.varparam[iG,2]

                term1a <- - diag(OmegaM1[[iC]] %*% dOmega[[iP1]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP2]][iIndex,iIndex,drop=FALSE])
                term2 <- - sum((residuals[iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP2]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP1]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                if(grid.varparam[iG,"deriv12"]){
                    term1b <- diag(OmegaM1[[iC]] %*% d2Omega[[iP1]][[iP2]][iIndex,iIndex,drop=FALSE])
                    term3 <- 1/2 * sum((residuals[iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% d2Omega[[iP1]][[iP2]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                }else if(grid.varparam[iG,"deriv21"]){
                    term1b <- diag(OmegaM1[[iC]] %*% d2Omega[[iP2]][[iP1]][iIndex,iIndex,drop=FALSE])
                    term3 <- 1/2 * sum((residuals[iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% d2Omega[[iP2]][[iP1]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                }else{
                    term1b <- 0
                    term3 <- 0
                }
                hessian[iP1,iP2,iC] <- hessian[iP1,iP2,iC] - 1/2 * sum( (1-leverage[iC,iIndex,drop=FALSE]) * (term1a + term1b) ) + term2 + term3
                hessian[iP2,iP1,iC] <- hessian[iP1,iP2,iC]
            }

            ## *** second derivative relative to the mean and variance parameters
            for(iG in index.hybridparam){ # iG <- 1
                iP1 <- grid.hybridparam[iG,1]
                iP2 <- grid.hybridparam[iG,2]

                term1 <- - sum((dmu[[iP1]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP2]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                if(!is.null(dmu[[iP2]]) && !is.null(dOmega[[iP1]])){
                    term2 <- - sum((dmu[[iP2]][iC,iIndex,drop=FALSE] %*% OmegaM1[[iC]] %*% dOmega[[iP1]][iIndex,iIndex,drop=FALSE] %*% OmegaM1[[iC]]) * residuals[iC,iIndex,drop=FALSE])
                }else{
                    term2 <- 0
                }

                hessian[iP1,iP2,iC] <- hessian[iP1,iP2,iC] + term1 + term2
                hessian[iP2,iP1,iC] <- hessian[iP1,iP2,iC]
            }
        }        
    }


    ## ** export
    return(hessian)
}

## * .dInformation2
#' @title Compute the First Derivative of the Expected Information Matrix
#' @description Compute the first derivative of the expected information matrix.
#' @name dInformation2-internal
#' 
#' @details \code{.dInformation2} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.dInformation2 <- function(dmu, d2mu, dOmega, d2Omega,
                           Omega, OmegaM1, n.corrected,
                           index.Omega, leverage, n.cluster,
                           name.param, name.3deriv){
    n.param <- length(name.param)
    index.deriv <- match(name.3deriv, name.param)

### ** prepare
    test.global <- is.null(index.Omega)
    if(!test.global){
        M.template <- Omega
        M.template[] <- 0
    }
    
    dInfo <-  array(0,
                    dim = c(n.param, n.param, length(name.3deriv)),
                    dimnames = list(name.param, name.param, name.3deriv))
    
### ** loop
    for(iDeriv in index.deriv){ # iDeriv <- 4
        for(iP1 in 1:n.param){ # iP1 <- 1
            for(iP2 in iP1:n.param){ # iP2 <- 1
                
                iNameD <- name.param[iDeriv]
                iName1 <- name.param[iP1]
                iName2 <- name.param[iP2]
                 ## cat(iNameD," ",iName1,"",iName2,"\n")
                
                ## *** identify relevant terms
                test.Omega1 <- !is.null(dOmega[[iNameD]]) && !is.null(dOmega[[iName1]]) && !is.null(dOmega[[iName2]])
                test.Omega2a <- !is.null(d2Omega[[iNameD]][[iName1]]) && !is.null(dOmega[[iName2]])
                test.Omega2b <- !is.null(d2Omega[[iName1]][[iNameD]]) && !is.null(dOmega[[iName2]])
                test.Omega3a <- !is.null(d2Omega[[iNameD]][[iName2]]) && !is.null(dOmega[[iName1]])
                test.Omega3b <- !is.null(d2Omega[[iName2]][[iNameD]]) && !is.null(dOmega[[iName1]])
                
                test.mu1a <- !is.null(d2mu[[iNameD]][[iName1]]) && !is.null(dmu[[iName2]])
                test.mu1b <- !is.null(d2mu[[iName1]][[iNameD]]) && !is.null(dmu[[iName2]])
                test.mu2a <- !is.null(d2mu[[iNameD]][[iName2]]) && !is.null(dmu[[iName1]])
                test.mu2b <- !is.null(d2mu[[iName2]][[iNameD]]) && !is.null(dmu[[iName1]])
                test.mu3 <- !is.null(dOmega[[iNameD]]) && !is.null(dmu[[iName1]]) && !is.null(dmu[[iName2]])

                if((test.Omega1 + test.Omega2a + test.Omega2b + test.Omega3a + test.Omega3b + test.mu1a + test.mu1b + test.mu2a + test.mu2b + test.mu3) == 0){
                    next
                }

                ## *** extract quantities for computations 
                if(test.mu1a){
                    d2mu.D1 <- d2mu[[iNameD]][[iName1]]
                }else if(test.mu1b){
                    d2mu.D1 <- d2mu[[iName1]][[iNameD]]
                }
                if(test.mu2a){
                    d2mu.D2 <- d2mu[[iNameD]][[iName2]]
                }else if(test.mu2b){
                    d2mu.D2 <- d2mu[[iName2]][[iNameD]]
                }
                if(test.Omega2a){
                    d2Omega.D1 <- d2Omega[[iNameD]][[iName1]]
                }else if(test.Omega2b){
                    d2Omega.D1 <- d2Omega[[iName1]][[iNameD]]
                }
                if(test.Omega3a){
                    d2Omega.D2 <- d2Omega[[iNameD]][[iName2]]
                }else{
                    d2Omega.D2 <- d2Omega[[iName2]][[iNameD]]
                }
                
                if(test.global){
                    ## *** Global: extract quantities for computations
                    if(!is.null(dOmega[[iNameD]])){
                        OmegaM1.dOmega.D <- OmegaM1 %*% dOmega[[iNameD]]
                    }
                    if(!is.null(dOmega[[iName1]])){
                        OmegaM1.dOmega.1 <- OmegaM1 %*% dOmega[[iName1]]
                    }
                    if(!is.null(dOmega[[iName2]])){
                        OmegaM1.dOmega.2 <- OmegaM1 %*% dOmega[[iName2]]
                    }                    

                    ## *** Global: compute
                    if(test.Omega1){
                        iDiag1 <- diag(OmegaM1.dOmega.D %*% OmegaM1.dOmega.1 %*% OmegaM1.dOmega.2)
                        iDiag2 <- diag(OmegaM1.dOmega.1 %*% OmegaM1.dOmega.D %*% OmegaM1.dOmega.2)
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - 1/2 * sum(iDiag1 * n.corrected + iDiag2 * n.corrected)
                    }

                    if(test.Omega2a || test.Omega2b){
                        iDiag <- diag(OmegaM1 %*% d2Omega.D1 %*% OmegaM1.dOmega.2)
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * n.corrected)
                    }

                    if(test.Omega3a || test.Omega3b){
                        iDiag <- diag(OmegaM1.dOmega.1 %*% OmegaM1 %*% d2Omega.D2)
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * n.corrected)
                    }

                    if(test.mu1a || test.mu1b){
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(d2mu.D1 %*% OmegaM1 * dmu[[iName2]])
                    }

                    if(test.mu2a || test.mu2b){
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(dmu[[iName1]] %*% OmegaM1 * d2mu.D2)
                    }
                    if(test.mu3){
                        dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - sum(dmu[[iName1]] %*% OmegaM1.dOmega.D %*% OmegaM1 * dmu[[iName2]])
                    }
                    
                }else{
                    for(iC in 1:n.cluster){ # iC <- 1
                        iIndex <- index.Omega[[iC]]
                        
                        if(!is.null(dOmega[[iNameD]])){
                            OmegaM1.dOmega.D <- OmegaM1[[iC]] %*% dOmega[[iNameD]][iIndex,iIndex]
                        }
                        if(!is.null(dOmega[[iName1]])){
                            OmegaM1.dOmega.1 <- OmegaM1[[iC]] %*% dOmega[[iName1]][iIndex,iIndex]
                        }
                        if(!is.null(dOmega[[iName2]])){
                            OmegaM1.dOmega.2 <- OmegaM1[[iC]] %*% dOmega[[iName2]][iIndex,iIndex]
                        }

                        if(test.Omega1){
                            iDiag1 <- diag(OmegaM1.dOmega.D %*% OmegaM1.dOmega.1 %*% OmegaM1.dOmega.2)
                            iDiag2 <- diag(OmegaM1.dOmega.1 %*% OmegaM1.dOmega.D %*% OmegaM1.dOmega.2)
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - 1/2 * sum((iDiag1+iDiag2) * (1 - leverage[iC,iIndex]))
                        }
                        if(test.Omega2a || test.Omega2b){
                            iDiag <- diag(OmegaM1[[iC]] %*% d2Omega.D1[iIndex,iIndex] %*% OmegaM1.dOmega.2)
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * (1 - leverage[iC,iIndex]))
                        }

                        if(test.Omega3a || test.Omega3b){
                            iDiag <- diag(OmegaM1.dOmega.1 %*% OmegaM1[[iC]] %*% d2Omega.D2[iIndex,iIndex])
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * (1 - leverage[iC,iIndex]))
                        }

                        if(test.mu1a || test.mu1b){
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(d2mu.D1[iC,iIndex] %*% OmegaM1[[iC]] * dmu[[iName2]][iC,iIndex])
                        }
                        
                        if(test.mu2a || test.mu2b){
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(dmu[[iName1]][iC,iIndex] %*% OmegaM1[[iC]] * d2mu.D2[iC,iIndex])
                        }
                        
                        if(test.mu3){                            
                            dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - sum(dmu[[iName1]][iC,iIndex] %*% OmegaM1.dOmega.D %*% OmegaM1[[iC]] * dmu[[iName2]][iC,iIndex])
                        }
                    }
                }
            }
            
        }
        ## *** Symmetrize
        dInfo[,,iNameD] <- symmetrize(dInfo[,,iNameD], update.upper = NULL)
    }

    ### ** export
    return(dInfo)
}

## * Documentation
#' @name validFCTs
#' @aliases validClass
#' @aliases validDimension
#' @aliases validInteger
#' @aliases validLogical
#' @aliases validNames
#' @aliases validNumeric
#' @aliases validPath
#' @title Check Arguments of a function.
#' 
#' @description Check the validity of the arguments in functions.
#' 
#' @param value1 the value of the (first) argument to be checked
#' @param value2 the second value of a second argument whose dimensions should be consistent with the first one
#' @param name1 the name of the (first) argument.
#' @param name2 the name of the second argument.
#' @param validClass the acceptable classes(s) for the argument. 
#' @param validDimension the acceptable dimension for the argument. If \code{NULL} then name2 is used as a reference.
#' @param valid.length the acceptable length(s) for the argument. If \code{NULL} no test is performed.
#' @param valid.values the acceptable value(s) for the argument. If \code{NULL} no test is performed. Can also be "character" or "character_or_logical".
#' @param super.classes uses the \code{is} function instead of \code{class} to test the class of the object.
#' @param refuse.NULL should an error be output if value is \code{NULL}.
#' @param refuse.NA should an error be output if value contains \code{NA}.
#' @param refuse.duplicates should an error be output if value contains duplicated values.
#' @param refuse.values values that must not appear in the argument
#' @param type For \code{validDimension}: the type of operator used to check the dimensions. For \code{validPath} either "dir" or "file" to check whether to path points to an existing directory or file.
#' @param required.values values that must appear in the argument
#' @param min the minimum acceptable value
#' @param max the maximum acceptable value
#' @param extension filter the files by the type of extension. 
#' @param method the name of the function using the argument.
#' @param check.fsep display a warning when the separator is not correctly specified in 
#' @param addPP add ": " after the name of the function in the error message.
#' 
#' @return An invisible \code{TRUE} or an error message.
#' 
#' @concept check
#' @keywords internal

## * validCharacter
#' @rdname validFCTs
validCharacter <- function(value1, name1 = as.character(substitute(value1)), valid.length, 
                           valid.values = "character", refuse.NULL = TRUE, refuse.duplicates = FALSE, 
                           method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  if(is.null(value1)){
    
    if(refuse.NULL == TRUE){
      stop(method, "\'", name1, "\' must not be NULL \n")
    }
    
  }else{
    
    #### check size
    n.value1 <- length(value1)
    
    if(!is.null(valid.length) && n.value1 %in% valid.length == FALSE){
      stop(method, "\'", name1, "\' must have length ", paste(valid.length, collapse = " or "), "  \n", 
           "length(", name1, ") : ", n.value1, "\n")
    }
    
    #### check duplicates
    if(refuse.duplicates == TRUE && any(duplicated(value1))){
      stop(method, "\'", name1, "\' contains duplicated values: ", "\"",paste(unique(value1[duplicated(value1)]), collapse = "\" \""), "\" \n")
    }
    
    #### check values
    if(identical(valid.values,"character")){
      
      if(any(is.character(value1) == FALSE)){
        stop(method, "\'", name1, "\' must be a ", if(n.value1 == 1){"character"}else{"vector of characters"}," \n", 
             "is(", name1, ") : ", paste(is(value1), collapse = " "), "\n")
      }
      
    } else if(identical(valid.values,"character_or_logical")){
      
      if(any( (is.character(value1) == FALSE) * (is.logical(value1) == FALSE) > 0 )){
        stop(method, "\'", name1, "\' must be a ", if(n.value1 == 1){"character or logical"}else{"vector of characters or logicals"}," \n", 
             "is(", name1, ") : ", paste(is(value1), collapse = " "), "\n")
      }
      
    } else if(!is.null(valid.values) && any(value1 %in% valid.values == FALSE)){
      
      stop(method, "wrong specification of \'", name1, "\' \n", 
           "valid values for \'", name1, "\' : ", if(refuse.NULL == FALSE){"NULL"}, " \"", paste(valid.values, collapse = "\" \""), "\" \n", 
           "refused value",if(sum(value1 %in% valid.values == FALSE)>1){"s"}," for \'", name1, "\' : \"", paste(value1[value1 %in% valid.values == FALSE], collapse = "\" \""), "\"\n")
      
    }
    
  }
  
  return(invisible(TRUE))
  
}

## * validClass
#' @rdname validFCTs
validClass <- function(value1, name1 = as.character(substitute(value1)), validClass, 
                       super.classes = TRUE, method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  if(super.classes == TRUE){
    
    if( all(is(value1) %in% validClass == FALSE) ){
      stop(method, "class of \'", name1, "\' must be one of the following \"", paste(validClass,collapse="\" \""), "\"  \n", 
           "proposed superclass : \"", paste(is(value1),collapse="\" \""), "\" \n")
    }  
    
  }else{
    
    if( class(value1) %in% validClass == FALSE){
      stop(method, "class of \'", name1, "\' must be \"", paste(validClass,collapse="\" \""),"\"  \n", 
           "proposed class : ", class(value1)[[1]], "\n")
    }  
    
  }
  
  return(invisible(TRUE))
  
}

## * validDimension
#' @rdname validFCTs
validDimension <- function(value1, value2 = NULL, name1 = as.character(substitute(value1)), name2 = as.character(substitute(value2)),
                           validDimension = NULL,
                           type = c("NROW","NCOL"), method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  n.type <- length(type)
  
  #### dimension 1
  testDimension <- sapply(1:n.type, function(x){
    do.call(type[x], list(value1))
  })
  
  #### dimension 2
  
  
  if(is.null(validDimension)){
    
    validDimension <- sapply(1:n.type, function(x){
      do.call(type[x], list(value2))
    })
    test.validDimension <- TRUE
    
  }else if(is.null(name2)){
    
    test.validDimension <- FALSE
    
  }else{
    
    test.validDimension <- TRUE
    
  }
  
  #### main
  for(iter_type in 1:n.type){
    
    if(testDimension[iter_type] != validDimension[iter_type]){
      
      if(test.validDimension){
        stop(method, "dimension mismatch between argument \'", name1, "\' and argument \'", name2, "\' \n", 
             type[iter_type],"(", name1, ") = ", testDimension[iter_type], " \n", 
             type[iter_type],"(", name2, ") = ", validDimension[iter_type], " \n")  
      }else{
        stop(method, "dimension mismatch between argument \'", name1, "\' and argument \'", name2, "\' \n", 
             type[iter_type],"(", name1, ") = ", testDimension[iter_type], " \n", 
             type[iter_type],"(", name2, ") = ", validDimension[iter_type], " \n")
        
      }
      
    }
    
  }
  
  return(invisible(TRUE))
}

## * validInteger
#' @rdname validFCTs
validInteger <- function(value1, name1 = as.character(substitute(value1)), valid.length, 
                         valid.values = NULL, min = NULL, max = NULL, 
                         refuse.NA = TRUE, refuse.NULL = TRUE, refuse.duplicates = FALSE, 
                         method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  validNumeric(value1 = value1, name1 = name1, valid.length = valid.length, min = min, max = max, 
               refuse.NA = refuse.NA, refuse.NULL = refuse.NULL, refuse.duplicates = refuse.duplicates, method = method)
  
  #### check integer
  if(!is.null(value1) && any(value1 %% 1 > 0)){
    stop(method, "\'", name1, "\' must contain integers not doubles \n",        
         "invalid value(s) in ", name1, " : ", paste(value1[value1 %% 1 > 0], collapse = " "), "\n")
  }
  
  return(invisible(TRUE))
}

## * validLogical
#' @rdname validFCTs
validLogical <- function(value1, name1 = as.character(substitute(value1)), valid.length, 
                         refuse.NULL = TRUE, refuse.NA = TRUE, 
                         method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  if(is.null(value1)){
    
    #### NULL
    if(refuse.NULL == TRUE){
      stop(method, "\'", name1, "\' must be logical ",if(refuse.NA == FALSE){"or NA"}," and not NULL \n")
    }
    
  }else{ 
    
    #### Size
    if(!is.null(valid.length) && length(value1) %in% valid.length == FALSE){
      stop(method, "\'", name1, "\' must have length ", paste(valid.length, collapse = " or "), "  \n", 
           "length(", name1, ") : ", length(value1), "\n")
    } 
    
    #### Type
    if(any(is.logical(value1) == FALSE)){
      stop(method, "\'", name1, "\' must be ", if(refuse.NULL == FALSE){"NULL or "}, if(refuse.NA == FALSE){"NA or "},"TRUE or FALSE \n",        
           "is(", name1, ") : ", paste(is(value1), collapse = " "), "\n")
    }
    
    if(refuse.NA == TRUE && any(is.na(value1)) ){
      stop(method, "\'", name1, "\' must be logical ",if(refuse.NULL == FALSE){"or NULL"}," and not NA \n")
    }
    
  }
  
  return(invisible(TRUE))
}

## * validNames
#' @rdname validFCTs
validNames <- function(value1, name1 = as.character(substitute(value1)), refuse.NULL = TRUE,
                       valid.length = NULL, valid.values = NULL, required.values = NULL, refuse.values = NULL,
                       method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  ## type
  if(is.matrix(value1)){
    value1 <- colnames(value1)
  }
  
  if(inherits(value1,"data.frame") || is.list(value1)){
    value1 <- names(value1)
  }
  
  ## tests
  if(is.null(value1)){
    
    if(refuse.NULL == TRUE){
      stop(method, "names of \'", name1, "\' must not be NULL \n")
    }
    
  }else{
    #### check size
    n.value1 <- length(value1)
    
    if(!is.null(valid.length) && n.value1 %in% valid.length == FALSE){
      stop(method, "\'", name1, "\' must have ", paste(valid.length, collapse = " or ")," names  \n", 
           "length(names(", name1, ")) : ", n.value1, "\n")
    }
    
    #### check content
    
    if(!is.null(required.values) && any(required.values %in% value1 == FALSE)){
      
      stop(method, "\'", name1, "\' must contains specific names \n",
           "missing names : \"",paste(required.values[required.values %in% value1 == FALSE], collapse = "\" \""),"\" \n", 
           "proposed names : \"", paste(value1, collapse = "\" \""), "\"\n")  
      
    }
    
    if(!is.null(valid.values) && any(value1 %in% valid.values == FALSE)){
      
      stop(method, "wrong specification of \'", name1, "\' \n", 
           "valid names for \'", name1, "\' : \"",paste(valid.values, collapse = "\" \""),"\" \n", 
           "refused names : \"", paste(value1[value1 %in% valid.values == FALSE], collapse = " "), "\"\n")  
      
    }
    
    if(!is.null(refuse.values) && any(value1 %in% refuse.values)){
      
      stop(method, "\'", name1, "\' contains forbidden names:", paste(value1[value1 %in% refuse.values], collapse = " "), "\"\n")  
      
    }
    
    if(any(duplicated(value1))){
      stop(method, "\'", name1, "\' must not contain duplicated names \n", 
           "duplicated names : \"", paste(value1[duplicated(value1)], collapse = " "), "\"\n")  
    }
    
  }
  
  return(invisible(TRUE))
  
}

## * validNumeric
#' @rdname validFCTs
validNumeric <- function(value1, name1 = as.character(substitute(value1)), valid.length,
                         valid.values = NULL , min = NULL, max = NULL,
                         refuse.NA = TRUE, refuse.NULL = TRUE, refuse.duplicates = FALSE, 
                         method = NULL, addPP = TRUE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  if(is.null(value1)){
    
    if(refuse.NULL == TRUE){
      stop(method, "\'", name1, "\' must not be NULL \n")
    }
    
  }else{
    
    #### check length
    if(!is.null(valid.length) && length(value1) %in% valid.length == FALSE){
      stop(method, "\'", name1, "\' must have length ", paste(valid.length, collapse = " or "), "  \n", 
           "length(", name1, ") : ", length(value1), "\n")
    }
    
    #### check NA
    if(refuse.NA == TRUE && any(is.na(value1))){
      stop(method, "\'", name1, "\' must not contain any NA \n", 
           "index of NA values : ", paste(which(is.na(value1)), collapse = " "), "\n")
    }
    
    #### check numeric
    if(any( (is.numeric(value1) == FALSE) * (is.na(value1) == FALSE) )){
      stop(method, "\'", name1, "\' must be numeric \n",        
           "is(", name1, ") : ", paste(is(value1), collapse = " "), "\n")
    }
    
    #### check duplicates
    if(refuse.duplicates == TRUE && any(duplicated(value1))){
      stop(method, "\'", name1, "\' contains duplicated values: ", paste(unique(value1[duplicated(value1)]), collapse = " "), "\n")
    }
    
    #### check min value1
    if(!is.null(min) && any(stats::na.omit(value1) < min)){
      stop(method, "\'", name1, "\' must be bigger than ", min, " \n",        
           "invalid value(s): ", paste(value1[stats::na.omit(value1) < min], collapse = " "), "\n")
    }
    
    #### check max value1
    if(!is.null(max) && any(stats::na.omit(value1) > max)){
      stop(method, "\'", name1, "\' must be smaller than ", max, " \n",        
           "invalid value(s): ", paste(value1[stats::na.omit(value1) > max], collapse = " "), "\n")
    }
    
    #### check valid values
    if(!is.null(valid.values) && any(value1 %in% valid.values == FALSE)){
      
      stop(method, "\'", name1, "\' contain invalid values \n", 
           "valid values for \'", name1, "\' : ", if(refuse.NULL == FALSE){"NULL"}, " \"", paste(valid.values, collapse = "\" \""), "\" \n", 
           "refused value",if(sum(value1 %in% valid.values == FALSE)>1){"s"}," for \'", name1, "\' : \"", paste(value1[value1 %in% valid.values == FALSE], collapse = " "), "\"\n")
      
    }
  }
  
  return(invisible(TRUE))
}

## * validPath
#' @rdname validFCTs
validPath <- function(value1, name1 = as.character(substitute(value1)), type,
                      method = NULL, addPP = TRUE, extension = NULL, check.fsep = FALSE){
  
  if(!is.null(method) && addPP){
    method <- paste0(method, ": ")
  }
  
  validCharacter(type, valid.length = 1, valid.values = c("file", "dir"))
  
  try_path <- switch(type,
                     file = file.exists(value1),
                     dir = dir.exists(value1)
  )
  
  if(try_path == FALSE){
    stop(method, "\'", name1, "\' does not lead to an existing ",switch(type,"file"="file","dir"="directory")," \n", 
         "proposed value: \"", value1, "\"\n", 
         "current path: ", getwd(), "\n")
  }
  
  
  if(type == "dir"){ 
    if(check.fsep == TRUE && substr(value1, start = nchar(value1), stop = nchar(value1)) != "/"){
      warning(method, "possible bad specification of \'", name1, "\' \n", 
              "it should end with a fsep (e.g. \"/\") \n")
    }
  }else if(type == "file" && !is.null(extension)){
    fileExtension <- tools::file_ext(value1) 
    if(fileExtension %in% extension == FALSE){
      stop(method, "\'", name1, "\' has not the expected extension \n", 
           "proposed extension: \"", fileExtension, "\" \n", 
           "expected extension: \"", paste(extension, collapse = "\" \""), "\"\n")
    }
  }
  
  return(invisible(TRUE))
}

### Utils.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 27 2018 (14:32) 
## Version: 
## Last-Updated: Jan 11 2022 (16:43) 
##           By: Brice Ozenne
##     Update #: 161
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * formula
## ** selectResponse (Documentation)
#' @title Response Variable of a Formula
#' @description Return the response variable contained in the formula.
#' @name selectResponse
#' 
#' @param object a formula
#' @param format [character] should an object of type call be returned (\code{format = "call"}),
#' or the names of the variables (\code{format = "vars"})
#' @param ... [internal] Only used by the generic method.
#'
#' @return See argument \code{format}.
#' 
#' @examples
#'
#' \dontrun{
#'
#' selectResponse <- lavaSearch2:::selectResponse
#' selectResponse.formula <- lavaSearch2:::selectResponse.formula
#' 
#' selectResponse(Y1~X1+X2)
#' selectResponse(Y1~X1+X2, format = "vars")
#' selectResponse(Surv(event,time)~X1+X2, format = "vars")
#' 
#' selectResponse(Y1~X1+Y1)
#' selectResponse(Y1+Y2~X1+Y1, format = "vars")
#' 
#' selectResponse(~X1+X2)
#' selectResponse(~X1+X2, format = "vars")
#' }
#' 
#' @rdname selectResponse
#' @keywords internal
`selectResponse` <-  function(object, ...) UseMethod("selectResponse")

## ** selectResponse.formula
#' @rdname selectResponse
#' @method selectResponse formula
selectResponse.formula <- function(object, format = "call", ...){
  
  match.arg(format, c("call","vars"))
  
  if(length(object)==3){
    res <- object[[2]]
    if(format == "vars"){
      res <- all.vars(res)
    }
  }else{
    res <- NULL
  }
  
  return(res)
}

## ** selectRegressor (Documentation)
#' @title Regressor of a Formula.
#' @description Return the regressor variables contained in the formula
#' @name selectRegressor
#' 
#' @param object a formula
#' @param format [character] should an object of format call be returned (\code{format = "call"}),
#' or the names of the variables (\code{format = "vars"})
#' @param ... [internal] Only used by the generic method.
#'
#' 
#' @examples
#'
#' \dontrun{
#'
#' selectRegressor <- lavaSearch2:::selectRegressor
#' selectRegressor.formula <- lavaSearch2:::selectRegressor.formula
#' 
#' selectRegressor(Y1~X1+X2)
#' selectRegressor(Y1~X1+X2, format = "vars")
#' 
#' selectRegressor(Y1~X1+Y1)
#' selectRegressor(Y1+Y2~X1+Y1, format = "vars")
#' 
#' selectRegressor(~X1+X2)
#' selectRegressor(~X1+X2, format = "vars")
#' 
#' }
#' @rdname selectRegressor
#' @keywords internal
`selectRegressor` <-  function(object, ...) UseMethod("selectRegressor")

## ** selectRegressor.formula
#' @rdname selectRegressor
#' @method selectRegressor formula
selectRegressor.formula <- function(object, format = "call", ...){
  
  match.arg(format, c("call","vars"))
  
  if(length(object)==3){
    res <- object[[3]]
    
  }else if(length(object)==2){
    res <- object[[2]]
  }else{
    res <- NULL
  }
  if(format == "vars"){
    res <- all.vars(res)
  }
  
  return(res)
}

## ** combineFormula
#' @title Combine formula
#' @description Combine formula by outcome
#' 
#' @param ls.formula a list of formula
#' @param as.formula should a list of formula be returned. Otherwise it will be a list of characters.
#' @param as.unique should regressors appears at most once in the formula
#' 
#' @examples
#' combineFormula(list(Y~X1,Y~X3+X5,Y1~X2))
#' lava.options(symbols = c("~",","))
#' combineFormula(list("Y~X1","Y~X3+X5","Y1~X2"))
#' lava.options(symbols = c("<-","<->"))
#' combineFormula(list("Y<-X1","Y<-X3+X5","Y1<-X2"))
#' 
#' combineFormula(list(Y~X1,Y~X3+X1,Y1~X2))
#' combineFormula(list(Y~X1,Y~X3+X1,Y1~X2), as.formula = FALSE)
#' combineFormula(list(Y~X1,Y~X3+X1,Y1~X2), as.unique = TRUE)
#' 
#' lava.options(symbols = c("~","~~"))
#' combineFormula(list("Y~X1","Y~X3","Y1~X2"))
#' 
#' @export
combineFormula <- function(ls.formula, as.formula = TRUE, as.unique = FALSE){
  
  if(length(ls.formula)==0){return(NULL)}
  ls.Vars <- initVarLinks(ls.formula, format = "list")
  
  ls.endogeneous <- ls.Vars$var1
  ls.X <- ls.Vars$var2
  endogenous <- unique(ls.endogeneous)
  n.endogeneous <- length(endogenous)
  
  ls.formula2 <- vector(n.endogeneous, mode = "list")
  for(iterE in 1:n.endogeneous){
    X <- unlist(ls.X[which(ls.endogeneous==endogenous[iterE])])
    if(as.unique){X <- unique(X)}
    txt <- paste(endogenous[iterE],"~",paste(X, collapse = " + "))
    if(as.formula){ls.formula2[[iterE]] <- stats::as.formula(txt)}else{ls.formula2[[iterE]] <- txt}
  }
  
  return(ls.formula2)
}



## ** formula2character
#' @title formula character conversion
#' @description Conversion of formula into character string or vice versa
#' @name convFormulaCharacter
#' 
#' @param f a formula.
#' @param type should the normal formula operator be used (\code{"formula"}) or the one of lava.option (\code{"symbols"} or \code{"symbol"}).
#' 
#' @examples
#' formula2character(Y1~X1+X2)
#' formula2character(Y1~X1+X2, type = "symbols")

#' @rdname convFormulaCharacter
#' @export
formula2character <- function(f, type = "formula"){
  
  match.arg(type, choices = c("formula", "symbols"))
  
  if(type == "formula"){
    txt <- paste(deparse(f), collapse = "+")
  }else {
    txt <- as.character(f)
    txt[1] <- lava.options()[[type]][1]
    txt <- paste(txt[2],txt[1],txt[3], sep = "")
  }
  
  return(gsub("[[:blank:]]","",txt))
  
}

## * Miscellaneous
## ** .allPermutations
## .allPermutations(1:3)
## .allPermutations(2:3)
.allPermutations <- function(vec){
    X <- lapply(vec, function(x){
        cbind(x, .allPermutations(setdiff(vec, x)))
    })
    return(unname(do.call(rbind,X)))
}
## ** .combination
#' @title Form all Unique Combinations Between two Vectors
#' @description Form all unique combinations between two vectors (removing symmetric combinations).
#' @name combination
#'
#' @param ... [vectors] elements to be combined.
#' @param levels [logical] should a label for each combination be output as an attribute named levels.
#'
#' @return A matrix, each row being a different combination.
#' 
#' @examples
#' .combination <- lavaSearch2:::.combination
#' 
#' .combination(1,1)
#' .combination(1:2,1:2)
#' .combination(c(1:2,1:2),1:2)
#' 
#' .combination(alpha = 1:2, beta = 3:4)
#' .combination(alpha = 1:2, beta = 3:4, gamma = 1:4)
#' .combination(alpha = 1:3, beta = 1:3, gamma = 1:3)
#'
#' @keywords internal
.combination <- function(..., levels = FALSE){

    ## ** normalize arguments
    dots <- list(...)
    test.null <- unlist(lapply(dots,is.null))    
    if(any(test.null)){
        return(NULL)
    }
    dots <- lapply(dots,unique)

    ## ** form all combinations
    grid <- expand.grid(dots, stringsAsFactors = FALSE) 
    
    ## ** remove identical combinations after permutations of the columns
    flatGrid <- apply(grid,1,function(iX){paste0(sort(iX),collapse = "")})    
    grid <- grid[!duplicated(flatGrid),]
    rownames(grid) <- NULL
    attr(grid,"levels") <- unname(flatGrid[!duplicated(flatGrid)])
    
    ## ** export
    return(grid)        
}

######################################################################
### Utils.R ends here

## * .onLoad
.onLoad <- function(lib, pkg="lavaSearch2") {

    lava::lava.options(search.calc.quantile.int = FALSE, ## hidden argument for modelsearch2
                       search.type.information = "E", ## hidden argument for modelsearch2
                       ## search.perm.stat = "exact", ## hidden argument for modelsearch2 (otherwise "exact")
                       method.estimate2 = "ols", ## hidden argument for sCorrect
                       factor.dRvcov = 1/2,
                       ssc = "residuals",
                       df = "satterthwaite",
                       df.robust = 1
                       )

    lava::addhook("gaussian_weight.estimate.hook", hook = "estimate.hooks")

}

## * .onAttach
.onAttach <- function(lib, pkg="lavaSearch2") {
    desc <- utils::packageDescription(pkg)
    packageStartupMessage(desc$Package, " version ",desc$Version)
}

lava_categorical2dummy <- get("categorical2dummy", envir = asNamespace("lava"), inherits = FALSE)
lava_estimate.lvm <- get("estimate.lvm", envir = asNamespace("lava"), inherits = FALSE)
lava_procdata.lvm <- get("procdata.lvm", envir = asNamespace("lava"), inherits = FALSE)

## * findNewLink
## ** doc findNewLink
#' @title Find all New Links Between Variables
#' @description Find all new links between variables (adapted from lava::modelsearch).
#' 
#' @name findNewLink
#' 
#' @param object a \code{lvm} object.
#' @param data [optional] a dataset used to identify the categorical variables when not specified in the \code{lvm} object.
#' @param exclude.var [character vector] all links related to these variables will be ignore.
#' @param type [character vector] the type of links to be considered: \code{"regression"}, \code{"covariance"}, or \code{"both"}, .
#' @param rm.latent_latent [logical] should the links relating two latent variables be ignored?
#' @param rm.endo_endo [logical] should the links relating two endogenous variables be ignored?
#' @param rm.latent_endo [logical] should the links relating one endogenous variable and one latent variable be ignored?
#' @param output [character] Specify \code{"names"} to return the names of the variables to link
#' or specify \code{"index"} to return their position.
#' @param ... [internal] only used by the generic method.
#'
#' @return A list containing:
#' \itemize{
#' \item M.links: a matrix with two columns indicating (by name or position) the exogenous and endogenous variable corresponding to each link.
#' \item links: the name of the additional possible links
#' \item directional: a logical vector indicating for each link whether the link is unidirectional (\code{TRUE}, i.e. regression link)
#' or bidirectional (\code{FALSE}, i.e. covariance link).
#' }
#' 
#' @examples
#' library(lava)
#' 
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' categorical(m,labels=c("M","F","MF")) <- ~X1
#' findNewLink(m, rm.endo = FALSE)
#' findNewLink(m, rm.endo = TRUE)
#' findNewLink(m, exclude.var = "X1")
#' 
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' 
#' findNewLink(m, rm.endo = FALSE)
#' findNewLink(m, rm.endo = TRUE)
#' findNewLink(m, rm.endo = TRUE, output = "index")
#' findNewLink(m, type = "covariance")
#' findNewLink(m, type = "regression")
#' 
#' @concept modelsearch
#' @export
`findNewLink` <-
  function(object, ...) UseMethod("findNewLink")

## ** method findNewLink.lvm
#' @export
#' @rdname findNewLink
findNewLink.lvm <- function(object,                            
                            data = NULL,
                            type = "both",
                            exclude.var = NULL,
                            rm.latent_latent= FALSE,
                            rm.endo_endo= FALSE,
                            rm.latent_endo= FALSE,
                            output = "names",
                            ...){

    match.arg(output, choices = c("names","index"))
    match.arg(type, choices = c("both","covariance","regression"))
    if(is.null(data)){        
        data <- lava::sim(object, n = 1)
    }
    
    ## *** convertion to dummy variable name for categorical variables
    xF <- lava_categorical2dummy(object, data)
    AP <- with(lava::index(xF$x), A + t(A) + P)
    latent.xF <-  latent(xF$x)
    endogenous.xF <- endogenous(xF$x)
    exogenous.xF <- exogenous(xF$x)

    if(!is.null(exclude.var)){
        exclude.var <- var2dummy(xF, exclude.var)
    }
    
    if( any(exclude.var %in% colnames(AP) == FALSE) ){
        wrong.var <- exclude.var[exclude.var %in% colnames(AP) == FALSE]
        stop("unknown variable to exclude \n",
             "variable(s): \"",paste(wrong.var, collapse = "\" \""),"\"\n")
    }

    ## *** loop over links
    restricted <- c()
    directional <- c()
    for (i in seq_len(ncol(AP) - 1)){
        for (j in seq(i + 1, nrow(AP))){

            var.i <- rownames(AP)[i]
            var.j <- rownames(AP)[j]
            
            if(!is.null(exclude.var) && (var.i %in% exclude.var || var.j %in% exclude.var)){
                next
            }

            isLatent.i <- var.i %in% latent.xF
            isLatent.j <- var.j %in% latent.xF
            isEndogenous.i <- var.i %in% endogenous.xF
            isEndogenous.j <- var.j %in% endogenous.xF
            isExogenous.i <- var.i %in% exogenous.xF
            isExogenous.j <- var.j %in% exogenous.xF

            if(rm.latent_latent && isLatent.i && isLatent.j){
                next
            }
            if(rm.endo_endo && isEndogenous.i && isEndogenous.j){
                next
            }
            if(isExogenous.i && isExogenous.j){
                next
            }
            if(rm.latent_endo && ( (isLatent.i && isEndogenous.j) || (isEndogenous.i && isLatent.j) )){
                next
            }
            iDirectional <- (isExogenous.i+isExogenous.j)>0
            if(type == "regression" && iDirectional == FALSE){
                next
            }
            if(type == "covariance" && iDirectional == TRUE){
                next
            }
            
            if (AP[j, i] == 0){
                restricted <- rbind(restricted, c(i, j))
                directional <- c(directional, iDirectional)
            }
        }
    }
    ## *** export  
    out <- list(M.links = restricted,
                links = NULL,
                directional = directional)

    if(!is.null(restricted)){
        M.names <- cbind(rownames(AP)[restricted[,1]],
                         colnames(AP)[restricted[,2]])
        out$links <- paste0(M.names[,1], lava.options()$symbols[2-directional],M.names[,2])
        if(output == "names"){
            out$M.links <- M.names
        }
    }
    
    return(out)  
}

## * addLink
## ** doc addLink
#' @title Add a New Link Between Two Variables in a LVM
#' @rdname addLink
#' @description Generic interface to add links to \code{lvm} objects.
#' 
#' @param object a \code{lvm} object.
#' @param var1 [character or formula] the exogenous variable of the new link or a formula describing the link to be added to the lvm.
#' @param var2 [character] the endogenous variable of the new link. Disregarded if the argument \code{var1} is a formula.
#' @param all.vars [internal] a character vector containing all the variables of the \code{lvm} object.
#' @param covariance [logical] is the link is bidirectional? Ignored if one of the variables non-stochastic (e.g. exogenous variables).
#' @param warnings [logical] Should a warning be displayed when no link is added?
#' @param ... [internal] only used by the generic method and from \code{addLink.lvm.reduced} to \code{addLink.lvm}.
#'
#' @details
#' The argument \code{all.vars} is useful for \code{lvm.reduce} object where the command \code{vars(object)} does not return all variables. The command \code{vars(object, xlp = TRUE)} must be used instead.
#'
#' Arguments \code{var1} and \code{var2} are passed to \code{initVarlink}.
#'
#' @examples
#' library(lava)
#' set.seed(10)
#' 
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' m2 <- m
#' 
#' addLink(m, x1 ~ y1, covariance = FALSE)
#' addLink(m, y1 ~ x1, covariance = FALSE)
#' coef(addLink(m, y1 ~ y2, covariance = TRUE))
#' 
#' addLink(m2, "x1", "y1", covariance = FALSE)
#' addLink(m2, "y1", "x1", covariance = FALSE)
#' newM <- addLink(m, "y1", "y2", covariance = TRUE)
#' coef(newM)
#'
#' @concept setter
#' @export
`addLink` <-
    function(object, ...) UseMethod("addLink")

## ** method addLink.lvm
#' @export
#' @rdname addLink
addLink.lvm <- function(object,
                        var1,
                        var2,
                        covariance,
                        all.vars = lava::vars(object),
                        warnings = FALSE,
                        ...){

    res <- initVarLink(var1, var2, format = "list")
    var1 <- res$var1
    var2 <- res$var2
    endogenous.object <- endogenous(object)
    exogenous.object <- exogenous(object)
    latent.object <- latent(object)
    
    if(var1 %in% all.vars == FALSE){
        if(warnings){
            warning("addLink.lvm: var1 does not match any variable in object, no link is added \n",
                    "var1: ",var1,"\n")
        }
    }
    
    ####
    if(is.na(var2)){
        
        intercept(object) <- stats::as.formula(paste0("~", var1))
        
    }else{
        
        if(var1 == var2){
            if(warnings){
                warning("addLink.lvm: var1 equals var2, no link is added \n",
                        "var1/2: ",var1,"\n")
            }
        }
        
        
        ## if(var2 %in% all.vars == FALSE){
        ##     if(warnings){
        ##         warning("addLink.lvm: var2 does not match any variable in object, no link is added \n",
        ##                 "var2: ",var2,"\n")
        ##     }
        ## }
        if(var1 %in% endogenous.object && var2 %in% endogenous.object){
            if(missing(covariance)){
                covariance <- TRUE
            }else if(covariance == FALSE){
                stop("addLink.lvm: cannot add a link between two endogenous variable when argument \'covariance\' is FALSE \n")
            }                
        }
    
       
        if(covariance){
            covariance(object) <- stats::as.formula(paste(var1, var2, sep = "~"))  
        }else if(var1 %in% endogenous.object || var2 %in% exogenous.object){
            regression(object) <- stats::as.formula(paste(var1, var2,  sep = "~"))
        }else if(var2 %in% endogenous.object || var1 %in% exogenous.object){
            regression(object) <- stats::as.formula(paste(var2, var1, sep = "~"))
        }else {
            if(var1 %in% latent.object){
                regression(object) <- stats::as.formula(paste(var1, var2, sep = "~"))  
            }else if(var2 %in% latent.object){
                regression(object) <- stats::as.formula(paste(var2, var1, sep = "~"))  
            }else{
                stop("unknow configuration \n")
            }
            
        }
    }
    
    return(object)
}

## ** method addLink.lvm.reduced
#' @export
#' @rdname addLink
addLink.lvm.reduced <- function(object, ...){
  return(addLink.lvm(object, all.vars = lava::vars(object, lp = FALSE, xlp = TRUE) , ...))
}

## * setLink
## ** Documentation - setLink
#' @title Set a Link to a Value
#' @name setLink
#' @description Generic interface to set a value to a link in a \code{lvm} object.
#' 
#' @param object a \code{lvm} object.
#' @param var1 [character or formula] the exogenous variable of the new link or a formula describing the link to be added to the lvm.
#' @param var2 [character] the endogenous variable of the new link. Disregarded if the argument \code{var1} is a formula.
#' @param value [numeric] the value at which the link should be set.
#' @param warnings [logical] should a warning be displayed if the link is not found in the \code{lvm} object.
#' @param ... [internal] only used by the generic method.
#'  
#' @examples
#' library(lava)
#' set.seed(10)
#' 
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' covariance(m) <- y1 ~ y2
#' 
#' m1 <- setLink(m, y3 ~ u, value = 1)
#' estimate(m1, lava::sim(m,1e2))
#' # m1 <- setLink(m, u ~ y3, value = 1)
#' 
#' m2 <- setLink(m, y1 ~ y2, value = 0.5)
#' estimate(m2, lava::sim(m,1e2))
#'
#' @concept setter
#' @export
`setLink` <-
  function(object, ...) UseMethod("setLink")

## ** method setLink.lvm
#' @rdname setLink
#' @export
setLink.lvm <- function(object, var1, var2, value, warnings = FALSE, ...){

    res <- initVarLink(var1, var2)
    var1 <- res$var1
    var2 <- res$var2
    object.coef <- stats::coef(object)
    
  #### set the link
  if(is.na(var2)){
    intercept(object, stats::as.formula(paste0("~",var1))) <- value
  }else if(paste(var1, var2, sep = "~") %in% object.coef){
    regression(object, stats::as.formula(paste(var1,var2, sep = "~"))) <- value
  }else if(paste(var1,var2, sep = ",") %in% object.coef){
    covariance(object, stats::as.formula(paste(var1,var2, sep = "~"))) <- value
  }else if(paste(var2,var1, sep = ",") %in% object.coef){
    covariance(object, stats::as.formula(paste(var1,var2, sep = "~"))) <- value
  }else{
    if(warnings){
      warning("setLink.lvm: no link was found from var1 to var2, no link is set \n",
              "var1: ",var1,"\n",
              "var2: ",var2,"\n")
    }
  }
  
  return(object)
}

### sCorrect-getVarCov2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 18 2019 (11:00) 
## Version: 
## Last-Updated: jan 23 2024 (10:25) 
##           By: Brice Ozenne
##     Update #: 85
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * getVarCov2
#' @title Residual Variance-Covariance Matrix With Small Sample Correction.
#' @description Reconstruct the residual variance-covariance matrix from a latent variable model. 
#' It is similar to \code{nlme::getVarCov} but with small sample correction.
#' @name getVarCov2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#' 
#' @return A matrix with as many rows and column as the number of endogenous variables
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the residuals.
#' 
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' n <- 101
#'
#' Y1 <- rnorm(n, mean = 0)
#' Y2 <- rnorm(n, mean = 0.3)
#' Id <- findInterval(runif(n), seq(0.1,1,0.1))
#' data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
#'            data.frame(Y=Y2,G="2",Id = Id)
#'            )
#'
#' #### latent variable models ####
#' library(lava)
#' e.lvm <- estimate(lvm(Y ~ G), data = data.df)
#' getVarCov2(e.lvm)
#' 
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`getVarCov2` <-
    function(object, ...) UseMethod("getVarCov2")

## * getVarCov2.lvmfit
#' @rdname getVarCov2
#' @export
getVarCov2.lvmfit <- function(object, ssc = lava.options()$ssc, ...){

    return(getVarCov2(estimate2(object, ssc = ssc, ...)))

}

## * getVarCov2.lvmfit2
#' @rdname getVarCov2
#' @export
getVarCov2.lvmfit2 <- function(object, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }
    

    Omega <- object$sCorrect$moment$Omega
    attr(Omega, "detail") <- NULL
    return(Omega)
    
}

######################################################################
### sCorrect-getVarCov2.R ends here

### summary.calibrateType1.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: apr 23 2018 (12:58) 
## Version: 
## Last-Updated: Jan 11 2022 (16:48) 
##           By: Brice Ozenne
##     Update #: 72
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * method summary.modelsearch2
#' @title Display the Type 1 Error Rate 
#' @description Display the type 1 error rate from the simulation results.
#'
#' @param object output of the \code{calibrateType1} function.
#' @param alpha [numeric, 0-1] the confidence levels.
#' @param robust [character] should the results be displayed for both model-based and robust standard errors (\code{TRUE}),
#' only  model-based standard error (\code{FALSE}), or only robust standard error (\code{"only"})?
#' @param type [character] should the type 1 error rate be diplayed (\code{"type1error"}) or the bias (\code{"bias")}.
#' @param digits [integer >0] the number of decimal places to use when displaying the summary.
#' @param log.transform [logical] should the confidence intervals be computed on the logit scale.
#' @param print should the summary be printed in the terminal.
#' @param ... [internal] only used by the generic method.
#' 
#' @method summary calibrateType1
#' @export
summary.calibrateType1 <- function(object, robust = FALSE, type = "type1error",
                                   alpha = 0.05, log.transform = TRUE,
                                   digits = 5, print = TRUE, ...){


    type <- match.arg(type, c("type1error","bias"))
    if(is.logical(robust)){robust <- as.character(robust)}
    robust <- match.arg(robust, c("TRUE","FALSE","only"))
    
    ## ** compute bias
    if(type == "bias"){
        stop("not implemented yet!")
    }
    
    ## ** compute type 1 error
    if(type == "type1error"){
        dfLong <- melt(object$p.value,
                       measure.vars = grep("^p.",names(object$p.value),value = TRUE),
                       value.name = "p.value",
                       variable.name = "method")
        dfS <- stats::aggregate(dfLong$p.value,
                                by = list(n = dfLong$n, method = dfLong$method, link = dfLong$link),
                                FUN = function(x){c(n.rep = length(x), type1error = mean(x<=alpha, na.rm = TRUE))},
                                simplify = FALSE)
        dfS <- cbind(dfS[,c("n","method","link")],
                     do.call(rbind,dfS[,"x"]))

        ## binom::binom.confint(x = round(dfS$type1error*dfS$n.rep), dfS$n.rep, method = "logit")
        
        logit.p <- log(dfS$type1error/(1-dfS$type1error))

        if(log.transform){
            ## delta method: ln(x/(1-x))' = ln(x)' - ln(1-x)' = 1/x + 1/(1-x) = 1/(x(1-x))
            var.logit.p <- 1/(dfS$n.rep*dfS$type1error*(1-dfS$type1error))

            ## on logit scale
            dfS$ci.inf <- logit.p + stats::qnorm(0.025) * sqrt(var.logit.p)
            dfS$ci.sup <- logit.p + stats::qnorm(0.975) * sqrt(var.logit.p)

            ## on original scale
            dfS$ci.inf <- 1/(1+exp(-dfS$ci.inf))
            dfS$ci.sup <- 1/(1+exp(-dfS$ci.sup))
        }else{
            dfS$ci.inf <- dfS$type1error + stats::qnorm(0.025) * sqrt(dfS$type1error*(1-dfS$type1error)/dfS$n.rep)
            dfS$ci.sup <- dfS$type1error + stats::qnorm(0.975) * sqrt(dfS$type1error*(1-dfS$type1error)/dfS$n.rep)
        }        

        ## names
        rownames(dfS) <- NULL
        dfS$correction <- sapply(as.character(dfS$method),switch,
                                 p.Ztest = "Gaussian approx.",
                                 p.Satt = "Satterthwaite approx.",
                                 p.SSC = "small sample correction",
                                 p.KR = "Satterthwaite approx. with small sample correction",
                                 p.robustZtest = "Gaussian approx.",
                                 p.robustSatt = "Satterthwaite approx.",
                                 p.robustSSC = "small sample correction",
                                 p.robustKR = "Satterthwaite approx. with small sample correction")
        
        dfS$statistic <- sapply(as.character(dfS$method),switch,
                                p.Ztest = "Wald",
                                p.Satt = "Wald",
                                p.SSC = "Wald",
                                p.KR = "Wald",
                                p.robustZtest = "robust Wald",
                                p.robustSatt = "robust Wald",
                                p.robustSSC = "robust Wald",
                                p.robustKR = "robust Wald")

        ## display
        if(print){
            seqN <- unique(dfS$n)
            seqRep <- stats::setNames(dfS$n.rep[duplicated(dfS$n) == FALSE],seqN)

            vecTrans <- c("Gaus" = "Gaussian approx.",
                          "Satt" = "Satterthwaite approx.",
                          "SSC" = "small sample correction",
                          "SSC + Satt" = "Satterthwaite approx. with small sample correction")

            ls.print <- lapply(seqN, function(iN){ # iN <- 73
                df.tempo <- dfS[dfS$n==iN, c("link","statistic","correction","type1error","ci.inf","ci.sup")]
                if(!is.null(digits)){
                    df.tempo$type1error <- round(df.tempo$type1error, digits = digits)
                    df.tempo$ci.inf <- round(df.tempo$ci.inf, digits = digits)
                    df.tempo$ci.sup <- round(df.tempo$ci.sup, digits = digits)
                }
                df.tempo$CI <- paste0("[",df.tempo$ci.inf," ; ",df.tempo$ci.sup,"]")
                df.tempo$ci.inf <- NULL
                df.tempo$ci.sup <- NULL

                if(robust == "FALSE"){
                    df.tempo <- df.tempo[df.tempo$statistic=="Wald",,drop=FALSE]
                }else if(robust == "only"){
                    df.tempo <- df.tempo[df.tempo$statistic=="robust Wald",,drop=FALSE]
                }
                
                df.tempo <- df.tempo[order(df.tempo$link,df.tempo$statistic),,drop=FALSE]
                df.tempo$statistic[duplicated(cbind(df.tempo$link,df.tempo$statistic))] <- ""
                df.tempo$link[duplicated(df.tempo$link)] <- ""

                rownames(df.tempo) <- NULL
                df.tempo$correction <- factor(df.tempo$correction,
                                              levels = as.character(vecTrans),
                                              labels = names(vecTrans))
                ## df.tempo$correction <- NULL
                
                return(df.tempo)
            })
            names(ls.print) <- as.character(seqN)
       
            cat("Estimated type 1 error rate [95% confidence interval] \n")
            lapply(seqN, function(iN){
                cat("  > sample size: ",iN," | number of simulations: ",seqRep[as.character(iN)],"\n",sep="")
                print(ls.print[[as.character(iN)]],row.names = FALSE)
            })
            cat("\n")
            cat("Corrections: Gaus = Gaussian approximation \n",
                "             SSC  = small sample correction \n",
                "             Satt = Satterthwaite approximation \n",sep="")
            
        }        
    }

    return(invisible(dfS))
}
######################################################################
### summary.calibrateType1.R ends here

## * Documentation
#' @title Extract Data From a Latent Variable Model
#' @description Extract data from a latent variable model.
#' @name extractData
#' 
#' @param object the fitted model.
#' @param design.matrix [logical] should the data be extracted after transformation (e.g. conversion of categorical variables to dummy variables)?
#' Otherwise the original data will be returned.
#' @param as.data.frame [logical] should the output be converted into a \code{data.frame} object?
#' @param rm.na [logical] should the lines containing missing values in the dataset be removed?
#' @param envir [environment] the environment from which to search the data.
#'
#' @return a dataset.
#' @concept extractor
#' @export
`extractData` <-
    function(object, design.matrix, as.data.frame, envir, rm.na){
        UseMethod("extractData", object)
    }

## * Example
#' @rdname extractData
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' n <- 101
#'
#' Y1 <- rnorm(n, mean = 0)
#' Y2 <- rnorm(n, mean = 0.3)
#' Id <- findInterval(runif(n), seq(0.1,1,0.1))
#' data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
#'            data.frame(Y=Y2,G="2",Id = Id)       
#'            )
#'
#' #### latent variable model ####
#' library(lava)
#' e.lvm <- estimate(lvm(Y ~ G), data = data.df)
#' extractData(e.lvm)
#' extractData(e.lvm, design.matrix = TRUE)
#' 


## * extractData.lvmfit
#' @export
extractData.lvmfit <- function(object, design.matrix = FALSE, as.data.frame = TRUE,
                               envir = environment(), rm.na = TRUE){
    ## ** check arguments
    if(!is.logical(design.matrix)){
        stop("Argument \'design.matrix\' must be of type logical")
    }
    if(!is.logical(as.data.frame)){
        stop("Argument \'as.data.frame\' must be of type logical")
    }

    ## ** extract data
    data <- object$data$model.frame
    if(!inherits(data, "data.frame")){
        data <- as.data.frame(data)
    }

    if(design.matrix){
        keep.cols <- intersect(c("(Intercept)",lava::vars(object)), names(data))
        data <- data[,keep.cols,drop=FALSE]
    }

    ## ** normalize data
    if(as.data.frame){
        data <- as.data.frame(data)        
    }

    ## ** remove missing values relative to the exogenous variables
    test.na <- rowSums(is.na(data[,lava::exogenous(object),drop=FALSE]))
    if(rm.na == TRUE && any(test.na>0)){ ## remove rows corresponding to missing values
        if(!inherits(object,"lvm.missing")){
            data <- data[setdiff(1:NROW(data),which(test.na>0)),,drop=FALSE]
        }else{
            test.na <- rowSums(is.na(data[,exogenous(object)])) > 0
            data <- data[setdiff(1:NROW(data),which(test.na>0)),,drop=FALSE]
            warnings("Missing values in the exogenous variables \n",
                     "May not extract the appropriate dataset \n")
        }
    }

    ## ** export
    return(data)
    
}

### methods-modelsearch2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: sep 22 2017 (16:43) 
## Version: 
## last-updated: feb 18 2019 (09:33) 
##           By: Brice Ozenne
##     Update #: 216
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * nStep
## ** documentation - nStep
#' @title Find the Number of Steps Performed During the Sequential Testing
#' @description Find the number of steps performed during the sequential testing.
#' @name nStep
#' 
#' @param object a \code{modelsearch2} object.
#' 
#' @return an integer.
#' 
#' @examples
#' \dontrun{
#' mSim <- lvm(Y~G+X1+X2)
#' addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
#' df.data <- lava::sim(mSim, 1e2)
#'
#' mBase <- lvm(Y~G)
#' addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
#' e.lvm <- estimate(mBase, data = df.data)
#' res <- modelsearch2(e.lvm, method.p.adjust = "holm")
#' nStep(res)
#' }
#' 
#' @concept modelsearch
#' @concept extractor
#' @export
`nStep` <-
  function(object) UseMethod("nStep")

## ** function - nStep
#' @rdname nStep
#' @export
nStep.modelsearch2 <- function(object){

    return(length(object$sequenceTest))
    
}

## * getStep
## ** documentation - getStep
#' @title Extract one Step From the Sequential Procedure
#' @description Extract one step from the sequential procedure.
#' @name getStep
#' 
#' @param object a \code{modelsearch2} object
#' @param step [integer >0] which test should be extracted?
#' @param slot [character] the element from the modelsearch2 object that should be extracted.
#'
#' @examples
#' \dontrun{
#' mSim <- lvm(Y~G+X1+X2)
#' addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
#' df.data <- lava::sim(mSim, 1e2)
#'
#' mBase <- lvm(Y~G)
#' addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
#' e.lvm <- estimate(mBase, data = df.data)
#' res <- modelsearch2(e.lvm, method.p.adjust = "holm")
#'
#' getStep(res)
#' getStep(res, slot = "sequenceTest")
#' getStep(res, step = 1)
#' }
#' @concept modelsearch
#' @concept extractor
#' @export
`getStep` <-
  function(object, step, slot) UseMethod("getStep")

## ** function - getStep
#' @rdname getStep
#' @export
getStep.modelsearch2 <- function(object, step = nStep(object), slot = NULL){
    
    ## ** normalize arguments
    lastStep <- nStep(object)
    if(any(step %in% 1:lastStep == FALSE)){
        stop("step must be an integer between 1 and ",lastStep,"\n")
    }
    if(!is.null(slot) && slot %in% names(object) == FALSE){
        stop("argument \'slot\' must be one of \"",paste(names(object),collapse="\" \""),"\"\n")
    }

    ## ** subset
    new.object <- object
    new.object$sequenceTest <- object$sequenceTest[step]
    new.object$sequenceModel <- object$sequenceModel[step]
    class(new.object) <- NULL

    ## ** export
    if(is.null(slot)){
        return(new.object)
    }else{
        if(is.list(new.object[[slot]])){
            return(new.object[[slot]][[1]])
        }else{
            return(new.object[[slot]])
        }
    }
}

## * getNewLink
## ** documentation - getNewLink
#' @title Extract the Links that Have Been Found by the modelsearch2.
#' @description Extract the links that have been found relevant by modelsearch2.
#' @name getNewLink
#' 
#' @param object a \code{modelsearch2} object.
#' @param step [logical] which test should be extracted?
#'
#' @return A character vector.
#' 
#' @examples
#' \dontrun{
#' mSim <- lvm(Y~G+X1+X2)
#' addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
#'
#' set.seed(10)
#' df.data <- lava::sim(mSim, 1e2)
#'
#' mBase <- lvm(Y~G)
#' addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
#' e.lvm <- estimate(mBase, data = df.data)
#' res <- modelsearch2(e.lvm, method.p.adjust = "holm")
#' getNewLink(res)
#' }
#' @concept modelsearch
#' @concept extractor
#' 
#' @export
`getNewLink` <-
  function(object, step) UseMethod("getNewLink")

## ** function - getNewLink
#' @rdname getNewLink
#' @export
getNewLink.modelsearch2 <- function(object, step = 1:nStep(object)){

    selected <- link <- NULL
    
    ## ** normalize arguments
    lastStep <- nStep(object)
    if(any(step %in% 1:lastStep == FALSE)){
        stop("step must be an integer between 1 and ",lastStep,"\n")
    }

    ## ** extract
    ls.link <- lapply(step, function(x){
        iStep <- getStep(object,step=x,slot="sequenceTest")
        return(subset(iStep, subset = selected == TRUE, select = "link", drop = TRUE))
    })

    return(unlist(ls.link))    
}

## * getNewModel
## ** documentation - getNewModel
#' @title Extract the Model that Has Been Retains by the modelsearch2.
#' @description Extract the model that has been retained by modelsearch2.
#' @name getNewModel
#' 
#' @param object a \code{modelsearch2} object.
#' @param step [integer >=0] the step at which the model should be extracted.
#' 0 returns the initial model, i.e. before adding any links.
#'
#' @return A \code{lvmfit} object.
#' 
#' @examples
#' \dontrun{
#' mSim <- lvm(Y~G+X1+X2)
#' addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
#'
#' set.seed(10)
#' df.data <- lava::sim(mSim, 1e2)
#'
#' mBase <- lvm(Y~G)
#' addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
#' e.lvm <- estimate(mBase, data = df.data)
#' res <- modelsearch2(e.lvm, method.p.adjust = "holm")
#' getNewModel(res)
#' }
#' @concept modelsearch
#' @concept extractor
#' 
#' @export
`getNewModel` <-
  function(object, step) UseMethod("getNewModel")

## ** function - getNewLink
#' @rdname getNewModel
#' @export
getNewModel.modelsearch2 <- function(object, step = nStep(object)){

    ## ** normalize arguments
    lastStep <- nStep(object)
    if(any(step %in% 0:lastStep == FALSE)){
        stop("Argument \'step\' must be an integer between 0 and ",lastStep,"\n")
    }
    if(length(step)!=1){
        stop("Argument \'step\' must have length one \n")
    }
    
    ## ** extract
    if(step>0){
        return(object$sequenceModel[[step]])
    }else{
        return(object$initialModel)
        }
}

## * autoplot
## ** documentation - autoplot
#' @title Display the Value of a Coefficient across the Steps.
#' @description Display the value of a coefficient across the steps.
#' @name autplot-modelsearch2
#' 
#' @param object a \code{modelsearch2} object.
#' @param param [character vector] the name of the coefficient(s) to be displayed.
#' @param ci [logical] should the confidence intervals of the coefficient(s) be displayed.
#' @param conf.level [numeric, 0-1] confidence level of the interval.
#' @param step [integer >0] the steps at which the coefficient value should be displayed.
#' @param plot [logical] should the graph be displayed?
#' @param add.0 [logical] should an horizontal line representing no effect be displayed?
#' @param ...  [internal] only used by the generic method.
#'
#' @return A list containing \itemize{
#' \item \code{plot}: a ggplot object.
#' \item \code{data}: the data used to generate the ggplot object.
#' }
#' 
#' @examples
#' \dontrun{
#' mSim <- lvm(Y~G+X1+X2+X3+X4+X5)
#' addvar(mSim) <- ~Z1+Z2
#'
#' set.seed(10)
#' df.data <- lava::sim(mSim, 1e2)
#'
#' mBase <- lvm(Y~G)
#' addvar(mBase) <- ~X1+X2+X3+X4+X5+Z1+Z2
#' e.lvm <- estimate(mBase, data = df.data)
#' res <- modelsearch2(e.lvm, method.p.adjust = "holm", alpha = 0.05)
#' autoplot(res, param = "Y~G")
#' autoplot(res, param = c("Y","Y~G"))
#' }
#' @concept modelsearch
#' @concept extractor
#' 

## ** code - autoplot
#' @rdname autoplot.modelsearch2
#' @method autoplot modelsearch2
#' @export
autoplot.modelsearch2 <- function(object, param, ci = TRUE, step = 0:nStep(object), conf.level = 0.95,
                                  plot = TRUE, add.0 = TRUE, ...){

    ## ** check arguments
    lastStep <- nStep(object)
    if(any(step %in% 0:lastStep == FALSE)){
        stop("Argument \'step\' must be an integer between 0 and ",lastStep,"\n")
    }
    if(any(diff(step)<=0)){
        stop("Argument \'step\' must be a vector of strictly increasing integers \n")
    }
    
    name.coef <- names(coef(object$initialModel))
    if(any(param %in% name.coef == FALSE)){
        stop("Argument \'param\' incorrect: some parameters do not belong to the initial model \n",
             "invalid parameters: \"",paste0(param[param %in% name.coef == FALSE], collapse ="\" \""),"\"\n")
    }

    ## ** get data
    n.points <- length(step)
    df.plot <- NULL

    for(iStep in 1:length(step)){
        iModel <- getNewModel(object, step = step[iStep])
        iCoef <- unname(cbind(coef(iModel)[param],stats::confint(iModel, level = conf.level)[param,,drop=FALSE]))
        colnames(iCoef) <- c("estimate", "ci.inf", "ci.sup")
        df.plot  <- rbind(df.plot,
                          data.frame(step = step[iStep], param  = param, iCoef)
                          )
    }

    ## ** graphical display
    gg <- ggplot2::ggplot(df.plot, aes_string(x = "step", y = "estimate"))
    if(add.0){
        gg <- gg + ggplot2::geom_hline(yintercept = 0, color = "red")
    }
    gg <- gg + ggplot2::geom_point() + ggplot2::geom_line()
    gg <- gg + ggplot2::facet_wrap(~param)
    if(ci){
        gg <- gg + ggplot2::geom_errorbar(aes_string(ymin = "ci.inf", ymax = "ci.sup"))        
        gg <- gg + ggplot2::ylab(paste0("estimate [",round(conf.level*100,2),"% confidence interval]"))
    }
    if(plot){
        print(gg)
    }

    ## ** export
    return(invisible(list(plot = gg,
                          data = df.plot)))

    

}

### symmetrize.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov  8 2017 (09:42) 
## Version: 
## Last-Updated: feb 19 2018 (09:33) 
##           By: Brice Ozenne
##     Update #: 19
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

#' @title Symmetrize a Matrix
#' @description Complete the upper (or lower) extra-diagonal terms  in order to obtain a symmetric matrix.
#' @param M a matrix.
#' @param update.upper [logical] should the upper extra diagonal terms be updated using the lower extra diagonal terms?
#' 
#' @examples
#' symmetrize <- lavaSearch2:::symmetrize
#'
#' ## example
#' M <- matrix(NA, 4, 4)
#' M[lower.tri(M)] <- 1:6
#'
#' symmetrize(M, update.upper = TRUE) # good
#'
#' M[upper.tri(M, diag = FALSE)] <- M[lower.tri(M, diag = FALSE)]
#' M # wrong
#' @keywords internal
symmetrize <- function(M, update.upper = TRUE){

    if(is.null(update.upper)){
        M <- (M+t(M))
        diag(M) <- diag(M)/2
    }else if(update.upper){
        M[upper.tri(M, diag = FALSE)] <- t(M)[upper.tri(M, diag = FALSE)]
    }else{
        M[lower.tri(M, diag = FALSE)] <- t(M)[lower.tri(M, diag = FALSE)]
    }
    return(M)    
}

##----------------------------------------------------------------------
### symmetrize.R ends here

### createGrid.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: aug 31 2017 (16:40) 
## Version: 
## last-updated: Jan 11 2022 (17:08) 
##           By: Brice Ozenne
##     Update #: 88
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:


## * Documentaiton - createGrid
#' @title Create a Mesh for the Integration
#' @description Create a mesh for the integration.
#' @name createGrid
#' 
#' @param n [integer >0] the number of points for the mesh in the x direction.
#' @param xmin [numeric] the minimal x value.
#' @param xmax [numeric] the maximal x value.
#' @param d.y [integer >0] the number of dimensions for the triangle.
#' @param d.z [integer >0] the number of dimensions.
#' @param zmax [numeric >0] the maximal z value (in absolute value).
#' @param fine [logical] should the mesh be displayed?
#' @param double [logical] should the grid be just outside the region of interest?
#' Otherwise it will be just inside.
#'
#' @details This create a mesh for integrating over a triangular surface using rectangles.
#' The domain is define by constrains on three types of variables:
#' \itemize{
#' \item the x variable: [unidimensional] that varies freely between [-xmax,-xmin] U [xmin,xmax].
#' \item the y variables: [dimension d.y] constrained to be lower in absolute value than the x variable.
#' In 2D this corresponds to 2 triangles, and in higher dimension to cones/hypercones.
#' \item the z variables: [dimension d.z] constrained to vary between [-zmax,zmax].
#' }
#' The intersection of these three conditions define the domain.
#'
#' The mesh is obtained slicing the triangles using rectangles.
#'
#' @return A list containing:
#' \itemize{
#' \item grid: a \code{data.frame} where each line corresponds to a point.
#' \item seqNames.min: a \code{character} giving the name of the x,y variables (min).
#' \item seqNames.max: a \code{character} giving the name of the x,y variables (max).
#' }
#' 
#' @examples
#' createGrid <- lavaSearch2:::createGrid
#' 
#' ## no z 
#' gridInt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
#'                          d.z = 0, fine = FALSE, double = FALSE)
#' gridExt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
#'                          d.z = 0, fine = FALSE, double = TRUE)
#' 
#' gridInt_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
#'                          d.z = 0, fine = FALSE, double = FALSE)
#' gridExt_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
#'                          d.z = 0, fine = FALSE, double = TRUE)
#' 
#' gridInt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
#'                          d.z = 0, fine = TRUE, double = FALSE)
#' 
#' ##  z
#' gridIntZ1_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
#'                            d.z = 1, zmax = 2, fine = FALSE, double = FALSE)
#' gridExtZ1_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
#'                            d.z = 1, zmax = 2, fine = FALSE, double = TRUE)
#'  
#' gridIntZ2_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
#'                            d.z = 2, zmax = 2, fine = FALSE, double = FALSE)
#' gridExtZ2_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
#'                            d.z = 2, zmax = 2, fine = FALSE, double = TRUE)
#'
#' @concept post-selection inference
#' @keywords internal

## * createGrid
#' @rdname createGrid
createGrid <- function(n,
                       xmin, xmax, d.y, 
                       d.z, zmax,
                       fine, double){

    y1.min <- y1.max <- NULL ## [:for CRAN check] subset
    
    ## ** find step along the x axis
    if(xmin[1]==0){
        by <- xmax/n
    }else{
        seqTry <- seq(0,xmax, length.out = n)
        by <- xmin/sum(seqTry<xmin)
    }
    seqPointsX <- seq(xmin,xmax, by = by)
    n.seqX <- length(seqPointsX)-1

    ## ** name variables
    y.minNames <- paste0("y",1:d.y,".min")
    y.maxNames <- paste0("y",1:d.y,".max")
    all.Names <- c("x.min","x.max",y.minNames,y.maxNames,"weight","index")

    seqNames.min <- c("x.min",y.minNames)
    seqNames.max <- c("x.max",y.maxNames)
    
    ## ** main grid
    grid.main <- NULL
    for(iX in 1:n.seqX){ #  iX <- 1
        grid.main <- rbind(grid.main,
                           c(seqPointsX[iX], seqPointsX[iX+1],
                             rep(-seqPointsX[iX+double],d.y), rep(seqPointsX[iX+double],d.y),
                             1, iX)
                           )
    }
    grid.main <- as.data.frame(grid.main)
    names(grid.main) <- all.Names
    
    ## gg.main <- ggplot(grid.main, aes(xmin = x.min, ymin = y1.min, xmax = x.max, ymax = y1.max, fill = index))
    ## gg.main <- gg.main + geom_rect()
    ## gg.main <- gg.main + geom_abline(slope = 1,color = "red") + geom_abline(slope = -1,color = "red")
    ## gg.main
    
    ## ** fine grid
    grid.fine <- NULL    
    if(fine){

        ls.mp <- lapply(1:d.y, function(x){c(-1,1)})
        grid.mp <- expand.grid(ls.mp)
        n.mp <- NROW(grid.mp)
        ls.index <- list(1:d.y, (d.y+1):(2*d.y))

        for(iX in 1:n.seqX){ # iX <- 1
            for(iMP in 1:n.mp){ # iMP <- 1

                index.neg <- which(grid.mp[iMP,]<0)
                ls.index2 <- ls.index
                for(i.neg in index.neg){
                    ls.index2[[1]][i.neg] <- ls.index[[2]][i.neg]
                    ls.index2[[2]][i.neg] <- ls.index[[1]][i.neg]    
                }
                M.seq <- cbind(rep(seqPointsX[iX],d.y)*grid.mp[iMP,],
                               rep(seqPointsX[iX+1],d.y)*grid.mp[iMP,])
                
                grid.fine <- rbind(grid.fine,
                                   c(seqPointsX[iX], seqPointsX[iX+1],
                                     as.numeric(M.seq[unlist(ls.index2)]),
                                     1/2, iX+1)
                                   )
            }
        }        
        grid.fine <- as.data.frame(grid.fine)
        names(grid.fine) <- all.Names        
        ## gg.fine <- ggplot(grid.fine, aes(xmin = x.min, ymin = y1.min, xmax = x.max, ymax = y1.max, fill = index))
        ## gg.fine <- gg.fine + geom_rect()
        ## gg.fine <- gg.fine + geom_abline(slope = 1,color = "red") + geom_abline(slope = -1,color = "red")
        ## gg.fine
    }

    ## ** Merge grids and remove empty cells
    grid.all <- rbind(grid.main, grid.fine)
    grid.all <- subset(grid.all, y1.min!=y1.max) # remove empty rectangles

    ## update index
    grid.all$index <- grid.all$index - min(grid.all$index) + 1
    ## ** duplicate grid  (negative x  and positive x)
    grid.all2 <- grid.all
    grid.all2$x.min <- -grid.all$x.max 
    grid.all2$x.max <- -grid.all$x.min # we need grid.all$x.min due to the previous line
    grid.all2$index <- grid.all2$index + max(grid.all2$index)
    grid <- rbind(grid.all,grid.all2)

    ## ** add z coordinate
    if(d.z>0){
        z.minNames <- paste0("z",1:d.z,".min")
        z.maxNames <- paste0("z",1:d.z,".max")

        grid[,z.minNames] <- -abs(zmax)
        grid[,z.maxNames] <- abs(zmax)
        seqNames.min <- c(seqNames.min, z.minNames)
        seqNames.max <- c(seqNames.max, z.maxNames)
    }

    return(list(grid = grid,
                seqNames.min = seqNames.min,
                seqNames.max = seqNames.max))
}

### sCorrect-summary.glht2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: maj  2 2018 (09:20) 
## Version: 
## Last-Updated: jan 24 2022 (11:11) 
##           By: Brice Ozenne
##     Update #: 228
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

#' @title Outcome of Linear Hypothesis Testing 
#' @description Estimates, p-values, and confidence intevals for linear hypothesis testing, possibly adjusted for multiple comparisons.
#' 
#' @param object a \code{glht2} object.
#' @param confint [logical] should confidence intervals be output
#' @param conf.level [numeric 0-1] level of the confidence intervals.
#' @param transform [function] function to backtransform the estimates, standard errors, null hypothesis, and the associated confidence intervals
#' (e.g. \code{exp} if the outcomes have been log-transformed).
#' @param seed [integer] value that will be set before adjustment for multiple comparisons to ensure reproducible results.
#' Can also be \code{NULL}: in such a case no seed is set.
#' @param rowname.rhs [logical] when naming the hypotheses, add the right-hand side (i.e. "X1-X2=0" instead of "X1-X2").
#' @param ... argument passed to \code{multcomp:::summary.glht}, e.g. argument \code{test} to choose the type of adjustment for multiple comparisons.
#' 

## * summary.glht2
#' @export
summary.glht2 <- function(object, confint = TRUE, conf.level = 0.95, transform = NULL, seed = NULL, rowname.rhs = TRUE, ...){
    if(!is.null(seed)){
        old.seed <- get0(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
        on.exit( assign(".Random.seed", old.seed, envir = .GlobalEnv, inherits = FALSE) )
        set.seed(seed)
    }
    
    keep.class <- class(object)
    object$test <- NULL
    object$confint <- NULL
    class(object) <- setdiff(keep.class, "glht2")
    keep.df <- object$df
    test.df <- any( (keep.df>0) * (!is.infinite(keep.df)) == 1 )
    object$df <- round(stats::median(object$df))
    output <- summary(object, ...)
    ## restaure df when possible
    method.adjust <- output$test$type
    if(NROW(object$linfct)==1){method.adjust <- "none"}
    if(test.df && method.adjust %in% c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none","univariate")){
        output$df <- keep.df
        output$test$pvalues <- stats::p.adjust(2*(1-stats::pt(abs(output$test$tstat), df = keep.df)), method = method.adjust)
    }
    
    name.hypo <- rownames(output$linfct)
    n.hypo <- length(name.hypo)
    if(confint && method.adjust %in% c("univariate","none","bonferroni","single-step")){
        if(method.adjust %in% c("none","univariate","bonferroni")){
            alpha <- switch(method.adjust,
                            "none" = 1-conf.level,
                            "univariate" = 1-conf.level,
                            "bonferroni" = (1-conf.level)/n.hypo)
            if(test.df){
                q <- stats::qt(1-alpha/2, df = output$df)
            }else{
                q <- stats::qnorm(1-alpha/2)
            }
            output$confint <- data.frame(matrix(NA, ncol = 3, nrow = n.hypo,
                                                dimnames = list(name.hypo, c("Estimate","lwr","upr"))))
            output$confint$Estimate <- as.double(output$test$coef)
            output$confint$lwr <- as.double(output$test$coef - q * output$test$sigma)
            output$confint$upr <- as.double(output$test$coef + q * output$test$sigma)
            ## range(confint(output, level = 1-alpha, calpha = univariate_calpha())$confint-output$confint)
        }else if(method.adjust == "single-step"){
            output <- confint(output, level = conf.level, calpha = multcomp::adjusted_calpha())
        }else{
            output$confint <- matrix(NA, nrow = n.hypo, ncol = 3,
                                     dimnames = list(name.hypo, c("Estimate","lwr","upr")))
        }
    }
    if(rowname.rhs){
        table2.rownames <- paste0(name.hypo, " == ", output$rhs)
    }else{
        table2.rownames <- name.hypo
    }
    output$table2 <- data.frame(matrix(NA, nrow = n.hypo, ncol = 7,
                                       dimnames = list(table2.rownames,
                                                       c("estimate","se","df","lower","upper","statistic","p.value"))
                                       ), stringsAsFactors = FALSE)
    output$table2$estimate <- output$test$coefficients
    output$table2$se <- output$test$sigma
    output$table2$df <- output$df
    output$table2$df[output$table2$df==0] <- Inf
    output$table2$lower <- output$confint[,"lwr"]
    output$table2$upper <- output$confint[,"upr"]
    output$table2$statistic <- output$test$tstat
    output$table2$p.value <- output$test$pvalues
    output$seed <- seed
    
    ## ** transformation
    output$transform <- transform
    output$table2 <- transformSummaryTable(output$table2,
                                           transform = transform)

    ## ** export    
    class(output) <- append(c("summary.glht2","summary.glht"),keep.class)
    return(output)
}

## * print.summary.glht2
#' @export
print.summary.glht2 <- function(x,
                                digits = max(3L, getOption("digits") - 2L),
                                digits.p.value = max(3L, getOption("digits") - 2L),
                                columns = c("estimate","se","df","lower","upper","statistic","p.value"),
                                ...){
    
    columns <- match.arg(columns, choices = c("estimate","se","df","lower","upper","statistic","p.value"), several.ok = TRUE)
    type <- x$type
    call <- if(isS4(x$model)){x$model@call}else{x$model$call}
    alternative <- x$alternativ
    type <- x$test$type
    txt.type <- switch(type,
                       "univariate" = "(CIs/p-values not adjusted for multiple comparisons)", 
                       "none" = "(CIs/p-values not adjusted for multiple comparisons)", 
                       "single-step" = paste0("(CIs/p-values adjusted for multiple comparisons -- single step max-test)"), 
                       "free" = paste0("(CIs/p-values adjusted for multiple comparisons -- step down max-test)"), 
                       "Westfall" = paste0("(CIs/p-values adjusted for multiple comparisons -- step down max-test with logical restrictions)"), 
                       paste0("(CIs/p-values adjusted for multiple comparisons -- ", type, " method)")
                       )
    txt.robust <- switch(as.character(x$robust),
                         "TRUE" = "Robust",
                         "FALSE" = "Model-based"
                         )

    ## txt.correction <- switch(as.character(x$ssc),
    ##                          "Cox" = " corrected for small sample bias (Cox correction)",
    ##                          "residuals" = " corrected for small sample bias (residual correction)",
    ##                          "NA" = ""
    ##                          )
    
    txt.alternative <- switch(alternative,
                              "less" = "one sided tests - inferiority",
                              "greater" = "one sided tests - superiority",
                              "two.sided" = "two sided tests")

    ## display
    cat("\n\t", "Simultaneous Tests for General Linear Hypotheses\n\n")
    if (!is.null(type)) {
        cat("Multiple Comparisons of Means (",txt.alternative,") \n\n", sep = "")
    }
    if (!is.null(call)) {
        cat("Fit: ")
        print(call)
        cat("Standard errors: ",txt.robust,"\n",sep="")
        cat("\n")
    }
    cat("Linear Hypotheses:\n")
    stats::printCoefmat(x$table2[,columns[columns %in% names(x$table2)],drop=FALSE], digits = digits,
                        has.Pvalue = "p.value" %in% columns,
                        P.values = "p.value" %in% columns,
                        eps.Pvalue = 10^{-digits.p.value})

    if(NROW(x$table2)>1){
        cat(txt.type,"\n")
    }
    error <- attr(x$test$pvalues,"error")
    if(!is.null(error) && error > 1e-12 && "p.value" %in% columns){
        txt.error <- paste0("Error when computing the adjusted p-value by numerical integration: ", signif(error, digits = digits))
        if(!is.null(x$seed)){
            txt.error <- paste0(txt.error," (seed ",x$seed,")")
        }
        cat(txt.error,"\n")
    }

    
    if(!is.null(x$global)){
        cat("\nGlobal test: p.value=",format.pval(x$global["p.value"], digits = digits, eps = 10^(-digits.p.value)),
            " (statistic=",round(x$global["statistic"], digits = digits),
            ", df=",round(x$global["df"], digits = digits),")\n",sep="")
    }
    ## if(nchar(txt.correction)>0){cat("(",txt.correction,")\n",sep="")}
    cat("\n")
    return(invisible(x))
}


######################################################################
### sCorrect-summary.glht2.R ends here

### residuals2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov  8 2017 (09:05) 
## Version: 
## Last-Updated: feb 11 2019 (13:25) 
##           By: Brice Ozenne
##     Update #: 935
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * documentation - residuals2
#' @title Extract Corrected Residuals
#' @description Extract correct residuals from a gaussian linear model.
#' @name residuals2
#' 
#' @param object a \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} object.
#' @param type [character] the type of residual to extract:
#' \code{"response"} for raw residuals,
#' \code{"studentized"} for studentized residuals,
#' \code{"normalized"} for normalized residuals.
#' @param param [named numeric vector] the fitted parameters.
#' @param data [data.frame] the data set.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#'
#' @details If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the residuals. \cr
#'
#' The raw residuals are defined by  observation minus the fitted value:
#' \deqn{
#' \varepsilon = (Y_1 - \mu_1, ..., Y_m - \mu_m)
#' }
#' The studentized residuals divided the raw residuals relative to each endogenous variable by the modeled variance of the endogenous variable.
#' \deqn{
#' \varepsilon_{stud} =(\frac{Y_1 - \mu_1}{\sigma_1}, ..., \frac{Y_m - \mu_m}{\sigma_m})
#' }
#' The normalized residuals multiply the raw residuals by the inverse of the square root of the modeled residual variance covariance matrix.
#' \deqn{
#' \varepsilon_{norm} = \varepsilon \Omega^{-1/2}
#' }
#' @return a matrix containing the residuals relative to each sample (in rows)
#' and each endogenous variable (in column).
#'
#' @examples
#' ## simulate data
#' set.seed(10)
#' m <- lvm(Y1~eta,Y2~eta,Y3~eta)
#' latent(m) <- ~eta
#' d <- lava::sim(m,20, latent = FALSE)
#'
#' ## standard linear model
#' e.lm <- lm(Y1~Y2, data = d)
#' sCorrect(e.lm) <- TRUE
#' 
#' sigma(e.lm)^2
#' mean(residuals(e.lm)^2)
#' mean(residuals2(e.lm)^2)
#' 
#' ## latent variable model
#' e.lvm <- estimate(m, data = d)
#' sCorrect(e.lvm) <- TRUE
#' mean(residuals2(e.lvm)^2)
#'
#' @concept small sample inference
#' @export
`residuals2` <-
    function(object, param, data, type) UseMethod("residuals2")

## * residuals2.lm2
#' @rdname residuals2
#' @export
residuals2.lm2 <- function(object, param = NULL, data = NULL, type = "response"){

    type <- match.arg(type, choices = c("response","studentized","normalized"), several.ok = FALSE)

    if(!is.null(param) || !is.null(data)){
        args <- object$sCorrect$args
        args$df <- FALSE
        object$sCorrect <- do.call(sCorrect,
                                   args = c(list(object, param = param, data = data),
                                            args))
    }
    if(type=="response"){
        residuals <- object$sCorrect$residuals
    }else if(type=="studentized"){
        residuals <- sweep(object$sCorrect$residuals,
                           STATS = sqrt(diag(object$sCorrect$Omega)),
                           FUN = "/",
                           MARGIN = 2)
        ## object$sCorrect$residuals/residuals
    }else if(type=="normalized"){
        residuals <- object$sCorrect$residuals %*% matrixPower(object$sCorrect$Omega, symmetric = TRUE, power = -1/2)
        colnames(residuals) <- colnames(object$sCorrect$residuals)
        ## object$sCorrect$residuals/residuals
        ## var(residuals)        
    }
    return(residuals)
}

## * residuals2.gls2
#' @rdname residuals2
#' @export
residuals2.gls2 <- residuals2.lm2

## * residuals2.lme2
#' @rdname residuals2
#' @export
residuals2.lme2 <- residuals2.lm2

## * residuals2.lvmfit2
#' @rdname residuals2
#' @export
residuals2.lvmfit2 <- residuals2.lm2


##----------------------------------------------------------------------
### residuals2.R ends here

### sampleRepeated.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 18 2019 (15:37) 
## Version: 
## Last-Updated: Jan 11 2022 (17:39) 
##           By: Brice Ozenne
##     Update #: 20
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Simulate Repeated Measurements over time
#' @description Simulate repeated measurements over time (one factor model).
#' @name sampleRepeated
#' 
#' @param n [integer] sample size.
#' @param n.Xcont [integer] number of continuous covariates acting on the latent variable.
#' @param n.Xcat [integer] number of categorical covariates acting on the latent variable.
#' @param n.rep [integer] number of measurement of the response variable.
#' @param format [character] should the dataset be returned in the \code{"long"} format or in the \code{"wide"} format.
#'
#' @return a \code{data.frame} object.
#'
#' @examples
#' 
#' sampleRepeated(10, format = "wide")
#' sampleRepeated(10, format = "long")

## * sampleRepeated
#' @rdname sampleRepeated
#' @export
sampleRepeated <- function(n, n.Xcont = 2, n.Xcat = 2, n.rep = 5, format = "long"){

    ## ** check arguments
    format <- match.arg(format, choices = c("wide","long"))
    
    ## ** define lvm
    m <- lava::lvm()
    idvars <- "id"
    distribution(m, ~id) <- function(n, ...){return(1:n)}    
    for(iY in 1:n.rep){ ## iY <- 1
        regression(m) <- stats::as.formula(paste0("Y",iY,"~eta"))
    }
    if(n.Xcont>0){
        for(iXcont in 1:n.Xcont){ ## iY <- 1
            regression(m) <- stats::as.formula(paste0("eta~X",iXcont))
        }
        idvars <- c(idvars,paste0("X",1:n.Xcont))
    }
    
    if(n.Xcat>0){
        for(iXcat in 1:n.Xcat){ ## iY <- 1
            regression(m) <- stats::as.formula(paste0("eta~Z",iXcat))
            categorical(m, labels=c("a","b","c")) <- paste0("Z",iXcat)
        }
        idvars <- c(idvars,paste0("Z",1:n.Xcat))
    }    
    latent(m) <- ~eta

    dW <- lava::sim(m, n = n, latent = FALSE)
    if(format == "wide"){
        return(dW)
    }else{
        dL <- stats::reshape(dW,
                             direction = "long",
                             idvar = idvars,
                             varying = list(paste0("Y",1:n.rep)),
                             v.names = "Y",
                             timevar = "time")
        rownames(dL) <- NULL
        return(dL)
    }
    
}

######################################################################
### sampleRepeated.R ends here

### mlf2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 29 2017 (12:56) 
## Version: 
## Last-Updated: feb 25 2019 (09:40) 
##           By: Brice Ozenne
##     Update #: 547
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## User 
### Code:


## * estfun
#' @title Extract Empirical Estimating Functions (lvmfit Object)
#' @description Extract the empirical estimating functions of a lvmfit object.
#' This function is for internal use but need to be public to enable its use by \code{multcomp::glht}.
#' @name estfun
#' 
#' @param x an \code{lvmfit} object.
#' @param ... arguments passed to methods.
#'
#' @details This function enables to use the \code{glht} function with lvmfit object.
#' Otherwise when calling \code{multcomp:::vcov.mmm} then \code{sandwich::sandwich} and then \code{sandwich::meat}, \code{sandwich::meat} will complain that \code{estfun} is not defined for \code{lvmfit} objects.
#'
#' @examples
#' library(multcomp)
#' 
#' #### generative model ####
#' mSim <- lvm(X ~ Age + 0.5*Treatment,
#'             Y ~ Gender + 0.25*Treatment,
#'             c(Z1,Z2,Z3) ~ eta, eta ~ 0.75*treatment,
#'             Age[40:5]~1)
#' latent(mSim) <- ~eta
#' categorical(mSim, labels = c("placebo","SSRI")) <- ~Treatment
#' categorical(mSim, labels = c("male","female")) <- ~Gender
#'
#' #### simulate data ####
#' n <- 5e1
#' set.seed(10)
#' df.data <- lava::sim(mSim, n = n, latent = FALSE)
#'
#' #### fit separate models ####
#' lmX <- lm(X ~ Age + Treatment, data = df.data)
#' lvmY <- estimate(lvm(Y ~ Gender + Treatment), data = df.data)
#' lvmZ <- estimate(lvm(c(Z1,Z2,Z3) ~ eta, eta ~ Treatment), 
#'                  data = df.data)
#'
#' #### create mmm object #### 
#' e.mmm <- mmm(X = lmX, Y = lvmY, Z = lvmZ)
#'
#' #### create contrast matrix ####
#' resC <- createContrast(e.mmm, var.test = "Treatment", add.variance = FALSE)
#'
#' #### adjust for multiple comparisons ####
#' e.glht <- glht(e.mmm, linfct = resC$mlf)
#' summary(e.glht)
#' @concept multiple comparison

## * estfun.lvmfit
#' @rdname estfun
#' @method estfun lvmfit
#' @export
estfun.lvmfit <- function(x, ...){
    U <- lava::score(x, indiv = TRUE)
    return(U)
}

## * estfun.gls
#' @rdname estfun
#' @method estfun gls
#' @export
estfun.gls <- function(x, ...){
    if(inherits(x,"gls2")){
        U <- score2(x)
    }else{
        U <- score2(x, bias.correct = FALSE)
    }
    return(U)
}

## * estfun.lme
#' @rdname estfun
#' @method estfun lme
#' @export
estfun.lme <- function(x, ...){
    if(inherits(x,"lme2")){
        U <- score2(x)
    }else{
        U <- score2(x, bias.correct = FALSE)
    }
    return(U)
}

## * Documentation - glht2
#' @title General Linear Hypothesis
#' @description Test general linear hypotheses and across latent variable models with small sample corrections.
#' @name glht2
#' 
#' @param model a \code{lvmfit} or \code{mmm} object.
#' The \code{mmm} object can only contain lm/gls/lme/lvmfit objects.
#' @param linfct [matrix or vector of character] the linear hypotheses to be tested. Same as the argument \code{par} of \code{\link{createContrast}}.
#' @param rhs [vector] the right hand side of the linear hypotheses to be tested.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias?
#' @param df [logical] should the degree of freedoms of the Wald statistic be computed using the Satterthwaite correction?
#' @param robust [logical] should robust standard error be used? 
#' Otherwise rescale the influence function with the standard error obtained from the information matrix.
#' @param cluster  [integer vector] the grouping variable relative to which the observations are iid.
#'
#' @details
#' Whenever the argument linfct is not a matrix, it is passed to the function \code{createContrast} to generate the contrast matrix and, if not specified, rhs. \cr \cr
#'
#' Since only one degree of freedom can be specify in a glht object and it must be an integer, the degree of freedom of the denominator of an F test simultaneously testing all hypotheses is retained, after rounding. \cr \cr
#'
#' Argument rhs and null are equivalent.
#' This redondance enable compatibility between \code{lava::compare}, \code{compare2}, \code{multcomp::glht}, and \code{glht2}.
#' @return A \code{glht} object.
#' 
#' @seealso
#' \code{\link{createContrast}} to create contrast matrices. \cr
#' \code{\link{sCorrect}} to pre-compute quantities for the small sample correction.
#' 
#' @concept multiple comparisons
#'
#' @examples
#' library(multcomp)
#' 
#' ## Simulate data
#' mSim <- lvm(c(Y1,Y2,Y3)~ beta * eta, Z1 ~ E, Z2 ~ E, Age[40:5]~1)
#' latent(mSim) <- "eta"
#' set.seed(10)
#' n <- 1e2
#'
#' df.data <- lava::sim(mSim, n, latent = FALSE, p = c(beta = 1))
#'
#' #### Inference on a single model ####
#' e.lvm <- estimate(lvm(Y1~E), data = df.data)
#' summary(glht2(e.lvm, linfct = c("Y1~E + Y1","Y1")))
#' 
#' #### Inference on separate models ####
#' ## fit separate models
#' lmX <- lm(Z1 ~ E, data = df.data)
#' lvmY <- estimate(lvm(Z2 ~ E + Age), data = df.data)
#' lvmZ <- estimate(lvm(c(Y1,Y2,Y3) ~ eta, eta ~ E), 
#'                  data = df.data)
#'
#' #### create mmm object #### 
#' e.mmm <- mmm(X = lmX, Y = lvmY, Z = lvmZ)
#'
#' #### create contrast matrix ####
#' resC <- createContrast(e.mmm, var.test = "E", add.variance = TRUE)
#'
#' #### adjust for multiple comparisons ####
#' e.glht2 <- glht2(e.mmm, linfct = resC$contrast, df = FALSE)
#' summary(e.glht2)
#'
#' @concept multiple comparison
#' @export
`glht2` <-
    function(model, linfct, rhs,
             bias.correct, df, robust, cluster) UseMethod("glht2")


## * glht2.lvmfit
#' @rdname glht2
#' @export
glht2.lvmfit <- function(model, linfct, rhs = 0,
                         bias.correct = TRUE, df = TRUE, robust = FALSE, cluster = NULL){

    if(robust==FALSE && !is.null(cluster)){
        stop("Argument \'cluster\' must be NULL when argument \'robust\' is FALSE \n")
    }
    
    ### ** define contrast matrix
    if(!is.matrix(linfct)){
        resC <- createContrast(model, par = linfct)
        linfct <- resC$contrast
        if("rhs" %in% names(match.call()) == FALSE){
            rhs <- resC$null
        }
    }

### ** pre-compute quantities for the small sample correction
    if(!inherits(model,"lvmfit2")){
        sCorrect(model, df = df) <- bias.correct
    }

### ** Wald test with small sample correction
    name.param <- colnames(linfct)
    n.param <- NCOL(linfct)
    n.hypo <- NROW(linfct)

    resWald <- compare2(model, contrast = linfct, null = rhs, as.lava = FALSE)
    ## update name according to multcomp, i.e. without second member
    rownames(linfct) <- .contrast2name(linfct, null = NULL) 

### ** Global degree of freedom
    if(df){
        df.global <- round(resWald["global","df"], digits = 0)
    }else{
        df.global <- 0
    }
    
### ** compute variance-covariance matrix
    if(robust){
        vcov.model <- crossprod(iid2(model, cluster = cluster))
    }else{
        vcov.model <- vcov2(model)
    }

### ** convert to the appropriate format
    out <- list(model = model,
                linfct = linfct,
                rhs = unname(rhs),
                coef = coef(model),
                vcov = vcov.model,
                df = df.global,
                alternative = "two.sided",
                type = NULL,
                robust = robust,
                bias.correct = bias.correct)
    class(out) <- c("glht2","glht")
        
    ### ** export
    return(out)
}


## * glht2.mmm
#' @rdname glht2
#' @export
glht2.mmm <- function (model, linfct, rhs = 0,
                       bias.correct = TRUE, df = TRUE, robust = FALSE, cluster = NULL){


    if(robust==FALSE && !is.null(cluster)){
        stop("Argument \'cluster\' must be NULL when argument \'robust\' is FALSE \n")
    }
    
    ### ** check the class of each model
    n.model <- length(model)
    name.model <- names(model)    
    if(is.null(name.model)){
        stop("Argument \'model\' must be named list. \n")
    }
    
    test.lm <- sapply(model, inherits, what = "lm")
    test.gls <- sapply(model, inherits, what = "gls")
    test.lme <- sapply(model, inherits, what = "lme")
    test.lvmfit <- sapply(model, inherits, what = "lvmfit")
    if(any(test.lm + test.gls + test.lme + test.lvmfit == 0)){
        index.wrong <- which(test.lm + test.gls + test.lme + test.lvmfit == 0)
        stop("Argument \'model\' must be a list of objects that inherits from lm/gls/lme/lvmfit. \n",
             "Incorrect element(s): ",paste(index.wrong, collapse = " "),".\n")
    }


### ** define the contrast matrix
    out <- list()
    if (is.character(linfct)){
        resC <- createContrast(model, par = linfct, add.variance = TRUE)
        contrast <- resC$contrast
        ls.contrast <- resC$ls.contrast
        if("rhs" %in% names(match.call()) == FALSE){
            rhs <- resC$null
        }
    }else if(is.matrix(linfct)){

        ls.contrast <- lapply(name.model, function(x){ ## x <- name.model[2]
            iColnames <- grep(paste0("^",x,": "), colnames(linfct), value = FALSE, fixed = FALSE)
            iRownames <- rowSums(linfct[,iColnames]!=0)>0
            linfct[iRownames, iColnames,drop=FALSE]            
        })
        names(ls.contrast) <- name.model
        contrast <- linfct
    }else{
        stop("Argument \'linfct\' must be a matrix or a vector of characters. \n",
             "Consider using  out <- createContrast(...) and pass out$contrast to linfct. \n")
    }

    ## ** check whether it is possible to compute df
    if(identical(df, TRUE)){
           
        ## does each model has the same df?
        ## test.df <- try(lapply(model, df.residual), silent = TRUE)
        ## if(inherits(test.df, "try-error")){
        ##     stop("Cannot check the degrees of freedom for each model - no \'df.residual\' method available \n",
        ##          "Consider setting the argument \'df\' to FALSE \n")
        ## }
        
        ## if(any(sapply(test.df,is.null))){
        ##     stop("Cannot compute residual degrees of freedom for all models \n",
        ##          "Consider setting the argument \'df\' to FALSE \n")
        ## }

        ## if(length(unique(unlist(test.df)))>1){
        ##     stop("Residual degrees of freedom differ across models \n",
        ##          "Consider setting the argument \'df\' to FALSE \n")
        ## }

        ## are linear hypothesis model specific?
        ls.testPerModel <- lapply(ls.contrast, function(iModel){
            rowSums(contrast[,colnames(iModel)]!=0)>0
        })
        
        if(any(Reduce("+",ls.testPerModel)>1)){
            stop("Cannot compute the degrees of freedom for tests performed across several models \n",
                 "Consider setting the argument \'df\' to FALSE \n")
        }    
    }
    
    ## ** Extract influence functions from all models    
    ls.res <- lapply(1:n.model, function(iM){ ## iM <- 1

### *** Pre-compute quantities
        if(!inherits(model[[iM]],"lm2") && !inherits(model[[iM]],"gls2") && !inherits(model[[iM]],"lme2") && !inherits(model[[iM]],"lvmfit2")){
            sCorrect(model[[iM]], df = df) <- bias.correct
        }
        out$param <- model[[iM]]$sCorrect$param
        name.param <- names(out$param)
        name.object.param <- paste0(name.model[iM],": ",name.param)
        out$param <- setNames(out$param, name.object.param)
        
### *** Compute df for each test
        if(df){
            ## here null does not matter since we only extract the degrees of freedom
            iContrast <- ls.contrast[[iM]]
            colnames(iContrast) <- name.param
        
            iWald <- compare2(model[[iM]], contrast = iContrast, as.lava = FALSE)
            out$df <- iWald[1:(NROW(iWald)-1),"df"]
        }else{
            out$df <- Inf
        }
### *** get iid decomposition
        index.missing <- model[[iM]]$na.action
        n.obs <- stats::nobs(model[[iM]]) + length(index.missing)

        out$iid <- matrix(NA, nrow = n.obs, ncol = length(name.param),
                          dimnames = list(NULL, name.param))
        
        out$iid[setdiff(1:n.obs,index.missing),] <- iid2(model[[iM]], robust = robust, cluster = cluster)
        
        colnames(out$iid) <- name.object.param
            
        return(out)
        
    })
    seq.df <- unlist(lapply(ls.res,"[[","df"))
    seq.param <- unlist(lapply(ls.res,"[[","param"))

    if(df){
        if(length(unique(seq.df))>1){
            warning("Unequal degrees of freedom for the Wald statistics \n",
                    "The median of the degrees of freedom is used.")
        }
        df.global <- round(stats::median(seq.df), digits = 0)
    }else{
        df.global <- 0
    }
        
    ls.iid <- lapply(ls.res,"[[","iid")
    n.obs <- unique(unlist(lapply(ls.iid, NROW)))
    if(length(n.obs)>1){
        stop("Mismatch between the number of observations in the iid \n",
                "Likely to be due to the presence of missing values \n")
        
    }
    M.iid <- do.call(cbind,ls.iid)
    if(any(is.na(M.iid))){
        M.iid[is.na(M.iid)] <- 0
    }
    vcov.model <- crossprod(M.iid)
    
    
### ** convert to the appropriate format
    out <- list(model = model,
                linfct = linfct,
                rhs = unname(rhs),
                coef = seq.param,
                vcov = vcov.model,
                df = df.global,
                alternative = "two.sided",
                type = NULL,
                robust = robust,
                bias.correct = bias.correct)
    class(out) <- c("glht2","glht")
        
    ### ** export
    return(out)    
}


## * .calcClosure
.calcClosure <- function(name, estimate, covariance, type, df){

    n.hypo <- length(name)
    correlation <- stats::cov2cor(covariance)

    ## ** create all possible hypotheses
    ls.closure <- lapply(n.hypo:1, function(iNtest){ ## iNtest <- 1  
        iList <- list(M = utils::combn(name, m = iNtest))
        iList$vec <- apply(iList$M, 2, paste, collapse = ",")
        return(iList)
    })

    ## ** compute all p.values
    for(iLevel in 1:length(ls.closure)){ ## iLevel <- 1
        ls.closure[[iLevel]]$test <- t(apply(ls.closure[[iLevel]]$M, 2, function(iHypo){
            index <- which(name %in% iHypo)
            if(type == "chisq"){
                return(.ChisqTest(estimate[index], covariance = covariance[index,index,drop=FALSE], df = df))
            }else if(type == "max"){
                return(.tTest(estimate[index],
                              covariance = covariance[index,index,drop=FALSE],
                              correlation = correlation[index,index,drop=FALSE], df = df))
            }
        }))
        rownames(ls.closure[[iLevel]]$test) <- ls.closure[[iLevel]]$vec
    }
    
    ## ** find all hypotheses in the closure related to an individual hypothesis
    ls.hypo <- vector(mode = "list", length = n.hypo)
    for(iHypo in 1:n.hypo){ ## iHypo <- 1
        ls.hypo[[iHypo]] <- do.call(rbind,lapply(ls.closure, function(iClosure){ ## iClosure <- 1
            iIndex <- which(colSums(iClosure$M==name[iHypo])>0)
            data.frame(hypothesis = iClosure$vec[iIndex],
                       statistic = as.double(iClosure$test[iIndex,"statistic"]),
                       p.value = as.double(iClosure$test[iIndex,"p.value"]))
        }))
    }
    names(ls.hypo) <- name
        
    ## ** adjusted p.values
    vec.p.value <- unlist(lapply(ls.hypo, function(x){max(x$p.value)}))
    return(list(closure = ls.closure,
                test = ls.hypo,
                p.value = vec.p.value))
    
}

## * .tTest
.tTest <- function(estimate, covariance, correlation, df, ...){
    df1 <- length(estimate)
    statistic <- max(abs(estimate/sqrt(diag(covariance))))
    if(is.null(df)){
        distribution <-  "gaussian"
    }else{
        distribution <- "student"
    }
    p.value <- .calcPmaxIntegration(statistic, p = df1, Sigma = correlation, df = df,
                                    distribution = distribution)
    return(c("statistic" = statistic,
             "p.value" = p.value))
}

## * .ChisqTest
.ChisqTest <- function(estimate, covariance, df, ...){
    df1 <- length(estimate)
    ## q * statistic ~ chisq or fisher
    statistic <- as.double(matrix(estimate, nrow = 1) %*% solve(covariance) %*% matrix(estimate, ncol = 1)) / df1
    if(!is.null(df)){
        return(c("statistic" = statistic,
                 "p.value" = 1-stats::pf(statistic, df1 = df1, df2 = df)))
    }else{
        return(c("statistic" = statistic,
                 "p.value" = 1-stats::pchisq(statistic, df = df1)))
        
    }
}


 

### summary.glht2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: maj  2 2018 (09:20) 
## Version: 
## Last-Updated: maj  2 2018 (16:58) 
##           By: Brice Ozenne
##     Update #: 34
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

summary.glht2 <- function(object, ...){

    class(object) <- setdiff(class(object), "glht2")
    output <- summary(object, ...)
    class(output) <- append("summary.glht2",class(output))
    
    return(output)
}

print.summary.glht2 <- function(object, ...){

    class(object) <- setdiff(class(object), "summary.glht2")
    output <- utils::capture.output(print(object))
    
    txt.robust <- switch(as.character(object$robust),
                         "TRUE" = "Robust standard errors",
                         "FALSE" = "Model-based standard errors"
                         )
    txt.correction <- switch(as.character(object$bias.correct),
                             "TRUE" = " corrected for small sample bias",
                             "FALSE" = ""
                             )
    output[length(output)] <- paste0("(",txt.robust,txt.correction,")\n")

    cat(paste0(output,collapse = "\n"))
    
    return(invisible(object))
}


######################################################################
### summary.glht2.R ends here

## * Documentation
#' @title Extract Data From a Model
#' 
#' @description Extract data from a model using \code{nlme::getData}, \code{riskRegression::coxDesign} or \code{model.frame}.. 
#' If it fails it will try to extract it by its name according to \code{model$call$data}.
#' 
#' @param object the fitted model.
#' @param design.matrix [logical] should the data be extracted after transformation (e.g. conversion of categorical variables to dummy variables)?
#' Otherwise the original data will be returned.
#' @param as.data.frame [logical] should the output be converted into a \code{data.frame} object?
#' @param envir [environment] the environment from which to search the data.
#'
#' @return a dataset.
#' 
#' @examples
#' set.seed(10)
#' n <- 101
#'
#' #### linear regression ####
#' Y1 <- rnorm(n, mean = 0)
#' Y2 <- rnorm(n, mean = 0.3)
#' Id <- findInterval(runif(n), seq(0.1,1,0.1))
#' data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
#'            data.frame(Y=Y2,G="2",Id = Id)
#'            )
#' m.lm <- lm(Y ~ G, data = data.df)
#' a <- extractData(m.lm, design.matrix = TRUE)
#' b <- extractData(m.lm, design.matrix = FALSE)
#' 
#' library(nlme)
#' m.gls <- gls(Y ~ G, weights = varIdent(form = ~ 1|Id), data = data.df)
#' c <- extractData(m.gls)
#' m.lme <- lme(Y ~ G, random = ~ 1|Id, data = data.df)
#' d <- extractData(m.lme)
#' 
#' library(lava)
#' e.lvm <- estimate(lvm(Y ~ G), data = data.df)
#' e <- extractData(e.lvm)
#' e <- extractData(e.lvm, design.matrix = TRUE)
#' 
#' #### survival #### 
#' library(survival)
#'
#' \dontrun{
#'   library(riskRegression) ## needs version >=1.4.3
#'   dt.surv <- sampleData(n, outcome = "survival")
#'   m.cox <- coxph(Surv(time, event) ~ X1 + X2, data = dt.surv, x = TRUE, y = TRUE)
#'   f <- extractData(m.cox, design.matrix = FALSE)
#'   f <- extractData(m.cox, design.matrix = TRUE)
#'   m.cox <- coxph(Surv(time, event) ~ strata(X1) + X2, data = dt.surv, x = TRUE, y = TRUE)
#'   f <- extractData(m.cox, design.matrix = TRUE)
#' }
#' 
#' #### nested fuuctions ####
#' fct1 <- function(m){
#'    fct2(m)
#' }
#' fct2 <- function(m){ 
#'    extractData(m)
#' }
#' g <- fct1(m.gls)
#' @concept extractor
#' @export
`extractData` <-
    function(object, design.matrix, as.data.frame, envir){
        UseMethod("extractData", object)
    }

## * method extractData.lm
#' @rdname extractData
#' @export
extractData.lm <- function(object, design.matrix = FALSE, as.data.frame = TRUE,
                           envir = environment()){
    ## ** check arguments
    validLogical(design.matrix, valid.length = 1)
    validLogical(as.data.frame, valid.length = 1)

    ## ** extract data
    if(design.matrix){
        data <- model.matrix(object)
    }else{
        ## cannot use model.frame because it only returned the part of the dataset relevant for fitting the model
        ## this is not enougth for modelsearch2
        ## data <- try(model.frame(object), silent = TRUE)
        ## data <- object$model
        data <- evalInParentEnv(object$call$data)
        if("function" %in% class(data)){
            stop("data has the same name as a function \n",
                 "consider renaming data before generating object \n")
        }
        if(!inherits(data, "data.frame")){
            stop("Could not extract the data from the model \n")
        } 
    }

    ## ** normalize data
    if(as.data.frame){
        data <- as.data.frame(data)        
    }

    ## ** export
    return(data)
}

## * method extractData.coxph
#' @rdname extractData
#' @export
extractData.coxph <- function(object, design.matrix = FALSE, as.data.frame = TRUE,
                              envir = environment()){
    ## ** check arguments
    validLogical(design.matrix, valid.length = 1)
    validLogical(as.data.frame, valid.length = 1)

    ## ** extract data
    if(design.matrix){
         tryPkg <- requireNamespace("riskRegression")
         if("try-error" %in% class(tryPkg)){
            stop(tryPkg)
        }else if(utils::packageVersion("riskRegression")<="1.4.3"){
            stop("riskRegression version must be > 1.4.3 \n",
                 "latest version available on Github at tagteam/riskRegression \n")
        }else{
            #### [:toUpdate]
            ##  data <- try(riskRegression::coxDesign(object), silent = TRUE)
            ##  strataVar <- riskRegression::coxVariableName(object)$stratavars.original

            ## this is a temporary modification waiting for the update of riskRegression on CRAN
            coxDesign.rr <- get("coxDesign", envir = asNamespace("riskRegression"), inherits = FALSE)
            coxVariableName.rr <- get("coxVariableName", envir = asNamespace("riskRegression"), inherits = FALSE)
            data <- try(coxDesign.rr(object), silent = TRUE)
            strataVar <- coxVariableName.rr(object)$stratavars.original
        } 
      
        if(length(strataVar)>0){         
            data2 <- evalInParentEnv(object$call$data)
            data <- cbind(as.data.frame(data),
                          as.data.frame(data2)[,strataVar,drop=FALSE])        
        }
    }else{
        data <- evalInParentEnv(object$call$data)
        
        if("function" %in% class(data)){
            stop("data has the same name as a function \n",
                 "consider renaming data before generating object \n")
        }
        if(!inherits(data, "data.frame")){
            stop("Could not extract the data from the model \n")
        } 
    }

    ## ** normalize data
    if(as.data.frame){
        data <- as.data.frame(data)        
    }

    ## ** export
    return(data)
    
}

## * method extractData.cph
#' @rdname extractData
#' @export
extractData.cph <- extractData.coxph

## * method extractData.lvmfit
#' @rdname extractData
#' @export
extractData.lvmfit <- function(object, design.matrix = FALSE, as.data.frame = TRUE,
                               envir = environment()){
    ## ** check arguments
    validLogical(design.matrix, valid.length = 1)
    validLogical(as.data.frame, valid.length = 1)

    ## ** extract data
    if(design.matrix){
        data <- object$data$model.frame
        keep.cols <- intersect(c("(Intercept)",lava::vars(object)), names(data))
        data <- data[,keep.cols,drop=FALSE]
    }else{
        data <- evalInParentEnv(object$call$data)
        
        if("function" %in% class(data)){
            stop("data has the same name as a function \n",
                 "consider renaming data before generating object \n")
        }
        
        if(!inherits(data, "data.frame")){
            data <- model.frame(object)
        }
    }

    ## ** normalize data
    if(as.data.frame){
        data <- as.data.frame(data)        
    }

    ## ** export
    return(data)
    
}

## * method extractData.gls
#' @rdname extractData
#' @export
extractData.gls <- function(object, design.matrix = FALSE, as.data.frame = TRUE,
                            envir = environment()){
    ## ** check arguments
    validLogical(design.matrix, valid.length = 1)
    validLogical(as.data.frame, valid.length = 1)

    ## ** extract data
    if(design.matrix){

        # assign the dataset to the object if not in the current environment
        name.data <- as.character(object$call$data)
        if((length(name.data) == 1) && (name.data %in% ls() == FALSE)){
            object$data <- evalInParentEnv(object$call$data)
        }
      
        data <- try(nlme::getData(object), silent = TRUE)

    }else{
        data <- evalInParentEnv(object$call$data)
        
        if("function" %in% class(data)){
            stop("data has the same name as a function \n",
                 "consider renaming data before generating object \n")
        }
        if(!inherits(data, "data.frame")){
            stop("Could not extract the data from the model \n")
        } 
    }

    ## ** normalize data
    if(as.data.frame){
        data <- as.data.frame(data)        
    }

    ## ** export
    return(data)
    
}

## * method extractData.lme
#' @rdname extractData
#' @export
extractData.lme <- extractData.gls

### matrixPower.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 23 2017 (16:52) 
## Version: 
## last-updated: Jan 11 2022 (16:00) 
##           By: Brice Ozenne
##     Update #: 60
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * matrixPower (documentation)
##' @title Power of a Matrix
##' @description Compute the power of a matrix.
##' @name matrixPower
##' 
##' @param object a matrix.
##' @param power [numeric] power to be applied to the matrix.
##' @param symmetric [logical] is the matrix symmetric? Argument passed to the function \code{eigen}.
##' @param tol [numeric >0] the threshold under which the eigenvalues are set to 0.
##' @param print.warning [logical] should a warning be print when some or the eigenvalues are not strictly positive.
##' 
##' @return A matrix.

## * matrixPower (example)
##' @rdname matrixPower
##' @examples
##' ## symmetric matrix
##' set.seed(10)
##' M <- matrix(rnorm(20*6),20,6)
##' Sigma <- var(M)
##' Sigma.half <- matrixPower(Sigma, power = 1/2, symmetric = TRUE)
##' round(Sigma.half %*% Sigma.half - Sigma,5)
##' 
##' iSigma <- matrixPower(Sigma, power = -1, symmetric = TRUE)
##' round(iSigma %*% Sigma,5)
##' 
##' iSigma.half <- matrixPower(Sigma, power = -1/2, symmetric = TRUE)
##' round(iSigma.half %*% iSigma.half - iSigma,5)
##' 
##' ## non symmetric matrix
##' set.seed(10)
##' M <- matrix(abs(rnorm(9)), 3, 3) + diag(1,3,3)
##' M-t(M)
##' 
##' iM <- matrixPower(M, power = -1, symmetric = FALSE)
##' round(iM %*% M,5)
##' 
##' iM.half <- matrixPower(M, power = -1/2, symmetric = FALSE)
##' round(iM.half %*% iM.half %*% M,5)
##' 

## * matrixPower (code)
##' @export
matrixPower <- function(object, power, symmetric, tol = 1e-12, print.warning = TRUE){
    object.eigen <- eigen(object, symmetric = symmetric)

    warn <- 0
    if(power<0 && any(object.eigen$values<tol)){
        if(print.warning){
            warning("small or negative eigenvalues are set to ",tol,"\n")
        }
        warn <- 1
        object.eigen$values[object.eigen$values < tol] <- tol
    }
    nRow <- length(object.eigen$values)
    D <- diag(object.eigen$values^power, nrow = nRow, ncol = nRow)

    if(symmetric){
        out <- object.eigen$vectors %*% D %*% t(object.eigen$vectors)
    }else{
        out <- object.eigen$vectors %*% D %*% solve(object.eigen$vectors)
    }

    if(warn==1){
        attr(out,"warning") <- "small or negative eigenvalues"
    }
    return(out)

}

#----------------------------------------------------------------------
### matrixPower.R ends here

### test-matrixPower.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 24 2017 (09:21) 
## Version: 
## last-updated: mar 13 2018 (13:24) 
##           By: Brice Ozenne
##     Update #: 16
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * header
rm(list = ls())
if(TRUE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)    
}

lava.options(symbols = c("~","~~"))

context("matrixPower")
matrixPower <- lavaSearch2:::matrixPower

## * tests
M <- matrix(rnorm(2e2),20,10)
Sigma <- var(M)

test_that("square root", {
    Sigma.half <- matrixPower(Sigma, power = 1/2, symmetric = FALSE)
    expect_true(all(abs(crossprod(Sigma.half)-Sigma) < 1e-12))
})
 
test_that("inverse", {
    Sigma.m1 <- matrixPower(Sigma, power = -1, symmetric = FALSE)
    expect_equal(Sigma.m1 %*% Sigma,diag(1,NROW(Sigma),NCOL(Sigma)))
})

#----------------------------------------------------------------------
### test-matrixPower.R ends here

### iidJack.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: jun 23 2017 (09:15) 
## Version: 
## last-updated: Jan 11 2022 (16:08) 
##           By: Brice Ozenne
##     Update #: 333
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * documentation - iidJack
#' @title Jackknife iid Decomposition from Model Object
#' @description Extract iid decomposition (i.e. influence function) from model object.
#'
#' @name iidJack
#' 
#' @param object a object containing the model.
#' @param data [data.frame] dataset used to perform the jackknife.
#' @param grouping [vector] variable defining cluster of observations that will be simultaneously removed by the jackknife.
#' @param cpus [integer >0] the number of processors to use.
#' If greater than 1, the fit of the model and the computation of the influence function for each jackknife sample is performed in parallel. 
#' @param keep.warnings [logical] keep warning messages obtained when estimating the model with the jackknife samples.
#' @param keep.error [logical]keep error messages obtained when estimating the model with the jackknife samples.
#' @param cl [cluster] a parallel socket cluster generated by \code{parallel::makeCluster}
#' that has been registered using \code{registerDoParallel}.
#' @param trace [logical] should a progress bar be used to trace the execution of the function
#' @param ... [internal] only used by the generic method.
#'
#' @return A matrix with in row the samples and in columns the parameters.
#' 
#' @examples
#' n <- 20
#' set.seed(10)
#' mSim <- lvm(c(Y1,Y2,Y3,Y4,Y5) ~ 1*eta)
#' latent(mSim) <- ~eta
#' categorical(mSim, K=2) <- ~G
#' transform(mSim, Id ~ eta) <- function(x){1:NROW(x)}
#' dW <- lava::sim(mSim, n, latent = FALSE)
#'
#' #### LVM ####
#' \dontrun{
#' m1 <- lvm(c(Y1,Y2,Y3,Y4,Y5) ~ 1*eta)
#' latent(m1) <- ~eta
#' regression(m1) <- eta ~ G
#' e <- estimate(m1, data = dW)
#'
#' iid1 <- iidJack(e)
#' iid2 <- iid(e)
#' attr(iid2, "bread") <- NULL
#'
#' apply(iid1,2,sd)
#' apply(iid2,2,sd)
#' quantile(iid2 - iid1)
#' }
#'
#'
#' @concept iid decomposition
#' @export
iidJack <- function(object,...) UseMethod("iidJack")

## * method iidJack.default
#' @rdname iidJack
#' @export
iidJack.default <- function(object, data = NULL, grouping = NULL, cpus = 1,
                            keep.warnings = TRUE, keep.error = TRUE,
                            cl = NULL, trace = TRUE, ...) {

    estimate.lvm <- lava_estimate.lvm

    ## ** test args
    init.cpus <- (cpus>1 && is.null(cl))
    if(init.cpus){
        test.package <- try(requireNamespace("foreach"), silent = TRUE)
        if(inherits(test.package,"try-error")){
            stop("There is no package \'foreach\' \n",
                 "This package is necessary when argument \'cpus\' is greater than 1 \n")
        }

        if(cpus > parallel::detectCores()){
            stop("Argument \'cpus\' is greater than the number of available CPU cores \n",
                 "available CPU cores: ",parallel::detectCores(),"\n")
        }
    }
    
    ## ** extract data
    if(is.null(data)){
        myData <- extractData(object, design.matrix = FALSE, as.data.frame = TRUE,
                              envir = parent.env(environment()))
    }else{ 
        myData <-  as.data.frame(data)
    }
    
    n.obs <- NROW(myData)
    if(any(class(object) %in% "lme")){
        getCoef <- nlme::fixef
    }else{
        getCoef <- coef
    }
    coef.x <- getCoef(object)
    names.coef <- names(coef.x)
    n.coef <- length(coef.x)

    ## ** update formula/model when defined by a variable and not in the current namespace
    if(length(object$call[[2]])==1){
        modelName <- as.character(object$call[[2]])
        if(modelName %in% ls() == FALSE){
            assign(modelName, value = evalInParentEnv(object$call[[2]]))
        }
    }

    ## ** define the grouping level for the data
    if(is.null(grouping)){
        myData$XXXgroupingXXX <- 1:n.obs
        grouping <- "XXXgroupingXXX"        
    }else{
        if(length(grouping)>1){
            stop("grouping must refer to only one variable \n")
        }
        if(grouping %in% names(myData) == FALSE){
            stop("variable defined in grouping not found in data \n")
        }
    }
    myData[,grouping] <- as.character(myData[,grouping])
    Ugrouping <- unique(myData[,grouping])
    n.group <- length(Ugrouping)

    ## ** warper
    warper <- function(i){ # i <- "31"
        newData <- subset(myData, subset = myData[[grouping]]!=i)
        xnew <- tryWithWarnings(stats::update(object, data = newData))
        if(!is.null(xnew$error)){
            xnew$value <- rep(NA, n.coef)
        }else{
            xnew$value <- getCoef(xnew$value)
        }
        return(xnew)
    }
    # warper("31")

    ## ** start parallel computation
    if(init.cpus){            
        ## define cluster
        if(trace>0){
            cl <- parallel::makeCluster(cpus, outfile = "")
        }else{
            cl <- parallel::makeCluster(cpus)
        }
        ## link to foreach
        doParallel::registerDoParallel(cl)
    }

    ## *** link to foreach
    doParallel::registerDoParallel(cl)

    ## ** parallel computations: get jackknife coef
    if(cpus>1){

        if(trace>0){
            pb <- utils::txtProgressBar(max = n.group, style = 3)          
        }
        ## *** packages/objects to export
        estimator <- as.character(object$call[[1]]) 

        vec.packages <- c("lava")
        possiblePackage <- gsub("package:","",utils::getAnywhere(estimator)$where[1])
        existingPackage <- as.character(utils::installed.packages()[,"Package"])

        ls.call <- as.list(object$call)
        test.length <- which(unlist(lapply(ls.call, length))==1)
        test.class <- which(unlist(lapply(ls.call, function(cc){
            (class(c) %in% c("numeric","character","logical")) == FALSE
        })))
        test.class <- which(unlist(lapply(ls.call, class)) %in% c("numeric","character","logical") == FALSE)
    
        indexExport <- intersect(test.class,test.length)
        toExport <- sapply(ls.call[indexExport], as.character)
    
        if(possiblePackage %in% existingPackage){
            vec.packages <- c(vec.packages,possiblePackage)
        }

        parallel::clusterCall(cl, fun = function(x){
            sapply(vec.packages, function(iP){
                suppressPackageStartupMessages(requireNamespace(iP, quietly = TRUE))
            })
        })
        
        ## *** objects to export
        if(length(object$call$data)==1){
            toExport <- c(toExport,as.character(object$call$data))
        }
        if(length(object$call$formula)==1){
            toExport <- c(toExport,as.character(object$call$formula))
        }
        if(length(object$call$fixed)==1){
            toExport <- c(toExport,as.character(object$call$fixed))        
        }
        toExport <- c(unique(toExport),"tryWithWarnings")

        ## *** parallel computation
        i <- NULL # [:for CRAN check] foreach
        resLoop <- foreach::`%dopar%`(
                                foreach::foreach(i = 1:n.group,
                                                 .export = toExport),{
                                                     
                                                     if(trace>0){utils::setTxtProgressBar(pb, i)}

                                                     return(warper(Ugrouping[i]))
                                                 })
    
        if(trace>0){close(pb)}

    }else{
        
        if(trace>0){
            test.package <- try(requireNamespace("pbapply"), silent = TRUE)
            if(inherits(test.package,"try-error")){
                stop("There is no package \'pbapply\' \n",
                     "This package is necessary when argument \'trace\' is TRUE \n")
            }
            resLoop <- pbapply::pblapply(Ugrouping, warper)
            
        }else{
            resLoop <- lapply(Ugrouping, warper)
        }
        

    }
    coefJack <- do.call(rbind, lapply(resLoop,"[[","value"))
    rownames(coefJack) <- 1:n.group
    
    ## ** end parallel computation
    if(init.cpus){
        parallel::stopCluster(cl)
    }
    
    ## ** post treatment: from jackknife coef to the iid decomposition
    # defined as (n-1)*(coef-coef(-i))
    # division by n to match output of lava, i.e. IF/n
    iidJack <- -(n.group-1)/n.group * sweep(coefJack, MARGIN = 2, STATS = coef.x, FUN = "-")
    colnames(iidJack) <- names.coef

    if(keep.warnings){
        ls.warnings <- lapply(resLoop,"[[","warnings")
        names(ls.warnings) <- 1:n.group
        ls.warnings <- Filter(Negate(is.null), ls.warnings)
        attr(iidJack,"warnings") <- ls.warnings
    }
    if(keep.error){
        ls.error <- lapply(resLoop,"[[","error")
        names(ls.error) <- 1:n.group
        ls.error <- Filter(Negate(is.null), ls.error)
        attr(iidJack,"error") <- ls.error
    }
    return(iidJack)
}
    
#----------------------------------------------------------------------
### iidJack.R ends here

### iid2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (13:16) 
## Version: 
## last-updated: jan 18 2022 (09:38) 
##           By: Brice Ozenne
##     Update #: 625
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - iid2
#' @title  Influence Function With Small Sample Correction.
#' @description  Extract the influence function from a latent variable model.
#' It is similar to \code{lava::iid} but with small sample correction.
#' @name iid2
#'
#' @param object,x a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param robust [logical] if \code{FALSE}, the influence function is rescaled such its the squared sum equals the model-based standard error (instead of the robust standard error).
#' Do not match the model-based correlation though.
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients (\code{"none"}, \code{"residual"}, \code{"cox"}). Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the variance-covariance matrix.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @return A matrix containing the 1st order influence function relative to each sample (in rows)
#' and each model coefficient (in columns).
#' 
#' @examples
#' #### simulate data ####
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- sim(m,n)
#'
#' #### latent variable model ####
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' iid.tempo <- iid2(e.lvm)
#'
#'
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`iid2` <-
  function(object, ...) UseMethod("iid2")

## * iid2.lvmfit
#' @rdname iid2
#' @export
iid2.lvmfit <- function(object, robust = TRUE, cluster = NULL, as.lava = TRUE, ssc = lava.options()$ssc,...){

    return(lava::iid(estimate2(object, ssc = ssc, ...), robust = robust, cluster = cluster, as.lava = as.lava))

}

## * iid2.lvmfit2
#' @rdname iid2
#' @export
iid2.lvmfit2 <- function(object, robust = TRUE, cluster = NULL, as.lava = TRUE, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    ## ** compute iid
    object.score <- score2(object, indiv = TRUE, cluster = cluster, as.lava = FALSE)
    object.vcov <- vcov2(object, as.lava = FALSE)
    object.iid <- object.score %*% object.vcov

    if(robust == FALSE){
        object.chol.vcov <- matrixPower(object.vcov, power = 1/2, symmetric = TRUE)
        object.chol.rvcov <- matrixPower(crossprod(object.iid[rowSums(!is.na(object.iid))>0,,drop=FALSE]), power = -1/2, symmetric = TRUE)

        object.iid <- object.iid %*% object.chol.rvcov %*% object.chol.vcov
    }

    ## restaure names
    colnames(object.iid) <- colnames(object.score)
    if(!is.null(cluster)){
        attr(object.iid,"cluster") <- attr(object.score,"cluster")
    }

    ## ** export
    if(as.lava){
        object.iid <- object.iid[,names(object$sCorrect$skeleton$originalLink2param),drop=FALSE]
        colnames(object.iid) <- as.character(object$sCorrect$skeleton$originalLink2param)
    }
    return(object.iid)

}

## * iid.lvmfit2
#' @rdname iid2
#' @export
iid.lvmfit2 <- function(x, robust = TRUE, cluster = NULL, as.lava = TRUE, ...){
    return(iid2(x, robust = robust, cluster = cluster, as.lava = as.lava, ...)) ## necessary as first argument of iid must be x 
}

## * Documentation - iid2plot
#' @title  Display the i.i.d. Decomposition
#' @description  Extract the i.i.d. decomposition and display it along with the corresponding coefficient.
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param param [character] name of one of the model parameters.
#'
#' @export
iid2plot <- function(object, param){

    all.param <- coef2(object)
    if(length(param) != 1){
        stop("Incorrect argument \'param\'. \n",
             "Should have length 1. \n")
    }
    if(param %in% names(all.param) == FALSE){
        stop("Incorrect argument \'param\'. \n",
             "Should be one of the model parameters. \n")
    }

    object.iid <- iid2(object)[,param]
    h <- graphics::hist(object.iid, main = paste0(param,"=",all.param[param]), xlab = "iid")

    return(invisible(h))
}

##----------------------------------------------------------------------
### iid2.R ends here

### effects2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  4 2019 (10:28) 
## Version: 
## Last-Updated: mar 12 2019 (16:24) 
##           By: Brice Ozenne
##     Update #: 75
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * effects2 (documentation)
#' @title Effects from a fitted model
#' @description Test whether a path in the latent variable model correspond to a null effect.
#' Similar to \code{lava::effects} but with small sample correction.
#' So far it only work for paths composed of two edges.
#' @name effects2
#'
#' @param object an object that inherits from lvmfit.
#' @param link [character vector] The path for which the effect should be assessed (e.g. \code{"A~B"}),
#' i.e. the effect of the right variable (B) on the left variable (A). 
#' @param df [logical] should the degree of freedoms of the Wald statistic be computed using the Satterthwaite correction?
#' Otherwise the degree of freedoms are set to \code{Inf}, i.e. a normal distribution is used instead of a Student's t distribution when computing the p-values.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias? Argument passed to \code{sCorrect}.
#' @param ...  [internal] only used by the generic method.
#' 
#' @concept small sample inference
#' @export
`effects2` <-
  function(object, link, ...) UseMethod("effects2")

## * effects2 (examples)
## TODO

## * compare2.lvmfit
#' @rdname effects2
#' @export
effects2.lvmfit <- function(object, link, df = TRUE, bias.correct = TRUE, ...){
    sCorrect(object, df = df) <- bias.correct    
    return(effects2.lvmfit2(object, link = link))
}

## * effects2 (code)
##' @rdname effects2
##' @export
effects2.lvmfit2 <- function(object, link, ...){
    n.link <- length(link)

    name.coef <- names(coef(object))
    link.direct <- link[link %in% name.coef]
    link.other <- setdiff(link, link.direct)

    object.summary2 <- summary2(object)$coef 
    
    out <- NULL
    if(length(link.direct)>0){        
        out <- rbind(out,
                     object.summary2[link.direct,])
    }

    if(length(link.other)>0){
        allCoef <- coefType(object, as.lava = FALSE)
        mu <- setNames(object.summary2[,"Estimate"],rownames(object.summary2))
        mu.se <- setNames(object.summary2[,"Std. Error"],rownames(object.summary2))
        mu.df <- setNames(object.summary2[,"df"],rownames(object.summary2))
        Sigma <- vcov2(object)
        dSigma <- object$sCorrect$dVcov.param
        
        for(iL in 1:length(link.other)){ ## iL <- 1
            iLink <- link.other[iL]
            iPath <- lava::path(object, to = as.formula(iLink))
            if(length(iPath$path[[1]])==0){
                stop("No path found \n")
            }else{
                iNode <- iPath$path[[1]]
                iN.node <- length(iNode)
                iLink <- paste0(iNode[-1], lava.options()$symbols[1], iNode[-iN.node], collpase = "")

                if(any(iLink %in% allCoef$originalLink == FALSE)){
                    stop("Part of the path could not be identified \n")
                }
                if(any(allCoef$type[allCoef$originalLink %in% iLink] != "regression")){
                    stop("Part of the path does not correspond to a regression link \n")
                }
                if(any(length(iLink)!=2)){
                    stop("Only implemented for path of length 2 \n")
                }                

                test.noconstrain <- is.na(allCoef$value[allCoef$originalLink %in% iLink])
                if(all(test.noconstrain)){
                    out <- rbind(out,
                                 .deltaMethod_product(mu = mu, Sigma = Sigma, dSigma = dSigma, link = iLink)
                                 )
                }else{
                    iLink.NA <- iLink[test.noconstrain==FALSE]
                    iLink.NNA <- iLink[test.noconstrain==TRUE]
                    iEffect <- prod(mu[iLink])
                    iEffect.se <- mu.se[iLink.NNA] * mu[iLink.NA]
                    iEffect.df <- mu.df[iLink.NNA]
                    
                    iRow <- c("Estimate" = iEffect,
                              "Std. Error" = iEffect.se,
                              "t-value" = iEffect/iEffect.se,
                              "P-value" = 2*(1-pt(abs(iEffect/iEffect.se), df = iEffect.df)),
                              "df" = iEffect.df
                              )
                    out <- rbind(out,
                                 iRow
                                 )                    
                }
                
            }

        }            
        
    }
    rownames(out) <- c(link.direct,link.other)

    return(out)
    
}

.deltaMethod_product <- function(mu,Sigma,dSigma,link){
    link1 <- link[1]
    link2 <- link[2]

    effect <- as.double(prod(mu[link]))
    effect.var <- as.double(Sigma[link1,link1] * mu[link2]^2 + Sigma[link2,link2] * mu[link1]^2 + 2 * Sigma[link1,link2] * mu[link1] * mu[link2])
    effect.se <- sqrt(effect.var)
    effect.Wald <- effect/effect.se

    if(!is.null(dSigma)){
        keep.param <- dimnames(dSigma)[[3]]
        Ilink1 <- as.numeric(keep.param %in% link1)
        Ilink2 <- as.numeric(keep.param %in% link2)
    
        dvar1 <- dSigma[link1,link1,] * mu[link2]^2 + Sigma[link1,link1] * 2 * Ilink2 * mu[link2]
        dvar2 <- dSigma[link2,link2,] * mu[link1]^2 + Sigma[link2,link2] * 2 * Ilink1 * mu[link1]
        dvar12 <- 2 * dSigma[link1,link2,] * mu[link1] * mu[link2] + 2 * Sigma[link1,link2] * (Ilink2 * mu[link2] + mu[link1] * Ilink1)
        dvar <- dvar1 + dvar2 + dvar12

        effect.df <- 2 * effect.var^2 / (t(dvar) %*% Sigma[keep.param,keep.param,drop=FALSE] %*% dvar)[1,1]
    }else{
        effect.df <- Inf
    }
    
    return(c("Estimate" = effect,
             "Std. Error" = effect.se,
             "t-value" = effect.Wald,
             "P-value" = 2*(1-pt(abs(effect.Wald), df = effect.df)),
             "df" = effect.df
             ))
}

######################################################################
### effects2.R ends here

### biasCoxSnell.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: aug  2 2019 (10:20) 
## Version: 
## Last-Updated: Jan 11 2022 (17:36) 
##           By: Brice Ozenne
##     Update #: 177
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * .init_sscCoxSnell
.init_sscCoxSnell <- function(object,...){
    return(object,...)
}

## * .sscCoxSnell
.sscCoxSnell <- function(object, ssc){
    param <- ssc$param0
    name.param <- names(param)
    
    ## ** compute JJK
    object$sCorrect$skeleton$grid.2varD2.1varD1
    JJK <- .calcJJK(object)

    ## print(range(.old_calcJJK(object)-JJK))
    ## print(sum(.old_calcJJK(object))-sum(JJK))

    ## ** least squares
    Y <- (1/2) * sapply(name.param, function(iP){sum(JJK[,,iP] * object$sCorrect$vcov.param)})
    X <- object$sCorrect$information
    ## information(e.lvm2)

    ## dd <- as.data.frame(cbind(Y = Y,X))
    ## names(dd) <- c("Y",paste0("X",1:9))
    ## e.lm <- lm(Y ~ -1+X1+X2+X3+X4+X5+X6+X7+X8+X9, data = dd)
    
    e.lm <- stats::lm.fit(y = Y, x = X)
    newparam <- param - e.lm$coefficient
    ## print(e.lm$coefficient)
    
    ## ** export
    attr(newparam,"JJK") <- JJK
    attr(newparam,"lm") <- e.lm
    return(newparam)

}


## * .calcJJK
.calcJJK <- function(object){

    ## ** extract information
    dmu <- object$sCorrect$dmoment$dmu
    d2mu <- object$sCorrect$d2moment$d2mu
    dOmega <- object$sCorrect$dmoment$dOmega
    d2Omega <- object$sCorrect$d2moment$d2Omega

    missing.pattern <- object$sCorrect$missing$pattern
    name.pattern <- object$sCorrect$missing$name.pattern
    unique.pattern <- object$sCorrect$missing$unique.pattern
    n.pattern <- length(name.pattern)
    OmegaM1 <- object$sCorrect$moment$OmegaM1.missing.pattern
    
    name.param <- names(object$sCorrect$param)
    n.param <- length(name.param)
    n.cluster <- object$sCorrect$cluster$n.cluster
    
    grid.2meanD1.1varD1 <- object$sCorrect$skeleton$grid.2meanD1.1varD1
    grid.2meanD2.1meanD1 <- object$sCorrect$skeleton$grid.2meanD2.1meanD1
    grid.2varD2.1varD1 <- object$sCorrect$skeleton$grid.2varD2.1varD1
    n.grid.2meanD1.1varD1 <- NROW(grid.2meanD1.1varD1)
    n.grid.2meanD2.1meanD1 <- NROW(grid.2meanD2.1meanD1)
    n.grid.2varD2.1varD1 <- NROW(grid.2varD2.1varD1)

    ## ** prepare output    
    JJK <-  array(0, dim = c(n.param,n.param,n.param),
                  dimnames = list(name.param,name.param,name.param))

    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        iOmegaM1 <- OmegaM1[[iPattern]]
        idmu <- .subsetList(dmu, indexRow = iIndex, indexCol = iY)
        idOmega <- .subsetList(dOmega, indexRow = iY, indexCol = iY)
        id2mu <- .subsetList2(d2mu, indexRow = iIndex, indexCol = iY)
        id2Omega <- .subsetList2(d2Omega, indexRow = iY, indexCol = iY)

        ## *** 1 second derivative and 1 first derivative regarding the variance
        if(n.grid.2varD2.1varD1>0){
            for(iGrid in 1:n.grid.2varD2.1varD1){ # iGrid <- 1
                iName1 <- grid.2varD2.1varD1[iGrid,"X"]
                iName2 <- grid.2varD2.1varD1[iGrid,"Y"]
                iName3 <- grid.2varD2.1varD1[iGrid,"Z"]

                ## term 1
                if(grid.2varD2.1varD1[iGrid,"d2XY"]){
                    d2.Var1 <- grid.2varD2.1varD1[iGrid,"d2XY.Var1"]
                    d2.Var2 <- grid.2varD2.1varD1[iGrid,"d2XY.Var2"]
                    iDiag <- diag(iOmegaM1 %*% id2Omega[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 %*% idOmega[[iName3]])
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    ## cat("a: ",iName1," ",iName2," ",iName3,"\n")
                }

                ## term 2
                if(grid.2varD2.1varD1[iGrid,"d2XZ"]){
                    d2.Var1 <- grid.2varD2.1varD1[iGrid,"d2XZ.Var1"]
                    d2.Var2 <- grid.2varD2.1varD1[iGrid,"d2XZ.Var2"]
                    iDiag <- diag(iOmegaM1 %*% id2Omega[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 %*% idOmega[[iName2]])
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    ## cat("b: ",iName1," ",iName2," ",iName3,"\n")
                }

                ## term 3
                if(grid.2varD2.1varD1[iGrid,"d2YZ"]){
                    d2.Var1 <- grid.2varD2.1varD1[iGrid,"d2YZ.Var1"]
                    d2.Var2 <- grid.2varD2.1varD1[iGrid,"d2YZ.Var2"]
                    iDiag <- diag(iOmegaM1 %*% id2Omega[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 %*% idOmega[[iName1]])
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + 1/2 * sum(iDiag * n.cluster)
                    ## cat(iGrid,") c: ",iName1," ",iName2," ",iName3," = ", 1/2 * sum(iDiag * n.cluster),"\n")
                }
            }
        }

        ## *** 1 second derivative and 1 first derivative regarding the mean
        if(n.grid.2meanD2.1meanD1>0){
            for(iGrid in 1:n.grid.2meanD2.1meanD1){ # iGrid <- 1
                iName1 <- grid.2meanD2.1meanD1[iGrid,"X"]
                iName2 <- grid.2meanD2.1meanD1[iGrid,"Y"]
                iName3 <- grid.2meanD2.1meanD1[iGrid,"Z"]

                ## term 4
                if(grid.2meanD2.1meanD1[iGrid,"d2XY"]){
                    d2.Var1 <- grid.2meanD2.1meanD1[iGrid,"d2XY.Var1"]
                    d2.Var2 <- grid.2meanD2.1meanD1[iGrid,"d2XY.Var2"]
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 * idmu[[iName3]])
                }

                ## term 5
                if(grid.2meanD2.1meanD1[iGrid,"d2XZ"]){
                    d2.Var1 <- grid.2meanD2.1meanD1[iGrid,"d2XZ.Var1"]
                    d2.Var2 <- grid.2meanD2.1meanD1[iGrid,"d2XZ.Var2"]
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 * idmu[[iName2]])
                }

                ## term 6
                if(grid.2meanD2.1meanD1[iGrid,"d2YZ"]){
                    d2.Var1 <- grid.2meanD2.1meanD1[iGrid,"d2YZ.Var1"]
                    d2.Var2 <- grid.2meanD2.1meanD1[iGrid,"d2YZ.Var2"]
                    JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + sum(id2mu[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 * idmu[[iName1]])
                }
            }
        }
        
        ## *** 2 first derivative regarding the mean and one regarding the variance
        if(n.grid.2meanD1.1varD1>0){
            for(iGrid in 1:n.grid.2meanD1.1varD1){ # iGrid <- 1

                ## term 7
                iName1 <- grid.2meanD1.1varD1[iGrid,"Z"]
                iName2 <- grid.2meanD1.1varD1[iGrid,"X"]
                iName3 <- grid.2meanD1.1varD1[iGrid,"Y"]
                value <- sum(idmu[[iName2]] %*% iOmegaM1 %*% idOmega[[iName1]] %*% iOmegaM1 * idmu[[iName3]])
                JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + value

                ## term 8 
                iName1 <- grid.2meanD1.1varD1[iGrid,"X"]
                iName2 <- grid.2meanD1.1varD1[iGrid,"Z"]
                iName3 <- grid.2meanD1.1varD1[iGrid,"Y"]
                JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - value
                
                ## term 9
                iName1 <- grid.2meanD1.1varD1[iGrid,"X"]
                iName2 <- grid.2meanD1.1varD1[iGrid,"Y"]
                iName3 <- grid.2meanD1.1varD1[iGrid,"Z"]
                JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - value
            }
        }
        

    }

    ## sum(abs(JJK)>0)
    return(JJK)
}


## * .old_calcJJK
.old_calcJJK <- function(object){

    ## ** extract information
    dmu <- object$sCorrect$dmoment$dmu
    d2mu <- object$sCorrect$d2moment$d2mu
    dOmega <- object$sCorrect$dmoment$dOmega
    d2Omega <- object$sCorrect$d2moment$d2Omega

    missing.pattern <- object$sCorrect$missing$pattern
    name.pattern <- object$sCorrect$missing$name.pattern
    unique.pattern <- object$sCorrect$missing$unique.pattern
    n.pattern <- length(name.pattern)
    OmegaM1 <- object$sCorrect$moment$OmegaM1.missing.pattern
    
    name.param <- names(object$sCorrect$param)
    n.param <- length(name.param)
    n.cluster <- object$sCorrect$cluster$n.cluster
    
    grid.2meanD1.1varD1 <- object$sCorrect$skeleton$grid.2meanD1.1varD1
    grid.2meanD2.1meanD1 <- object$sCorrect$skeleton$grid.2meanD2.1meanD1
    grid.2varD2.1varD1 <- object$sCorrect$skeleton$grid.2varD2.1varD1
    n.grid.2meanD1.1varD1 <- NROW(grid.2meanD1.1varD1)
    n.grid.2meanD2.1meanD1 <- NROW(grid.2meanD2.1meanD1)
    n.grid.2varD2.1varD1 <- NROW(grid.2varD2.1varD1)

    ## ** prepare output    
    JJK <-  array(0, dim = c(n.param,n.param,n.param),
                  dimnames = list(name.param,name.param,name.param))

    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        iOmegaM1 <- OmegaM1[[iPattern]]
        idmu <- .subsetList(dmu, indexRow = iIndex, indexCol = iY)
        idOmega <- .subsetList(dOmega, indexRow = iY, indexCol = iY)
        id2mu <- .subsetList2(d2mu, indexRow = iIndex, indexCol = iY)
        id2Omega <- .subsetList2(d2Omega, indexRow = iY, indexCol = iY)

        for(iParam1 in 1:n.param){ ## iParam1 <- 1
            for(iParam2 in 1:n.param){ ## iParam2 <- 1
                for(iParam3 in 1:n.param){ ## iParam3 <- 1

                    iName1 <- name.param[iParam1]
                    iName2 <- name.param[iParam2]
                    iName3 <- name.param[iParam3]
                    
                    ## *** 1 second derivative and 1 first derivative regarding the variance

                    ## term 1
                    if(!is.null(idOmega[[iName3]]) && !is.null(id2Omega[[iName1]][[iName2]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName1]][[iName2]] %*% iOmegaM1 %*% idOmega[[iName3]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    }else if(!is.null(idOmega[[iName3]]) && !is.null(id2Omega[[iName2]][[iName1]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName2]][[iName1]] %*% iOmegaM1 %*% idOmega[[iName3]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    }

                    ## term 2
                    if(!is.null(idOmega[[iName2]]) && !is.null(id2Omega[[iName1]][[iName3]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName1]][[iName3]] %*% iOmegaM1 %*% idOmega[[iName2]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    }else if(!is.null(idOmega[[iName2]]) && !is.null(id2Omega[[iName3]][[iName1]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName3]][[iName1]] %*% iOmegaM1 %*% idOmega[[iName2]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - 1/2 * sum(iDiag * n.cluster)
                    }

                    ## term 3
                    if(!is.null(idOmega[[iName1]]) && !is.null(id2Omega[[iName2]][[iName3]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName2]][[iName3]] %*% iOmegaM1 %*% idOmega[[iName1]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + 1/2 * sum(iDiag * n.cluster)
                        ## cat("c: ",iName1," ",iName2," ",iName3," = ", 1/2 * sum(iDiag * n.cluster),"\n")
                    }else if(!is.null(idOmega[[iName1]]) && !is.null(id2Omega[[iName3]][[iName2]])){
                        iDiag <- diag(iOmegaM1 %*% id2Omega[[iName3]][[iName2]] %*% iOmegaM1 %*% idOmega[[iName1]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + 1/2 * sum(iDiag * n.cluster)
                        ## cat("c: ",iName1," ",iName2," ",iName3," = ", 1/2 * sum(iDiag * n.cluster),"\n")
                    }
        
                    ## *** 1 second derivative and 1 first derivative regarding the mean

                    ## term 4
                    if(!is.null(idmu[[iName3]]) && !is.null(id2mu[[iName1]][[iName2]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[iName1]][[iName2]] %*% iOmegaM1 * idmu[[iName3]])
                    }else if(!is.null(idmu[[iName3]]) && !is.null(id2mu[[iName2]][[iName1]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[iName2]][[iName1]] %*% iOmegaM1 * idmu[[iName3]])
                    }

                    ## term 5
                    if(!is.null(idmu[[iName2]]) && !is.null(id2mu[[iName1]][[iName3]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[iName1]][[iName3]] %*% iOmegaM1 * idmu[[iName2]])
                    }else if(!is.null(idmu[[iName2]]) && !is.null(id2mu[[iName3]][[iName1]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - sum(id2mu[[iName3]][[iName1]] %*% iOmegaM1 * idmu[[iName2]])
                    }

                    ## term 6
                    if(!is.null(idmu[[iName1]]) && !is.null(id2mu[[iName2]][[iName3]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + sum(id2mu[[iName2]][[iName3]] %*% iOmegaM1 * idmu[[iName1]])
                    }else if(!is.null(idmu[[iName1]]) && !is.null(id2mu[[iName3]][[iName2]])){
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + sum(id2mu[[iName3]][[iName2]] %*% iOmegaM1 * idmu[[iName1]])
                    }

                    ## *** 2 first derivative regarding the mean and one regarding the variance

                    ## term 7
                    if(!is.null(idmu[[iName2]]) && !is.null(idOmega[[iName1]]) && !is.null(idmu[[iName3]])){
                        value <- sum(idmu[[iName2]] %*% iOmegaM1 %*% idOmega[[iName1]] %*% iOmegaM1 * idmu[[iName3]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] + value
                    }

                    ## term 8 
                    if(!is.null(idmu[[iName1]]) && !is.null(idOmega[[iName2]]) && !is.null(idmu[[iName3]])){
                        value <- sum(idmu[[iName1]] %*% iOmegaM1 %*% idOmega[[iName2]] %*% iOmegaM1 * idmu[[iName3]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - value
                    }
                
                    ## term 9
                    if(!is.null(idmu[[iName2]]) && !is.null(idOmega[[iName3]]) && !is.null(idmu[[iName1]])){
                        value <- sum(idmu[[iName2]] %*% iOmegaM1 %*% idOmega[[iName3]] %*% iOmegaM1 * idmu[[iName1]])
                        JJK[iName1,iName2,iName3] <- JJK[iName1,iName2,iName3] - value
                    }
                }
            }
        }        

    }

    return(JJK)
}
######################################################################
### biasCoxSnell.R ends here

### leverage.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 19 2018 (17:58) 
## Version: 
## Last-Updated: feb 11 2019 (13:26) 
##           By: Brice Ozenne
##     Update #: 36
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * documentation - leverage2
#' @title Extract Leverage Values
#' @description Extract leverage values from a Gaussian linear model. 
#' @name leverage2
#' 
#' @param object a \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} object.
#' @param param [optional] the fitted parameters.
#' @param data [optional] the data set.
#' @param ... arguments to be passed to \code{sCorrect}.
#'
#' @details The leverage are defined as the partial derivative of the fitted values with respect to the observations.
#' \deqn{
#' leverage_i = \frac{\partial \hat{Y}_i}{\partial Y_i}
#' }
#' See Wei et al. (1998). \cr \cr
#' 
#' If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the residuals.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#' 
#' @return a matrix containing the leverage relative to each sample (in rows)
#' and each endogenous variable (in column).
#'
#' @references Bo-Cheng Wei et al., Generalized Leverage and its applications (1998), Scandinavian Journal of Statistics 25:1:25-37.
#' 
#' @examples
#' ## simulate data
#' set.seed(10)
#' m <- lvm(Y1~eta,Y2~eta,Y3~eta)
#' latent(m) <- ~eta
#' d <- lava::sim(m,20, latent = FALSE)
#'
#' ## standard linear model
#' e.lm <- lm(Y1~Y2, data = d)
#'
#' sCorrect(e.lm) <- TRUE
#' range(as.double(leverage2(e.lm)) - influence(e.lm)$hat)
#'
#' ## latent variable model
#' e.lvm <- estimate(m, data = d)
#' sCorrect(e.lvm) <- TRUE
#' leverage2(e.lvm)
#' 
#' @concept small sample inference
#' @export
`leverage2` <-
    function(object, ...) UseMethod("leverage2")

## * leverage2.lm
#' @rdname leverage2
#' @export
leverage2.lm <- function(object, param = NULL, data = NULL, ...){

    sCorrect(object, param = param, data = data, df = FALSE, ...) <- TRUE

    ### ** export
    return(object$sCorrect$leverage)
}

## * leverage2.gls
#' @rdname leverage2
#' @export
leverage2.gls <- leverage2.lm

## * leverage2.lme
#' @rdname leverage2
#' @export
leverage2.lme <- leverage2.lm

## * leverage2.lvmfit
#' @rdname leverage2
#' @export
leverage2.lvmfit <- leverage2.lm

## * leverage2.lm2
#' @rdname leverage2
#' @export
leverage2.lm2 <- function(object, param = NULL, data = NULL, ...){

    if(!is.null(param) || !is.null(data)){
        args <- object$sCorrect$args
        args$df <- FALSE
        object$sCorrect <- do.call(sCorrect,
                                   args = c(list(object, param = param, data = data),
                                            args))
    }
    
    return(object$sCorrect$leverage)   
}

## * leverage2.gls2
#' @rdname leverage2
#' @export
leverage2.gls2 <- leverage2.lm2

## * leverage2.lme2
#' @rdname leverage2
#' @export
leverage2.lme2 <- leverage2.lm2

## * leverage2.lvmfit2
#' @rdname leverage2
#' @export
leverage2.lvmfit2 <- leverage2.lm2



##----------------------------------------------------------------------
### leverage.R ends here

### sCorrect-hessian2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: dec 11 2019 (14:09) 
## Version: 
## Last-Updated: jan 23 2024 (10:25) 
##           By: Brice Ozenne
##     Update #: 144
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - hessian2
#' @title  Hessian With Small Sample Correction.
#' @description  Extract the hessian from a latent variable model, with small sample correction
#' @name hessian2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param indiv [logical] If \code{TRUE}, the hessian relative to each observation is returned. Otherwise the total hessian is returned.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the hessian.
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#' 
#' @return An array containing the second derivative of the likelihood relative to each sample (dim 3)
#' and each pair of model coefficients (dim 1,2).
#' 
#' @examples
#' #### simulate data ####
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- Sequence.lvm(0)
#' set.seed(10)
#' d <- lava::sim(m,n)
#'
#' #### latent variable models ####
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' hessian2(e.lvm)
#'
#' @concept small sample inference
#' @export
`hessian2` <-
  function(object, indiv, cluster, as.lava, ...) UseMethod("hessian2")

## * hessian2.lvmfit
#' @rdname hessian2
#' @export
hessian2.lvmfit <- function(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ssc = lava.options()$ssc, ...){

    return(hessian2(estimate2(object, ssc = ssc, hessian = TRUE, ...), cluster = cluster, as.lava = as.lava))

}


## * hessian2.lvmfit2
#' @rdname hessian2
#' @export
hessian2.lvmfit2 <- function(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...){
    
    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    ## ** define cluster
    if(is.null(cluster)){
        n.cluster <- object$sCorrect$cluster$n.cluster
        cluster.index <- 1:n.cluster
    }else{
        if(!is.numeric(cluster)){
            data <- object$sCorrect$data
            if(length(cluster)==1){                
                if(cluster %in% names(data) == FALSE){
                    stop("Invalid \'cluster\' argument \n",
                         "Could not find variable \"",cluster,"\" in argument \'data\' \n")
                }
                cluster <- data[[cluster]]
            }
            cluster.index <- as.numeric(factor(cluster, levels = unique(cluster)))            
        }else{
            cluster.index <- as.numeric(factor(cluster, levels = unique(cluster)))
        }

        n.cluster <- length(unique(cluster.index))
    }

    ## ** get hessian
    hessian <- object$sCorrect$hessian
    if(is.null(hessian)){
        hessian <- .hessian2(dmu = object$sCorrect$dmoment$dmu,
                             dOmega = object$sCorrect$dmoment$dOmega,
                             d2mu = object$sCorrect$d2moment$d2mu,
                             d2Omega = object$sCorrect$d2moment$d2Omega,
                             epsilon = object$sCorrect$residuals,                                     
                             OmegaM1 = object$sCorrect$moment$OmegaM1.missing.pattern,
                             missing.pattern = object$sCorrect$missing$pattern,
                             unique.pattern = object$sCorrect$missing$unique.pattern,
                             name.pattern = object$sCorrect$missing$name.pattern,
                             grid.mean = object$sCorrect$skeleton$grid.dmoment$mean, 
                             grid.var = object$sCorrect$skeleton$grid.dmoment$var, 
                             grid.hybrid = object$sCorrect$skeleton$grid.dmoment$hybrid, 
                             name.param = object$sCorrect$skeleton$Uparam,
                             leverage = object$sCorrect$leverage,
                             n.cluster = object$sCorrect$cluster$n.cluster,
                             weights = object$sCorrect$weights)

        hessian.name <- stats::setNames(names(object$sCorrect$skeleton$originalLink2param),object$sCorrect$skeleton$originalLink2param)[object$sCorrect$skeleton$Uparam]
        dimnames(hessian) <- list(as.character(hessian.name),
                                  as.character(hessian.name),
                                  NULL)
    }

    if(!is.null(cluster)){ ## aggregate hessian by cluster
        hessianSave <- hessian
        hessian <- array(0, dim = dim(hessian),
                         dimnames = dimnames(hessian))
        for(i in 1:length(cluster)){
            hessian[,,cluster[i]] <- hessian[,,cluster[i]] + hessianSave[,,i]
        }
    }

    ## ** export
    hessian <- hessian[names(object$sCorrect$skeleton$originalLink2param),
                       names(object$sCorrect$skeleton$originalLink2param),
                      ,
                       drop=FALSE]
    if(as.lava == FALSE){
        dimnames(hessian) <- list(as.character(object$sCorrect$skeleton$originalLink2param),
                                  as.character(object$sCorrect$skeleton$originalLink2param),
                                  NULL)
    }
    if(indiv==FALSE){
        hessian <- apply(hessian, 1:2, sum)
        
    }else if(!is.null(cluster)){
        index2.cluster <- tapply(1:length(cluster),cluster,list)
        attr(hessian,"cluster") <- names(index2.cluster)
    }
    
    return(hessian)
}

## * .hessian2
#' @title Compute the Hessian Matrix From the Conditional Moments
#' @description Compute the Hessian matrix from the conditional moments.
#' @name hessian2-internal
#' 
#' @details \code{calc_hessian} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.hessian2 <- function(dmu, dOmega, d2mu, d2Omega, epsilon, OmegaM1,
                      missing.pattern, unique.pattern, name.pattern,
                      grid.mean, grid.var, grid.hybrid, name.param,
                      leverage, n.cluster, weights){
    if(lava.options()$debug){cat(".hessian2\n")}

    ## ** Prepare
    n.grid.mean <- NROW(grid.mean)
    n.grid.var <- NROW(grid.var)
    n.grid.hybrid <- NROW(grid.hybrid)
    n.param <- length(name.param)
    n.pattern <- length(name.pattern)

    hessian <- array(NA, dim = c(n.param, n.param, n.cluster),
                     dimnames = list(name.param,name.param,NULL))
    if(length(dmu)>0){
        index.mean <- 1:n.grid.mean
    }else{
        index.mean <- NULL
    }
    if(length(dOmega)>0){
        index.var <- 1:n.grid.var
    }else{
        index.var <- NULL
    }
    if(length(dmu)>0 && length(dOmega)>0){
        index.hybrid <- 1:n.grid.hybrid
    }else{
        index.hybrid <- NULL
    } 

    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        iOmegaM1 <- OmegaM1[[iPattern]]
        iEpsilon <- epsilon[iIndex,iY,drop=FALSE]
        idmu <- .subsetList(dmu, indexRow = iIndex, indexCol = iY)
        idOmega <- .subsetList(dOmega, indexRow = iY, indexCol = iY)
        id2mu <- .subsetList2(d2mu, indexRow = iIndex, indexCol = iY)
        id2Omega <- .subsetList2(d2Omega, indexRow = iY, indexCol = iY)
        if(!is.null(leverage)){
            iLeverage <- leverage[iIndex,iY,drop=FALSE]
        }
        hessian[,,iIndex] <- 0 ##  ## initialize (keep NA for missing values)

        ## *** second derivative relative to the mean parameters
        for(iG in index.mean){ # iG <- 1
            iP1 <- grid.mean[iG,1]
            iP2 <- grid.mean[iG,2]
            if(grid.mean[iG,"d2.12"]){
                term1 <- rowSums((id2mu[[iP1]][[iP2]] %*% iOmegaM1) * iEpsilon)
            }else if(grid.mean[iG,"d2.21"]){
                term1 <- rowSums((id2mu[[iP2]][[iP1]] %*% iOmegaM1) * iEpsilon)
            }else{
                term1 <- 0
            }
            term2 <- -rowSums((idmu[[iP1]] %*% iOmegaM1) * idmu[[iP2]])
            hessian[iP1,iP2,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE] + term1 + term2
            hessian[iP2,iP1,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE]
        }
        ## *** second derivative relative to the variance parameters
        for(iG in index.var){ # iG <- 2
            iP1 <- grid.var[iG,1]
            iP2 <- grid.var[iG,2]

            term1a <- - diag(iOmegaM1 %*% idOmega[[iP1]] %*% iOmegaM1 %*% idOmega[[iP2]])
            term2 <- - rowSums((iEpsilon %*% iOmegaM1 %*% idOmega[[iP2]] %*% iOmegaM1 %*% idOmega[[iP1]] %*% iOmegaM1) * iEpsilon)
            if(grid.var[iG,"d2.12"]){
                term1b <- diag(iOmegaM1 %*% id2Omega[[iP1]][[iP2]])
                term3 <- 1/2 * rowSums((iEpsilon %*% iOmegaM1 %*% id2Omega[[iP1]][[iP2]] %*% iOmegaM1) * iEpsilon)
            }else if(grid.var[iG,"d2.21"]){
                term1b <- diag(iOmegaM1 %*% id2Omega[[iP2]][[iP1]])
                term3 <- 1/2 * rowSums((iEpsilon %*% iOmegaM1 %*% id2Omega[[iP2]][[iP1]] %*% iOmegaM1) * iEpsilon)
            }else{
                term1b <- 0
                term3 <- 0
            }
            if(is.null(leverage)){
                hessian[iP1,iP2,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE] - 1/2 * rep(sum(term1a + term1b), length(iIndex)) + term2 + term3
            }else{
                hessian[iP1,iP2,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE] - 1/2 * rowSums( sweep(1-iLeverage, FUN = "*", STATS = term1a + term1b, MARGIN = 2) ) + term2 + term3
            }
            hessian[iP2,iP1,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE]
        }
        
        ## *** second derivative relative to the mean and variance parameters
        for(iG in index.hybrid){ # iG <- 1
            iP1 <- grid.hybrid[iG,1]
            iP2 <- grid.hybrid[iG,2]

            if(!is.null(idmu[[iP1]]) && !is.null(idOmega[[iP2]])){
                term1 <- - rowSums((idmu[[iP1]] %*% iOmegaM1 %*% idOmega[[iP2]] %*% iOmegaM1) * iEpsilon)
            }else{
                term1 <- 0
            }
            if(!is.null(idmu[[iP2]]) && !is.null(idOmega[[iP1]])){
                term2 <- - rowSums((idmu[[iP2]] %*% iOmegaM1 %*% idOmega[[iP1]] %*% iOmegaM1) * iEpsilon)
            }else{
                term2 <- 0
            }
            
            hessian[iP1,iP2,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE] + term1 + term2
            hessian[iP2,iP1,iIndex] <- hessian[iP1,iP2,iIndex,drop=FALSE]
        }
    }

    ## ** export
    if(!is.null(weights)){
        for(iI in 1:length(weights)){
            hessian[,,iI] <- weights[iI] * hessian[,,iI]
        }
    }
    return(hessian)
}

## * .subsetList
.subsetList <- function(object, indexRow, indexCol){
    if(length(object)==0){
        return(object)
    }else{    
        return(lapply(object, FUN = function(iL){iL[indexRow,indexCol,drop=FALSE]}))
    }
}
## * .subsetList2
.subsetList2 <- function(object, indexRow, indexCol){
    if(length(object)==0){
        return(object)
    }else{    
        return(lapply(object, FUN = .subsetList, indexRow = indexRow, indexCol = indexCol))
    }
}


######################################################################
### sCorrect-hessian2.R ends here

### var2dummy.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: jun 22 2017 (16:03) 
## Version: 
## last-updated: sep 28 2018 (11:57) 
##           By: Brice Ozenne
##     Update #: 41
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * var2dummy
#' @title Convert Variable Names to Dummy Variables Names.
#' @description When dealing with categorical variables, the \code{estimate} function convert the categorical variables into dummy variables.
#' This function convert a set of variable names to their corresponding name in the model with dummy variables
#' @name var2dummy
#' 
#' @param object a \code{lvm} object.
#' @param var [character] the variable to be transformed.
#' @param data [data.frame] dataset according to which the model should be updated.
#' @param rm.first.factor [logical] should the first level of each categorical variable be ignored?
#' @param ... [internal] additional arguments to be passed from \code{var2dummy.lvm} to \code{var2dummy.list}.
#' 
#' @examples
#'
#' \dontrun{
#' var2dummy <- lavaSearch2:::var2dummy
#' var2dummy.list <- lavaSearch2:::var2dummy.list
#' var2dummy.lvm <- lavaSearch2:::var2dummy.lvm
#' 
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u ~ X1+X2
#' var2dummy(m, var = c("X1","X2"))
#' categorical(m,labels=c("M","F","MF")) <- ~X1
#' var2dummy(m, var = c("X1","X2"))
#' categorical(m,labels=c("1","2","3")) <- ~X2
#' var2dummy(m, var = c("X1","X2"))
#' }
#' 
#' @keywords internal
`var2dummy` <-
    function(object,...) UseMethod("var2dummy")

## * var2dummy.list
#' @rdname var2dummy
#' @export
var2dummy.list <- function(object, var, rm.first.factor = TRUE, ...){

    var <- stats::setNames(var,var)
    ## convertion to dummy variable name for categorical variables
    factor.var <- names(object$x$attributes$labels)
    
    if(!is.null(var) && any(var %in% factor.var)){
        subvar <- var[var %in% factor.var]
        for(iFactor in subvar){ # iFactor <- "X1"
            newvar <- paste0(iFactor,object$x$attributes$labels[[iFactor]])
            if(rm.first.factor){newvar <- newvar[-1]}
            newvar <- stats::setNames(newvar, rep(iFactor, length(newvar)))
            var <- c(var[names(var)!=iFactor],newvar)            
        }
    }
    return(var)
}

## * var2dummy.lvm
#' @rdname var2dummy
#' @export
var2dummy.lvm <- function(object, data = NULL, ...){

    if(is.null(data)){
        data <- lava::sim(object, n = 1)
    }
    object2 <- lava_categorical2dummy(object, data)
    
    ## recover attributes for models not defined using categorical
    obsvars <- setdiff(lava::vars(object2$x),lava::latent(object))
    
    if(any(obsvars %in% names(data) == FALSE)){
        missing.vars <- obsvars[obsvars %in% names(data) == FALSE]
        test.num <- sapply(1:NCOL(data), function(col){is.numeric(data[[col]])})
        possible.match <- names(data)[test.num==FALSE]
        n.possible.match <- length(possible.match)
        
        ls.labels <- list()
        for(iMatch in 1:n.possible.match){ # iMatch <- 1
            iVar <- possible.match[iMatch]
            iLabel <- levels(as.factor(data[[iVar]]))
            if(any(paste0(iVar,iLabel) %in% missing.vars)){
                ls.labels[[iVar]] <- iLabel
            }
        }

        object2$x$attributes$labels <- ls.labels
    }
    
    res <- var2dummy(object2, ...)
    return(res)
}

#----------------------------------------------------------------------
### var2dummy.R ends here

### createContrast.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan 31 2018 (12:05) 
## Version: 
## Last-Updated: jan 18 2022 (10:38) 
##           By: Brice Ozenne
##     Update #: 491
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - createContrast
#' @title Create Contrast matrix
#' @description Returns a contrast matrix corresponding an object.
#' The contrast matrix will contains the hypotheses in rows and the model coefficients in columns.
#' @name createContrast
#' 
#' @param object a \code{lvmfit} object or a list of  a \code{lvmfit} objects.
#' @param linfct [vector of characters] expression defining the linear hypotheses to be tested.
#' Can also be a regular expression (of length 1) that is used to identify the coefficients to be tested using \code{grep}.
#' See the examples section.
#' @param ... Argument to be passed to \code{.createContrast}:
#' \itemize{
#' \item diff.first [logical] should the contrasts between the first and any of the other coefficients define the null hypotheses.
#' \item add.rowname [logical] add rownames to the contrast matrix and names to the right-hand side.
#' \item rowname.rhs [logical] when naming the hypotheses, add the right-hand side (i.e. "X1-X2=0" instead of "X1-X2").
#' \item sep [character vector of length2] character surrounding the left part of the row names.
#' }
#'
#' @details
#' One can initialize an empty contrast matrix setting the argument\code{linfct} to \code{character(0)}. \cr \cr
#'
#' @return A list containing
#' \itemize{
#' \item{contrast} [matrix] a contrast matrix corresponding to the left hand side of the linear hypotheses.
#' \item{null} [vector] the right hand side of the linear hypotheses.
#' \item{Q} [integer] the rank of the contrast matrix.
#' \item{ls.contrast} [list, optional] the contrast matrix corresponding to each submodel.
#' Only present when the \code{argument} object is a list of models.
#' }
#' @examples
#' ## Simulate data
#' mSim <- lvm(X ~ Age + Treatment,
#'             Y ~ Gender + Treatment,
#'             c(Z1,Z2,Z3) ~ eta, eta ~ treatment,
#'             Age[40:5]~1)
#' latent(mSim) <- ~eta
#' categorical(mSim, labels = c("placebo","SSRI")) <- ~Treatment
#' categorical(mSim, labels = c("male","female")) <- ~Gender
#' n <- 1e2
#' set.seed(10)
#' df.data <- lava::sim(mSim,n)
#'
#' ## Estimate separate models
#' lmX <- lava::estimate(lvm(X ~ -1 + Age + Treatment), data = df.data)
#' lmY <- lava::estimate(lvm(Y ~ -1 + Gender + Treatment), data = df.data)
#' lvmZ <- lava::estimate(lvm(c(Z1,Z2,Z3) ~ -1 + 1*eta, eta ~ -1 + Treatment), 
#'                  data = df.data)
#'
#' ## Contrast matrix for a given model
#' createContrast(lmX, linfct = "X~Age")
#' createContrast(lmX, linfct = c("X~Age=0","X~Age+5*X~TreatmentSSRI=0"))
#' createContrast(lmX, linfct = c("X~Age=0","X~Age+5*X~TreatmentSSRI=0"), sep = NULL)
#' createContrast(lmX, linfct = character(0))
#'
#' ## Contrast matrix for the join model
#' ls.lvm <- list(X = lmX, Y = lmY, Z = lvmZ)
#' createContrast(ls.lvm, linfct = "TreatmentSSRI=0")
#' createContrast(ls.lvm, linfct = "TreatmentSSRI=0", rowname.rhs = FALSE)
#' createContrast(ls.lvm, linfct = character(0))
#'
#' ## Contrast for multigroup models
#' m <- lava::lvm(Y~Age+Treatment)
#' e <- lava::estimate(list(m,m), data = split(df.data, df.data$Gender))
#' print(coef(e))
#' createContrast(e, linfct = "Y~TreatmentSSRI@1 - Y~TreatmentSSRI@2 = 0")
#' createContrast(e, linfct = "Y~TreatmentSSRI@2 - Y~TreatmentSSRI@1 = 0")
#' 
#' @export
`createContrast` <-
    function(object, ...) UseMethod("createContrast")

## * createContrast.character
#' @rdname createContrast
#' @export
createContrast.character <- function(object, ...){

    object.lhs <- strsplit(object,split = "=")[[1]] ## rm rhs
    object.term.lhs <- base::trimws(strsplit(object.lhs[[1]],split = "\\+|\\-")[[1]], which = "both") ## split with +/-
    object.coefname <- sapply(object.term.lhs, function(iE){tail(strsplit(iE,split="*",fixed=TRUE)[[1]],1)}) ## remove *

    return(.createContrast(object, object.coefname, ...))
        
}

## * createContrast.lvmfit
#' @rdname createContrast
#' @export
createContrast.lvmfit <- function(object, linfct, ...){
    name.param <- names(coef(object))
    if(!identical(class(linfct),"character")){
        stop("Argument \'linfct\' must be of type character (or vector of character) \n")
    }
    ## if(length(linfct)==1 && all(linfct %in% name.param == FALSE) & all(sapply(linfct, function(iL){any(grepl(iL, name.param, fixed = TRUE))}))){
    ##     linfct <- grep(linfct, name.param,  value = TRUE)
    ## }
    out <- .createContrast(linfct = linfct, name.param = name.param, ...)
    return(out)
    
}

## * createContrast.lvmfit
#' @rdname createContrast
#' @export
createContrast.lvmfit2 <- function(object, linfct, ...){
    if(!identical(class(linfct),"character")){
        stop("Argument \'linfct\' must be of type character (or vector of character) \n")
    }
    ## if(length(linfct)==1){
        ## name.param <- names(coef(object, as.lava = TRUE))
        ## name.param2 <- names(coef(object, as.lava = FALSE))

    ##     if(all(name.param != linfct) & all(sapply(linfct, function(iL){any(grepl(iL, name.param, fixed = TRUE))}))){
    ##         linfct <- grep(linfct, name.param,  value = TRUE)
    ##     }
    ## }else{
    ##     name.param <- names(coef(object))
    ## }
    name.param <- names(coef(object, as.lava = TRUE))
    name.param2 <- names(coef(object, as.lava = FALSE))
    if(any(name.param!=name.param2)){
        if(any(sapply(setdiff(name.param2,name.param), function(iCoef){any(grepl(iCoef, linfct, fixed = TRUE))}))){
            out <- .createContrast(linfct = linfct, name.param = name.param2, ...)
        }else{
            out <- .createContrast(linfct = linfct, name.param = name.param, ...)
        }
        
    }else{
        out <- .createContrast(linfct = linfct, name.param = name.param, ...)
    }
    
    return(out)
    
}

## * createContrast.list
#' @rdname createContrast
#' @export
createContrast.list <- function(object, linfct = NULL, ...){

    if(!identical(class(linfct),"character")){
        stop("Argument \'linfct\' must be of type character (or vector of character) \n")
    }

    ## ** find the names of the coefficients
    name.model <- names(object)
    n.model <- length(name.model)
    if(is.null(name.model)){
        stop("Incorrect argument  \'object\' \n",
             "Each element of the list must be named \n")
    }

    ls.coefname <- lapply(name.model, function(iModel){ ## list by model
        iResC <- createContrast(object[[iModel]],
                                linfct = character(0))
        return(colnames(iResC$contrast))
    })
    names(ls.coefname) <- name.model

    ls.object.coefname <- lapply(name.model, function(iModel){ ## list by model with model name
        paste0(iModel,": ", ls.coefname[[iModel]])
    })    
    names(ls.object.coefname) <- name.model
    
    object.coefname <- unname(unlist(ls.object.coefname)) ## vector
    n.coef <- length(object.coefname)
        
    ## ** create full contrast matrix
    if(length(linfct)==1){

        ## isolate left hand side of the linfct and remove multiplicative factor (e.g. 2*Age=0 -> Age)
        linfct.contrast <- createContrast(linfct, ...)
        linfct.coefname <- colnames(linfct.contrast$contrast)
        linfct.rhs <- as.double(linfct.contrast$null)
        linfct.factor <- as.double(linfct.contrast$contrast)
        
        ## name associated with each hypothesis
        name.linfct <- names(linfct)
        if(is.null(name.linfct)){
            name.linfct <- rownames(linfct.contrast$contrast)
        }
        ## generate contrast matrix with only 0
        out <- list(contrast = matrix(0, nrow = 0, ncol = n.coef,
                                      dimnames = list(NULL, object.coefname)),
                    null = numeric(0),
                    Q = 0)
        
        ## regressor corresponding to each coefficient
        tableCoef <- data.frame(name = unlist(lapply(object, function(iO){rownames(coef(iO, type = 9, labels = 2))})),
                                model = unlist(lapply(1:n.model, function(iO){rep(name.model[iO], NROW(coef(object[[iO]], type = 9, labels = 2)))})),
                                type = unlist(lapply(object, function(iO){attr(coef(iO, type = 9, labels = 2),"type")})),
                                to = unlist(lapply(object, function(iO){attr(coef(iO, type = 9, labels = 2),"var")})),
                                from = unlist(lapply(object, function(iO){attr(coef(iO, type = 9, labels = 2),"from")})))
        tableCoef <- tableCoef[tableCoef$type=="regression",]
        tableCoef$group <- interaction(tableCoef$model,tableCoef$to,drop=TRUE)
        tableCoef$group <- as.numeric(factor(tableCoef$group,unique(tableCoef$group)))

        ## fill contrast matrix
        for(iG in 1:max(tableCoef$group)){ ## iG <- 3
            if(all(linfct.coefname %in% tableCoef[tableCoef$group==iG,"from"])){

                iTemplate <- .createContrast(linfct, tableCoef[tableCoef$group==iG,"from"], ...)

                iRow <- matrix(0, nrow = 1, ncol = n.coef, dimnames = list(NULL,object.coefname))
                if(!is.null(rownames(iTemplate$contrast))){
                        rownames(iRow) <- paste0(unique(tableCoef[tableCoef$group==iG,"model"]),": ",name.linfct)
                }
                iRow[,paste0(tableCoef[tableCoef$group==iG,"model"],": ",tableCoef[tableCoef$group==iG,"name"])] <- unname(iTemplate$contrast)

                out$contrast <- rbind(out$contrast,iRow)
                out$null <- c(out$null, unname(iTemplate$null))
                out$Q <- out$Q+1

                tableCoef[tableCoef$group==iG,"contrast"] <- as.double(iTemplate$contrast)
            }
        }

    }else{
        out <- .createContrast(linfct, name.param = object.coefname, ...)
    }

    ## ** create contrast matrix relative to each model
    out$mlf <- lapply(name.model, function(iModel){ ## iModel <- name.model[1]
        ## only keep columns corresponding to coefficients belonging the the current model
        iContrast <- out$contrast[,ls.object.coefname[[iModel]],drop=FALSE]

        ## update name by removing the name of the model
        colnames(iContrast) <- ls.coefname[[iModel]]

        ## remove lines in the contrast matrix containing only 0
        if(NROW(iContrast)>0){
          index.n0 <- which(rowSums(iContrast!=0)!=0)
          return(iContrast[index.n0,,drop=FALSE])
        }else{
          return(iContrast)
        }
    })
    names(out$mlf) <- name.model    
    class(out$mlf) <- "mlf"

    ## remove right hand side from the names (like in multicomp)
    if(!is.null(list(...)$rowname.rhs) && list(...)$rowname.rhs==FALSE){
        out$mlf <- lapply(out$mlf, function(x){ ## x <- name.model[1]
            if(NROW(x)>0){
                if("sep" %in% names(list(...))){
                    rownames(x) <- .contrast2name(x, null = NULL, sep = list(...)$sep)
                }else{
                    rownames(x) <- .contrast2name(x, null = NULL)
                }
            }
            return(x)
        })
            
        class(out$mlf) <- "mlf"
    }
    names(out$null) <- rownames(out$contrast)
   
    ## ** export
    return(out)    
}

## * createContrast.mmm
#' @rdname createContrast
#' @export
createContrast.mmm <- createContrast.list

## * .createContrast
.createContrast <- function(linfct, name.param, diff.first = FALSE, add.rowname = TRUE, rowname.rhs = TRUE, sep = c("[","]"), ...){

    n.param <- length(name.param)
    dots <- list(...)
    if(length(dots)>0){
        txt.args <- paste(names(dots), collapse = "\" \"")
        txt.s <- if(length(dots)>1){"s"}else{""}
        txt.verb <- if(length(dots)>1){"are"}else{"is"}
        warning("Extra argument",txt.s," \"",txt.args,"\" ",txt.verb," ignored. \n")
    }

    if(diff.first){
        linfct <- paste0(linfct[-1]," - ",linfct[1])
    }

    n.hypo <- length(linfct)
    name.hypo <- names(linfct)
    if(any(nchar(linfct)==0)){
        stop("Argument contains empty character string(s) instead of an expression involving the model mean coefficients \n")
    }
    null <- rep(NA, n.hypo)
    contrast <- matrix(0, nrow = n.hypo, ncol = n.param,
                       dimnames = list(NULL,name.param))

    if(n.hypo>0){
        for(iH in 1:n.hypo){ # iH <- 1
            iTempo.eq <- strsplit(linfct[iH], split = "=", fixed = TRUE)[[1]]
            if(length(iTempo.eq)==1){ ## set null to 0 when second side of the equation is missing
                iTempo.eq <- c(iTempo.eq,"0")
            }

            null[iH] <- as.numeric(trim(iTempo.eq[2]))
            iRh.plus <- strsplit(iTempo.eq[[1]], split = "+", fixed = TRUE)[[1]]
            iRh <- trim(unlist(sapply(iRh.plus, strsplit, split = "-", fixed = TRUE)))
            iRh <- iRh[iRh!="",drop=FALSE]
                            
            ls.iRh <- lapply(strsplit(iRh, split = "*", fixed = TRUE), trim)
                    
            iN.tempo <- length(ls.iRh)

            for(iCoef in 1:iN.tempo){ # iCoef <- 1

                if(length(ls.iRh[[iCoef]])==1){
                    iFactor <- 1
                    iName <- ls.iRh[[iCoef]][1]                
                }else{
                    iFactor <- as.numeric(ls.iRh[[iCoef]][1])
                    iName <- ls.iRh[[iCoef]][2]
                }
            
                if(iName %in% name.param == FALSE){
                    txt.message <- paste0("unknown coefficient ",iName," in hypothesis ",iH,"\n")
                    possibleMatch <- grep(iName, name.param, fixed = TRUE, value = TRUE)
                    if(all(is.na(possibleMatch)) || length(possibleMatch)==0){
                        possibleMatch <- pmatch(iName, table = name.param)
                    }else if(length(linfct) == 1 && length(possibleMatch)>1){ ##
                        message("Guessing the contrast based on the string \"",linfct,"\" (",length(possibleMatch)," coefficients found). \n")
                        return(.createContrast(linfct = possibleMatch, name.param = name.param, diff.first = diff.first, add.rowname = add.rowname, rowname.rhs = rowname.rhs, sep = sep, ...))
                    }
                    
                    if(all(is.na(possibleMatch)) || length(possibleMatch)==0){
                        possibleMatch <- agrep(iName, name.param, ignore.case = TRUE,value = TRUE)
                    }
                    if(all(!is.na(possibleMatch)) && length(possibleMatch)>0){
                        txt.message <- c(txt.message,
                                         paste0("candidates: \"",paste(possibleMatch, collapse = "\" \""),"\"\n"))
                    }
                    stop(txt.message)                    
                    
                    
                    
                }

                ## identify if it is a minus sign
                iBeforeCoef <- strsplit(iTempo.eq[[1]], split = ls.iRh[iCoef])[[1]][1]
                if(iCoef > 1){
                    iBeforeCoef <- strsplit(iBeforeCoef, split = ls.iRh[iCoef-1])[[1]][2]
                }
                test.sign <- length(grep("-",iBeforeCoef))>0
                contrast[iH,iName] <- c(1,-1)[test.sign+1] * iFactor
            }
        }
        if(add.rowname){
            if(is.null(name.hypo)){
                name.hypo <- .contrast2name(contrast, null = if(rowname.rhs){null}else{NULL}, sep = sep)
            }
            rownames(contrast) <- name.hypo
            null <- stats::setNames(null, name.hypo)
        }
    }
    return(list(contrast = contrast,
                null = null,
                Q = n.hypo))
}

## * .contrast2name
#' @title Create Rownames for a Contrast Matrix
#' @description Create rownames for a contrast matrix using the coefficients and the names of the coefficients. The rownames will be [value * name] == null, e.g. [beta + 4*alpha] = 0.
#' @name contrast2name
#'
#' @param contrast [matrix] a contrast matrix defining the left hand side of the linear hypotheses to be tested.
#' @param null [vector, optional] the right hand side of the linear hypotheses to be tested.
#' @param sep [character of length 2, optional] character used in rownames to wrap the left hand side of the equation.
#'
#' @details When argument \code{NULL} is null then the rownames will not be put into brackets and the right hand side will not be added to the name.
#'
#' @return a character vector.
#' 
#' @keywords internal
.contrast2name <- function(contrast, null = NULL, sep = c("[","]")){
    contrast.names <- colnames(contrast)
    
    df.index <- as.data.frame(which(contrast != 0, arr.ind = TRUE))
    df.index$col <- contrast.names[df.index$col]
    index.order <- order(df.index$row,match(df.index$col,contrast.names))
    df.index <- df.index[index.order,]
    df.index$nrow <- unlist(tapply(df.index$row, df.index$row, function(x){1:length(x)}))

    ## find coef value to each  coefficient
    df.index$coef <- contrast[which(contrast!=0)][index.order]
    df.index$coefname <- as.character(df.index$coef)

    ## add positive sign
    df.index[df.index$coef>0,"coefname"] <- paste0("+",df.index$coefname[df.index$coef>0])
    
    ## add multiplicative symbol
    index.Npm1 <- which(df.index$coefname %in% c("","+1","-1") == FALSE)
    df.index[index.Npm1, "coefname"] <- paste0(df.index[index.Npm1, "coefname"],"*")

    ## simplify
    df.index[df.index$coefname == "+1" & df.index$nrow == 1, "coefname"] <- ""
    df.index[df.index$coefname == "+1", "coefname"] <- "+"            
    df.index[df.index$coefname == "-1", "coefname"] <- "-"
    df.index[df.index$nrow == 1, "coefname"] <- gsub("+","",df.index[df.index$nrow == 1, "coefname"], fixed = TRUE)

    ## add space between coefficients
    df.index$coefname <- gsub("+"," + ",df.index$coefname, fixed = TRUE)
    df.index$coefname <- gsub("-"," - ",df.index$coefname, fixed = TRUE)
    df.index[df.index$coefname == " - " & df.index$nrow == 1, "coefname"] <- "- "

    ## paste together value and coefficient names
    df.index$rowname <- paste0(df.index$coefname,df.index$col)
    ## paste together names from the same hypothesis
    out <- unlist(tapply(df.index$rowname,df.index$row,paste,collapse=""))

    ## add right hand side
    if(!is.null(null)){
        ## out <- paste0("[",out,"] = ",null)
        out <- paste0(sep[1],out,sep[2]," = ",null)
    }else if(!is.null(sep)){
        out <- paste0(sep[1],out,sep[2])
    }

    return(as.character(out))
}


##----------------------------------------------------------------------
### createContrast.R ends here

### compare2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan 30 2018 (14:33) 
## Version: 
## Last-Updated: jan 23 2024 (10:31) 
##           By: Brice Ozenne
##     Update #: 904
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - compare2
#' @title Test Linear Hypotheses With Small Sample Correction
#' @description Test Linear Hypotheses using Wald statistics in a latent variable model.
#' Similar to \code{lava::compare} but with small sample correction.
#' @name compare2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param linfct [matrix or vector of character] the linear hypotheses to be tested. Same as the argument \code{par} of \code{\link{createContrast}}.
#' @param rhs [vector] the right hand side of the linear hypotheses to be tested.
#' @param robust [logical] should the robust standard errors be used instead of the model based standard errors?
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param as.lava [logical] should the output be similar to the one return by \code{lava::compare}?
#' @param F.test [logical] should a joint test be performed?
#' @param conf.level [numeric 0-1] level of the confidence intervals.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details The \code{linfct} argument and \code{rhs} specify the set of linear hypotheses to be tested. They can be written:
#' \deqn{
#'   linfct * \theta = rhs
#' }
#' where \eqn{\theta} is the vector of the model coefficients. \cr
#' The \code{par} argument must contain expression(s) involving the model coefficients.
#' For example \code{"beta = 0"} or \code{c("-5*beta + alpha = 3","-alpha")} are valid expressions if alpha and beta belong to the set of model coefficients.
#' A contrast matrix and the right hand side will be generated inside the function. \cr
#' 
#' When directly specified, the contrast matrix must contain as many columns as there are coefficients in the model (mean and variance coefficients).
#' Each hypothesis correspond to a row in the contrast matrix. \cr
#'
#' The rhs vector should contain as many elements as there are row in the contrast matrix. \cr
#' 
#' @seealso \code{\link{createContrast}} to create contrast matrices. \cr
#' \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#' 
#' @return If \code{as.lava=TRUE} an object of class \code{htest}.
#' Otherwise a \code{data.frame} object.

## * example - compare2
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' mSim <- lvm(Y~0.1*X1+0.2*X2)
#' categorical(mSim, labels = c("a","b","c")) <- ~X1
#' transform(mSim, Id~Y) <- function(x){1:NROW(x)}
#' df.data <- lava::sim(mSim, 1e2)
#'
#' #### with lvm ####
#' m <- lvm(Y~X1+X2)
#' e.lvm <- estimate(m, df.data)
#' 
#' compare2(e.lvm, linfct = c("Y~X1b","Y~X1c","Y~X2"))
#' compare2(e.lvm, linfct = c("Y~X1b","Y~X1c","Y~X2"), robust = TRUE)
#' 
#' @concept inference
#' @keywords smallSampleCorrection
#' @export
`compare2` <-
    function(object, linfct, rhs,
             robust, cluster,
             as.lava, F.test,
             conf.level, ...) UseMethod("compare2")

## * compare2.lvmfit
#' @rdname compare2
#' @export
compare2.lvmfit <- function(object, linfct = NULL, rhs = NULL,
                            robust = FALSE, cluster = NULL,
                            as.lava = TRUE, F.test = TRUE,
                            conf.level = 0.95,
                            ssc = lava.options()$ssc, df = lava.options()$df, ...){

    return(compare(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...),
                    linfct = linfct, rhs = rhs, robust = robust, cluster = cluster, as.lava = as.lava, F.test = F.test, conf.level = conf.level)
           )

}

## * compare2.lvmfit2
#' @rdname compare2
#' @export
compare2.lvmfit2 <- function(object, linfct = NULL, rhs = NULL,
                              robust = FALSE, cluster = NULL,
                              as.lava = TRUE, F.test = TRUE,
                              conf.level = 0.95, ...){
   
    dots <- list(...)
    if(any(names(dots)=="par")){
        stop("Argument \'par\' is no longer used as it has been replaced by \'linfct\'. \n")
    }
    if(length(dots[names(dots) %in% "sep" == FALSE])>0){
        warning("Argument(s) \'",paste(setdiff(names(dots),"sep"),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }
    if(is.null(linfct)){ ## necessary for lava::gof to wo
        object0 <- object
        class(object0) <- setdiff(class(object0),"lvmfit2")
        return(lava::compare(object0))
    }
    if(!is.logical(robust)){ 
        stop("Argument \'robust\' should be TRUE or FALSE \n")
    }
    if(robust==FALSE && !is.null(cluster)){
        stop("Argument \'cluster\' must be NULL when argument \'robust\' is FALSE \n")
    }

    ## ** extract information
    df <- object$sCorrect$df
    
    ## 0-order: param
    param <- coef(object, as.lava = FALSE)
    n.param <- length(param)
    name.param <- names(param)

    ## 1-order: score
    if(robust){
        score <- score(object, cluster = cluster, as.lava = FALSE, indiv = TRUE)
    }else{
        score <- NULL
    }
    
    ## 2-order: variance covariance
    vcov.param <- vcov(object, as.lava = FALSE)
    warn <- attr(vcov.param, "warning")
    attr(vcov.param, "warning") <- NULL
    if(robust){
        rvcov.param <- vcov(object, robust = TRUE, cluster = cluster, as.lava = FALSE)
    }

    ## 3-order: derivative of the variance covariance matrices
    if(df == "satterthwaite"){
        dVcov.param <- object$sCorrect$dVcov.param[names(object$sCorrect$skeleton$originalLink2param),
                                                   names(object$sCorrect$skeleton$originalLink2param),
                                                   names(object$sCorrect$skeleton$originalLink2param),
                                                   drop=FALSE]
        dimnames(dVcov.param) <- list(as.character(object$sCorrect$skeleton$originalLink2param),
                                      as.character(object$sCorrect$skeleton$originalLink2param),
                                      as.character(object$sCorrect$skeleton$originalLink2param))
        keep.param <- dimnames(dVcov.param)[[3]]

        if(robust && (lava.options()$df.robust != 1)){
            if(!is.null(cluster) || is.null(object$sCorrect$dRvcov.param)){
                ## update derivative according to cluster
                hessian <- hessian2(object, cluster = cluster, as.lava = FALSE)
                dRvcov.param <- .dRvcov.param(score = score,
                                              hessian = hessian,
                                              vcov.param = vcov.param,
                                              dVcov.param = dVcov.param,
                                              n.param = n.param,
                                              name.param = name.param)
                                              
            }else{
                dRvcov.param <- object$sCorrect$dRvcov.param[names(object$sCorrect$skeleton$originalLink2param),
                                                             names(object$sCorrect$skeleton$originalLink2param),
                                                             names(object$sCorrect$skeleton$originalLink2param),
                                                             drop=FALSE]
                dimnames(dRvcov.param) <- list(as.character(object$sCorrect$skeleton$originalLink2param),
                                               as.character(object$sCorrect$skeleton$originalLink2param),
                                               as.character(object$sCorrect$skeleton$originalLink2param))
            }
        }
    }

    ## ** normalize linear hypotheses
    if(!is.matrix(linfct)){
        res.C <- createContrast(object, linfct = linfct, rowname.rhs = FALSE, ...)
        if(any(colnames(res.C$contrast)!=name.param) && all(colnames(res.C$contrast) == names(object$sCorrect$skeleton$originalLink2param))){
            colnames(res.C$contrast) <- as.character(object$sCorrect$skeleton$originalLink2param)
        }
        linfct <- res.C$contrast
        if(is.null(rhs)){
            rhs <- res.C$null
        }else{
            if(length(rhs)!=length(res.C$null)){
                stop("Incorrect argument \'rhs\' \n",
                     "Must have length ",length(res.C$null),"\n")
            }
            rhs <- stats::setNames(rhs, names(res.C$null))
        }
        name.hypoShort <- rownames(linfct)
        name.hypo <- paste0(name.hypoShort," = ",rhs)
    }else{
        if(is.null(colnames(linfct))){
            stop("Argument \'linfct\' must have column names \n")
        }
        if(NCOL(linfct) != n.param){
            stop("Argument \'linfct\' should be a matrix with ",n.param," columns \n")
        }
        if(any(colnames(linfct) %in% name.param == FALSE)){
            txt <- setdiff(colnames(linfct), name.param)
            stop("Argument \'linfct\' has incorrect column names \n",
                 "invalid name(s): \"",paste(txt, collapse = "\" \""),"\"\n")
        }
        if(any(name.param %in% colnames(linfct) == FALSE)){
            txt <- setdiff(name.param, colnames(linfct))
            stop("Argument \'linfct\' has incorrect column names \n",
                 "missing name(s): \"",paste(txt, collapse = "\" \""),"\"\n")
        }
        ## reorder columns according to coefficients
        linfct <- linfct[,name.param,drop=FALSE]
        if(F.test && any(abs(svd(linfct)$d)<1e-10)){
            stop("Argument \'linfct\' is singular \n")
        }
        if(is.null(rhs)){
            rhs <- stats::setNames(rep(0,NROW(linfct)),rownames(linfct))
        }else if(length(rhs)!=NROW(linfct)){
            stop("The length of argument \'rhs\' must match the number of rows of argument \'linfct' \n")
        }
        if(is.null(rownames(linfct))){
            rownames(linfct) <- .contrast2name(linfct, null = rhs)
            rhs <- stats::setNames(rhs, rownames(linfct))
        }
        name.hypo <- rownames(linfct)
        name.hypoShort <- sapply(strsplit(name.hypo, split = " = ", fixed = TRUE),"[[",1)
    }

    n.hypo <- length(name.hypo)
    linfct <- linfct[,names(param),drop=FALSE] ## column in contrast may not be in the same order as param

    ## ** Univariate Wald test
    ## coefficient (used for F.test and lava export)
    C.p <- linfct %*% param
    C.p.rhs <- C.p - rhs

    ## variance (used for F.test and lava export)
    if(robust){
        C.vcov.C <- linfct %*% rvcov.param %*% t(linfct)
    }else{
        C.vcov.C <- linfct %*% vcov.param %*% t(linfct)
    }

    ## df
    if(df == "satterthwaite"){
        df.Wald  <- dfSigma(contrast = linfct,
                            score = score,
                            vcov = vcov.param,
                            rvcov = rvcov.param,
                            dVcov = dVcov.param,
                            dRvcov = dRvcov.param,
                            keep.param = keep.param,                            
                            type = if(robust){lava.options()$df.robust}else{1})

        ##
        ## 2 * vcov.param["Y","Y"]^2 / (vcov.param["Y~~Y","Y~~Y"]*dVcov.param["Y","Y","Y~~Y"]^2)
        ## 
    }else{
        df.Wald <- rep(Inf, n.hypo)
    }

    ## ** Multivariate Wald test
    error <- NULL
    if(F.test){
        iC.vcov.C <- try(solve(C.vcov.C), silent = TRUE)
        if(!inherits(iC.vcov.C,"try-error")){
            stat.F <- t(C.p.rhs) %*% iC.vcov.C %*% (C.p.rhs) / n.hypo

            ## df (independent t statistics)
            if(df == "satterthwaite"){
                svd.tempo <- eigen(iC.vcov.C)
                D.svd <- diag(svd.tempo$values, nrow = n.hypo, ncol = n.hypo)
                P.svd <- svd.tempo$vectors
     
                C.anova <- sqrt(D.svd) %*% t(P.svd) %*% linfct

                nu_m  <- dfSigma(contrast = C.anova,
                                 score = score,
                                 vcov = vcov.param,
                                 rvcov = rvcov.param,
                                 dVcov = dVcov.param,
                                 dRvcov = dRvcov.param,
                                 keep.param = keep.param,                            
                                 type = if(robust){lava.options()$df.robust}else{1})
                EQ <- sum(nu_m/(nu_m-2))
                df.F <- 2*EQ / (EQ - n.hypo)
            }else{
                df.F <- Inf
            }
            ## store
            F.res <- c("statistic" = as.numeric(stat.F),
                       "df" = df.F,
                       "p.value" = 1 - stats::pf(stat.F,
                                                 df1 = n.hypo,
                                                 df2 = df.F)
                       )
        }else{
            warning("Unable to invert the variance-covariance matrix after application of the contrasts \n")
            error <- iC.vcov.C
        }
    }

    ## ** export
    if(as.lava == TRUE){
        level.inf <- (1-conf.level)/2
        level.sup <- 1-level.inf

        level.inf.label <- paste0(100*level.inf,"%")
        level.sup.label <- paste0(100*level.sup,"%")

        df.estimate <- matrix(NA, nrow = n.hypo, ncol = 5,
                              dimnames = list(name.hypoShort,c("Estimate", "Std.Err", "df", level.inf.label, level.sup.label)))
        df.estimate[,"Estimate"] <- C.p
        df.estimate[,"Std.Err"] <- sqrt(diag(C.vcov.C))
        df.estimate[,"df"] <- df.Wald
        df.estimate[,level.inf.label] <- df.estimate[,"Estimate"] + stats::qt(level.inf, df = df.estimate[,"df"]) * df.estimate[,"Std.Err"]
        df.estimate[,level.sup.label] <- df.estimate[,"Estimate"] + stats::qt(level.sup, df = df.estimate[,"df"]) * df.estimate[,"Std.Err"]

        dimnames(C.vcov.C) <- list(name.hypoShort,name.hypoShort)
        out <- list(statistic = stats::setNames(F.res["statistic"],"F-statistic"),
                    parameter = stats::setNames(round(F.res["df"],2), paste0("df1 = ",n.hypo,", df2")), ## NOTE: cannot not be change to coefficients because of lava
                    p.value = F.res["p.value"],
                    method = c("- Wald test -", "", "Null Hypothesis:", name.hypo),
                    estimate = df.estimate,
                    vcov = C.vcov.C,
                    coef = stats::setNames(C.p[,1], name.hypoShort),
                    null = stats::setNames(rhs, name.hypoShort),
                    cnames = name.hypo                    
                    )
        if(robust){
            colnames(out$estimate)[2] <- "robust SE"
        }
        rownames(linfct) <- name.hypo
        attr(out, "B") <- linfct
        class(out) <- "htest"
    }else{
        if(length(unique(df.Wald))==1){
            df.Wald <- df.Wald[1]
        }
        out <- list(model = object,
                    linfct = linfct,
                    rhs = unname(rhs),
                    coef = param,
                    vcov = if(robust){rvcov.param}else{vcov.param},
                    df = df.Wald,
                    alternative = "two.sided",
                    type = NULL,
                    robust = robust,
                    ssc = object$sCorrect$ssc$type,
                    global = if(F.test){F.res}else{NULL})
        class(out) <- c("glht2","glht")
    }
    attr(out,"warning") <- warn
    attr(out,"error") <- error
    return(out)

}

## * compare.lvmfit2
#' @rdname compare2
#' @export
compare.lvmfit2 <- compare2.lvmfit2

## * dfSigma
##' @title Degree of Freedom for the Chi-Square Test
##' @description Computation of the degrees of freedom of the chi-squared distribution
##' relative to the model-based variance
##'
##' @param contrast [numeric vector] the linear combination of parameters to test
##' @param score [numeric matrix] the individual score for each parameter.
##' @param vcov [numeric matrix] the model-based variance-covariance matrix of the parameters.
##' @param rvcov [numeric matrix] the robust variance-covariance matrix of the parameters.
##' @param dVcov [numeric array] the first derivative of the model-based variance-covariance matrix of the parameters.
##' @param dRvcov [numeric array] the first derivative of the robust variance-covariance matrix of the parameters.
##' @param keep.param [character vector] the name of the parameters with non-zero first derivative of their variance parameter.
##' @param type [integer] 1 corresponds to the Satterthwaite approximation of the the degrees of freedom applied to the model-based variance,
##' 2 to the Satterthwaite approximation of the the degrees of freedom applied to the robust variance,
##' 3 to the approximation described in (Pan, 2002) section 2 and 3.1.
##'
##' @references
##' Wei Pan and Melanie M. Wall, Small-sample adjustments in using the sandwich variance estiamtor in generalized estimating equations. Statistics in medicine (2002) 21:1429-1441.
##' 
dfSigma <- function(contrast, score, vcov, rvcov, dVcov, dRvcov, keep.param, type){
    if(type==1){
        C.vcov.C <- rowSums(contrast %*% vcov * contrast) ## variance matrix of the linear combination
        C.dVcov.C <- sapply(keep.param, function(x){
            rowSums(contrast %*% dVcov[,,x] * contrast)
        })
        numerator <- 2 *(C.vcov.C)^2
        denom <- rowSums(C.dVcov.C %*% vcov[keep.param,keep.param,drop=FALSE] * C.dVcov.C)
        df <- numerator/denom
    }else if(type==2){
        C.rvcov.C <- rowSums(contrast %*% rvcov * contrast) ## variance matrix of the linear combination
        C.dRvcov.C <- sapply(keep.param, function(x){
            rowSums(contrast %*% dRvcov[,,x] * contrast)
        })
        numerator <- 2 *(C.rvcov.C)^2
        denom <- rowSums(C.dRvcov.C %*% rvcov[keep.param,keep.param,drop=FALSE] * C.dRvcov.C)
        df <- numerator/denom
    }else if(type==3){
        vcov.S <- contrast %*% vcov
        index.var <- diag(matrix(1:NROW(contrast)^2,NROW(contrast),NROW(contrast)))
        
        K <- NROW(score)
        ls.Pi <- lapply(1:K, function(iC){as.double(tcrossprod(score[iC,]))})
        M.Pi <- do.call(rbind,ls.Pi)
        M.Pi_center <- sweep(M.Pi, MARGIN = 2, STATS = colMeans(M.Pi), FUN = "-")
        ## M.Pi_center - M.Pi
        T <- t(M.Pi_center) %*% M.Pi_center / (K*(K-1))
        ## range(var(M.Pi)/K - T)
        eq.3 <- K^2 * (vcov.S %x% vcov.S) %*% T %*%  (vcov.S %x% vcov.S)

        Vs <- (vcov.S %x% vcov.S) %*% Reduce("+",ls.Pi)
        ## range(Vs - as.double(rvcov))
        ## range(Vs[index.var] - diag(rvcov))
        df <- 2*Vs[index.var]^2/sapply(index.var, function(iIndex){eq.3[iIndex,iIndex]})
        
    }
    
    return(stats::setNames(df, rownames(contrast)))
}


##----------------------------------------------------------------------
### compare2.R ends here

### test-compare2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 20 2017 (10:22) 
## Version: 
## last-updated: jan 23 2024 (11:31) 
##           By: Brice Ozenne
##     Update #: 237
#----------------------------------------------------------------------
## 
### Commentary: 
## Test battery:
## - linear regression:
##   compare2.lvm and compare2.lm
##   * check whether df and sigma match with what is expected
##     with and without small sample correction
##
## - mixed model (compound symmetry)
##   compare2.lvm, compare2.gls, and compare2.lme
##   * compare Satterthwaite correction with lmerTest (Wald test)
##   * compare with lava::compare (F-test)
##   * small sample correction: assess internal consistency between compare2.lvm, compare2.gls, and compare2.lme
##
## - mixed model (unstructured)
##   compare2.lvm, compare2.gls, and compare2.lme
##   * Satterthwaite correction: consistency between using or not numerical derivative
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

library(clubSandwich)
library(nlme)
library(lme4)
library(lmerTest)
library(pbkrtest)
lava.options(symbols = c("~","~~"))

context("compare2")

## * simulation
n <- 5e1
mSim <- lvm(c(Y1~eta1,Y2~eta1+X2,Y3~eta1+X1,
              Z1~eta2,Z2~eta2,Z3~eta2+X3))
regression(mSim) <- eta1~X1+Gender
latent(mSim) <- ~eta1+eta2
categorical(mSim, labels = c("Male","Female")) <- ~Gender
transform(mSim, Id~Y1) <- function(x){1:NROW(x)}
set.seed(10)
d <- lava::sim(mSim, n = n, latent = FALSE)
dL <- reshape2::melt(d, id.vars = c("Id","X1","X2","X3","Gender"),
                     measure.vars = c("Y1","Y2","Y3","Z1","Z2","Z3"))
dLred <- dL[dL$variable %in% c("Y1","Y2","Y3"),]

## * linear regression

e.lvm <- estimate(lvm(Y1~X1+X2), data = d)
e.lm <- lm(Y1~X1+X2, data = d)
e.gls <- nlme::gls(Y1~X1+X2, data = d, method = "ML")

## vcov(e.lvm)

### ** clubSandwich
cS.vcov <- vcovCR(e.lm, type = "CR0", cluster = d$Id)
cS.df <- coef_test(e.lm, vcov = cS.vcov, test = "Satterthwaite", cluster = 1:NROW(d))
cS.df
## cS.df$df is very suspect: should be the same for all coefficient and close to n-p

### ** compare2
test_that("linear regression: df",{
    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(e.lvm, linfct = name.param, ssc = "none", as.lava = FALSE)
        
    ## test value
    n.param <- length(coef(e.lm))
    df.GS <- c(rep(n,n.param), n/4)
    expect_equal(unname(df.lvm$df), df.GS)

    sigma2 <- coef(e.lvm)["Y1~~Y1"]
    iXX <- solve(crossprod(model.matrix(e.lm)))
    std.GS <- c(sqrt(diag(iXX*sigma2)),sqrt(2*sigma2^2/e.lvm$data$n))
    expect_equal(unname(sqrt(diag(df.lvm$vcov))), unname(std.GS))
})

test_that("linear regression: df adjusted",{
    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(e.lvm, linfct = name.param, as.lava = FALSE)

    ## test value
    n.param <- length(coef(e.lvm))-1
    df.GS <- c(rep(n-n.param,n.param), (n-n.param)/4)
    expect_equal(unname(df.lvm$df), df.GS)

    sigma2 <- sigma(e.lm)^2
    iXX <- solve(crossprod(model.matrix(e.lm)))
    std.GS <- c(sqrt(diag(iXX*sigma2)),sqrt(2*sigma2^2/(n-n.param)))
    expect_equal(unname(sqrt(diag(df.lvm$vcov))), unname(std.GS))
})

## * multiple linear regression lvm
## ** model fit
ls.lm <- list(lm(Y1~X1,d),lm(Y2~X2,d),lm(Y3~X1+X3,d))
e.lvm <- estimate(lvm(Y1~X1,Y2~X2,Y3~X1+X3), data = d)

## e.lvm2 <- estimate(lvm(Y1[mu:sigma1]~ beta1*X1 + beta2*X2,
##                        Y2[mu:sigma2]~ beta1*X1 + beta2*X2,
##                        Y3[mu:sigma3]~ beta1*X1 + beta2*X2),
##                        data = d)
## e.gls <- gls(value ~ X1 + X2,
##              data = dLred,
##              weight = varIdent(form = ~1|variable),
##              method = "ML")

## test_that("gls equivalent to lvm", {
##     expect_equal(as.double(logLik(e.lvm2)), as.double(logLik(e.gls)))
## })

## ** compare2
test_that("multiple linear regression: df adjusted",{
    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(estimate2(e.lvm), linfct = name.param, as.lava = FALSE, sep = c("",""))

    ## 
    X <- list(as.matrix(cbind(1,d[,c("X1")])),
              as.matrix(cbind(1,d[,c("X2")])),
              as.matrix(cbind(1,d[,c("X1","X3")])))
    sigma2 <- list(sigma(lm(Y1~X1,data=d))^2,
                   sigma(lm(Y2~X2,data=d))^2,
                   sigma(lm(Y3~X1+X3,data=d))^2)
    std.GS <- mapply(X, sigma2, FUN = function(x,y){
        c(sqrt(diag(solve(crossprod(x))*y)),sqrt(2*y^2/(n-NCOL(x))))
    })

    name.coef.lvm <- names(coef(e.lvm))
    expect_equal(unname(sqrt(diag(df.lvm$vcov))[c("Y1","Y1~X1","Y1~~Y1")]), unname(std.GS[[1]]), tol = 1e-7)
    expect_equal(unname(sqrt(diag(df.lvm$vcov))[c("Y2","Y2~X2","Y2~~Y2")]), unname(std.GS[[2]]), tol = 1e-7)
    expect_equal(unname(sqrt(diag(df.lvm$vcov))[c("Y3","Y3~X1","Y3~X3","Y3~~Y3")]), unname(std.GS[[3]]), tol = 1e-7)

    ## test value
    df.GS <- lapply(X, function(x){
        c(rep(n - NCOL(x),NCOL(x)),
          (n - NCOL(x))/4)
    })
    expect_equal(unname(df.lvm$df[c("Y1","Y1~X1","Y1~~Y1")]), unname(df.GS[[1]]), tol = 1e-7)
    expect_equal(unname(df.lvm$df[c("Y2","Y2~X2","Y2~~Y2")]), unname(df.GS[[2]]), tol = 1e-7)
    expect_equal(unname(df.lvm$df[c("Y3","Y3~X1","Y3~X3","Y3~~Y3")]), unname(df.GS[[3]]), tol = 1e-7)

})

## * Mixed model: Compound symmetry
m <- lvm(c(Y1[mu1:sigma]~1*eta,
           Y2[mu2:sigma]~1*eta,
           Y3[mu3:sigma]~1*eta,
           eta~X1+Gender)) 
e.lvm <- estimate(m, d)
## compare2(e.lvm)

## e.lmer <- lmerTest::lmer(value ~ variable + X1 + Gender + (1|Id),
##                           data = dLred, REML = FALSE)
## e2.lmer <- update(e.lmer, REML = TRUE)
e.lme <- nlme::lme(value ~ variable + X1 + Gender, random = ~ 1|Id,
                   data = dLred, method = "ML")

e.gls <- nlme::gls(value ~ variable + X1 + Gender,
                   correlation = corCompSymm(form=~ 1|Id),
                   data = dLred, method = "ML")

## ** clubSandwich - bug
## expect_equal(logLik(e.lmer),logLik(e.lme))
expect_equal(-259.840820322626, as.double(logLik(e.lme)))
coef_test(e.lme, vcov = "CR0", test = "Satterthwaite", cluster = dLred$Id)
## strange that same type of coef have very different degrees of freedom

## ** compare 
expect_equal(-259.840820322626,as.double(logLik(e.lvm)))

test_that("mixed model: Satterthwaite ",{

    ## does not work when running test
    ## GS <- summary(e.lmer)$coef[,c("df","t value")]
    GS <- matrix(c(91.83528558, 100.00000017, 100.00000016, 49.99999976, 49.99999977, -1.00221818, 0.63398721, 1.67029848, 10.09702494, 3.08796567), 
                 nrow = 5, 
                 ncol = 2, 
                 dimnames = list(c("(Intercept)", "variableY2", "variableY3", "X1", "GenderFemale"),c("df", "t value")) 
                 ) 

    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(e.lvm, linfct = name.param, ssc = "none", as.lava = FALSE)
    expect_equal(as.double(GS[,"df"]),
                 as.double(df.lvm$df[1:5]), tol = 1e-4) ## needed for CRAN
    expect_equal(as.double(GS[,"t value"]),
                 as.double(summary(df.lvm, test = multcomp::adjusted("none"))$table2[1:5,"statistic"]), tol = 1e-8) ## needed for CRAN

    ## F test
    ## GS <- lmerTest::contestMD(e.lmer, L = diag(1,5,5), rhs = 0, ddf = "Satterthwaite")
    GS <- data.frame("Sum Sq" = c(166.27553944),"Mean Sq" = c(33.25510789),"NumDF" = c(5),"DenDF" = c(76.52281817),"F value" = c(23.33447979),"Pr(>F)" = c(0),
                     check.names = FALSE)
    name.param <- names(coef(e.lvm))    
    df.F <- compare2(e.lvm, linfct = name.param[1:5], ssc = "none", as.lava = FALSE, F.test = TRUE)
    
    expect_equal(GS[["DenDF"]], unname(df.F$global["df"]), tol = 1e-5)
    expect_equal(GS[["F value"]], unname(df.F$global["statistic"]), tol = 1e-8)
})

test_that("mixed model: KR-like correction",{

    ## does not work when running test
    ## GS <- summary(e2.lmer, ddf = "Kenward-Roger")$coef[,c("df","t value")]
    ## get_Lb_ddf(e2.lmer, c(0,1,0,0,0))
    ## get_Lb_ddf(e2.lmer, c(0,0,0,1,0))
    GS <- matrix(c(85.06326979, 98, 98, 47, 47, -0.97751931, 0.62761532, 1.65351114, 9.78942877, 2.99389376), 
                 nrow = 5, 
                 ncol = 2, 
                 dimnames = list(c("(Intercept)", "variableY2", "variableY3", "X1", "GenderFemale"),c("df", "t value")) 
                 ) 
    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(e.lvm, linfct = name.param, as.lava = FALSE)

    ## expect_equal(as.double(GS[,"df"]),
    ##              as.double(df.lvm$df[1:5]), tol = 1) ## difference of 1 or 2 df
    expect_equal(as.double(GS[,"t value"]),
                 as.double(summary(df.lvm, test = multcomp::adjusted("none"))$table2[1:5,"statistic"]), tol = 1e-5) ## needed for CRAN

    previous.value <- data.frame("estimate" = c(-0.25588154, 0.15137028, 0.39879913, 1.48076547, 0.92411608, 1.45423356, 0.6594628), 
                                 "se" = c(0.26176621, 0.24118321, 0.24118321, 0.15126168, 0.30866695, 0.20917549, 0.24297288), 
                                 "df" = c(87.2184581, 96.66666667, 96.66666667, 48.33333333, 48.33333333, 24.16666667, 14.29181404), 
                                 "lower" = c(-0.77615186, -0.32733248, -0.07990363, 1.17668766, 0.30361014, 1.02267408, 0.13933428), 
                                 "upper" = c(0.26438879, 0.63007304, 0.87750189, 1.78484328, 1.54462201, 1.88579304, 1.17959131), 
                                 "statistic" = c(-0.97751933, 0.62761531, 1.65351113, 9.78942913, 2.99389387, 6.95221787, 2.71414157), 
                                 "p.value" = c(0.3310157, 0.53173576, 0.10147095, 0, 0.00433193, 3.3e-07, 0.01654271))
    expect_equivalent(previous.value, summary(df.lvm, test = multcomp::adjusted("none"))$table2, tol = 1e-5)
})

### ** compare to SAS
if(FALSE){
    ## setwd("c:/Users/hpl802/AppData/Roaming/R")
    write.table(dLred, file = "mydata.txt", row.names = FALSE)
    ## /* Define path */
    ## %Let NomEtude = %Str(C:\Users\hpl802\AppData\Roaming\R\);

    ## /* Define path to file */
    ## FILENAME Fichier "&NomEtude.%Str(mydata.txt)";

    ## /* Importation of the data */
    ## Data mydata; 
    ## Infile Fichier FirstObs=2 obs=61; /* if no FirstObs : will read all first lines */
    ## input gpr $ animal $ week weight; 
    ## Run;

    ## /* display data */
    ## PROC SGPANEL DATA=mydata;
    ## PANELBY gpr;
    ## SERIES X=week Y=weight / GROUP=animal;
    ## RUN;

    ## /* Fit mixed model */
    ## PROC Mixed DATA = mydata;
    ## Class gpr animal week ;
    ## Model weight = week gpr*week / SOLUTION DDFM=KR;
    ## Repeated week / SUBJECT = animal TYPE = CS R RCORR;
    ## RUN; 
}

## * Mixed model: Unstructured with different variance
m <- lvm(Y1~1*eta,
         Y2~1*eta,
         Y3~1*eta,
         eta~X1+Gender)
covariance(m) <- Y1~Y2
covariance(m) <- Y1~Y3
e.lvm <- estimate(m, d)

e.lme <- lme(value ~ variable + X1 + Gender,
             random =~ 1|Id,
             correlation = corSymm(),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

e.gls <- gls(value ~ variable + X1 + Gender,
             correlation = corSymm(form=~ 1|Id),
             weights = varIdent(form =~ 1|variable),
             data = dLred,
             method = "ML")

test_that("lme/gls equivalent to lvm", {
    expect_equal(as.double(logLik(e.lvm)), as.double(logLik(e.lme)))
    expect_equal(as.double(logLik(e.gls)), as.double(logLik(e.lme)))
})

## ** compare
test_that("UN mixed model: df",{
    name.param <- names(coef(e.lvm))
    df.lvm <- compare2(e.lvm, linfct = name.param, ssc = "none", as.lava = FALSE)

    ##                          estimate       std  statistic       df      p-value
    ## [eta] = 0              -0.2530247 0.2459609 -1.0287194 61.99195 3.076087e-01
    ## [Y2] = 0                0.1513703 0.2248199  0.6732956 50.00000 5.038597e-01
    ## [Y3] = 0                0.3987991 0.2286753  1.7439534 50.00000 8.731285e-02
    ## [eta~X1] = 0            1.4498392 0.1465743  9.8914990 50.00000 2.318146e-13
    ## [eta~GenderFemale] = 0  0.9213738 0.2991017  3.0804696 50.00000 3.355219e-03
    ## [Y1~~Y1] = 0            1.3533853 0.4323206  3.1305133 29.40589 3.923193e-03
    ## [eta~~eta] = 0          0.4391486 0.3092283  1.4201436 21.64808 1.698083e-01
    ## [Y2~~Y2] = 0            1.6200992 0.4392003  3.6887476 13.94845 2.444719e-03
    ## [Y3~~Y3] = 0            1.7889734 0.4646774  3.8499254 13.62954 1.850575e-03
    ## [Y1~~Y2] = 0            0.2231421 0.3296648  0.6768757 24.12389 5.049242e-01
    ## [Y1~~Y3] = 0            0.2638691 0.3376548  0.7814760 23.84905 4.422119e-01
    ## global                         NA        NA 17.0357449 34.39628 7.882273e-11

    previous.value <- data.frame("estimate" = c(-0.25302471, 0.15137028, 0.39879913, 1.4498392, 0.92137382, 1.35338532, 0.43914858, 1.62009916, 1.78897339, 0.22314209, 0.26386911), 
                                 "se" = c(0.24596086, 0.22481994, 0.22867534, 0.14657427, 0.29910174, 0.43232057, 0.30922829, 0.43920032, 0.46467742, 0.32966479, 0.33765478), 
                                 "df" = c(61.99194948, 50, 50, 50, 50, 29.40589311, 21.64808397, 13.94844762, 13.62954237, 24.12389256, 23.8490532), 
                                 "lower" = c(-0.74469474, -0.30019386, -0.0605088, 1.15543612, 0.3206103, 0.46972029, -0.20275661, 0.67778142, 0.78979205, -0.45706773, -0.43324959), 
                                 "upper" = c(0.23864531, 0.60293442, 0.85810706, 1.74424227, 1.52213734, 2.23705035, 1.08105378, 2.5624169, 2.78815472, 0.90335192, 0.9609878), 
                                 "statistic" = c(-1.02871942, 0.6732956, 1.74395338, 9.89149902, 3.08046963, 3.13051333, 1.42014364, 3.68874764, 3.84992537, 0.67687572, 0.78147601), 
                                 "p.value" = c(0.30760871, 0.50385966, 0.08731285, 0, 0.00335522, 0.00392319, 0.16980825, 0.00244472, 0.00185058, 0.50492415, 0.44221192))
    expect_equivalent(previous.value, summary(df.lvm, test = multcomp::adjusted("none"))$table2, tol = 1e-5)
})


#----------------------------------------------------------------------
### test-compare2.R ends here

### compare2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan 30 2018 (14:33) 
## Version: 
## Last-Updated: mar  7 2019 (11:41) 
##           By: Brice Ozenne
##     Update #: 592
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - compare2
#' @title Test Linear Hypotheses with small sample correction
#' @description Test Linear Hypotheses using a multivariate Wald statistic.
#' Similar to \code{lava::compare} but with small sample correction.
#' @name compare2
#'
#' @param object an object that inherits from lm/gls/lme/lvmfit.
#' @param df [logical] should the degree of freedoms of the Wald statistic be computed using the Satterthwaite correction?
#' Otherwise the degree of freedoms are set to \code{Inf}, i.e. a normal distribution is used instead of a Student's t distribution when computing the p-values.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias? Argument passed to \code{sCorrect}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param par [vector of characters] expression defining the linear hypotheses to be tested.
#' See the examples section. 
#' @param contrast [matrix] a contrast matrix defining the left hand side of the linear hypotheses to be tested.
#' @param robust [logical] should the robust standard errors be used instead of the model based standard errors?
#' @param null,rhs [vector] the right hand side of the linear hypotheses to be tested.
#' @param as.lava [logical] should the output be similar to the one return by \code{lava::compare}?
#' @param F.test [logical] should a joint test be performed?
#' @param level [numeric 0-1] the confidence level of the confidence interval.
#' @param ...  [internal] only used by the generic method.
#'
#' @details The \code{par} argument or the arguments \code{contrast} and \code{null} (or equivalenty \code{rhs})
#' specify the set of linear hypotheses to be tested. They can be written:
#' \deqn{
#'   contrast * \theta = null
#' }
#' where \eqn{\theta} is the vector of the model coefficients. \cr
#' The \code{par} argument must contain expression(s) involving the model coefficients.
#' For example \code{"beta = 0"} or \code{c("-5*beta + alpha = 3","-alpha")} are valid expressions if alpha and beta belong to the set of model coefficients.
#' A contrast matrix and the right hand side will be generated inside the function. \cr
#' 
#' When directly specified, the contrast matrix must contain as many columns as there are coefficients in the model (mean and variance coefficients).
#' Each hypothesis correspond to a row in the contrast matrix. \cr
#'
#' The null vector should contain as many elements as there are row in the contrast matrix. \cr
#' 
#' Argument rhs and null are equivalent.
#' This redondance enable compatibility between \code{lava::compare}, \code{compare2}, \code{multcomp::glht}, and \code{glht2}.
#'
#' @seealso \code{\link{createContrast}} to create contrast matrices. \cr
#' \code{\link{sCorrect}} to pre-compute quantities for the small sample correction.
#' 
#' @return If \code{as.lava=TRUE} an object of class \code{htest}.
#' Otherwise a \code{data.frame} object.

## * example - compare2
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' mSim <- lvm(Y~0.1*X1+0.2*X2)
#' categorical(mSim, labels = c("a","b","c")) <- ~X1
#' transform(mSim, Id~Y) <- function(x){1:NROW(x)}
#' df.data <- lava::sim(mSim, 1e2)
#'
#' #### with lm ####
#' ## direct use of compare2
#' e.lm <- lm(Y~X1+X2, data = df.data)
#' anova(e.lm)
#' compare2(e.lm, par = c("X1b=0","X1c=0"))
#' 
#' ## or first compute the derivative of the information matrix
#' sCorrect(e.lm) <- TRUE
#' 
#' ## and define the contrast matrix
#' C <- createContrast(e.lm, par = c("X1b=0","X1c=0"), add.variance = TRUE)
#'
#' ## run compare2
#' compare2(e.lm, contrast = C$contrast, null = C$null)
#' compare2(e.lm, contrast = C$contrast, null = C$null, robust = TRUE)
#' 
#' #### with gls ####
#' library(nlme)
#' e.gls <- gls(Y~X1+X2, data = df.data, method = "ML")
#'
#' ## first compute the derivative of the information matrix
#' sCorrect(e.gls, cluster = 1:NROW(df.data)) <- TRUE
#' 
#' compare2(e.gls, par = c("5*X1b+2*X2 = 0","(Intercept) = 0"))
#' 
#' #### with lvm ####
#' m <- lvm(Y~X1+X2)
#' e.lvm <- estimate(m, df.data)
#' 
#' compare2(e.lvm, par = c("-Y","Y~X1b+Y~X1c"))
#' compare2(e.lvm, par = c("-Y","Y~X1b+Y~X1c"), robust = TRUE)
#' @concept small sample inference
#' @export
`compare2` <-
  function(object, df, bias.correct, ...) UseMethod("compare2")

## * compare2.lm
#' @rdname compare2
#' @export
compare2.lm <- function(object, df = TRUE, bias.correct = TRUE, ...){
    sCorrect(object, df = df) <- bias.correct
    return(.compare2(object, ...))
}

## * compare2.gls
#' @rdname compare2
#' @export
compare2.gls <- function(object, df = TRUE, bias.correct = TRUE, cluster = NULL, ...){
    sCorrect(object, df = df, cluster = cluster) <- bias.correct
    return(.compare2(object, ...))
}

## * compare2.lme
#' @rdname compare2
#' @export
compare2.lme <- compare2.lm

## * compare2.lvmfit
#' @rdname compare2
#' @export
compare2.lvmfit <- function(object, df = TRUE, bias.correct = TRUE, cluster = NULL, ...){
    sCorrect(object, df = df) <- bias.correct
    return(.compare2(object, cluster = cluster, ...))
}

## * compare2.lm2
#' @rdname compare2
#' @export
compare2.lm2 <- function(object, ...){
    return(.compare2(object, ...))
}

## * compare2.gls2
#' @rdname compare2
#' @export
compare2.gls2 <- function(object, ...){
    return(.compare2(object, ...))
}

## * compare2.lme2
#' @rdname compare2
#' @export
compare2.lme2 <- function(object, ...){
    return(.compare2(object, ...))
}

## * compare2.lvmfit2
#' @rdname compare2
#' @export
compare2.lvmfit2 <- function(object, ...){
    return(.compare2(object, ...))
}

## * .compare2
#' @rdname compare2
.compare2 <- function(object, par = NULL, contrast = NULL, null = NULL, rhs = NULL,
                      robust = FALSE, cluster = NULL, df = object$sCorrect$args$df,
                      as.lava = TRUE, F.test = TRUE, level = 0.95){


    if(!is.null(null) && !is.null(rhs)){
        stop("Arguments \'null\' and \'rhs\' should not be both specified \n")
    }
    if(!is.logical(robust)){ 
        stop("Argument \'robust\' should be TRUE or FALSE \n")
    }
    if(!is.logical(df) && (robust == FALSE || df %in% c(0:3) == FALSE)){     ## 2-3 hidden values
        stop("Argument \'df\' should be TRUE or FALSE \n")
    }

    if(robust){
        factor.dRvcov <- lava.options()$factor.dRvcov

        if(!is.null(cluster)){
            
            if(length(cluster)==1){
                ## reconstruct cluster variable
                if(inherits(object,"lvmfit")){
                    data <- object$data$model.frame
                }else{
                    data <- extractData(object)
                }
                
                if(cluster %in%  names(data) == FALSE){
                    stop("Could not find variable ",cluster," (argument \'cluster\') in argument \'data\' \n")
                }else{
                    cluster <- data[[cluster]]
                }            
                
            }else if(stats::nobs(object)!=length(cluster)){
                stop("length of argument \'cluster\' does not match number of rows of the score matrix \n")
            }
            ls.indexCluster <- tapply(1:length(cluster),cluster,list)
            n.cluster <- length(ls.indexCluster)
        }else{
            n.cluster <- stats::nobs(object)
        }
    }

    
    ## ** extract information
    ## 0-order: param
    param <- object$sCorrect$param

    n.param <- length(param)
    name.param <- names(param)

    ## 1-order: score
    if(robust){
        score <- object$sCorrect$score
    }
    
    ## 2-order: variance covariance
    vcov.param <- vcov2(object)
    attr(vcov.param, "warning") <- NULL
    warn <- attr(vcov2(object), "warning")
    
    if(robust){
        rvcov.param <- crossprod(iid2(object, cluster = cluster))
        hessian <- object$sCorrect$hessian
    }

    ## 3-order: derivative of the variance covariance
    if(df>0){
        dVcov.param <- object$sCorrect$dVcov.param
        keep.param <- dimnames(dVcov.param)[[3]]
    }
    
    ## ** Prepare for the robust case 
    if(df>1 && robust){ ## not used if df=1

        ## update the score/hessian/derivative at the cluster level
        if(!is.null(cluster)){            
            scoreSave <- score
            hessianSave <- hessian

            score <- matrix(NA, nrow = n.cluster, ncol = NCOL(score),
                            dimnames = list(NULL, colnames(score)))
            hessian <- array(NA, dim = c(NCOL(score), NCOL(score), n.cluster),
                             dimnames = list(colnames(score), colnames(score), NULL))            
            for(iCluster in 1:n.cluster){ ## iCluster <- 1
                score[iCluster,] <- colSums(scoreSave[ls.indexCluster[[iCluster]],,drop=FALSE])
                hessian[,,iCluster] <- apply(hessianSave[,,ls.indexCluster[[iCluster]],drop=FALSE],1:2,sum)
            }
            ## compute derivative
            name.3deriv <- dimnames(dVcov.param)[[3]]
            dRvcov.param <- array(NA, dim = c(n.param,n.param,n.param), dimnames = list(name.param,name.param,name.param))
            for(iP in 1:n.param){ ## iP <- 1
                ## if(name.param[iP] %in% name.3deriv){
                    ## term1 <- dVcov.param[,,name.param[iP]] %*% crossprod(score) %*% vcov.param
                ## }else{
                    ## term1 <- matrix(0, nrow = n.param, ncol = n.param)
                ## }
                ## term2 <- vcov.param %*% hessian[iP,,] %*% score %*% vcov.param
                ## dRvcov.param[,,iP] <- term1 + t(term1) + term2 + t(term2)

                term2 <- vcov.param %*% hessian[iP,,] %*% score %*% vcov.param
                dRvcov.param[,,iP] <- term2 + t(term2)
            }
        }else{
            dRvcov.param <- object$sCorrect$dRvcov.param
        }
    }
    
    ### ** normalize linear hypotheses
    if(!is.null(par)){
        
        if(!is.null(contrast)){
            stop("Argument \'par\' and argument \'contrast\' should not simultaneously specified")
        }else if(!is.null(null)){
            stop("Argument \'par\' and argument \'null\' should not simultaneously specified")
        }else{
            res.C <- createContrast(par, name.param = name.param, add.rowname = TRUE)
            contrast <- res.C$contrast
            null <- res.C$null
        }
        
    }else{
        
        if(is.null(contrast)){
            stop("Argument \'contrast\' and argument \'par\' cannot be both NULL \n",
                 "Please specify the null hypotheses using one of the two arguments \n")
        }
        if(is.null(colnames(contrast))){
            stop("Argument \'contrast\' must have column names \n")
        }
        if(any(colnames(contrast) %in% name.param == FALSE)){
            txt <- setdiff(colnames(contrast), name.param)
            stop("Argument \'contrast\' has incorrect column names \n",
                 "invalid name(s): \"",paste(txt, collapse = "\" \""),"\"\n")
        }
        if(any(name.param %in% colnames(contrast) == FALSE)){
            txt <- setdiff(name.param, colnames(contrast))
            stop("Argument \'contrast\' has incorrect column names \n",
                 "missing name(s): \"",paste(txt, collapse = "\" \""),"\"\n")
        }
        if(NCOL(contrast) != n.param){
            stop("Argument \'contrast\' should be a matrix with ",n.param," columns \n")
        }
        ## reorder columns according to coefficients
        contrast <- contrast[,name.param,drop=FALSE]
        if(any(abs(svd(contrast)$d)<1e-10)){
            stop("Argument \'contrast\' is singular \n")
        }
        if(is.null(null)){
            null <- setNames(rep(0,NROW(contrast)),rownames(contrast))
        }else if(length(null)!=NROW(contrast)){
            stop("The length of argument \'null\' does not match the number of rows of argument \'contrast' \n")
        }
        if(is.null(rownames(contrast))){
            rownames(contrast) <- .contrast2name(contrast, null = null)
            null <- setNames(null, rownames(contrast))
        }
    }
    
    ### ** prepare export
    name.hypo <- rownames(contrast)
    n.hypo <- NROW(contrast)

    df.table <- as.data.frame(matrix(NA, nrow = n.hypo, ncol = 5,
                                     dimnames = list(name.hypo,
                                                     c("estimate","std","statistic","df","p-value"))
                                     ))

    ## ** Univariate Wald test
    C.p <- (contrast %*% param) - null
    if(robust){
        C.vcov.C <- contrast %*% rvcov.param %*% t(contrast)
    }else{
        C.vcov.C <- contrast %*% vcov.param %*% t(contrast)
    }
    sd.C.p <- sqrt(diag(C.vcov.C))
    stat.Wald <- C.p/sd.C.p

    ## store
    df.table$estimate <- as.numeric(C.p)
    df.table$std <- as.numeric(sd.C.p)
    df.table$statistic <- as.numeric(stat.Wald)

    ##  degrees of freedom
    if(df>0 && !is.null(dVcov.param)){

        ## univariate
        if(robust == FALSE){
            df.Wald  <- dfSigma(contrast = contrast,
                                vcov = vcov.param,
                                dVcov = dVcov.param,
                                keep.param = keep.param)
        }else if(robust == TRUE){

            if(df == TRUE){
                df.Wald <- dfSigma(contrast = contrast,
                                   vcov = vcov.param,
                                   dVcov = dVcov.param,
                                   keep.param = keep.param)
            }else if(df == 2){
                df.Wald  <- dfSigma(contrast = contrast,
                                    vcov = rvcov.param,
                                    dVcov = dRvcov.param * factor.dRvcov,
                                    keep.param = name.param)
            }else if(df == 3){
                df.Wald <- dfSigmaRobust(contrast = contrast,
                                         vcov = vcov.param,
                                         rvcov = rvcov.param,
                                         score = score)
            }
        }
    }else{
        df.Wald <- rep(Inf, n.hypo)
        df.F <- Inf
    }

    ## store
    df.table$df <- as.numeric(df.Wald)
    df.table$`p-value` <- as.numeric(2*(1-stats::pt(abs(df.table$statistic), df = df.table$df)))
    
    ## ** Multivariate Wald test
    df.table <- rbind(df.table, global = rep(NA,5))
    error <- NULL
    if(F.test){
        ## statistic
        iC.vcov.C <- try(solve(C.vcov.C), silent = TRUE)
        
        if(!inherits(iC.vcov.C,"try-error")){
            stat.F <- t(C.p) %*% iC.vcov.C %*% (C.p) / n.hypo

            ## df (independent t statistics)
            if(df>0){
                svd.tempo <- eigen(iC.vcov.C)
                D.svd <- diag(svd.tempo$values, nrow = n.hypo, ncol = n.hypo)
                P.svd <- svd.tempo$vectors
     
                C.anova <- sqrt(D.svd) %*% t(P.svd) %*% contrast

                if(df == TRUE){
                    nu_m <- dfSigma(contrast = C.anova,
                                    vcov = vcov.param,
                                    dVcov = dVcov.param,
                                    keep.param = keep.param)
                } else if(df == 2){
                    nu_m <- dfSigma(contrast = C.anova,
                                    vcov = rvcov.param,
                                    dVcov = dRvcov.param * factor.dRvcov,
                                    keep.param = keep.param)
                } else if(df == 3){
                    nu_m <- dfSigmaRobust(contrast = C.anova,
                                          vcov = vcov.param,
                                          rvcov = rvcov.param,
                                          score = score)
                }
        
                EQ <- sum(nu_m/(nu_m-2))
                df.F <- 2*EQ / (EQ - n.hypo)
            }else{
                df.F <- Inf
            }
            ## store
            df.table["global", "statistic"] <- as.numeric(stat.F)
            df.table["global", "df"] <- df.F
            df.table["global", "p-value"] <- 1 - stats::pf(df.table["global", "statistic"],
                                                           df1 = n.hypo,
                                                           df2 = df.table["global", "df"])
        }else{
            warning("Unable to compute the degrees of freedom for the F-test \n")
            error <- iC.vcov.C
        }
    }

    ## ** export
    if(as.lava == TRUE){
        level.inf <- (1-level)/2
        level.sup <- 1-level.inf

        level.inf.label <- paste0(100*level.inf,"%")
        level.sup.label <- paste0(100*level.sup,"%")

        df.estimate <- matrix(NA, nrow = n.hypo, ncol = 5,
                              dimnames = list(name.hypo,c("Estimate", "Std.Err", "df", level.inf.label, level.sup.label)))
        df.estimate[,"Estimate"] <- df.table[name.hypo,"estimate"]
        df.estimate[,"Std.Err"] <- df.table[name.hypo,"std"]
        df.estimate[,"df"] <- df.table[name.hypo,"df"]
        df.estimate[,level.inf.label] <- df.table[name.hypo,"estimate"] + stats::qt(level.inf, df = df.table[name.hypo,"df"]) * df.table[name.hypo,"std"]
        df.estimate[,level.sup.label] <- df.table[name.hypo,"estimate"] + stats::qt(level.sup, df = df.table[name.hypo,"df"]) * df.table[name.hypo,"std"]

        out <- list(statistic = setNames(df.table["global","statistic"],"F-statistic"),
                    parameter = setNames(round(df.table["global","df"],2), paste0("df1 = ",n.hypo,", df2")), ## NOTE: cannot not be change to coefficients because of lava
                    p.value = df.table["global","p-value"],
                    method = c("- Wald test -", "", "Null Hypothesis:", name.hypo),
                    estimate = df.estimate,
                    vcov = C.vcov.C,
                    coef = C.p[,1],
                    null = null,
                    cnames = name.hypo                    
                    )
        if(robust){
            colnames(out$estimate)[2] <- "robust SE"
        }        
        attr(out, "B") <- contrast
        class(out) <- "htest"
    }else{
        out <- df.table
        attr(out, "warning") <- warn
        attr(out, "contrast") <- contrast
    }
    attr(out,"error") <- error
    return(out)
}

## * dfSigma
##' @title Degree of Freedom for the Chi-Square Test
##' @description Computation of the degrees of freedom of the chi-squared distribution
##' relative to the model-based variance
##'
##' @param contrast [numeric vector] the linear combination of parameters to test
##' @param vcov [numeric matrix] the variance-covariance matrix of the parameters.
##' @param dVcov [numeric array] the first derivative of the variance-covariance matrix of the parameters.
##' @param keep.param [character vector] the name of the parameters with non-zero first derivative of their variance parameter.
##' 
dfSigma <- function(contrast, vcov, dVcov, keep.param){
    ## iLink <- "LogCau~eta"
    C.vcov.C <- rowSums(contrast %*% vcov * contrast) ## variance matrix of the linear combination
    ## C.vcov.C - vcov[iLink,iLink]

    C.dVcov.C <- sapply(keep.param, function(x){
        rowSums(contrast %*% dVcov[,,x] * contrast)
    })
    ## C.dVcov.C - dVcov[iLink,iLink,]
    numerator <- 2 *(C.vcov.C)^2
    ## numerator - 2*vcov[iLink,iLink]^2
    denom <- rowSums(C.dVcov.C %*% vcov[keep.param,keep.param,drop=FALSE] * C.dVcov.C)
    ## denom - t(dVcov[iLink,iLink,]) %*% vcov[keep.param,keep.param,drop=FALSE] %*% dVcov[iLink,iLink,]
    df <- numerator/denom
    return(df)
}

## * dfSigmaRobust
##' @title Degree of Freedom for the Robust Chi-Square Test
##' @description Computation of the degrees of freedom of the chi-squared distribution
##' relative to the robust-based variance
##'
##' @param contrast [numeric vector] the linear combination of parameters to test
##' @param vcov [numeric matrix] the variance-covariance matrix of the parameters.
##' @param rvcov [numeric matrix] the robust variance-covariance matrix of the parameters.
##' @param score [numeric matrix] the individual score for each parameter.
##'
##' @details When contrast is the identity matrix, this function compute the moments of the sandwich estimator
##' and the degrees of freedom of the approximate t-test as described in (Pan, 2002) section 2 and 3.1.
##'
##' @references
##' Wei Pan and Melanie M. Wall, Small-sample adjustments in using the sandwich variance estiamtor in generalized estimating equations. Statistics in medicine (2002) 21:1429-1441.
##' 
dfSigmaRobust <- function(contrast, vcov, rvcov, score){
    
    ## ** prepare
    n <- NROW(score)

    ## apply contrasts
    vcov.S <- vcov %*% t(contrast)

    ## ** compute moments of rvcov
    ## fast
    E.score2 <- crossprod(score)
    iid.score2 <- lapply(1:n, function(iRow){
        (tcrossprod(score[iRow,]) - E.score2/n)^2
    })

    var.rvcov <- t(vcov.S^2) %*% Reduce("+",iid.score2) %*% vcov.S^2

    ## slow    
    E.rvcov <- contrast %*% rvcov %*% t(contrast)
    ## iid.rvcov <- lapply(1:n, function(iRow){
    ##     (t(vcov.S) %*% tcrossprod(score[iRow,]) %*% vcov.S - E.rvcov/n)^2
    ## })
    ## var.rvcov <- Reduce("+",iid.rvcov)

    ## ** export
    df.rvcov <- (2*E.rvcov^2)/var.rvcov
    
    return(setNames(diag(df.rvcov), rownames(contrast)))
}


##----------------------------------------------------------------------
### compare2.R ends here

### transformSummaryTable.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb  3 2020 (18:29) 
## Version: 
## Last-Updated: jan 24 2022 (10:47) 
##           By: Brice Ozenne
##     Update #: 29
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

#' @title Apply Transformation to Summary Table
#' @description Update summary table according to a transformation, e.g. log-transformtion.
#' P-values are left unchanged but estimates, standard errors, and confidence intervals are updated.
#'
#' @param object A data.frame with columns estimate, se, lower, upper.
#' @param transform the name of a transformation or a function.
#'
#' @return a data.frame
#' @export
transformSummaryTable <- function(object, transform = NULL){
    if(is.null(transform)){
        return(object)
    }else if(identical(transform,"atanh")){
        transform <- atanh
        dtransform <- function(x){1/(1-x^2)}
    }else if(identical(transform,"exp")){
        transform <- exp
        dtransform <- function(x){exp(x)}
    }else if(identical(transform,"log")){
        transform <- log
        dtransform <- function(x){1/x}
    }else if(identical(transform,"loglog")){
        transform <- function(x){log(log(x))}
        dtransform <- function(x){1/(-x*log(x))}
    }else if(identical(transform,"cloglog")){
        transform <- function(x){log(log(1-x))}
        dtransform <- function(x){1/(-(1-x)*log(1-x))}
    }else if(!is.null(attr(transform,"derivative"))){
        dtransform <- attr(transform,"derivative")
    }else{
        dtransform <- function(x){diag(numDeriv::jacobian(transform, x))}
    }
    object[,"se"] <- object[,"se"]*dtransform(object[,"estimate"])
    object[,"estimate"] <- transform(object[,"estimate"])
    if("lower" %in% names(object)){
        object[,"lower"] <- transform(object[,"lower"])
    }
    if("upper" %in% names(object)){
        object[,"upper"] <- transform(object[,"upper"])
    }
    if("null" %in% names(object)){
        object[,"null"] <- transform(object[,"null"])
    }
    if("null" %in% names(object)){
        object[,"null"] <- transform(object[,"null"])
    }
    return(object)
}

######################################################################
### transformSummaryTable.R ends here

### sCorrect-dInformation2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: dec 11 2019 (14:09) 
## Version: 
## Last-Updated: Jan 17 2022 (18:44) 
##           By: Brice Ozenne
##     Update #: 349
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * .dInformation2
#' @title Compute the First Derivative of the Expected Information Matrix
#' @description Compute the first derivative of the expected information matrix.
#' @name .dinformation2-internal
#' 
#' @details \code{calc_dinformation} will perform the computation individually when the
#' argument \code{index.Omega} is not null.
#' 
#' @keywords internal
.dInformation2 <- function(dmu, dOmega, d2mu, d2Omega, OmegaM1,
                           missing.pattern, unique.pattern, name.pattern,
                           grid.3varD1, grid.2meanD1.1varD1, grid.2meanD2.1meanD1, grid.2varD2.1varD1,
                           name.param, leverage, n.cluster, weights){

    if(lava.options()$debug){cat(".dInformation2\n")}
    if(!is.null(weights)){
        stop(".dInformation2 does not support weights. \n")
    }
    symmetrize <- TRUE

    ## ** Prepare
    n.param <- length(name.param)
    n.pattern <- length(name.pattern)

    n.grid.3varD1 <- NROW(grid.3varD1)
    if(symmetrize && n.grid.3varD1>0){
        grid.3varD1 <- grid.3varD1[grid.3varD1$duplicatedXYZ==FALSE,,drop=FALSE]
        n.grid.3varD1 <- NROW(grid.3varD1)
    }

    n.grid.2meanD1.1varD1 <- NROW(grid.2meanD1.1varD1)
    if(symmetrize && n.grid.3varD1>0){
        grid.2meanD1.1varD1 <- grid.2meanD1.1varD1[grid.2meanD1.1varD1$duplicatedXY==FALSE,,drop=FALSE]
        n.grid.2meanD1.1varD1 <- NROW(grid.2meanD1.1varD1)
    }

    n.grid.2varD2.1varD1 <- NROW(grid.2varD2.1varD1)
    if(symmetrize && n.grid.2varD2.1varD1>0){
        grid.2varD2.1varD1 <- grid.2varD2.1varD1[grid.2varD2.1varD1$d2XZ+grid.2varD2.1varD1$d2YZ>0,,drop=FALSE]
        n.grid.2varD2.1varD1 <- NROW(grid.2varD2.1varD1)

        grid.2varD2.1varD1[,"d2XZ"] <- grid.2varD2.1varD1[,"d2XZ"]*(1-grid.2varD2.1varD1[,"duplicatedXZ"])
        grid.2varD2.1varD1[,"d2YZ"] <- grid.2varD2.1varD1[,"d2YZ"]*(1-grid.2varD2.1varD1[,"duplicatedYZ"])
    }
    
    n.grid.2meanD2.1meanD1 <- NROW(grid.2meanD2.1meanD1)
    if(symmetrize && n.grid.2meanD2.1meanD1>0){
        grid.2meanD2.1meanD1 <- grid.2meanD2.1meanD1[grid.2meanD2.1meanD1$d2XZ+grid.2meanD2.1meanD1$d2YZ>0,,drop=FALSE]
        n.grid.2meanD2.1meanD1 <- NROW(grid.2meanD2.1meanD1)

        grid.2meanD2.1meanD1[,"d2XZ"] <- grid.2meanD2.1meanD1[,"d2XZ"]*(1-grid.2meanD2.1meanD1[,"duplicatedXZ"])
        grid.2meanD2.1meanD1[,"d2YZ"] <- grid.2meanD2.1meanD1[,"d2YZ"]*(1-grid.2meanD2.1meanD1[,"duplicatedYZ"])
    }

    dInfo <- array(0,
                   dim = c(n.param, n.param, n.param),
                   dimnames = list(name.param, name.param, name.param))
    
    ## ** loop over missing data pattern
    for(iP in 1:n.pattern){ ## iP <- 1
        iPattern <- name.pattern[iP]
        iIndex <- missing.pattern[[iPattern]]
        iY <- which(unique.pattern[iP,]==1)

        iOmegaM1 <- OmegaM1[[iPattern]]
        if(!is.null(leverage)){
            iN.corrected <- length(iIndex) - colSums(leverage[iIndex,iY,drop=FALSE])
        }else{
            iN.corrected <- length(iIndex)
        }
        idmu <- .subsetList(dmu, indexRow = iIndex, indexCol = iY)
        idOmega <- .subsetList(dOmega, indexRow = iY, indexCol = iY)
        id2mu <- .subsetList2(d2mu, indexRow = iIndex, indexCol = iY)
        id2Omega <- .subsetList2(d2Omega, indexRow = iY, indexCol = iY)

        ## *** 3 first derivative regarding the variance
        if(n.grid.3varD1>0){
            for(iGrid in 1:n.grid.3varD1){ # iGrid <- 1
                iName1 <- grid.3varD1[iGrid,"X"]
                iName2 <- grid.3varD1[iGrid,"Y"]
                iName3 <- grid.3varD1[iGrid,"Z"]

                ## term 1
                iDiag1 <- diag(iOmegaM1 %*% idOmega[[iName3]] %*% iOmegaM1 %*% idOmega[[iName2]] %*% iOmegaM1 %*% idOmega[[iName1]])
                iDiag2 <- diag(iOmegaM1 %*% idOmega[[iName2]] %*% iOmegaM1 %*% idOmega[[iName3]] %*% iOmegaM1 %*% idOmega[[iName1]])
                dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] - 1/2 * sum(iDiag1 * iN.corrected + iDiag2 * iN.corrected)

                ## symmetrize (XYZ = XZY = YXZ = YZX = ZXY = ZYX)
                if(symmetrize){
                    dInfo[iName1,iName3,iName2] <- dInfo[iName1,iName2,iName3]
                    dInfo[iName2,iName1,iName3] <- dInfo[iName1,iName2,iName3]
                    dInfo[iName2,iName3,iName1] <- dInfo[iName1,iName2,iName3]
                    dInfo[iName3,iName1,iName2] <- dInfo[iName1,iName2,iName3]
                    dInfo[iName3,iName2,iName1] <- dInfo[iName1,iName2,iName3]
                }
            }
        }

        ## *** 2 first derivative regarding the mean and one regarding the variance
        if(n.grid.2meanD1.1varD1>0){
            for(iGrid in 1:n.grid.2meanD1.1varD1){ # iGrid <- 1
                iName1 <- grid.2meanD1.1varD1[iGrid,"X"]
                iName2 <- grid.2meanD1.1varD1[iGrid,"Y"]
                iName3 <- grid.2meanD1.1varD1[iGrid,"Z"]

                ## term 4
                dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] - sum(idmu[[iName1]] %*% iOmegaM1 %*% idOmega[[iName3]] %*% iOmegaM1 * idmu[[iName2]])
                
                ## symmetrize (XYZ = YXZ)
                if(symmetrize){
                    dInfo[iName2,iName1,iName3] <- dInfo[iName1,iName2,iName3]
                }

            }
        }
        
        ## *** 1 second derivative and 1 first derivative regarding the variance
        if(n.grid.2varD2.1varD1>0){
            for(iGrid in 1:n.grid.2varD2.1varD1){ # iGrid <- 1
                iName1 <- grid.2varD2.1varD1[iGrid,"X"]
                iName2 <- grid.2varD2.1varD1[iGrid,"Y"]
                iName3 <- grid.2varD2.1varD1[iGrid,"Z"]
               
                ## term 2
                if(grid.2varD2.1varD1[iGrid,"d2YZ"]){
                    d2.Var1 <- grid.2varD2.1varD1[iGrid,"d2YZ.Var1"]
                    d2.Var2 <- grid.2varD2.1varD1[iGrid,"d2YZ.Var2"]
                    iDiag <- 1/2 * sum(diag(iOmegaM1 %*% id2Omega[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 %*% idOmega[[iName1]]) * iN.corrected)
                    dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] + iDiag

                    ## symmetrize (XYZ = XZY)
                    if(symmetrize && (iName2 != iName3)){
                        dInfo[iName1,iName3,iName2] <- dInfo[iName1,iName3,iName2] + iDiag
                    }
                }

                ## term 3
                if(grid.2varD2.1varD1[iGrid,"d2XZ"]){
                    d2.Var1 <- grid.2varD2.1varD1[iGrid,"d2XZ.Var1"]
                    d2.Var2 <- grid.2varD2.1varD1[iGrid,"d2XZ.Var2"]
                    iDiag <- 1/2 * sum(diag(iOmegaM1 %*% idOmega[[iName2]] %*% iOmegaM1 %*% id2Omega[[d2.Var1]][[d2.Var2]]) * iN.corrected)
                    dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] + iDiag

                    ## symmetrize (XYZ = ZYX)
                    if(symmetrize && (iName1 != iName3)){
                        dInfo[iName3,iName2,iName1] <- dInfo[iName3,iName2,iName1] + iDiag
                    }
                }

            }
        }

        ## *** 1 second derivative and 1 first derivative regarding the mean
        if(n.grid.2meanD2.1meanD1>0){
            for(iGrid in 1:n.grid.2meanD2.1meanD1){ # iGrid <- 1
                iName1 <- grid.2meanD2.1meanD1[iGrid,"X"]
                iName2 <- grid.2meanD2.1meanD1[iGrid,"Y"]
                iName3 <- grid.2meanD2.1meanD1[iGrid,"Z"]

                ## term 5
                if(grid.2meanD2.1meanD1[iGrid,"d2XZ"]){
                    d2.Var1 <- grid.2meanD2.1meanD1[iGrid,"d2XZ.Var1"]
                    d2.Var2 <- grid.2meanD2.1meanD1[iGrid,"d2XZ.Var2"]
                    iDiag <- sum(id2mu[[d2.Var1]][[d2.Var2]] %*% iOmegaM1 * idmu[[iName2]])
                    dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] + iDiag

                    ## symmetrize (XYZ=ZYX)
                    if(symmetrize && (iName1 != iName3)){
                        dInfo[iName3,iName2,iName1] <- dInfo[iName3,iName2,iName1] + iDiag
                    }
                }

                ## term 6
                if(grid.2meanD2.1meanD1[iGrid,"d2YZ"]){
                    d2.Var1 <- grid.2meanD2.1meanD1[iGrid,"d2YZ.Var1"]
                    d2.Var2 <- grid.2meanD2.1meanD1[iGrid,"d2YZ.Var2"]
                    iDiag <- sum(idmu[[iName1]] %*% iOmegaM1 * id2mu[[d2.Var1]][[d2.Var2]])
                    dInfo[iName1,iName2,iName3] <- dInfo[iName1,iName2,iName3] + iDiag

                    ## symmetrize (XYZ = XZY)
                    if(symmetrize && (iName2 != iName3)){
                        dInfo[iName1,iName3,iName2] <- dInfo[iName1,iName3,iName2] + iDiag
                    }
                }
            }
        }
    }
    
    ## dInfo.bis <- .old_dInformation2(dmu = dmu, dOmega = dOmega, d2mu = d2mu, d2Omega = d2Omega, OmegaM1 = OmegaM1,
    ##                                 missing.pattern = missing.pattern, unique.pattern = unique.pattern, name.pattern = name.pattern, 
    ##                                 grid.dInformation = expand.grid(X = name.param, Y = name.param, Z = name.param, duplicated = FALSE, stringsAsFactors = FALSE),
    ##                                 name.param = name.param, name.param.dInformation = name.param,
    ##                                 leverage = leverage, n.cluster = n.cluster)
    ## dInfo.ter <- .old_dInformation2(dmu = dmu, dOmega = dOmega, d2mu = d2mu, d2Omega = d2Omega, OmegaM1 = OmegaM1,
    ##                                 missing.pattern = missing.pattern, unique.pattern = unique.pattern, name.pattern = name.pattern, 
    ##                                 grid.dInformation = expand.grid(X = name.param[2:3], Y = name.param[2:3], Z = name.param[2:3], duplicated = FALSE, stringsAsFactors = FALSE),
    ##                                 name.param = name.param[2:3], name.param.dInformation = name.param[2:3],
    ##                                 leverage = leverage, n.cluster = n.cluster)

    ## dInfo.bis[,,1]-t(dInfo.bis[,,1])
    ## dInfo.bis[,,2]-t(dInfo.bis[,,2])
    ## dInfo.bis[,,3]-t(dInfo.bis[,,3])
    ## print(range(dInfo.bis - dInfo))
    ## print(range(dInfo.ter  - dInfo))
    
    ## ** export
    return(dInfo)
}

## * .dVcov.param
.dVcov.param <- function(vcov.param, dInformation, n.param, name.param){
    dVcov.param <- array(0,
                         dim = c(n.param,n.param,n.param),
                         dimnames = list(name.param,name.param,name.param)
                         )
    
    for(iP in name.param){ ## iP <- "Y1"
        if(any(dInformation[,,iP]!=0)){
            dVcov.param[,,iP] <- - vcov.param %*% dInformation[,,iP] %*% vcov.param
        }
    }
    
    return(dVcov.param)
}

## * .dRvcov.param
.dRvcov.param <- function(score, hessian, vcov.param, dVcov.param, n.param, name.param){

    dRvcov.param <- array(0,
                          dim = c(n.param,n.param,n.param),
                          dimnames = list(name.param,name.param,name.param)
                          )
    
    score2_vcov.param <- crossprod(score) %*% vcov.param
    score_vcov.param <- score %*% vcov.param
        
    for(iP in name.param){ ## iP <- 1
        if(any(dVcov.param[,,iP]!=0)){
            term1 <- dVcov.param[,,iP] %*% score2_vcov.param
        }else{
            term1 <- matrix(0, nrow = n.param, ncol = n.param)
        }
        term2 <- vcov.param %*% hessian[iP,,] %*% score_vcov.param
        ## dRvcov.param[,,iP] <- term2 + t(term2) ## (what was lavaSearch2 doing before version 2.0.0)
        dRvcov.param[,,iP] <- term1 + t(term1) + term2 + t(term2)
    }

    return(dRvcov.param)
}



## * .old_dInformation2
## .old_dInformation2 <- function(dmu, dOmega, d2mu, d2Omega, OmegaM1,
##                                missing.pattern, unique.pattern, name.pattern, 
##                                grid.dInformation, name.param, name.param.dInformation,
##                                leverage, n.cluster){

##     if(lava.options()$debug){cat(".dInformation2\n")}

##     ## ** Prepare
##     n.grid <- NROW(grid.dInformation)
##     n.param <- length(name.param)
##     n.param.dInformation <- length(name.param.dInformation)
##     n.pattern <- length(name.pattern)

##     dInfo <-  array(0,
##                     dim = c(n.param, n.param, n.param.dInformation),
##                     dimnames = list(name.param, name.param, name.param.dInformation))

##     index.duplicated <- which(grid.dInformation$duplicated)
##     index.Nduplicated <- setdiff(1:n.grid, index.duplicated)
##     ## grid.dInformation[index.Nduplicated,,drop=FALSE]
    
##     ## ** loop over missing data pattern
##     for(iP in 1:n.pattern){ ## iP <- 1
##         iPattern <- name.pattern[iP]
##         iOmegaM1 <- OmegaM1[[iPattern]]
##         iIndex <- missing.pattern[[iPattern]]
##         iY <- which(unique.pattern[iP,]==1)

##         if(!is.null(leverage)){
##             iN.corrected <- length(iIndex) - colSums(leverage[iIndex,iY,drop=FALSE])
##         }else{
##             iN.corrected <- length(iIndex)
##         }
##         for(iGrid in index.Nduplicated){ # iGrid <- 1
##             iName1 <- grid.dInformation[iGrid,"X"]
##             iName2 <- grid.dInformation[iGrid,"Y"]
##             iNameD <- grid.dInformation[iGrid,"Z"]
##             ## cat("* ", iNameD," ",iName1,"",iName2,"\n")

##             ## *** identify relevant terms
##             test.Omega1 <- !is.null(dOmega[[iNameD]]) && !is.null(dOmega[[iName1]]) && !is.null(dOmega[[iName2]])
##             test.Omega2a <- !is.null(d2Omega[[iNameD]][[iName1]]) && !is.null(dOmega[[iName2]])
##             test.Omega2b <- !is.null(d2Omega[[iName1]][[iNameD]]) && !is.null(dOmega[[iName2]])
##             test.Omega3a <- !is.null(d2Omega[[iNameD]][[iName2]]) && !is.null(dOmega[[iName1]])
##             test.Omega3b <- !is.null(d2Omega[[iName2]][[iNameD]]) && !is.null(dOmega[[iName1]])

##             test.mu1a <- !is.null(d2mu[[iNameD]][[iName1]]) && !is.null(dmu[[iName2]])
##             test.mu1b <- !is.null(d2mu[[iName1]][[iNameD]]) && !is.null(dmu[[iName2]])
##             test.mu2a <- !is.null(d2mu[[iNameD]][[iName2nn]]) && !is.null(dmu[[iName1]])
##             test.mu2b <- !is.null(d2mu[[iName2]][[iNameD]]) && !is.null(dmu[[iName1]])
##             test.mu3 <- !is.null(dOmega[[iNameD]]) && !is.null(dmu[[iName1]]) && !is.null(dmu[[iName2]])

##             if((test.Omega1 + test.Omega2a + test.Omega2b + test.Omega3a + test.Omega3b + test.mu1a + test.mu1b + test.mu2a + test.mu2b + test.mu3) == 0){
##                 next
##             }
##             ## *** extract quantities for computations 
##             if(test.mu1a){
##                 d2mu.D1 <- d2mu[[iNameD]][[iName1]][iIndex,iY,drop=FALSE]
##             }else if(test.mu1b){
##                 d2mu.D1 <- d2mu[[iName1]][[iNameD]][iIndex,iY,drop=FALSE]
##             }
##             if(test.mu2a){
##                 d2mu.D2 <- d2mu[[iNameD]][[iName2]][iIndex,iY,drop=FALSE]
##             }else if(test.mu2b){
##                 d2mu.D2 <- d2mu[[iName2]][[iNameD]][iIndex,iY,drop=FALSE]
##             }
##             if(test.Omega2a){
##                 d2Omega.D1 <- d2Omega[[iNameD]][[iName1]][iY,iY,drop=FALSE]
##             }else if(test.Omega2b){
##                 d2Omega.D1 <- d2Omega[[iName1]][[iNameD]][iY,iY,drop=FALSE]
##             }
##             if(test.Omega3a){
##                 d2Omega.D2 <- d2Omega[[iNameD]][[iName2]][iY,iY,drop=FALSE]
##             }else{
##                 d2Omega.D2 <- d2Omega[[iName2]][[iNameD]][iY,iY,drop=FALSE]
##             }
                
##             ## *** pre-compute 
##             if(!is.null(dOmega[[iName1]])){
##                 iOmegaM1.dOmega.1 <- iOmegaM1 %*% dOmega[[iName1]][iY,iY,drop=FALSE]
##             }
##             if(!is.null(dOmega[[iName2]])){
##                 iOmegaM1.dOmega.2 <- iOmegaM1 %*% dOmega[[iName2]][iY,iY,drop=FALSE]
##             }                    
##             if(!is.null(dOmega[[iNameD]])){
##                 iOmegaM1.dOmega.D <- iOmegaM1 %*% dOmega[[iNameD]][iY,iY,drop=FALSE]
##             }
            
##             ## *** evaluate contributions to dInformation
##             if(test.Omega1){

##                 iOmegaM1.dOmega.1 %*% iOmegaM1.dOmega.D %*% iOmegaM1.dOmega.2
                
##                 iDiag1 <- diag(iOmegaM1.dOmega.D %*% iOmegaM1.dOmega.2 %*% iOmegaM1.dOmega.1)
##                 iDiag2 <- diag(iOmegaM1.dOmega.2 %*% iOmegaM1.dOmega.D %*% iOmegaM1.dOmega.1)
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - 1/2 * sum((iDiag1 + iDiag2) * iN.corrected)
##                 ## if(iName1!=iName2){
##                 ##     cat(iName1,"/",iName2,"/",iNameD,": ",iDiag1," ",iDiag2,"\n")
##                 ## }
##             }

##             if(test.Omega3a || test.Omega3b){
##                 iDiag <- diag(iOmegaM1 %*% d2Omega.D2 %*% iOmegaM1.dOmega.1)
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * iN.corrected)                
##             }
            
##             if(test.Omega2a || test.Omega2b){
##                 iDiag <- diag(iOmegaM1.dOmega.2 %*% iOmegaM1 %*% d2Omega.D1)
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + 1/2 * sum(iDiag * iN.corrected)
##             }

##             if(test.mu1a || test.mu1b){
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(d2mu.D1 %*% iOmegaM1 * dmu[[iName2]][iIndex,iY,drop=FALSE])
##             }

##             if(test.mu2a || test.mu2b){
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] + sum(dmu[[iName1]][iIndex,iY,drop=FALSE] %*% iOmegaM1 * d2mu.D2)
##             }
                    
##             if(test.mu3){
##                 dInfo[iName1,iName2,iNameD] <- dInfo[iName1,iName2,iNameD] - sum(dmu[[iName1]][iIndex,iY,drop=FALSE] %*% iOmegaM1.dOmega.D %*% iOmegaM1 * dmu[[iName2]][iIndex,iY,drop=FALSE])
##             }

##         }
##     }

##     ### ** export
##     return(dInfo)
## }

######################################################################
### sCorrect-dInformation2.R ends here

### evalInParentEnv.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt  5 2017 (10:48) 
## Version: 
## last-updated: sep 28 2018 (12:05) 
##           By: Brice Ozenne
##     Update #: 12
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

#' @title Find Object in the Parent Environments
#' 
#' @description Search an object in the parent environments. For internal use.
#' 
#' @param name [character] the name of the object to get.
#'
#' @concept extractor
#' @keywords internal
evalInParentEnv <- function(name){
  
  ### ** find parent environments
  frames <- sys.status()
  all.frames <- sapply(1:length(frames$sys.frames), function(x){identical(parent.frame(x),globalenv())})
  index.parents <- which(all.frames==FALSE)
  n.parents <- length(index.parents)
  
  ### ** look in parent environments
  iParent <- 1
  res <- NULL
  cv <- FALSE
  while((iParent <= n.parents) && (cv == FALSE)){ # iParent <- 1
    
    res <- try(eval(name, envir = parent.frame(iParent)), silent = TRUE)
      
      if("try-error" %in% class(res)){
        iParent <- iParent + 1     
      }else{
        cv <- TRUE
      }
    
  }
  
  return(res)
}


#----------------------------------------------------------------------
### evalInParentEnv.R ends here

### effects2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar  4 2019 (10:28) 
## Version: 
## Last-Updated: jan 23 2024 (10:24) 
##           By: Brice Ozenne
##     Update #: 386
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * effects2 (documentation)
#' @title Effects Through Pathways With Small Sample Correction 
#' @description Test whether a path in the latent variable model correspond to a null effect.
#' Similar to \code{lava::effects} but with small sample correction (if any).
#' So far it only work for a single path related two variable composed of one or two edges.
#' @name effects2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param linfct [character vector] The path for which the effect should be assessed (e.g. \code{"A~B"}),
#' i.e. the effect of the right variable (B) on the left variable (A). 
#' @param robust [logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param conf.level [numeric, 0-1] level of the confidence intervals.
#' @param from,to alternative to argument \code{linfct}. See \code{lava::effects}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object.
#' 
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
#' 
#' @return A data.frame with a row per path.
#' 
#' @concept inference
#' @keywords smallSampleCorrection
#' @export
`effects2` <-
  function(object, linfct, robust, cluster, conf.level, ...) UseMethod("effects2")

## * effects2 (examples)
## TODO

## * effects2.lvmfit
#' @rdname effects2
#' @export
effects2.lvmfit <- function(object, linfct, robust = FALSE, cluster = NULL, conf.level = 0.95, to = NULL, from = NULL, df = lava.options()$df, ssc = lava.options()$ssc, ...){

    return(effects2(estimate2(object, ssc = ssc, df = df, dVcov.robust = robust, ...), linfct = linfct, to = to, from = from, robust = robust, cluster = cluster, conf.level = conf.level))

}

## * effects2.lvmfit2
#' @rdname effects2
#' @export
effects2.lvmfit2 <- function(object, linfct, robust = FALSE, cluster = NULL, conf.level = 0.95, to = NULL, from = NULL, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }
    object0 <- object
    class(object0) <- setdiff(class(object0),"lvmfit2")
        
    ## ** identify path
    if(!is.null(to) || !is.null(from)){
        n.hypo <- 1

        if(!missing(linfct)){
            stop("Cannot specify argument \'linfct\' at the same time as argument \'from\' or \'to\'. \n")
        }
        e.effects <- effects(object0, from = from, to = to)
        pathEffect <- stats::setNames(list(stats::setNames(list(e.effects$path), paste0(e.effects["to"],"~",e.effects["from"]))),paste0(e.effects["to"],"~",e.effects["from"]))
        type <- "total"
        null <- 0

    }else{
        n.hypo <- length(linfct)
        pathEffect <- vector(mode = "list", length = n.hypo)
        if(is.null(names(linfct))){
            names(pathEffect) <- linfct
        }else{
            if(any(duplicated(names(linfct)))){
                stop("Duplicated names for argument \'linfct\'. \n")
            }
            names(pathEffect) <- names(linfct)
        }
        type <- rep(as.character(NA), n.hypo)
        null <- rep(as.numeric(NA), n.hypo)

        for(iH in 1:n.hypo){

            if(grepl("|",linfct[iH], fixed=TRUE)){
                type[iH] <- base::trimws(strsplit(linfct[iH],split="|",fixed=TRUE)[[1]][2], which = "both")
                type[iH] <- match.arg(type[iH], c("indirect","direct","total"))
                linfct[iH] <- strsplit(linfct[iH],split="|",fixed=TRUE)[[1]][1]
            }else{
                type[iH] <- "total"
            }
            
            ## extract left and right side of the equation
            if(length(grep("=",linfct[iH]))>1){
                stop("Each element of argument \'linfct\' should contain at most one \'=\' sign.\n",
                     "Something like: coef1-2*coef2=0. \n")
            }
            iContrast <- createContrast(linfct[iH])
            if(iContrast$null==0){
                iContrast <- createContrast(linfct[iH], rowname.rhs = FALSE)
            }

            null[iH] <- unname(iContrast$null)
            iLHS.hypo_factor <- as.double(iContrast$contrast)
            iLHS.hypo_coef <- unname(colnames(iContrast$contrast))
            iN.param <- length(iLHS.hypo_coef)

            pathEffect[[iH]] <- stats::setNames(vector(mode = "list", length = iN.param), iLHS.hypo_coef)
            attr(pathEffect[[iH]], "factor") <- iLHS.hypo_factor
            
            for(iCoef in 1:iN.param){ ## iCoef <- 1
                ## extract all paths for each coefficient
                pathEffect[[iH]][[iCoef]] <- effects(object0, stats::as.formula(iLHS.hypo_coef[[iCoef]]))$path
                if(length(pathEffect[[iH]][[iCoef]])==0){
                    stop("Could not find path relative to coefficient ",iLHS.hypo_coef[[iCoef]]," (linfct=",linfct[iH],"). \n")
                }else if(type[iH]=="direct" && any(sapply(pathEffect[[iH]][[iCoef]],length)>2)){
                    pathEffect[[iH]][[iCoef]] <- pathEffect[[iH]][[iCoef]][sapply(pathEffect[[iH]][[iCoef]],length)==2]
                    if(length(pathEffect[[iH]][[iCoef]])==0){
                        stop("Could not find direct path relative to coefficient ",iLHS.hypo_coef[[iCoef]]," (linfct=",linfct[iH],"). \n")
                    }
                }else if(type[iH]=="indirect" && any(sapply(pathEffect[[iH]][[iCoef]],length)>2)){
                    pathEffect[[iH]][[iCoef]] <- pathEffect[[iH]][[iCoef]][sapply(pathEffect[[iH]][[iCoef]],length)>2]
                    if(length(pathEffect[[iH]][[iCoef]])==0){
                        stop("Could not find indirect path relative to coefficient ",iLHS.hypo_coef[[iCoef]]," (linfct=",linfct[iH],"). \n")
                    }
                }
            }
            
        }
    }

    ## ** extract information
    ## 0-order: param
    object.param <- coef(object, as.lava = FALSE)
    object.paramAll <- coef2(object, type = 9, labels = 1)[,"Estimate"]
    name.param <- names(object.param)
    n.param <- length(name.param)

    ## 1-order: score
    if(robust){
        object.score <- score(object, cluster = cluster, as.lava = FALSE)
    }

    ## 2-order: variance covariance
    object.vcov.param <- vcov(object, as.lava = FALSE)
    if(robust){
        object.rvcov.param <- vcov(object, robust = TRUE, cluster = cluster, as.lava = FALSE)
    }
    
    test.df <- (object$sCorrect$df == "satterthwaite")
    if(test.df){
        object.dVcov.param <- object$sCorrect$dVcov.param

        if(robust && (lava.options()$df.robust != 1)){

            if(!is.null(cluster)){ ## update derivative according to cluster
                object.dRvcov.param <- .dRvcov.param(score = object.score,
                                                     hessian = hessian2(object, cluster = cluster),
                                                     vcov.param = object.vcov.param,
                                                     dVcov.param = object.dVcov.param,
                                                     n.param = n.param,
                                                     name.param = name.param)
                                              
            }else{
                dRvcov.param <- object$sCorrect$dRvcov.param
            }
        }
    }

    coefEffect <- pathEffect

    ## ** identify coefficients corresponding to path
    for(iH in 1:n.hypo){ ## iH <- 1
        for(iCoef in 1:iN.param){ ## iCoef <- 1
            iN.path <- length(pathEffect[[iH]][[iCoef]])
            for(iPath in 1:iN.path){ ## iPath <- 1
                coefEffect[[iH]][[iCoef]][[iPath]] <- paste(pathEffect[[iH]][[iCoef]][[iPath]][-1], pathEffect[[iH]][[iCoef]][[iPath]][-length(pathEffect[[iH]][[iCoef]][[iPath]])], sep = lava.options()$symbols[1])
                if(any(coefEffect[[iH]][[iCoef]][[iPath]] %in% name.param) == FALSE){
                    stop("Incorrect path: ",paste(pathEffect[[iH]][[iCoef]][[iPath]], collapse="->"),"\n",
                         "Could not find coefficient: \"",paste(coefEffect[[iH]][[iCoef]][[iPath]][coefEffect[[iH]][[iCoef]][[iPath]] %in% name.param == FALSE], collapse = "\" \""),"\".\n")
                }
            }
        }
    }

    ## ** point estimate
    vec.beta <- stats::setNames(rep(NA, length = n.hypo), names(pathEffect))
    for(iH in 1:n.hypo){ ## iH <- 1
        iValue.param <- lapply(coefEffect[[iH]], function(iCoef){ ## for each coefficient (e.g. Y~E1 - Y~E2 = 0)
            iValue.path <- lapply(iCoef, function(iName){prod(object.paramAll[iName])}) ## get effect through each path corresponding to a coefficient (e.g. Y~E: Y~E and Y~X and X~E, i.e. \beta1 and \beta2*\beta3)
            return(do.call("sum", iValue.path)) ## return total effect (e.g. \beta1 + \beta2*\beta3)
        })
        if(is.null(attr(coefEffect[[iH]],"factor"))){
            vec.beta[iH] <- unlist(iValue.param)
        }else{
            vec.beta[iH] <- sum(attr(coefEffect[[iH]],"factor") * unlist(iValue.param))
        }
    }

    ## ** variance
    ## *** partial derivative
    dbeta.dtheta <- matrix(NA, nrow = n.hypo, ncol = n.param, dimnames = list(names(pathEffect), name.param))
    for(iH in 1:n.hypo){ ## iH <- 1
        iValue.param <- lapply(coefEffect[[iH]], function(iCoef){  ## iCoef <- coefEffect[[iH]][[1]] ## for each coefficient (e.g. Y~E1 - Y~E2 = 0)
            iDValue.path <- colSums(do.call(rbind,lapply(iCoef, function(iName){ ## iName <- iCoef[[1]] ## get derivative through each path corresponding to a coefficient (e.g. Y~E: Y~E and Y~X and X~E, i.e. \beta1 and \beta2*\beta3)
                iDeriv <- stats::setNames(rep(0, n.param), name.param)
                iDeriv[intersect(iName,name.param)] <- prod(object.paramAll[iName])/object.paramAll[intersect(iName,name.param)]
                return(iDeriv)
            })))
        })
        if(is.null(attr(coefEffect[[iH]],"factor"))){
            dbeta.dtheta[iH,] <- iValue.param[[1]]
        }else{
            dbeta.dtheta[iH,] <- attr(coefEffect[[iH]],"factor") %*% do.call(rbind,iValue.param)
        }   
    }

    if(robust){
        Mvcov.beta <- dbeta.dtheta %*% object.rvcov.param %*% t(dbeta.dtheta)
    }else{
        Mvcov.beta <- dbeta.dtheta %*% object.vcov.param %*% t(dbeta.dtheta)
    }

    ## ** compute df
    if(test.df){
        vec.df <- dfSigma(contrast = dbeta.dtheta,
                          score = object.score,
                          vcov = object.vcov.param,
                          rvcov = object.rvcov.param,
                          dVcov = object.dVcov.param,
                          dRvcov = object.dRvcov.param,
                          keep.param = dimnames(object.dVcov.param)[[3]],                            
                          type = if(robust){lava.options()$df.robust}else{1})
    }else{
        vec.df <- rep(0, n.hypo)
    }

    ## ** gather everything in glht object
    linfct2 <- diag(1, ncol = n.hypo, nrow = n.hypo)
    dimnames(linfct2) <- list(names(pathEffect),names(pathEffect))

    out <- list(model = object,
                linfct = linfct2,
                rhs = null,
                coef = vec.beta,
                vcov = Mvcov.beta,
                df = vec.df,
                alternative = "two.sided",
                type = NULL,
                robust = robust,
                ssc = object$sCorrect$ssc$type,
                grad = dbeta.dtheta,
                path = pathEffect,
                global = NULL)
    class(out) <- c("glht2","glht")

    ## ** export
    return(out)

}

## * effects.lvmfit2
#' @rdname effects2
#' @export
effects.lvmfit2 <- effects2.lvmfit2

######################################################################
### effects2.R ends here

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

OLS_cpp <- function(X, y) {
    .Call(`_lavaSearch2_OLS_cpp`, X, y)
}

OLS2_cpp <- function(X, y) {
    .Call(`_lavaSearch2_OLS2_cpp`, X, y)
}

wildBoot_cpp <- function(iid, lsIndexModel, nSample, nObs, nModel, p) {
    .Call(`_lavaSearch2_wildBoot_cpp`, iid, lsIndexModel, nSample, nObs, nModel, p)
}

### getVarCov2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: mar 27 2018 (09:55) 
## Version: 
## Last-Updated: jul 25 2019 (10:13) 
##           By: Brice Ozenne
##     Update #: 33
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * getVarCov2

#' @title Reconstruct the Conditional Variance Covariance Matrix
#' @description Reconstruct the conditional variance covariance matrix from a nlme or lvm model.
#' Only compatible with specific correlation and variance structure.
#' @name getVarCov2
#'
#' @param object a \code{gls} or \code{lme} object
#' @param param [numeric vector] values for the model parameters.
#' @param data [data.frame] the data set.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param ... [internal] only used by the generic method.
#' 
#' @details The compound symmetry variance-covariance matrix in a gls model is of the form:
#' \tabular{cccc}{
#' \eqn{\Sigma =} \tab \eqn{\sigma^2} \tab \eqn{\sigma^2 \rho} \tab \eqn{\sigma^2 \rho} \cr
#' \tab . \tab \eqn{\sigma^2} \tab \eqn{\sigma^2 \rho} \cr
#' \tab . \tab . \tab \eqn{\sigma^2}
#' }
#'
#' The unstructured variance-covariance matrix in a gls model is of the form:
#'  \tabular{cccc}{
#' \eqn{\Sigma =} \tab \eqn{\sigma^2} \tab \eqn{\sigma^2 \sigma_2 \rho_{1,2}} \tab \eqn{\sigma^2 \sigma_3 \rho_{1,3}} \cr
#' \tab . \tab \eqn{\sigma^2 \sigma_2^2} \tab \eqn{\sigma^2 \sigma_2 \sigma_3 \rho_{2,3}} \cr
#' \tab . \tab . \tab \eqn{\sigma^2 \sigma_3^2}
#' }
#' @return A list containing the residual variance-covariance matrix in the element Omega.
#' 
#' @examples
#' 
#' ## simulate data 
#' library(nlme)
#' n <- 5e1
#' mSim <- lvm(c(Y1~1*eta,Y2~1*eta,Y3~1*eta,eta~G))
#' latent(mSim) <- ~eta
#' transform(mSim,Id~Y1) <- function(x){1:NROW(x)}
#' set.seed(10)
#' dW <- lava::sim(mSim,n,latent = FALSE)
#' dW <- dW[order(dW$Id),,drop=FALSE]
#' dL <- reshape2::melt(dW,id.vars = c("G","Id"), variable.name = "time")
#' dL <- dL[order(dL$Id),,drop=FALSE]
#' dL$Z1 <- rnorm(NROW(dL))
#' dL$time.num <- as.numeric(as.factor(dL$time))
#' 
#' #### iid model #### 
#' e1.gls <- nlme::gls(Y1 ~ G, data = dW, method = "ML")
#' getVarCov2(e1.gls, cluster = 1:n)$Omega
#' 
#' #### heteroschedasticity ####
#' dW$group <- rbinom(n, size = 1, prob = 1/2)
#' dW$repetition <- as.numeric(as.factor(dW$group))
#' e2a.gls <- nlme::gls(Y1 ~ G, data = dW, method = "ML",
#'                     weights = varIdent(form =~ repetition|group))
#' getVarCov2(e2a.gls, cluster = 1:n)$Omega
#'
#' 
#' e2b.gls <- nlme::gls(value ~ 0+time + time:G,
#'                    weight = varIdent(form = ~ time.num|time),
#'                    data = dL, method = "ML")
#' getVarCov2(e2b.gls, cluster = "Id")$Omega
#'
#' #### compound symmetry ####
#' e3.gls <- nlme::gls(value ~ time + G,
#'                    correlation = corCompSymm(form = ~1| Id),
#'                    data = dL, method = "ML")
#' getVarCov2(e3.gls)$Omega
#' 
#' #### unstructured ####
#' e4.gls <- nlme::gls(value ~ time,
#'                     correlation = corSymm(form = ~time.num| Id),
#'                     weight = varIdent(form = ~ 1|time),
#'                     data = dL, method = "ML")
#' getVarCov2(e4.gls)$Omega
#'
#' #### lvm model ####
#' m <- lvm(c(Y1~1*eta,Y2~1*eta,Y3~1*eta,eta~G))
#' latent(m) <- ~eta
#' e <- estimate(m, dW)
#' getVarCov2(e)
#' 
#' @concept extractor
#' 
#' @export
`getVarCov2` <-
    function(object, ...) UseMethod("getVarCov2")

## * getVarCov2.gls
#' @rdname getVarCov2
#' @export
getVarCov2.gls <- function(object, data = NULL, cluster, ...){

    ## ** data
    if(is.null(data)){
        data <- extractData(object, design.matrix = FALSE, as.data.frame = TRUE,
                            envir = parent.env(environment()))
    }

    ## ** endogenous variable
    formula.object <- .getFormula2(object)
    name.Y <- all.vars(stats::update(formula.object, ".~1"))

    ## ** extractors   
    res.cluster <- .getCluster2(object, data = data, cluster = cluster)
    res.param <- .coef2(object)
    res.index <- .getIndexOmega2(object,
                                 param = res.param,
                                 attr.param = attributes(res.param),
                                 name.Y = name.Y,
                                 cluster = res.cluster$cluster,
                                 levels.cluster = res.cluster$levels.cluster,
                                 data = data)
    res.Omega <- .getVarCov2(object,
                             param = res.param,
                             attr.param = attributes(res.param),
                             name.endogenous = res.index$name.endogenous,
                             n.endogenous = res.index$n.endogenous,
                             ref.group = res.index$ref.group)

    ## ** export
    return(c(res.cluster,
             list(param = res.param),
             res.index,
             list(Omega = res.Omega))
           )
}

## * getVarCov2.lme
#' @rdname getVarCov2
#' @export
getVarCov2.lme <- getVarCov2.gls

## * getVarCov2.lvmfit
#' @rdname getVarCov2
#' @export
getVarCov2.lvmfit <- function(object, data = NULL, param = NULL, ...){

    if(inherits(object, "lvmfit2")){
        return(object$sCorrect$Omega)
    }else{
        name.latent <- latent(object)
        n.latent <- length(name.latent)

        ## ** Prepare
        if(is.null(object$conditionalMoment)){
            name.endogenous <- endogenous(object)
            if(is.null(param)){
                param <- coef(object)
            }
            if(is.null(data)){
                data <- as.data.frame(object$data$model.frame)
            }

            object$conditionalMoment <- conditionalMoment(object,
                                                          data = data,
                                                          param = param,
                                                          name.endogenous = name.endogenous,
                                                          name.latent = name.latent,
                                                          first.order = FALSE,
                                                          second.order = FALSE,
                                                          usefit = TRUE)
        }

        ## ** Compute Omega
        if(n.latent>0){
            Omega <- object$conditionalMoment$value$tLambda.tiIB.Psi.iIB %*% object$conditionalMoment$value$Lambda + object$conditionalMoment$value$Sigma
        }else{
            Omega <- object$conditionalMoment$value$Sigma
        }

        return(Omega)
    }
}



##----------------------------------------------------------------------
### getVarCov2.R ends here

#library(lavaSearch2)
suppressPackageStartupMessages(library("testthat"))
suppressPackageStartupMessages(library("nlme"))
suppressPackageStartupMessages(library("multcomp"))
suppressPackageStartupMessages(library("Matrix"))
test_check("lavaSearch2")

### iid2.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (13:16) 
## Version: 
## last-updated: feb 11 2019 (13:26) 
##           By: Brice Ozenne
##     Update #: 482
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * Documentation - iid2
#' @title  Extract corrected i.i.d. decomposition
#' @description  Extract corrected i.i.d. decomposition from a gaussian linear model.
#' @name iid2
#'
#' @param object a linear model or a latent variable model
#' @param param [named numeric vector] the fitted parameters.
#' @param data [data.frame] the data set.
#' @param cluster [integer vector] the grouping variable relative to which the observations are iid.
#' @param bias.correct [logical] should the standard errors of the coefficients be corrected for small sample bias? Only relevant if the \code{sCorrect} function has not yet be applied to the object.
#' @param robust [logical] if \code{FALSE}, the i.i.d. decomposition is rescaled such its the squared sum equals the model-based standard error (instead of the robust standard error).
#' @param ... arguments to be passed to \code{sCorrect}.
#'
#' @details If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the influence function.
#'
#' @seealso \code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
#'
#' @return A matrix containing the 1st order influence function relative to each sample (in rows)
#' and each model coefficient (in columns).
#' 
#' @examples
#' n <- 5e1
#' p <- 3
#' X.name <- paste0("X",1:p)
#' link.lvm <- paste0("Y~",X.name)
#' formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))
#'
#' m <- lvm(formula.lvm)
#' distribution(m,~Id) <- sequence.lvm(0)
#' set.seed(10)
#' d <- sim(m,n)
#'
#' ## linear model
#' e.lm <- lm(formula.lvm,data=d)
#' iid.tempo <- iid2(e.lm, bias.correct = FALSE)
#' range(iid.tempo[,1:4]-iid(e.lm))
#' 
#' ## latent variable model
#' e.lvm <- estimate(lvm(formula.lvm),data=d)
#' iid.tempo <- iid2(e.lvm, bias.correct = FALSE)
#' range(iid.tempo-iid(e.lvm))
#' ## difference due to the use of the observed info matrix vs. the expected one.
#'
#' ## rescale i.i.d using model-based standard error
#' iid.tempo <- iid2(e.lvm, robust = FALSE, bias.correct = FALSE)
#' diag(crossprod(iid.tempo))-diag(vcov(e.lvm))
#'
#' @concept small sample inference
#' @concept iid decomposition
#' @export
`iid2` <-
  function(object, ...) UseMethod("iid2")

## * iid2.lm
#' @rdname iid2
#' @export
iid2.lm <- function(object,
                    param = NULL, data = NULL, bias.correct = TRUE, ...){

    sCorrect(object, param = param, data = data, df = FALSE) <- bias.correct

    iid.tempo <- iid2(object, ...)

    ## ** keep track of the NA
    if(!is.null(object$na.action)){
        n.NA <- length(object$na.action)
        iid.withNA <- matrix(NA, nrow = NROW(iid.tempo) + n.NA, ncol = NCOL(iid.tempo),
                             dimnames = list(NULL, colnames(iid.tempo)))
        iid.withNA[-object$na.action,] <- iid.tempo
    }else{
        iid.withNA <- iid.tempo
    }
    
## ** export
    return(iid.withNA)
}

## * iid2.gls
#' @rdname iid2
#' @export
iid2.gls <- function(object, cluster = NULL,
                     param = NULL, data = NULL, bias.correct = TRUE, ...){

    sCorrect(object, cluster = cluster, param = param, data = data, df = FALSE) <- bias.correct

    if(!is.null(cluster)){
        cluster <- object$sCorrect$args$cluster
    }

### ** export
    return(iid2(object, ...))
}

## * iid2.lme
#' @rdname iid2
#' @export
iid2.lme <- iid2.lm

## * iid2.lvmfit
#' @rdname iid2
#' @export
iid2.lvmfit <- iid2.lm


## * iid2.lm2
#' @rdname iid2
#' @export
iid2.lm2 <- function(object, cluster = NULL,
                     param = NULL, data = NULL, robust = TRUE, ...){

    ## ** compute the score
    if(!is.null(param) || !is.null(data)){
        args <- object$sCorrect$args
        args$df <- FALSE
        object$sCorrect <- do.call(sCorrect,
                                   args = c(list(object, param = param, data = data),
                                            args))
    }else if(is.null(object$sCorrect$score)){
        stop("set argument \'score\' to TRUE when calling sCorrect \n")
    }

    ## ** compute iid
    if(is.null(cluster)){
        scoreCluster <- score2(object)
    }else{
        n.obs <- stats::nobs(object)

        if(length(cluster) != n.obs){
            stop("Argument \'cluster\' does not have the correct length (n=",n.obs,") \n")
        }
        scoreIndiv <- score2(object)
        scoreCluster <- do.call("rbind",tapply(1:n.obs,cluster, function(iIndex){
            colSums(scoreIndiv[iIndex,,drop=FALSE])
        }))
    }
    res <- scoreCluster %*% object$sCorrect$vcov.param
    if(robust == FALSE){
        vec.sigma <- sqrt(diag(object$sCorrect$vcov.param))
        vec.sigma.robust <- sqrt(apply(res^2,2,sum))
        res <- sweep(res, MARGIN = 2, FUN = "*", STATS = vec.sigma/vec.sigma.robust)
    }
    
    ## ** export
    return(res)

}

## * iid2.gls2
#' @rdname iid2
#' @export
iid2.gls2 <- iid2.lm2

## * iid2.lme2
#' @rdname iid2
#' @export
iid2.lme2 <- iid2.lm2

## * iid2.lvmfit2
#' @rdname iid2
#' @export
iid2.lvmfit2 <- function(object, cluster = NULL, data = NULL, ...){

    ## try to find cluster in data
    if(!is.null(object$cluster)){
        if(!is.null(cluster)){
            stop("Argument \'cluster\' must be NULL when the object already contain a cluster \n")
        }
        cluster <- object$cluster
    }else if(!is.null(cluster) && length(cluster) == 1){
        if(!is.null(data)){
            if(cluster %in%  names(data) == FALSE){
                stop("Could not find variable ",cluster," (argument \'cluster\') in argument \'data\' \n")
            }else{
                cluster <- data[[cluster]]
            }            
        }else{
            if(cluster %in% names(object$data$model.frame) == FALSE){
                stop("Could not find variable ",cluster," (argument \'cluster\') in argument object$data \n")
            }else{
                cluster <- object$data$model.frame[[cluster]]
            }            
        }
    }

    return(iid2.lm2(object = object, data = data, cluster = cluster, ...))
}



##----------------------------------------------------------------------
### iid2.R ends here

## * initVarLink (Documentation)
#' @title Normalize var1 and var2
#' @name initVarLink
#' @description Convert var1 and var2 from formula or covariance to character.
#' 
#' @param var1 [character or formula] the exogenous variable of the new link or a formula describing the link.
#' @param var2 [character] the endogenous variable of the new link.
#' Disregarded if the argument \code{var1} is a formula.
#' @param rep.var1 [logical] should var1 be duplicated to match var2 length.
#' Only active if \code{format = "list"}.
#' @param format [character] should the name of the variable be returned (\code{format = "list"}),
#' a vector of character formula (\code{format = "txt.formula"}),
#' or a list of formula (\code{format = "formula"}).
#' @param Slink [character] the symbol for regression link.
#' @param Scov [character] the symbol for covariance link.
#' @param ... argument to be passed to \code{initVarLink}.
#'
#' @return See argument \code{format}.
#' 
#' @examples
#' initVarLink(y ~ x1)
#' initVarLink("y ~ x1")
#' initVarLink(y ~ x1 + x2)
#' initVarLink("y ~ x1 + x2")
#' initVarLink(y ~ x1 + x2, rep.var1 = TRUE)
#' initVarLink(y ~ x1 + x2, rep.var1 = TRUE, format = "formula")
#' initVarLink(y ~ x1 + x2, rep.var1 = TRUE, format = "txt.formula")
#' initVarLink("y", "x1", format = "formula")
#'
#' initVarLink("y ~ x1:0|1")
#'
#' initVarLinks(y ~ x1)
#' initVarLinks("y ~ x1")
#' initVarLinks(c("y ~ x1","y~ x2"))
#' initVarLinks(c(y ~ x1,y ~ x2))
#' initVarLinks(c("y ~ x1","y ~ x2"), format = "formula")
#' initVarLinks(c(y ~ x1,y ~ x2), format = "formula")
#' initVarLinks(c("y ~ x1","y~ x2"), format = "txt.formula")
#' initVarLinks(c(y ~ x1,y ~ x2), format = "txt.formula")

## * initVarLink
#' @rdname initVarLink
#' @export
initVarLink <- function(var1, var2, rep.var1 = FALSE, format = "list",
                         Slink = c(lava.options()$symbols[1],"~"),
                         Scov = lava.options()$symbols[2]){

    format <- match.arg(format, c("list","txt.formula","formula"))
    test.formula <- (class(var1) == "formula")
    test.covariance <- sapply(Scov,grepl,x=var1,fixed=TRUE)
    test.regression <- sapply(Slink,grepl,x=var1,fixed=TRUE)

    if(missing(var2)){
        if(test.formula){
            var2 <- selectRegressor(var1, format = "vars")
            var1 <- selectResponse(var1, format = "vars")
            sep <- if(format == "formula"){"~"}else{Slink}
        }else if(any(test.covariance)){ ## covariance
            Scov <- Scov[test.covariance][1]
            varSplit <- strsplit(var1, split = Scov)[[1]]
            var1 <- trimws(varSplit[1])
            var2 <- trimws(varSplit[2])
            sep <- if(format == "formula"){"~"}else{Scov}
        } else if(any(test.regression)){ ## regression
            Slink <- Slink[test.regression][1]
            varSplit <- strsplit(var1, split = Slink)[[1]]
            var1 <- trimws(varSplit[1])
            var2 <- trimws(varSplit[2])
            sep <- if(format == "formula"){"~"}else{Slink}
        } else {
            var1 <- var1
            var2 <- NA
        }
    }else{

        if(!is.character(var1) || !is.character(var2)){
            stop("\'var1\' and \'var2\' must be characters when both are specified \n")
        }
        sep <- if(format == "formula"){"~"}else{Slink}        
    }
  
  
#### convert to format
    if(format == "formula"){
        n.var2 <- length(var2)
        var1 <- rep(var1, times = n.var2)
        res <- sapply(1:n.var2, function(i){
            stats::as.formula(paste(var1[i], var2[i], sep = sep))
        })
    
  }else if(format == "txt.formula"){
    n.var2 <- length(var2)
    var1 <- rep(var1, times = n.var2)
    res <- sapply(1:n.var2, function(i){paste(var1[i], var2[i], sep = sep)})
    
  }else if(format == "list"){
    if(rep.var1 && !missing(var1)){var1 <- rep(var1, length(var2))}
    res <- list(var1 = var1,
                var2 = if(!missing(var2)){var2}else{NULL} 
    )
  }
 
  ## export 
  return(res)
}

## * initVarLinks
#' @rdname initVarLink
#' @export
initVarLinks <- function(var1, format = "list",...){
        
    if("formula" %in% class(var1)){
        res <- initVarLink(var1, rep.var1 = TRUE, format = format,
                           ...)
    }else {
        res <- sapply(var1, function(x){
            initVarLink(x, rep.var1 = TRUE, format = format,
                        ...)
        })
        if(format == "list"){
            res <- list(var1 = unname(unlist(res["var1",])),
                        var2 = unname(unlist(res["var2",])))
        }else{
            res <- unname(unlist(res))
        }
    }

    return(res)
    
}

## * header
## rm(list = ls())
if(TRUE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))

context("initVarLink")

## * tests

lava.options(symbols = c("~","~~"))

test_that("initVar - two different variables",{
  l <- list(var1 = "a", var2 = "b")  
  f <- a~b
  # regression
  expect_equal(l, initVarLink(var1 = a~b))
  expect_equal(l, initVarLink(var1 = a ~ b))
  expect_equal(list(f),initVarLink(var1 = "a ~ b", format = "formula"))
  # covariance
  expect_equal(l, initVarLink(var1 = "a~~b"))
  # var
  expect_equal(l, initVarLink(var1 = "a", var2 = "b"))
})

test_that("initVar - one repeated variable",{
  
  l <- list(var1 = "X1", var2 = "X1")
  # regression
  expect_equal(l, initVarLink(var1 = X1~X1))
  expect_equal(l, initVarLink(var1 = "X1~~X1"))
})

test_that("initVar - no response variable",{
  
  l1 <- list(var1 = NULL, var2 = "X1")
  l12 <- list(var1 = NULL, var2 = c("X1","X2"))
  # regression
  expect_equal(l1, initVarLink(var1 = ~X1))
  expect_equal(l12, initVarLink(var1 = ~X1+X2))
})


initVarLink(var1 = a ~ b+c+d*e, format = "list")
initVarLink(var1 = a ~ b+c+d*e, format = "txt.formula")
initVarLink(var1 = a ~ b+c+d*e, format = "formula")




initVarLink(var1 = Y~X1+X2)
initVarLink(var1 = Y~X1+X2, rep.var1 = TRUE)
initVarLink(var1 = Y~X1+X2, format = "formula")
initVarLink(var1 = Y~X1+X2, format = "txt.formula")

lava.options(symbols = c("<-","<->"))
initVarLink(var1 = "Y<-X1+X2", rep.var1 = TRUE)
initVarLink(var1 = "Y<-X1+X2", format = "formula")
initVarLink(var1 = "Y<-X1+X2", format = "txt.formula")

### p.adjust2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: dec 19 2019 (11:28) 
## Version: 
## Last-Updated: Jan 11 2022 (17:38) 
##           By: Brice Ozenne
##     Update #: 24
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:


p.adjust2 <- function(p, method, vcov.param = NULL){

    traditional.method <- c("holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none")
    new.method <- c("AB1","AB2")
    all.method <- c(traditional.method,new.method)
    
    if(length(method)!=1){
        stop("Argument \'method\' must have length 1 \n")
    }
    
    if(method %in% traditional.method){
        out <- stats::p.adjust(p = p, method = method)
    }else{
        if(method %in% new.method == FALSE){
            stop("Argument \'method\' must be one of \"",paste(all.method,collapse = "\" \""),"\"\n")
        }
        
        if(method %in% c("AB1","AB2") && is.null(vcov.param)){
            stop("Argument \'vcov.param\' must not be NULL when argument \'method\' is \"AB1\" or \"AB2\" \n")
        }
        if(method %in% c("AB1","AB2")){
            
            if(is.null(names(p))){
                stop("Argument \'p\' must be named \n")
            }
            if(!all(names(p) %in% colnames(vcov.param))){
                stop("The column names of argument \'vcov.param\' must match the names of argument \'p\' \n")
            }

            ## compute average correlation
            M.rho <- stats::cov2cor(vcov.param)
            diag(M.rho) <- NA
            r <- colMeans(abs(M.rho[,names(p),drop=FALSE]), na.rm = TRUE)
            k <- length(p)
            out <- switch(method,
                          "AB1" = pmin(1, p*(k-(k-1)*sqrt(abs(r)))),
                          "AB2" = pmin(1, p*k^(1-sqrt(abs(r))))
                          )
            attr(out,"r") <- r
                   
            
        }
    }

    names(out) <- names(p)

    return(out)
}

######################################################################
### p.adjust2.R ends here

### sCorrect-estimate2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan  3 2018 (14:29) 
## Version: 
## Last-Updated: jan 23 2024 (10:30) 
##           By: Brice Ozenne
##     Update #: 2169
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - estimate2
#' @title  Satterthwaite Correction and Small Sample Correction
#' @description Correct the bias of the ML estimate of the variance and compute the first derivative of the information matrix.
#' @name estimate2
#'
#' @param object a \code{lvm} object.
#' @param param [numeric vector, optional] the values of the parameters at which to perform the correction.
#' @param data [data.frame, optional] the dataset relative to which the correction should be performed.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param df [character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
#' Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
#' Only relevant when using a \code{lvmfit} object. 
#' @param tol.max [numeric >0] the largest acceptable absolute difference between two succesive estimates of the bias correction.
#' @param iter.max [integer >0] the maximum number of iterations used to estimate the bias correction.
#' @param derivative [character] should the first derivative of the information matrix be computed using a formula (\code{"analytic"}) or numerical derivative (\code{"numeric"})?
#' @param hessian [logical] should the hessian be stored? Can be \code{NULL} to indicate only if computed during the small sample correction.
#' @param dVcov.robust [logical] should the first derivative of robust variance-covariance matrix be stored?
#' @param trace [logical] should the execution of the function be traced.
#' @param ...  arguments passed to \code{lava::estimate} when using a \code{lvm} object.
#'
#' @details The argument \code{value} is equivalent to the argument \code{bias.correct} of the function \code{summary2}.
#' 
#' @concept estimator
#' @keywords smallSampleCorrection
#' 
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' dW <- sampleRepeated(10, format = "wide")
#' 
#' #### latent variable model ####
#' m.lvm <- lvm(Y1~X1+X2+Z1)
#'
#' e2.lvm <- estimate2(m.lvm, data = dW)
#' summary2(e2.lvm)
#' 
#' @export
`estimate2` <-
    function(object, param, data,
             ssc, df,
             derivative, hessian, dVcov.robust,
             iter.max, tol.max, trace , ...) UseMethod("estimate2")


## * estimate2.lvm
#' @rdname estimate2
#' @export
estimate2.lvm <- function(object, param = NULL, data = NULL,
                          ssc = lava.options()$ssc, df = lava.options()$df,
                          derivative = "analytic", hessian = FALSE, dVcov.robust = FALSE,
                          iter.max = 100, tol.max = 1e-6, trace = 0, ...){

    if(!is.null(param)){
        warning("Argument \'param\' not used with lvm objects. \n")
    }
    
    out <- lava::estimate(x = object, data = data, ...)
    return(estimate2(out, param = NULL, data = NULL,
                     ssc = ssc, df = df,
                     derivative = derivative, hessian = hessian, dVcov.robust = FALSE,
                     iter.max = iter.max, tol.max = tol.max, trace = trace))

}

## * estimate2.lvmfit
#' @rdname estimate2
#' @export
estimate2.lvmfit <- function(object, param = NULL, data = NULL,
                             ssc = lava.options()$ssc, df = lava.options()$df,
                             derivative = "analytic", hessian = FALSE, dVcov.robust = FALSE,
                             iter.max = 100, tol.max = 1e-6, trace = 0, ...){

    ## ** preliminary tests
    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    if("multigroupfit" %in% class(object)){
        stop("estimate2 cannot handle multigroup models \n")
    }

    if(inherits(object,"lvmfit") && length(object$model$attributes$ordinal)>0){
        name.t <- names(object$model$attributes$ordinal)
        stop("estimate2 does not handle ordinal variables \n",
             "ordinal variable(s): \"",paste(name.t, collapse = "\" \""),"\"\n")
    }
    
    if(inherits(object,"lvmfit") && length(object$model$attributes$transform)>0){
        name.t <- names(object$model$attributes$transform)
        stop("estimate2 does not handle transformed variables \n",
             "transformed variable(s): \"",paste(name.t, collapse = "\" \""),"\"\n")
    }

    ## arguments
    if(identical(ssc,FALSE) || identical(ssc,NA)){
        ssc <- "none"
    }
    if(identical(df,FALSE) || identical(df,NA)){
        df <- "none"
    }
    ssc <- match.arg(tolower(ssc), c("none","residuals","residuals0","cox"))
    df <- match.arg(tolower(df), c("none","satterthwaite"))

    if(df %in% "satterthwaite" || ssc %in% "cox"){
        second.order <- TRUE
    }else if(ssc %in% c("residuals","residuals0")){
        second.order <- FALSE
    }else{
        second.order <- FALSE
    }

    ## ** initialize object
    if(trace>0){cat("Initialization:")}
    object$sCorrect <- moments2(object, data = data, param = param, Psi = NULL,
                                initialize = TRUE, usefit = TRUE,
                                score = TRUE, information = TRUE, hessian = hessian, vcov = TRUE,
                                dVcov = (ssc == "cox")  || (ssc == "none" && df == "satterthwaite"), dVcov.robust = (ssc == "none" && df == "satterthwaite" && dVcov.robust),
                                residuals = TRUE, leverage = FALSE, derivative = derivative)  ## setting leverage to FALSE is like initialization to 0
    if(trace>0){cat(" done \n")}

    ## ** bias correction    
    if(ssc != "none"){

        ## *** initialize bias correction
        if(trace>0){cat("Initialize bias correction \n")}
        if(ssc=="Cox"){
            object.ssc <- list(type = "Cox",
                               param0 = object$sCorrect$param,
                               Omega0 = object$sCorrect$moment$Omega)
        }else if(ssc %in% c("residuals","residuals0")){
            object.ssc <- .init_sscResiduals(object)
        }
        
        ## *** perform bias correction
        if(trace>0){cat("Perform bias correction \n")}
        iCV <- FALSE
        iIter <- 0
        iTol <- Inf
        iiParam <- object$sCorrect$param
        ## cat(iTol," (",iiParam,") \n")
            
        while(iCV == FALSE && iIter < iter.max){
            if(trace>0){cat("*")}

            ## bias correction
            if(ssc == "Cox"){
                iParam <- .sscCoxSnell(object, ssc = object.ssc)
                object.ssc$JJK <- attr(iParam,"JJK")
                object.ssc$lm <- attr(iParam,"lm")
                object.ssc$Psi <- object$sCorrect$moment$Omega - object.ssc$Omega0
            }else if(ssc %in% c("residuals","residuals0")){
                iParam <- .sscResiduals(object, ssc = object.ssc)
                object.ssc$Omega <- attr(iParam,"Omega")
                object.ssc$Psi <- attr(iParam,"Psi")
                ## use previous Omega to compute leverage and residuals
                attr(object.ssc$Omega,"Omega.leverage") <- object$sCorrect$moment$Omega
                attr(object.ssc$Omega,"dOmega.leverage") <- object$sCorrect$dmoment$dOmega
                attr(object.ssc$Omega,"Omega.residuals") <- object$sCorrect$moment$Omega
            }
            ## object.ssc$Omega0 + object.ssc$Psi - object.ssc$Omega

            ## cv criteria
            iIter <- iIter + 1
            iTol <- max(abs(iParam-iiParam))
            ## cat(iTol," (",iParam,") \n")
            iiParam <- iParam
            iCV <- iTol <= tol.max

            ## update moments
            ## if ssc=="residuals0" then do not rescale the residuals according the the bias
            if(iCV==FALSE && iIter < iter.max){
                object$sCorrect <- moments2(object, param = iParam, Psi = if(ssc!="residuals0"){object.ssc$Psi}else{NULL}, Omega = object.ssc$Omega, 
                                            initialize = FALSE, usefit = TRUE,
                                            score = TRUE, information = TRUE, hessian = FALSE, vcov = TRUE,
                                            dVcov = (ssc == "cox"), dVcov.robust = FALSE,
                                            residuals = TRUE, leverage = TRUE, derivative = derivative)
            }else{
                object$sCorrect <- moments2(object, param = iParam, Psi = if(ssc!="residuals0"){object.ssc$Psi}else{NULL}, Omega = object.ssc$Omega, 
                                            initialize = FALSE, usefit = TRUE,
                                            score = TRUE, information = TRUE, hessian = hessian, vcov = TRUE,
                                            dVcov = (df == "satterthwaite"), dVcov.robust = dVcov.robust,
                                            residuals = TRUE, leverage = TRUE, derivative = derivative)

                object$sCorrect$ssc <- c(object.ssc,
                                         cv = iCV,
                                         iter = iIter,
                                         tol = iTol,
                                         iter.max = iter.max,
                                         tol.max = tol.max
                                         )
                
            }
        }
        
        ## *** assess convergence
        if(iCV == FALSE){
            warning("small sample correction did not reach convergence after ",iIter," iteration",if(iIter>1){"s"}else{""},". \n")
        }
        if(trace > 0){
            cat("\n")
        }

    }else{
        object$sCorrect$ssc$type <- "none"
    }

    ## ** degrees of freedom    
    object$sCorrect$df <- df ## degrees of freedom are computed later (by compare2)
    
    ## ** restaure original param order
    name.param <- object$sCorrect$name.param
    if(!is.null(name.param)){
        object$sCorrect$param <- object$sCorrect$param[name.param]
        names(object$sCorrect$param) <- names(name.param)

        if(!is.null(object$sCorrect$score)){
            object$sCorrect$score <- object$sCorrect$score[,name.param,drop=FALSE]
            colnames(object$sCorrect$score) <- names(name.param)
        }
        if(!is.null(object$sCorrect$information)){
            object$sCorrect$information <- object$sCorrect$information[name.param,name.param,drop=FALSE]
            dimnames(object$sCorrect$information) <- list(names(name.param),names(name.param))
        }
        if(!is.null(object$sCorrect$vcov.param)){
            object$sCorrect$vcov.param <- object$sCorrect$vcov.param[name.param,name.param,drop=FALSE]
            dimnames(object$sCorrect$vcov.param) <- list(names(name.param),names(name.param))
        }
        if(!is.null(object$sCorrect$hessian)){
            object$sCorrect$hessian <- object$sCorrect$hessian[name.param,name.param,,drop=FALSE]
            dimnames(object$sCorrect$hessian) <- list(names(name.param),names(name.param),NULL)
        }
        if(!is.null(object$sCorrect$dInformation)){
            object$sCorrect$dInformation <- object$sCorrect$dInformation[name.param,name.param,name.param,drop=FALSE]
            dimnames(object$sCorrect$dInformation) <- list(names(name.param),names(name.param),names(name.param))
        }
        if(!is.null(object$sCorrect$dVcov.param)){
            object$sCorrect$dVcov.param <- object$sCorrect$dVcov.param[name.param,name.param,name.param,drop=FALSE]
            dimnames(object$sCorrect$dVcov.param) <- list(names(name.param),names(name.param),names(name.param))
        }
        if(!is.null(object$sCorrect$dRvcov.param)){
            object$sCorrect$dRvcov.param <- object$sCorrect$dRvcov.param[name.param,name.param,name.param,drop=FALSE]
            dimnames(object$sCorrect$dRvcov.param) <- list(names(name.param),names(name.param),names(name.param))
        }
    }
    
    ## ** export
    class(object) <- append("lvmfit2",class(object))
    return(object)    
}

## * estimate2.list
#' @rdname estimate2
#' @export
estimate2.list <- function(object, ...){
    object.class <- class(object)
    object <- lapply(object, estimate2, ...)
    class(object) <- object.class
    return(object)
}

## * estimate2.mmm
#' @rdname estimate2
#' @export
estimate2.mmm <- estimate2.list

##----------------------------------------------------------------------
### sCorrect.R ends here

## * modelsearch2 (documentation)
#' @title Data-driven Extension of a Latent Variable Model
#' @description Procedure adding relationship between variables that are supported by the data.
#' @name modelsearch2
#' 
#' @param object a \code{lvmfit} object.
#' @param link [character, optional for \code{lvmfit} objects] the name of the additional relationships to consider when expanding the model. Should be a vector containing strings like "Y~X". See the details section.
#' @param data [data.frame, optional] the dataset used to identify the model
#' @param method.p.adjust [character] the method used to adjust the p.values for multiple comparisons.
#' Can be any method that is valid for the \code{stats::p.adjust} function (e.g. \code{"fdr"}).
#' Can also be \code{"max"}, \code{"fastmax"}, or \code{"gof"}.
#' @param method.maxdist [character] the method used to estimate the distribution of the max statistic.
#' \code{"resampling"} resample the score under the null to estimate the null distribution.
#' \code{"bootstrap"} performs a wild bootstrap of the iid decomposition of the score to estimate the null distribution.
#' \code{"approximate"} attemps to identify the latent gaussian variable corresponding to each score statistic (that is chi-2 distributed).
#' It approximates the correlation matrix between these latent gaussian variables and uses numerical integration to compute the distribution of the max.
#' @param n.sample [integer, >0] number of samples used in the resampling approach.
#' @param na.omit should tests leading to NA for the test statistic be ignored. Otherwise this will stop the selection process.
#' @param alpha [numeric 0-1] the significance cutoff for the p-values.
#' When the p-value is below, the corresponding link will be added to the model
#' and the search will continue. Otherwise the search will stop.
#' @param nStep the maximum number of links that can be added to the model.
#' @param trace [logical] should the execution of the function be traced?
#' @param cpus the number of cpus that can be used for the computations.
#'
#' @details
#' method.p.adjust = \code{"max"} computes the p-values based on the distribution of the max statistic.
#' This max statistic is the max of the square root of the score statistic.
#' The p-value are computed integrating the multivariate normal distribution.
#' 
#' method.p.adjust = \code{"fastmax"} only compute the p-value for the largest statistic.
#' It is faster than \code{"max"} and lead to identical results.
#' 
#' method.p.adjust = \code{"gof"} keep adding links until the chi-squared test (of correct specification of the covariance matrix) is no longer significant.
#' @return A list containing:
#' \itemize{
#' \item sequenceTest: the sequence of test that has been performed.
#' \item sequenceModel: the sequence of models that has been obtained.
#' \item sequenceQuantile: the sequence of rejection threshold. Optional. 
#' \item sequenceIID: the influence functions relative to each test. Optional. 
#' \item sequenceSigma: the covariance matrix relative to each test. Optional. 
#' \item initialModel: the model before the sequential search.
#' \item statistic: the argument \code{statistic}.
#' \item method.p.adjust: the argument \code{method.p.adjust}.
#' \item alpha: [numeric 0-1] the significance cutoff for the p-values.
#' \item cv: whether the procedure has converged.
#' } 
#'
#' @concept modelsearch
#' @export
`modelsearch2` <-
    function(object, link, data,
             method.p.adjust, method.maxdist, n.sample, na.omit, 
             alpha,  nStep, trace, cpus) UseMethod("modelsearch2")


## * modelsearch2 (example)
#' @rdname modelsearch2
#' @examples
#'
#' ## simulate data 
#' mSim <- lvm()
#' regression(mSim) <- c(y1,y2,y3,y4)~u
#' regression(mSim) <- u~x1+x2
#' categorical(mSim,labels=c("A","B","C")) <- "x2"
#' latent(mSim) <- ~u
#' covariance(mSim) <- y1~y2
#' transform(mSim, Id~u) <- function(x){1:NROW(x)}
#'
#' set.seed(10)
#' df.data <- lava::sim(mSim, n = 1e2, latent = FALSE)
#' 
#' ## only identifiable extensions
#' m <- lvm(c(y1,y2,y3,y4)~u)
#' latent(m) <- ~u
#' addvar(m) <- ~x1+x2
#' 
#' e <- estimate(m, df.data)
#'
#' \dontrun{
#' resSearch <- modelsearch(e)
#' resSearch
#'
#' resSearch2 <- modelsearch2(e, nStep = 2)
#' resSearch2
#' }
#' \dontshow{
#' search.link <- c("u~x1","u~x2","y1~x1","y1~x2","y1~~y2","y1~~y3")
#' resSearch2 <- modelsearch2(e, nStep = 2, link = search.link)
#' resSearch2
#' }
#'
#' ## some extensions are not identifiable
#' m <- lvm(c(y1,y2,y3)~u)
#' latent(m) <- ~u
#' addvar(m) <- ~x1+x2 
#'
#' e <- estimate(m, df.data)
#'
#' \dontrun{
#' resSearch <- modelsearch(e)
#' resSearch
#' resSearch2 <- modelsearch2(e)
#' resSearch2
#' }
#'
#' ## for instance
#' mNI <- lvm(c(y1,y2,y3)~u)
#' latent(mNI) <- ~u
#' covariance(mNI) <- y1~y2
#' ## estimate(mNI, data = df.data)
#' ## does not converge
#'
#' 
#' 

## * modelsearch2.lvmfit (code)
#' @export
modelsearch2.lvmfit <- function(object, link = NULL, data = NULL, 
                                method.p.adjust = "fastmax", method.maxdist = "approximate", n.sample = 1e5, na.omit = TRUE, 
                                alpha = 0.05, nStep = NULL, 
                                trace = TRUE, cpus = 1){

    ## ** check arguments
    ## object
    if(any(is.na(model.frame(object))) && method.p.adjust %in% c("max","fastmax")){
        warning("Missing values - the iid decomposition of the test statistics will only be computed on complete data \n")
    }
    
    ## methods
    method.p.adjust <- match.arg(method.p.adjust, c("fastmax", "max", "holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "none","gof"))    
    if(method.p.adjust == "gof" ){
        method.p.adjust <- "none"
        stop.gof <- TRUE
    }else{
        stop.gof <- FALSE
    }

    if(n.sample<0 || (n.sample %% 1 != 0) ){
        stop("Argument \'n.sample\' must be a positive integer \n")
    }
    method.maxdist <- match.arg(method.maxdist, c("approximate","resampling","bootstrap"))    

    ## cpus
    if(is.null(cpus)){ cpus <- parallel::detectCores()}

        if(is.null(cpus) || cpus > 1){
        test.package <- try(requireNamespace("foreach"), silent = TRUE)
        if(inherits(test.package,"try-error")){
            stop("There is no package \'foreach\' \n",
                 "This package is necessary when argument \'cpus\' is greater than 1 \n")
        }
    }
    
    if(!is.null(cpus) && cpus>1){
        if(cpus > parallel::detectCores()){
            stop("Argument \'cpus\' is greater than the number of available CPU cores \n",
                 "available CPU cores: ",parallel::detectCores(),"\n")
        }
    }

    ## ## extra arguments 
    ## dots <- list(...)
    ## if(length(dots)>0){
    ##     stop("modelsearch2 does not take any extra arguments \n",
    ##          "name of the extra arguments: \"",paste(names(dots), collapse = "\" \""),"\" \n")
    ## }

    ## ** prepare
    ## *** data
    if(is.null(data)){
        data <- as.data.frame(stats::model.frame(object, all = TRUE))
    }
    
    ## *** normalize the links
    if(is.null(link)){
        res.find <- do.call(findNewLink,
                            args = c(list(object$model,
                                          data = data,
                                          output = "names")))
        directive <- res.find$directional
        restricted <- res.find$M.links
        link <- res.find$link
        if(is.null(link)){
            stop("Automatic search has not found any possible additional link \n",
                 "Consider specifying manually the argument \'link\' \n")
        }
    }else{
        resLink <- .initializeLinks(object, data = data, link = link)
        object <- resLink$object
        link <- resLink$link
        directive <- resLink$directive
        restricted <- resLink$restricted
    }    
    
    ## ** initialization
    if(is.null(nStep)){
        nStep <- NROW(restricted)
    }
    iStep <- 1
    iRestricted <- restricted
    iDirective <- directive
    iLink <- link
    iObject <- object

    ## update of the model
    add.args <- setdiff(names(object$call), c("","object","data","control"))
    ls.call <- lapply(add.args, function(arg){object$call[[arg]]})
    names(ls.call) <- add.args

    ls.call$data <- data
    if(!is.null(data)){
        index.cols <- which(names(data)%in%names(ls.call$data)==FALSE)
        if(length(index.cols)>0){
            ls.call$data <- cbind(ls.call$data,
                                  subset(as.data.frame(data), select = index.cols))
        }
    }
    if(is.null(object$control)){
        ls.call$control <- list()
    }else{
        ls.call$control <- object$control
    }
    ls.call$control$trace <- FALSE
    
    ## output
    ls.seqTests <- list()
    ls.seqModels <- list()
    ls.seqIID <- list() # only for method.p.adjust = "max"
    ls.seqSigma <- list() # only for method.p.adjust = "max"
    vec.seqQuantile <- NULL # only for method.p.adjust = "max"
    
    ## criterion    
    cv <- FALSE

    ## define cluster
    if(cpus>1){
        if(trace>0){
            cl <- parallel::makeCluster(cpus, outfile = "")
        }else{
            cl <- parallel::makeCluster(cpus)
        }
        doParallel::registerDoParallel(cl)
        
        vec.packages <- c("lavaSearch2","lava")
        parallel::clusterCall(cl, fun = function(x){
            sapply(vec.packages, function(iP){
                suppressPackageStartupMessages(attachNamespace(iP)) ## requireNamespace did not worked
            })
        })
        
    }else{
        cl <- NULL
    }

    ## ** display a summary of the call
    if(trace>0){

        cat("\n",
            "** Sequential variable selection using the score statistic ** \n",
            " Number of possible additional links         : ",length(link)," \n",
            " Maximum number of steps                     : ",nStep,"\n",        
            " Adjustment method for multiple comparisons  : ",method.p.adjust,"\n",
            " Confidence level                            : ",1-alpha,"\n",
            " Number of cpus                              : ",cpus,"\n\n",
            sep="")
    }

    ## ** Forward search
    if(stop.gof){
        if(trace>0){
            cat("p.Chi-squared test = ",gof(iObject)$fit$p.value,"\n", sep = "")
        }
        if(gof(iObject)$fit$p.value >= alpha){
            cv <- TRUE
        }
    }
        
    while(iStep <= nStep && NROW(iRestricted)>0 && cv==FALSE){
        if(trace >= 1){cat("Step ",iStep,":\n",sep="")}

        
        
        resStep <- .oneStep_scoresearch(iObject, data = data,
                                        restricted = iRestricted, link = iLink, directive = iDirective,
                                        method.p.adjust = method.p.adjust, method.maxdist = method.maxdist, n.sample = n.sample,
                                        cl = cl, trace = trace)

        ## ** update according the most significant p.value
        ## *** check convergence
        if(stop.gof){
            cv <- FALSE
            test.na <- FALSE
        }else if(na.omit || method.p.adjust == "fastmax"){
            cv <- all(stats::na.omit(resStep$test$adjusted.p.value) > alpha)
            test.na <- FALSE
        }else{
            cv <- all(resStep$test$adjusted.p.value > alpha)
            if(is.na(cv)){                    
                cv <- TRUE
                test.na <- TRUE
            }else{
                test.na <- FALSE
            }
        }

        ## *** identify most promising test
        index.maxTest <- which.max(abs(resStep$test$statistic))[1]
        resStep$test$selected <- FALSE
        resStep$test[index.maxTest,"selected"] <- (resStep$test[index.maxTest,"adjusted.p.value"] <= alpha)
        resStep$test$nTests <- NROW(resStep$test)
        resStep$test <- resStep$test[order(resStep$test$statistic),]

        ## *** update the model
        if(cv==FALSE){
            ls.call$x <- addLink(iObject$model, var1 = iRestricted[index.maxTest,1], var2 = iRestricted[index.maxTest,2],
                                 covariance = 1-iDirective[index.maxTest])

            ## first attempt
            ls.call$control$start <- stats::coef(iObject)
            suppressWarnings(
                iObject <- tryCatch(do.call(lava::estimate, args = ls.call),
                                    error = function(x){x},
                                    finally = function(x){x})
            )

            ## second attempt
            if(inherits(iObject,"error") || iObject$opt$convergence>0){
                ls.call$control$start <- NULL
                suppressWarnings(
                    iObject <- do.call(lava::estimate, args = ls.call)
                )
                if(inherits(iObject,"error") || iObject$opt$convergence>0){
                    stop("Estimation of the extended latent variable model did not converge \n")
                }
            }

            ## update links
            iLink <- iLink[-index.maxTest]
            iRestricted <- iRestricted[-index.maxTest,,drop=FALSE]
            iDirective <- iDirective[-index.maxTest]
        }
        


        ## *** update the output
        ls.seqTests[[iStep]] <- resStep$test
        ls.seqModels[[iStep]] <- iObject
        if(method.p.adjust %in% c("max","fastmax")){
            ls.seqIID[[iStep]] <- resStep$iid
            ls.seqSigma[[iStep]] <- resStep$Sigma
            vec.seqQuantile <- c(vec.seqQuantile,resStep$test$quantile[1])
        }


        ## *** display results
        if(trace > 0){
            rowSelected <- NROW(resStep$test)
            if(cv==FALSE){
                cat("add ",as.character(resStep$test[rowSelected, "link"]),
                    " (statistic = ",resStep$test[rowSelected, "statistic"],
                    ", adjusted.p.value = ",resStep$test[rowSelected, "adjusted.p.value"],
                    ")\n",sep="")
            }else{
                if(test.na){
                    cat("NA among the test statistics \n")
                }else{
                    cat("no variable to add",
                        " (statistic = ",resStep$test[rowSelected, "statistic"],
                        ", adjusted.p.value = ",resStep$test[rowSelected, "adjusted.p.value"],
                        ")\n",sep="")
                }
            }
        }

        ## *** check convergence gof
        if(stop.gof){
            if(trace>0){
                cat("p.Chi-squared test = ",gof(iObject)$fit$p.value,"\n", sep = "")
            }
            if(gof(iObject)$fit$p.value >= alpha){
                cv <- TRUE
            }
        }
    

        
        iStep <- iStep + 1
    }

    if(cpus>1){
        parallel::stopCluster(cl)
    }

    ## ** export
    if(length(ls.seqIID)==0){ls.seqIID <- NULL}
    if(length(ls.seqSigma)==0){ls.seqSigma <- NULL}
    output <- list(sequenceTest = ls.seqTests,
                   sequenceModel = ls.seqModels,
                   sequenceQuantile = vec.seqQuantile,
                   sequenceIID = ls.seqIID,
                   sequenceSigma = ls.seqSigma,
                   initialModel = object,
                   method.p.adjust = method.p.adjust,
                   alpha = alpha,
                   cv = cv)
    class(output) <- "modelsearch2"
    return(output)
}

## * .initializeLinks
.initializeLinks <- function(object, data, link){
    restricted <- do.call(cbind,initVarLinks(link))
    directive <- rep(TRUE, length(link))

    ## ** identify covariance link
    index.Ndir <- grep(lava.options()$symbols[2],link,fixed=TRUE)
    if(length(index.Ndir)>0){
        directive[index.Ndir] <- FALSE
    }
   
    ## ** get all vars
    if(is.null(data)){            
        data <- evalInParentEnv(object$call$data)             
        if(is.null(data)){
            data <- lava::sim(object,1)
        }            
    }
    ## ** take care of categorical variables
    iData <- try(eval(object$call$d), silent = TRUE)
    if(!inherits(iData, "data.frame")){
        iData <- evalInParentEnv(object$call$data)
        if(!inherits(iData, "data.frame")){
            stop("Could not identify argument data in object$call \n")
        }
    }

    M.linkvar <- do.call(rbind,lapply(1:NROW(restricted), function(row){ ## row <- 2
        data.frame(Y = unname(restricted[row,1]),
                   X = unname(var2dummy(object$model,
                                        data = iData,
                                        var = restricted[row,2])),
                   dir = directive[row],
                   stringsAsFactors = FALSE)
    }))

    ## check no missing covariance links
    index.regression <- which(M.linkvar[,"dir"]==TRUE)
    if(length(index.regression)>0){
        index.covariance <- which(M.linkvar[index.regression,"X"] %in% lava::endogenous(object) + M.linkvar[index.regression,"X"] %in% lava::endogenous(object) ==2)
        if(length(index.covariance)>0){
            stop("Covariance links must be indicated with the symbol \"",lava.options()$symbols[2],"\" \n",
                 "Possible covariance links: ",paste0(link[index.regression][index.covariance], collapse = " "),"\n")
        }
    }
    restricted2 <- as.matrix(M.linkvar[,1:2,drop=FALSE])
    directive <- M.linkvar[,3]
    link <- paste0(restricted2[,1],lava.options()$symbols[2-directive],restricted2[,2])

    ## ** check links
    allVars.link <- setdiff(unique(as.vector(restricted2)), lava::latent(object$model))
    allVars.model <- lava::vars(object$model)
    allVars.data <- names(data)

    if(any(allVars.link %in% allVars.model == FALSE)){
        missing.var <- allVars.link[allVars.link %in% allVars.model == FALSE]
        if(any(allVars.link %in% allVars.data == FALSE)){
            missing.var <- allVars.link[allVars.link %in% allVars.data == FALSE]
            stop("Some links contains variables that are not in the latent variable model \n",
                 "variables(s) : \"",paste(missing.var,collapse ="\" \""),"\"\n")
        }

    }
    
    ## ** check covariance links
    if(any(directive==FALSE)){
        if(any(restricted2[directive==FALSE,1] %in% lava::exogenous(object)) || any(restricted2[directive==FALSE,2] %in% lava::exogenous(object))){
            wrong <- union(which(restricted2[directive==FALSE,1] %in% exogenous(object)),
                           which(restricted2[directive==FALSE,2] %in% exogenous(object)))
            stop("Covariance links can only relate endogenous variables \n",
                 "Covariance link(s) involving exogenous variables: ,", paste(link[wrong], collapse = " ; "),"\n")
        }
    }
    
     return(list(object = object,
                link = link,
                directive = directive,
                restricted = restricted2))
}
## * .oneStep_scoresearch
.oneStep_scoresearch  <- function(object, data,
                                  restricted, link, directive,
                                  method.p.adjust, alpha, method.maxdist, n.sample,
                                  cl, trace){

    ## ** initialization
    n.link <- NROW(restricted)
    coef.object <- coef(object)
    namecoef.object <- names(coef.object)
    ncoef.object <- length(coef.object)
    type.information <- lava.options()$search.type.information
    type.statistic <- lava.options()$search.sample.stat
    
    ## ** warper
    warper <- function(iterI){ # iterI <- 1

        out <- list(table = data.frame(statistic = as.numeric(NA),
                                       p.value = as.numeric(NA),
                                       adjusted.p.value = as.numeric(NA),
                                       dp.Info = as.numeric(NA),
                                       stringsAsFactors = FALSE),
                    iid = NULL)

        ## *** define extended model
        newModel <- addLink(object$model, var1 = restricted[iterI,1], var2 = restricted[iterI,2],
                            covariance = 1-directive[iterI])

        ## *** remove useless variables
        Mlink <- newModel$M + (newModel$cov - diag(1, NROW(newModel$cov), NCOL(newModel$cov)))
        noLink.var <- names(which((rowSums(Mlink)==0)+(colSums(Mlink)==0)==2))
        if(length(noLink.var)>0){
            rmvar(newModel) <- noLink.var
        }

        ## *** compute sufficient statistics
        ## necessary otherwise information can have a weird behavior, e.g.
        ## library(lava)
        ##
        ## mSim <- lvm(Y ~ X1, X2 ~ eta, X3 ~ eta, X4 ~ eta)
        ## latent(mSim) <- ~eta
        ## d <- sim(mSim, 100)
        ##
        ## m <- lvm(Y~X1+X2) 
        ## e <- estimate(m, d)
        ## information(e, data = d, p = coef(e)) ## gold standard
        ## information(m, data = d, p = coef(e)) ## issue
        ##
        ## fix
        ## ss <- lava:::procdata.lvm(m, data = d, missing = FALSE) 
        ## mm <- lava::fixsome(m, measurement.fix=TRUE, S=ss$S, mu=ss$mu, n = ss$n, debug=FALSE)
        ## information(mm, data = d, p = coef(e)) ## ok
        suffStat <- lava_procdata.lvm(newModel, data = data, missing = inherits(object,"lvm.missing")) 
        newModel <- lava::fixsome(newModel, measurement.fix=TRUE, S=suffStat$S, mu=suffStat$mu, n = suffStat$n, debug=FALSE)

        ## *** define constrained coefficients
        coef0.new <- stats::setNames(rep(0, ncoef.object+1), coef(newModel))
        coef0.new[namecoef.object] <- coef.object

        ## *** compute the iid decomposition and statistic
        namecoef.newobject <- names(coef0.new)
        Info <- lava::information(newModel, p = coef0.new, n = NROW(data), type = type.information, data = data)
        dimnames(Info) <- list(namecoef.newobject,namecoef.newobject)
        
        if(method.p.adjust %in% c("max","fastmax")){
            iid.score <- lava::score(newModel, p = coef0.new, data = data, indiv = TRUE)
            ## rm na
            iid.score <- iid.score[rowSums(is.na(iid.score))==0,]
            if(method.maxdist == "approximate"){
                ## compute decomposition
                out$iid <-  iid.score %*% solve(Info) %*% cbind(colSums(iid.score))
                colnames(out$iid) <- link[iterI]
                ## out$iid <- out$iid/sqrt(sum(out$iid^2))
                out$table$statistic <- sum(out$iid)
                out$table$dp.Info <- TRUE
                
            }else if(method.maxdist %in% c("resampling","bootstrap")){
                n.sample <- NROW(iid.score)

                InfoM12 <- matrixPower(Info, power  = -1/2, symmetric = TRUE, tol = 1e-15, print.warning = FALSE)
                out$table$dp.Info <- !("warning" %in% names(attributes(InfoM12)))
                dimnames(InfoM12) <- list(namecoef.newobject,namecoef.newobject)
                
                ## initial version
                ## linComb <- cbind(1, -solve(Info[link[iterI],link[iterI],drop=FALSE]) %*% Info[link[iterI],namecoef.object,drop=FALSE]) %*% Info[c(link[iterI],namecoef.object),c(link[iterI],namecoef.object)]
                if(FALSE){
                    InfoM1 <- crossprod(InfoM12)
                    dimnames(InfoM1) <- list(namecoef.newobject,namecoef.newobject)
                    linComb <- cbind(1, -Info[link[iterI],namecoef.object,drop=FALSE] %*% solve(Info[namecoef.object,namecoef.object,drop=FALSE])) %*% Info[c(link[iterI],namecoef.object),c(link[iterI],namecoef.object)]
                    iid.theta <- iid.score %*% InfoM1
                    colnames(iid.theta) <- namecoef.newobject
                    out$iid <- iid.theta[,link[iterI]] %*% linComb[,namecoef.newobject,drop=FALSE] %*% InfoM12
                    colnames(out$iid) <- paste0(link[iterI],":",namecoef.newobject)
                }
                ## short version
                out$iid <- (iid.score[,link[iterI],drop=FALSE] - iid.score[,namecoef.object,drop=FALSE] %*% solve(Info[namecoef.object,namecoef.object,drop=FALSE]) %*% Info[namecoef.object,link[iterI],drop=FALSE]) %*% InfoM12[link[iterI],,drop=FALSE]
                colnames(out$iid) <- paste0(link[iterI],":",namecoef.newobject)

                out$table$statistic <- as.double(crossprod(colSums(out$iid))) ## first order approximation (almost identical to exact value)
            }
        }else{
            ## ee.lvm <- estimate(newModel, data = data)
            ## SS <- score(ee.lvm, p = coef0.new)
            ## II <- information(ee.lvm, p = coef0.new)
            ## SS %*% solve(I) %*% t(SS)
            score <- lava::score(newModel, p = coef0.new, indiv = FALSE, data = data)
            
            out$table$statistic <- as.double(score %*% solve(Info) %*% t(score))
            ## range(Info - II)
            ## range(score - SS)
        }
        return(out)
    }

    ## ** compute score tests
    if(trace>0){
        cat(" - compute score test for all possible additional links \n")
    }
    
    if(!is.null(cl)){
    
        if(trace > 0){
            pb <- utils::txtProgressBar(max = n.link, style = 3) 
        }

        
        ## get influence function
        i <- NULL # [:for CRAN check] foreach
        res <- foreach::`%dopar%`(
                            foreach::foreach(i = 1:n.link,
                                             .export = c("lava_procdata.lvm"),
                                             .combine = function(res1,res2){
                                                 res <- list(table = rbind(res1$table,res2$table),
                                                             iid = cbind(res1$iid,res2$iid))
                                                 return(res)
                                             }), {
                                                 if(trace>0){utils::setTxtProgressBar(pb, i)}
                                                 return(warper(i))
                                             })

        if(trace>0){close(pb)}
        
    }else{
        
        if(trace>0){
            resApply <- pbapply::pblapply(1:n.link, warper)            
        }else{
            resApply <- lapply(1:n.link, warper)
        }
        res <- list(table = do.call(rbind, lapply(resApply,"[[","table")),
                    iid = do.call(cbind,lapply(resApply,"[[","iid")))
        
    }
    ## index.iid <- unlist(lapply(1:n.link, function(iL){ ## iL <- 1
    ## rep(iL,times = NCOL(res$iid[[iL]]))
    ## }))
    table.test <- data.frame(link = link, res$table, error = NA, stringsAsFactors = FALSE)
    iid.link <- res$iid

    ## ** p.value
    statistic <- as.numeric(table.test[,"statistic"])
    if(any(statistic<0)){
        stop("Negative score statistic \n")
    }
    ## univariate rejection area
    table.test[,"p.value"] <- 1-stats::pchisq(statistic, df = 1)

    ## ** adjusted p.value
    if(method.p.adjust %in% c("fastmax","max")){
        if(method.maxdist == "approximate"){
            outDistMax <- .approxMaxDistChi2(table = table.test, iid = iid.link, statistic = statistic, method.p.adjust = method.p.adjust,
                                            link = link, n.link = n.link,
                                            search.calc.quantile.int = lava.options()$search.calc.quantile.int, alpha = alpha,
                                            cl = cl, trace = trace)
        }else if(method.maxdist %in% c("resampling","bootstrap")){
            outDistMax <- .sampleMaxDistChi2(table = table.test, iid = iid.link, statistic = statistic, method.p.adjust = method.p.adjust,
                                        link = link, n.link = n.link, n.sample = n.sample, method = method.maxdist,
                                        cl = cl, trace = trace)
        }

        table.test <- outDistMax$table
        Sigma <- outDistMax$Sigma

    }else{        
        table.test[, "adjusted.p.value"] <- stats::p.adjust(table.test$p.value, method = method.p.adjust)
        table.test[, "quantile"] <- as.numeric(NA)
        Sigma <- NULL        
    }


    return(list(test = table.test,
                Sigma = Sigma,
                iid = iid.link))
}


## * approxMaxDistChi2
.approxMaxDistChi2 <- function(table, iid, statistic, method.p.adjust,
                               link, n.link,
                               search.calc.quantile.int, alpha,
                               cl, trace){
        Sigma <- stats::cor(iid)
        dimnames(Sigma) <- list(link,link)
        if(method.p.adjust == "fastmax"){
            index.maxstat <- which.max(statistic)
        
            resInt <- .calcPmaxIntegration(statistic = sqrt(statistic[index.maxstat]), p = n.link,
                                           Sigma = Sigma, df = NULL, distribution = "gaussian")
        
            table[index.maxstat, "adjusted.p.value"] <- as.double(resInt)
            table[index.maxstat, "error"] <- attr(resInt,"error")
        }else if(method.p.adjust == "max"){

            if(is.null(cl)){
                if(trace>0){      
                    ls.resInt <- pbapply::pblapply(1:n.link, function(i){
                        .calcPmaxIntegration(statistic = sqrt(statistic[i]), p = n.link,
                                             Sigma = Sigma, df = NULL, distribution = "gaussian")
                    })            
                }else{
                    ls.resInt <- lapply(1:n.link, function(i){
                        .calcPmaxIntegration(statistic = sqrt(statistic[i]), p = n.link,
                                             Sigma = Sigma, df = NULL, distribution = "gaussian")
                    })            
                }
            }else{
            
                if(trace>0){
                    pb <- utils::txtProgressBar(max = n.link, style = 3)                   
                }

                ## export package
                parallel::clusterCall(cl, fun = function(x){
                    suppressPackageStartupMessages(requireNamespace("mvtnorm", quietly = TRUE))
                })
        
                value <- NULL # [:for CRAN check] foreach
                ls.resInt <- foreach::`%dopar%`(
                                          foreach::foreach(value = 1:n.link,
                                                           .export = c(".calcPmaxIntegration")),
                                          {
                                              if(trace>0){utils::setTxtProgressBar(pb, value)}
                                              return(.calcPmaxIntegration(statistic = sqrt(statistic[value]), p = n.link,
                                                                          Sigma = Sigma, df = NULL, distribution = "gaussian"))
                                          })

                if(trace>0){close(pb)}
            }
            names(ls.resInt) <- link
        
            table[, "adjusted.p.value"] <- unlist(lapply(ls.resInt,as.double))
            table[, "error"] <- unlist(lapply(ls.resInt,attr,"error"))
        }
        
        if(lava.options()$search.calc.quantile.int){
            table[, "quantile"] <- .calcQmaxIntegration(alpha = alpha, p = n.link,
                                                             Sigma = Sigma,
                                                             df = NULL, distribution = "gaussian")
        }

        return(list(table = table,
                    Sigma = Sigma))
}
## * sampleMaxDistChi2
.sampleMaxDistChi2 <- function(table, iid, statistic, method.p.adjust,
                               link, n.link, n.sample, method,
                               cl, trace){
    p <- NCOL(iid)
    n <- NROW(iid)
    ls.name <- strsplit(colnames(iid), split = ":")
    vec.model <- unlist(lapply(ls.name,"[[",1))
    Umodel <- unique(vec.model)
    ls.indexModel <- tapply(1:length(vec.model),vec.model,list)

    ## ** sampling under H0

    ## *** resampling    
    if(method == "resampling"){
        Sigma <- crossprod(iid)
        sample2 <- mvtnorm::rmvnorm(n.sample, mean = rep(0,p), sigma = Sigma)^2
        M.scoreStat <- do.call(cbind,lapply(1:n.link, function(iModel){ ## iModel <- 1
            return(rowSums(sample2[,ls.indexModel[[iModel]],drop=FALSE]))
        }))
    }

    ## *** wild bootstrap
    if(method == "bootstrap"){
        Sigma <- NULL
        M.scoreStat <- wildBoot_cpp(iid = iid,
                                    lsIndexModel = lapply(ls.indexModel, function(x){x-1}),
                                    nSample = n.sample,
                                    nObs = n,
                                    nModel = n.link,
                                    p = p)
    }
    ## *** check
    ## apply(M.scoreStat,2, function(x){1-mean(x <= qchisq(0.95, df = 1))})
    ## hist(M.scoreStat[,1], freq = FALSE)
    ## points(seq(0,15,0.1), dchisq(seq(0,15,0.1), df = 1), col = "red", type = "l")
    
    
    ## ** p-value for each statistic
    p.value <- colMeans(sweep(M.scoreStat, MARGIN = 2, FUN = ">", STATS = statistic)) + 1/2 * colMeans(sweep(M.scoreStat, MARGIN = 2, FUN = "==", STATS = statistic))
    ## mean(M.scoreStat[,1] > statistic[1])

    ## ** p-value for the max statistic
    maxScoreStat <- apply(M.scoreStat,1,max)
    p.value.max <- sapply(statistic, function(iT){mean( maxScoreStat>iT + 0.5*(maxScoreStat==iT))})
    ## p.value.max / p.value
    table[, "p.value"] <- p.value
    table[, "adjusted.p.value"] <- p.value.max
    table[, "error"] <- NA
    return(list(table = table,
                Sigma = Sigma))

}



## * iidConstrainscore (obsolete)
## iidConstrainScore <- function(object, newobject){
##     if(all(sapply(newobject$mean,function(x){all(is.na(x))}))){
##         suffStat <- lava:::procdata.lvm(newobject, data = object$data$model.frame, missing = FALSE) 
##         newobject <- lava::fixsome(newobject, measurement.fix=TRUE, S=suffStat$S, mu=suffStat$mu, n = suffStat$n, debug=FALSE)
##     }
    
##     param0 <- coef(object)
##     name.param0 <- names(param0)
##     name.param <- coef(newobject)
##     n.param <- length(name.param)

##     newparam <- setNames(rep(0,n.param), name.param)
##     newparam[name.param0] <- param0
##     extraparam <- setdiff(name.param, name.param0)

##     S <- score(newobject, p = newparam, data = object$data$model.frame, indiv = TRUE)
##     I <- information(newobject, p = newparam, data = object$data$model.frame)
##     dimnames(I) <- list(name.param,name.param)
##     iInfo <- solve(I)
    
##     linComb <- cbind(1, solve(I[extraparam,extraparam,drop=FALSE]) %*% I[extraparam,name.param0,drop=FALSE]) %*% I[c(extraparam,name.param0),c(extraparam,name.param0)]
##     iid <- S %*% iInfo
##     out <- sweep(iid, MARGIN = 2, FUN = "*", STATS = as.double(linComb[,name.param]))
##     return(out/sqrt(object$data$n))
## }

### test2-modelsearch2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: jan 22 2018 (11:45) 
## Version: 
## Last-Updated: Jan 12 2022 (14:47) 
##           By: Brice Ozenne
##     Update #: 25
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(FALSE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))

context("lavaSearch2")

## * example with one additional link
n <- 100
m.sim <- lvm(Y~E+0*X1)
m <- lvm(Y~E)
addvar(m) <- ~X1

set.seed(12)
df.sim <- lava::sim(m.sim, n=100, latent = FALSE)
e.base <- estimate(m, data = df.sim)

test_that("Score 1 link",{
    GS.score <- modelsearch(e.base, silent = TRUE)
    index.coef <- which(GS.score$res[,"Index"]=="Y~X1")

    search.holm <- modelsearch2(e.base, method.p.adjust = "holm", trace = 0)
    ## c("E", "hessian", "varS", "outer", "sandwich", "robust", "num"),   "outer"
    
    expect_equal(as.double(GS.score$test[index.coef,"Test Statistic"]),
                 as.double(search.holm$sequenceTest[[1]][1,"statistic"]), tol = 1e-9)
    expect_equal(as.double(GS.score$test[index.coef,"P-value"]),
                 as.double(search.holm$sequenceTest[[1]][1,"p.value"]), tol = 1e-9)


    search.approx <- modelsearch2(e.base, method.p.adjust = "fastmax", trace = 0,
                                  method.maxdist = "approximate")
    search.resampling <- modelsearch2(e.base, method.p.adjust = "fastmax", trace = 0,
                                      method.maxdist = "resampling")
    search.bootstrap <- modelsearch2(e.base, method.p.adjust = "fastmax", trace = 0,
                                      method.maxdist = "bootstrap")
    
    expect_equal(search.approx$sequenceTest[[1]][1,"statistic"],
                 search.holm$sequenceTest[[1]][1,"statistic"])
    expect_equal(search.approx$sequenceTest[[1]][1,"statistic"],
                 search.resampling$sequenceTest[[1]][1,"statistic"], tol = 1e-3)
    expect_equal(search.approx$sequenceTest[[1]][1,"statistic"],
                 search.bootstrap$sequenceTest[[1]][1,"statistic"], tol = 1e-3)

    expect_equal(round(search.resampling$sequenceTest[[1]][1,"p.value"],2),
                 0.24)
    expect_equal(round(search.bootstrap$sequenceTest[[1]][1,"p.value"],2),
                 0.24)

    
})


##----------------------------------------------------------------------
### test2-modelsearch2.R ends here

### sCorrect-skeleton.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov  8 2017 (10:35) 
## Version: 
## Last-Updated: jan 23 2024 (13:46) 
##           By: Brice Ozenne
##     Update #: 1699
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation - skeleton
#' @title Pre-computation for the Score
#' @description Pre-compute quantities that are necessary to compute the score of a lvm model.
#' @name skeleton
#' 
#' @param object a \code{lvm} object.
#' @param X [matrix] design matrix containing the covariates for each endogeneous and latent variable.
#' @param endogenous [character vector] the name of the endogeneous variables.
#' @param latent [character vector] the name of the latent variables.
#' 
#' @details
#' When the user specifies names for the coefficients (e.g. Y1[mu:sigma]) or uses constraints (Y1~beta*X1), \code{as.lava=FALSE} will use the names specified by the user (e.g. mu, sigma, beta)
#' while \code{as.lava=TRUE} will use the name of the first link defining the coefficient.
#'
#' @examples
#' \dontrun{
#' skeleton <- lavaSearch2::skeleton
#' skeleton.lvm <- lavaSearch2::skeleton.lvm
#' skeleton.lvmfit <- lavaSearch2::skeleton.lvmfit
#' 
#' ## without constrain
#' m <- lvm(Y1~X1+X2+eta,Y2~X3+eta,Y3~eta)
#' latent(m) <- ~eta
#' 
#' e <- estimate(m, lava::sim(m,1e2))
#' M.data <- as.matrix(model.frame(e))
#'
#' skeleton(e$model, as.lava = TRUE,
#'          name.endogenous = endogenous(e), n.endogenous = 3,
#'          name.latent = latent(e), 
#'          update.value = FALSE)
#' skeleton(e, data = M.data, p = pars(e), as.lava = TRUE,
#'          name.endogenous = endogenous(e), n.endogenous = 3,
#'          name.latent = latent(e), 
#'          update.value = TRUE)
#'
#' ## with constrains
#' m <- lvm(Y[mu:sigma] ~ beta*X1+X2)
#' e <- estimate(m, lava::sim(m,1e2))
#' M.data <- as.matrix(model.frame(e))
#'
#' skeleton(e$model, as.lava = TRUE,
#'          name.endogenous = "Y", n.endogenous = 1,
#'          name.latent = NULL, 
#'          update.value = FALSE)$skeleton
#' 
#' skeleton(e, data = M.data, p = pars(e), as.lava = FALSE,
#'          name.endogenous = "Y", n.endogenous = 1,
#'          name.latent = NULL, 
#'          update.value = FALSE)$skeleton
#' 
#'}
#' @concept small sample inference
#' @concept derivative of the score equation
#' @keywords internal


## * skeleton
#' @rdname skeleton
#' @export
skeleton <- function(object, X,
                     endogenous, latent,
                     n.cluster, index.Omega){
    if(lava.options()$debug){cat("skeleton \n")}

    n.endogenous <- length(endogenous)
    n.latent <- length(latent)
    n.obs <- NROW(X)
    obsByEndoInX <- tapply(1:n.obs,X$XXendogenousXX,list)

    ## ** extract table type
    type <- coefType(object, as.lava = FALSE)
    type.theta <- type[type$marginal == FALSE,,drop=FALSE]
    theta.value <- list()
    theta.param <- list()

    ## ** prepare value and param
    if("nu" %in% type.theta$detail){
        type.nu <- type.theta[type.theta$detail %in% "nu",,drop=FALSE]

        theta.value$nu <- stats::setNames(rep(NA, n.endogenous), endogenous)
        if(any(!is.na(type.nu$value))){
            theta.value$nu[type.nu[!is.na(type.nu$value),"Y"]] <- type.nu[!is.na(type.nu$value),"value"]
        }

        theta.param$nu <- stats::setNames(rep(as.character(NA), n.endogenous), endogenous)
        if(any(!is.na(type.nu$param))){
            theta.param$nu[type.nu[!is.na(type.nu$param),"Y"]] <- type.nu[!is.na(type.nu$param),"param"]
        }
        theta.param$Xnu <- matrix(NA,
                                  nrow = n.cluster, ncol = n.endogenous, byrow = TRUE,
                                  dimnames = list(NULL,endogenous))
        for(iC in 1:n.cluster){
            theta.param$Xnu[iC,index.Omega[[iC]]] <- 1
        }        
    }

    if("alpha" %in% type.theta$detail){
        type.alpha <- type.theta[type.theta$detail %in% "alpha",,drop=FALSE]
        theta.value$alpha <- stats::setNames(rep(NA, n.latent), latent)
        theta.param$alpha <- stats::setNames(rep(as.character(NA), n.latent), latent)
        theta.param$Xalpha <- rep(1, times = n.cluster)
        
        if(any(!is.na(type.alpha$value))){
            theta.value$alpha[type.alpha[!is.na(type.alpha$value),"Y"]] <- type.alpha[!is.na(type.alpha$value),"value"]
        }
        if(any(!is.na(type.alpha$param))){
            theta.param$alpha[type.alpha[!is.na(type.alpha$param),"Y"]] <- type.alpha[!is.na(type.alpha$param),"param"]
        }
    }

    if("K" %in% type.theta$detail){
        type.K <- type.theta[type.theta$detail %in% "K",,drop=FALSE]
        K.exogenous <- unique(type.K$X)
        
        theta.value$K <- matrix(0, nrow = length(K.exogenous), ncol = n.endogenous,
                                dimnames = list(K.exogenous, endogenous))
        theta.param$K <- matrix(as.character(NA), nrow = length(K.exogenous), ncol = n.endogenous,
                                dimnames = list(K.exogenous, endogenous))

        theta.param$XK <- lapply(endogenous, function(iE){ ## iE <- endogenous[1]
            iXK <- matrix(NA, nrow = n.cluster, ncol = length(K.exogenous),
                          dimnames = list(NULL, K.exogenous))
            iXK[X[obsByEndoInX[[iE]],"XXclusterXX"],] <- as.matrix(X[obsByEndoInX[[iE]],K.exogenous,drop=FALSE])
            return(iXK)
        })
        names(theta.param$XK) <- endogenous
        
        for(iK in 1:NROW(type.K)){ ## iK <- 1
            theta.value$K[type.K[iK,"X"],type.K[iK,"Y"]] <- type.K[iK,"value"]
            theta.param$K[type.K[iK,"X"],type.K[iK,"Y"]] <- type.K[iK,"param"]
        }
    }

    if("Gamma" %in% type.theta$detail){
        type.Gamma <- type.theta[type.theta$detail %in% "Gamma",,drop=FALSE]
        Gamma.exogenous <- unique(type.Gamma$X)

        theta.value$Gamma <- matrix(0, nrow = length(Gamma.exogenous), ncol = n.latent,
                                    dimnames = list(Gamma.exogenous, latent))
        theta.param$Gamma <- matrix(as.character(NA), nrow = length(Gamma.exogenous), ncol = n.latent,
                                    dimnames = list(Gamma.exogenous, latent))
        theta.param$XGamma <- lapply(obsByEndoInX[latent], function(iIndex){as.matrix(X[iIndex,Gamma.exogenous,drop=FALSE])})
        for(iGamma in 1:NROW(type.Gamma)){ ## iGamma <- 1
            theta.value$Gamma[type.Gamma[iGamma,"X"],type.Gamma[iGamma,"Y"]] <- type.Gamma[iGamma,"value"]
            theta.param$Gamma[type.Gamma[iGamma,"X"],type.Gamma[iGamma,"Y"]] <- type.Gamma[iGamma,"param"]
        }
    }

    if("Lambda" %in% type.theta$detail){
        type.Lambda <- type.theta[type.theta$detail %in% "Lambda",,drop=FALSE]
        
        theta.value$Lambda <- matrix(0, nrow = n.latent, ncol = n.endogenous,
                                     dimnames = list(latent, endogenous))
        theta.param$Lambda <- matrix(as.character(NA), nrow = n.latent, ncol = n.endogenous,
                                     dimnames = list(latent, endogenous))

        for(iLambda in 1:NROW(type.Lambda)){ ## iLambda <- 1
            theta.value$Lambda[type.Lambda[iLambda,"X"],type.Lambda[iLambda,"Y"]] <- type.Lambda[iLambda,"value"]
            theta.param$Lambda[type.Lambda[iLambda,"X"],type.Lambda[iLambda,"Y"]] <- type.Lambda[iLambda,"param"]
        }
    }

    if("B" %in% type.theta$detail){
        type.B <- type.theta[type.theta$detail %in% "B",,drop=FALSE]
        
        theta.value$B <- matrix(0, nrow = n.latent, ncol = n.latent,
                                     dimnames = list(latent, latent))
        theta.param$B <- matrix(as.character(NA), nrow = n.latent, ncol = n.latent,
                                     dimnames = list(latent, latent))

        for(iB in 1:NROW(type.B)){ ## iB <- 1
            theta.value$B[type.B[iB,"X"],type.B[iB,"Y"]] <- type.B[iB,"value"]
            theta.param$B[type.B[iB,"X"],type.B[iB,"Y"]] <- type.B[iB,"param"]
        }
    }

    if(any(c("Sigma_var", "Sigma_cov") %in% type.theta$detail)){
        type.Sigma_var <- type.theta[type.theta$detail %in% "Sigma_var",,drop=FALSE]
        type.Sigma_cov <- type.theta[type.theta$detail %in% "Sigma_cov",,drop=FALSE]

        theta.value$Sigma <- matrix(0, nrow = n.endogenous, ncol = n.endogenous,
                                     dimnames = list(endogenous, endogenous))
        theta.param$Sigma <- matrix(as.character(NA), nrow = n.endogenous, ncol = n.endogenous,
                                     dimnames = list(endogenous, endogenous))

        if(NROW(type.Sigma_var)>0){
            for(iSigma_var in 1:NROW(type.Sigma_var)){ ## iSigma_var <- 1
                theta.value$Sigma[type.Sigma_var[iSigma_var,"X"],type.Sigma_var[iSigma_var,"Y"]] <- type.Sigma_var[iSigma_var,"value"]
                theta.param$Sigma[type.Sigma_var[iSigma_var,"X"],type.Sigma_var[iSigma_var,"Y"]] <- type.Sigma_var[iSigma_var,"param"]
            }
        }
        if(NROW(type.Sigma_cov)>0){
            for(iSigma_cov in 1:NROW(type.Sigma_cov)){ ## iSigma_cov <- 1
                theta.value$Sigma[type.Sigma_cov[iSigma_cov,"X"],type.Sigma_cov[iSigma_cov,"Y"]] <- type.Sigma_cov[iSigma_cov,"value"]
                theta.value$Sigma[type.Sigma_cov[iSigma_cov,"Y"],type.Sigma_cov[iSigma_cov,"X"]] <- type.Sigma_cov[iSigma_cov,"value"]
                theta.param$Sigma[type.Sigma_cov[iSigma_cov,"X"],type.Sigma_cov[iSigma_cov,"Y"]] <- type.Sigma_cov[iSigma_cov,"param"]
                theta.param$Sigma[type.Sigma_cov[iSigma_cov,"Y"],type.Sigma_cov[iSigma_cov,"X"]] <- type.Sigma_cov[iSigma_cov,"param"]
            }
        }
    }
    
    if(any(c("Psi_var", "Psi_cov") %in% type.theta$detail)){
        type.Psi_var <- type.theta[type.theta$detail %in% "Psi_var",,drop=FALSE]
        type.Psi_cov <- type.theta[type.theta$detail %in% "Psi_cov",,drop=FALSE]
        
        theta.value$Psi <- matrix(0, nrow = n.latent, ncol = n.latent,
                                     dimnames = list(latent, latent))
        theta.param$Psi <- matrix(as.character(NA), nrow = n.latent, ncol = n.latent,
                                     dimnames = list(latent, latent))

        if(NROW(type.Psi_var)>0){
            for(iPsi_var in 1:NROW(type.Psi_var)){ ## iPsi_var <- 1
                theta.value$Psi[type.Psi_var[iPsi_var,"X"],type.Psi_var[iPsi_var,"Y"]] <- type.Psi_var[iPsi_var,"value"]
                theta.param$Psi[type.Psi_var[iPsi_var,"X"],type.Psi_var[iPsi_var,"Y"]] <- type.Psi_var[iPsi_var,"param"]
            }
        }
        if(NROW(type.Psi_cov)>0){
            for(iPsi_cov in 1:NROW(type.Psi_cov)){ ## iPsi_cov <- 1
                theta.value$Psi[type.Psi_cov[iPsi_cov,"X"],type.Psi_cov[iPsi_cov,"Y"]] <- type.Psi_cov[iPsi_cov,"value"]
                theta.value$Psi[type.Psi_cov[iPsi_cov,"Y"],type.Psi_cov[iPsi_cov,"X"]] <- type.Psi_cov[iPsi_cov,"value"]
                theta.param$Psi[type.Psi_cov[iPsi_cov,"X"],type.Psi_cov[iPsi_cov,"Y"]] <- type.Psi_cov[iPsi_cov,"param"]
                theta.param$Psi[type.Psi_cov[iPsi_cov,"Y"],type.Psi_cov[iPsi_cov,"X"]] <- type.Psi_cov[iPsi_cov,"param"]
            }
        }
    }

    ## ** original link
    type.originalLink <- type[!is.na(type$originalLink),,drop=FALSE]
    originalLink2param <- stats::setNames(type.originalLink$param,type.originalLink$originalLink)
    if(inherits(object,"lvm")){
        originalLink2param <- originalLink2param[coef(object)]
    }else if(inherits(object,"lvmfit")){
        originalLink2param <- originalLink2param[names(coef(object))]
    }
    
    ## ** type of parameters
    type.param <- type[!is.na(type$param),,drop=FALSE]
    type.mean <- c("nu","alpha","K","Gamma","Lambda","B")
    type.var <- c("Lambda","B","Sigma_var","Sigma_cov","Psi_var","Psi_cov")

    Uparam <- as.character(originalLink2param)
    Uparam.mean <- unique(type.param[type.param$detail %in% type.mean,"param"])
    Uparam.variance <- unique(type.param[type.param$detail %in% type.var,"param"])
    Uparam.hybrid <- intersect(Uparam.mean,Uparam.variance)
    
    ## ** to update
    toUpdate <- c("nu" = "nu" %in% type.param$detail,
                  "K" = "K" %in% type.param$detail,
                  "Lambda" = "Lambda" %in% type.param$detail,
                  "Sigma" = ("Sigma_cov" %in% type.param$detail) || ("Sigma_var" %in% type.param$detail),
                  "alpha" = "alpha" %in% type.param$detail,
                  "Gamma" = "Gamma" %in% type.param$detail,
                  "B" = "B" %in% type.param$detail,
                  "Psi" = ("Psi_cov" %in% type.param$detail) || ("Psi_var" %in% type.param$detail)
                  )    

    ## ** residuals
    theta.param$endogenous <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                                    dimnames = list(NULL, endogenous))
    for(iEndo in 1:length(endogenous)){ ## iEndo <- 1
        iIndexLong <- obsByEndoInX[[endogenous[iEndo]]]
        iIndexWide <- X[iIndexLong,"XXclusterXX"]
        theta.param$endogenous[iIndexWide,endogenous[iEndo]] <- X[iIndexLong,"XXvalueXX"]
    }

    ## ** export
    return(list(param = theta.param,
                value = theta.value,
                type = type,
                Uparam = Uparam,
                Uparam.mean = Uparam.mean,
                Uparam.variance = Uparam.variance,
                Uparam.hybrid = Uparam.hybrid,
                toUpdate.moment = toUpdate,
                originalLink2param = originalLink2param,
                obsByEndoInX = obsByEndoInX)
           )
}

## * skeletonDtheta
#' @rdname skeleton
skeletonDtheta <- function(object, X,
                           endogenous, latent,
                           missing.pattern, unique.pattern, name.pattern,
                           n.cluster, index.Omega){
    if(lava.options()$debug){cat("skeletonDtheta \n")}

    n.endogenous <- length(endogenous)
    n.latent <- length(latent)
    type <- object$type

    type <- type[!is.na(type$param),]
    name.param <- unique(type$param)
    n.param <- length(name.param)

    ## ** Compute partial derivative with respect to the matrices of parameters
    dOmega.dparam <- list() ## derivative regarding the variance
    dmat.dparam <- list() ## derivative regarding the matrix of parameters

    if("nu" %in% type$detail){
        type.nu <- type[type$detail == "nu",]
        Utype.nu <- unique(type.nu$param)
        nUtype.nu <- length(Utype.nu)

        dmat.dparam$nu <- stats::setNames(vector(mode = "list", length = nUtype.nu), Utype.nu)
        
        for(iNu in Utype.nu){ ## iNu <- Utype.nu[1]
            dnu.dparam <- as.numeric(endogenous %in% type.nu[type.nu$param==iNu,"Y"])

            dmat.dparam$nu[[iNu]] <- matrix(NA,
                                            nrow = n.cluster, ncol = n.endogenous, byrow = TRUE,
                                            dimnames = list(NULL,endogenous))
            for(iP in name.pattern){ ## iP <- name.pattern[1]
                iIndex <- missing.pattern[[iP]]
                iY <- which(unique.pattern[iP,]==1)
                dmat.dparam$nu[[iNu]][iIndex,iY] <- rep(1, times = length(iIndex)) %o% dnu.dparam[iY]
            }
        }
    }

    if("alpha" %in% type$detail){
        type.alpha <- type[type$detail == "alpha",]
        Utype.alpha <- unique(type.alpha$param)
        nUtype.alpha <- length(Utype.alpha)
        dmat.dparam$alpha <- stats::setNames(vector(mode = "list", length = nUtype.alpha), Utype.alpha)

        for(iAlpha in Utype.alpha){ ## iAlpha <- Utype.alpha[1]
            dmat.dparam$alpha[[iAlpha]] <- matrix(as.numeric(latent %in% type.alpha[type.alpha$param==iAlpha,"Y"]),
                                                  nrow = n.cluster, ncol = n.latent, byrow = TRUE,
                                                  dimnames = list(NULL,latent))
        }
    }

    if("K" %in% type$detail){
        type.K <- type[type$detail == "K",]
        Utype.K <- unique(type.K$param)
        nUtype.K <- length(Utype.K)

        for(iK in Utype.K){ ## iK <- Utype.K[1]
            iType.K <- type.K[type.K$param==iK,]
            dmat.dparam$K[[iK]] <- matrix(0,
                                          nrow = n.cluster, ncol = n.endogenous,
                                          dimnames = list(NULL,endogenous))

            for(iX in 1:NROW(iType.K)){ ## iY <- 5
                iY <- match(iType.K$Y[iX],endogenous)
                iIndex <- which(X$XXendogenousXX==endogenous[iY])
                dmat.dparam$K[[iK]][X[iIndex,"XXclusterXX"],iY] <- dmat.dparam$K[[iK]][X[iIndex,"XXclusterXX"],iY] + X[X$XXendogenousXX==endogenous[iY],iType.K$X[iX]]
            }
        }
    }

    if("Gamma" %in% type$detail){
        type.Gamma <- type[type$detail == "Gamma",]
        Utype.Gamma <- unique(type.Gamma$param)
        nUtype.Gamma <- length(Utype.Gamma)
        dmat.dparam$Gamma <- stats::setNames(vector(mode = "list", length = nUtype.Gamma), Utype.Gamma)

        for(iGamma in Utype.Gamma){ ## iGamma <- Utype.Gamma[1]
            iType.Gamma <- type.Gamma[type.Gamma$param==iGamma,]
            dmat.dparam$Gamma[[iGamma]] <- matrix(0,
                                                  nrow = n.cluster, ncol = n.latent,
                                                  dimnames = list(NULL,latent))

            for(iX in 1:NROW(iType.Gamma)){ ## iLatent <- 5
                iLatent <- match(iType.Gamma$Y[iX],latent)
                dmat.dparam$Gamma[[iGamma]][,iLatent] <- dmat.dparam$Gamma[[iGamma]][,iLatent] + X[X$XXendogenousXX==latent[iLatent],iType.Gamma$X[iX]]
            }
        }
    }

    if("Lambda" %in% type$detail){
        type.Lambda <- type[type$detail == "Lambda",]
        Utype.Lambda <- unique(type.Lambda$param)
        nUtype.Lambda <- length(Utype.Lambda)
        dmat.dparam$Lambda <- stats::setNames(vector(mode = "list", length = nUtype.Lambda), Utype.Lambda)

        for(iLambda in Utype.Lambda){ ## iLambda <- Utype.Lambda[1]
            dmat.dparam$Lambda[[iLambda]] <- matrix(as.double(object$param$Lambda %in% iLambda),
                                                    nrow = n.latent, ncol = n.endogenous,
                                                    dimnames = list(latent,endogenous))
        }
    }

    if("B" %in% type$detail){
        type.B <- type[type$detail == "B",]
        Utype.B <- unique(type.B$param)
        nUtype.B <- length(Utype.B)
        dmat.dparam$B <- stats::setNames(vector(mode = "list", length = nUtype.B), Utype.B)

        for(iB in Utype.B){ ## iB <- Utype.B[1]
            dmat.dparam$B[[iB]] <- matrix(as.double(object$param$B %in% iB),
                                          nrow = n.latent, ncol = n.latent,
                                          dimnames = list(latent,latent))
        }
    }

    if(any(c("Sigma_var","Sigma_cov") %in% type$detail)){
        type.Sigma <- type[type$detail %in% c("Sigma_var","Sigma_cov"),]
        Utype.Sigma <- unique(type.Sigma$param)
        nUtype.Sigma <- length(Utype.Sigma)
        dmat.dparam$Sigma <- stats::setNames(vector(mode = "list", length = nUtype.Sigma), Utype.Sigma)

        for(iSigma in Utype.Sigma){ ## iSigma <- Utype.Sigma[1]
            dmat.dparam$Sigma[[iSigma]] <- matrix(as.double(object$param$Sigma %in% iSigma),
                                                  nrow = n.endogenous, ncol = n.endogenous,
                                                  dimnames = list(endogenous,endogenous))
        }
    }

    if(any(c("Psi_var","Psi_cov") %in% type$detail)){
        type.Psi <- type[type$detail %in% c("Psi_var","Psi_cov"),]
        Utype.Psi <- unique(type.Psi$param)
        nUtype.Psi <- length(Utype.Psi)
        dmat.dparam$Psi <- stats::setNames(vector(mode = "list", length = nUtype.Psi), Utype.Psi)

        for(iPsi in Utype.Psi){ ## iB <- Utype.B[1]
            dmat.dparam$Psi[[iPsi]] <- matrix(as.double(object$param$Psi %in% iPsi),
                                              nrow = n.latent, ncol = n.latent,
                                              dimnames = list(latent,latent))
        }
    }

    ## ** Store derivative with respect to the mean/variance
    if(length(dmat.dparam$nu)+length(dmat.dparam$K) > 0){
        dmu.dparam <- c(dmat.dparam$nu, dmat.dparam$K)
    }else{
        dmu.dparam <- list()
    }

    if(length(dmat.dparam$Sigma) > 0){
        dOmega.dparam <- dmat.dparam$Sigma
    }else{
        dOmega.dparam <- list()
    }

    ## ** pairs of parameters to be considered
    grid.param <- list(mean = .combination(object$Uparam.mean, object$Uparam.mean),
                       var = .combination(object$Uparam.var, object$Uparam.var),
                       hybrid = .combination(object$Uparam.mean, object$Uparam.var))

    ## ** export
    return(c(object,
             list(
                 dmu.dparam = dmu.dparam,
                 dOmega.dparam = dOmega.dparam,
                 dmat.dparam = dmat.dparam,
                 grid.dmoment = grid.param
                 ))
           )
}

## * skeletonDtheta2
#' @rdname skeleton
skeletonDtheta2 <- function(object){
    if(lava.options()$debug){cat("skeletonDtheta2 \n")}

    type.param <- object$type[!is.na(object$type$param),,drop=FALSE]
    grid.param <- list()
    
    ## ** identify all combinations of coefficients with second derivative
    grid.param$mean <- list()

    grid.param$mean$alpha.B <- .combinationDF(type.param,
                                        detail1 = "alpha", name1 = "alpha",
                                        detail2 = "B", name2 = "B")

    grid.param$mean$alpha.Lambda <- .combinationDF(type.param,
                                                 detail1 = "alpha", name1 = "alpha",
                                                 detail2 = "Lambda", name2 = "Lambda")

    grid.param$mean$Gamma.B <- .combinationDF(type.param,
                                        detail1 = "Gamma", name1 = "Gamma",
                                        detail2 = "B", name2 = "B")

    grid.param$mean$Gamma.Lambda <- .combinationDF(type.param,
                                             detail1 = "Gamma", name1 = "Gamma",
                                             detail2 = "Lambda", name2 = "Lambda")
    
    grid.param$mean$Lambda.B <- .combinationDF(type.param,
                                        detail1 = "Lambda", name1 = "Lambda",
                                        detail2 = "B", name2 = "B")

    grid.param$mean$B.B <- .combinationDF(type.param,
                                    detail1 = "B", name1 = "B1",
                                    detail2 = "B", name2 = "B2")

    grid.param$var <- list()
    
    grid.param$var$Psi.Lambda <- .combinationDF(type.param,
                                           detail1 = c("Psi_var","Psi_cov"), name1 = "Psi",
                                           detail2 = "Lambda", name2 = "Lambda")

    grid.param$var$Psi.B <- .combinationDF(type.param,
                                      detail1 = c("Psi_var","Psi_cov"), name1 = "Psi",
                                      detail2 = "B", name2 = "B")

    grid.param$var$Lambda.B <- .combinationDF(type.param,
                                             detail1 = "Lambda", name1 = "Lambda",
                                             detail2 = "B", name2 = "B")

    grid.param$var$Lambda.Lambda <- .combinationDF(type.param,
                                              detail1 = "Lambda", name1 = "Lambda1",
                                              detail2 = "Lambda", name2 = "Lambda2")

    grid.param$var$B.B <- .combinationDF(type.param,
                                        detail1 = "B", name1 = "B1",
                                        detail2 = "B", name2 = "B2")

    
    grid.param$mean <- grid.param$mean[lengths(grid.param$mean)>0]
    grid.param$var <- grid.param$var[lengths(grid.param$var)>0]
    
    ## ** create d2mu and d2Omega
    if(length(grid.param$n.mean)>0){
        grid.tempo <- lapply(grid.param$mean, function(x){
            if(NROW(x)>0){
                colnames(x) <- c("x","y")
            }
            return(x)
        })
        collapseGrid <- do.call(rbind, grid.tempo)
        name.tempo <- as.character(unique(collapseGrid[[1]]))
        d2mu <- lapply(name.tempo, function(x){
            iIndex <- which(collapseGrid[[1]]==x)
            v <- vector(mode = "list", length(iIndex))
            names(v) <- collapseGrid[[2]][iIndex]
            return(v)
        })
        names(d2mu) <- name.tempo
    }else{
        d2mu <- list()
    }

    if(length(grid.param$var)>0){
        grid.tempo <- lapply(grid.param$var, function(x){
            if(NROW(x)>0){
                colnames(x) <- c("x","y")
            }
            return(x)
        })
        collapseGrid <- do.call(rbind, grid.tempo)
        name.tempo <- as.character(unique(collapseGrid[[1]]))
        d2Omega <- lapply(name.tempo, function(x){
            iIndex <- which(collapseGrid[[1]]==x)
            v <- vector(mode = "list", length(iIndex))
            names(v) <- collapseGrid[[2]][iIndex]
            return(v)
        })
        names(d2Omega) <- name.tempo
    }else{
        d2Omega <- list()
    }

    ## ** update grid.dmoment
    if(NROW(object$grid.dmoment$mean)>0){
        object$grid.dmoment$mean$d2.12 <- FALSE
        object$grid.dmoment$mean$d2.21 <- FALSE
    }
    
    if(length(grid.param$mean)>0){
        for(iType in names(grid.param$mean)){  ## iType <- names(grid.param$mean)[1]
            for(iRow in 1:NROW(grid.param$mean[[iType]])){ ## iRow <- 1
                iName1 <- grid.param$mean[[iType]][iRow,1]
                iName2 <- grid.param$mean[[iType]][iRow,2]
                iIndex1 <- which((object$grid.dmoment$mean$Var1==iName1)*(object$grid.dmoment$mean$Var2==iName2)==1)
                iIndex2 <- which((object$grid.dmoment$mean$Var1==iName2)*(object$grid.dmoment$mean$Var2==iName1)==1)
                if(length(iIndex1)>0){
                    object$grid.dmoment$mean[iIndex1,"d2.12"] <- TRUE
                }else if(length(iIndex2)>0){
                    object$grid.dmoment$mean[iIndex2,"d2.21"] <- TRUE
                }
            }
        }
    }

    if(NROW(object$grid.dmoment$var)>0){
        object$grid.dmoment$var$d2.12 <- FALSE
        object$grid.dmoment$var$d2.21 <- FALSE
    }
    if(length(grid.param$var)>0){
        for(iType in names(grid.param$var)){  ## iType <- names(grid.param$var)[1]
            for(iRow in 1:NROW(grid.param$var[[iType]])){ ## iRow <- 1
                iName1 <- grid.param$var[[iType]][iRow,1]
                iName2 <- grid.param$var[[iType]][iRow,2]
                iIndex1 <- which((object$grid.dmoment$var$Var1==iName1)*(object$grid.dmoment$var$Var2==iName2)==1)
                iIndex2 <- which((object$grid.dmoment$var$Var1==iName2)*(object$grid.dmoment$var$Var2==iName1)==1)
                if(length(iIndex1)>0){
                    object$grid.dmoment$var[iIndex1,"d2.12"] <- TRUE
                }else if(length(iIndex2)>0){
                    object$grid.dmoment$var[iIndex2,"d2.21"] <- TRUE
                }
            }
        }
    }

    ## ** Parameters in dInformation/JJK
    grid.3varD1 <- .duplicatedGrid(expand.grid(X = object$Uparam.var, Y = object$Uparam.var, Z = object$Uparam.var, stringsAsFactors = FALSE))
    
    grid.2meanD1.1varD1 <- .duplicatedGrid(expand.grid(X = object$Uparam.mean, Y = object$Uparam.mean, Z = object$Uparam.var, stringsAsFactors = FALSE))

    grid.2meanD2.1meanD1 <- .expand.grid2(X = object$Uparam.mean, Y = object$Uparam.mean, Z = object$Uparam.mean,
                                          d2 = object$grid.dmoment$mean)
    grid.2varD2.1varD1 <- .expand.grid2(X = object$Uparam.var, Y = object$Uparam.var, Z = object$Uparam.var,
                                        d2 = object$grid.dmoment$var)

    ## ** Export    
    return(c(object,
             list(d2mu.dparam = d2mu,
                  d2Omega.dparam = d2Omega,
                  grid.d2moment = grid.param,
                  grid.3varD1 = grid.3varD1,
                  grid.2meanD1.1varD1 = grid.2meanD1.1varD1,
                  grid.2meanD2.1meanD1 = grid.2meanD2.1meanD1,
                  grid.2varD2.1varD1 = grid.2varD2.1varD1))
           )
}



## * helpers
## ** .combinationDF
.combinationDF <- function(data,
                           detail1, detail2,
                           name1, name2){

    detail <- NULL # [:for CRAN check] subset
    
    if(any(detail1 %in% data$detail) && any(detail2 %in% data$detail) ){
        ls.args <- list(subset(data, subset = detail %in% detail1, select = "param", drop = TRUE),
                        subset(data, subset = detail %in% detail2, select = "param", drop = TRUE))
        names(ls.args) <- c(name1,name2)
    
        return(do.call(.combination, args = ls.args))
        
    }else{
        
        return(numeric(0))
        
    }
}

## ** .expand.grid2
.expand.grid2 <- function(X,Y,Z,d2){
    if(NROW(d2)==0 || sum(d2$d2.12+d2$d2.21)==0){return(NULL)}

    ## find available derivatives
    index.12 <- which(d2[,"d2.12"])
    index.21 <- which(d2[,"d2.21"])

    available.d2 <- rbind(data.frame(Var1 = d2[index.12,1], Var2 = d2[index.12,2], stringsAsFactors = FALSE),
                          data.frame(Var1 = d2[index.21,2], Var2 = d2[index.21,1], stringsAsFactors = FALSE))
    name.available.d2 <- as.character(interaction(available.d2))
    n.available.d2 <- length(name.available.d2)

    unavailable.d2 <- data.frame(Var1 = available.d2[,2], Var2 = available.d2[,1], stringsAsFactors = FALSE)
    name2position <- rbind(data.frame(name = name.available.d2, position = 1:n.available.d2, stringsAsFactors = FALSE),
                           data.frame(name = as.character(interaction(unavailable.d2)), position = 1:n.available.d2, stringsAsFactors = FALSE))
    unavailable.d2 <- unavailable.d2[as.character(interaction(unavailable.d2)) %in% name.available.d2 == FALSE,,drop=FALSE]
    
    ## generate grid (combination between X and YZ)
    grid <- expand.grid(X = X, Y = Y, Z = Z, stringsAsFactors = FALSE)

    ## find second order derivative
    name.gridXY <- as.character(interaction(grid[,c("X","Y")]))
    d2XY <- available.d2[name2position$position[match(name.gridXY, name2position$name)],]
    names(d2XY) <- c("d2XY.Var1","d2XY.Var2")
        
    name.gridXZ <- as.character(interaction(grid[,c("X","Z")]))
    d2XZ <- available.d2[name2position$position[match(name.gridXZ, name2position$name)],]
    names(d2XZ) <- c("d2XZ.Var1","d2XZ.Var2")

    name.gridYZ <- as.character(interaction(grid[,c("Y","Z")]))
    d2YZ <- available.d2[name2position$position[match(name.gridYZ, name2position$name)],]
    names(d2YZ) <- c("d2YZ.Var1","d2YZ.Var2")
    
    grid <- cbind(grid,d2XY,d2XZ,d2YZ)
    rownames(grid) <- NULL
    
    ## find duplicated
    grid <- cbind(.duplicatedGrid(grid[,c("X","Y","Z")]), grid[,-(1:3)])
    grid$d2XY <- !is.na(grid$d2XY.Var1)
    grid$d2XZ <- !is.na(grid$d2XZ.Var1)
    grid$d2YZ <- !is.na(grid$d2YZ.Var1)
    return(grid[grid$d2XY+grid$d2XZ+grid$d2YZ>0,,drop=FALSE])
}

## ** .duplicatedGrid
.duplicatedGrid <- function(grid){
    if(NROW(grid)==0){
        return(NULL)
    }else if(NROW(grid)==1){
        grid$duplicatedXY <- FALSE
        grid$duplicatedXZ <- FALSE
        grid$duplicatedYZ <- FALSE
        grid$duplicatedXYZ <- FALSE
        return(grid)
    }
    grid <- as.data.frame(grid)
    
    ## normalize grid
    labels <- unique(unlist(grid))
    n.labels <- length(labels)
    grid.num <- cbind(apply(grid, 2, match, labels), index =  1:NROW(grid))
    p <- NCOL(grid)
    
    ## warper
    warper <- function(M, M.max, p){
        newlevel <- rep(NA, M.max)
        newlevel[1] <- 1
        for(iX in 2:M.max){
            newlevel[iX] <- p * newlevel[iX-1] + 1
        }
        M[] <- newlevel[unlist(M)]
        return(duplicated(rowSums(M)))
    }
    
    ## find duplicated
    duplicatedXY <- do.call(rbind,by(grid.num[,c(1:2,4),drop=FALSE], INDICES = grid.num[,3], FUN = function(iGrid){
        return(cbind(warper(iGrid[,1:2], M.max = n.labels, p = p),iGrid[,3]))
    }))
    grid$duplicatedXY[duplicatedXY[,2]] <- as.logical(duplicatedXY[,1])

    duplicatedXZ <- do.call(rbind,by(grid.num[,c(1,3,4),drop=FALSE], INDICES = grid.num[,2], FUN = function(iGrid){
        return(cbind(warper(iGrid[,1:2], M.max = n.labels, p = p),iGrid[,3]))
    }))
    grid$duplicatedXZ[duplicatedXZ[,2]] <- as.logical(duplicatedXZ[,1])

    duplicatedYZ <- do.call(rbind,by(grid.num[,c(2:4),drop=FALSE], INDICES = grid.num[,1], FUN = function(iGrid){
        return(cbind(warper(iGrid[,1:2], M.max = n.labels, p = p),iGrid[,3]))
    }))
    grid$duplicatedYZ[duplicatedYZ[,2]] <- as.logical(duplicatedYZ[,1])

    
    grid$duplicatedXYZ <- warper(grid.num[,1:3,drop=FALSE], M.max = n.labels, p = p)

    return(grid)

}

### sCorrect-getGroups2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 18 2019 (10:58) 
## Version: 
## Last-Updated: Jan 17 2022 (18:43) 
##           By: Brice Ozenne
##     Update #: 170
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Reconstruct the Cluster variable
#' @description Reconstruct the cluster variable.
#' Similar to \code{nlme::getGroups}.
#' @noRd
#'
#' @param object a \code{lvmfit} object.
#' @param data dataset.
#' @param index.Omega  [list] for each cluster, the position of the observed endogenous variables (i.e. how to subset the residual variance-covariance matrix).
#' @param endogenous [character vector] name of the endogenous variables.
#' @param ... [internal] Only used by the generic method.
#'  
#' @return A list containing:
#' \itemize{
#' \item index.cluster: the cluster index for each observation.
#' \item name.cluster: a unique identifier for each cluster.
#' \item n.cluster: the number of clusters.
#' }
#' 
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' dW <- sampleRepeated(10, format = "wide")
#' set.seed(10)
#' dL <- sampleRepeated(10, format = "long")
#' dL$time2 <- paste0("visit",dL$time)
#' 
#' #### latent variable model ####
#' e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
#' .getGroups2(e.lvm, data = dW)
#' 
#' @concept extractor
#' @keywords internal
`.getGroups2` <-
    function(object, data, index.Omega, endogenous) UseMethod(".getGroups2")

## * .getGroups2.lvm
.getGroups2.lvm <- function(object, data = NULL, index.Omega = NULL, endogenous = NULL){
    if(is.null(data)){
        data <- extractData(object)
    }
    if(is.null(index.Omega)){
        index.Omega <- .getIndexOmega(object, data = data)
    }
    if(is.null(endogenous)){
        endogenous <- lava::endogenous(object)
    }
    n.endogenous <- length(endogenous)

    ## ** find clusters
    n.cluster <- NROW(data)
    name.cluster <- 1:n.cluster
    missing <- any(is.na(index.Omega))
    index.cluster <- unlist(lapply(name.cluster, rep, times = n.endogenous))

    index.Omega <- tapply(index.Omega, index.cluster, function(iVec){list(stats::na.omit(iVec))})
    Uindex.Omega <- unique(index.Omega[sapply(index.Omega,length)>0])
    return(list(index.cluster = index.cluster,
                name.cluster = name.cluster,
                n.cluster = n.cluster,
                index.Omega = index.Omega,
                index2endogenous = stats::setNames(as.list(Uindex.Omega),Uindex.Omega)
                ))
    
}

## * .getGroups2.lvmfit
.getGroups2.lvmfit <- .getGroups2.lvm








######################################################################
### sCorrect-getGroups2.R ends here

### sCorrect-updateMoment.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: dec 10 2019 (09:58) 
## Version: 
## Last-Updated: jan 17 2022 (14:57) 
##           By: Brice Ozenne
##     Update #: 232
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * updateMoment
updateMoment <- function(skeleton, value, toUpdate,
                         name.pattern, unique.pattern,
                         param, Omega, endogenous, latent, n.cluster){
    if(lava.options()$debug){cat("updateMoment \n")}
    ## remove attributes used by moments2
    attr(Omega,"Omega.residuals") <- NULL
    attr(Omega,"Omega.leverage") <- NULL
    attr(Omega,"dOmega.leverage") <- NULL

    n.endogenous <- length(endogenous)
    n.latent <- length(latent)
    
    ## ** Update with the current values
    name.update  <- names(toUpdate[toUpdate==TRUE])
    if(length(name.update)>0){
        for(iUpdate in name.update){ ## iUpdate <- "Sigma"
            if(iUpdate == "SigmaValue"){
                index.update <- which(!is.na(skeleton$SigmaParam))
                skeleton$SigmaValue[index.update] <- param[skeleton$SigmaParam[index.update]]
                value$Sigma <- apply(skeleton$SigmaValue, MARGIN = 1:2, FUN = prod)
                attr(value$Sigma,"detail") <- skeleton$SigmaValue
            }else{
                index.update <- which(!is.na(skeleton[[iUpdate]]))
                value[[iUpdate]][index.update] <- param[skeleton[[iUpdate]][index.update]]
            }
        }
    }

    ## ** Pre-compute relevant quantities
    if(n.latent>0){
        ## alpha + X\Gamma
        value$alpha.XGamma <- matrix(0, nrow = n.cluster, ncol = n.latent,
                                     dimnames = list(NULL, latent))
        if("alpha" %in% names(value)){
            value$alpha.XGamma <- value$alpha.XGamma + skeleton$Xalpha %o% value$alpha
        }
        if("Gamma" %in% names(value)){
            value$alpha.XGamma <- value$alpha.XGamma + do.call(cbind,lapply(latent, function(iL){skeleton$XGamma[[iL]] %*% value$Gamma[,iL]}))
        }

        ## (I-B)^{-1}
        if("B" %in% names(value)){
            value$iIB <- solve(diag(1, nrow = n.latent, ncol = n.latent) - value$B)
        }else{
            value$iIB <- diag(1, nrow = n.latent, ncol = n.latent)
            dimnames(value$iIB) <- list(latent,latent)
        }

        ## (alpha + X\Gamma) (I-B)^{-1}
        value$alpha.XGamma.iIB <- value$alpha.XGamma %*% value$iIB
        
        ## (I-B)^{-1} \Lambda
        value$iIB.Lambda <-  value$iIB %*% value$Lambda
        value$tLambda.tiIB <-  t(value$iIB.Lambda)

        ## \Psi (I-B)^{-1}
        value$Psi.iIB <- value$Psi %*% value$iIB

        ## (I-B)^{-t} \Psi (I-B)^{-1}
        value$tiIB.Psi.iIB <-  t(value$iIB) %*% value$Psi

        ## \Lambda^t (I-B)^{-t} \Psi (I-B)^{-1}
        value$tLambda.tiIB.Psi.iIB <- t(value$iIB.Lambda) %*% value$Psi.iIB

        ## (I-B)^{-t} \Psi (I-B)^{-1} \Lambda
        value$tiIB.Psi.iIB.Lambda <- t(value$tLambda.tiIB.Psi.iIB) 
    }

    ## ** Compute mean
    value$mu <- matrix(0, nrow = n.cluster, ncol = n.endogenous,
                       dimnames = list(NULL,endogenous))

    if("nu" %in% names(value)){
        value$mu <- value$mu + sweep(skeleton$Xnu, MARGIN = 2, FUN = "*", STATS = value$nu)
    }
    if("K" %in% names(value)){
        value$mu <- value$mu + do.call(cbind,lapply(endogenous, function(iE){skeleton$XK[[iE]] %*% value$K[,iE]})) ## iE <- endogenous[1]
    }
    if(n.latent>0){
        value$mu <- value$mu + value$alpha.XGamma %*% value$iIB.Lambda
    }

    ## ** Compute variance
    Omega.param <- matrix(0, nrow = n.endogenous, ncol = n.endogenous, 
                          dimnames = list(endogenous,endogenous))
    if("Sigma" %in% names(value)){
        Omega.param <- Omega.param + value$Sigma
    }
    if("Psi" %in% names(value)){
        Omega.param <- Omega.param + value$tLambda.tiIB.Psi.iIB %*% value$Lambda
    }

    if(!is.null(Omega)){
        value$Omega <- Omega        
        attr(value$Omega,"discrepancy") <- Omega-Omega.param
    }else{
        value$Omega <- Omega.param
    }

    value$Omega.missing.pattern <- lapply(1:length(name.pattern), function(iM){ ## iM <- 1
        iIndex <- which(unique.pattern[iM,]==1)
        return(value$Omega[iIndex,iIndex,drop=FALSE])
    })
    names(value$Omega.missing.pattern) <- name.pattern
    value$OmegaM1.missing.pattern <- lapply(value$Omega.missing.pattern, solve)

    ## ** Export
    return(value)
}

## * updateDMoment
updateDMoment <- function(moment, skeleton, param){
    if(lava.options()$debug){cat("updateDMoment \n")}

    ## ** import information
    dmu <- skeleton$dmu.dparam
    dOmega <- skeleton$dOmega.dparam

    iIB.Lambda <- moment$iIB.Lambda
    tLambda.tiIB <- moment$tLambda.tiIB
    alpha.XGamma.iIB <- moment$alpha.XGamma.iIB
    tiIB.Psi.iIB.Lambda <- moment$tiIB.Psi.iIB.Lambda
    tLambda.tiIB.Psi.iIB <- moment$tLambda.tiIB.Psi.iIB
    Sigma <- moment$Sigma
    attr(Sigma,"detail") <- NULL

    ## ** Compute partial derivative regarding the mean
    ## NOTE: no "nu", "K", or "Gamma" as the partial derivative is independent of the parameter values
    ##       and can therefore be computed once for all
    
    if("alpha" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$alpha)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dmu[[iParam]] <- skeleton$dmat.dparam$alpha[[iParam]] %*% iIB.Lambda
        }
    }

    if("Gamma" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$Gamma)   
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dmu[[iParam]] <- skeleton$dmat.dparam$Gamma[[iParam]] %*% iIB.Lambda
        }
    }

    if("Lambda" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$Lambda)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dmu[[iParam]] <- alpha.XGamma.iIB %*% skeleton$dmat.dparam$Lambda[[iParam]]
        }
    }

    if("B" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$B)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dmu[[iParam]] <- alpha.XGamma.iIB %*% skeleton$dmat.dparam$B[[iParam]] %*% iIB.Lambda
        }
    }
    
    ## ** Compute partial derivative regarding the variance
    if("Lambda" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$Lambda)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dOmega[[iParam]] <- t(skeleton$dmat.dparam$Lambda[[iParam]]) %*% tiIB.Psi.iIB.Lambda + tLambda.tiIB.Psi.iIB %*% skeleton$dmat.dparam$Lambda[[iParam]]
        }
    }

    if("B" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$B)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dOmega[[iParam]] <- t(iIB.Lambda) %*% t(skeleton$dmat.dparam$B[[iParam]]) %*% tiIB.Psi.iIB.Lambda + tLambda.tiIB.Psi.iIB %*% skeleton$dmat.dparam$B[[iParam]] %*% iIB.Lambda
        }
    }

    if("Psi" %in% names(skeleton$dmat.dparam)){
        iName.param <- names(skeleton$dmat.dparam$Psi)
        for(iParam in iName.param){ ## iParam <- iName.param[1]
            dOmega[[iParam]] <- tLambda.tiIB %*% skeleton$dmat.dparam$Psi[[iParam]] %*% iIB.Lambda
        }
    }

   ## *** Export
    return(list(dmu = dmu, dOmega = dOmega))

}


## * updateD2Moment
updateD2Moment <- function(moment, skeleton, param){
    if(lava.options()$debug){cat("updateD2Moment \n")}

    ## ** Import information
    d2mu <- skeleton$d2mu.dparam
    d2Omega <- skeleton$d2Omega.dparam

    dalpha <- skeleton$dmat.dparam$alpha
    dLambda <- skeleton$dmat.dparam$Lambda
    dGamma <- skeleton$dmat.dparam$Gamma
    dB <- skeleton$dmat.dparam$B
    dPsi <- skeleton$dmat.dparam$Psi

    Psi <- moment$Psi
    Lambda <- moment$Lambda
    iIB <- moment$iIB
    Psi.iIB <- moment$Psi.iIB
    iIB.Lambda <- moment$iIB.Lambda
    tLambda.tiIB <- moment$tLambda.tiIB
    alpha.XGamma.iIB <- moment$alpha.XGamma.iIB
    tiIB.Psi.iIB <- moment$tiIB.Psi.iIB
    tiIB.Psi.iIB.Lambda <- moment$tiIB.Psi.iIB.Lambda
    tLambda.tiIB.Psi.iIB <- moment$tLambda.tiIB.Psi.iIB
    
    grid.mean <- skeleton$grid.d2moment$mean
    grid.var <- skeleton$grid.d2moment$var
    names.grid.mean <- names(grid.mean)
    names.grid.var <- names(grid.var)
    
    ## ** Compute partial derivative regarding the mean
    ## NOTE: no "nu", "K", or "Gamma" as the partial derivative is 0

    if("alpha.B" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$alpha.B)){ # iP <- 1
            iName1 <- grid.mean$alpha.B[iP,"alpha"]
            iName2 <- grid.mean$alpha.B[iP,"B"]
            d2mu[[iName1]][[iName2]] <- dalpha[[iName1]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
        }
    }
    
    if("alpha.Lambda" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$alpha.Lambda)){ # iP <- 1
            iName1 <- grid.mean$alpha.Lambda[iP,"alpha"]
            iName2 <- grid.mean$alpha.Lambda[iP,"Lambda"]
            d2mu[[iName1]][[iName2]] <- dalpha[[iName1]] %*% iIB %*% dLambda[[iName2]]
        }
    }

    if("Gamma.B" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$Gamma.B)){ # iP <- 1
                iName1 <- grid.mean$Gamma.B[iP,"Gamma"]
                iName2 <- grid.mean$Gamma.B[iP,"B"]
                d2mu[[iName1]][[iName2]] <- dGamma[[iName1]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
        }
    }

    if("Gamma.Lambda" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$Gamma.Lambda)){ # iP <- 1
                iName1 <- grid.mean$Gamma.Lambda[iP,"Gamma"]
                iName2 <- grid.mean$Gamma.Lambda[iP,"Lambda"]                
                d2mu[[iName1]][[iName2]] <- dGamma[[iName1]] %*% iIB %*% dLambda[[iName2]]
        }
    }

    if("Lambda.B" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$Lambda.B)){ # iP <- 1
                iName1 <- grid.mean$Lambda.B[iP,"Lambda"]
                iName2 <- grid.mean$Lambda.B[iP,"B"]
                d2mu[[iName1]][[iName2]] <- alpha.XGamma.iIB %*% dB[[iName2]] %*% iIB %*% dLambda[[iName1]]
        }
    }

    if("B.B" %in% names.grid.mean){
        for(iP in 1:NROW(grid.mean$B.B)){ # iP <- 1
                iName1 <- grid.mean$B.B[iP,"B1"]
                iName2 <- grid.mean$B.B[iP,"B2"]

                term1 <- alpha.XGamma.iIB %*% dB[[iName2]] %*% iIB %*% dB[[iName1]] %*% iIB.Lambda
                term2 <- alpha.XGamma.iIB %*% dB[[iName1]] %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
                d2mu[[iName1]][[iName2]] <- term1 + term2
        }
    }

    ## ** Compute partial derivative regarding the variance
    if("Psi.Lambda" %in% names.grid.var){
        for(iP in 1:NROW(grid.var$Psi.Lambda)){ # iP <- 1
            iName1 <- grid.var$Psi.Lambda[iP,"Psi"]
            iName2 <- grid.var$Psi.Lambda[iP,"Lambda"]

            term1 <- t(dLambda[[iName2]]) %*% t(iIB) %*% dPsi[[iName1]] %*% iIB.Lambda                
            d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
        }
    }
    
    if("Psi.B" %in% names.grid.var){
        for(iP in 1:NROW(grid.var$Psi.B)){ # iP <- 1
                iName1 <- grid.var$Psi.B[iP,"Psi"]
                iName2 <- grid.var$Psi.B[iP,"B"]

                term1 <- t(iIB.Lambda) %*% t(dB[[iName2]]) %*% t(iIB) %*% dPsi[[iName1]] %*% iIB.Lambda
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
        }
    }
    
    if("Lambda.B" %in% names.grid.var){
        for(iP in 1:NROW(grid.var$Lambda.B)){ # iP <- 1
                iName1 <- grid.var$Lambda.B[iP,"Lambda"]
                iName2 <- grid.var$Lambda.B[iP,"B"]

                term1 <- t(dLambda[[iName1]]) %*% t(iIB) %*% t(dB[[iName2]]) %*% t(iIB) %*% Psi %*% iIB.Lambda
                term2 <- t(dLambda[[iName1]]) %*% t(iIB) %*% Psi %*% iIB %*% dB[[iName2]] %*% iIB.Lambda
                ## term2 <- tLambda.tiIB.Psi.iIB %*% dB[[iName2]] %*% iIB %*% dLambda[[iName1]]                
                d2Omega[[iName1]][[iName2]] <- term1 + t(term1) + term2 + t(term2)
        }
    }

    if("Lambda.Lambda" %in% names.grid.var){
        for(iP in 1:NROW(grid.var$Lambda.Lambda)){ # iP <- 1
            iName1 <- grid.var$Lambda.Lambda[iP,"Lambda1"]
            iName2 <- grid.var$Lambda.Lambda[iP,"Lambda2"]
                
            term1 <- t(dLambda[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% dLambda[[iName2]]
            d2Omega[[iName1]][[iName2]] <- term1 + t(term1)
        }
    }

    if("B.B" %in% names.grid.var){
        for(iP in 1:NROW(grid.var$B.B)){ # iP <- 1
            iName1 <- grid.var$B.B[iP,"B1"]
            iName2 <- grid.var$B.B[iP,"B2"]

            term1 <- t(iIB.Lambda) %*% t(dB[[iName2]]) %*% t(iIB) %*% t(dB[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% Lambda
            term2 <- t(iIB.Lambda) %*% t(dB[[iName1]]) %*% t(iIB) %*% t(dB[[iName2]]) %*% t(iIB) %*% Psi.iIB %*% Lambda
            term3 <- t(iIB.Lambda) %*% t(dB[[iName1]]) %*% t(iIB) %*% Psi.iIB %*% dB[[iName2]] %*% iIB %*% Lambda
            d2Omega[[iName1]][[iName2]] <- term1 + t(term1) + term2 + t(term2) + term3 + t(term3)
        }
    }

    ## ** Export
    return(list(d2mu = d2mu, d2Omega = d2Omega))

}



######################################################################
### sCorrect-updateMoment.R ends here

### sCorrect-nobs2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: Jan  4 2022 (14:37) 
## Version: 
## Last-Updated: jan 23 2024 (10:26) 
##           By: Brice Ozenne
##     Update #: 24
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * documentation - nobs2
#' @title Effective Sample Size.
#' @description Extract the effective sample size, i.e. sample size minus the loss in degrees of freedom caused by the estimation of the parameters.
#' @name nobs2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object.
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the leverage.
#' 
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#' 
#' @return Numeric vector of length the number of endogenous variables.
#' 
#' @concept extractor
#' @keywords smallSampleCorrection
#' 
#' @export
`nobs2` <-
    function(object, ssc, ...) UseMethod("nobs2")

## * nobs2.lvmfit
#' @rdname nobs2
#' @export
nobs2.lvmfit <- function(object, ssc = lava.options()$ssc, ...){

    return(nobs2(estimate2(object, ssc = ssc, ...)))

}

## * nobs2.lvmfit2
#' @rdname nobs2
#' @export
nobs2.lvmfit2 <- function(object, ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    return(stats::nobs(object) - colSums(leverage2(object, format = "wide")))
}

##----------------------------------------------------------------------
### sCorrect-nobs2.R ends here

### sCorrect-sscResiduals.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb 16 2018 (16:38) 
## Version: 
## Last-Updated: jan 17 2022 (14:06) 
##           By: Brice Ozenne
##     Update #: 1226
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * .init_sscResiduals
.init_sscResiduals <- function(object){

    out <- list()

    ## ** extract info
    endogenous <- object$sCorrect$endogenous
    n.endogenous <- length(endogenous)
    latent <- object$sCorrect$latent
    n.latent <- length(latent)
    
    type <- object$sCorrect$skeleton$type

    index.var <- which(type$detail %in% c("Sigma_var","Sigma_cov","Psi_var","Psi_cov"))
    index.param <- which(!is.na(type$param))
    type.param <- type[intersect(index.var,index.param),,drop=FALSE]
    
    
    Omega <- object$sCorrect$moment$Omega
    ## name.var <- object$sCorrect$name.param[object$sCorrect$name.param %in% unique(type.param$param)] ## make sure to keep the same order as in the original vector of parameters
    name.var <- unique(type.param$param)
    n.var <- length(name.var)

    ## ** subset residual variance-covariance
    index.upper.tri <- data.frame(index = which(upper.tri(Omega, diag = TRUE)),
                                  which(upper.tri(Omega, diag = TRUE), arr.ind = TRUE)
                                  )
    name.rhs <- paste(endogenous[index.upper.tri[,"row"]],
                      lava.options()$symbols[2],
                      endogenous[index.upper.tri[,"col"]],
                      sep = "")
    n.rhs <- length(name.rhs)

    ## ** fixed part of the variance-covariance matrix
    index.value <- which(!is.na(type$value))
    index.var2 <- which(type$detail %in% c("Lambda","B","Sigma_var","Sigma_cov","Psi_var","Psi_cov"))
    type.var.constrain <- type[intersect(index.value, index.var2),,drop=FALSE]

    Omega.constrain <- matrix(0, nrow = n.endogenous, ncol = n.endogenous, 
                              dimnames = list(endogenous,endogenous))

    if(NROW(type.var.constrain)>0){
        ## extract parameters with value fixed by the user
        value <- object$sCorrect$skeleton$value

        ## add fixed Sigma value
        if(any(c("Sigma_var","Sigma_cov") %in% type.var.constrain$detail)){
            addSigma <- value$Sigma
            addSigma[is.na(addSigma)] <- 0
            Omega.constrain <- Omega.constrain + addSigma
        }
        if(any(c("Psi_var","Psi_cov") %in% type.var.constrain$detail)){
            Psi.constrain <- value$Psi
            Lambda.constrain <- value$Lambda
            if(any(is.na(value$B))){
                stop("Current implementation cannot handle constraints in Psi when there are B parameters.\n")
            }
            if(any(is.na(value$Lambda))){
                stop("Current implementation cannot handle constraints in Psi when there are lambda parameters.\n")
            }
            if("B" %in% names(value)){
                iIB.constrain <- solve(diag(1, nrow = n.latent, ncol = n.latent) - value$B)
            }else{
                iIB.constrain <- diag(1, nrow = n.latent, ncol = n.latent)
            }

            addPsi <- t(Lambda.constrain) %*% t(iIB.constrain) %*% Psi.constrain %*% iIB.constrain %*% Lambda.constrain
            addPsi[is.na(addPsi)] <- 0
            Omega.constrain <- Omega.constrain + addPsi
        }

    }

    ## ** Design matrix
    A <- matrix(0, nrow = n.rhs, ncol = n.var,
                dimnames = list(name.rhs, name.var))

    ## *** Sigma_var and Sigma_cov
    if(any(type.param$detail %in% c("Sigma_var","Sigma_cov"))){
        type.Sigma <- type.param[type.param$detail %in% c("Sigma_var","Sigma_cov"),,drop=FALSE]
        for(iRow in 1:NROW(type.Sigma)){ ## iRow <- 1 
            A[paste0(type.Sigma[iRow,"Y"],"~~",type.Sigma[iRow,"X"]),type.Sigma[iRow,"param"]] <- 1
        }
    }
    attr(A,"name") <- name.var

    ## *** Psi_var and Psi_cov
    if(any(type.param$detail %in% c("Psi_var","Psi_cov"))){
        type.Psi <- type.param[type.param$detail %in% c("Psi_var","Psi_cov"),,drop=FALSE]
        index.Psi <- cbind(row = match(type.Psi$X, latent),
                           col = match(type.Psi$Y, latent))
        rownames(index.Psi) <- type.Psi$param
    }else{
        index.Psi <- NULL
    }
    
    return(list(type = "residuals",
                param0 = object$sCorrect$param,
                Omega0 = object$sCorrect$moment$Omega,
                residuals0 = object$sCorrect$residuals,
                index.upper.tri = index.upper.tri,
                name.rhs = name.rhs,
                name.var = name.var,
                A = A,
                Omega.constrain = Omega.constrain,
                index.Psi = index.Psi
                ))
}

## * .sscResiduals
#' @title Compute Bias Corrected Quantities.
#' @description Compute bias corrected residuals variance covariance matrix
#' and information matrix.
#' Also provides the leverage values and corrected sample size when adjust.n is set to TRUE.
#' @name estimate2
#' 
#' @keywords internal
.sscResiduals <- function(object, ssc, algorithm = "2"){
    algorithm <- match.arg(as.character(algorithm), choices = c("1","2"))

    ## ** initial values (i.e. non bias corrected)
    Omega0 <- ssc$Omega0 
    Omega.constrain <- ssc$Omega.constrain
    param0 <- ssc$param0
    residuals0 <- ssc$residuals0

    ## ** current values
    Omega <- object$sCorrect$moment$Omega 
    epsilon <- object$sCorrect$residuals
    leverage <- object$sCorrect$leverage
    dmu <- object$sCorrect$dmoment$dmu
    dOmega <- object$sCorrect$dmoment$dOmega
    vcov.param <- object$sCorrect$vcov.param
    
    endogenous <- object$sCorrect$endogenous
    n.endogenous <- length(endogenous)
    n.cluster <- object$sCorrect$cluster$n.cluster
    param.mean <- object$sCorrect$skeleton$Uparam.mean
    param.var <- object$sCorrect$skeleton$Uparam.var
    param.hybrid <- intersect(param.mean,param.var)
    missing.pattern <- object$sCorrect$missing$pattern
    name.pattern <- object$sCorrect$missing$name.pattern
    n.pattern <- length(name.pattern)
    unique.pattern <- object$sCorrect$missing$unique.pattern

    if(length(param.mean)==0){
        stop("No mean parameter. No small sample correction needed. \n",
             "Consider setting  \'ssc\' to NA. \n")
    }
    ## ** Step (i-ii) compute individual and average bias
    dmu <- aperm(abind::abind(dmu[param.mean], along = 3), perm = c(3,2,1))
    vcov.muparam <- vcov.param[param.mean,param.mean,drop=FALSE]
    Psi <- matrix(0, nrow = n.endogenous, ncol = n.endogenous,
                  dimnames = list(endogenous, endogenous))
    n.Psi <- matrix(0, nrow = n.endogenous, ncol = n.endogenous,
                    dimnames = list(endogenous, endogenous))

    ## ls.Psi <- vector(mode = "list", length = n.cluster)
    for(iP in 1:n.pattern){ ## iP <- 1
        iY <- which(unique.pattern[iP,]>0)
        
        for(iC in missing.pattern[[iP]]){ ## iC <- 1
            ## individual bias
            if(length(param.mean)==1){
                iPsi <- vcov.muparam[1,1] * tcrossprod(dmu[,iY,iC])
            }else{
                iPsi <- t(dmu[,iY,iC])  %*% vcov.muparam %*% dmu[,iY,iC]
            }

            ## ls.Psi[[iC]] <- iPsi
            ## cumulated bias            
            Psi[iY,iY] <- Psi[iY,iY] + iPsi
            n.Psi[iY,iY] <- n.Psi[iY,iY] + 1
        }
    }
    ## take the average
    Psi[n.Psi>0] <- Psi[n.Psi>0]/n.Psi[n.Psi>0]
    
    ## ** Step (iii): compute corrected residuals and effective sample size
    ## done in estimate2 via moments2
    
    ## ** Step (iv): bias-corrected residual covariance matrix
    Omega.adj <- Omega0 + Psi

    ## ** Step (v): bias-corrected variance parameters
    A <- ssc$A

    ## *** right hand side of the equation
    index.upper.tri <- ssc$index.upper.tri[,"index"]
    eq.rhs <- stats::setNames((Omega.adj-Omega.constrain)[index.upper.tri],
                              ssc$name.rhs)

    ## *** left hand side of the equation
    index.Psi <- ssc$index.Psi

    if(NROW(index.Psi)>0){
        Z <- object$sCorrect$moment$iIB %*% object$sCorrect$moment$Lambda
        tZ <- t(Z)
        n.index.Psi <- NROW(index.Psi)
    
        ## A = t(Z) Psi Z + Sigma
        ## (t(Z) Psi Z)_{ij} = \sum_{k,l} Z_{k,i} Psi_{k,l} Z_{l,j}
        ## (t(Z) Psi Z)_{ij} regarding param_(k,l) = Z_{k,i} Z_{l,j}
        for(iPsi in 1:n.index.Psi){ ## iPsi <- 3
            iNamePsi <- rownames(index.Psi)[iPsi]
            iRowPsi <- index.Psi[iPsi,"row"]
            iColPsi <- index.Psi[iPsi,"col"]
            A[,iNamePsi] <- A[,iNamePsi] + (tZ[,index.Psi[iPsi,"row"]] %o% Z[index.Psi[iPsi,"col"],])[index.upper.tri]
            if(iRowPsi!=iColPsi){
                A[,iNamePsi] <- A[,iNamePsi] + (tZ[,index.Psi[iPsi,"col"]] %o% Z[index.Psi[iPsi,"row"],])[index.upper.tri]
            }
        }
        
    }

    ## *** solve equation
    ## microbenchmark::microbenchmark(svd = {asvd <- svd(A) ; asvd$v %*% diag(1/asvd$d) %*% t(asvd$u) %*% eq.rhs;},
    ## qr = qr.coef(qr(A), eq.rhs),
    ## Rcpp = OLS_cpp(A, eq.rhs),
    ## RcppTry = try(OLS_cpp(A, eq.rhs)[,1], silent = TRUE),
    ## Rcpp2 = OLS2_cpp(A, eq.rhs),
    ## OLS1 = solve(crossprod(A), crossprod(A, eq.rhs)),
    ## OLS2 = solve(t(A) %*% A) %*% t(A) %*% eq.rhs,
    ## OLS_stats = stats::lsfit(x = A, y = eq.rhs),
    ## OLS_LINPACK = .Call(stats:::C_Cdqrls, x = A, y = eq.rhs, tolerance = 1e-7, FALSE)$coefficients, times = 500)
    if(lava.options()$method.estimate2=="svd"){
        asvd <- svd(A)
        iSolution <- try((asvd$v %*% diag(1/asvd$d) %*% t(asvd$u) %*% eq.rhs)[,1], silent = TRUE)
    }else if(lava.options()$method.estimate2=="ols"){
        iSolution <- try(OLS_cpp(A, eq.rhs)[,1], silent = TRUE)
    }else{
        stop("unknown OLS methods \n")
    }
    
    if(inherits(iSolution, "try-error")){
        if(abs(det(t(A) %*% A)) <  1e-10){            
            stop("Singular matrix: cannot update the estimates \n")
        }else{
            stop(iSolution)
        }
    }
    names(iSolution) <- attr(A,"name")

    ## *** update parameters
    ## param0[ssc$name.var] - iSolution
    param0[names(iSolution)] <- iSolution

    ## ** Step (vi-vii): update derivatives and information matrix
    ## done in estimate2 via moments2
        
    ## ** Export
    attr(param0,"Omega") <- Omega.adj 
    attr(param0,"Psi") <- Psi
    return(param0)
}



##----------------------------------------------------------------------
### sCorrect-sscResiduals.R ends here

### sCorrect-coef2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 18 2019 (10:14) 
## Version: 
## Last-Updated: jan 23 2024 (10:23) 
##           By: Brice Ozenne
##     Update #: 321
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Model Coefficients With Small Sample Correction
#' @description Extract the coefficients from a latent variable model.
#' Similar to \code{lava::compare} but with small sample correction.
#' @name coef2
#'
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param as.lava [logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the model coefficients.
#' 
#' @return A numeric vector named with the names of the coefficients.
#' 
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#' 
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`coef2` <-
    function(object, as.lava, ...) UseMethod("coef2")


## * Examples
#' @rdname coef2
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' dW <- sampleRepeated(10, format = "wide")
#' set.seed(10)
#' dL <- sampleRepeated(10, format = "long")
#' dL$time2 <- paste0("visit",dL$time)
#' 
#' #### latent variable models ####
#' e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
#' coef(e.lvm)
#' coef2(e.lvm)
#' coef2(e.lvm, as.lava = FALSE)

## * coef2.lvmfit
#' @export
coef2.lvmfit <- function(object, as.lava = TRUE, ssc = lava.options()$ssc, ...){

    return(coef(estimate2(object, ssc = ssc, ...), as.lava = as.lava))

}

## * coef2.lvmfit2
#' @export
coef2.lvmfit2 <- function(object, as.lava = TRUE, ...){
    dots <- list(...)
    if(any(names(dots) %in% c("type","symbol","labels") == FALSE)){
        if(length(dots)>0){
            warning("Argument(s) \'",paste(setdiff(names(dots), c("type","symbol","labels")), collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
        }
    }
    if(length(dots)>1){ ## for the print function

        ## new values
        res <- model.tables(object, as.lava = TRUE)

        ## extract structure from lava
        object0 <- object
        class(object0) <- setdiff(class(object), "lvmfit2")
        out <- do.call(stats::coef, args = c(list(object0),dots)) ## this does not necessarily output the full parameter name
        ## Y1~~Y1 may be abreviated into Y1 which is confusion with Y1 the intercept
        dots$symbol <- NULL
        out.names <- intersect(rownames(do.call(stats::coef, args = c(list(object0),dots))), ## full name
                               rownames(res))
        index.out.names <- match(out.names, rownames(do.call(stats::coef, args = c(list(object0),dots))))

        ## rownames(res) <- as.character(object$sCorrect$skeleton$originalLink2param)
        out[index.out.names,"Estimate"] <- res[out.names,"estimate"]
        out[index.out.names,"Std. Error"] <- res[out.names,"se"]
        out[index.out.names,3] <- res[out.names,"statistic"] ## use 3 instead of Z value / Z-value
        if(object$sCorrect$df=="satterthwaite"){ 
            colnames(out)[3] <- "t value"
            if(colnames(out)[4]=="Pr(>|z|)"){ 
                colnames(out)[4] <- "Pr(>|t|)"
            }
        }
        out[index.out.names,4] <- res[out.names,"p.value"] ## use 4 instead of P-value / Pr(>|z|)

    }else{        
        out <- object$sCorrect$param[names(object$sCorrect$skeleton$originalLink2param)]
        if(as.lava==FALSE){
            names(out) <- as.character(object$sCorrect$skeleton$originalLink2param)
        }
    }
    return(out)
}

## * coef.lvmfit2
#' @export
coef.lvmfit2 <- coef2.lvmfit2

######################################################################
### sCorrect-coef2.R ends here

### calibrateType1.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: apr  5 2018 (10:23) 
## Version: 
## Last-Updated: Jan 11 2022 (16:51) 
##           By: Brice Ozenne
##     Update #: 852
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:


## * Documentation - calibrateType1
##' @title Simulation Study Assessing Bias and Type 1 Error
##' @description Perform a simulation study over one or several sample size
##' to assess the bias of the estimate
##' and the type 1 error of the Wald test and robust Wald test
##' @name calibrateType1
##' 
##' @param object a \code{lvm} object defining the model to be fitted.
##' @param warmup [list of lvm] a list of \code{lvm} objects that will be sequentially fitted with for
##' starting values the parameter of the previous model in the list (if any). The parameters of the final
##' model of the list are used to initialize the fit of the model of interest (i.e. object).
##' @param param [character vector] names of the coefficient whose value will be tested. 
##' @param null [numeric vector] vector of null hypotheses, one for each model coefficient.
##' By default a vector of 0. 
##' @param n [integer vector, >0] sample size(s) considered in the simulation study.
##' @param n.rep [integer, >0] number of simulations per sample size.
##' @param cluster  [integer vector] the grouping variable relative to which the observations are iid.
##' Will be passed to \code{lava::estimate}.
##' @param correction [logical] should the type 1 error after correction be computed?
##' @param generative.object [lvm] object defining the statistical model generating the data.
##' @param generative.coef [name numeric vector] values for the parameters of the generative model.
##' Can also be \code{NULL}: in such a case the coefficients are set to default values decided by lava (usually 0 or 1).
##' @param true.coef [name numeric vector] expected values for the parameters of the fitted model.
##' @param n.true [integer, >0] sample size at which the estimated coefficients will be a reliable approximation of the true coefficients.
##' @param round.true [integer, >0] the number of decimal places to be used for the true value of the coefficients. No rounding is done if \code{NULL}.
##' @param bootstrap [logical] should bootstrap resampling be performed?
##' @param n.bootstrap [integer, >0] the number of bootstrap sample to be used for each bootstrap.
##' @param checkType1 [logical] returns an error if the coefficients associated to the null hypotheses do not equal 0.
##' @param checkType2 [logical] returns an error if the coefficients associated to the null hypotheses equal 0.
##' @param dir.save [character] path to the directory were the results should be exported.
##' Can also be \code{NULL}: in such a case the results are not exported.
##' @param F.test [logical] should a multivariate Wald test be perform testing simultaneously all the null hypotheses?
##' @param label.file [character] element to include in the file name.
##' @param seed [integer, >0] value that will be set before adjustment for multiple comparisons to ensure reproducible results.
##' Can also be \code{NULL}: in such a case no seed is set.
##' @param cpus [integer >0] the number of processors to use.
##' If greater than 1, the simulations are performed in parallel. 
##' @param trace [integer] should the execution of the function be trace. Can be 0, 1 or 2.
##' @param ... [internal] Only used by the generic method.
##' 
##' @return An object of class \code{calibrateType1}.
##' @seealso \code{link{autoplot.calibrateType1}} for a graphical display of the bias or of the type 1 error.
##' 
##' @author Brice Ozenne
##'
##' @examples
##' \dontrun{
##' #### simulate data ####
##' m.Sim <- lvm(c(Y1[mu1:sigma]~1*eta,
##'                Y2[mu2:sigma]~1*eta,
##'                Y3[mu3:sigma]~1*eta,
##'                eta~beta1*Group+beta2*Gender))
##' latent(m.Sim) <- ~eta
##' categorical(m.Sim, labels = c("M","F")) <- ~Gender
##'
##' d <- lava::sim(m.Sim, 1e2)
##'
##' #### calibrate type 1 error on the estimated model ####
##' m <- lvm(Y1~eta,
##'          Y2~eta,
##'          Y3~eta,
##'          eta~Group+Gender)
##' e <- lava::estimate(m, data = d)
##' res <- calibrateType1(e, param = "eta~Group", n.rep = 100)
##' res <- calibrateType1(e, param = c("eta~Group","Y1~eta"), F.test = TRUE, n.rep = 100)
##' res <- calibrateType1(e, param = "eta~Group", n.rep = 100, cpus = 4)
##' summary(res)
##' }
##' 
##' @export
`calibrateType1` <-
    function(object, param, n.rep, ...) UseMethod("calibrateType1")

## * calibrateType1.lvm
##' @rdname calibrateType1
##' @export
calibrateType1.lvm <- function(object, param, n.rep, n, correction = TRUE, warmup = NULL, null = NULL,
                               F.test = FALSE, cluster = NULL,
                               generative.object = NULL, generative.coef = NULL, 
                               true.coef = NULL, n.true = 1e6, round.true = 2,              
                               bootstrap = FALSE, n.bootstrap = 1e3,
                               checkType1 = FALSE, checkType2 = FALSE,
                               dir.save = NULL, label.file = NULL,             
                               seed = NULL, cpus = 1, trace = 2, ...){

### ** test
    if(cpus>1){
        test.package <- try(requireNamespace("foreach"), silent = TRUE)
        if(inherits(test.package,"try-error")){
            stop("There is no package \'foreach\' \n",
                 "This package is necessary when argument \'cpus\' is greater than 1 \n")
        }

        if(cpus > parallel::detectCores()){
            stop("Argument \'cpus\' is greater than the number of available CPU cores \n",
                 "available CPU cores: ",parallel::detectCores(),"\n")
        }
    }
    
### ** prepare
    n.n <- length(n)

    ## *** generative model
    if(is.null(generative.object)){
        generative.object <- object
    }
    if(any(lava::manifest(object) %in% lava::manifest(generative.object) == FALSE)){
        missingVar <- lava::manifest(object)[lava::manifest(object) %in% lava::manifest(generative.object) == FALSE]
        stop("The object contains manifest variables that are not in the generative model \n",
             "missing manifest variables: \"",paste0(missingVar, collapse = "\" \""),"\" \n")
    }
    df.type_generative <- coefType(generative.object, as.lava = FALSE)
    name.param_generative <- df.type_generative[!is.na(df.type_generative$lava),"param"]
    n.param_generative <- length(name.param_generative)

    ## *** coef generative
    if(!is.null(generative.coef) && any(names(generative.coef) %in% name.param_generative == FALSE)){
        extraParam <- names(generative.coef)[names(generative.coef) %in% name.param_generative == FALSE]
        stop("Invalid argument \'generative.coef\': some of the coefficient names do not match those of the generative model \n",
             "extra coefficients: \"",paste0(extraParam, collapse = "\" \""),"\"\n")
    }

    ## *** coef of the fitted model
    if(is.null(true.coef)){
        if(trace>1){
            cat("  Estimate true coefficients using a sample size of n=",n.true," ", sep="")
        }
        e.true <- lava::estimate(object,
                                 data = lava::sim(generative.object, n = n.true, p = generative.coef, latent = FALSE))
        coef.true <- coef(e.true)
        if(!is.null(round.true)){
            coef.true <- round(coef.true, digits = round.true)
        }
        if(trace>1){
            cat("- done \n")
        }
    }else{
        if(trace>1){
            cat("  Check true coefficients ")
        }
        n.true <- n[n.n]
        e.true <- suppressWarnings(lava::estimate(object, cluster = cluster,
                                                  data = lava::sim(generative.object, n = n.true, p = generative.coef, latent = FALSE)))
        name.test <- names(coef(e.true))
        if(e.true$opt$convergence>0 && trace>1){
            cat("(incorrect convergence of the model) ")
        }
        
        if(!identical(sort(name.test),sort(names(true.coef)))){
            extraNames <- setdiff(names(true.coef),name.test)
            missingNames <- setdiff(name.test,names(true.coef))
            stop("Names of the coefficients in argument \'true.coef\' does not matches those of the estimated coefficients \n",
                 "missing names: \"",paste0(missingNames, collapse = "\" \""),"\" \n",
                 "extra names: \"",paste0(extraNames, collapse = "\" \""),"\" \n")
        }
        
        coef.true <- true.coef
        if(trace>1){
            cat("- done \n")
        }
        
    }
    name.coef <- names(coef.true)
    n.coef <- length(name.coef)
       
    ## *** type of the coef of the fitted model
    df.type <- coefType(e.true, as.lava = FALSE)
    df.type <- df.type[df.type$name %in% name.coef,]
    type.coef <- stats::setNames(df.type$detail, df.type$name)

    ## *** null hypothesis
    n.param <- length(param)
    if(any(param %in% name.coef == FALSE)){
        incorrect.name <- param[param %in% name.coef == FALSE]
        possible.name <- setdiff(name.coef, param)
        ls.name <- lapply(incorrect.name, function(iN){
            dist.tempo <- utils::adist(x = iN, y = possible.name)
            return(possible.name[which.min(dist.tempo)])
        })
        ex.name <- unique(unlist(ls.name))

        stop("Invalid argument \'param\': some of the coefficient names does not match those of the estimate model \n",
             "incorrect names: \"",paste(incorrect.name, collapse = "\" \""),"\" \n",
             "example of valid names: \"",paste(ex.name, collapse = "\" \""),"\"\n")
    }

    if(checkType1 && any(coef.true[param]!=0)){
        txtCoef <- paste(param[coef.true[param]!=0], collapse = "\" \"")
        stop("Control type 1 error: coefficients \"",txtCoef,"\" are not 0 while their belong to the param hypothesis\n")
    }
    if(checkType2 && any(coef.true[param]==0)){
        txtCoef <- paste(param[coef.true[param]==0], collapse = "\" \"")
        stop("Control type 2 error: coefficients \"",txtCoef,"\" are 0 while their belong to the param hypothesis\n")
    }
    res.C <- .createContrast(param,
                             name.param = name.coef,
                             add.rowname = TRUE,
                             rowname.rhs = FALSE)
    contrast <- res.C$contrast
    if(is.null(null)){
        rhs <- res.C$null
    }else{
        if(length(null)!=n.param){
            stop("Argument \'null\' have the same length as argument \'param\' \n")
        }
        if(!is.null(names(null)) && any(names(null)!=param)){
            stop("If named, argument \'null\' must have for name argument \'param\' \n")
        }
        rhs <- null
    }

    ## *** warmup
    if(!is.null(warmup) && inherits(warmup,"lvm")){
        stop("Argument \'warmup\' must be a list of lvm objects \n")
    }
 
### ** display
    if(trace>1){
        cat("  Settings: \n")
        cat("  > simulation for n=",paste(n,collapse = " "),"\n",sep="")
        cat("  > model: \n")
        print(object)
        cat("  > expected coefficients: \n")
        print(coef.true)
        cat("  > null hypotheses: \n")
        print(rhs)
        if(bootstrap){
            cat("  > bootstrap: ",bootstrap,"\n")
        }
        if(!is.null(seed)){
            cat("  > seed: ",seed,"\n")
        }        
    }

    
### ** loop
    store.coef <- paste0(rownames(contrast)," == ",rhs)
    if(F.test){
        store.coef <- c(store.coef, "global")
    }
    n.store <- length(store.coef)
    
    if(trace>1){
        cat("\n")
        cat(" Perform simulation: \n")
    }
    if(cpus>1){
        ## *** define cluster
        if(trace>0){
            cl <- suppressMessages(parallel::makeCluster(cpus, outfile = ""))
            pb <- utils::txtProgressBar(max = n.rep, style = 3)          
        }else{
            cl <- parallel::makeCluster(cpus)
        }
        ## *** link to foreach
        doParallel::registerDoParallel(cl)

        ## *** export package
        parallel::clusterCall(cl, fun = function(x){
            suppressPackageStartupMessages(requireNamespace("lava", quietly = TRUE))
            suppressPackageStartupMessages(requireNamespace("lavaSearch2", quietly = TRUE))
        })
        
        ## *** seed
        cpus.name <- unlist(parallel::clusterApply(cl = cl, 1:cpus, function(x){
            myName <- paste(Sys.info()[["nodename"]], Sys.getpid(), sep="-")
            return(myName)
        }))
        if(length(seed)==0){
            seed <- rep(as.numeric(NA), cpus)
        }else if(length(seed)==1){
            set.seed(seed)
            seed <- sample(1:1e5,size=cpus,replace=FALSE)                
        }else{
            if(length(seed)!=cpus){
                stop("Length of argument \'seed\' does not match argument \'cpus\' \n")
            }            
        }
        names(seed) <- cpus.name

        
        ## *** parallel computation
        toExport <- c(".warperType1", "cpus.name")

        iRep <- NULL # [:for CRAN check] foreach
        resSim <- foreach::`%dopar%`(
                               foreach::foreach(iRep = 1:n.rep,
                                                .export = toExport,
                                                .packages = c("lavaSearch2")),{ # iRep <- 1

                                                    if(trace>0){utils::setTxtProgressBar(pb, iRep)}

                                                    myName <- paste(Sys.info()[['nodename']], Sys.getpid(), sep='-')
                                                    iSeed <- as.double(seed[myName])
                                                    
                                                    ls.pvalue <- vector(mode = "list", length = n.n)
                                                    ls.estimate <- vector(mode = "list", length = n.n)
                                                    iIndex <- 1
                                                    
                                                    for(iN in n){ # iN <- n[1]
                                                        iRes <- .warperType1(iRep,
                                                                             n = iN,
                                                                             correction = correction,
                                                                             generative.object = generative.object,
                                                                             generative.coef = generative.coef,
                                                                             object = object,
                                                                             warmup = warmup,
                                                                             cluster = cluster,
                                                                             coef.true = coef.true,
                                                                             type.coef = type.coef,
                                                                             name.coef = name.coef,
                                                                             store.coef = store.coef,
                                                                             n.coef = n.coef,
                                                                             n.store = n.store,
                                                                             F.test = F.test,
                                                                             param = param,
                                                                             contrast = contrast,
                                                                             rhs = rhs,
                                                                             bootstrap = bootstrap,
                                                                             n.bootstrap = n.bootstrap,
                                                                             seed = iSeed,
                                                                             ...)

                                                        ls.pvalue[[iIndex]] <- iRes$pvalue
                                                        ls.estimate[[iIndex]] <- iRes$estimate
                                                        iIndex <- iIndex + 1
                                                    }
                                                    
                                                    return(list(pvalue = do.call("rbind",ls.pvalue),
                                                                estimate = do.call("rbind",ls.estimate)))
                                                })
        
        parallel::stopCluster(cl)
        if(trace>0){close(pb)}
        ## *** post process
        dt.pvalue <- do.call("rbind",lapply(resSim,"[[","pvalue"))
        dt.pvalue <- dt.pvalue[order(dt.pvalue$n,dt.pvalue$rep),,drop=FALSE]
        dt.estimate <- do.call("rbind",lapply(resSim,"[[","estimate"))
        dt.estimate <- dt.estimate[order(dt.estimate$n,dt.estimate$rep),,drop=FALSE]
        
    }else{

        ## *** filename
        if(is.null(label.file)){label.file <- seed}
        filename_tempo.pvalue <- paste0("type1error-S",label.file,"(tempo).rds")
        filename_tempo.estimate <- paste0("estimate-S",label.file,"(tempo).rds")
        filename.pvalue <- gsub("\\(tempo\\)","",filename_tempo.pvalue)
        filename.estimate <- gsub("\\(tempo\\)","",filename_tempo.estimate)

        if(!is.null(dir.save)){
            if (dir.exists(dir.save) == FALSE) {
                stop("Argument \'dir.save\' does not lead to an existing directory \n")
            }            
        }

        if(!is.null(dir.save)){
            cat("  > export results in ",dir.save,"\n")
        }

        if(trace>0){
            test.package <- try(requireNamespace("pbapply"), silent = TRUE)
            if(inherits(test.package,"try-error")){
                stop("There is no package \'pbapply\' \n",
                     "This package is necessary when argument \'trace\' is TRUE \n")
            }
            FCTapply <- pbapply::pblapply
        }else{
            FCTapply <- lapply
        }
        if(!is.null(seed)){
            set.seed(seed)
        }else{
            seed <- as.numeric(NA)
        }
        
        ## *** sequential simulation
        dt.pvalue <- NULL
        dt.estimate <- NULL

        for(iN in n){ ## iN <- n[1]

            if(trace>0){
                cat("  > sample size=",iN,"\n", sep = "")                    
            }

            resSim <- do.call(FCTapply, args = list(X = 1:n.rep, FUN = function(iRep){
                .warperType1(iRep,
                             n = iN,
                             correction = correction,
                             generative.object = generative.object,
                             generative.coef = generative.coef,
                             object = object, warmup = warmup, cluster = cluster,                             
                             coef.true = coef.true, type.coef = type.coef, name.coef = name.coef,
                             store.coef = store.coef, n.coef = n.coef, n.store = n.store,
                             F.test = F.test, param = param, contrast = contrast, rhs = rhs,
                             bootstrap = bootstrap,
                             n.bootstrap = n.bootstrap,
                             seed = seed,
                             ...)
            }))
            dt.pvalue <- rbind(dt.pvalue,
                               do.call("rbind",lapply(resSim,"[[","pvalue"))
                               )
            dt.estimate <- rbind(dt.estimate,
                                 do.call("rbind",lapply(resSim,"[[","estimate")))
                             
            
            ## export (tempo)
            if(!is.null(dir.save)){
                saveRDS(dt.pvalue, file = file.path(dir.save,filename_tempo.pvalue))
                saveRDS(dt.estimate, file = file.path(dir.save,filename_tempo.estimate))
            }
            
        }   
    }

    ## ** export
    if(!is.null(dir.save)){
        saveRDS(dt.pvalue, file = file.path(dir.save,filename.pvalue))
        saveRDS(dt.estimate, file = file.path(dir.save,filename.estimate))
    }
    out <- list(estimate = dt.estimate,
                p.value = dt.pvalue,
                e.true = e.true,
                param = param)
    class(out) <- append("calibrateType1",class(out))
    return(out)


}

## * calibrateType1.lvmfit
##' @rdname calibrateType1
##' @export
calibrateType1.lvmfit <- function(object, param, n.rep, correction = TRUE, F.test = FALSE,
                                  bootstrap = FALSE, n.bootstrap = 1e3,
                                  seed = NULL, trace = 2, cpus = 1, ...){

    ## ** Prepare
    ## *** model
    object.model <- object$model

    ## *** coef
    coef.true <- coef(object)
    name.coef <- names(coef.true)
    if(any(param %in% name.coef == FALSE)){
        txt <- param[param %in% name.coef == FALSE]
        txt2 <- setdiff(name.coef, param)
        stop("Argument \'param\' does not match the names of the model coefficients \n",
             "Incorrect param: \"",paste(txt, collapse = "\" \""),"\" \n",
             "Possible param: \"",paste(txt2, collapse = "\" \""),"\" \n")
    }
    coef.true[param] <- 0

    ## *** data
    n <- object$data$n

    ## ** Run
    out <- calibrateType1(object.model,
                          param = param,
                          n.rep = n.rep,
                          n = n,
                          correction = correction,
                          F.test = F.test,
                          generative.object = object.model,
                          generative.coef = coef.true, 
                          true.coef = coef.true,              
                          bootstrap = bootstrap,
                          n.bootstrap = n.bootstrap,
                          checkType1 = FALSE,
                          checkType2 = FALSE,
                          dir.save = NULL,
                          label.file = NULL,             
                          seed = seed,
                          cpus = cpus,
                          trace = trace,
                          ...)


    ## ** Export
    return(out)
}

## * .warperType1
.warperType1 <- function(iRep, n, correction, generative.object, generative.coef,
                         object, warmup, cluster,
                         coef.true, type.coef, name.coef, store.coef, n.coef, n.store,
                         F.test, param, contrast, rhs,
                         bootstrap, n.bootstrap,
                         seed,
                         ...){

    ls.iE <- list(estimate.truth = coef.true)  ## temporary
    ls.iP <- list()  ## temporary
    out <- list()
    dots <- list(...)
    if("control" %in%  names(dots) == FALSE){
        dots$control <- list()
    }
    ## ** simulation
    dt.sim <- lava::sim(generative.object, n = n, p = generative.coef, latent = FALSE)

    ## ** initialisation
    if(!is.null(warmup)){
        n.warmup <- length(warmup)
        for(iW in 1:n.warmup){ ## iW <- 1
            iE.lvm <- suppressWarnings(do.call(lava::estimate, args = c(list(warmup[[iW]], data = dt.sim, cluster = cluster), dots)))
            if(("convergence" %in% names(iE.lvm$opt)) && (iE.lvm$opt$convergence==1)){return(list(pvalue=NULL,estimate=NULL))} ## exclude lvm that has not converged
            dots$control$start <- coef(iE.lvm)
        }
    }
    
    ## ** model fit
    e.lvm <- suppressWarnings(do.call(lava::estimate, args = c(list(object, data = dt.sim, cluster = cluster), dots)))
    
    ## check correct convergence of the latent variable model
    if(("convergence" %in% names(e.lvm$opt)) && (e.lvm$opt$convergence==1)){return(list(pvalue=NULL,estimate=NULL))} ## exclude lvm that has not converged
    if(any(eigen(getVarCov2(e.lvm, ssc = FALSE, df = FALSE))$values<=0)){return(list(pvalue=NULL,estimate=NULL))} ## exclude lvm where the residual covariance matrix is not semipositive definite
    ratio_sd_beta <- sqrt(diag(vcov(e.lvm)))/(abs(coef(e.lvm))+1)
    if(max(stats::na.omit(ratio_sd_beta))>1e3){return(list(pvalue=NULL,estimate=NULL))} ## exclude if standard error much larger than coefficient
    if(inherits(suppressWarnings(try(summary(e.lvm)$coef, silent = TRUE)), "try-error")){return(list(pvalue=NULL,estimate=NULL))} ## exclude lvm where we cannot compute the summary

    ## ** corrections
    e.lvm <- estimate2(e.lvm, df = "none", ssc = "none")
    if(correction){
        e.lvm.Satt <- try(estimate2(e.lvm, df = "satterthwaite", ssc = "none"), silent = TRUE)
        e.lvm.KR <- try(suppressWarnings(estimate2(e.lvm, df = "satterthwaite", ssc = "residuals")), silent = TRUE)
    }else{
        e.lvm.Satt <- 1
        class(e.lvm.Satt) <- "try-error"
        e.lvm.KR <- 1
        class(e.lvm.KR) <- "try-error"
    }
    
    ## ** extract p.values
    eS.ML <- summary2(e.lvm, robust = FALSE)$coef
    F.ML <- summary(compare2(e.lvm, robust = FALSE,
                             linfct = contrast, rhs = rhs, F.test = F.test, as.lava = FALSE))$table2

    eS.robustML <- summary2(e.lvm, robust = TRUE)$coef
    F.robustML <- summary(compare2(e.lvm, robust = TRUE,
                                   linfct = contrast, rhs = rhs, F.test = F.test, as.lava = FALSE))$table2
    

    if(!inherits(e.lvm.Satt,"try-error")){
        
        eS.Satt <- summary2(e.lvm.Satt, robust = FALSE)$coef
        F.Satt <- summary(compare2(e.lvm.Satt, robust = FALSE,
                                   linfct = contrast, rhs = rhs, F.test = F.test,
                                   as.lava = FALSE))$table2
            
        eS.robustSatt <- summary2(e.lvm.Satt, robust = TRUE)$coef
        F.robustSatt <- summary(compare2(e.lvm.Satt, robust = TRUE,
                                  linfct = contrast, rhs = rhs, F.test = F.test,
                                  as.lava = FALSE))$table2
    }
    
    if(!inherits(e.lvm.KR,"try-error")){

        eS.SSC <- summary2(e.lvm.KR, robust = FALSE)$coef
        F.SSC <- summary(compare2(e.lvm.KR, robust = FALSE, 
                                  linfct = contrast, rhs = rhs, F.test = F.test,
                                  as.lava = FALSE))$table2
        
        eS.robustSSC <- summary2(e.lvm.KR, robust = TRUE)$coef
        F.robustSSC <- summary(compare2(e.lvm.KR, robust = TRUE, 
                                        linfct = contrast, rhs = rhs, F.test = F.test,
                                        as.lava = FALSE))$table2

        eS.KR <- summary2(e.lvm.KR, robust = FALSE)$coef
        F.KR <- summary(compare2(e.lvm.KR, robust = FALSE,
                                 linfct = contrast, rhs = rhs, F.test = F.test,
                                 as.lava = FALSE))$table2
        
        eS.robustKR <- summary2(e.lvm.KR, robust = TRUE)$coef
        F.robustKR <- summary(compare2(e.lvm.KR, robust = TRUE,
                                       linfct = contrast, rhs = rhs, F.test = F.test,
                                       as.lava = FALSE))$table2
    }
    
    ## ** store

    ## *** estimates
    ls.iE$estimate.ML <- eS.ML[name.coef,"Estimate"]
    if(!inherits(e.lvm.KR,"try-error")){
        ls.iE$estimate.MLcorrected <- eS.KR[name.coef,"Estimate"]
    }else{
        ls.iE$estimate.MLcorrected <- rep(as.numeric(NA), n.coef)
    }

    ## *** standard errors
    if(is.null(cluster)){
        ls.iE$se.ML <- eS.ML[name.coef,"Std. Error"]
        if(!inherits(e.lvm.KR,"try-error")){
            ls.iE$se.MLcorrected <- eS.KR[name.coef,"Std. Error"]
        }else{
            ls.iE$se.MLcorrected <- rep(as.numeric(NA), n.coef)
        }

        ls.iE$se.robustML <- eS.robustML[name.coef,"robust SE"]
        if(!inherits(e.lvm.KR,"try-error")){
            ls.iE$se.robustMLcorrected <- eS.robustKR[name.coef,"robust SE"]
        }else{
            ls.iE$se.robustMLcorrected <- rep(as.numeric(NA), n.coef)
        }
    }

    ## *** degree of freedom
    if(!inherits(e.lvm.Satt,"try-error")){
        if(is.null(cluster)){
            ls.iE$df.ML <- eS.Satt[name.coef,"df"]
        }
        ls.iE$df.robustML <- eS.robustSatt[name.coef,"df"]
    }
    
    if(!inherits(e.lvm.KR,"try-error")){
        if(is.null(cluster)){
            ls.iE$df.MLcorrected <- eS.KR[name.coef,"df"]
        }
        ls.iE$df.robustMLcorrected <- eS.robustKR[name.coef,"df"]
    }
    
    ## *** p-value
    if(is.null(cluster)){
        ls.iP$p.Ztest <- F.ML[store.coef,"p.value"]
        if(!inherits(e.lvm.Satt,"try-error")){
            ls.iP$p.Satt <- F.Satt[store.coef,"p.value"]
        }
        if(!inherits(e.lvm.KR,"try-error")){
            ls.iP$p.SSC <- F.SSC[store.coef,"p.value"]
            ls.iP$p.KR <- F.KR[store.coef,"p.value"]
        }
    }
    ls.iP$p.robustZtest <- F.robustML[store.coef,"p.value"]
    if(!inherits(e.lvm.Satt,"try-error")){
        ls.iP$p.robustSatt <- F.robustSatt[store.coef,"p.value"]
    }
    if(!inherits(e.lvm.KR,"try-error")){
        ls.iP$p.robustSSC <- F.robustSSC[store.coef,"p.value"]
        ls.iP$p.robustKR <- F.robustKR[store.coef,"p.value"]
    }

    ## *** niter.correct / warning
    if(!inherits(e.lvm.KR,"try-error")){
        test.warning <- inherits(attr(e.lvm.KR$sCorrect,"warning"),"try-error")
        niter.correct <- as.double(e.lvm.KR$sCorrect$ssc$iter)
    }else{
        test.warning <- NA
        niter.correct <- as.double(NA)
    }
    
    ## ** bootstrap
    if(bootstrap>0){
        e.boot <- eval(parse(text = "butils::bootReg(e.lvm, type = \"coef\", n.boot = n.bootstrap"))

        index.coef.boot <- match(param, name.coef)
        boot.perc <- summary(e.boot, p.value = TRUE, type = "perc", print = FALSE, index = index.coef.boot)
        boot.stud <- summary(e.boot, p.value = TRUE, type = "stud", print = FALSE, index = index.coef.boot)
        boot.bca <- summary(e.boot, p.value = TRUE, type = "bca", print = FALSE, index = index.coef.boot)

        ls.iP$p.bootPerc <- boot.perc[store.coef,"p.value"]
        ls.iP$p.bootStud <- boot.stud[store.coef,"p.value"]
        ls.iP$p.bootBca <- boot.bca[store.coef,"p.value"]
    }else{
        n.bootstrap <- as.numeric(NA)
    }

    ## ** collect results
    ## estimates
    ## use rep to avoid warning
    ## return(list(n = n,
    ## rep = iRep,
    ## seed = seed,
    ## ninter = niter.correct,
    ## warning = test.warning,
    ## name = names(coef.true),
    ## type = type.coef[name.coef]))
    df1 <- data.frame(n = n,
                      rep = iRep, n.coef,
                      seed = seed, n.coef,
                      niter = niter.correct,
                      warning = test.warning,
                      name = names(coef.true),
                      type = unname(type.coef[name.coef]),
                      stringsAsFactors = FALSE)
    rownames(df1) <- NULL

    df2 <- do.call(cbind,ls.iE)
    rownames(df2) <- NULL

    out$estimate <- cbind(df1, df2)

    ## p.value
    df1 <- data.frame(n = n,
                      rep = iRep,
                      seed = seed,
                      nboot = n.bootstrap,
                      niter = niter.correct,
                      warning = test.warning,
                      link = unname(store.coef),
                      stringsAsFactors = FALSE)
    rownames(df1) <- NULL

    df2 <- do.call(cbind,ls.iP)
    rownames(df2) <- NULL

    out$pvalue <- cbind(df1, df2)

    ## ** export
    return(out)
}
    
######################################################################
### calibrateType1.R ends here

## * .mergeCalibrateType1
.mergeCalibrateType1 <- function(list){

    
    if(any(sapply(list,function(iL){inherits(iL,"calibrateType1")})==FALSE)){
        stop("Argument \'list\' must only contain \"calibrateType1\" objects \n")
    }

    newObject <- list[[1]]
    newObject$estimate <- do.call(rbind,lapply(list,"[[","estimate"))
    newObject$p.value <- do.call(rbind,lapply(list,"[[","p.value"))

    return(newObject)
}

### sCorrect-residuals2.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: nov 18 2019 (11:17) 
## Version: 
## Last-Updated: jan 23 2024 (10:26) 
##           By: Brice Ozenne
##     Update #: 141
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Residuals With Small Sample Correction.
#' @description Extract residuals from a latent variable model.
#' Similar to \code{stats::residuals} but with small sample correction.
#' @name residuals2
#' 
#' @param object a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).
#' @param type [character] the type of residual to extract:
#' \code{"response"} for raw residuals,
#' \code{"studentized"} for studentized residuals,
#' \code{"normalized"} for normalized residuals.
#' @param format [character] Use \code{"wide"} to return the residuals in the wide format (one row relative to each sample).
#' Otherwise use \code{"long"} to return the residuals in the long format.
#' @param ssc [character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
#' correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
#' Only relevant when using a \code{lvmfit} object. 
#' @param ... additional argument passed to \code{estimate2} when using a \code{lvmfit} object. 
#'
#' @seealso \code{\link{estimate2}} to obtain \code{lvmfit2} objects.
#'
#' @details When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the residuals.
#'
#' The raw residuals are defined by  observation minus the fitted value:
#' \deqn{
#' \varepsilon = (Y_1 - \mu_1, ..., Y_m - \mu_m)
#' }
#' The studentized residuals divided the raw residuals relative to each endogenous variable by the modeled variance of the endogenous variable.
#' \deqn{
#' \varepsilon_{stud} =(\frac{Y_1 - \mu_1}{\sigma_1}, ..., \frac{Y_m - \mu_m}{\sigma_m})
#' }
#' The normalized residuals multiply the raw residuals by the inverse of the square root of the modeled residual variance covariance matrix.
#' \deqn{
#' \varepsilon_{norm} = \varepsilon \Omega^{-1/2}
#' }
#' @return a matrix containing the residuals relative to each sample (in rows)
#' and each endogenous variable (in column).
#' 
#' @concept extractor
#' @keywords smallSampleCorrection
#' @export
`residuals2` <-
    function(object, type, format, ssc, ...) UseMethod("residuals2")

## * Examples
#' @rdname residuals2
#' @examples
#' #### simulate data ####
#' set.seed(10)
#' n <- 101
#'
#' Y1 <- rnorm(n, mean = 0)
#' Y2 <- rnorm(n, mean = 0.3)
#' Id <- findInterval(runif(n), seq(0.1,1,0.1))
#' data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
#'            data.frame(Y=Y2,G="2",Id = Id)
#'            )
#'
#' #### latent variable models ####
#' library(lava)
#' e.lvm <- estimate(lvm(Y ~ G), data = data.df)
#' residuals(e.lvm)
#' residuals2(e.lvm)
#' residuals(e.lvm) - residuals2(e.lvm)
#'

## * residuals2.lvmfit
#' @export
residuals2.lvmfit <- function(object, type = "response", format = "wide", ssc = lava.options()$ssc, ...){

    return(residuals(estimate2(object, ssc = ssc, ...), type = type, format = format))

}

## * residuals2.lvmfit2
#' @export
residuals2.lvmfit2 <- function(object, type = "response", format = "wide", ...){

    dots <- list(...)
    if(length(dots)>0){
        warning("Argument(s) \'",paste(names(dots),collapse="\' \'"),"\' not used by ",match.call()[1],". \n")
    }

    format <- match.arg(format, choices = c("long","wide"))

    residuals <- .normalizeResiduals(epsilon = object$sCorrect$residuals,
                                     Omega = object$sCorrect$moment$Omega,
                                     type = type,
                                     missing.pattern = object$sCorrect$missing$pattern,
                                     unique.pattern = object$sCorrect$missing$unique.pattern,
                                     Omega.missing.pattern = object$sCorrect$moment$Omega.missing.pattern)
    if(format == "wide"){
        return(residuals)
    }else if(format == "long"){
        endogenous <- colnames(residuals)
        n.endogenous <- length(endogenous)
        
        residualsW <- data.frame(1:NROW(residuals), residuals)
        names(residualsW) <- c("cluster",endogenous)
        residualsL <- stats::na.omit(stats::reshape(residualsW,
                                                    idvar = "cluster",
                                                    direction = "long",
                                                    varying = list(endogenous),
                                                    timevar = "endogenous",
                                                    v.names = "residual"))

        rownames(residualsL) <- NULL
        residualsL$endogenous <- factor(residualsL$endogenous, levels = 1:n.endogenous, labels = endogenous)
        reorder <- match(interaction(object$sCorrect$old2new.order$XXclusterXX.old, object$sCorrect$old2new.order$XXendogenousXX.old),
                         interaction(residualsL$cluster,residualsL$endogenous))
        return(residualsL[reorder,"residual"])
    }
}

## * residuals.lvmfit2
#' @export
residuals.lvmfit2 <- residuals2.lvmfit2

## * .normalizeResiduals
.normalizeResiduals <- function(epsilon, Omega, type,
                                missing.pattern, unique.pattern, Omega.missing.pattern){
    type <- match.arg(type, choices = c("response","studentized","normalized"), several.ok = FALSE)

    if(type %in% c("studentized")){
        epsilon <- sweep(epsilon,
                         STATS = sqrt(diag(Omega)),
                         FUN = "/",
                         MARGIN = 2)
        ## object$sCorrect$residuals/epsilon
    }else if(type=="normalized"){
        name.endogenous <- colnames(epsilon)
        if(any(is.na(epsilon))==FALSE){
            epsilon <- epsilon %*% matrixPower(Omega, symmetric = TRUE, power = -1/2)
        }else{
            iOmegaHalf.missing.pattern <- lapply(Omega.missing.pattern,matrixPower,symmetric = TRUE, power = -1/2)
            for(iP in names(missing.pattern)){
                iY <- which(unique.pattern[iP,]==1)
                for(iC in missing.pattern[[iP]]){ ## iC <- 1
                    epsilon[iC,iY] <- epsilon[iC,iY] %*% iOmegaHalf.missing.pattern[[iP]]
                }
            }
            
        }
        colnames(epsilon) <- name.endogenous
    }

    return(epsilon)
}

## * .adjustResiduals
.adjustResiduals <- function(epsilon, Psi, Omega, 
                             name.pattern, missing.pattern, unique.pattern,
                             endogenous, n.endogenous, n.cluster){
    if(is.null(Psi)){return(epsilon)}
    n.endogenous <- length(endogenous)

    epsilon.adj <- matrix(NA, nrow = n.cluster, ncol = n.endogenous,
                          dimnames = list(NULL, endogenous))
    n.pattern <- NROW(unique.pattern)
    
    for(iP in 1:n.pattern){ ## iP <- 1 
        iIndex <- missing.pattern[[iP]]
        iY <- which(unique.pattern[iP,]==1)
        
        iOmega <- Omega[iY,iY,drop=FALSE]
        iPsi <- Psi[iY,iY,drop=FALSE]

        iOmega.chol <- matrixPower(iOmega, symmetric = TRUE, power = 1/2)
        iH <- iOmega %*% iOmega - iOmega.chol %*% iPsi %*% iOmega.chol
        iHM1 <- tryCatch(matrixPower(iH, symmetric = TRUE, power = -1/2), warning = function(w){w})
        if(inherits(iHM1,"warning")){
            stop("Cannot compute the adjusted residuals \n",
                 "Estimated bias too large compared to the estimated variance-covariance matrix \n",
                 "Consider setting argument \'adjust.n\' to FALSE when calling sCorrect \n")
        }
        epsilon.adj[iIndex,iY] <- epsilon[iIndex,iY,drop=FALSE] %*% iOmega.chol %*% iHM1 %*% iOmega.chol
    }

    return(epsilon.adj)
}



######################################################################
### sCorrect-residuals2.R ends here

### coefByType.R --- 
##----------------------------------------------------------------------
## Author: Brice Ozenne
## Created: feb  5 2018 (11:37) 
## Version: 
## Last-Updated: okt  4 2018 (17:42) 
##           By: Brice Ozenne
##     Update #: 32
##----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
##----------------------------------------------------------------------
## 
### Code:

## * Documentation
#' @title Extract the Coefficient by Type
#' @description Extract specific types of coefficient from a \code{lvm} object:
#' covariance coefficient(s) (\code{coefCov}),
#' extra parameter(s) (\code{coefExtra}),
#' position in the list of models for each coefficient (\code{coefIndexModel}),
#' intercept coefficient(s) (\code{coefIntercept}),
#' coefficient(s) that are used as reference (\code{coefRef}),
#' regression coefficient(s) (\code{coefReg}),
#' variance coefficient(s) (\code{coefVar}).
#'  
#' @name coefByType
#' 
#' @param object a lvm model or a fitted lvm model 
#' @param value should the name of the coefficient be returned? Else return the coefficients
#' @param keep.var should the variance coefficients be returned?
#' @param ... arguments to be passed to \code{}
#'
#' @return A vector containing the names of the positions of the coefficients.
#' 
#' @examples 
#' #### regression ####
#' m <- lvm(Y~X1+X2)
#' e <- estimate(m, lava::sim(m, 1e2))
#' 
#' coefCov(m)
#' coefCov(m, value = TRUE)
#'
#' coefCov(m, keep.var = TRUE)
#' coefCov(m, value = TRUE, keep.var = TRUE)
#'
#' coefIndexModel(m)
#' coefIndexModel(e)
#' 
#' coefIntercept(m)
#' coefIntercept(m, value = TRUE)
#'
#' coefReg(m)
#' coefReg(m, value = TRUE)
#' 
#' #### LVM ####
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x1+x2
#' latent(m) <- ~u
#' covariance(m) <- y1~y2
#'
#' m.Sim <- m
#' categorical(m.Sim, labels = c("a","b","c")) <- ~x2
#' e <- estimate(m, lava::sim(m.Sim, 1e2))
#'
#' coefCov(m)
#' coefCov(m, value = TRUE) 
#'
#' coefCov(m, keep.var = TRUE)
#' coefCov(m, value = TRUE, keep.var = TRUE)
#' 
#' coefExtra(m)
#'
#' coefIndexModel(m)
#' coefIndexModel(e)
#'
#' ## additional categorical variable 
#' categorical(m, labels = as.character(1:3)) <- "X1"
#' 
#' coefExtra(m)
#' coefExtra(m, value = TRUE)
#'
#' ## additional categorical variable
#' categorical(m, labels = as.character(1:3)) <- "x1"
#'
#' coefIntercept(m)
#' coefIntercept(m, value = TRUE)
#' coefIntercept(e)
#'
#' coefReg(e, value = TRUE)
#' 
#' #### multigroup ####
#' m <- lvm(Y~X1+X2)
#' eG <- estimate(list(m,m), list(lava::sim(m, 1e2), lava::sim(m, 1e2)))
#' 
#' coefIndexModel(eG)
#'
#' @concept extractor

## * coefCov
#' @rdname coefByType
#' @export
`coefCov` <-
  function(object, value, keep.var, ...) UseMethod("coefCov")

## ** coefCov.lvm
#' @rdname coefByType
#' @export
coefCov.lvm <- function(object, value = FALSE, keep.var = FALSE, ...){

    res <- retainType(type = coefType(object, ...),
                      validType = c("covariance", if(keep.var){"variance"}else{NULL}),
                      value = value)

    return(res)
}

## ** coefCov.lvmfit
#' @rdname coefByType
#' @export
coefCov.lvmfit <- coefCov.lvm

## ** coefCov.multigroup
#' @rdname coefByType
#' @export
coefCov.multigroup <- coefCov.lvm

## * coefExtra

#' @rdname coefByType
#' @export
`coefExtra` <-
  function(object, value, ...) UseMethod("coefExtra")

## ** coefExtra.lvm
#' @rdname coefByType
#' @export
coefExtra.lvm <- function(object, value = FALSE, ...){
    
    res <- retainType(type = coefType(object, ...),
                      validType = "extra",
                      value = value) 
    
    return(res)    
}

## ** coefExtra.lvmfit
#' @rdname coefByType
#' @export
coefExtra.lvmfit <- coefExtra.lvm

## ** coefExtra.multigroup
#' @rdname coefByType
#' @export
coefExtra.multigroup <- coefExtra.lvm

## * coefIndexModel
#' @rdname coefByType
#' @export
`coefIndexModel` <-
  function(object,...) UseMethod("coefIndexModel")

## ** coefIndexModel.lvm
#' @rdname coefByType
#' @export
coefIndexModel.lvm <- function(object, ...){
    name.coef <- stats::coef(object)
    index <- rep(1, length(name.coef))
    names(index) <- name.coef
    return(index)
}

## ** coefIndexModel.lvmfit
#' @rdname coefByType
#' @export
coefIndexModel.lvmfit <- function(object, ...){
    name.coef <- names(stats::coef(object))
    index <- rep(1, length(name.coef))
    names(index) <- name.coef
    return(index)
}

## ** coefIndexModel.multigroup
#' @rdname coefByType
#' @export
coefIndexModel.multigroup <- function(object, ...){
    n.model <- length(object$lvm)

    ## new coef names
    allCoef <- object$name
    n.allCoef <- length(allCoef)
    index.AllCoef <- object$coef
  
    index <- stats::setNames(rep(NA, n.allCoef), allCoef)
  
    for(iModel in 1:n.model){ # iModel <- 1
        index[index.AllCoef[[iModel]]] <- iModel
    }
  
    #### export
    return(index)
}

## ** coefIndexModel.multigroupfit
#' @rdname coefByType
#' @export
coefIndexModel.multigroupfit <- function(object, ...){
    out <- coefIndexModel(object$model)
    return(out[names(stats::coef(object))])
}
  
## * coefIntercept

#' @rdname coefByType
#' @export
`coefIntercept` <-
  function(object, value, ...) UseMethod("coefIntercept")

## ** coefIntercept.lvm
#' @rdname coefByType
#' @export
coefIntercept.lvm <- function(object, value = FALSE, ...){ 

    res <- retainType(type = coefType(object, ...),
                      validType = "intercept",
                      value = value)

    return(res)
}

## ** coefIntercept.lvmfit
#' @rdname coefByType
#' @export
coefIntercept.lvmfit <- coefIntercept.lvm

## ** coefIntercept.multigroup
#' @rdname coefByType
#' @export
coefIntercept.multigroup <- coefIntercept.lvm

## * coefRef
#' @rdname coefByType
#' @export
`coefRef` <-
  function(object, value, ...) UseMethod("coefRef")

## ** coefRef.lvmfit
#' @rdname coefByType
#' @export
coefRef.lvmfit <- function(object, value = FALSE, ...){
    
    res <- retainType(type = attr(coefType(object, ...), "reference"),
                      validType = TRUE,
                      value = value)

    return(res)    
}

## * coefReg
#' @rdname coefByType
#' @export
`coefReg` <-
  function(object, value, ...) UseMethod("coefReg")

## ** coefReg.lvm
#' @rdname coefByType
#' @export
coefReg.lvm <- function(object, value = FALSE, ...){
    
     res <- retainType(type = coefType(object, ...),
                      validType = "regression",
                      value = value)

     return(res)
}

## ** coefReg.lvmfit
#' @rdname coefByType
#' @export
coefReg.lvmfit <- coefReg.lvm

## ** coefReg.multigroup
#' @rdname coefByType
#' @export
coefReg.multigroup <- coefReg.lvm

## * coefVar

#' @rdname coefByType
#' @export
`coefVar` <-
  function(object, value, ...) UseMethod("coefVar")

## ** coefVar.lvm
#' @rdname coefByType
#' @export
coefVar.lvm <- function(object, value = FALSE, ...){ 

    res <- retainType(type = coefType(object, ...),
                      validType = "variance",
                      value = value)

    return(res)
}

## ** coefVar.lvmfit
#' @rdname coefByType
#' @export
coefVar.lvmfit <- coefVar.lvm

## ** coefVar.multigroup
#' @rdname coefByType
#' @export
coefVar.multigroup <- coefVar.lvm

## * retainType  (needed for coefCov/Latent/Ref)
retainType <- function(type, validType, value){
  index.var <- which(type %in% validType)
  
  if(length(index.var)>0){
      if(value){
          return(names(type)[index.var])
      }else{
          return(index.var)
      }
  }else{
      return(NULL)
  }
}


##----------------------------------------------------------------------
### coefByType.R ends here

### test-coefType.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 12 2017 (14:52) 
## Version: 
## last-updated: Jan 12 2022 (12:32) 
##           By: Brice Ozenne
##     Update #: 82
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * header
## rm(list = ls())
if(TRUE){ ## already called in test-all.R
    library(testthat)
    library(lavaSearch2)
}

lava.options(symbols = c("~","~~"))

context("coefType")

## * linear regression
## ** only continous variables
m <- lvm(Y~X1+X2)
e <- estimate(m, lava::sim(m, 1e2))


df.truth <- data.frame(name = c("Y","Y~X1","Y~X2","Y~~Y"),
                       type = c("intercept","regression","regression","variance"),
                       letter = c("nu","K","K","Sigma_var"),
                       stringsAsFactors = FALSE)

test_that("coefType - lm", {

    test <- coefType(m, as.lava = TRUE)
    GS <- setNames(df.truth$type,df.truth$name)
    expect_equal(test,GS)

    test <- subset(coefType(m, as.lava = FALSE), !is.na(lava))
    expect_equal(as.character(test$detail),df.truth$letter)
    expect_equal(test$name,df.truth$name)

    test <- coefType(e, as.lava = TRUE)
    GS <- setNames(df.truth$type,df.truth$name)[names(test)]
    expect_equal(test,GS)

    test <- subset(coefType(e, as.lava = FALSE), !is.na(lava))
    expect_equal(as.character(test$detail),df.truth$letter)
    expect_equal(test$name,df.truth$name)

})

test_that("coefCov - lm", {
    expect_true(is.null(coefCov(m)))
    expect_true(is.null(coefCov(m, value = TRUE)))

    expect_equal(coefCov(m, keep.var = TRUE),4)
    expect_equal(coefCov(m, value = TRUE, keep.var = TRUE),"Y~~Y")
})

test_that("coefIndexModel - lm", {
    expect_true(all(coefIndexModel(m)==1))
    expect_true(all(coefIndexModel(e)==1))
})

test_that("coefIntercept - lm", {
    expect_equal(coefIntercept(m, keep.var = TRUE),1)
    expect_equal(coefIntercept(m, value = TRUE, keep.var = TRUE),"Y")
}) 

test_that("coefReg - lm", {
    expect_equal(coefReg(m, keep.var = TRUE),2:3)
    expect_equal(coefReg(m, value = TRUE, keep.var = TRUE),c("Y~X1","Y~X2"))
}) 

## ** 1 categorical variable
m <- lvm(Y~X1+X2+X3)
mSim <- m
categorical(mSim, labels =c("a","b","c")) <- ~X2
d <- lava::sim(mSim, 1e2)
e <- estimate(m, d)

df.truth <- data.frame(name = c("Y","Y~X1","Y~X2b","Y~X2c","Y~X3","Y~~Y"),
                       type = c("intercept","regression","regression","regression","regression","variance"),
                       letter = c("nu","K","K","K","K","Sigma_var"),
                       stringsAsFactors = FALSE
                       )

test_that("coefType - lm", {
    test <- coefType(mSim, as.lava = TRUE)
    expect_equal(names(test),as.character(coef(mSim)))

    test <- coefType(e, as.lava = TRUE)
    GS <- setNames(df.truth$type,df.truth$name)[names(coef(e))]
    expect_equal(test,GS)
    
    test <- subset(coefType(e, as.lava = FALSE), !is.na(lava))
    expect_equal(as.character(test$detail),df.truth$letter)
    expect_equal(test$name,df.truth$name)
})

## ** Several categorical variables
m <- lvm(Y1~X1+X2,Y2~X1+X3)
mSim <- m
categorical(mSim, labels = c("a","b","c")) <- "X1"
d <- lava::sim(mSim, 1e2)
e <- estimate(m, d)

df.truth <- data.frame(name = c("Y1","Y2",
                                "Y1~X1b","Y1~X1c","Y1~X2",
                                "Y2~X1b","Y2~X1c","Y2~X3",
                                "Y1~~Y1","Y2~~Y2"),
                       stringsAsFactors = FALSE)
df.truth$type <- c(rep("intercept",2),rep("regression",6),rep("variance",2))
df.truth$letter <- c(rep("nu",2),rep("K",6),rep("Sigma_var",2))

test_that("coefType - lm", {

    test <- coefType(mSim, as.lava = TRUE)
    expect_equal(names(test),as.character(coef(mSim)))

    test <- coefType(e, as.lava = TRUE)
    GS <- setNames(df.truth$type,df.truth$name)[names(coef(e))]
    expect_equal(test,GS)
    
    test <- subset(coefType(e, as.lava = FALSE), !is.na(lava))
    expect_equal(as.character(test$detail),df.truth$letter)
    expect_equal(test$name,df.truth$name)
})

## ** with extra variable

m.sim <- lvm(Y1 ~ 1)
categorical(m.sim, labels = c("control","concussion")) <- ~group

test_that("coefType - lm with extra variable", {
    expect_equal(unname(coefType(m.sim)), c("intercept", "variance", "extra"))    
})

## m.sim <- lvm(Y1 ~ 1, group ~ 1) ## ERROR
## categorical(m.sim, labels = c("control","concussion")) <- ~group
## lava:::categorical2dummy(m.sim, sim(m.sim, 1e2))$x$M
## no group concussion column

## * lvm
## ** categorical variable
m <- lvm(c(Y1,Y2,Y3)~eta1, c(Z1,Z2,Z3)~eta2)
regression(m) <- Y1~X1
regression(m) <- eta2~eta1+X2
covariance(m) <- Y1~Z1
latent(m) <- ~eta1+eta2
mSim <- m
categorical(mSim, labels =c("a","b","c")) <- ~X2
d <- lava::sim(mSim, 1e2, latent = FALSE)
e <- estimate(m, d)

df.truth <- rbind(c("Y2","intercept","nu"),
                  c("Y3","intercept","nu"),
                  c("Z2","intercept","nu"),
                  c("Z3","intercept","nu"),
                  ##
                  c("eta1","intercept","alpha"),
                  c("eta2","intercept","alpha"),
                  ##
                  c("Y1~X1","regression","K"),
                  ##
                  c("eta2~X2b","regression","Gamma"),
                  c("eta2~X2c","regression","Gamma"),
                  ##
                  c("Y2~eta1","regression","Lambda"),
                  c("Y3~eta1","regression","Lambda"),
                  c("Z2~eta2","regression","Lambda"),
                  c("Z3~eta2","regression","Lambda"),
                  ##
                  c("eta2~eta1","regression","B"),
                  ##
                  c("Y1~~Y1","variance","Sigma_var"),
                  c("Y2~~Y2","variance","Sigma_var"),
                  c("Y3~~Y3","variance","Sigma_var"),
                  c("Z1~~Z1","variance","Sigma_var"),
                  c("Z2~~Z2","variance","Sigma_var"),
                  c("Z3~~Z3","variance","Sigma_var"),
                  ##
                  c("Y1~~Z1","covariance","Sigma_cov"),
                  ##
                  c("eta1~~eta1","variance","Psi_var"),
                  c("eta2~~eta2","variance","Psi_var")
                  )

df.truth <- as.data.frame(df.truth, stringsAsFactors = FALSE)
names(df.truth) <- c("name","type","detail")

test_that("coefType - lvm", {

    test <- coefType(mSim, as.lava = TRUE)
    expect_equal(names(test),as.character(coef(mSim)))
    
    test <- coefType(e, as.lava = TRUE)
    GS <- setNames(df.truth$type,df.truth$name)[names(coef(e))]
    expect_equal(test,GS)

    test <- subset(coefType(e, as.lava = FALSE), !is.na(lava))
    expect_equal(as.character(test$detail),df.truth$detail)
    expect_equal(test$name,df.truth$name)
})
## ** constrains (0 mean 1 loading)

m <- lvm(c(Y1~0+1*eta1,Y2~0+1*eta1,Y3~0+1*eta1,
           Z1~0+1*eta2,Z2~0+1*eta2,Z3~0+1*eta2))
latent(m) <- ~eta1 + eta2
covariance(m) <- Y1~Y2

e <- estimate(m, lava::sim(m,1e2))


df.truth <- rbind(data.frame(name = "Y1", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Y2", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Y3", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z1", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z2", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z3", type = "intercept", detail = "nu", fixed = TRUE, stringsAsFactors = FALSE),
                  ##
                  data.frame(name = "eta1", type = "intercept", detail = "alpha", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "eta2", type = "intercept", detail = "alpha", fixed = FALSE, stringsAsFactors = FALSE),
                  ##
                  data.frame(name = "Y1~eta1", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Y2~eta1", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Y3~eta1", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z1~eta2", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z2~eta2", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  data.frame(name = "Z3~eta2", type = "regression", detail = "Lambda", fixed = TRUE, stringsAsFactors = FALSE),
                  ##
                  data.frame(name = "Y1~~Y1", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "Y2~~Y2", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "Y3~~Y3", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "Z1~~Z1", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "Z2~~Z2", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "Z3~~Z3", type = "variance", detail = "Sigma_var", fixed = FALSE, stringsAsFactors = FALSE),
                  ##
                  data.frame(name = "Y1~~Y2", type = "covariance", detail = "Sigma_cov", fixed = FALSE, stringsAsFactors = FALSE),
                  ##
                  data.frame(name = "eta1~~eta1", type = "variance", detail = "Psi_var", fixed = FALSE, stringsAsFactors = FALSE),
                  data.frame(name = "eta2~~eta2", type = "variance", detail = "Psi_var", fixed = FALSE, stringsAsFactors = FALSE)
                  )
df.truth <- as.data.frame(df.truth)

test_that("coefType - constrains 0/1", {

    test <- coefType(m, as.lava = TRUE)
    expect_equal(names(test),as.character(coef(m)))

    test <- coefType(e, as.lava = TRUE)
    GS <- setNames(df.truth[df.truth$fixed==FALSE,"type"],df.truth[df.truth$fixed==FALSE,"name"])[names(coef(e))]
    expect_equal(test,GS)
    
    test <- coefType(e, as.lava = FALSE)
    expect_equal(as.character(test$detail),df.truth$detail)
    expect_equal(test$name,df.truth$name)    
})


## ** constrains (common mean and variance)

m <- lvm(c(Y1[mu:sigma]~beta*eta1,Y2[mu:sigma]~1*eta1,Y3~1*eta1+X1),eta1~X1)
mSim <- m
categorical(mSim, labels = as.character(1:3)) <- ~X1

d <- lava::sim(mSim, latent=FALSE, 1e2)
e <- estimate(m, d)

test_that("coefType - constrains mean/variance", {
    test <- coefType(m, as.lava = TRUE)
    expect_equal(names(test),as.character(coef(m)))

    test <- coefType(e, as.lava = TRUE)
    #GS <- setNames(df.truth[fixed==FALSE,type],df.truth[fixed==FALSE,name])[names(coef(e))]
    #expect_equal(test,GS)
 
    coefType(e, as.lava = FALSE)
})

 
#----------------------------------------------------------------------
### test-coefType.R ends here

### checkData.R --- 
#----------------------------------------------------------------------
## author: Brice Ozenne
## created: okt 26 2017 (14:25) 
## Version: 
## last-updated: aug  7 2018 (10:54) 
##           By: Brice Ozenne
##     Update #: 38
#----------------------------------------------------------------------
## 
### Commentary: 
## 
### Change Log:
#----------------------------------------------------------------------
## 
### Code:

## * documentation - checkData
#' @title Check that Validity of the Dataset
#' @description Check whether the dataset can be used to fit the \code{lvm} object.
#' 
#' @name checkData
#' 
#' @param object a \code{lvm} object.
#' @param data [data.frame] the dataset used to obtain the object.
#' @param trace [logical] when \code{TRUE}, the outcome of the check will be displayed. 
#'
#' @return Invisible  \code{TRUE} or \code{FALSE}.
#' 
#' @examples 
#' m <- lvm()
#' regression(m) <- c(y1,y2,y3)~u
#' regression(m) <- u~x
#' latent(m) <- ~u
#'
#' d <- lava::sim(m,1e2)
#'
#' try(checkData(m, data = d)) # return an error
#' 
#' checkData(m, data = d[,-4])
#' 
#' try(checkData(m, data = d[,-(3:4)])) # return an error
#'
#' @concept diagnostic
#' @export
`checkData` <-
  function(object, data, trace) UseMethod("checkData")

## * checkData.lvm
#' @rdname checkData
#' @export
checkData.lvm <- function(object, data, trace = TRUE){

    ## ** normalize arguments
    data <- as.data.frame(data)
        
    ## ** check missing names
    vars <- lava::vars(object)
    latent <- lava::latent(object)    
    missingVars <- vars[vars %in% names(data) == FALSE]

    if(length(latent) == 0){
        
        if(length(missingVars)>0){
            if(trace){
                cat("Missing variable in data: ",paste(missingVars, collapse = " "),"\n", sep ="")
            }
            return(invisible(FALSE))
        }else{
            if(trace){
                cat("No issue detected \n")
            }
            return(invisible(TRUE))
        }
        
    }else{
        
        if(!identical(sort(latent),sort(missingVars))){
            if(trace){
                cat("Wrong specification of the latent variables \n",
                    "latent variables according to the LVM: ",paste(latent, collapse = " "),"\n",
                    "missing variables in data: ",paste(missingVars, collapse = " "),"\n", sep = "")
            }
            return(invisible(FALSE))
        }else{
            if(trace){
                cat("No issue detected \n")
            }
            return(invisible(TRUE))
        }
        
    }
}


#----------------------------------------------------------------------
### checkData.R ends here

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coefByType.R
\name{coefByType}
\alias{coefByType}
\alias{coefCov}
\alias{coefCov.lvm}
\alias{coefCov.lvmfit}
\alias{coefCov.multigroup}
\alias{coefExtra}
\alias{coefExtra.lvm}
\alias{coefExtra.lvmfit}
\alias{coefExtra.multigroup}
\alias{coefIndexModel}
\alias{coefIndexModel.lvm}
\alias{coefIndexModel.lvmfit}
\alias{coefIndexModel.multigroup}
\alias{coefIndexModel.multigroupfit}
\alias{coefIntercept}
\alias{coefIntercept.lvm}
\alias{coefIntercept.lvmfit}
\alias{coefIntercept.multigroup}
\alias{coefRef}
\alias{coefRef.lvmfit}
\alias{coefReg}
\alias{coefReg.lvm}
\alias{coefReg.lvmfit}
\alias{coefReg.multigroup}
\alias{coefVar}
\alias{coefVar.lvm}
\alias{coefVar.lvmfit}
\alias{coefVar.multigroup}
\title{Extract the Coefficient by Type}
\usage{
coefCov(object, value, keep.var, ...)

\method{coefCov}{lvm}(object, value = FALSE, keep.var = FALSE, ...)

\method{coefCov}{lvmfit}(object, value = FALSE, keep.var = FALSE, ...)

\method{coefCov}{multigroup}(object, value = FALSE, keep.var = FALSE, ...)

coefExtra(object, value, ...)

\method{coefExtra}{lvm}(object, value = FALSE, ...)

\method{coefExtra}{lvmfit}(object, value = FALSE, ...)

\method{coefExtra}{multigroup}(object, value = FALSE, ...)

coefIndexModel(object, ...)

\method{coefIndexModel}{lvm}(object, ...)

\method{coefIndexModel}{lvmfit}(object, ...)

\method{coefIndexModel}{multigroup}(object, ...)

\method{coefIndexModel}{multigroupfit}(object, ...)

coefIntercept(object, value, ...)

\method{coefIntercept}{lvm}(object, value = FALSE, ...)

\method{coefIntercept}{lvmfit}(object, value = FALSE, ...)

\method{coefIntercept}{multigroup}(object, value = FALSE, ...)

coefRef(object, value, ...)

\method{coefRef}{lvmfit}(object, value = FALSE, ...)

coefReg(object, value, ...)

\method{coefReg}{lvm}(object, value = FALSE, ...)

\method{coefReg}{lvmfit}(object, value = FALSE, ...)

\method{coefReg}{multigroup}(object, value = FALSE, ...)

coefVar(object, value, ...)

\method{coefVar}{lvm}(object, value = FALSE, ...)

\method{coefVar}{lvmfit}(object, value = FALSE, ...)

\method{coefVar}{multigroup}(object, value = FALSE, ...)
}
\arguments{
\item{object}{a lvm model or a fitted lvm model}

\item{value}{should the name of the coefficient be returned? Else return the coefficients}

\item{keep.var}{should the variance coefficients be returned?}

\item{...}{arguments to be passed to \code{}}
}
\value{
A vector containing the names of the positions of the coefficients.
}
\description{
Extract specific types of coefficient from a \code{lvm} object:
covariance coefficient(s) (\code{coefCov}),
extra parameter(s) (\code{coefExtra}),
position in the list of models for each coefficient (\code{coefIndexModel}),
intercept coefficient(s) (\code{coefIntercept}),
coefficient(s) that are used as reference (\code{coefRef}),
regression coefficient(s) (\code{coefReg}),
variance coefficient(s) (\code{coefVar}).
}
\examples{
#### regression ####
m <- lvm(Y~X1+X2)
e <- estimate(m, lava::sim(m, 1e2))

coefCov(m)
coefCov(m, value = TRUE)

coefCov(m, keep.var = TRUE)
coefCov(m, value = TRUE, keep.var = TRUE)

coefIndexModel(m)
coefIndexModel(e)

coefIntercept(m)
coefIntercept(m, value = TRUE)

coefReg(m)
coefReg(m, value = TRUE)

#### LVM ####
m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x1+x2
latent(m) <- ~u
covariance(m) <- y1~y2

m.Sim <- m
categorical(m.Sim, labels = c("a","b","c")) <- ~x2
e <- estimate(m, lava::sim(m.Sim, 1e2))

coefCov(m)
coefCov(m, value = TRUE) 

coefCov(m, keep.var = TRUE)
coefCov(m, value = TRUE, keep.var = TRUE)

coefExtra(m)

coefIndexModel(m)
coefIndexModel(e)

## additional categorical variable 
categorical(m, labels = as.character(1:3)) <- "X1"

coefExtra(m)
coefExtra(m, value = TRUE)

## additional categorical variable
categorical(m, labels = as.character(1:3)) <- "x1"

coefIntercept(m)
coefIntercept(m, value = TRUE)
coefIntercept(e)

coefReg(e, value = TRUE)

#### multigroup ####
m <- lvm(Y~X1+X2)
eG <- estimate(list(m,m), list(lava::sim(m, 1e2), lava::sim(m, 1e2)))

coefIndexModel(eG)

}
\concept{extractor}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare2.R
\name{dfSigmaRobust}
\alias{dfSigmaRobust}
\title{Degree of Freedom for the Robust Chi-Square Test}
\usage{
dfSigmaRobust(contrast, vcov, rvcov, score)
}
\arguments{
\item{contrast}{[numeric vector] the linear combination of parameters to test}

\item{vcov}{[numeric matrix] the variance-covariance matrix of the parameters.}

\item{rvcov}{[numeric matrix] the robust variance-covariance matrix of the parameters.}

\item{score}{[numeric matrix] the individual score for each parameter.}
}
\description{
Computation of the degrees of freedom of the chi-squared distribution
relative to the robust-based variance
}
\details{
When contrast is the identity matrix, this function compute the moments of the sandwich estimator
and the degrees of freedom of the approximate t-test as described in (Pan, 2002) section 2 and 3.1.
}
\references{
Wei Pan and Melanie M. Wall, Small-sample adjustments in using the sandwich variance estiamtor in generalized estimating equations. Statistics in medicine (2002) 21:1429-1441.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IntDensTri.R
\name{autoplot.intDensTri}
\alias{autoplot.intDensTri}
\title{2D-display of the Domain Used to Compute the Integral}
\usage{
\method{autoplot}{intDensTri}(object, coord.plot = c("x", "y1"), plot = TRUE, ...)
}
\arguments{
\item{object}{output of the function \code{intDensTri}.}

\item{coord.plot}{[character vector] the x and y coordinates. Can be \code{"x"}, \code{"y1"} to \code{"yd"}, \code{"z"} if \code{zmin} was specified when calling \code{intDensTri}.}

\item{plot}{[logical] should the plot be displayed?}

\item{...}{[internal] Only used by the generic method.}
}
\value{
A \code{ggplot} object.
}
\description{
2D-display of the domain used to compute the integral.
}
\seealso{
\code{\link{intDensTri}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils-nlme.R
\name{getVarCov2-internal}
\alias{getVarCov2-internal}
\alias{.getVarCov2}
\alias{.getVarCov2.gls}
\alias{.getVarCov2.lme}
\title{Reconstruct the Marginal Variance Covariance Matrix from a nlme Model}
\usage{
.getVarCov2(object, ...)

\method{.getVarCov2}{gls}(object, param, attr.param, name.endogenous,
  n.endogenous, ref.group, ...)

\method{.getVarCov2}{lme}(object, param, attr.param, ...)
}
\arguments{
\item{object}{a \code{gls} or \code{lme} object}

\item{...}{[internal] Only used by the generic method.}

\item{param}{[numeric vector] the mean and variance coefficients.}

\item{attr.param}{[character vector] the type of each coefficients (mean or variance).}

\item{name.endogenous}{[character vector] name of each repetition of the endogenous variable.}

\item{n.endogenous}{[integer >0] number of repetitions of the endogenous variable.}

\item{ref.group}{[character vector] the levels of the variable defining the variance component in a generic covariance matrix.}
}
\value{
[matrix] the marginal variance covariance matrix for a full sample.
}
\description{
Reconstruct the marginal variance covariance matrix from a nlme model.
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-extractData.R
\name{extractData}
\alias{extractData}
\alias{extractData.lvmfit}
\title{Extract Data From a Latent Variable Model}
\usage{
extractData(object, design.matrix, as.data.frame, envir, rm.na)

\method{extractData}{lvmfit}(
  object,
  design.matrix = FALSE,
  as.data.frame = TRUE,
  envir = environment(),
  rm.na = TRUE
)
}
\arguments{
\item{object}{the fitted model.}

\item{design.matrix}{[logical] should the data be extracted after transformation (e.g. conversion of categorical variables to dummy variables)?
Otherwise the original data will be returned.}

\item{as.data.frame}{[logical] should the output be converted into a \code{data.frame} object?}

\item{envir}{[environment] the environment from which to search the data.}

\item{rm.na}{[logical] should the lines containing missing values in the dataset be removed?}
}
\value{
a dataset.
}
\description{
Extract data from a latent variable model.
}
\examples{
#### simulate data ####
set.seed(10)
n <- 101

Y1 <- rnorm(n, mean = 0)
Y2 <- rnorm(n, mean = 0.3)
Id <- findInterval(runif(n), seq(0.1,1,0.1))
data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
           data.frame(Y=Y2,G="2",Id = Id)       
           )

#### latent variable model ####
library(lava)
e.lvm <- estimate(lvm(Y ~ G), data = data.df)
extractData(e.lvm)
extractData(e.lvm, design.matrix = TRUE)

}
\concept{extractor}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conditionalMoment.R
\name{conditionalMoment}
\alias{conditionalMoment}
\alias{conditionalMoment.lm}
\alias{conditionalMoment.gls}
\alias{conditionalMoment.lme}
\alias{conditionalMoment.lvm}
\alias{conditionalMoment.lvmfit}
\title{Prepare the Computation of score2}
\usage{
conditionalMoment(object, ...)

\method{conditionalMoment}{lm}(object, data, param, name.endogenous,
  first.order, second.order, ...)

\method{conditionalMoment}{gls}(object, data, formula, param, attr.param,
  ref.group, first.order, second.order, index.Omega, vec.OmegaMat, cluster,
  n.cluster, name.endogenous, n.endogenous, ...)

\method{conditionalMoment}{lme}(object, attr.param, ...)

\method{conditionalMoment}{lvm}(object, data, first.order, second.order,
  name.endogenous, name.latent, ...)

\method{conditionalMoment}{lvmfit}(object, data, param, first.order,
  second.order, usefit, ...)
}
\arguments{
\item{object, x}{a latent variable model.}

\item{...}{[internal] only used by the generic method or by the <- methods.}

\item{data}{[data.frame] data set.}

\item{param, p}{[numeric vector] the fitted coefficients.}

\item{name.endogenous}{[character vector, optional] name of the endogenous variables}

\item{second.order}{[logical] should the terms relative to the third derivative of the likelihood be be pre-computed?}

\item{formula}{[formula] two-sided linear formula.}

\item{attr.param}{[character vector] the type of each coefficient
(e.g. mean or variance coefficient).}

\item{ref.group}{[character vector] the levels of the variable defining the variance component in a generic covariance matrix.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{n.cluster}{[integer >0] the number of i.i.d. observations.}

\item{n.endogenous}{[integer >0] the number of outcomes.}

\item{name.latent}{[character vector, optional] name of the latent variables}

\item{usefit, value}{[logical] If TRUE the coefficients estimated by the model are used to pre-compute quantities. Only for lvmfit objects.}
}
\description{
Compute the conditional mean and variance,
and their first and second derivative regarding the model parameters.
}
\details{
For lvmfit objects, there are two levels of pre-computation:
\itemize{
\item a basic one that do no involve the model coefficient (\code{conditionalMoment.lvm}).
\item an advanced one that require the model coefficients (\code{conditionalMoment.lvmfit}). 
}
}
\examples{
m <- lvm(Y1~eta,Y2~eta,Y3~eta)
latent(m) <- ~eta

d <- lava::sim(m,1e2)
e <- estimate(m, d)

## basic pre-computation
res1 <- conditionalMoment(e, data = d,
                         first.order = FALSE, second.order = FALSE,
                         name.endogenous = endogenous(e),
                         name.latent = latent(e), usefit = FALSE)
res1$skeleton$Sigma

## full pre-computation
res2 <- conditionalMoment(e, param = coef(e), data = d,
                         first.order = FALSE, second.order = FALSE,
                         name.endogenous = endogenous(e),
                         name.latent = latent(e), usefit = TRUE
)
res2$value$Sigma

}
\concept{derivative of the score equation}
\concept{small sample inference}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkData.R
\name{checkData}
\alias{checkData}
\alias{checkData.lvm}
\title{Check that Validity of the Dataset}
\usage{
checkData(object, data, trace)

\method{checkData}{lvm}(object, data, trace = TRUE)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{data}{[data.frame] the dataset used to obtain the object.}

\item{trace}{[logical] when \code{TRUE}, the outcome of the check will be displayed.}
}
\value{
Invisible  \code{TRUE} or \code{FALSE}.
}
\description{
Check whether the dataset can be used to fit the \code{lvm} object.
}
\examples{
m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x
latent(m) <- ~u

d <- lava::sim(m,1e2)

try(checkData(m, data = d)) # return an error

checkData(m, data = d[,-4])

try(checkData(m, data = d[,-(3:4)])) # return an error

}
\concept{diagnostic}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrixPower.R
\name{matrixPower}
\alias{matrixPower}
\title{Power of a Matrix}
\usage{
matrixPower(object, power, symmetric, tol = 1e-12, print.warning = TRUE)
}
\arguments{
\item{object}{a matrix.}

\item{power}{[numeric] power to be applied to the matrix.}

\item{symmetric}{[logical] is the matrix symmetric? Argument passed to the function \code{eigen}.}

\item{tol}{[numeric >0] the threshold under which the eigenvalues are set to 0.}

\item{print.warning}{[logical] should a warning be print when some or the eigenvalues are not strictly positive.}
}
\value{
A matrix.
}
\description{
Compute the power of a matrix.
}
\examples{
## symmetric matrix
set.seed(10)
M <- matrix(rnorm(20*6),20,6)
Sigma <- var(M)
Sigma.half <- matrixPower(Sigma, power = 1/2, symmetric = TRUE)
round(Sigma.half \%*\% Sigma.half - Sigma,5)

iSigma <- matrixPower(Sigma, power = -1, symmetric = TRUE)
round(iSigma \%*\% Sigma,5)

iSigma.half <- matrixPower(Sigma, power = -1/2, symmetric = TRUE)
round(iSigma.half \%*\% iSigma.half - iSigma,5)

## non symmetric matrix
set.seed(10)
M <- matrix(abs(rnorm(9)), 3, 3) + diag(1,3,3)
M-t(M)

iM <- matrixPower(M, power = -1, symmetric = FALSE)
round(iM \%*\% M,5)

iM.half <- matrixPower(M, power = -1/2, symmetric = FALSE)
round(iM.half \%*\% iM.half \%*\% M,5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-glht2.R
\name{glht2}
\alias{glht2}
\alias{glht2.lvmfit}
\alias{glht2.lvmfit2}
\alias{glht2.mmm}
\alias{glht.lvmfit2}
\title{General Linear Hypothesis Testing With Small Sample Correction}
\usage{
glht2(object, ...)

\method{glht2}{lvmfit}(
  object,
  linfct,
  rhs = NULL,
  robust = FALSE,
  cluster = NULL,
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  ...
)

\method{glht2}{lvmfit2}(object, linfct, rhs = NULL, robust = FALSE, cluster = NULL, ...)

\method{glht2}{mmm}(object, linfct, rhs = 0, robust = FALSE, cluster = NULL, ...)

\method{glht}{lvmfit2}(model, linfct, rhs = NULL, robust = FALSE, cluster = NULL, ...)
}
\arguments{
\item{object, model}{a \code{lvmfit}, \code{lvmfit2}, or \code{mmm} object.}

\item{...}{[logical] arguments passed to lower level methods.}

\item{linfct}{[matrix or vector of character] the linear hypotheses to be tested. Same as the argument \code{par} of \code{\link{createContrast}}.}

\item{rhs}{[vector] the right hand side of the linear hypotheses to be tested.}

\item{robust}{[logical] should robust standard error be used? 
Otherwise rescale the influence function with the standard error obtained from the information matrix.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}
}
\value{
A \code{glht} object.
}
\description{
Test linear hypotheses on coefficients from a latent variable models with small sample corrections.
}
\details{
Whenever the argument linfct is not a matrix, it is passed to the function \code{createContrast} to generate the contrast matrix and, if not specified, rhs. \cr \cr

Since only one degree of freedom can be specify in a glht object and it must be an integer, the degree of freedom of the denominator of an F test simultaneously testing all hypotheses is retained, after rounding. \cr \cr

Argument rhs and null are equivalent.
This redondance enable compatibility between \code{lava::compare}, \code{compare2}, \code{multcomp::glht}, and \code{glht2}.
}
\examples{
library(multcomp)

## Simulate data
mSim <- lvm(c(Y1,Y2,Y3)~ beta * eta, Z1 ~ E, Z2 ~ E, Age[40:5]~1)
latent(mSim) <- "eta"
set.seed(10)
n <- 1e2

df.data <- lava::sim(mSim, n, latent = FALSE, p = c(beta = 1))

#### Inference on a single model ####
e.lvm <- estimate(lvm(Y1~E), data = df.data)
summary(glht2(e.lvm, linfct = c("Y1~E + Y1","Y1")))

#### Inference on separate models ####
## fit separate models
lvmX <- estimate(lvm(Z1 ~ E), data = df.data)
lvmY <- estimate(lvm(Z2 ~ E + Age), data = df.data)
lvmZ <- estimate(lvm(c(Y1,Y2,Y3) ~ eta, eta ~ E), 
                 data = df.data)

#### create mmm object #### 
e.mmm <- mmm(X = lvmX, Y = lvmY, Z = lvmZ)

#### create contrast matrix ####
resC <- createContrast(e.mmm, linfct = "E")

#### adjust for multiple comparisons ####
e.glht2 <- glht2(e.mmm, linfct = c(X="E"), df = FALSE)
summary(e.glht2)

}
\seealso{
\code{\link{createContrast}} to create contrast matrices. \cr
\code{\link{estimate2}} to pre-compute quantities for the small sample correction.
}
\concept{multiple comparison}
\concept{multiple comparisons}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.calibrateType1.R
\name{summary.calibrateType1}
\alias{summary.calibrateType1}
\title{Display the Type 1 Error Rate}
\usage{
\method{summary}{calibrateType1}(
  object,
  robust = FALSE,
  type = "type1error",
  alpha = 0.05,
  log.transform = TRUE,
  digits = 5,
  print = TRUE,
  ...
)
}
\arguments{
\item{object}{output of the \code{calibrateType1} function.}

\item{robust}{[character] should the results be displayed for both model-based and robust standard errors (\code{TRUE}),
only  model-based standard error (\code{FALSE}), or only robust standard error (\code{"only"})?}

\item{type}{[character] should the type 1 error rate be diplayed (\code{"type1error"}) or the bias (\code{"bias")}.}

\item{alpha}{[numeric, 0-1] the confidence levels.}

\item{log.transform}{[logical] should the confidence intervals be computed on the logit scale.}

\item{digits}{[integer >0] the number of decimal places to use when displaying the summary.}

\item{print}{should the summary be printed in the terminal.}

\item{...}{[internal] only used by the generic method.}
}
\description{
Display the type 1 error rate from the simulation results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2.R
\name{modelsearch2}
\alias{modelsearch2}
\alias{modelsearch2.lvmfit}
\title{Data-driven Extension of a Latent Variable Model}
\usage{
modelsearch2(
  object,
  link,
  data,
  method.p.adjust,
  method.maxdist,
  n.sample,
  na.omit,
  alpha,
  nStep,
  trace,
  cpus
)

\method{modelsearch2}{lvmfit}(
  object,
  link = NULL,
  data = NULL,
  method.p.adjust = "fastmax",
  method.maxdist = "approximate",
  n.sample = 1e+05,
  na.omit = TRUE,
  alpha = 0.05,
  nStep = NULL,
  trace = TRUE,
  cpus = 1
)
}
\arguments{
\item{object}{a \code{lvmfit} object.}

\item{link}{[character, optional for \code{lvmfit} objects] the name of the additional relationships to consider when expanding the model. Should be a vector containing strings like "Y~X". See the details section.}

\item{data}{[data.frame, optional] the dataset used to identify the model}

\item{method.p.adjust}{[character] the method used to adjust the p.values for multiple comparisons.
Can be any method that is valid for the \code{stats::p.adjust} function (e.g. \code{"fdr"}).
Can also be \code{"max"}, \code{"fastmax"}, or \code{"gof"}.}

\item{method.maxdist}{[character] the method used to estimate the distribution of the max statistic.
\code{"resampling"} resample the score under the null to estimate the null distribution.
\code{"bootstrap"} performs a wild bootstrap of the iid decomposition of the score to estimate the null distribution.
\code{"approximate"} attemps to identify the latent gaussian variable corresponding to each score statistic (that is chi-2 distributed).
It approximates the correlation matrix between these latent gaussian variables and uses numerical integration to compute the distribution of the max.}

\item{n.sample}{[integer, >0] number of samples used in the resampling approach.}

\item{na.omit}{should tests leading to NA for the test statistic be ignored. Otherwise this will stop the selection process.}

\item{alpha}{[numeric 0-1] the significance cutoff for the p-values.
When the p-value is below, the corresponding link will be added to the model
and the search will continue. Otherwise the search will stop.}

\item{nStep}{the maximum number of links that can be added to the model.}

\item{trace}{[logical] should the execution of the function be traced?}

\item{cpus}{the number of cpus that can be used for the computations.}
}
\value{
A list containing:
\itemize{
\item sequenceTest: the sequence of test that has been performed.
\item sequenceModel: the sequence of models that has been obtained.
\item sequenceQuantile: the sequence of rejection threshold. Optional. 
\item sequenceIID: the influence functions relative to each test. Optional. 
\item sequenceSigma: the covariance matrix relative to each test. Optional. 
\item initialModel: the model before the sequential search.
\item statistic: the argument \code{statistic}.
\item method.p.adjust: the argument \code{method.p.adjust}.
\item alpha: [numeric 0-1] the significance cutoff for the p-values.
\item cv: whether the procedure has converged.
}
}
\description{
Procedure adding relationship between variables that are supported by the data.
}
\details{
method.p.adjust = \code{"max"} computes the p-values based on the distribution of the max statistic.
This max statistic is the max of the square root of the score statistic.
The p-value are computed integrating the multivariate normal distribution.

method.p.adjust = \code{"fastmax"} only compute the p-value for the largest statistic.
It is faster than \code{"max"} and lead to identical results.

method.p.adjust = \code{"gof"} keep adding links until the chi-squared test (of correct specification of the covariance matrix) is no longer significant.
}
\examples{

## simulate data 
mSim <- lvm()
regression(mSim) <- c(y1,y2,y3,y4)~u
regression(mSim) <- u~x1+x2
categorical(mSim,labels=c("A","B","C")) <- "x2"
latent(mSim) <- ~u
covariance(mSim) <- y1~y2
transform(mSim, Id~u) <- function(x){1:NROW(x)}

set.seed(10)
df.data <- lava::sim(mSim, n = 1e2, latent = FALSE)

## only identifiable extensions
m <- lvm(c(y1,y2,y3,y4)~u)
latent(m) <- ~u
addvar(m) <- ~x1+x2

e <- estimate(m, df.data)

\dontrun{
resSearch <- modelsearch(e)
resSearch

resSearch2 <- modelsearch2(e, nStep = 2)
resSearch2
}
\dontshow{
search.link <- c("u~x1","u~x2","y1~x1","y1~x2","y1~~y2","y1~~y3")
resSearch2 <- modelsearch2(e, nStep = 2, link = search.link)
resSearch2
}

## some extensions are not identifiable
m <- lvm(c(y1,y2,y3)~u)
latent(m) <- ~u
addvar(m) <- ~x1+x2 

e <- estimate(m, df.data)

\dontrun{
resSearch <- modelsearch(e)
resSearch
resSearch2 <- modelsearch2(e)
resSearch2
}

## for instance
mNI <- lvm(c(y1,y2,y3)~u)
latent(mNI) <- ~u
covariance(mNI) <- y1~y2
## estimate(mNI, data = df.data)
## does not converge



}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-information2.R
\name{information2}
\alias{information2}
\alias{information2.lvmfit}
\alias{information2.lvmfit2}
\alias{information.lvmfit2}
\title{Expected Information With Small Sample Correction.}
\usage{
information2(object, as.lava, ssc, ...)

\method{information2}{lvmfit}(object, as.lava = TRUE, ssc = lava.options()$ssc, ...)

\method{information2}{lvmfit2}(object, as.lava = TRUE, ...)

\method{information}{lvmfit2}(x, ...)
}
\arguments{
\item{object, x}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{as.lava}{[logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}
}
\value{
A matrix with as many rows and columns as the number of coefficients.
}
\description{
Extract the expected information matrix from a latent variable model.
Similar to \code{lava::information} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the information matrix.
}
\examples{
#### simulate data ####
n <- 5e1
p <- 3
X.name <- paste0("X",1:p)
link.lvm <- paste0("Y~",X.name)
formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))

m <- lvm(formula.lvm)
distribution(m,~Id) <- Sequence.lvm(0)
set.seed(10)
d <- lava::sim(m,n)

#### linear models ####
e.lm <- lm(formula.lvm,data=d)

#### latent variable models ####
e.lvm <- estimate(lvm(formula.lvm),data=d)
information(e.lvm)
information2(e.lvm)
information2(e.lvm)[1:4,1:4] -  solve(vcov(e.lm))

}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createContrast.R
\name{createContrast}
\alias{createContrast}
\alias{createContrast.character}
\alias{createContrast.lvmfit}
\alias{createContrast.lvmfit2}
\alias{createContrast.list}
\alias{createContrast.mmm}
\title{Create Contrast matrix}
\usage{
createContrast(object, ...)

\method{createContrast}{character}(object, ...)

\method{createContrast}{lvmfit}(object, linfct, ...)

\method{createContrast}{lvmfit2}(object, linfct, ...)

\method{createContrast}{list}(object, linfct = NULL, ...)

\method{createContrast}{mmm}(object, linfct = NULL, ...)
}
\arguments{
\item{object}{a \code{lvmfit} object or a list of  a \code{lvmfit} objects.}

\item{...}{Argument to be passed to \code{.createContrast}:
\itemize{
\item diff.first [logical] should the contrasts between the first and any of the other coefficients define the null hypotheses.
\item add.rowname [logical] add rownames to the contrast matrix and names to the right-hand side.
\item rowname.rhs [logical] when naming the hypotheses, add the right-hand side (i.e. "X1-X2=0" instead of "X1-X2").
\item sep [character vector of length2] character surrounding the left part of the row names.
}}

\item{linfct}{[vector of characters] expression defining the linear hypotheses to be tested.
Can also be a regular expression (of length 1) that is used to identify the coefficients to be tested using \code{grep}.
See the examples section.}
}
\value{
A list containing
\itemize{
\item{contrast} [matrix] a contrast matrix corresponding to the left hand side of the linear hypotheses.
\item{null} [vector] the right hand side of the linear hypotheses.
\item{Q} [integer] the rank of the contrast matrix.
\item{ls.contrast} [list, optional] the contrast matrix corresponding to each submodel.
Only present when the \code{argument} object is a list of models.
}
}
\description{
Returns a contrast matrix corresponding an object.
The contrast matrix will contains the hypotheses in rows and the model coefficients in columns.
}
\details{
One can initialize an empty contrast matrix setting the argument\code{linfct} to \code{character(0)}. \cr \cr
}
\examples{
## Simulate data
mSim <- lvm(X ~ Age + Treatment,
            Y ~ Gender + Treatment,
            c(Z1,Z2,Z3) ~ eta, eta ~ treatment,
            Age[40:5]~1)
latent(mSim) <- ~eta
categorical(mSim, labels = c("placebo","SSRI")) <- ~Treatment
categorical(mSim, labels = c("male","female")) <- ~Gender
n <- 1e2
set.seed(10)
df.data <- lava::sim(mSim,n)

## Estimate separate models
lmX <- lava::estimate(lvm(X ~ -1 + Age + Treatment), data = df.data)
lmY <- lava::estimate(lvm(Y ~ -1 + Gender + Treatment), data = df.data)
lvmZ <- lava::estimate(lvm(c(Z1,Z2,Z3) ~ -1 + 1*eta, eta ~ -1 + Treatment), 
                 data = df.data)

## Contrast matrix for a given model
createContrast(lmX, linfct = "X~Age")
createContrast(lmX, linfct = c("X~Age=0","X~Age+5*X~TreatmentSSRI=0"))
createContrast(lmX, linfct = c("X~Age=0","X~Age+5*X~TreatmentSSRI=0"), sep = NULL)
createContrast(lmX, linfct = character(0))

## Contrast matrix for the join model
ls.lvm <- list(X = lmX, Y = lmY, Z = lvmZ)
createContrast(ls.lvm, linfct = "TreatmentSSRI=0")
createContrast(ls.lvm, linfct = "TreatmentSSRI=0", rowname.rhs = FALSE)
createContrast(ls.lvm, linfct = character(0))

## Contrast for multigroup models
m <- lava::lvm(Y~Age+Treatment)
e <- lava::estimate(list(m,m), data = split(df.data, df.data$Gender))
print(coef(e))
createContrast(e, linfct = "Y~TreatmentSSRI@1 - Y~TreatmentSSRI@2 = 0")
createContrast(e, linfct = "Y~TreatmentSSRI@2 - Y~TreatmentSSRI@1 = 0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calcType1postSelection.R
\name{calcType1postSelection}
\alias{calcType1postSelection}
\title{Compute the Type 1 Error After Selection [EXPERIMENTAL]}
\usage{
calcType1postSelection(
  level,
  mu,
  Sigma,
  quantile.previous,
  distribution,
  df,
  n = 10,
  correct = TRUE,
  ...
)
}
\arguments{
\item{level}{[numeric 0-1] expected coverage.}

\item{mu}{[numeric vector] the expectation of the joint distribution of the test statistics}

\item{Sigma}{[matrix] the variance-covariance of the joint distribution of the test statistics.}

\item{quantile.previous}{[numeric] significance quantile used at the previous step.}

\item{distribution}{[character] distribution of the test statistics.
Can be \code{"pmvnorm"} (normal distribution) or \code{"pvmt"} (Student's t distribution)}

\item{df}{[integer > 0] the degree of freedom of the joint Student's t distribution.
Only used when \code{distribution="pvmt"}.}

\item{n}{[integer > 0] number of points for the numerical integration}

\item{correct}{[logical] if true, correct the level to account for previous testings.}

\item{...}{arguments passed to \code{\link{intDensTri}}.}
}
\value{
[numeric] the type 1 error.
}
\description{
Compute the type 1 error after selection [EXPERIMENTAL].
}
\details{
The number of tests at the current step (i.e. after selection) is assumed to be
one less than the number of tests at the previous step (i.e. before selection).

Arguments \code{mu} and \code{Sigma} must contain the moments for the vector of test statistics
before and after selection (in that order).
}
\examples{
library(mvtnorm)
n <- 350

#### only 2 tests
Sigma <- rbind(c(1,0,0),c(0,1,1),c(0,1,1))
z2 <- qmvnorm(0.95, mean = rep(0,2), sigma = Sigma[1:2,1:2], tail = "both.tails")$quantile

## no selection since strong effect
mu <- c(10,0,0)
calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
                        mu = mu, Sigma = Sigma, correct = TRUE)

## strong selection
\dontrun{
mu <- c(0,0,0)
levelC <- calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
                        mu = mu, Sigma = Sigma)
print(levelC) # more liberal than without selection
calcType1postSelection(levelC, quantile.previous = z2, distribution = "gaussian",
                        mu = mu, Sigma = Sigma, correct = FALSE)
}

#### 3 tests
Sigma <- diag(1,5,5)
Sigma[4,2] <- 1
Sigma[2,4] <- 1
Sigma[5,3] <- 1
Sigma[3,5] <- 1

z2 <- qmvnorm(0.95, mean = mu[1:3], sigma = Sigma[1:3,1:3], tails = "both.tails")$quantile

## no selection since strong effect
\dontrun{
mu <- c(10,0,0,0,0)
calcType1postSelection(0.95, quantile.previous = z2, distribution = "gaussian",
                        mu = mu, Sigma = Sigma, correct = TRUE)

## strong selection
mu <- c(0,0,0,0,0)
levelC <- calcType1postSelection(0.95, quantile.previous = z2,
                        mu = mu, Sigma = Sigma, distribution = "gaussian")
calcType1postSelection(levelC, quantile.previous = z2, distribution = "gaussian",
                        mu = mu, Sigma = Sigma, correct = FALSE)
}

}
\author{
Brice Ozenne
}
\concept{post-selection inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-methods.R
\name{nStep}
\alias{nStep}
\alias{nStep.modelsearch2}
\title{Find the Number of Steps Performed During the Sequential Testing}
\usage{
nStep(object)

\method{nStep}{modelsearch2}(object)
}
\arguments{
\item{object}{a \code{modelsearch2} object.}
}
\value{
an integer.
}
\description{
Find the number of steps performed during the sequential testing.
}
\examples{
\dontrun{
mSim <- lvm(Y~G+X1+X2)
addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
df.data <- lava::sim(mSim, 1e2)

mBase <- lvm(Y~G)
addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
e.lvm <- estimate(mBase, data = df.data)
res <- modelsearch2(e.lvm, method.p.adjust = "holm")
nStep(res)
}

}
\concept{extractor}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-getVarCov2.R
\name{getVarCov2}
\alias{getVarCov2}
\alias{getVarCov2.lvmfit}
\alias{getVarCov2.lvmfit2}
\title{Residual Variance-Covariance Matrix With Small Sample Correction.}
\usage{
getVarCov2(object, ...)

\method{getVarCov2}{lvmfit}(object, ssc = lava.options()$ssc, ...)

\method{getVarCov2}{lvmfit2}(object, ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}
}
\value{
A matrix with as many rows and column as the number of endogenous variables
}
\description{
Reconstruct the residual variance-covariance matrix from a latent variable model. 
It is similar to \code{nlme::getVarCov} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the residuals.
}
\examples{
#### simulate data ####
set.seed(10)
n <- 101

Y1 <- rnorm(n, mean = 0)
Y2 <- rnorm(n, mean = 0.3)
Id <- findInterval(runif(n), seq(0.1,1,0.1))
data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
           data.frame(Y=Y2,G="2",Id = Id)
           )

#### latent variable models ####
library(lava)
e.lvm <- estimate(lvm(Y ~ G), data = data.df)
getVarCov2(e.lvm)

}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/initVarlink.R
\name{initVarLink}
\alias{initVarLink}
\alias{initVarLinks}
\title{Normalize var1 and var2}
\usage{
initVarLink(
  var1,
  var2,
  rep.var1 = FALSE,
  format = "list",
  Slink = c(lava.options()$symbols[1], "~"),
  Scov = lava.options()$symbols[2]
)

initVarLinks(var1, format = "list", ...)
}
\arguments{
\item{var1}{[character or formula] the exogenous variable of the new link or a formula describing the link.}

\item{var2}{[character] the endogenous variable of the new link.
Disregarded if the argument \code{var1} is a formula.}

\item{rep.var1}{[logical] should var1 be duplicated to match var2 length.
Only active if \code{format = "list"}.}

\item{format}{[character] should the name of the variable be returned (\code{format = "list"}),
a vector of character formula (\code{format = "txt.formula"}),
or a list of formula (\code{format = "formula"}).}

\item{Slink}{[character] the symbol for regression link.}

\item{Scov}{[character] the symbol for covariance link.}

\item{...}{argument to be passed to \code{initVarLink}.}
}
\value{
See argument \code{format}.
}
\description{
Convert var1 and var2 from formula or covariance to character.
}
\examples{
initVarLink(y ~ x1)
initVarLink("y ~ x1")
initVarLink(y ~ x1 + x2)
initVarLink("y ~ x1 + x2")
initVarLink(y ~ x1 + x2, rep.var1 = TRUE)
initVarLink(y ~ x1 + x2, rep.var1 = TRUE, format = "formula")
initVarLink(y ~ x1 + x2, rep.var1 = TRUE, format = "txt.formula")
initVarLink("y", "x1", format = "formula")

initVarLink("y ~ x1:0|1")

initVarLinks(y ~ x1)
initVarLinks("y ~ x1")
initVarLinks(c("y ~ x1","y~ x2"))
initVarLinks(c(y ~ x1,y ~ x2))
initVarLinks(c("y ~ x1","y ~ x2"), format = "formula")
initVarLinks(c(y ~ x1,y ~ x2), format = "formula")
initVarLinks(c("y ~ x1","y~ x2"), format = "txt.formula")
initVarLinks(c(y ~ x1,y ~ x2), format = "txt.formula")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-skeleton.R
\name{skeleton}
\alias{skeleton}
\alias{skeletonDtheta}
\alias{skeletonDtheta2}
\title{Pre-computation for the Score}
\usage{
skeleton(object, X, endogenous, latent, n.cluster, index.Omega)

skeletonDtheta(
  object,
  X,
  endogenous,
  latent,
  missing.pattern,
  unique.pattern,
  name.pattern,
  n.cluster,
  index.Omega
)

skeletonDtheta2(object)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{X}{[matrix] design matrix containing the covariates for each endogeneous and latent variable.}

\item{endogenous}{[character vector] the name of the endogeneous variables.}

\item{latent}{[character vector] the name of the latent variables.}
}
\description{
Pre-compute quantities that are necessary to compute the score of a lvm model.
}
\details{
When the user specifies names for the coefficients (e.g. Y1[mu:sigma]) or uses constraints (Y1~beta*X1), \code{as.lava=FALSE} will use the names specified by the user (e.g. mu, sigma, beta)
while \code{as.lava=TRUE} will use the name of the first link defining the coefficient.
}
\examples{
\dontrun{
skeleton <- lavaSearch2::skeleton
skeleton.lvm <- lavaSearch2::skeleton.lvm
skeleton.lvmfit <- lavaSearch2::skeleton.lvmfit

## without constrain
m <- lvm(Y1~X1+X2+eta,Y2~X3+eta,Y3~eta)
latent(m) <- ~eta

e <- estimate(m, lava::sim(m,1e2))
M.data <- as.matrix(model.frame(e))

skeleton(e$model, as.lava = TRUE,
         name.endogenous = endogenous(e), n.endogenous = 3,
         name.latent = latent(e), 
         update.value = FALSE)
skeleton(e, data = M.data, p = pars(e), as.lava = TRUE,
         name.endogenous = endogenous(e), n.endogenous = 3,
         name.latent = latent(e), 
         update.value = TRUE)

## with constrains
m <- lvm(Y[mu:sigma] ~ beta*X1+X2)
e <- estimate(m, lava::sim(m,1e2))
M.data <- as.matrix(model.frame(e))

skeleton(e$model, as.lava = TRUE,
         name.endogenous = "Y", n.endogenous = 1,
         name.latent = NULL, 
         update.value = FALSE)$skeleton

skeleton(e, data = M.data, p = pars(e), as.lava = FALSE,
         name.endogenous = "Y", n.endogenous = 1,
         name.latent = NULL, 
         update.value = FALSE)$skeleton

}
}
\concept{derivative of the score equation}
\concept{small sample inference}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Objective_gaussian_weight.R
\docType{data}
\name{gaussian_weight}
\alias{gaussian_weight}
\alias{gaussian_weight.estimate.hook}
\alias{gaussian_weight_method.lvm}
\alias{gaussian_weight_logLik.lvm}
\alias{gaussian_weight_objective.lvm}
\alias{gaussian_weight_score.lvm}
\alias{gaussian_weight_gradient.lvm}
\alias{gaussian_weight_hessian.lvm}
\title{Estimate LVM With Weights}
\format{
An object of class \code{character} of length 1.
}
\usage{
gaussian_weight.estimate.hook(x, data, estimator, ...)

gaussian_weight_method.lvm

gaussian_weight_logLik.lvm(object, type = "cond", p, data, weights, ...)

gaussian_weight_objective.lvm(x, ...)

gaussian_weight_score.lvm(
  x,
  data,
  p,
  S,
  n,
  mu = NULL,
  weights = NULL,
  debug = FALSE,
  reindex = FALSE,
  mean = TRUE,
  constrain = TRUE,
  indiv = FALSE,
  ...
)

gaussian_weight_gradient.lvm(...)

gaussian_weight_hessian.lvm(x, p, n, weights = NULL, ...)
}
\arguments{
\item{x, object}{A latent variable model}

\item{data}{dataset}

\item{estimator}{name of the estimator to be used}

\item{...}{passed to lower level functions.}

\item{type}{must be "cond"}

\item{p}{parameter value}

\item{weights}{weight associated to each iid replicate.}

\item{S}{empirical variance-covariance matrix between variable}

\item{n}{number of iid replicates}

\item{mu}{empirical mean}

\item{debug, reindex, mean, constrain, indiv}{additional arguments not used}
}
\description{
Estimate LVM with weights.
}
\examples{
#### linear regression with weights ####

## data
df <- data.frame(Y = c(1,2,2,1,2),
                 X = c(1,1,2,2,2),
                 missing = c(0,0,0,0,1),
                 weights = c(1,1,2,1,NA))

## using lm
e.lm.GS <- lm(Y~X, data = df)
e.lm.test <- lm(Y~X, data = df[df$missing==0,], weights = df[df$missing==0,"weights"])

## using lvm
m <- lvm(Y~X)
e.GS <- estimate(m, df)
## e.lava.test <- estimate(m, df[df$missing==0,], weights = df[df$missing==0,"weights"])
## warnings!!
e.test <- estimate(m, data = df[df$missing==0,],
                   weights = df[df$missing==0,"weights"],
                   estimator = "gaussian_weight")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoplot.calibrateType1.R
\name{autoplot_calibrateType1}
\alias{autoplot_calibrateType1}
\alias{autoplot.calibrateType1}
\title{Graphical Display of the Bias or Type 1 Error}
\usage{
\method{autoplot}{calibrateType1}(
  object,
  type = "bias",
  plot = TRUE,
  color.threshold = "red",
  type.bias = "absolute",
  alpha = 0.05,
  nrow.legend = NULL,
  name2label = NULL,
  color = NULL,
  keep.method = NULL,
  ...
)
}
\arguments{
\item{object}{output of \code{\link{calibrateType1}}.}

\item{type}{[character] if type equals \code{"bias"} the bias will be displayed.
Otherwise if it equals \code{"type1error"} the type 1 error will be displayed.}

\item{plot}{[logical] should the plot be displayed?}

\item{color.threshold}{[character] the color for the line representing the expected value(s).}

\item{type.bias}{[character] if type.bias equals \code{"absolute"} the absolute bias will be used.
Otherwise if it equals \code{"relative"} the relative bias will be used.
Only relevant when type equals \code{"bias"}.}

\item{alpha}{[numeric, 0-1] the significance threshold to consider.
Only relevant when type equals \code{"type1error"}.}

\item{nrow.legend}{[integer, >0] the number of rows for the legend.
Only relevant when type equals \code{"type1error"}.}

\item{name2label}{[named character vector] the label for the legend.
The vector should contain the method names (see details).
Only relevant when type equals \code{"type1error"}.}

\item{color}{[character vector] a vector of colours to be used to color the lines.
Only relevant when type equals \code{"type1error"}.}

\item{keep.method}{[character vector] the methods names for which the type 1 error should be displayed.
Only relevant when type equals \code{"type1error"}.}

\item{...}{[internal] Only used by the generic method.}
}
\value{
An list containing:
\itemize{
\item plot: a ggplot object.
\item data: the dataset used to generate the ggplot object.
}
}
\description{
Graphical display of the bias or type 1 error
for the output of \code{\link{calibrateType1}}.
}
\details{
Method names:
\itemize{
\item \code{p.Ztest}
\item \code{p.Satt}
\item \code{p.KR}
\item \code{p.robustZtest}
\item \code{p.robustSatt}
\item \code{p.robustKR}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/var2dummy.R
\name{var2dummy}
\alias{var2dummy}
\alias{var2dummy.list}
\alias{var2dummy.lvm}
\title{Convert Variable Names to Dummy Variables Names.}
\usage{
var2dummy(object, ...)

\method{var2dummy}{list}(object, var, rm.first.factor = TRUE, ...)

\method{var2dummy}{lvm}(object, data = NULL, ...)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{...}{[internal] additional arguments to be passed from \code{var2dummy.lvm} to \code{var2dummy.list}.}

\item{var}{[character] the variable to be transformed.}

\item{rm.first.factor}{[logical] should the first level of each categorical variable be ignored?}

\item{data}{[data.frame] dataset according to which the model should be updated.}
}
\description{
When dealing with categorical variables, the \code{estimate} function convert the categorical variables into dummy variables.
This function convert a set of variable names to their corresponding name in the model with dummy variables
}
\examples{

\dontrun{
var2dummy <- lavaSearch2:::var2dummy
var2dummy.list <- lavaSearch2:::var2dummy.list
var2dummy.lvm <- lavaSearch2:::var2dummy.lvm

m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u ~ X1+X2
var2dummy(m, var = c("X1","X2"))
categorical(m,labels=c("M","F","MF")) <- ~X1
var2dummy(m, var = c("X1","X2"))
categorical(m,labels=c("1","2","3")) <- ~X2
var2dummy(m, var = c("X1","X2"))
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-score2.R
\name{score2}
\alias{score2}
\alias{score2.lvmfit}
\alias{score2.lvmfit2}
\alias{score.lvmfit2}
\title{Score With Small Sample Correction}
\usage{
score2(object, indiv, cluster, as.lava, ...)

\method{score2}{lvmfit}(
  object,
  indiv = FALSE,
  cluster = NULL,
  as.lava = TRUE,
  ssc = lava.options()$ssc,
  ...
)

\method{score2}{lvmfit2}(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...)

\method{score}{lvmfit2}(x, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...)
}
\arguments{
\item{object, x}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{indiv}{[logical] If \code{TRUE}, the score relative to each observation is returned. Otherwise the total score is returned.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{as.lava}{[logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}
}
\value{
When argument indiv is \code{TRUE}, a matrix containing the score relative to each sample (in rows)
and each model coefficient (in columns). Otherwise a numeric vector of length the number of model coefficients.
}
\description{
Extract the (individual) score a the latent variable model.
Similar to \code{lava::score} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
}
\examples{
#### simulate data ####
n <- 5e1
p <- 3
X.name <- paste0("X",1:p)
link.lvm <- paste0("Y~",X.name)
formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))

m <- lvm(formula.lvm)
distribution(m,~Id) <- Sequence.lvm(0)
set.seed(10)
d <- lava::sim(m,n)

#### linear models ####
e.lm <- lm(Y~X1+X2+X3, data = d)

#### latent variable models ####
m.lvm <- lvm(formula.lvm)
e.lvm <- estimate(m.lvm,data=d)
e2.lvm <- estimate2(m.lvm,data=d)
score.tempo <- score(e2.lvm, indiv = TRUE)
colSums(score.tempo)

}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-residuals2.R
\name{residuals2}
\alias{residuals2}
\alias{residuals2.lvmfit}
\title{Residuals With Small Sample Correction.}
\usage{
residuals2(object, type, format, ssc, ...)

\method{residuals2}{lvmfit}(
  object,
  type = "response",
  format = "wide",
  ssc = lava.options()$ssc,
  ...
)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{type}{[character] the type of residual to extract:
\code{"response"} for raw residuals,
\code{"studentized"} for studentized residuals,
\code{"normalized"} for normalized residuals.}

\item{format}{[character] Use \code{"wide"} to return the residuals in the wide format (one row relative to each sample).
Otherwise use \code{"long"} to return the residuals in the long format.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}
}
\value{
a matrix containing the residuals relative to each sample (in rows)
and each endogenous variable (in column).
}
\description{
Extract residuals from a latent variable model.
Similar to \code{stats::residuals} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the residuals.

The raw residuals are defined by  observation minus the fitted value:
\deqn{
\varepsilon = (Y_1 - \mu_1, ..., Y_m - \mu_m)
}
The studentized residuals divided the raw residuals relative to each endogenous variable by the modeled variance of the endogenous variable.
\deqn{
\varepsilon_{stud} =(\frac{Y_1 - \mu_1}{\sigma_1}, ..., \frac{Y_m - \mu_m}{\sigma_m})
}
The normalized residuals multiply the raw residuals by the inverse of the square root of the modeled residual variance covariance matrix.
\deqn{
\varepsilon_{norm} = \varepsilon \Omega^{-1/2}
}
}
\examples{
#### simulate data ####
set.seed(10)
n <- 101

Y1 <- rnorm(n, mean = 0)
Y2 <- rnorm(n, mean = 0.3)
Id <- findInterval(runif(n), seq(0.1,1,0.1))
data.df <- rbind(data.frame(Y=Y1,G="1",Id = Id),
           data.frame(Y=Y2,G="2",Id = Id)
           )

#### latent variable models ####
library(lava)
e.lvm <- estimate(lvm(Y ~ G), data = data.df)
residuals(e.lvm)
residuals2(e.lvm)
residuals(e.lvm) - residuals2(e.lvm)

}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcov2.R
\name{vcov2}
\alias{vcov2}
\alias{vcov2.lm}
\alias{vcov2.gls}
\alias{vcov2.lme}
\alias{vcov2.lvmfit}
\alias{vcov2.lm2}
\alias{vcov2.gls2}
\alias{vcov2.lme2}
\alias{vcov2.lvmfit2}
\title{Extract the Variance Covariance Matrix of the Model Parameters}
\usage{
vcov2(object, ...)

\method{vcov2}{lm}(object, param = NULL, data = NULL, bias.correct = TRUE, ...)

\method{vcov2}{gls}(object, param = NULL, data = NULL, bias.correct = TRUE, ...)

\method{vcov2}{lme}(object, param = NULL, data = NULL, bias.correct = TRUE, ...)

\method{vcov2}{lvmfit}(object, param = NULL, data = NULL, bias.correct = TRUE, ...)

\method{vcov2}{lm2}(object, param = NULL, data = NULL, ...)

\method{vcov2}{gls2}(object, param = NULL, data = NULL, ...)

\method{vcov2}{lme2}(object, param = NULL, data = NULL, ...)

\method{vcov2}{lvmfit2}(object, param = NULL, data = NULL, ...)
}
\arguments{
\item{object}{a linear model or a latent variable model}

\item{...}{arguments to be passed to \code{sCorrect}.}

\item{param}{[optional] the fitted parameters.}

\item{data}{[optional] the data set.}

\item{bias.correct}{[logical] should the standard errors of the coefficients be corrected for small sample bias? Only relevant if the \code{sCorrect} function has not yet be applied to the object.}
}
\value{
A matrix.
}
\description{
Extract the variance covariance matrix of the model parameters from a Gaussian linear model.
}
\details{
If argument \code{p} or \code{data} is not null, then the small sample size correction is recomputed to correct the influence function.
}
\examples{
n <- 5e1
p <- 3
X.name <- paste0("X",1:p)
link.lvm <- paste0("Y~",X.name)
formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))

m <- lvm(formula.lvm)
distribution(m,~Id) <- Sequence.lvm(0)
set.seed(10)
d <- lava::sim(m,n)

## linear model
e.lm <- lm(formula.lvm,data=d)
vcov.tempo <- vcov2(e.lm, bias.correct = TRUE)
vcov.tempo[rownames(vcov(e.lm)),colnames(vcov(e.lm))]/vcov(e.lm)

## latent variable model
e.lvm <- estimate(lvm(formula.lvm),data=d)
vcov.tempo <- vcov2(e.lvm, bias.correct = FALSE)
vcov.tempo/vcov(e.lvm)

}
\seealso{
\code{\link{sCorrect}} to obtain \code{lm2}, \code{gls2}, \code{lme2}, or \code{lvmfit2} objects.
}
\concept{small sample inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-calcDistMax.R
\name{calcDistMax}
\alias{calcDistMax}
\alias{calcDistMaxIntegral}
\alias{calcDistMaxBootstrap}
\title{Adjust the p.values Using the Quantiles of the Max Statistic}
\usage{
calcDistMaxIntegral(
  statistic,
  iid,
  df,
  iid.previous = NULL,
  quantile.previous = NULL,
  quantile.compute = lava.options()$search.calc.quantile.int,
  alpha,
  cpus = 1,
  cl = NULL,
  trace
)

calcDistMaxBootstrap(
  statistic,
  iid,
  iid.previous = NULL,
  quantile.previous = NULL,
  method,
  alpha,
  cpus = 1,
  cl = NULL,
  n.sim,
  trace,
  n.repmax = 100
)
}
\arguments{
\item{statistic}{[numeric vector] the observed Wald statistic.
Each statistic correspond to a null hypothesis (i.e. a coefficient) that one wish to test.}

\item{iid}{[matrix] zero-mean iid decomposition of the coefficient used to compute the statistic.}

\item{df}{[numeric] the degree of freedom defining the multivariate Student's t distribution.
If \code{NULL} the multivariate Gaussian distribution will be used instead.}

\item{iid.previous}{[matrix, EXPERIMENTAL] zero-mean iid decomposition of previously tested coefficient.}

\item{quantile.previous}{[numeric, EXPERIMENTAL] rejection quantiles of the previously tested hypotheses. If not \code{NULL} the values should correspond the variable in to the first column(s) of the argument \code{iid.previous}.}

\item{quantile.compute}{[logical] should the rejection quantile be computed?}

\item{alpha}{[numeric 0-1] the significance cutoff for the p-values.
When the p-value is below, the corresponding link will be retained.}

\item{cpus}{[integer >0] the number of processors to use.
If greater than 1, the computation of the p-value relative to each test is performed in parallel.}

\item{cl}{[cluster] a parallel socket cluster generated by \code{parallel::makeCluster}
that has been registered using \code{registerDoParallel}.}

\item{trace}{[logical] should the execution of the function be traced?}

\item{method}{[character] the method used to compute the p-values.}

\item{n.sim}{[integer >0] the number of bootstrap simulations used to compute each p-values.
Disregarded when the p-values are computed using numerical integration.}

\item{n.repmax}{[integer >0] the maximum number of rejection for each bootstrap sample before switching to a new bootstrap sample.
Only relevant when conditioning on a previous test.
Disregarded when the p-values are computed using numerical integration.}
}
\value{
A list containing
\itemize{
\item p.adjust: the adjusted p-values.
\item z: the rejection threshold.
\item Sigma: the correlation matrix between the test statistic.
\item correctedLevel: the alpha level corrected for conditioning on previous tests.
}
}
\description{
Adjust the p.values using the quantiles of the max statistic.
}
\examples{
library(mvtnorm)

set.seed(10)
n <- 100
p <- 4
link <- letters[1:p]
n.sim <- 1e3 # number of bootstrap simulations 

#### test - not conditional ####
X.iid <- rmvnorm(n, mean = rep(0,p), sigma = diag(1,p))
colnames(X.iid) <- link
statistic <- setNames(1:p,link)


r1 <- calcDistMaxIntegral(statistic = statistic, iid = X.iid, 
            trace = FALSE, alpha = 0.05, df = 1e6) 

r3 <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
            method = "residual",
            trace = FALSE, alpha = 0.05, n.sim = n.sim)

r4 <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
            method = "wild",
            trace = FALSE, alpha = 0.05, n.sim = n.sim)

rbind(integration = c(r1$p.adjust, quantile = r1$z),
      bootResidual = c(r3$p.adjust, quantile = r3$z),
      bootWild    = c(r4$p.adjust, quantile = r4$z))

#### test - conditional ####
\dontrun{
Z.iid <- rmvnorm(n, mean = rep(0,p+1), sigma = diag(1,p+1))
seqQuantile <- qmvnorm(p = 0.95, delta = rep(0,p+1), sigma = diag(1,p+1), 
                    tail = "both.tails")$quantile

r1c <- calcDistMaxIntegral(statistic = statistic, iid = X.iid,
            iid.previous = Z.iid, quantile.previous =  seqQuantile, 
            trace = FALSE, alpha = 0.05, df = NULL)

r3c <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
            iid.previous = Z.iid, quantile.previous =  seqQuantile, method = "residual",
            trace = FALSE, alpha = 0.05, n.sim = n.sim)

r4c <- calcDistMaxBootstrap(statistic = statistic, iid = X.iid,
            iid.previous = Z.iid, quantile.previous =  seqQuantile, method = "wild",
            trace = FALSE, alpha = 0.05, n.sim = n.sim)

rbind(integration = c(r1c$p.adjust, quantile = r1c$z),
      bootResidual = c(r3c$p.adjust, quantile = r3c$z),
      bootWild    = c(r4c$p.adjust, quantile = r4c$z))
}
}
\concept{modelsearch}
\concept{post-selection inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils-nlme.R
\name{getIndexOmega2-internal}
\alias{getIndexOmega2-internal}
\alias{.getIndexOmega2}
\alias{.getIndexOmega2.gls}
\alias{.getIndexOmega2.lme}
\title{Extract the name of the endogenous variables}
\usage{
.getIndexOmega2(object, ...)

\method{.getIndexOmega2}{gls}(object, param, attr.param, name.Y, cluster,
  levels.cluster, data)

\method{.getIndexOmega2}{lme}(object, param, attr.param, name.Y, cluster,
  levels.cluster, data)
}
\arguments{
\item{object}{a \code{gls} or \code{lme} object.}

\item{...}{[internal] Only used by the generic method.}

\item{param}{[numeric vector] the mean and variance coefficients.}

\item{attr.param}{[character vector] the type of each coefficients (mean or variance).}

\item{name.Y}{[character] name of the endogenous variable.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.
Only required for \code{gls} models with no correlation argument.}

\item{data}{[data.frame] the data set.}
}
\value{
A list containing:
\itemize{
\item index.Omega: [list of integer vector] For each cluster of observations,
the index of the endogenous variable relative to each observation.
\item n.endogenous: [integer] the number of endogenous variables.
\item name.endogenous: [character vector] the name of the endogenous variables.
\item ref.group: [character vector] the levels of the variable defining the variance component in a generic covariance matrix.
}
}
\description{
Extract the name of the endogenous variables from a nlme model.
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/information2.R
\name{dInformation2-internal}
\alias{dInformation2-internal}
\alias{.dInformation2}
\title{Compute the First Derivative of the Expected Information Matrix}
\usage{
.dInformation2(
  dmu,
  d2mu,
  dOmega,
  d2Omega,
  Omega,
  OmegaM1,
  n.corrected,
  index.Omega,
  leverage,
  n.cluster,
  name.param,
  name.3deriv
)
}
\description{
Compute the first derivative of the expected information matrix.
}
\details{
\code{.dInformation2} will perform the computation individually when the
argument \code{index.Omega} is not null.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-hessian2.R
\name{hessian2}
\alias{hessian2}
\alias{hessian2.lvmfit}
\alias{hessian2.lvmfit2}
\title{Hessian With Small Sample Correction.}
\usage{
hessian2(object, indiv, cluster, as.lava, ...)

\method{hessian2}{lvmfit}(
  object,
  indiv = FALSE,
  cluster = NULL,
  as.lava = TRUE,
  ssc = lava.options()$ssc,
  ...
)

\method{hessian2}{lvmfit2}(object, indiv = FALSE, cluster = NULL, as.lava = TRUE, ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{indiv}{[logical] If \code{TRUE}, the hessian relative to each observation is returned. Otherwise the total hessian is returned.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{as.lava}{[logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}
}
\value{
An array containing the second derivative of the likelihood relative to each sample (dim 3)
and each pair of model coefficients (dim 1,2).
}
\description{
Extract the hessian from a latent variable model, with small sample correction
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the hessian.
}
\examples{
#### simulate data ####
n <- 5e1
p <- 3
X.name <- paste0("X",1:p)
link.lvm <- paste0("Y~",X.name)
formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))

m <- lvm(formula.lvm)
distribution(m,~Id) <- Sequence.lvm(0)
set.seed(10)
d <- lava::sim(m,n)

#### latent variable models ####
e.lvm <- estimate(lvm(formula.lvm),data=d)
hessian2(e.lvm)

}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{small sample inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lavaSearch2-package.R
\docType{package}
\name{lavaSearch2}
\alias{lavaSearch2}
\alias{lavaSearch2,}
\alias{lavaSearch2-package}
\title{Tools for Model Specification in the Latent Variable Framework}
\description{
The package contains three main functionalities:
\itemize{
\item \code{\link{compare2}}, \code{\link{summary2}}: Wald tests/robust Wald tests/F-tests/robust F-tests with improved control of the type 1 error in small samples.
\item \code{\link{glht2}}: adjustment for multiple comparisons when doing inference for multiple latent variable models.
\item \code{\link{modelsearch2}}: searching for local dependencies with adjustment for multiple comparisons.
}
It contains other useful functions such as:
\itemize{
\item \code{\link{calibrateType1}}: simulation study of the type 1 error of Wald tests.
\item \code{\link{createContrast}}: user-friendly function generating a contrast matrix.
\item \code{\link{getVarCov2}}: reconstruct the conditional variance covariance matrix.
\item \code{\link{iidJack}}: extract the jackknife iid decomposition.
}
}
\details{
The latent variable models (LVM) considered in this package can be written \cr
as a measurement model:
\deqn{Y_i = \nu + \eta_i \Lambda + X_i K + \epsilon_i}
and a structural model:
\deqn{\eta_i = \alpha + \eta_i B + X_i \Gamma + \zeta_i}
where \eqn{\Sigma} is the variance covariance matrix of the residuals \eqn{\epsilon}, \cr
and \eqn{\Psi}   is the variance covariance matrix of the residuals \eqn{\zeta}. \cr

The corresponding conditional mean is:
\deqn{
\mu_i(\theta) = E[Y_i|X_i] = \nu + (\alpha + X_i \Gamma) (1-B)^{-1} \Lambda + X_i K
}
\deqn{
\Omega(\theta) = Var[Y_i|X_i] = \Lambda^{t} (1-B)^{-t} \Psi (1-B)^{-1} \Lambda + \Sigma
}

The package aims to provides tool for testing linear hypotheses on the model coefficients
\eqn{\nu}, \eqn{\Lambda}, \eqn{K}, \eqn{\Sigma},
\eqn{\alpha}, \eqn{B}, \eqn{\Gamma}, \eqn{\Psi}.
Searching for local dependency enable to test whether the proposed model is too simplistic and if so to identify which additional coefficients should be added to the model.
}
\section{Limitations}{


'lavaSearch2' has been design for Gaussian latent variable models.
This means that it may not work / give valid results:
\itemize{
\item in presence of censored or binary outcomes.
\item with stratified models (i.e. object of class \code{multigroup}).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defineCategoricalLink.R
\name{defineCategoricalLink}
\alias{defineCategoricalLink}
\alias{defineCategoricalLink.lvm}
\title{Identify Categorical Links in LVM}
\usage{
defineCategoricalLink(object, link, data)

\method{defineCategoricalLink}{lvm}(object, link = NULL, data = NULL)
}
\arguments{
\item{object}{a \code{lvm} model.}

\item{link}{[character] the links to be analyzed. If \code{NULL}, all the coefficients from the lvm model are used instead.}

\item{data}{[data.frame] the dataset that will be used to fit the model. If \code{NULL}, a simulated data will be generated from the model.}
}
\value{
a \code{data.frame} with a description of each link in rows. \cr
The column factitious identify the links that will be replaced with other links
(e.g. "Y1~X1" becomes "Y1~X1b" and "Y1~X1c"). \cr
}
\description{
Identify categorical links in latent variable models.
}
\examples{
\dontrun{
defineCategoricalLink <- lavaSearch2:::defineCategoricalLink
defineCategoricalLink.lvm <- lavaSearch2:::defineCategoricalLink.lvm

## linear model
m <- lvm(Y1~X1+X2,Y2~X1+X3)
categorical(m, K = 3) <- "X1"
try(defineCategoricalLink(m)) # error

categorical(m, K = 3, labels = 1:3) <- "X1"
defineCategoricalLink(m)
defineCategoricalLink(m, "Y~X1")
defineCategoricalLink(m, "X1:0|1")
defineCategoricalLink(m, "X1:1|2")
defineCategoricalLink(m, c("X1:0|1", "X1:1|2"))
defineCategoricalLink(m, c("Y~X1","Y~X2"))
defineCategoricalLink(m, c("Y~X2","Y~X1"))

## latent variable model
m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x1+x2
latent(m) <- ~u
covariance(m) <- y1~y2
categorical(m, labels = as.character(1:3)) <- "X1"

defineCategoricalLink(m)
}

}
\concept{setter}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-effects2.R
\name{effects2}
\alias{effects2}
\alias{effects2.lvmfit}
\alias{effects2.lvmfit2}
\alias{effects.lvmfit2}
\title{Effects Through Pathways With Small Sample Correction}
\usage{
effects2(object, linfct, robust, cluster, conf.level, ...)

\method{effects2}{lvmfit}(
  object,
  linfct,
  robust = FALSE,
  cluster = NULL,
  conf.level = 0.95,
  to = NULL,
  from = NULL,
  df = lava.options()$df,
  ssc = lava.options()$ssc,
  ...
)

\method{effects2}{lvmfit2}(
  object,
  linfct,
  robust = FALSE,
  cluster = NULL,
  conf.level = 0.95,
  to = NULL,
  from = NULL,
  ...
)

\method{effects}{lvmfit2}(
  object,
  linfct,
  robust = FALSE,
  cluster = NULL,
  conf.level = 0.95,
  to = NULL,
  from = NULL,
  ...
)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{linfct}{[character vector] The path for which the effect should be assessed (e.g. \code{"A~B"}),
i.e. the effect of the right variable (B) on the left variable (A).}

\item{robust}{[logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{conf.level}{[numeric, 0-1] level of the confidence intervals.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{from, to}{alternative to argument \code{linfct}. See \code{lava::effects}.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}
}
\value{
A data.frame with a row per path.
}
\description{
Test whether a path in the latent variable model correspond to a null effect.
Similar to \code{lava::effects} but with small sample correction (if any).
So far it only work for a single path related two variable composed of one or two edges.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
}
\concept{inference}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-compare2.R
\name{dfSigma}
\alias{dfSigma}
\title{Degree of Freedom for the Chi-Square Test}
\usage{
dfSigma(contrast, score, vcov, rvcov, dVcov, dRvcov, keep.param, type)
}
\arguments{
\item{contrast}{[numeric vector] the linear combination of parameters to test}

\item{score}{[numeric matrix] the individual score for each parameter.}

\item{vcov}{[numeric matrix] the model-based variance-covariance matrix of the parameters.}

\item{rvcov}{[numeric matrix] the robust variance-covariance matrix of the parameters.}

\item{dVcov}{[numeric array] the first derivative of the model-based variance-covariance matrix of the parameters.}

\item{dRvcov}{[numeric array] the first derivative of the robust variance-covariance matrix of the parameters.}

\item{keep.param}{[character vector] the name of the parameters with non-zero first derivative of their variance parameter.}

\item{type}{[integer] 1 corresponds to the Satterthwaite approximation of the the degrees of freedom applied to the model-based variance,
2 to the Satterthwaite approximation of the the degrees of freedom applied to the robust variance,
3 to the approximation described in (Pan, 2002) section 2 and 3.1.}
}
\description{
Computation of the degrees of freedom of the chi-squared distribution
relative to the model-based variance
}
\references{
Wei Pan and Melanie M. Wall, Small-sample adjustments in using the sandwich variance estiamtor in generalized estimating equations. Statistics in medicine (2002) 21:1429-1441.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{selectResponse}
\alias{selectResponse}
\alias{selectResponse.formula}
\title{Response Variable of a Formula}
\usage{
selectResponse(object, ...)

\method{selectResponse}{formula}(object, format = "call", ...)
}
\arguments{
\item{object}{a formula}

\item{...}{[internal] Only used by the generic method.}

\item{format}{[character] should an object of type call be returned (\code{format = "call"}),
or the names of the variables (\code{format = "vars"})}
}
\value{
See argument \code{format}.
}
\description{
Return the response variable contained in the formula.
}
\examples{

\dontrun{

selectResponse <- lavaSearch2:::selectResponse
selectResponse.formula <- lavaSearch2:::selectResponse.formula

selectResponse(Y1~X1+X2)
selectResponse(Y1~X1+X2, format = "vars")
selectResponse(Surv(event,time)~X1+X2, format = "vars")

selectResponse(Y1~X1+Y1)
selectResponse(Y1+Y2~X1+Y1, format = "vars")

selectResponse(~X1+X2)
selectResponse(~X1+X2, format = "vars")
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-dInformation2.R
\name{.dinformation2-internal}
\alias{.dinformation2-internal}
\alias{.dInformation2}
\title{Compute the First Derivative of the Expected Information Matrix}
\usage{
.dInformation2(
  dmu,
  dOmega,
  d2mu,
  d2Omega,
  OmegaM1,
  missing.pattern,
  unique.pattern,
  name.pattern,
  grid.3varD1,
  grid.2meanD1.1varD1,
  grid.2meanD2.1meanD1,
  grid.2varD2.1varD1,
  name.param,
  leverage,
  n.cluster,
  weights
)
}
\description{
Compute the first derivative of the expected information matrix.
}
\details{
\code{calc_dinformation} will perform the computation individually when the
argument \code{index.Omega} is not null.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibrateType1.R
\name{calibrateType1}
\alias{calibrateType1}
\alias{calibrateType1.lvm}
\alias{calibrateType1.lvmfit}
\title{Simulation Study Assessing Bias and Type 1 Error}
\usage{
calibrateType1(object, param, n.rep, ...)

\method{calibrateType1}{lvm}(
  object,
  param,
  n.rep,
  n,
  correction = TRUE,
  warmup = NULL,
  null = NULL,
  F.test = FALSE,
  cluster = NULL,
  generative.object = NULL,
  generative.coef = NULL,
  true.coef = NULL,
  n.true = 1e+06,
  round.true = 2,
  bootstrap = FALSE,
  n.bootstrap = 1000,
  checkType1 = FALSE,
  checkType2 = FALSE,
  dir.save = NULL,
  label.file = NULL,
  seed = NULL,
  cpus = 1,
  trace = 2,
  ...
)

\method{calibrateType1}{lvmfit}(
  object,
  param,
  n.rep,
  correction = TRUE,
  F.test = FALSE,
  bootstrap = FALSE,
  n.bootstrap = 1000,
  seed = NULL,
  trace = 2,
  cpus = 1,
  ...
)
}
\arguments{
\item{object}{a \code{lvm} object defining the model to be fitted.}

\item{param}{[character vector] names of the coefficient whose value will be tested.}

\item{n.rep}{[integer, >0] number of simulations per sample size.}

\item{...}{[internal] Only used by the generic method.}

\item{n}{[integer vector, >0] sample size(s) considered in the simulation study.}

\item{correction}{[logical] should the type 1 error after correction be computed?}

\item{warmup}{[list of lvm] a list of \code{lvm} objects that will be sequentially fitted with for
starting values the parameter of the previous model in the list (if any). The parameters of the final
model of the list are used to initialize the fit of the model of interest (i.e. object).}

\item{null}{[numeric vector] vector of null hypotheses, one for each model coefficient.
By default a vector of 0.}

\item{F.test}{[logical] should a multivariate Wald test be perform testing simultaneously all the null hypotheses?}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.
Will be passed to \code{lava::estimate}.}

\item{generative.object}{[lvm] object defining the statistical model generating the data.}

\item{generative.coef}{[name numeric vector] values for the parameters of the generative model.
Can also be \code{NULL}: in such a case the coefficients are set to default values decided by lava (usually 0 or 1).}

\item{true.coef}{[name numeric vector] expected values for the parameters of the fitted model.}

\item{n.true}{[integer, >0] sample size at which the estimated coefficients will be a reliable approximation of the true coefficients.}

\item{round.true}{[integer, >0] the number of decimal places to be used for the true value of the coefficients. No rounding is done if \code{NULL}.}

\item{bootstrap}{[logical] should bootstrap resampling be performed?}

\item{n.bootstrap}{[integer, >0] the number of bootstrap sample to be used for each bootstrap.}

\item{checkType1}{[logical] returns an error if the coefficients associated to the null hypotheses do not equal 0.}

\item{checkType2}{[logical] returns an error if the coefficients associated to the null hypotheses equal 0.}

\item{dir.save}{[character] path to the directory were the results should be exported.
Can also be \code{NULL}: in such a case the results are not exported.}

\item{label.file}{[character] element to include in the file name.}

\item{seed}{[integer, >0] value that will be set before adjustment for multiple comparisons to ensure reproducible results.
Can also be \code{NULL}: in such a case no seed is set.}

\item{cpus}{[integer >0] the number of processors to use.
If greater than 1, the simulations are performed in parallel.}

\item{trace}{[integer] should the execution of the function be trace. Can be 0, 1 or 2.}
}
\value{
An object of class \code{calibrateType1}.
}
\description{
Perform a simulation study over one or several sample size
to assess the bias of the estimate
and the type 1 error of the Wald test and robust Wald test
}
\examples{
\dontrun{
#### simulate data ####
m.Sim <- lvm(c(Y1[mu1:sigma]~1*eta,
               Y2[mu2:sigma]~1*eta,
               Y3[mu3:sigma]~1*eta,
               eta~beta1*Group+beta2*Gender))
latent(m.Sim) <- ~eta
categorical(m.Sim, labels = c("M","F")) <- ~Gender

d <- lava::sim(m.Sim, 1e2)

#### calibrate type 1 error on the estimated model ####
m <- lvm(Y1~eta,
         Y2~eta,
         Y3~eta,
         eta~Group+Gender)
e <- lava::estimate(m, data = d)
res <- calibrateType1(e, param = "eta~Group", n.rep = 100)
res <- calibrateType1(e, param = c("eta~Group","Y1~eta"), F.test = TRUE, n.rep = 100)
res <- calibrateType1(e, param = "eta~Group", n.rep = 100, cpus = 4)
summary(res)
}

}
\seealso{
\code{link{autoplot.calibrateType1}} for a graphical display of the bias or of the type 1 error.
}
\author{
Brice Ozenne
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{addLink}
\alias{addLink}
\alias{addLink.lvm}
\alias{addLink.lvm.reduced}
\title{Add a New Link Between Two Variables in a LVM}
\usage{
addLink(object, ...)

\method{addLink}{lvm}(
  object,
  var1,
  var2,
  covariance,
  all.vars = lava::vars(object),
  warnings = FALSE,
  ...
)

\method{addLink}{lvm.reduced}(object, ...)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{...}{[internal] only used by the generic method and from \code{addLink.lvm.reduced} to \code{addLink.lvm}.}

\item{var1}{[character or formula] the exogenous variable of the new link or a formula describing the link to be added to the lvm.}

\item{var2}{[character] the endogenous variable of the new link. Disregarded if the argument \code{var1} is a formula.}

\item{covariance}{[logical] is the link is bidirectional? Ignored if one of the variables non-stochastic (e.g. exogenous variables).}

\item{all.vars}{[internal] a character vector containing all the variables of the \code{lvm} object.}

\item{warnings}{[logical] Should a warning be displayed when no link is added?}
}
\description{
Generic interface to add links to \code{lvm} objects.
}
\details{
The argument \code{all.vars} is useful for \code{lvm.reduce} object where the command \code{vars(object)} does not return all variables. The command \code{vars(object, xlp = TRUE)} must be used instead.

Arguments \code{var1} and \code{var2} are passed to \code{initVarlink}.
}
\examples{
library(lava)
set.seed(10)

m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x1+x2
latent(m) <- ~u
m2 <- m

addLink(m, x1 ~ y1, covariance = FALSE)
addLink(m, y1 ~ x1, covariance = FALSE)
coef(addLink(m, y1 ~ y2, covariance = TRUE))

addLink(m2, "x1", "y1", covariance = FALSE)
addLink(m2, "y1", "x1", covariance = FALSE)
newM <- addLink(m, "y1", "y2", covariance = TRUE)
coef(newM)

}
\concept{setter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{setLink}
\alias{setLink}
\alias{setLink.lvm}
\title{Set a Link to a Value}
\usage{
setLink(object, ...)

\method{setLink}{lvm}(object, var1, var2, value, warnings = FALSE, ...)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{...}{[internal] only used by the generic method.}

\item{var1}{[character or formula] the exogenous variable of the new link or a formula describing the link to be added to the lvm.}

\item{var2}{[character] the endogenous variable of the new link. Disregarded if the argument \code{var1} is a formula.}

\item{value}{[numeric] the value at which the link should be set.}

\item{warnings}{[logical] should a warning be displayed if the link is not found in the \code{lvm} object.}
}
\description{
Generic interface to set a value to a link in a \code{lvm} object.
}
\examples{
library(lava)
set.seed(10)

m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x1+x2
latent(m) <- ~u
covariance(m) <- y1 ~ y2

m1 <- setLink(m, y3 ~ u, value = 1)
estimate(m1, lava::sim(m,1e2))
# m1 <- setLink(m, u ~ y3, value = 1)

m2 <- setLink(m, y1 ~ y2, value = 0.5)
estimate(m2, lava::sim(m,1e2))

}
\concept{setter}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-compare2.R
\name{compare2}
\alias{compare2}
\alias{compare2.lvmfit}
\alias{compare2.lvmfit2}
\alias{compare.lvmfit2}
\title{Test Linear Hypotheses With Small Sample Correction}
\usage{
compare2(
  object,
  linfct,
  rhs,
  robust,
  cluster,
  as.lava,
  F.test,
  conf.level,
  ...
)

\method{compare2}{lvmfit}(
  object,
  linfct = NULL,
  rhs = NULL,
  robust = FALSE,
  cluster = NULL,
  as.lava = TRUE,
  F.test = TRUE,
  conf.level = 0.95,
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  ...
)

\method{compare2}{lvmfit2}(
  object,
  linfct = NULL,
  rhs = NULL,
  robust = FALSE,
  cluster = NULL,
  as.lava = TRUE,
  F.test = TRUE,
  conf.level = 0.95,
  ...
)

\method{compare}{lvmfit2}(
  object,
  linfct = NULL,
  rhs = NULL,
  robust = FALSE,
  cluster = NULL,
  as.lava = TRUE,
  F.test = TRUE,
  conf.level = 0.95,
  ...
)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{linfct}{[matrix or vector of character] the linear hypotheses to be tested. Same as the argument \code{par} of \code{\link{createContrast}}.}

\item{rhs}{[vector] the right hand side of the linear hypotheses to be tested.}

\item{robust}{[logical] should the robust standard errors be used instead of the model based standard errors?}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{as.lava}{[logical] should the output be similar to the one return by \code{lava::compare}?}

\item{F.test}{[logical] should a joint test be performed?}

\item{conf.level}{[numeric 0-1] level of the confidence intervals.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}
}
\value{
If \code{as.lava=TRUE} an object of class \code{htest}.
Otherwise a \code{data.frame} object.
}
\description{
Test Linear Hypotheses using Wald statistics in a latent variable model.
Similar to \code{lava::compare} but with small sample correction.
}
\details{
The \code{linfct} argument and \code{rhs} specify the set of linear hypotheses to be tested. They can be written:
\deqn{
  linfct * \theta = rhs
}
where \eqn{\theta} is the vector of the model coefficients. \cr
The \code{par} argument must contain expression(s) involving the model coefficients.
For example \code{"beta = 0"} or \code{c("-5*beta + alpha = 3","-alpha")} are valid expressions if alpha and beta belong to the set of model coefficients.
A contrast matrix and the right hand side will be generated inside the function. \cr

When directly specified, the contrast matrix must contain as many columns as there are coefficients in the model (mean and variance coefficients).
Each hypothesis correspond to a row in the contrast matrix. \cr

The rhs vector should contain as many elements as there are row in the contrast matrix. \cr
}
\examples{
#### simulate data ####
set.seed(10)
mSim <- lvm(Y~0.1*X1+0.2*X2)
categorical(mSim, labels = c("a","b","c")) <- ~X1
transform(mSim, Id~Y) <- function(x){1:NROW(x)}
df.data <- lava::sim(mSim, 1e2)

#### with lvm ####
m <- lvm(Y~X1+X2)
e.lvm <- estimate(m, df.data)

compare2(e.lvm, linfct = c("Y~X1b","Y~X1c","Y~X2"))
compare2(e.lvm, linfct = c("Y~X1b","Y~X1c","Y~X2"), robust = TRUE)

}
\seealso{
\code{\link{createContrast}} to create contrast matrices. \cr
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{inference}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/iidJack.R
\name{iidJack}
\alias{iidJack}
\alias{iidJack.default}
\title{Jackknife iid Decomposition from Model Object}
\usage{
iidJack(object, ...)

\method{iidJack}{default}(
  object,
  data = NULL,
  grouping = NULL,
  cpus = 1,
  keep.warnings = TRUE,
  keep.error = TRUE,
  cl = NULL,
  trace = TRUE,
  ...
)
}
\arguments{
\item{object}{a object containing the model.}

\item{...}{[internal] only used by the generic method.}

\item{data}{[data.frame] dataset used to perform the jackknife.}

\item{grouping}{[vector] variable defining cluster of observations that will be simultaneously removed by the jackknife.}

\item{cpus}{[integer >0] the number of processors to use.
If greater than 1, the fit of the model and the computation of the influence function for each jackknife sample is performed in parallel.}

\item{keep.warnings}{[logical] keep warning messages obtained when estimating the model with the jackknife samples.}

\item{keep.error}{[logical]keep error messages obtained when estimating the model with the jackknife samples.}

\item{cl}{[cluster] a parallel socket cluster generated by \code{parallel::makeCluster}
that has been registered using \code{registerDoParallel}.}

\item{trace}{[logical] should a progress bar be used to trace the execution of the function}
}
\value{
A matrix with in row the samples and in columns the parameters.
}
\description{
Extract iid decomposition (i.e. influence function) from model object.
}
\examples{
n <- 20
set.seed(10)
mSim <- lvm(c(Y1,Y2,Y3,Y4,Y5) ~ 1*eta)
latent(mSim) <- ~eta
categorical(mSim, K=2) <- ~G
transform(mSim, Id ~ eta) <- function(x){1:NROW(x)}
dW <- lava::sim(mSim, n, latent = FALSE)

#### LVM ####
\dontrun{
m1 <- lvm(c(Y1,Y2,Y3,Y4,Y5) ~ 1*eta)
latent(m1) <- ~eta
regression(m1) <- eta ~ G
e <- estimate(m1, data = dW)

iid1 <- iidJack(e)
iid2 <- iid(e)
attr(iid2, "bread") <- NULL

apply(iid1,2,sd)
apply(iid2,2,sd)
quantile(iid2 - iid1)
}


}
\concept{iid decomposition}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-information2.R
\name{information2-internal}
\alias{information2-internal}
\alias{.information2}
\title{Compute the Expected Information Matrix From the Conditional Moments}
\usage{
.information2(
  dmu,
  dOmega,
  OmegaM1,
  missing.pattern,
  unique.pattern,
  name.pattern,
  grid.mean,
  grid.var,
  name.param,
  leverage,
  weights = NULL,
  n.cluster
)
}
\description{
Compute the expected information matrix from the conditional moments.
}
\details{
\code{calc_information} will perform the computation individually when the
argument \code{index.Omega} is not null.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-iid2.R
\name{iid2}
\alias{iid2}
\alias{iid2.lvmfit}
\alias{iid2.lvmfit2}
\alias{iid.lvmfit2}
\title{Influence Function With Small Sample Correction.}
\usage{
iid2(object, ...)

\method{iid2}{lvmfit}(
  object,
  robust = TRUE,
  cluster = NULL,
  as.lava = TRUE,
  ssc = lava.options()$ssc,
  ...
)

\method{iid2}{lvmfit2}(object, robust = TRUE, cluster = NULL, as.lava = TRUE, ...)

\method{iid}{lvmfit2}(x, robust = TRUE, cluster = NULL, as.lava = TRUE, ...)
}
\arguments{
\item{object, x}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{robust}{[logical] if \code{FALSE}, the influence function is rescaled such its the squared sum equals the model-based standard error (instead of the robust standard error).
Do not match the model-based correlation though.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{as.lava}{[logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients (\code{"none"}, \code{"residual"}, \code{"cox"}). Only relevant when using a \code{lvmfit} object.}
}
\value{
A matrix containing the 1st order influence function relative to each sample (in rows)
and each model coefficient (in columns).
}
\description{
Extract the influence function from a latent variable model.
It is similar to \code{lava::iid} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the variance-covariance matrix.
}
\examples{
#### simulate data ####
n <- 5e1
p <- 3
X.name <- paste0("X",1:p)
link.lvm <- paste0("Y~",X.name)
formula.lvm <- as.formula(paste0("Y~",paste0(X.name,collapse="+")))

m <- lvm(formula.lvm)
distribution(m,~Id) <- Sequence.lvm(0)
set.seed(10)
d <- sim(m,n)

#### latent variable model ####
e.lvm <- estimate(lvm(formula.lvm),data=d)
iid.tempo <- iid2(e.lvm)


}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tryWithWarnings.R
\name{tryWithWarnings}
\alias{tryWithWarnings}
\title{Run an Expression and Catch Warnings and Errors}
\usage{
tryWithWarnings(expr)
}
\arguments{
\item{expr}{the line of code to be evaluated}
}
\value{
A list containing:
\itemize{
\item value the result of the evaluation of the expression
\item warnings warning(s) generated during the evaluation of the expression
\item error error generated during the evaluation of the expression
}
}
\description{
Similar to \code{try} but also returns warnings.
}
\details{
from https://stackoverflow.com/questions/4948361/how-do-i-save-warnings-and-errors-as-output-from-a-function
}
\examples{
FctTest <- function(x){
  return(log(x))
}
tryWithWarnings(FctTest(-1))
tryWithWarnings(FctTest(1))
tryWithWarnings(FctTest(xxxx))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createContrast.R
\name{contrast2name}
\alias{contrast2name}
\alias{.contrast2name}
\title{Create Rownames for a Contrast Matrix}
\usage{
.contrast2name(contrast, null = NULL, sep = c("[", "]"))
}
\arguments{
\item{contrast}{[matrix] a contrast matrix defining the left hand side of the linear hypotheses to be tested.}

\item{null}{[vector, optional] the right hand side of the linear hypotheses to be tested.}

\item{sep}{[character of length 2, optional] character used in rownames to wrap the left hand side of the equation.}
}
\value{
a character vector.
}
\description{
Create rownames for a contrast matrix using the coefficients and the names of the coefficients. The rownames will be [value * name] == null, e.g. [beta + 4*alpha] = 0.
}
\details{
When argument \code{NULL} is null then the rownames will not be put into brackets and the right hand side will not be added to the name.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{combineFormula}
\alias{combineFormula}
\title{Combine formula}
\usage{
combineFormula(ls.formula, as.formula = TRUE, as.unique = FALSE)
}
\arguments{
\item{ls.formula}{a list of formula}

\item{as.formula}{should a list of formula be returned. Otherwise it will be a list of characters.}

\item{as.unique}{should regressors appears at most once in the formula}
}
\description{
Combine formula by outcome
}
\examples{
combineFormula(list(Y~X1,Y~X3+X5,Y1~X2))
lava.options(symbols = c("~",","))
combineFormula(list("Y~X1","Y~X3+X5","Y1~X2"))
lava.options(symbols = c("<-","<->"))
combineFormula(list("Y<-X1","Y<-X3+X5","Y1<-X2"))

combineFormula(list(Y~X1,Y~X3+X1,Y1~X2))
combineFormula(list(Y~X1,Y~X3+X1,Y1~X2), as.formula = FALSE)
combineFormula(list(Y~X1,Y~X3+X1,Y1~X2), as.unique = TRUE)

lava.options(symbols = c("~","~~"))
combineFormula(list("Y~X1","Y~X3","Y1~X2"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-moments2.R
\name{vcov2-internal}
\alias{vcov2-internal}
\alias{.info2vcov}
\title{Inverse the Information Matrix}
\usage{
.info2vcov(information, attr.info = FALSE)
}
\arguments{
\item{information}{[matrix] information matrix to be inverted.}

\item{attr.info}{[logical] should the information matrix be returned as an attribute?}
}
\description{
Compute the inverse of the information matrix.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-nobs2.R
\name{nobs2}
\alias{nobs2}
\alias{nobs2.lvmfit}
\alias{nobs2.lvmfit2}
\title{Effective Sample Size.}
\usage{
nobs2(object, ssc, ...)

\method{nobs2}{lvmfit}(object, ssc = lava.options()$ssc, ...)

\method{nobs2}{lvmfit2}(object, ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}
}
\value{
Numeric vector of length the number of endogenous variables.
}
\description{
Extract the effective sample size, i.e. sample size minus the loss in degrees of freedom caused by the estimation of the parameters.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the leverage.
}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{convFormulaCharacter}
\alias{convFormulaCharacter}
\alias{formula2character}
\title{formula character conversion}
\usage{
formula2character(f, type = "formula")
}
\arguments{
\item{f}{a formula.}

\item{type}{should the normal formula operator be used (\code{"formula"}) or the one of lava.option (\code{"symbols"} or \code{"symbol"}).}
}
\description{
Conversion of formula into character string or vice versa
}
\examples{
formula2character(Y1~X1+X2)
formula2character(Y1~X1+X2, type = "symbols")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-methods.R
\name{getNewLink}
\alias{getNewLink}
\alias{getNewLink.modelsearch2}
\title{Extract the Links that Have Been Found by the modelsearch2.}
\usage{
getNewLink(object, step)

\method{getNewLink}{modelsearch2}(object, step = 1:nStep(object))
}
\arguments{
\item{object}{a \code{modelsearch2} object.}

\item{step}{[logical] which test should be extracted?}
}
\value{
A character vector.
}
\description{
Extract the links that have been found relevant by modelsearch2.
}
\examples{
\dontrun{
mSim <- lvm(Y~G+X1+X2)
addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6

set.seed(10)
df.data <- lava::sim(mSim, 1e2)

mBase <- lvm(Y~G)
addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
e.lvm <- estimate(mBase, data = df.data)
res <- modelsearch2(e.lvm, method.p.adjust = "holm")
getNewLink(res)
}
}
\concept{extractor}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transformSummaryTable.R
\name{transformSummaryTable}
\alias{transformSummaryTable}
\title{Apply Transformation to Summary Table}
\usage{
transformSummaryTable(object, transform = NULL)
}
\arguments{
\item{object}{A data.frame with columns estimate, se, lower, upper.}

\item{transform}{the name of a transformation or a function.}
}
\value{
a data.frame
}
\description{
Update summary table according to a transformation, e.g. log-transformtion.
P-values are left unchanged but estimates, standard errors, and confidence intervals are updated.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package-butils-valid.R
\name{validFCTs}
\alias{validFCTs}
\alias{validCharacter}
\alias{validClass}
\alias{validDimension}
\alias{validInteger}
\alias{validLogical}
\alias{validNames}
\alias{validNumeric}
\alias{validPath}
\title{Check Arguments of a function.}
\usage{
validCharacter(
  value1,
  name1 = as.character(substitute(value1)),
  valid.length,
  valid.values = "character",
  refuse.NULL = TRUE,
  refuse.duplicates = FALSE,
  method = NULL,
  addPP = TRUE
)

validClass(
  value1,
  name1 = as.character(substitute(value1)),
  validClass,
  super.classes = TRUE,
  method = NULL,
  addPP = TRUE
)

validDimension(
  value1,
  value2 = NULL,
  name1 = as.character(substitute(value1)),
  name2 = as.character(substitute(value2)),
  validDimension = NULL,
  type = c("NROW", "NCOL"),
  method = NULL,
  addPP = TRUE
)

validInteger(
  value1,
  name1 = as.character(substitute(value1)),
  valid.length,
  valid.values = NULL,
  min = NULL,
  max = NULL,
  refuse.NA = TRUE,
  refuse.NULL = TRUE,
  refuse.duplicates = FALSE,
  method = NULL,
  addPP = TRUE
)

validLogical(
  value1,
  name1 = as.character(substitute(value1)),
  valid.length,
  refuse.NULL = TRUE,
  refuse.NA = TRUE,
  method = NULL,
  addPP = TRUE
)

validNames(
  value1,
  name1 = as.character(substitute(value1)),
  refuse.NULL = TRUE,
  valid.length = NULL,
  valid.values = NULL,
  required.values = NULL,
  refuse.values = NULL,
  method = NULL,
  addPP = TRUE
)

validNumeric(
  value1,
  name1 = as.character(substitute(value1)),
  valid.length,
  valid.values = NULL,
  min = NULL,
  max = NULL,
  refuse.NA = TRUE,
  refuse.NULL = TRUE,
  refuse.duplicates = FALSE,
  method = NULL,
  addPP = TRUE
)

validPath(
  value1,
  name1 = as.character(substitute(value1)),
  type,
  method = NULL,
  addPP = TRUE,
  extension = NULL,
  check.fsep = FALSE
)
}
\arguments{
\item{value1}{the value of the (first) argument to be checked}

\item{name1}{the name of the (first) argument.}

\item{valid.length}{the acceptable length(s) for the argument. If \code{NULL} no test is performed.}

\item{valid.values}{the acceptable value(s) for the argument. If \code{NULL} no test is performed. Can also be "character" or "character_or_logical".}

\item{refuse.NULL}{should an error be output if value is \code{NULL}.}

\item{refuse.duplicates}{should an error be output if value contains duplicated values.}

\item{method}{the name of the function using the argument.}

\item{addPP}{add ": " after the name of the function in the error message.}

\item{validClass}{the acceptable classes(s) for the argument.}

\item{super.classes}{uses the \code{is} function instead of \code{class} to test the class of the object.}

\item{value2}{the second value of a second argument whose dimensions should be consistent with the first one}

\item{name2}{the name of the second argument.}

\item{validDimension}{the acceptable dimension for the argument. If \code{NULL} then name2 is used as a reference.}

\item{type}{For \code{validDimension}: the type of operator used to check the dimensions. For \code{validPath} either "dir" or "file" to check whether to path points to an existing directory or file.}

\item{min}{the minimum acceptable value}

\item{max}{the maximum acceptable value}

\item{refuse.NA}{should an error be output if value contains \code{NA}.}

\item{required.values}{values that must appear in the argument}

\item{refuse.values}{values that must not appear in the argument}

\item{extension}{filter the files by the type of extension.}

\item{check.fsep}{display a warning when the separator is not correctly specified in}
}
\value{
An invisible \code{TRUE} or an error message.
}
\description{
Check the validity of the arguments in functions.
}
\concept{check}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/symmetrize.R
\name{symmetrize}
\alias{symmetrize}
\title{Symmetrize a Matrix}
\usage{
symmetrize(M, update.upper = TRUE)
}
\arguments{
\item{M}{a matrix.}

\item{update.upper}{[logical] should the upper extra diagonal terms be updated using the lower extra diagonal terms?}
}
\description{
Complete the upper (or lower) extra-diagonal terms  in order to obtain a symmetric matrix.
}
\examples{
symmetrize <- lavaSearch2:::symmetrize

## example
M <- matrix(NA, 4, 4)
M[lower.tri(M)] <- 1:6

symmetrize(M, update.upper = TRUE) # good

M[upper.tri(M, diag = FALSE)] <- M[lower.tri(M, diag = FALSE)]
M # wrong
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils-nlme.R
\name{getCluster2-internal}
\alias{getCluster2-internal}
\alias{.getCluster2}
\alias{.getCluster2.gls}
\alias{.getCluster2.lme}
\title{Reconstruct the Cluster Variable from a nlme Model}
\usage{
.getCluster2(object, ...)

\method{.getCluster2}{gls}(object, cluster, data, ...)

\method{.getCluster2}{lme}(object, ...)
}
\arguments{
\item{object}{a \code{gls} or \code{lme} object.}

\item{...}{[internal] Only used by the generic method.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.
Only required for \code{gls} models with no correlation argument.}

\item{data}{[data.frame] the data set.}
}
\value{
A list containing:
\itemize{
\item cluster: the cluster index for each observation.
\item n.cluster: the number of clusters.
}
}
\description{
Reconstruct the cluster variable from a nlme model.
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-summary.glht2.R
\name{summary.glht2}
\alias{summary.glht2}
\title{Outcome of Linear Hypothesis Testing}
\usage{
\method{summary}{glht2}(
  object,
  confint = TRUE,
  conf.level = 0.95,
  transform = NULL,
  seed = NULL,
  rowname.rhs = TRUE,
  ...
)
}
\arguments{
\item{object}{a \code{glht2} object.}

\item{confint}{[logical] should confidence intervals be output}

\item{conf.level}{[numeric 0-1] level of the confidence intervals.}

\item{transform}{[function] function to backtransform the estimates, standard errors, null hypothesis, and the associated confidence intervals
(e.g. \code{exp} if the outcomes have been log-transformed).}

\item{seed}{[integer] value that will be set before adjustment for multiple comparisons to ensure reproducible results.
Can also be \code{NULL}: in such a case no seed is set.}

\item{rowname.rhs}{[logical] when naming the hypotheses, add the right-hand side (i.e. "X1-X2=0" instead of "X1-X2").}

\item{...}{argument passed to \code{multcomp:::summary.glht}, e.g. argument \code{test} to choose the type of adjustment for multiple comparisons.}
}
\description{
Estimates, p-values, and confidence intevals for linear hypothesis testing, possibly adjusted for multiple comparisons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-leverage2.R
\name{leverage2}
\alias{leverage2}
\alias{leverage2.lvmfit}
\alias{leverage2.lvmfit2}
\title{Leverage With Small Sample Correction.}
\usage{
leverage2(object, format, ssc, ...)

\method{leverage2}{lvmfit}(object, format = "wide", ssc = lava.options()$ssc, ...)

\method{leverage2}{lvmfit2}(object, format = "wide", ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{format}{[character] Use \code{"wide"} to return the residuals in the wide format (one row relative to each sample).
Otherwise use \code{"long"} to return the residuals in the long format.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}
}
\value{
a matrix containing the leverage relative to each sample (in rows)
and each endogenous variable (in column).
}
\description{
Extract leverage values from a latent variable model, with small sample correction.
}
\details{
The leverage are defined as the partial derivative of the fitted values with respect to the observations.
\deqn{
leverage_i = \frac{\partial \hat{Y}_i}{\partial Y_i}
}
See Wei et al. (1998). \cr \cr

When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the leverage.
}
\examples{
#### simulate data ####
set.seed(10)
m <- lvm(Y1~eta,Y2~eta,Y3~eta)
latent(m) <- ~eta
d <- lava::sim(m,20, latent = FALSE)

#### latent variable models ####
e.lvm <- estimate(m, data = d)
leverage2(e.lvm)

}
\references{
Bo-Cheng Wei et al., Generalized Leverage and its applications (1998), Scandinavian Journal of Statistics 25:1:25-37.
}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{estimator}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createGrid.R
\name{createGrid}
\alias{createGrid}
\title{Create a Mesh for the Integration}
\usage{
createGrid(n, xmin, xmax, d.y, d.z, zmax, fine, double)
}
\arguments{
\item{n}{[integer >0] the number of points for the mesh in the x direction.}

\item{xmin}{[numeric] the minimal x value.}

\item{xmax}{[numeric] the maximal x value.}

\item{d.y}{[integer >0] the number of dimensions for the triangle.}

\item{d.z}{[integer >0] the number of dimensions.}

\item{zmax}{[numeric >0] the maximal z value (in absolute value).}

\item{fine}{[logical] should the mesh be displayed?}

\item{double}{[logical] should the grid be just outside the region of interest?
Otherwise it will be just inside.}
}
\value{
A list containing:
\itemize{
\item grid: a \code{data.frame} where each line corresponds to a point.
\item seqNames.min: a \code{character} giving the name of the x,y variables (min).
\item seqNames.max: a \code{character} giving the name of the x,y variables (max).
}
}
\description{
Create a mesh for the integration.
}
\details{
This create a mesh for integrating over a triangular surface using rectangles.
The domain is define by constrains on three types of variables:
\itemize{
\item the x variable: [unidimensional] that varies freely between [-xmax,-xmin] U [xmin,xmax].
\item the y variables: [dimension d.y] constrained to be lower in absolute value than the x variable.
In 2D this corresponds to 2 triangles, and in higher dimension to cones/hypercones.
\item the z variables: [dimension d.z] constrained to vary between [-zmax,zmax].
}
The intersection of these three conditions define the domain.

The mesh is obtained slicing the triangles using rectangles.
}
\examples{
createGrid <- lavaSearch2:::createGrid

## no z 
gridInt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
                         d.z = 0, fine = FALSE, double = FALSE)
gridExt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
                         d.z = 0, fine = FALSE, double = TRUE)

gridInt_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
                         d.z = 0, fine = FALSE, double = FALSE)
gridExt_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
                         d.z = 0, fine = FALSE, double = TRUE)

gridInt_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
                         d.z = 0, fine = TRUE, double = FALSE)

##  z
gridIntZ1_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
                           d.z = 1, zmax = 2, fine = FALSE, double = FALSE)
gridExtZ1_2d <- createGrid(5, d.y = 1, xmin = 0, xmax = 4, 
                           d.z = 1, zmax = 2, fine = FALSE, double = TRUE)
 
gridIntZ2_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
                           d.z = 2, zmax = 2, fine = FALSE, double = FALSE)
gridExtZ2_4d <- createGrid(5, d.y = 3, xmin = 0, xmax = 4, 
                           d.z = 2, zmax = 2, fine = FALSE, double = TRUE)

}
\concept{post-selection inference}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-iid2.R
\name{iid2plot}
\alias{iid2plot}
\title{Display the i.i.d. Decomposition}
\usage{
iid2plot(object, param)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{param}{[character] name of one of the model parameters.}
}
\description{
Extract the i.i.d. decomposition and display it along with the corresponding coefficient.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils-nlme.R
\name{coef2-internal}
\alias{coef2-internal}
\alias{.coef2}
\alias{.coef2.lm}
\alias{.coef2.gls}
\alias{.coef2.lme}
\title{Export Mean and Variance Coefficients}
\usage{
.coef2(object)

\method{.coef2}{lm}(object)

\method{.coef2}{gls}(object)

\method{.coef2}{lme}(object)
}
\arguments{
\item{object}{a \code{lm}, \code{gls} or \code{lme} object.}

\item{name.Y}{[character] the name of the endogenous variable. Used to name certain variance parameters.}
}
\value{
A numeric vector named with the names of the coefficient with three attributes:
\itemize{
\item mean.coef: the name of the mean coefficients.
\item var.coef: the name of the variance coefficients.
\item cor.coef:  the name of the correlation coefficients.
}
}
\description{
Export mean and variance coefficients
from a \code{lm}, \code{gls}, or \code{lme} object.
}
\details{
The variance coefficients that are exported are the residual variance of each outcome. 
This is \eqn{\sigma^2} for the first one and \eqn{k^2 \sigma^2} for the remaining ones.
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{combination}
\alias{combination}
\alias{.combination}
\title{Form all Unique Combinations Between two Vectors}
\usage{
.combination(..., levels = FALSE)
}
\arguments{
\item{...}{[vectors] elements to be combined.}

\item{levels}{[logical] should a label for each combination be output as an attribute named levels.}
}
\value{
A matrix, each row being a different combination.
}
\description{
Form all unique combinations between two vectors (removing symmetric combinations).
}
\examples{
.combination <- lavaSearch2:::.combination

.combination(1,1)
.combination(1:2,1:2)
.combination(c(1:2,1:2),1:2)

.combination(alpha = 1:2, beta = 3:4)
.combination(alpha = 1:2, beta = 3:4, gamma = 1:4)
.combination(alpha = 1:3, beta = 1:3, gamma = 1:3)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-methods.R
\name{getNewModel}
\alias{getNewModel}
\alias{getNewModel.modelsearch2}
\title{Extract the Model that Has Been Retains by the modelsearch2.}
\usage{
getNewModel(object, step)

\method{getNewModel}{modelsearch2}(object, step = nStep(object))
}
\arguments{
\item{object}{a \code{modelsearch2} object.}

\item{step}{[integer >=0] the step at which the model should be extracted.
0 returns the initial model, i.e. before adding any links.}
}
\value{
A \code{lvmfit} object.
}
\description{
Extract the model that has been retained by modelsearch2.
}
\examples{
\dontrun{
mSim <- lvm(Y~G+X1+X2)
addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6

set.seed(10)
df.data <- lava::sim(mSim, 1e2)

mBase <- lvm(Y~G)
addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
e.lvm <- estimate(mBase, data = df.data)
res <- modelsearch2(e.lvm, method.p.adjust = "holm")
getNewModel(res)
}
}
\concept{extractor}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect.R
\name{sCorrect}
\alias{sCorrect}
\alias{sCorrect.default}
\alias{sCorrect<-}
\alias{sCorrect<-.default}
\title{Depreciated Method For Small Sample Correction}
\usage{
sCorrect(object, ...)

\method{sCorrect}{default}(object, ...)

sCorrect(x, ...) <- value

\method{sCorrect}{default}(x, ...) <- value
}
\arguments{
\item{object, x}{a \code{lvmfit} object.}

\item{...}{not used.}

\item{value}{not used.}
}
\description{
Depreciated method for small sample correction, now replaced by the \code{\link{estimate2}} method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-getIndexOmega.R
\name{getIndexOmega}
\alias{getIndexOmega}
\alias{.getIndexOmega}
\alias{.getIndexOmega.lvm}
\alias{.getIndexOmega.lvmfit}
\title{Identify the Endogenous Variables}
\usage{
.getIndexOmega(object, data, ...)

\method{.getIndexOmega}{lvm}(object, data, ...)

\method{.getIndexOmega}{lvmfit}(object, data, ...)
}
\arguments{
\item{object}{a \code{lvmfit} object.}

\item{data}{dataset.}

\item{...}{[internal] Only used by the generic method.}
}
\description{
Identify the endogenous variables, i.e., returns a vector with length the number of observations,
whose values are the index of the repetitions.
}
\examples{
\dontrun{
#### simulate data ####
set.seed(10)
dW <- sampleRepeated(10, format = "wide")
set.seed(10)
dL <- sampleRepeated(10, format = "long")
dL$time2 <- paste0("visit",dL$time)

#### lvm model ####
e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
## lavaSearch2:::.getIndexOmega(e.lvm, data = dW)
}
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Scorrect-confint2.R, R/Scorrect-model.tables.R
\name{confint2}
\alias{confint2}
\alias{confint2.lvmfit}
\alias{model.tables2}
\title{Confidence Intervals With Small Sample Correction}
\usage{
confint2(object, robust, cluster, transform, as.lava, conf.level, ...)

\method{confint2}{lvmfit}(
  object,
  robust = FALSE,
  cluster = NULL,
  transform = NULL,
  as.lava = TRUE,
  conf.level = 0.95,
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  ...
)

model.tables2(object, robust, cluster, transform, as.lava, conf.level, ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{robust}{[logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{transform}{[function] transformation to be applied.}

\item{as.lava}{[logical] when \code{TRUE} uses the same names as when using \code{stats::coef}.}

\item{conf.level}{[numeric, 0-1] level of the confidence intervals.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}
}
\value{
A data.frame with a row per coefficient.

A data.frame with a row per coefficient.
}
\description{
Extract confidence intervals of the coefficients from a latent variable model.
Similar to \code{lava::confint} but with small sample correction.

Extract estimate, standard error, confidence intervals and p-values associated to each coefficient of a latent variable model.
Similar to \code{lava::confint} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.

When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the confidence intervals.
}
\examples{
#### simulate data ####
set.seed(10)
dW <- sampleRepeated(10, format = "wide")
set.seed(10)
dL <- sampleRepeated(10, format = "long")
dL$time2 <- paste0("visit",dL$time)

#### latent variable models ####
e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
confint(e.lvm)
confint2(e.lvm)
confint2(e.lvm, as.lava = FALSE)
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/evalInParentEnv.R
\name{evalInParentEnv}
\alias{evalInParentEnv}
\title{Find Object in the Parent Environments}
\usage{
evalInParentEnv(name)
}
\arguments{
\item{name}{[character] the name of the object to get.}
}
\description{
Search an object in the parent environments. For internal use.
}
\concept{extractor}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-summary.R
\name{summary.modelsearch2}
\alias{summary.modelsearch2}
\title{summary Method for modelsearch2 Objects}
\usage{
\method{summary}{modelsearch2}(object, print = TRUE, ...)
}
\arguments{
\item{object}{output of the \code{modelsearch2} function.}

\item{print}{should the summary be printed in the terminal.}

\item{...}{[internal] only used by the generic method.}
}
\description{
summary method for modelsearch2 objects.
}
\details{
The column \code{dp.Info} contains the percentage of extended models (i.e. model with one additional link)
for which the information matrix evaluated at the value of the parameters of the initial model is non positive definie.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-methods.R
\name{autplot-modelsearch2}
\alias{autplot-modelsearch2}
\alias{autoplot.modelsearch2}
\title{Display the Value of a Coefficient across the Steps.}
\usage{
\method{autoplot}{modelsearch2}(
  object,
  param,
  ci = TRUE,
  step = 0:nStep(object),
  conf.level = 0.95,
  plot = TRUE,
  add.0 = TRUE,
  ...
)
}
\arguments{
\item{object}{a \code{modelsearch2} object.}

\item{param}{[character vector] the name of the coefficient(s) to be displayed.}

\item{ci}{[logical] should the confidence intervals of the coefficient(s) be displayed.}

\item{step}{[integer >0] the steps at which the coefficient value should be displayed.}

\item{conf.level}{[numeric, 0-1] confidence level of the interval.}

\item{plot}{[logical] should the graph be displayed?}

\item{add.0}{[logical] should an horizontal line representing no effect be displayed?}

\item{...}{[internal] only used by the generic method.}
}
\value{
A list containing \itemize{
\item \code{plot}: a ggplot object.
\item \code{data}: the data used to generate the ggplot object.
}
}
\description{
Display the value of a coefficient across the steps.
}
\examples{
\dontrun{
mSim <- lvm(Y~G+X1+X2+X3+X4+X5)
addvar(mSim) <- ~Z1+Z2

set.seed(10)
df.data <- lava::sim(mSim, 1e2)

mBase <- lvm(Y~G)
addvar(mBase) <- ~X1+X2+X3+X4+X5+Z1+Z2
e.lvm <- estimate(mBase, data = df.data)
res <- modelsearch2(e.lvm, method.p.adjust = "holm", alpha = 0.05)
autoplot(res, param = "Y~G")
autoplot(res, param = c("Y","Y~G"))
}
}
\concept{extractor}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-score2.R
\name{score2-internal}
\alias{score2-internal}
\alias{.score2}
\title{Compute the Corrected Score.}
\usage{
.score2(
  dmu,
  dOmega,
  epsilon,
  OmegaM1,
  missing.pattern,
  unique.pattern,
  name.pattern,
  name.param,
  name.meanparam,
  name.varparam,
  n.cluster,
  weights
)
}
\arguments{
\item{n.cluster}{[integer >0] the number of observations.}
}
\description{
Compute the corrected score.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-estimate2.R, R/sCorrect-sscResiduals.R
\name{estimate2}
\alias{estimate2}
\alias{estimate2.lvm}
\alias{estimate2.lvmfit}
\alias{estimate2.list}
\alias{estimate2.mmm}
\alias{.sscResiduals}
\title{Satterthwaite Correction and Small Sample Correction}
\usage{
estimate2(
  object,
  param,
  data,
  ssc,
  df,
  derivative,
  hessian,
  dVcov.robust,
  iter.max,
  tol.max,
  trace,
  ...
)

\method{estimate2}{lvm}(
  object,
  param = NULL,
  data = NULL,
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  derivative = "analytic",
  hessian = FALSE,
  dVcov.robust = FALSE,
  iter.max = 100,
  tol.max = 1e-06,
  trace = 0,
  ...
)

\method{estimate2}{lvmfit}(
  object,
  param = NULL,
  data = NULL,
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  derivative = "analytic",
  hessian = FALSE,
  dVcov.robust = FALSE,
  iter.max = 100,
  tol.max = 1e-06,
  trace = 0,
  ...
)

\method{estimate2}{list}(object, ...)

\method{estimate2}{mmm}(object, ...)

.sscResiduals(object, ssc, algorithm = "2")
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{param}{[numeric vector, optional] the values of the parameters at which to perform the correction.}

\item{data}{[data.frame, optional] the dataset relative to which the correction should be performed.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}

\item{derivative}{[character] should the first derivative of the information matrix be computed using a formula (\code{"analytic"}) or numerical derivative (\code{"numeric"})?}

\item{hessian}{[logical] should the hessian be stored? Can be \code{NULL} to indicate only if computed during the small sample correction.}

\item{dVcov.robust}{[logical] should the first derivative of robust variance-covariance matrix be stored?}

\item{iter.max}{[integer >0] the maximum number of iterations used to estimate the bias correction.}

\item{tol.max}{[numeric >0] the largest acceptable absolute difference between two succesive estimates of the bias correction.}

\item{trace}{[logical] should the execution of the function be traced.}

\item{...}{arguments passed to \code{lava::estimate} when using a \code{lvm} object.}
}
\description{
Correct the bias of the ML estimate of the variance and compute the first derivative of the information matrix.

Compute bias corrected residuals variance covariance matrix
and information matrix.
Also provides the leverage values and corrected sample size when adjust.n is set to TRUE.
}
\details{
The argument \code{value} is equivalent to the argument \code{bias.correct} of the function \code{summary2}.
}
\examples{
#### simulate data ####
set.seed(10)
dW <- sampleRepeated(10, format = "wide")

#### latent variable model ####
m.lvm <- lvm(Y1~X1+X2+Z1)

e2.lvm <- estimate2(m.lvm, data = dW)
summary2(e2.lvm)

}
\concept{estimator}
\keyword{internal}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/IntDensTri.R
\name{intDensTri}
\alias{intDensTri}
\title{Integrate a Gaussian/Student Density over a Triangle}
\usage{
intDensTri(
  mu,
  Sigma,
  df,
  n,
  x.min,
  z.max = NULL,
  type = "double",
  proba.min = 1e-06,
  prune = NULL,
  distribution = "pmvnorm"
)
}
\arguments{
\item{mu}{[numeric vector] the expectation of the joint distribution.}

\item{Sigma}{[matrix] the variance-covariance of the joint distribution.}

\item{df}{[integer > 0] the degree of freedom of the joint Student's t distribution.
Only used when \code{distribution="pvmt"}.}

\item{n}{[integer > 0] number of points for the numerical integration.}

\item{x.min}{[numeric] the minimum value along the x axis.}

\item{z.max}{[numeric vector, optional] the maximum value along the z axis.
Define the dimension of Z.}

\item{type}{[character] the type of mesh to be used.
Can be \code{\"raw\"}, \code{\"double\"}, or \code{\"fine\"}.}

\item{proba.min}{[numeric 0-1] the probability used to find the maximum value along the x axis.
Only used if \code{prune} is not specified.}

\item{prune}{[integer >0] number of standard deviations after which the domain ends along the x axis.}

\item{distribution}{[character] type of joint distribution.
Can be \code{"pmvnorm"} (normal distribution) or \code{"pvmt"} (Student's t distribution)}
}
\value{
A numeric.
}
\description{
Consider a univariate random variable X,
two multivariate random variables Y and Z,
and t1 and t2 two real numbers.
This function can compute either
P[|X|>t1,|X]>|Y1|,...,|X]>|Yp|] if zmin is not specified,
P[|Z1|<t2,...,|Zq|<t2,|X|>t1,|X]>|Y1|,...,|X]>|Yp|] if zmin is specified.
}
\details{
Argument \code{type}: \itemize{
\item \code{\"raw\"}: mesh with points inside the domain
\item \code{\"double\"}: mesh with points outside the domain
\item \code{\"fine\"}: mesh with points inside the domain plus additional rectangles trying to fill the missing domain.
}

Argument \code{Sigma} and \code{mu}:
define the mean and variance-covariance of the random variables X, Y, Z
(in this order). The length of the argument \code{z.max} is used to define the dimension of Z.
The dimension of X is always 1.
}
\examples{
library(mvtnorm)

p <- 2
Sigma <- diag(p)
mu <- rep(0, p)

## bivariate normal distribution
z2 <- qmvt(0.975, mean = mu, sigma = Sigma, df = 1e3)$quantile

# compute integral
intDensTri(mu = mu, Sigma = Sigma, n=5, x.min=0, type = "fine")$value-1/2
intDensTri(mu = mu, Sigma = Sigma, n=30, x.min=0, type = "raw")$value-1/2
intDensTri(mu = mu, Sigma = Sigma, n=50, x.min=0, type = "raw")$value-1/2

intDensTri(mu = mu, Sigma = Sigma, df = 5, n=5, x.min=0, distribution = "pmvt")$value-1/2
res <- intDensTri(mu = mu, Sigma = Sigma, df = 5, n=10, x.min=0, distribution = "pmvt")
res$value-1/2
ggplot2::autoplot(res)

## trivariate normal distribution
\dontrun{
p <- 3
Sigma <- diag(p)
mu <- rep(0, p)

res2 <- intDensTri(mu = mu, Sigma = Sigma, n=5, x.min = 0, z.max = 10)
ggplot2::autoplot(res2)
ggplot2::autoplot(res2, coord.plot = c("x","z1"))
res2
}

#### when the distribution is far from 0
\dontrun{
eq1 <- intDensTri(mu = c(10,0), Sigma = diag(1,2), 
                  x.min = 2, n=10)
eq1$value-1
ggplot2::autoplot(eq1)

eq2 <- intDensTri(mu = c(10,0,0), Sigma = diag(1,3),
                  x.min=2, z.max = 10, type = "raw",
                  n=10)
ggplot2::autoplot(eq2, coord.plot = c("y1","z1"))
eq2$value-1

## more variables
p <- 5
Sigma <- diag(p)
mu <- rep(0, p)

res2 <- intDensTri(mu = mu, Sigma = Sigma, n=5, x.min = 1, z.max = c(2,2))
res2$grid
}

}
\concept{post-selection inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multcomp.R
\name{estfun}
\alias{estfun}
\alias{estfun.lvmfit}
\alias{estfun.gls}
\alias{estfun.lme}
\title{Extract Empirical Estimating Functions (lvmfit Object)}
\usage{
\method{estfun}{lvmfit}(x, ...)

\method{estfun}{gls}(x, ...)

\method{estfun}{lme}(x, ...)
}
\arguments{
\item{x}{an \code{lvmfit} object.}

\item{...}{arguments passed to methods.}
}
\description{
Extract the empirical estimating functions of a lvmfit object.
This function is for internal use but need to be public to enable its use by \code{multcomp::glht}.
}
\details{
This function enables to use the \code{glht} function with lvmfit object.
Otherwise when calling \code{multcomp:::vcov.mmm} then \code{sandwich::sandwich} and then \code{sandwich::meat}, \code{sandwich::meat} will complain that \code{estfun} is not defined for \code{lvmfit} objects.
}
\examples{
library(multcomp)

#### generative model ####
mSim <- lvm(X ~ Age + 0.5*Treatment,
            Y ~ Gender + 0.25*Treatment,
            c(Z1,Z2,Z3) ~ eta, eta ~ 0.75*treatment,
            Age[40:5]~1)
latent(mSim) <- ~eta
categorical(mSim, labels = c("placebo","SSRI")) <- ~Treatment
categorical(mSim, labels = c("male","female")) <- ~Gender

#### simulate data ####
n <- 5e1
set.seed(10)
df.data <- lava::sim(mSim, n = n, latent = FALSE)

#### fit separate models ####
lmX <- lm(X ~ Age + Treatment, data = df.data)
lvmY <- estimate(lvm(Y ~ Gender + Treatment), data = df.data)
lvmZ <- estimate(lvm(c(Z1,Z2,Z3) ~ eta, eta ~ Treatment), 
                 data = df.data)

#### create mmm object #### 
e.mmm <- mmm(X = lmX, Y = lvmY, Z = lvmZ)

#### create contrast matrix ####
resC <- createContrast(e.mmm, var.test = "Treatment", add.variance = FALSE)

#### adjust for multiple comparisons ####
e.glht <- glht(e.mmm, linfct = resC$mlf)
summary(e.glht)
}
\concept{multiple comparison}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampleRepeated.R
\name{sampleRepeated}
\alias{sampleRepeated}
\title{Simulate Repeated Measurements over time}
\usage{
sampleRepeated(n, n.Xcont = 2, n.Xcat = 2, n.rep = 5, format = "long")
}
\arguments{
\item{n}{[integer] sample size.}

\item{n.Xcont}{[integer] number of continuous covariates acting on the latent variable.}

\item{n.Xcat}{[integer] number of categorical covariates acting on the latent variable.}

\item{n.rep}{[integer] number of measurement of the response variable.}

\item{format}{[character] should the dataset be returned in the \code{"long"} format or in the \code{"wide"} format.}
}
\value{
a \code{data.frame} object.
}
\description{
Simulate repeated measurements over time (one factor model).
}
\examples{

sampleRepeated(10, format = "wide")
sampleRepeated(10, format = "long")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelsearch2-methods.R
\name{getStep}
\alias{getStep}
\alias{getStep.modelsearch2}
\title{Extract one Step From the Sequential Procedure}
\usage{
getStep(object, step, slot)

\method{getStep}{modelsearch2}(object, step = nStep(object), slot = NULL)
}
\arguments{
\item{object}{a \code{modelsearch2} object}

\item{step}{[integer >0] which test should be extracted?}

\item{slot}{[character] the element from the modelsearch2 object that should be extracted.}
}
\description{
Extract one step from the sequential procedure.
}
\examples{
\dontrun{
mSim <- lvm(Y~G+X1+X2)
addvar(mSim) <- ~Z1+Z2+Z3+Z4+Z5+Z6
df.data <- lava::sim(mSim, 1e2)

mBase <- lvm(Y~G)
addvar(mBase) <- ~X1+X2+Z1+Z2+Z3+Z4+Z5+Z6
e.lvm <- estimate(mBase, data = df.data)
res <- modelsearch2(e.lvm, method.p.adjust = "holm")

getStep(res)
getStep(res, slot = "sequenceTest")
getStep(res, step = 1)
}
}
\concept{extractor}
\concept{modelsearch}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-moments2.R
\name{moments2}
\alias{moments2}
\alias{moments2.lvm}
\alias{moments2.lvmfit}
\title{Compute Key Quantities of a Latent Variable Model}
\usage{
moments2(
  object,
  param,
  data,
  weights,
  Omega,
  Psi,
  initialize,
  usefit,
  update.dmoment,
  update.d2moment,
  score,
  information,
  hessian,
  vcov,
  dVcov,
  dVcov.robust,
  residuals,
  leverage,
  derivative
)

\method{moments2}{lvm}(
  object,
  param = NULL,
  data = NULL,
  weights = NULL,
  Omega = NULL,
  Psi = NULL,
  initialize = TRUE,
  usefit = TRUE,
  update.dmoment = TRUE,
  update.d2moment = TRUE,
  score = TRUE,
  information = TRUE,
  hessian = TRUE,
  vcov = TRUE,
  dVcov = TRUE,
  dVcov.robust = TRUE,
  residuals = TRUE,
  leverage = TRUE,
  derivative = "analytic"
)

\method{moments2}{lvmfit}(
  object,
  param = NULL,
  data = NULL,
  weights = NULL,
  Omega = NULL,
  Psi = NULL,
  initialize = TRUE,
  usefit = TRUE,
  update.dmoment = TRUE,
  update.d2moment = TRUE,
  score = TRUE,
  information = TRUE,
  hessian = TRUE,
  vcov = TRUE,
  dVcov = TRUE,
  dVcov.robust = TRUE,
  residuals = TRUE,
  leverage = TRUE,
  derivative = "analytic"
)
}
\arguments{
\item{object}{a latent variable model.}

\item{param}{[numeric vector] value of the model parameters if different from the estimated ones.}

\item{data}{[data.frame] dataset if different from the one used to fit the model.}

\item{Psi}{[matrix]  Average first order bias in the residual variance. Only necessary for computing adjusted residuals.}

\item{initialize}{[logical] Pre-compute quantities dependent on the data but not on the parameters values.}

\item{usefit}{[logical] Compute key quantities based on the parameter values.}

\item{update.dmoment}{[logical] should the first derivative of the moments be computed/updated?}

\item{update.d2moment}{[logical] should the second derivative of the the moments be computed/updated?}

\item{score}{[logical] should the score be output?}

\item{information}{[logical] should the expected information be output?}

\item{hessian}{[logical] should the hessian be output?}

\item{vcov}{[logical] should the variance-covariance matrix based on the expected information be output?}

\item{dVcov}{[logical] should the derivative of the variance-covariance matrix be output?}

\item{dVcov.robust}{[logical]  should the derivative of the robust variance-covariance matrix be output?}
}
\description{
Compute conditional mean, conditional variance, their first and second derivative regarding model parameters, as well as various derivatives of the log-likelihood.
}
\details{
For lvmfit objects, there are two levels of pre-computation:
\itemize{
\item a basic one that do no involve the model coefficient (\code{conditionalMoment.lvm}).
\item an advanced one that require the model coefficients (\code{conditionalMoment.lvmfit}). 
}
}
\examples{
m <- lvm(Y1~eta,Y2~eta,Y3~eta)
latent(m) <- ~eta

d <- lava::sim(m,1e2)
e <- estimate(m, d)

## basic pre-computation
res1 <- moments2(e, data = d, initialize = TRUE, usefit = FALSE,
                score = TRUE, information = TRUE, hessian = TRUE, vcov = TRUE,
                dVcov = TRUE, dVcov.robust = TRUE, residuals = TRUE, leverage = FALSE,
                derivative = "analytic")
res1$skeleton$param$Sigma

## full pre-computation
res2 <- moments2(e, param = coef(e), data = d, initialize = TRUE, usefit = TRUE,
                score = TRUE, information = TRUE, hessian = TRUE, vcov = TRUE,
                dVcov = TRUE, dVcov.robust = TRUE, residuals = TRUE, leverage = FALSE,
                derivative = "analytic")
res2$moment$Omega

}
\concept{derivative of the score equation}
\concept{small sample inference}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-summary2.R
\name{summary2}
\alias{summary2}
\alias{summary2.lvmfit}
\alias{summary2.lvmfit2}
\alias{summary.lvmfit2}
\title{Latent Variable Model Summary After Small Sample Correction}
\usage{
summary2(object, robust, cluster, digit, ...)

\method{summary2}{lvmfit}(
  object,
  robust = FALSE,
  cluster = NULL,
  digit = max(5, getOption("digit")),
  ssc = lava.options()$ssc,
  df = lava.options()$df,
  ...
)

\method{summary2}{lvmfit2}(
  object,
  robust = FALSE,
  cluster = NULL,
  digit = max(5, getOption("digit")),
  ...
)

\method{summary}{lvmfit2}(
  object,
  robust = FALSE,
  cluster = NULL,
  digit = max(5, getOption("digit")),
  ...
)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{robust}{[logical] should robust standard errors be used instead of the model based standard errors? Should be \code{TRUE} if argument cluster is not \code{NULL}.}

\item{cluster}{[integer vector] the grouping variable relative to which the observations are iid.}

\item{digit}{[integer > 0] the number of decimal places to use when displaying the summary.}

\item{...}{[logical] arguments passed to lower level methods.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}

\item{df}{[character] method used to estimate the degree of freedoms of the Wald statistic: Satterthwaite \code{"satterthwaite"}. 
Otherwise (\code{"none"}/\code{FALSE}/\code{NA}) the degree of freedoms are set to \code{Inf}.
Only relevant when using a \code{lvmfit} object.}
}
\description{
Summarize a fitted latent variable model.
Similar to \code{stats::summary} with small sample correction.
}
\details{
\code{summary2} is the same as \code{summary}
except that it first computes the small sample correction (but does not store it).
So if \code{summary2} is to be called several times,
it is more efficient to pre-compute the quantities for the small sample correction
using \code{sCorrect} and then call \code{summary2}.

\code{summary2} returns an object with an element \code{table2} containing the estimates, standard errors, degrees of freedom,
upper and lower limits of the confidence intervals, test statistics, and p-values.
}
\examples{
#### simulate data ####
m <- lvm(Y~X1+X2)
set.seed(10)
d <- lava::sim(m, 2e1)

#### latent variable models ####
e.lvm <- estimate(m, data = d)
summary(e.lvm)$coef

summary2(e.lvm)
summary2(e.lvm, ssc = "none")

}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{small sample inference}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sCorrect-coef2.R
\name{coef2}
\alias{coef2}
\alias{coef2.lvmfit}
\title{Model Coefficients With Small Sample Correction}
\usage{
coef2(object, as.lava, ...)

\method{coef2}{lvmfit}(object, as.lava = TRUE, ssc = lava.options()$ssc, ...)
}
\arguments{
\item{object}{a \code{lvmfit} or \code{lvmfit2} object (i.e. output of \code{lava::estimate} or \code{lavaSearch2::estimate2}).}

\item{as.lava}{[logical] if \code{TRUE}, uses the same names as when using \code{stats::coef}.}

\item{...}{additional argument passed to \code{estimate2} when using a \code{lvmfit} object.}

\item{ssc}{[character] method used to correct the small sample bias of the variance coefficients: no correction (\code{"none"}/\code{FALSE}/\code{NA}),
correct the first order bias in the residual variance (\code{"residual"}), or correct the first order bias in the estimated coefficients \code{"cox"}).
Only relevant when using a \code{lvmfit} object.}
}
\value{
A numeric vector named with the names of the coefficients.
}
\description{
Extract the coefficients from a latent variable model.
Similar to \code{lava::compare} but with small sample correction.
}
\details{
When argument object is a \code{lvmfit} object, the method first calls \code{estimate2} and then extract the model coefficients.
}
\examples{
#### simulate data ####
set.seed(10)
dW <- sampleRepeated(10, format = "wide")
set.seed(10)
dL <- sampleRepeated(10, format = "long")
dL$time2 <- paste0("visit",dL$time)

#### latent variable models ####
e.lvm <- estimate(lvm(c(Y1,Y2,Y3) ~ 1*eta + X1, eta ~ Z1), data = dW)
coef(e.lvm)
coef2(e.lvm)
coef2(e.lvm, as.lava = FALSE)
}
\seealso{
\code{\link{estimate2}} to obtain \code{lvmfit2} objects.
}
\concept{extractor}
\keyword{smallSampleCorrection}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coefType.R
\name{coefType}
\alias{coefType}
\alias{coefType.lvm}
\alias{coefType.lvmfit}
\alias{coefType.multigroup}
\title{Extract the Type of Each Coefficient}
\usage{
coefType(object, as.lava, ...)

\method{coefType}{lvm}(object, as.lava = TRUE, data = NULL, ...)

\method{coefType}{lvmfit}(object, as.lava = TRUE, ...)

\method{coefType}{multigroup}(object, as.lava = TRUE, ...)
}
\arguments{
\item{object}{a \code{lvm} or \code{lvmfit} object.}

\item{as.lava}{[logical] export the type of coefficients mimicking \code{lava:::coef}.}

\item{...}{arguments to be passed to \code{lava::coef}}

\item{data}{[data.frame, optional] the dataset. Help to identify the categorical variables.}
}
\value{
\code{coefType} returns a \code{data.frame} when \code{as.lava=FALSE}:
\itemize{
\item name: name of the link
\item Y: outcome variable
\item X: regression variable in the design matrix (could be a transformation of the original variables, e.g. dichotomization).
\item data: original variable
\item type: type of link
\item value: if TRUE, the value of the link is set and not estimated.
\item marginal: if TRUE, the value of the link does not impact the estimation.
\item detail: a more detailed description of the type of link (see the details section)
\item lava: name of the coefficient in lava
}
When \code{as.lava=TRUE}, \code{coefType} returns a named vector containing the type of each coefficient.
}
\description{
Extract the type of each coefficient of a \code{lvm} object.
}
\details{
A lvm can be written as a measurement model:
\deqn{Y_i = \nu + \Lambda \eta_i + K X_i + \epsilon_i}
and a structural model:
\deqn{\eta_i = \alpha + B \eta_i + \Gamma X_i + \zeta_i}
where \eqn{\Psi}   is the variance covariance matrix of the residuals \eqn{\zeta} \cr
and   \eqn{\Sigma} is the variance covariance matrix of the residuals \eqn{\epsilon}. \cr \cr

\code{coefType} either returns the Latin/Greek letter corresponding to the coefficients
or it groups them:
\itemize{
\item intercept: \eqn{\nu} and \eqn{\alpha}.
\item regression: \eqn{\Lambda}, \eqn{K}, \eqn{B}, and \eqn{\Gamma}.
\item covariance: extra-diagonal terms of \eqn{\Sigma} and \eqn{\Psi}.
\item variance: diagonal of \eqn{\Sigma} and \eqn{\Psi}.
}

A link denotes a relationship between two variables.
The coefficient are used to represent the strength of the association between two variable, i.e. the strength of a link.
A coefficient may corresponds to the strength of one or several link.
}
\examples{
#### regression ####
m <- lvm(Y~X1+X2)
e <- estimate(m, lava::sim(m, 1e2))

coefType(m)
coefType(e)

#### LVM ####
m <- lvm()
regression(m) <- c(y1,y2,y3)~u
regression(m) <- u~x1+x2
latent(m) <- ~u
covariance(m) <- y1~y2

m.Sim <- m
categorical(m.Sim, labels = c("a","b","c")) <- ~x2
e <- estimate(m, lava::sim(m.Sim, 1e2))

coefType(m)
coefType(e)

## additional categorical variables 
categorical(m, labels = as.character(1:3)) <- "X1"

coefType(m, as.lava = FALSE)

#### LVM with constrains ####
m <- lvm(c(Y1~0+1*eta1,Y2~0+1*eta1,Y3~0+1*eta1,
          Z1~0+1*eta2,Z2~0+1*eta2,Z3~0+1*eta2))
latent(m) <- ~eta1 + eta2
e <- estimate(m, lava::sim(m,1e2))

coefType(m)
coefType(e)

#### multigroup ####
m <- lvm(Y~X1+X2)
eG <- estimate(list(m,m), list(lava::sim(m, 1e2), lava::sim(m, 1e2)))
coefType(eG)

}
\concept{extractor}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Utils.R
\name{selectRegressor}
\alias{selectRegressor}
\alias{selectRegressor.formula}
\title{Regressor of a Formula.}
\usage{
selectRegressor(object, ...)

\method{selectRegressor}{formula}(object, format = "call", ...)
}
\arguments{
\item{object}{a formula}

\item{...}{[internal] Only used by the generic method.}

\item{format}{[character] should an object of format call be returned (\code{format = "call"}),
or the names of the variables (\code{format = "vars"})}
}
\description{
Return the regressor variables contained in the formula
}
\examples{

\dontrun{

selectRegressor <- lavaSearch2:::selectRegressor
selectRegressor.formula <- lavaSearch2:::selectRegressor.formula

selectRegressor(Y1~X1+X2)
selectRegressor(Y1~X1+X2, format = "vars")

selectRegressor(Y1~X1+Y1)
selectRegressor(Y1+Y2~X1+Y1, format = "vars")

selectRegressor(~X1+X2)
selectRegressor(~X1+X2, format = "vars")

}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.R
\name{clean}
\alias{clean}
\alias{clean.lvm}
\title{Simplify a lvm object}
\usage{
clean(x, ...)

\method{clean}{lvm}(x, rm.exo = TRUE, rm.endo = TRUE, rm.latent = TRUE, ...)
}
\arguments{
\item{x}{\code{lvm}-object}

\item{...}{additional arguments to lower level functions}

\item{rm.exo}{should exogenous variables with no links be removed from the object?}

\item{rm.endo}{should endogenous variables with no links be removed from the object?}

\item{rm.latent}{should latent variables with no links be removed from the object?}
}
\description{
Remove variables with no link.
}
\examples{

m <- lvm()
m <- regression(m, x=paste0("x",1:5),y="y1")
m <- regression(m, x=paste0("x",1:5),y="y2")
covariance(m) <- y1~y2

cancel(m) <- y1 ~ x1
cancel(m) <- y2 ~ x1
clean(m)

m <- lvm(y1 ~ eta + x1, y2 ~ eta, y3 ~ eta + x2)
latent(m) <- ~eta
clean(m)
m
cancel(m) <- y1 ~ eta
cancel(m) <- y2 ~ eta
cancel(m) <- y3 ~ eta
clean(m)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link.R
\name{findNewLink}
\alias{findNewLink}
\alias{findNewLink.lvm}
\title{Find all New Links Between Variables}
\usage{
findNewLink(object, ...)

\method{findNewLink}{lvm}(
  object,
  data = NULL,
  type = "both",
  exclude.var = NULL,
  rm.latent_latent = FALSE,
  rm.endo_endo = FALSE,
  rm.latent_endo = FALSE,
  output = "names",
  ...
)
}
\arguments{
\item{object}{a \code{lvm} object.}

\item{...}{[internal] only used by the generic method.}

\item{data}{[optional] a dataset used to identify the categorical variables when not specified in the \code{lvm} object.}

\item{type}{[character vector] the type of links to be considered: \code{"regression"}, \code{"covariance"}, or \code{"both"}, .}

\item{exclude.var}{[character vector] all links related to these variables will be ignore.}

\item{rm.latent_latent}{[logical] should the links relating two latent variables be ignored?}

\item{rm.endo_endo}{[logical] should the links relating two endogenous variables be ignored?}

\item{rm.latent_endo}{[logical] should the links relating one endogenous variable and one latent variable be ignored?}

\item{output}{[character] Specify \code{"names"} to return the names of the variables to link
or specify \code{"index"} to return their position.}
}
\value{
A list containing:
\itemize{
\item M.links: a matrix with two columns indicating (by name or position) the exogenous and endogenous variable corresponding to each link.
\item links: the name of the additional possible links
\item directional: a logical vector indicating for each link whether the link is unidirectional (\code{TRUE}, i.e. regression link)
or bidirectional (\code{FALSE}, i.e. covariance link).
}
}
\description{
Find all new links between variables (adapted from lava::modelsearch).
}
\examples{
library(lava)

m <- lvm()
regression(m) <- c(y1,y2,y3)~u
categorical(m,labels=c("M","F","MF")) <- ~X1
findNewLink(m, rm.endo = FALSE)
findNewLink(m, rm.endo = TRUE)
findNewLink(m, exclude.var = "X1")

regression(m) <- u~x1+x2
latent(m) <- ~u

findNewLink(m, rm.endo = FALSE)
findNewLink(m, rm.endo = TRUE)
findNewLink(m, rm.endo = TRUE, output = "index")
findNewLink(m, type = "covariance")
findNewLink(m, type = "regression")

}
\concept{modelsearch}
